!------------------------------------------------------------------------------
! THIS FILE CONTAINS 5 MODULES
! 1 - RADCONS
! 2 - RADCAL_VAR
! 3 - RADACL
! 4 - SPECDATA
! 5 - MIEV
!------------------------------------------------------------------------------


MODULE RADCONS
!------------------------------------------------------------------------------
! RADIATION PARAMETERS

USE PRECISION_PARAMETERS
IMPLICIT NONE

REAL(EB), ALLOCATABLE, DIMENSION(:,:) :: DLN,ORIENTATION_FACTOR
REAL(EB), ALLOCATABLE, DIMENSION(:)   :: BBFRAC, WL_LOW, WL_HIGH
REAL(EB), ALLOCATABLE, DIMENSION(:)   :: DLX, DLY, DLZ, DLB, RSA

INTEGER, ALLOCATABLE, DIMENSION(:,:) :: DLM
INTEGER, ALLOCATABLE, DIMENSION(:)   :: NRP

REAL(EB) :: RADTMP, PATH_LENGTH
REAL(EB) :: DGROUP_A, DGROUP_B, WEIGH_CYL
REAL(EB) :: DPHI0, FOUR_SIGMA, RPI_SIGMA, LTSTEP, RTMPMAX, RTMPMIN
REAL(EB) :: MIE_MINIMUM_DIAMETER,MIE_MAXIMUM_DIAMETER

INTEGER :: TIME_STEP_INCREMENT,NMIEANG
INTEGER :: NRDMIE, NLMBDMIE, MIE_NDG
INTEGER :: NRT,NCO,UIIDIM,NLAMBDAT,NKAPPAT,NKAPPAZ

LOGICAL :: WIDE_BAND_MODEL,WSGG_MODEL
REAL(EB) :: SOOT_DENSITY

INTEGER :: N_RADCAL_ARRAY_SIZE,RADCAL_SPECIES_INDEX(16),N_KAPPA_T=44,N_KAPPA_Y=50
REAL(EB), ALLOCATABLE, DIMENSION(:,:) :: Z2RADCAL_SPECIES
REAL(EB), ALLOCATABLE, DIMENSION(:,:,:,:) :: RADCAL_SPECIES2KAPPA
CHARACTER(LABEL_LENGTH) :: RADCAL_SPECIES_ID(16)='NULL'


!------------------------------------------------------------------------------
!
!     BBFRAC    FRACTION OF BLACKBODY RADIATION
!     DLX       MEAN X-COMPONENT OF THE CONTROL ANGLE RAY VECTOR
!     DLY       MEAN Y-COMPONENT OF THE CONTROL ANGLE RAY VECTOR
!     DLZ       MEAN Z-COMPONENT OF THE CONTROL ANGLE RAY VECTOR
!     DLB       MEAN BOTTOM-COMPONENT OF RAYN VECTOR (CYLINDRICAL CASE)
!     DLM       MIRRORING INDEXES
!     DLN       WALL NORMAL MATRIX
!     DPHI0     OPENING ANGLE OF THE CYLINDRICAL DOMAIN
!     E_WALL    WALL EMISSIVITY
!     ILW       RADIATION INTENSITIES ON SOLID MIRRORS AND MESH INTERFACES.
!               INTENSITY INTEGRALS (BAND SPECIFIC OR ANGLE INCREMENT) FOR SOLID AND OPEN WALLS
!     INRAD_W   INCIDENT RADIATIVE HEAT FLUX ON A CELL (QRADIN = E_WALL*INRAD_W)
!     R50       ARRAY OF PARTICLE RADII CORRESPONDING TO THE MEDIAN DIAMETERS
!               OF THE DISTRIBUTIONS USED IN THE GENERATION OF WQABS AND WQSCA ARRAYS.
!     MIE_NDG   NUMBER OF PARTICLE RADII IN WQABS AND WQSCA ARRAYS
!     NLMBDMIE  NUMBER OF WAVE LENGTHS IN MIE CALCULATIONS
!     NMIEANG   NUMBER OF ANGLE BINS IN FORWARD SCATTERING INTEGRATION
!     NUMBER_RADIATION_ANGLES
!     NRA       TOTAL NUMBER OF RADIATION CONTROL ANGLES
!     NRDMIE    NUMBER OF PARTICLE RADII IN MIE CALCULATIONS
!     NRT       NUMBER OF RADIATION THETA ANGLES
!     NRP       NUMBER OF RADIATION PHI ANGLES ON EACH THETA BAND
!     NUMBER_SPECTRA_BANDS
!     NSB       NUMBER OF SPECTRAL BANDS (1=GRAY, 6=WIDE BAND, 9=WIDE BAND W. CH4)
!     OUTRAD_W  EMITTED INTENSITY FROM A WALL (OUTRAD_W = QRADOUT/PI)
!     ORIENTATION_FACTOR
!               0.5 IF THE PARTICLE ORIENTATION MOST CLOSELY ALIGNS WITH THE SOLID ANGLE; 0 OTHERWISE
!     PHIUP     UPPER LIMIT OF SOLID ANGLE COMPONENT PHI
!     PHILOW    LOWER LIMIT OF SOLID ANGLE COMPONENT PHI
!     RADTMP    RADIATION TEMPERATURE FOR ABSORPTION PROPERTIES (MIE)
!     QRADIN    ABSORBED RADIATIVE HEAT FLUX INTO A SURFACE CELL (SOLID WALL OR OPEN)
!     QRADOUT   EMITTED RADIATIVE HEAT FLUX FROM A SURFACE (SOLID WALL OR OPEN)
!     RSA       ARRAY OF SOLID ANGLES
!     RTMPMAX   MAXIMUM TEMPERATURE FOR TABULATION OF RADIATIVE PROPERTIES
!     RTMPMIN   MINIMUM TEMPERATURE FOR TABULATION OF RADIATIVE PROPERTIES
!     THETAUP   UPPER LIMIT OF SOLID ANGLE COMPONENT THETA
!     THETALOW  LOWER LIMIT OF SOLID ANGLE COMPONENT THETA
!     UII       INTEGRATED INTENSITY
!     UIID      PARTS OF UII  IF WIDE_BAND_MODEL = TRUE, UIID CONTAINS THE BAND SPECIFIC INTENSITY
!                             IF WIDE_BAND_MODEL/= TRUE, UIID CONTAINS THE ANGLE_INCREMENTS OF INTENSITY
!     WEIGH_CYL IN CYLINDRICAL COORDINATES, ALL INTENSITIES REPRESENT TWO ACTUAL CONTROL ANGLES
!     WL_LOW    LOWER WAVELENGTH LIMIT OF THE SPECTRAL BANDS
!     WL_HIGH   UPPER WAVELENGTH LIMIT OF THE SPECTRAL BANDS
!     WQABS     ABSORPTION EFFICIENCY FACTOR ARRAY
!     WQSCA     SCATTERING EFFICIENCY FACTOR ARRAY
!
!     PATH_LENGTH             MEAN PATH LENGTH FOR THE GRAY GAS ABS. COEF.
!     ANGLE_INCREMENT         HOW MANY ANGLES ARE SKIPPED ON EACH UPDATE
!     TIME_STEP_INCREMENT     HOW OFTEN IS THE RADIATION SOLVER CALLED
!
!------------------------------------------------------------------------------

END MODULE RADCONS


MODULE RADCAL_VAR

!------------------------------------------------------------------------------
! VARIABLES
! OMMIN : (REAL) MINIMUM WAVE NUMBER IN SPECTRUM, CM-1
! OMMAX : (REAL) MAXIMUM WAVE NUMBER IN SPECTRUM, CM-1
! NOM   : (INTEGER) NUMBER OF WAVELENGTH INTERVALS
! GC    : (REAL) COLLISION BROADENED HALF-WIDTH
! LAMBDA: (REAL) WAVELENGTH, IN MICROMETER
!------------------------------------------------------------------------------
! MODULE WRAPPER FOR RADCAL SUBROUTINE

USE PRECISION_PARAMETERS
USE GLOBAL_CONSTANTS, ONLY: AEROSOL_AL2O3
IMPLICIT NONE

CHARACTER(255) :: MESSAGE
INTEGER, PARAMETER :: N_RADCAL_SPECIES = 16      ! NUMBER OF RADCAL SPECIES (INCLUDING SOOT) PLUS NITROGEN AND OXYGEN
INTEGER, PARAMETER :: N_RADCAL_SPECIES_GAS=N_RADCAL_SPECIES-1 ! NUMBER OF RADCAL SPECIES (INCLUDING SOOT) PLUS NITROGEN AND OXYGEN

TYPE ELEMENTS
   CHARACTER(32)           :: RADCAL_ID
   CHARACTER(32)           :: ID
   CHARACTER(2048)         :: COMMENTS
   REAL(EB), POINTER, DIMENSION(:) :: BANDS
   REAL(EB), POINTER, DIMENSION(:) :: TEMP_EXP
END TYPE ELEMENTS

TYPE(ELEMENTS), DIMENSION(N_RADCAL_SPECIES) :: RADCAL_SPECIES

REAL(EB), ALLOCATABLE, DIMENSION(:,:) :: PARTIAL_PRESSURES_ATM
REAL(EB), ALLOCATABLE, DIMENSION(:)   :: TEMP_GAS
REAL(EB), ALLOCATABLE, DIMENSION(:)   :: SEGMENT_LENGTH_M
REAL(EB), ALLOCATABLE, DIMENSION(:)   :: TOTAL_PRESSURE_ATM

CHARACTER(255) :: CHID_RADCAL  ! RADCAL CASE ID
CHARACTER(255) :: TITLE_RADCAL ! RADCAL CASE TITLE

REAL(EB) :: OMMIN, OMMAX, LAMBDAMIN, LAMBDAMAX
REAL(EB) :: TWALL, TAU, XTOT, FV

INTEGER:: NPT

REAL(EB), ALLOCATABLE, DIMENSION(:,:) :: GAMMA, SD15, SD, SD7, SD3

INTEGER, PARAMETER :: N_TEMP_C3H6=7,  N_BAND_C3H6=3, &
                      N_TEMP_C3H8=7,  N_BAND_C3H8=2, &
                      N_TEMP_C7H16=7, N_BAND_C7H16=2, &
                      N_TEMP_C7H8=7,  N_BAND_C7H8=5, &
                      N_TEMP_CH4=23,  N_BAND_CH4=3, &
                      N_TEMP_CH3OH=7, N_BAND_CH3OH=4, &
                      N_TEMP_C5H8O2=7,N_BAND_C5H8O2=6, &
                      N_TEMP_C2H6=7,  N_BAND_C2H6=3, &
                      N_TEMP_C2H4=7,  N_BAND_C2H4=4

! 1: GOODY, 2: MALKMUS, 3: ELSASSER
INTEGER, PARAMETER :: I_MODEL_C3H6=2,I_MODEL_C3H8=2,I_MODEL_C7H16=2,I_MODEL_C7H8=2,I_MODEL_CH3OH=2,I_MODEL_C5H8O2=2, &
                      I_MODEL_C2H6=2,I_MODEL_C2H4=2,I_MODEL_CO2=1,I_MODEL_H2O=1,I_MODEL_CO=1,I_MODEL_CH4=3

REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_C3H6_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_C3H8_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_C7H16_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_C7H8_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_CH3OH_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_C5H8O2_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_C2H6_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_CH4_TEMP
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: SD_C2H4_TEMP

REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_C3H6
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_C3H8
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_C7H16
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_C2H6
REAL(EB), ALLOCATABLE, DIMENSION(:)    :: BE_C2H4

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_CH4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_C3H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_C3H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_C7H16
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_C2H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: OM_BND_C2H4

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_CH4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_CH4

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_C3H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_C3H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD3_C3H6

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_C3H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_C3H8

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_C7H16
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_C7H16

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD3_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD4_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD5_C7H8

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD3_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD4_CH3OH

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD3_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD4_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD5_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD6_C5H8O2

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_C2H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_C2H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD3_C2H6

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD1_C2H4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD2_C2H4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD3_C2H4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: SD4_C2H4

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_C3H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_C3H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD3_C3H6

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_C3H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_C3H8

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_C7H16
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_C7H16

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD3_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD4_C7H8
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD5_C7H8

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD3_CH3OH
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD4_CH3OH

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD3_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD4_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD5_C5H8O2
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD6_C5H8O2

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_C2H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_C2H6
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD3_C2H6

REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD1_C2H4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD2_C2H4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD3_C2H4
REAL(EB), ALLOCATABLE, DIMENSION(:,:)  :: GAMMAD4_C2H4

REAL(EB), ALLOCATABLE, DIMENSION(:)  :: AB, WAVE_NUMBER, LAMBDA
REAL(EB), ALLOCATABLE, DIMENSION(:)  :: INCIDENT_RADIANCE

REAL(EB), ALLOCATABLE, DIMENSION(:,:) :: TTAU

INTEGER :: NOM

REAL(EB), PARAMETER :: M_TO_CM = 100.0_EB  ! CONVERTION FACTOR METERS -> CENTIMETERS
REAL(EB), PARAMETER :: CM_TO_M = 0.01_EB   ! CONVERTION FACTOR CENTIMETERS -> METERS

CHARACTER(30) :: RADCAL_ID
REAL(EB) :: PFUEL

INTEGER, PARAMETER :: I_CO=3, I_CO2=1, I_H2O=2, I_N2=14, I_O2=15, I_FV=16, I_C2H4=5, I_C2H6=6, I_C3H6=7, I_C3H8=8, &
                      I_C7H8=9, I_C7H16=10, I_CH3OH=11, I_CH4=4, I_CH4_OLD=13, I_MMA=12

END MODULE RADCAL_VAR


MODULE RADCAL_CALC

USE PRECISION_PARAMETERS
USE GLOBAL_CONSTANTS, ONLY: SIGMA
USE COMP_FUNCTIONS, ONLY: SHUTDOWN
USE RADCONS, ONLY: RPI_SIGMA
USE RADCAL_VAR

PUBLIC INIT_RADCAL,RCALLOC,RCDEALLOC,PLANCK,SUB_RADCAL

CONTAINS

!==============================================================================
SUBROUTINE INIT_RADCAL
!==============================================================================
IMPLICIT NONE
!------------------------------------------------------------------------------
! LOCAL VARIABLES

REAL(EB) :: DOM, OMEGA

INTEGER :: I_WAVENUMB

IF((OMMAX<1100._EB).AND.(OMMIN<OMMAX)) THEN
   NOM = INT((OMMAX-OMMIN)/5._EB)
ELSEIF(OMMIN>5000._EB) THEN
   NOM = INT((OMMAX-OMMIN)/50._EB)
ELSEIF(OMMIN<1100._EB.AND.OMMAX>5000._EB) THEN
   NOM = INT((1100._EB-OMMIN)/5._EB)+INT((5000._EB-1100._EB)/25._EB) +INT((OMMAX-5000._EB)/50._EB)
ELSEIF(OMMIN<1100._EB) THEN
   NOM = INT((1100._EB-OMMIN)/5._EB)+INT((OMMAX-1100._EB)/25._EB)
ELSEIF(OMMAX>5000._EB) THEN
   NOM = INT((5000._EB-OMMIN)/25._EB)+INT((OMMAX-5000._EB)/50._EB)
ELSE
   NOM = INT((OMMAX-OMMIN)/25._EB)
ENDIF

ALLOCATE(INCIDENT_RADIANCE(NOM)); INCIDENT_RADIANCE   = 0.0_EB
ALLOCATE(TTAU(0:NPT,NOM))       ; TTAU                = 1.0_EB

ALLOCATE(LAMBDA(NOM))      ; LAMBDA      = 0.0_EB
ALLOCATE(AB(NOM))          ; AB          = 0.0_EB
ALLOCATE(WAVE_NUMBER(NOM)) ; WAVE_NUMBER = 0.0_EB

! POPULATE VECTOR WAVE_NUMBER AND LAMBDA

DOM    = 5.0_EB
OMEGA  = OMMIN-DOM

DO I_WAVENUMB = 1, NOM
   OMEGA=OMEGA+DOM
   IF(OMEGA > 1100._EB) OMEGA = OMEGA + 20._EB
   IF(OMEGA > 5000._EB) OMEGA = OMEGA + 25._EB

   WAVE_NUMBER(I_WAVENUMB) = OMEGA            ! WAVENUMBER IN CM^-1
   LAMBDA(I_WAVENUMB)      = 10000.0_EB/OMEGA ! WAVELENGTH IN MICRON
ENDDO

!------------------------------------------------------------------------------
END SUBROUTINE INIT_RADCAL


!==============================================================================
SUBROUTINE SUB_RADCAL(EFFECTIVE_ABSORPTION,PLANCK_MEAN_ABSORPTION,RADIANCE,    &
                     TOTAL_TRANSMISSIVITY)
!==============================================================================
!
! INPUT
!------
! IO   : (INTEGER) OUPUT FILE UNIT

! OUTPUT
!-------
! EFFECTIVE_ABSORPTION   : (REAL) EFFECTIVE ABSORPTION COEFFICIENT,   UNITS: 1/CM
! PLANCK_MEAN_ABSORPTION : (REAL) PLANCK MEAN ABSORPTION COEFFICIENT, UNITS: 1/CM
! RADIANCE               : (REAL) TOTAL INCIDENT POWER PER UNIT OF AREA PER
!                                 UNIT OF SOLID ANGLE, UNITS: W/M2/STR
! TOTAL_TRANSMISSIVITY   : (REAL) TOTAL TRANSMISSIVITY, COMPUTED ONLY WHEN TWALL
!                                 (BLACKBODY SOURCE) IS STRICTLY GREATER THAN 0
!                                 DIMENSIONLESS NUMBER
!
! LOCAL
!------
! A_COLLISION : (REAL) COLLISION BROADENED FINE STRUCTURE PARAMETER
!               (LINE_HALF_WIDTH/LINE_SPACING)
! A_DOPPLER   : (REAL) DOPPLER BROADENED FINE STRUCTURE PARAMETER
!               (LINE_HALF_WIDTH/LINE_SPACING)
! AZOTEMP     : (REAL) RATIO REFERENCE TEMPERATURE (273K) OVER LOCAL TEMPERATURE
! OMEGA       : (REAL) WAVENUMBER
! OPTICAL_THICKNESS: (REAL) OPTICAL THICKNESS OF THE ITH SPECIES,  UNITS: CMSTP
! PTOT        : (REAL) TOTAL PRESSURE IN ATM
! TEMP4       : (REAL) LOCAL TEMPERATURE RAISED TO THE POWER 4
!
! X_COLLISION : (REAL) ! OPTICAL DEPTH FOR A PURE COLLISION CURVE OF GROWTH
! X_DOPPLER   : (REAL) ! OPTICAL DEPTH FOR A PURE DOPPLER CURVE OF GROWTH
! X_PARTICLE  : (REAL) ! OPTICAL DEPTH FOR PARTICLE
!------------------------------------------------------------------------------

IMPLICIT NONE

! VARIABLES PASSED OUTPUT
REAL(EB), INTENT(OUT) :: EFFECTIVE_ABSORPTION, PLANCK_MEAN_ABSORPTION,       &
                        RADIANCE, TOTAL_TRANSMISSIVITY

! LOCAL VARIABLES

REAL(EB), ALLOCATABLE, DIMENSION(:)   :: PATH_LENGTH_CM, AZOTEMP

REAL(EB), ALLOCATABLE, DIMENSION(:)   :: TAUS, TAUL, X_PARTICLE, AB_PLANCK,  &
                                       AB_TAU, BB_SPECT, BB_GAS, BB_WALL

REAL(EB), ALLOCATABLE, DIMENSION(:,:) :: CURTIS_XSTAR, CURTIS_ACOLLISION,    &
                                       CURTIS_ADOPPLER, GC,                &
                                       OPTICAL_THICKNESS, OPTICAL_DEPTH

REAL(EB) :: RSL, RSS, OMEGA, WAVE_LENGTH,  DAMBDA, LTERM, TEMP4, &
            AVG_TEMP, TOTAL_LENGTH_CM, INT_BB_GAS, INT_BB_WALL

INTEGER :: I_WAVENUMB, KMAX, KMIN, I_PATH

! INITIALIZATION

INT_BB_GAS            = 0.0_EB
INT_BB_WALL           = 0.0_EB
INCIDENT_RADIANCE     = 0.0_EB
RADIANCE              = 0.0_EB
TOTAL_TRANSMISSIVITY  = 1.0_EB  ! INITIALIZATION VALUES TO ARBITRARY
                                 ! LARGE NUMBER SO THAT USER KNOWS WHEN IT
                                 ! HAS NOT BEEN CALCULATED
!------------------------------------------------------------------------------
! ALLOCATE ARRAYS NEEDED FOR CALCULATION OF RADIATION PROPERTIES AT A GIVEN
! WAVENUMBER: GC, SPECIES_PRESSURE_ATM, OPTICAL_THICKNESS

IF(ALLOCATED(CURTIS_XSTAR))         DEALLOCATE(CURTIS_XSTAR)
ALLOCATE(CURTIS_XSTAR(1:N_RADCAL_SPECIES,0:NPT));      CURTIS_XSTAR      = 0.0_EB

IF(ALLOCATED(CURTIS_ACOLLISION))    DEALLOCATE(CURTIS_ACOLLISION)
ALLOCATE(CURTIS_ACOLLISION(1:N_RADCAL_SPECIES,0:NPT)); CURTIS_ACOLLISION = 0.0_EB

IF(ALLOCATED(CURTIS_ADOPPLER))      DEALLOCATE(CURTIS_ADOPPLER)
ALLOCATE(CURTIS_ADOPPLER(1:N_RADCAL_SPECIES,0:NPT));   CURTIS_ADOPPLER   = 0.0_EB

IF(ALLOCATED(OPTICAL_THICKNESS))    DEALLOCATE(OPTICAL_THICKNESS)
ALLOCATE(OPTICAL_THICKNESS(1:N_RADCAL_SPECIES,1:NPT)); OPTICAL_THICKNESS = 0.0_EB

IF(ALLOCATED(OPTICAL_DEPTH))         DEALLOCATE(OPTICAL_DEPTH)
ALLOCATE(OPTICAL_DEPTH(1:N_RADCAL_SPECIES,0:NPT));      OPTICAL_DEPTH    = 0.0_EB

IF(ALLOCATED(GC))                   DEALLOCATE(GC)
ALLOCATE(GC(1:N_RADCAL_SPECIES,1:NPT));                 GC               = 0.0_EB

IF(ALLOCATED(PATH_LENGTH_CM))       DEALLOCATE(PATH_LENGTH_CM)
ALLOCATE(PATH_LENGTH_CM(1:NPT))

! TO COMPUTE THE EFFECTIVE ABSORPTION COEFFICIENT
IF (ALLOCATED(BB_WALL)) DEALLOCATE(BB_WALL)
ALLOCATE(BB_WALL(NOM)); BB_WALL = 0.0_EB

IF (ALLOCATED(BB_GAS)) DEALLOCATE(BB_GAS)
ALLOCATE(BB_GAS(NOM));  BB_GAS  = 0.0_EB

! CONVERT PATH LENGTH FROM METERS TO CM
PATH_LENGTH_CM = M_TO_CM*SEGMENT_LENGTH_M

! COMPUTE RATIO STANDARD TEMPERATURE OVER CONSIDERED TEMPERATURE
IF(ALLOCATED(AZOTEMP))              DEALLOCATE(AZOTEMP)
ALLOCATE(AZOTEMP(1:NPT))
AZOTEMP = 273._EB/TEMP_GAS

!------------------------------------------------------------------------------
! COMPUTE THE SEGMENTS PROPERTIES THAT ARE INDEPENDENT OF THE FREQUENCY
! STORE THEM IN ARRAYS WITH DIMENSIONS (NPT,N_RADCAL_SPECIES)

INIT_PATH: DO I_PATH = 1, NPT  ! LOOP OVER ELEMENTS OF THE PATH LENGTH.

!------------------------------------------------------------------------------
! TEST. IF PATH_LENGTH_CM SET TO ZERO (BAD SEGMENT), CYCLE
! ELSE, PROCEED WITH CALCULATION

   IF (PATH_LENGTH_CM(I_PATH)<=TWO_EPSILON_EB) CYCLE

!------------------------------------------------------------------------------
! COMPUTE SPECIES OPTICAL THICKNESS AND COLLISION BROADENING

   OPTICAL_THICKNESS(:,I_PATH) = AZOTEMP(I_PATH)*PATH_LENGTH_CM(I_PATH)*               &
                                                PARTIAL_PRESSURES_ATM(:,I_PATH)

   GC(:,I_PATH)                = COLLISION_BROADENING( PARTIAL_PRESSURES_ATM(:,I_PATH), &
                                                      TOTAL_PRESSURE_ATM(I_PATH),      &
                                                      AZOTEMP(I_PATH)                )

ENDDO INIT_PATH

! COMPUTE TOTAL LENGTH OF THE LINE OF SIGHT
TOTAL_LENGTH_CM = SUM(PATH_LENGTH_CM)

! COMPUTE LENGTH WEIGHTED AVERAGE TEMP4 AND AVG_TEMP ALONG LINE OF SIGHT

TEMP4    = SUM(TEMP_GAS**4*PATH_LENGTH_CM)/TOTAL_LENGTH_CM
AVG_TEMP = SUM(TEMP_GAS*PATH_LENGTH_CM)/TOTAL_LENGTH_CM

!------------------------------------------------------------------------------
! LOOP SPECTRUM COMPUTES EACH SPECTRAL CONTRIBUTION,
! LOOP OVER EACH WAVE NUMBER INTERVALS
! THIS LOOP CAN BE PARALLIZED WITH OPENMP

!!                   BELOW ARE OPENMP INSTRUCTIONS

!$OMP PARALLEL PRIVATE (I_WAVENUMB, I_PATH) FIRSTPRIVATE(CURTIS_XSTAR,CURTIS_ACOLLISION,CURTIS_ADOPPLER,OPTICAL_DEPTH,XTOT)
!$OMP DO
LSPECTRUM: DO I_WAVENUMB = 1, NOM
   CURTIS_XSTAR      = 0.0_EB
   CURTIS_ADOPPLER   = 0.0_EB
   OPTICAL_DEPTH     = 0.0_EB
   CURTIS_ACOLLISION = 0.0_EB
!------------------------------------------------------------------------------
! LOOP OVER THE DIFFERENT SEGMENTS OF THE LINE OF SIGHT

   PATH: DO I_PATH = 1,NPT  ! LOOP OVER ELEMENTS OF THE PATH LENGTH.
! EVALUATE THE COMBINED SPECTRAL TRANSMISSIVITY AND RADIANCE
! ARGUMENTS OUT: X (VECTOR)
      CALL SPECIES_OPTICAL_DEPTH(PATH_LENGTH_CM(I_PATH), PARTIAL_PRESSURES_ATM(:,I_PATH),   &
                                 TOTAL_PRESSURE_ATM(I_PATH), GC(:,I_PATH), TEMP_GAS(I_PATH),&
                                 WAVE_NUMBER(I_WAVENUMB), OPTICAL_THICKNESS(:,I_PATH),      &
                                 CURTIS_XSTAR(:,I_PATH-1:I_PATH),                           &
                                 CURTIS_ACOLLISION(:,I_PATH-1:I_PATH),                      &
                                 CURTIS_ADOPPLER(:,I_PATH-1:I_PATH),                        &
                                 OPTICAL_DEPTH(:,I_PATH-1:I_PATH))
      XTOT = SUM(OPTICAL_DEPTH(:,I_PATH))
      TTAU(I_PATH,I_WAVENUMB) = EXP(-XTOT)
   ENDDO PATH

   ! AB : WEAKLINE LIMITS OPTICAL DEPTH PER UNIT LENGTH, UNITS IN CM-1
   IF (TOTAL_LENGTH_CM.GT.0.0_EB) THEN
      AB(I_WAVENUMB) = SUM(CURTIS_XSTAR(:,NPT))/TOTAL_LENGTH_CM
   ELSE
      AB(I_WAVENUMB) = 0.0_EB
   ENDIF

! INTEGRATE THE INCIDENT RADIANCE OVER THE WHOLE LINE OF SIGHT
   INCIDENT_RADIANCE(I_WAVENUMB) = 0.0_EB

   DO I_PATH = 1, NPT
      ! RECALL: TTAU(0,I_WAVENUMB) = 1.0_EB

      INCIDENT_RADIANCE(I_WAVENUMB) = INCIDENT_RADIANCE(I_WAVENUMB) +                       &
                                    (TTAU(I_PATH-1,I_WAVENUMB)-TTAU(I_PATH,I_WAVENUMB))*  &
                                    PLANCK_WN(TEMP_GAS(I_PATH),WAVE_NUMBER(I_WAVENUMB))
   ENDDO

   INCIDENT_RADIANCE(I_WAVENUMB) = INCIDENT_RADIANCE(I_WAVENUMB) +  &
                                 TTAU(NPT,I_WAVENUMB)*PLANCK_WN(TWALL,WAVE_NUMBER(I_WAVENUMB))

   BB_GAS(I_WAVENUMB)  = PLANCK_WN(AVG_TEMP,WAVE_NUMBER(I_WAVENUMB))
   BB_WALL(I_WAVENUMB) = PLANCK_WN(TWALL,   WAVE_NUMBER(I_WAVENUMB))
ENDDO LSPECTRUM
!$OMP END DO
!$OMP END PARALLEL

!------------------------------------------------------------------------------
! INTEGRATE THE RADIANCE OVER THE SPECTRUM

RADIANCE    = INTEGRATION(WAVE_NUMBER,INCIDENT_RADIANCE)
INT_BB_GAS  = INTEGRATION(WAVE_NUMBER,BB_GAS)
INT_BB_WALL = INTEGRATION(WAVE_NUMBER,BB_WALL)

!------------------------------------------------------------------------------
! SOOT CONTRIBUTION
! DETERMINE SOOT RADIANCE FOR SHORT (BETWEEN OMMAX AND 25000) AND
! LONG WAVELENGTHS (BETWEEN 5 AND OMMIN)
! THIS CORRESPONDS TO THE RANGES BETWEEN 400 NM (VISIBLE) TO OMMAX (USER DEPENDENT)
! AND OMMIN (USER DEPENDENT) AND 2000 MICRONS

RSL  = 0._EB
RSS  = 0._EB

KMAX = INT(OMMIN)
KMIN = INT(OMMAX)

!------------------------------------------------------------------------------
LONG: DO I_WAVENUMB = 5, KMAX, 5 ! LOOP OVER LONG WAVELENGTHS

   OMEGA       = REAL(I_WAVENUMB,EB)
   WAVE_LENGTH = 10000._EB/OMEGA
   DAMBDA      = 10000._EB/(OMEGA-2.5_EB)-10000._EB/(OMEGA+2.5_EB)

! INITIALIZE TAUL: TRANSMISSIVITY FOR LONG WAVELENGTHS

   IF (ALLOCATED(TAUL)) DEALLOCATE(TAUL)
   ALLOCATE(TAUL(0:NPT));     TAUL       = 1.0_EB

   IF (ALLOCATED(X_PARTICLE)) DEALLOCATE(X_PARTICLE)
   ALLOCATE(X_PARTICLE(NPT)); X_PARTICLE = 0.0_EB

   ! LOOP OVER THE LINE OF SIGHT ELEMENTS
   DO I_PATH = 1, NPT

      CALL POD(OMEGA,PARTIAL_PRESSURES_ATM(I_FV,I_PATH),PATH_LENGTH_CM(I_PATH), X_PARTICLE(I_PATH),TEMP_GAS(I_PATH))

      TAUL(I_PATH) = EXP(-SUM(X_PARTICLE(1:I_PATH)))
      RSL = RSL + (TAUL(I_PATH-1)-TAUL(I_PATH))*PLANCK(TEMP_GAS(I_PATH),WAVE_LENGTH)*DAMBDA

   ENDDO

! ADD CONTRIBUTION OF THE INCIDENT RADIANCE FROM THE WALL
   RSL = RSL + TAUL(NPT)*PLANCK(TWALL,WAVE_LENGTH)*DAMBDA

! CALCULATE CONTRIBUTION TO INT_BB_WALL & INT_BB_GAS
   INT_BB_WALL = INT_BB_WALL + PLANCK(TWALL,    WAVE_LENGTH)*DAMBDA
   INT_BB_GAS  = INT_BB_GAS  + PLANCK(AVG_TEMP, WAVE_LENGTH)*DAMBDA

ENDDO LONG

!------------------------------------------------------------------------------
SHORT: DO I_WAVENUMB = KMIN, 25000, 100 ! LOOP OVER SHORT WAVELENGTHS

   OMEGA       = REAL(I_WAVENUMB,EB)
   WAVE_LENGTH = 10000._EB/OMEGA
   DAMBDA      = 10000._EB/(OMEGA-50._EB)-10000._EB/(OMEGA+50._EB)

! INITIALIZE TAUS: TRANSMISSIVITY FOR SHORT WAVELENGTHS

   IF (ALLOCATED(TAUS)) DEALLOCATE(TAUS)
   ALLOCATE(TAUS(0:NPT));     TAUS       = 1.0_EB

   IF (ALLOCATED(X_PARTICLE)) DEALLOCATE(X_PARTICLE)
   ALLOCATE(X_PARTICLE(NPT)); X_PARTICLE = 0.0_EB

   ! LOOP OVER THE LINE OF SIGHT ELEMENTS

   DO I_PATH = 1, NPT

      CALL POD(OMEGA,PARTIAL_PRESSURES_ATM(I_FV,I_PATH),PATH_LENGTH_CM(I_PATH), X_PARTICLE(I_PATH),TEMP_GAS(I_PATH))

      TAUS(I_PATH) = EXP(-SUM(X_PARTICLE(1:I_PATH)))
      RSS = RSS + (TAUS(I_PATH-1)-TAUS(I_PATH))*PLANCK(TEMP_GAS(I_PATH),WAVE_LENGTH)*DAMBDA

   ENDDO

! ADD CONTRIBUTION OF THE INCIDENT RADIANCE FROM THE WALL
   RSS = RSS + TAUS(NPT)*PLANCK(TWALL,WAVE_LENGTH)*DAMBDA

! CALCULATE CONTRIBUTION TO INT_BB_WALL & INT_BB_GAS
   INT_BB_WALL = INT_BB_WALL + PLANCK(TWALL,    WAVE_LENGTH)*DAMBDA
   INT_BB_GAS  = INT_BB_GAS  + PLANCK(AVG_TEMP, WAVE_LENGTH)*DAMBDA

ENDDO SHORT

! END LOOP OVER SPECIES
!------------------------------------------------------------------------------

! ADD THE CONTRIBUTION OF SOOT OVER LARGER WAVE LENGTH DOMAIN
RADIANCE = RADIANCE + RSS + RSL

!------------------------------------------------------------------------------
! DEALLOCATE SOME VARIABLES (GC,X,OPTICAL_THICKNESS)

IF(ALLOCATED(BB_GAS))            DEALLOCATE(BB_GAS)
IF(ALLOCATED(BB_WALL))           DEALLOCATE(BB_WALL)
IF(ALLOCATED(GC))                DEALLOCATE(GC)
IF(ALLOCATED(OPTICAL_DEPTH))     DEALLOCATE(OPTICAL_DEPTH)
IF(ALLOCATED(OPTICAL_THICKNESS)) DEALLOCATE(OPTICAL_THICKNESS)
IF(ALLOCATED(CURTIS_XSTAR))      DEALLOCATE(CURTIS_XSTAR)
IF(ALLOCATED(CURTIS_ACOLLISION)) DEALLOCATE(CURTIS_ACOLLISION)
IF(ALLOCATED(CURTIS_ADOPPLER))   DEALLOCATE(CURTIS_ADOPPLER)
IF(ALLOCATED(X_PARTICLE))        DEALLOCATE(X_PARTICLE)
IF(ALLOCATED(TAUS))              DEALLOCATE(TAUS)
IF(ALLOCATED(TAUL))              DEALLOCATE(TAUL)
!------------------------------------------------------------------------------

IF (TOTAL_LENGTH_CM > 0.0_EB) THEN

   IF(ABS(TWALL-AVG_TEMP)<=SPACING(TWALL)) THEN
      LTERM = 1.0_EB
   ELSE
      LTERM = MIN(1.0_EB,MAX(1E-30_EB,(RADIANCE-INT_BB_GAS)/(INT_BB_WALL-INT_BB_GAS)))
   ENDIF
   !IF (LTERM.GT.1.0_EB) THEN
   !   WRITE(*,*) "LTERM: ", LTERM
   !   WRITE(*,*) "RADIANCE/RPI_SIGMA ", RADIANCE/RPI_SIGMA
   !   WRITE(*,*) "TWALL**4 ", TWALL**4
   !   WRITE(*,*) "TEMP**4  ", TEMP4
   !ENDIF

   EFFECTIVE_ABSORPTION = -1._EB/TOTAL_LENGTH_CM*LOG(LTERM)
ELSE
   EFFECTIVE_ABSORPTION = 0.0_EB
ENDIF

!------------------------------------------------------------------------------
! COMPUTES THE MEAN ABSORPTION COEFFICIENTS OVER THE WHOLE PATHLENGTH
! INTEGRATE PLANCK_MEAN_ABSORPTION*PLANCK OVER LAMBDA

IF (ALLOCATED(AB_PLANCK)) DEALLOCATE(AB_PLANCK)
ALLOCATE(AB_PLANCK(NOM))

DO I_WAVENUMB = 1, NOM
   AB_PLANCK(I_WAVENUMB) = AB(I_WAVENUMB)*PLANCK(AVG_TEMP,LAMBDA(I_WAVENUMB))
ENDDO

PLANCK_MEAN_ABSORPTION = INTEGRATION(LAMBDA,AB_PLANCK)

! DIVIDE BY SIGMA*PI*TEMP**4 TO GET FINAL COEFFICIENT PLANCK_MEAN_ABSORPTION

IF (TEMP4 > TWO_EPSILON_EB) THEN
   PLANCK_MEAN_ABSORPTION = PLANCK_MEAN_ABSORPTION/(RPI_SIGMA*TEMP4)
ELSE
   PLANCK_MEAN_ABSORPTION = 0.0_EB
ENDIF

!------------------------------------------------------------------------------
! COMPUTES THE TOTAL TRANSMISSIVITY OVER THE WHOLE PATHLENGTH. IT IS CALCULATED
! ONLY IF TWALL IS STRICTLY GREATER THAN 0
! INTEGRATE TAU*PLANCK OVER LAMBDA

IF (TWALL**4 > TWO_EPSILON_EB) THEN

   IF (ALLOCATED(AB_TAU)) DEALLOCATE(AB_TAU)
   ALLOCATE(AB_TAU(NOM))

   IF (ALLOCATED(BB_SPECT)) DEALLOCATE(BB_SPECT)
   ALLOCATE(BB_SPECT(NOM))


   DO I_WAVENUMB = 1, NOM
      BB_SPECT(I_WAVENUMB) = PLANCK_WN(TWALL,WAVE_NUMBER(I_WAVENUMB))
      AB_TAU(I_WAVENUMB)   = TTAU(NPT,I_WAVENUMB)*BB_SPECT(I_WAVENUMB)
   ENDDO

   TOTAL_TRANSMISSIVITY = INTEGRATION(WAVE_NUMBER,AB_TAU)/INTEGRATION(WAVE_NUMBER,BB_SPECT)

ENDIF

!------------------------------------------------------------------------------
! DEALLOCATE AB_PLANCK & AB_TAU
IF (ALLOCATED(AB_PLANCK)) DEALLOCATE(AB_PLANCK)
IF (ALLOCATED(AB_TAU))    DEALLOCATE(AB_TAU)
IF (ALLOCATED(BB_SPECT))  DEALLOCATE(BB_SPECT)

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE SUB_RADCAL


!==============================================================================
FUNCTION COLLISION_BROADENING(SPECIES_PRESSURE_ATM,PTOT,AZOTEMP) RESULT(GC)
!==============================================================================
!! THIS FUNCTION COMPUTES THE COLLISION BROADENING HALF-WIDTH AT HALF-HEIGHT.
!! (CALCULATION PROCEEDS IN ACCORDANCE WITH THE SLG MODEL, TABLE 5-18, IN NASA SP-3080._EB)
!! BASED ON EQ 5-34 FROM NASA REPORT SP-3080
!! RETURNS VECTOR GC THAT CONTAINS THE COLLISION BROADENING HALF-WIDTH AT
!! HALF-HEIGHT INCLUDING FOREIGN AND SELF-BROADENING COLLISION COEFFICIENTS.
!! GC UNITS IN CM-1
! CHECKED ON 02/11/13
!------------------------------------------------------------------------------

! ARGUMENTS IN
REAL(EB), DIMENSION(:), INTENT(IN) :: SPECIES_PRESSURE_ATM ! SPECIES PARTIAL PRESSURE, UNITS: ATM
REAL(EB), INTENT(IN) :: PTOT         ! TOTAL PRESSURE, UNITS: ATM
REAL(EB), INTENT(IN) :: AZOTEMP      ! RECALL AZOTEMP = 273/T, UNITS: NONE

! ARGUMENT OUT
REAL(EB), ALLOCATABLE, DIMENSION(:) :: GC ! UNITS IN CM-1

! LOCAL VARIABLES
REAL(EB) :: P_FUEL

INTEGER, DIMENSION(6) :: I_INDICES
INTEGER :: I, II, I_FUEL

I_FUEL = I_C2H4

! POPULATE I_INDICES USED FOR COMPUTATIONS OF GC
! [NOTE: THE TOTAL INTENSITY CALCULATED IS THAT WHICH LEAVES INTERVAL J=1.
! P(I,J) IS PARTIAL PRESSURE, ATM, OF SPECIES I IN  INTERVAL J.
! I=1,2,3,4,5, OR 6 IMPLIES SPECIES IS CO2, H2O, FUEL, CO, O2, OR N2, RESP.]

I_INDICES(1) = I_CO2
I_INDICES(2) = I_H2O
I_INDICES(3) = I_FUEL     ! FUEL
I_INDICES(4) = I_CO
I_INDICES(5) = I_O2
I_INDICES(6) = I_N2

! [NOTE: THE TOTAL INTENSITY CALCULATED IS THAT WHICH LEAVES INTERVAL J=1.
! P(I,J) IS PARTIAL PRESSURE, ATM, OF SPECIES I IN  INTERVAL J.
! I=1,2,3,4,5, OR 6 IMPLIES SPECIES IS CO2, H2O, CH4, CO, O2, OR N2, RESP.]

IF(ALLOCATED(GC)) DEALLOCATE(GC)
ALLOCATE(GC(N_RADCAL_SPECIES))
GC = 0.0_EB

! COMPUTE FUEL PRESSURE. REMOVE CONTRIBUTION OF NON FUEL SPECIES:
! CO2, CO, H2O, O2, N2

P_FUEL = PTOT - (SPECIES_PRESSURE_ATM(I_CO2)+SPECIES_PRESSURE_ATM(I_CO)+      &
               SPECIES_PRESSURE_ATM(I_H2O)+SPECIES_PRESSURE_ATM(I_N2)+      &
               SPECIES_PRESSURE_ATM(I_O2))

! COMPUTE COLLISIONAL BROADENING HALF-WIDTH AT HALF HEIGHT, EQ 5-34 NASA REPORT SP-3080
! INITIALIZATION

! CO2(II=1), H2O(II=2), FUEL(II=3), CO(II=4), O2 (II=5), N2 (II=6)
DO I = 1, 4

! IF THE SPECIES IS NOT PRESENT, I.E. SPECIES_PRESSURE_ATM(I_INDICES(I))==0.0_EB
! THEN SET GC TO ZERO AND CYCLE. ELSE PROCEED WITH CALCULATION

   IF (SPECIES_PRESSURE_ATM(I_INDICES(I))==0.0_EB.AND.I/=3) CYCLE

! INCLUDE NON-RESONANT FOREIGN AND SELF-BROADENING COLLISIONS
   DO II = 1, 6
      IF (I_INDICES(II) == I_FUEL) THEN
         GC(I_INDICES(I)) = GC(I_INDICES(I))+GAMMA(I,II)*P_FUEL*SQRT(AZOTEMP)
      ELSE
         GC(I_INDICES(I)) = GC(I_INDICES(I))+GAMMA(I,II)*SPECIES_PRESSURE_ATM(I_INDICES(II))*SQRT(AZOTEMP)
      ENDIF
   ENDDO
! INCLUDE RESONANT SELF-BROADENING COLLISIONS
   IF (I_INDICES(I) /= I_FUEL) THEN
      GC(I_INDICES(I))=GC(I_INDICES(I))+GAMMA(I,7)*SPECIES_PRESSURE_ATM(I_INDICES(I))*AZOTEMP
   ELSE
      GC(I_INDICES(I))=GC(I_INDICES(I))+GAMMA(I,7)*P_FUEL*AZOTEMP
   ENDIF
ENDDO

! COMPUTE THE COLLISION BROADENING COEFFICIENT FOR THE FUEL

DO I = 1, N_RADCAL_SPECIES
   IF (I/=I_CO2.AND.I/=I_O2.AND.I/=I_CO.AND.I/=I_H2O.AND.I/=I_N2.AND.I/=I_FV) &
      GC(I) = GC(I_FUEL)
ENDDO

!------------------------------------------------------------------------------
END FUNCTION COLLISION_BROADENING

!==============================================================================
SUBROUTINE SPECIES_OPTICAL_DEPTH(PATH_LENGTH_CM,P_ATM,PTOT,GC,TEMP,OMEGA,          &
                              OPTICAL_THICKNESS, CURTIS_XSTAR,                  &
                              CURTIS_ACOLLISION, CURTIS_ADOPPLER,OPTICAL_DEPTH)
!==============================================================================
!! THIS SUBROUTINE COMPUTES THE SPECIES OPTICAL_DEPTH FOR ALL THE SPECIES
!! GIVEN A WAVENUMBER OMEGA AND SPECIES OPTICAL_THICKNESS.
!! RETURNS THE OPTICAL DEPTH FOR EACH SPECIES USING THE CURTIS GODSON APPROXIMATION
!
!
! ARGUMENTS IN:
! ------------
! PATH_LENGTH_CM   : SCALAR, PHYSICAL LENGTH, IN CM
! P_ATM            : VECTOR, (NRADCAL_SPECIES) SPECIES PARTIAL PRESSURE, IN ATM
! PTOT             : SCALAR, TOTAL PRESSURE, IN ATM
! GC               : VECTOR, COLLISION-BROADENED HALF-WIDTH VALUES, IN CM-1
! TEMP             : SCALAR, LOCAL TEMPERATURE, IN K
! OMEGA            : SCALAR, WAVENUMBER, IN CM-1
! OPTICAL_THICKNESS: VECTOR (NRADCAL_SPECIES) ELEMENTS, UNITS IN STP.CM.ATM
!
! ARGUMENTS INOUT:
!-----------------
! VARIBLES USED TO COMPUTE THE CURTIS-GODSON APPROXIMATION PARAMETERS
! THE FIRST VALUES (I.E.) (NRADCAL_SPECIES,1) WAS EITHER COMPUTED PREVIOUSLY
! OR IS EQUAL TO 0 (FOR THE FIRST POINT)
!
! CURTIS_XSTAR      : ARRAY OF VALUES OF XSTART.     DIMENSION: (NRADCAL_SPECIES,2).
! CURTIS_ACOLLISION : ARRAY OF VALUES OF ACOLLISION. DIMENSION: (NRADCAL_SPECIES,2).
! CURTIS_ADOPPLER   : ARRAY OF VALUES OF ADOPPLER.   DIMENSION: (NRADCAL_SPECIES,2).
! OPTICAL_DEPTH     : ARRAY OF VALUES OF ELEMENTS OPTICAL DEPTH, DIMENSIONLESS
!                     DIMENSION: (NRADCAL_SPECIES,2).
!
!------------------------------------------------------------------------------

IMPLICIT NONE

! VARIABLES PASSED IN
REAL(EB), DIMENSION(:), INTENT(IN) :: P_ATM, OPTICAL_THICKNESS, GC
REAL(EB), INTENT(IN) :: PTOT, TEMP, OMEGA, PATH_LENGTH_CM

! VARIABLES PASSED INOUT
REAL(EB), DIMENSION(:,:), INTENT(INOUT) :: CURTIS_XSTAR, CURTIS_ACOLLISION,   &
                                          CURTIS_ADOPPLER, OPTICAL_DEPTH

! LOCAL
REAL(EB) :: X_DOPPLER, X_COLLISION, XSTAR
REAL(EB) :: SDWEAK,    A_COLLISION, A_DOPPLER

INTEGER  :: I_MODEL ! MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER
INTEGER  :: I

! INITIALIZE X_DOPPLER, X_COLLISION
X_DOPPLER   = 0.0_EB
X_COLLISION = 0.0_EB

! INITIALIZE CURTIS_XSTAR(:,2), CURTIS_ACOLLISION(:,2),
!            CURTIS_ADOPPLER(:,2), OPTICAL_DEPTH(:,2)
! THESE VALUES WILL BE UPDATED IF SDWEAK OF A GIVEN SPECIES IS GREATER THAN TINY_EB

CURTIS_XSTAR(:,2)      = CURTIS_XSTAR(:,1)
CURTIS_ACOLLISION(:,2) = CURTIS_ACOLLISION(:,1)
CURTIS_ADOPPLER(:,2)   = CURTIS_ADOPPLER(:,1)
OPTICAL_DEPTH(:,2)     = OPTICAL_DEPTH(:,1)

! LOOP SPECIES COMPUTES THE CONTRIBUTION OF EACH SPECIES TO TAU

LSPECIES: DO I = 1, N_RADCAL_SPECIES
   IF (P_ATM(I) <= TWO_EPSILON_EB) CYCLE LSPECIES

   I_MODEL = 1 ! ENFORCE GOODY MODEL WHEN NOT SPECIFIED

   SDWEAK      = TINY_EB  ! AVERAGE SPECTRAL ABSORPTION  COEFFICIENT. UNITS: (ATM^-1.CM^-1)
   A_COLLISION = 1.0_EB  ! FINE STRUCTURE PARAMETER. DIMENSIONLESS
   A_DOPPLER   = 1.0_EB

   ! IF THE SPECIES IS PRESENT IN THIS SEGMENT, GET SDWEAK, A_COLLISION, A_DOPPLER.
   ! OTHERWISE USE 0.0_EB, 1.0_EB, 1.0_EB

   SELECT CASE (I)
      CASE (I_CO2)
         CALL CO2(    OMEGA,TEMP,GC(I_CO2), SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_H2O)
         CALL H2O(    OMEGA,TEMP,GC(I_H2O), SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_CO)
         CALL CO(     OMEGA,TEMP,GC(I_CO),  SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_CH4_OLD)
         CALL CH4_OLD(OMEGA,TEMP,P_ATM(I_CH4_OLD),  PTOT,GC(I_CH4_OLD),SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_CH4)
         CALL CH4(    OMEGA,TEMP,P_ATM(I_CH4),  PTOT,GC(I_CH4),SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_C3H8)
         CALL C3H8(   OMEGA,TEMP,P_ATM(I_C3H8), PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_C7H16)
         CALL C7H16(  OMEGA,TEMP,P_ATM(I_C7H16),PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_CH3OH)
         CALL CH3OH(  OMEGA,TEMP,P_ATM(I_CH3OH),PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_C7H8)
         CALL C7H8(   OMEGA,TEMP,P_ATM(I_C7H8), PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_C3H6)
         CALL C3H6(   OMEGA,TEMP,P_ATM(I_C3H6), PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_MMA)
         CALL C5H8O2( OMEGA,TEMP,P_ATM(I_MMA),  PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_C2H6)
         CALL C2H6(   OMEGA,TEMP,P_ATM(I_C2H6), PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_C2H4)
         CALL C2H4(   OMEGA,TEMP,P_ATM(I_C2H4), PTOT,SDWEAK,A_COLLISION,A_DOPPLER,I_MODEL)
      CASE (I_FV)
      CASE DEFAULT
         CYCLE LSPECIES
      END SELECT

   ! IF (SDWEAK <= TINY_EB) THEN THE SPECIES DOES NOT PARTICIPATE ON THIS SEGMENTS. SKIP THE CALCULATION OF
   ! OPTICAL_DEPTH(:,2). IF NOT, PROCEED WITH CALCULATION OF X_COLLISION, X_DOPPLER, ETC...

   IF (I/=I_FV) THEN

      IF(SDWEAK <= TINY_EB) CYCLE LSPECIES

      XSTAR             = SDWEAK*OPTICAL_THICKNESS(I) ! OPTICAL DEPTH FOR THE WEAK LINE LIMIT.  SP-3080 EQ. 5-27
      CURTIS_XSTAR(I,2) = CURTIS_XSTAR(I,1) + XSTAR
      IF (CURTIS_XSTAR(I,2)>=TINY_EB) THEN
      ! COLLISION BROADENING
         CURTIS_ACOLLISION(I,2) = 1.0_EB/CURTIS_XSTAR(I,2)*(CURTIS_ACOLLISION(I,1)*CURTIS_XSTAR(I,1)+ &
         A_COLLISION*XSTAR)
      ! DOPPLER BROADENING
         CURTIS_ADOPPLER(I,2) = 1.0_EB/CURTIS_XSTAR(I,2)*(CURTIS_ADOPPLER(I,1)*CURTIS_XSTAR(I,1)+ &
         A_DOPPLER*XSTAR)
      ELSE
         CURTIS_ACOLLISION(I,2) = 1.0_EB
         CURTIS_ADOPPLER(I,2)   = 1.0_EB
      ENDIF

      IF (CURTIS_XSTAR(I,2) >= 1.E-20_EB .AND. CURTIS_ACOLLISION(I,2)>0._EB) THEN
!------------------------------------------------------------------------------
! COMPUTE THE OPTICAL DEPTH FOR THE ITH SPECIES USING RANDOM BAND MODEL
! COMPOSED OF LINES OF DOPPLER-LORENTZ SHAPE WITH:
! RANDOM DISTRIBUTION OF EITHER EQUAL STRENGTH LINES FOR DOPPLER OR DECAYING
! USE APPROPRIATE MODEL FOR LORENTZ LINE (ELSASSER,GOODY,MALKMUS)
! COMPUTE OPTICAL DEPTH FOR PURE DOPPLER CURVE OF GROWTH
         X_DOPPLER = GROWTH_DOPPLER(CURTIS_XSTAR(I,2),CURTIS_ADOPPLER(I,2),I_MODEL)

! LORENTZ OPTICAL DEPTH
! COMPUTE OPTICAL DEPTH FOR PURE COLLISION CURVE OF GROWTH
         SELECT CASE (I_MODEL)
            CASE(1)  ! GOODY MODEL
               X_COLLISION = GOODY(CURTIS_XSTAR(I,2),CURTIS_ACOLLISION(I,2))
            CASE(2)  ! MALKMUS MODEL
               X_COLLISION = MALKMUS(CURTIS_XSTAR(I,2),CURTIS_ACOLLISION(I,2))
            CASE(3)  ! ELSASSER MODEL
               X_COLLISION = ELSASSER(CURTIS_XSTAR(I,2),CURTIS_ACOLLISION(I,2))
            CASE DEFAULT
               X_COLLISION = GOODY(CURTIS_XSTAR(I,2),CURTIS_ACOLLISION(I,2))
         END SELECT

      ! COMPUTE OPTICAL DEPTH
         OPTICAL_DEPTH(I,2) = COMBINED_LINES(X_COLLISION, X_DOPPLER, CURTIS_XSTAR(I,2))

      ELSE ! XSTAR < 1.E-20_EB VERY WEAK LINE
         OPTICAL_DEPTH(I,2) = CURTIS_XSTAR(I,2)
      ENDIF ! CONDITION (XSTAR >= 1.E-20_EB)

   ! DETERMINE OPTICAL DEPTH OF SOOT
   ELSE ! TREATMENT FOR SOOT VOLUME FRACTION
      CALL POD(OMEGA, P_ATM(I),PATH_LENGTH_CM,XSTAR,TEMP)
      CURTIS_XSTAR(I,2)  = CURTIS_XSTAR(I,1) + XSTAR
      OPTICAL_DEPTH(I,2) = CURTIS_XSTAR(I,2)

   ENDIF

END DO LSPECIES

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE SPECIES_OPTICAL_DEPTH

!==============================================================================
ELEMENTAL FUNCTION GROWTH_DOPPLER(XSTAR,A_DOPPLER,I_MODEL) RESULT(X_DOPPLER)
!==============================================================================
!! THIS FUNCTION COMPUTES THE DOPPLER CURVE OF GROWTH.
!! TWO CASES
!! 1) EQUALLY INTENSE, RANDOMLY SPACED DOPPLER LINE (I_MODEL =1 OR 3)
!! 2) GEOMETRICALLY DECAYING STRENGTH, RANDOMLY SPACED DOPPLER LINE (I_MODEL=2)
!
! VARIABLES IN:
! XSTAR: OPTICAL PATHLENGTH * MEAN ABSORPTION COEFFICIENT
! A_DOPPLER: DOPPLER FINE STRUCTURE PARAMETER
! I_MODEL: MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER
!
! VARIABLE OUT
! X_DOPPLER

! VARIABLES IN:
REAL(EB), INTENT(IN) :: XSTAR
REAL(EB), INTENT(IN) :: A_DOPPLER
INTEGER,  INTENT(IN) :: I_MODEL

! VARIABLE OUT
REAL(EB) :: X_DOPPLER

! SOURCE: ! SP-3080, CHAP 3.2.5 E AND F. SEE PAGE 105.

SELECT CASE(I_MODEL)
CASE(1) ! GOODY MODEL: EQUALLY INTENSE, RANDOMLY SPACED DOPPLER LINE
      X_DOPPLER = 1.7_EB*A_DOPPLER*SQRT(LOG(1._EB+(0.58824_EB*XSTAR/A_DOPPLER)**2))
CASE(2) ! MALKMUS: GEOMETRICALLY DECAYING STRENGTH, RANDOMLY SPACED DOPPLER LINE
      X_DOPPLER = 0.937_EB*A_DOPPLER*(LOG(1.0_EB+(1.07_EB*XSTAR/A_DOPPLER)**(0.6666_EB)))* &
                                 SQRT(LOG(1.0_EB+(1.07_EB*XSTAR/A_DOPPLER)**(0.6666_EB)))
CASE(3)
      X_DOPPLER = 1.7_EB*A_DOPPLER*SQRT(LOG(1._EB+(0.58824_EB*XSTAR/A_DOPPLER)**2))
CASE DEFAULT
      X_DOPPLER = 1.7_EB*A_DOPPLER*SQRT(LOG(1._EB+(0.58824_EB*XSTAR/A_DOPPLER)**2))
END SELECT
!------------------------------------------------------------------------------
END FUNCTION GROWTH_DOPPLER

!==============================================================================
ELEMENTAL FUNCTION COMBINED_LINES(X_COLLISION, X_DOPPLER, XSTAR) RESULT(OPTICAL_DEPTH)
!==============================================================================
!! COMPUTE THE COMBINED COLLISION AND DOPPLER OPTICAL DEPTHS AND
! THE OPTICAL DEPTH BASED ON EXPRESSION GIVEN IN SP-3080 EQ 5-26 & EQ 5-25 (P220)
!
! VARIABLES IN:
! X_COLLISION : LORENZ CURVE OF GROWTH
! X_DOPPLER: DOPPLER CURVE OF GROWTH
! XSTAR: OPTICAL PATHLENGTH * MEAN ABSORPTION COEFFICIENT
!
! VARIABLE OUT
! OPTICAL_DEPTH

! VARIABLES PASSED IN
REAL(EB), INTENT(IN) :: X_COLLISION
REAL(EB), INTENT(IN) :: X_DOPPLER
REAL(EB), INTENT(IN) :: XSTAR

! VARIABLE PASSED OUTPUT
REAL(EB) :: OPTICAL_DEPTH

! LOCAL VARIABLES
REAL(EB) :: Y_DOPPLER, Y_COLLISION, Y_COMBINED

! COMPUTE THE COMBINED COLLISION AND DOPPLER OPTICAL DEPTHS Y_COMBINED

Y_DOPPLER   = MAX(1._EB-(X_DOPPLER/XSTAR)**2,TWO_EPSILON_EB)
Y_COLLISION = MAX(1._EB-(X_COLLISION/XSTAR)**2,TWO_EPSILON_EB)
! SP-3080 EQ. 5-26
Y_COMBINED = MAX(1._EB/Y_COLLISION**2+1._EB/Y_DOPPLER**2-1._EB,1._EB)
! FINALLY COMPUTE OPTICAL DEPTH
! SP-3080 EQ. 5-25
OPTICAL_DEPTH = XSTAR*SQRT(1._EB-(Y_COMBINED**(-0.5_EB)))

!------------------------------------------------------------------------------
END FUNCTION COMBINED_LINES


!==============================================================================
ELEMENTAL FUNCTION GOODY(XSTAR,A_COLLISION) RESULT(X_GOODY)
!==============================================================================
!! FUNCTION COMPUTES THE EQUIVALENT LINE WIDTH OVER THE AVERAGE LINE SPACING
!! USING THE GOODY STATISTICAL MODEL
!
! VARIABLES IN:
! XSTAR       : OPTICAL PATHLENGTH * MEAN ABSORPTION COEFFICIENT
! A_COLLISION : FINE STRUCTURE PARAMETER
!
! RETURNS X_GOODY

IMPLICIT NONE

REAL(EB), INTENT(IN) :: XSTAR
REAL(EB), INTENT(IN) :: A_COLLISION

REAL(EB)  :: X_GOODY

IF (XSTAR<1.0E-15_EB) THEN ! WEAK LINE LIMIT
X_GOODY = XSTAR
ELSEIF(XSTAR>1.0E+5_EB) THEN ! STRONG LINE LIMIT
X_GOODY = SQRT(4.0_EB*XSTAR*A_COLLISION)
ELSE                        ! FORMULATION TO AVOID DIVISION BY ZERO
X_GOODY = SQRT(A_COLLISION)*XSTAR/SQRT(A_COLLISION+0.25_EB*XSTAR)
ENDIF

RETURN
!------------------------------------------------------------------------------
END FUNCTION GOODY

!==============================================================================
ELEMENTAL FUNCTION MALKMUS(XSTAR,A_COLLISION) RESULT(X_MALKMUS)
!==============================================================================
!
!! FUNCTION COMPUTES THE EQUIVALENT LINE WIDTH OVER THE AVERAGE LINE SPACING
!! USING THE MALKMUS STATISTICAL MODEL
!
! VARIABLES IN:
! XSTAR       : OPTICAL PATHLENGTH * MEAN ABSORPTION COEFFICIENT
! A_COLLISION : FINE STRUCTURE PARAMETER
!
! RETURNS X_MALKMUS (OPTICAL DEPTH FOR A PURE COLLISION CURVE)

IMPLICIT NONE

! VARIABLES PASSED IN
REAL(EB), INTENT(IN) :: XSTAR
REAL(EB), INTENT(IN) :: A_COLLISION

REAL(EB)  :: X_MALKMUS

IF    (XSTAR < 1.0E-15_EB) THEN ! WEAK LINE LIMIT
X_MALKMUS = XSTAR
ELSEIF(XSTAR > 1.0E+5_EB ) THEN ! STRONG LINE LIMIT
X_MALKMUS = SQRT(4.0_EB*XSTAR*A_COLLISION)
ELSE             ! FORMULATION TO AVOID DIVISION BY A_COLLISION IF EQUAL TO ZERO
X_MALKMUS = SQRT(4.0_EB*A_COLLISION**2+4.0_EB*A_COLLISION*XSTAR)-2.0_EB*A_COLLISION
ENDIF

RETURN
!------------------------------------------------------------------------------
END FUNCTION MALKMUS

!==============================================================================
FUNCTION ELSASSER(XSTAR,A_COLLISION) RESULT(X_ELSASSER)
!==============================================================================
!! FUNCTION COMPUTES THE EQUIVALENT LINE WIDTH OVER THE AVERAGE LINE SPACING
!! USING THE ELSASSER STATISTICAL MODEL
!
! VARIABLES IN:
! XSTAR       : OPTICAL PATHLENGTH * MEAN ABSORPTION COEFFICIENT
! A_COLLISION : FINE STRUCTURE PARAMETER
!
! RETURNS X_ELSASSER
!------------------------------------------------------------------------------

IMPLICIT NONE

! VARIABLES PASSED IN
REAL(EB), INTENT(IN) :: XSTAR
REAL(EB), INTENT(IN) :: A_COLLISION

! VARIABLES PASSED OUT
REAL(EB)  :: X_ELSASSER

! LOCAL VARIABLES

REAL(EB) :: X_COLLISION

! THE FOLLOWING COMPUTES THE OPTICAL DEPTH, X_COLLISION, FOR ANY SPECIES REPRESENTED
! BY THE ELSASSER MODEL USING THE GODSON EQUATION AND AN APPROXIMATION TO THE LADENBERG-REICHE
! FUNCTION AS RECOMMENDED BY BROSMER AND TIEN (JQSRT 33,P 521), EQ. 2, PAGE 523

X_COLLISION = SQRT(4.0_EB*A_COLLISION)*XSTAR/SQRT(16.0_EB/PI*A_COLLISION+XSTAR)

X_ELSASSER  = -LOG(1.0_EB-ERF(X_COLLISION)+TWO_EPSILON_EB)

!------------------------------------------------------------------------------
RETURN
END FUNCTION ELSASSER

!==============================================================================
PURE FUNCTION PARTITION_FUNCTION(TRANSITION_WN,DEGENERACY,TEMP)
!==============================================================================
!! THIS FUNCTION CALCULATE THE PARTITION_FUNCTION OF AN HARMONIC
!! QUANTUM OSCILLATOR KNOWING THE DIFFERENT FREQUENCIES AND DEGENERACIES
!------------------------------------------------------------------------------

IMPLICIT NONE

! VARIABLE PASSED IN
REAL(EB), DIMENSION(:), INTENT(IN) :: TRANSITION_WN ! WAVENUMBER ASSOCIATED WITH
                                                   ! A VIBRATION MODE

REAL(EB), DIMENSION(:), INTENT(IN) :: DEGENERACY  ! DENEGERACY OF ENERGY LEVEL
                                                ! ASSOCIATED WITH A VIBRATION
                                                ! MODE

REAL(EB), INTENT(IN) :: TEMP                     ! TEMPERATURE IN KELVIN

! VARIABLE PASSED OUT
REAL(EB) :: PARTITION_FUNCTION

! LOCAL VARIABLES
REAL(EB), PARAMETER :: Q2 = 1.4388_EB   ! SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB) :: Q2_OVER_T
INTEGER  :: N_MODE
INTEGER  :: I_MODE

Q2_OVER_T = Q2/TEMP
N_MODE    = SIZE(TRANSITION_WN)

PARTITION_FUNCTION = 1.0_EB

DO I_MODE = 1, N_MODE
PARTITION_FUNCTION = PARTITION_FUNCTION* &
         (1.0_EB-EXP(-Q2_OVER_T*TRANSITION_WN(I_MODE)))**(DEGENERACY(I_MODE))
ENDDO

!------------------------------------------------------------------------------
RETURN
END FUNCTION PARTITION_FUNCTION


!==============================================================================
PURE SUBROUTINE APPROX_VIB_ROT(TRANSITION_MATRIX,LINE_STRENGTH,OM,TEMP,OMEGA,  &
                              BCNT,BE,GC1, GD, GDINV,GDDINV,SDWEAK,LINE_SPACING)
!==============================================================================
!! JUST OVERLAPPING SPECTRAL LINES
!
!
!------------------------------------------------------------------------------

IMPLICIT NONE

! VARIABLE PASSED IN
REAL(EB), DIMENSION(:,:), INTENT(IN) :: TRANSITION_MATRIX ! SQUARE MATRIX
REAL(EB), DIMENSION(:),   INTENT(IN) :: BCNT ! BAND CENTER IN WAVENUMBERS (CM-1)
REAL(EB), DIMENSION(:),   INTENT(IN) :: OM
REAL(EB), DIMENSION(:),   INTENT(IN) :: LINE_STRENGTH

REAL(EB), INTENT(IN) :: OMEGA ! WAVENUMBER OF SOUGHT PROPERTIES
REAL(EB), INTENT(IN) :: BE    ! ROTATIONAL CONSTANT
REAL(EB), INTENT(IN) :: TEMP  ! TEMPERATURE IN KELVIN
REAL(EB), INTENT(IN) :: GC1
REAL(EB), INTENT(IN) :: GD
REAL(EB), INTENT(IN), OPTIONAL :: LINE_SPACING ! OPTIONAL

! VARIABLES PASSED OUT
REAL(EB), INTENT(OUT) :: SDWEAK
REAL(EB), INTENT(OUT) :: GDINV
REAL(EB), INTENT(OUT) :: GDDINV

! LOCAL VARIABLES
REAL(EB), DIMENSION(:), ALLOCATABLE :: TRANSITION_WN, ATOT
REAL(EB), PARAMETER :: Q2 = 1.4388_EB   ! SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: T0 = 300._EB     ! REFERENCE TEMPERATURE, IN KELVIN

REAL(EB) :: T0OT
REAL(EB) :: Q2OT
REAL(EB) :: DINV

INTEGER  :: N_TRANSITIONS
INTEGER  :: N_VIBRATIONS

INTEGER :: I_TRANSITION

SDWEAK  =  0.0_EB
T0OT    =  T0/TEMP
Q2OT    = -Q2/TEMP

N_TRANSITIONS = SIZE(TRANSITION_MATRIX,2)
N_VIBRATIONS  = SIZE(TRANSITION_MATRIX,1)

IF (.NOT.PRESENT(LINE_SPACING)) THEN
DINV = 1.0_EB/(4.0_EB*BE)
ELSE
DINV = LINE_SPACING
ENDIF

! COMPUTE THE ASSOCIATED TRANSITION WAVENUMBERS
! ALLOCATE VARIABLE TRANSITION_WN

IF (ALLOCATED(TRANSITION_WN)) DEALLOCATE(TRANSITION_WN)
ALLOCATE(TRANSITION_WN(N_VIBRATIONS))

! ALLOCATE VARIABLE ATOT

IF (ALLOCATED(ATOT)) DEALLOCATE(ATOT)
ALLOCATE(ATOT(N_VIBRATIONS))

! CALCULATE THE VECTOR OF TRANISTION WAVENUMBERS
TRANSITION_WN = MATMUL(TRANSITION_MATRIX,OM)

! COMPUTE THE DIFFERENT LINE INTENSITIES AT TEMPERATURE TEMP

DO I_TRANSITION = 1, N_TRANSITIONS
ATOT(I_TRANSITION) = LINE_STRENGTH(I_TRANSITION)*T0OT*(1.0_EB-EXP(Q2OT*TRANSITION_WN(I_TRANSITION)))/ &
PARTITION_FUNCTION(OM,TRANSITION_MATRIX(I_TRANSITION,:),TEMP)
ENDDO

! COMPUTE THE MEAN ABSORPTION COEFFICIENT USING EQ 11-44 FROM PENNER IN QUANTITATIVE MOLECULAR SPECTROMETRY, 1955
! SDWEAK = SUM(ATOT*PROPABILITY)
! JUST OVERLAPPING SPECTRAL LINE

DO I_TRANSITION = 1, N_TRANSITIONS
SDWEAK = SDWEAK + ATOT(I_TRANSITION)*(-Q2OT/(4._EB*BE))*ABS(OMEGA-BCNT(I_TRANSITION))* &
                  EXP(Q2OT/(4._EB*BE)*(OMEGA-BCNT(I_TRANSITION))**2)
ENDDO

GDINV  = GC1*DINV
GDDINV = GD*DINV
!EXPRESS S/D AT STP, AS IS IN NASA SP-3080
SDWEAK = SDWEAK*TEMP/273._EB

!------------------------------------------------------------------------------
RETURN
END SUBROUTINE APPROX_VIB_ROT

!==============================================================================
ELEMENTAL SUBROUTINE CO2(OMEGA,TEMP,GC1,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER
IMPLICIT NONE

! VARIABLES PASSED IN
REAL(EB), INTENT(IN) :: OMEGA ! WAVENUMBER, UNITS IN CM-1
REAL(EB), INTENT(IN) :: TEMP  ! TEMPERATURE, UNITS IN KELVIN
REAL(EB), INTENT(IN) :: GC1   ! HALF WIDTH AT HALF HEIGHT FOR LORENTZ LINE, UNITS IN CM-1

! VARIABLES PASSED OUT
REAL(EB), INTENT(OUT):: SDWEAK  ! NARROW BAND MEAN ABS(ORTION COEFFICIENT, UNITS IN CM-1
REAL(EB), INTENT(OUT):: GDINV   ! LINE WIDTH TO LINE SPACING RATION FOR LORENTZ LINES
REAL(EB), INTENT(OUT):: GDDINV  ! LINE WIDTH TO LINE SPACING RATION FOR DOPPLER
INTEGER,  INTENT(OUT):: I_MODEL ! MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCAL VARIABLES
INTEGER  :: I,J,K,L

REAL(EB), PARAMETER :: WM_CO2 = 44._EB      ! MOLAR MASS (IN GRAM) OF CO2
REAL(EB), PARAMETER :: BE     = 0.391635_EB ! ROTATIONAL CONSTANT CO2
REAL(EB), PARAMETER :: Q2     = 1.4388_EB   ! SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: T0     = 300._EB     ! REFERENCE TEMPERATURE, IN KELVIN

REAL(EB), PARAMETER :: OM1 = 1354.91_EB  ! FUNDAMENTAL FREQUENCY FOR TRANSITION (000)->(100), UNITS CM-1
REAL(EB), PARAMETER :: OM2 =  673.0_EB   ! FUNDAMENTAL FREQUENCY FOR TRANSITION (000)->(010), UNITS CM-1
REAL(EB), PARAMETER :: OM3 = 2396.49_EB  ! FUNDAMENTAL FREQUENCY FOR TRANSITION (000)->(001), UNITS CM-1

REAL(EB), DIMENSION(3) :: ATOT, BCNT
REAL(EB), DIMENSION(:), ALLOCATABLE  :: OM, LINE_STRENGTH
REAL(EB), DIMENSION(:,:), ALLOCATABLE :: TRANSITION_MATRIX

REAL(EB) :: AA,BB,CC,QQ,EE,FF,GG, &
         SMINUS,SPLUS,SDSTRG,GD, &
         X13,X23,X33,XBAR,OM12,ALPHA,OMPRIM,V3,GAM,  &
         OMVV3,DELTA,V,OMVBAR,F1,F2,UNFLO1,UNFLO2,UNFLO3,TEST, &
         VBAR1,OMA,OMB,TTEMP,TT,T1,TW,WW,TEMP1,TEMP2,TEMP3,    &
         DINV,A,B,D,G,W1,DINV1,DINV2,DINV3,Q2OT,T0OT,Q2OT0

I_MODEL = I_MODEL_CO2

! INITIALIZE SDWEAK, GDINV, GDDINV

SDWEAK = 0._EB
GDINV  = 1._EB
GDDINV = 1._EB

IF (OMEGA > 5725.0_EB) THEN
SDWEAK = 0.0_EB
GDINV  = 1.0_EB
GDDINV = 1.0_EB
ELSE

! COMPUTE DOPPLER BROADENING HALF-WIDTHS SP-3080 EQ: 5-35
GD = 5.94E-6_EB*OMEGA*SQRT(TEMP/(273.0_EB*WM_CO2))

IF(OMEGA>4550.0_EB) THEN
! CONTRIBUTION TO 2.0 MICRON BAND FROM:
! (000)-(041)
! (000)-(121),
! (000)-(201) TRANSITIONS.

! BAND CENTER

   BCNT(1) = 4860.5_EB ! (000)-(041)
   BCNT(2) = 4983.5_EB ! (000)-(121)
   BCNT(3) = 5109.0_EB ! (000)-(201)

   IF (ALLOCATED(TRANSITION_MATRIX)) DEALLOCATE(TRANSITION_MATRIX)
   ALLOCATE(TRANSITION_MATRIX(3,3))

   IF(ALLOCATED(LINE_STRENGTH)) DEALLOCATE(LINE_STRENGTH)
   ALLOCATE(LINE_STRENGTH(3))

   IF(ALLOCATED(OM)) DEALLOCATE(OM)
   ALLOCATE(OM(3))

   ! ! COMPUTE THE ASSOCIATED TRANSITION WAVENUMBERS

   TRANSITION_MATRIX(1,1:3) = [0,4,1] ! (000)-(041)
   TRANSITION_MATRIX(2,1:3) = [1,2,1] ! (000)-(121)
   TRANSITION_MATRIX(3,1:3) = [2,0,1] ! (000)-(201)

   OM(1) = OM1
   OM(2) = OM2
   OM(3) = OM3

!! COMPUTE THE INTEGRATED INTENSITY OF THE BAND
!! 0.272: INTEGRATED INTENSITY (IN CM-2-ATM-1) OF BAND BCNT(1) = 4860.5_EB ((041)-(000)) AT T=300K
!! 1.01 : INTEGRATED INTENSITY (IN CM-2-ATM-1) OF BAND BCNT(2) = 4983.5_EB ((121)-(000)) AT T=300K
!! 0.426: INTEGRATED INTENSITY (IN CM-2-ATM-1) OF BAND BCNT(3) = 5109  _EB ((201)-(000)) AT T=300K
!! VALUES FROM PENNER AND VARANASI JQSRT VOL 4 PP 799-806, 1964

   LINE_STRENGTH = [0.272_EB,1.01_EB,0.426_EB]

   CALL APPROX_VIB_ROT(TRANSITION_MATRIX,LINE_STRENGTH,OM,TEMP,OMEGA,BCNT,BE,GC1, &
                        GD,GDINV,GDDINV,SDWEAK)

ELSEIF((OMEGA<=4550._EB).AND.(OMEGA>3800._EB)) THEN
   SDWEAK = 0.0_EB
   GDINV  = 1.0_EB
   GDDINV = 1.0_EB

ELSEIF((OMEGA<=3800._EB).AND.(OMEGA>3050._EB)) THEN
   B   = 0.391635_EB
   A   = 0.0030875_EB

   X13 = -19.37_EB
   X23 = -12.53_EB
   X33 = -12.63_EB

   T0OT = T0/TEMP

   Q2OT  = -Q2/TEMP
   Q2OT0 = -Q2/T0

   XBAR  = 0.5_EB*(.5_EB*X13+X23)
   OM12  = 0.5_EB*(.5_EB*OM1+OM2)

   SDWEAK = 0._EB
   SDSTRG = 0._EB

!CALCULATE ABSORPTION COEF. AND LINE SPACING PARAMETER FOR 2.7 MICRON BAND
   L = 1
!CONTRIBUTION TO 2.7 MICRON BAND FROM (000)-(021) AND (010)-(031) TRANS.
   ALPHA  = 28.5_EB
   OMPRIM = 2.0_EB*OM2+OM3

   L120: DO
      AA = ALPHA*B*Q2/(A*(1._EB-EXP(OM3*Q2OT0))*(1._EB-EXP(OM12*Q2OT0))**3*(1._EB+EXP(OM12*Q2OT0))*(1._EB-EXP(OMPRIM*Q2OT0)))
      BB = (1._EB-EXP(Q2OT*OMEGA))*(1._EB-EXP(Q2OT*OM3))* (1._EB-EXP(OM12*Q2OT))**3*(1._EB+EXP(OM12*Q2OT)) &
            *(1._EB-EXP(Q2OT*OMPRIM))
      CC = AA*BB*OMEGA*T0/TEMP**2

      L102: DO J = 1, 20
         V = REAL(J-1,EB)
         IF ((J/2*2)==J) G = (V+1.0_EB)*(V+3.0_EB)/4.0_EB ! REPLACE WITH MODULO 2
         IF ((J/2*2)/=J) G = (V+2.0_EB)*(V+2.0_EB)/4.0_EB ! REPLACE WITH MODULO 2

         VBAR1 = -1.0_EB+(V+3.0_EB)*(V+4.0_EB)/(V+2.0_EB)/6.0_EB

         IF(J/2*2==J) VBAR1 = -1.0_EB+(V+5.0_EB)/6.0_EB ! REPLACE WITH MODULO 2

         L101: DO K = 1, 10
            V3  = REAL(K-1,EB)
            QQ  = (V3+1)*G*EXP((V3*OM3+V*OM12)*Q2OT)*(VBAR1+1.0_EB)
            GAM = B-A*(V3+1._EB)

            IF(L==2) THEN
               OMVV3 = 3728.0_EB-5.0_EB*V-47._EB*V3
               IF (V<=TWO_EPSILON_EB) OMVV3 = 3715._EB-47._EB*V3
            ELSE
               OMVV3 = 3598.0_EB-18.0_EB*V-47.0_EB*V3
               IF (V<=TWO_EPSILON_EB) OMVV3 = 3613.0_EB-47.0_EB*V3
            ENDIF

            DELTA = A*(OMEGA-OMVV3)

            IF ((GAM**2)<=DELTA) CYCLE L102

            D      = 2.0_EB*SQRT(GAM**2-DELTA)
            OMVBAR = OMVV3*(1.0_EB-EXP(OMVV3*Q2OT))
            F1     = GAM - 0.5_EB*D
            F2     = GAM + 0.5_EB*D
            EE     = Q2*GAM/(A**2*TEMP)
            UNFLO1 = EE*DELTA*(1.0_EB + 0.5_EB*A/GAM)

            IF(UNFLO1<=-78.0_EB) CYCLE L102

            UNFLO2 = EE*2.0_EB*GAM*F1

            IF(UNFLO2>=78.0_EB) CYCLE L102

            FF     = EXP(EE*DELTA*(1.0_EB+0.5_EB*A/GAM))
            SMINUS = CC*QQ/OMVBAR*ABS(F1)*FF*EXP(-EE*2.0_EB*GAM*F1)
            UNFLO3 = EE*2.0_EB*GAM*F2

            IF(UNFLO3>=78.0_EB) THEN
               SPLUS = 0.0_EB
            ELSE
               SPLUS = CC*QQ/OMVBAR*ABS(F2)*FF*EXP(-EE*2.0_EB*GAM*F2)
            ENDIF

            GG     = SDWEAK
            SDWEAK = (SMINUS+SPLUS)/D+SDWEAK
            TEST   = (SDWEAK-GG)/SDWEAK
            IF(TEST<.0001_EB) CYCLE L102
            SDSTRG = SQRT(0.5_EB*G)*(SQRT(SMINUS)+SQRT(SPLUS))/D+SDSTRG
         ENDDO L101
      ENDDO L102

      IF(L==2) EXIT L120
!CONTRIBUTION TO 2.7 MICRON BAND FROM (000)-(101) AND (010)-(111) TRANS.
      ALPHA  = 42.3_EB
      OMPRIM = OM1+OM3
      L      = 2
   ENDDO L120
!CALCULATE ABSORPTION COEF AND LINE SPACING PARAMETER FOR 4.3 MICRON BAND
   IF(SDWEAK<=TINY_EB) THEN
      SDWEAK = 0.0_EB
      GDINV  = 1.0_EB
      GDDINV = 1.0_EB
   ELSE
      DINV   = SDSTRG*SDSTRG/SDWEAK
      GDINV  = GC1*DINV
      GDDINV = GD*DINV
!***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
      SDWEAK = SDWEAK*TEMP/273.0_EB
!         ENDIF
   ENDIF

ELSEIF((OMEGA<=3050.0_EB).AND.(OMEGA>2474.0_EB)) THEN
   SDWEAK = 0.0_EB
   GDINV  = 1.0_EB
   GDDINV = 1.0_EB

ELSEIF((OMEGA<=2474.0_EB).AND.(OMEGA>1975.0_EB)) THEN

B = .391635_EB
A = .0030875_EB

X13 = -19.37_EB
X23 = -12.53_EB
X33 = -12.63_EB

XBAR = 0.5_EB*(0.5_EB*X13+X23)
OM12 = 0.5_EB*(0.5_EB*OM1+OM2)

SDWEAK=0.0_EB
SDSTRG=0.0_EB

! DIPOLE APPROXIMATION

IF(OMEGA<=2395.0_EB) THEN
      ALPHA  = 2700.0_EB
      OMPRIM = OM3
      AA     = ALPHA*B*Q2/(A*(1.0_EB-EXP(-OM3*Q2/T0))*(1.0_EB-EXP(-OM12*Q2 /T0))**3* &
                              (1.0_EB+EXP(-OM12*Q2/T0))*(1.0_EB-EXP(-OMPRIM*Q2/T0)))
      BB     = (1._EB-EXP(-Q2*OMEGA/TEMP))*(1.0_EB-EXP(-Q2*OM3/TEMP))*(1.0_EB-EXP(-OM12*Q2/TEMP))**3*&
               (1._EB+EXP(-OM12*Q2/TEMP))*(1.0_EB-EXP(-Q2*OMPRIM/TEMP))
      CC     = AA*BB*OMEGA*T0/TEMP**2

      L202A: DO J = 1, 20
         V  = REAL(J-1,EB)

         IF(J/2*2==J) G = 0.25_EB*(V+1.0_EB)*(V+3.0_EB) ! REPLACE WITH MODULO
         IF(J/2*2/=J) G = 0.25_EB*(V+2.0_EB)*(V+2.0_EB) ! REPLACE WITH MODULO

         L201A: DO K=1,10

            V3    = REAL(K-1,EB)
            QQ    = (V3+1.0_EB)*G*EXP(-(V3*OM3+V*OM12)*Q2/TEMP)
            GAM   = B-A*(V3+1.0_EB)
            OMVV3 = OM3 + 0.5_EB*X13 + X23 + 2.0_EB*X33 + XBAR*V + 2.0_EB*X33*V3
            DELTA = A*(OMEGA-OMVV3)

            IF(GAM*GAM<=DELTA) CYCLE L202A

            D      = 2.0_EB*SQRT(GAM**2-DELTA)
            OMVBAR = OMVV3*(1._EB-EXP(-OMVV3*Q2/TEMP))

            F1     = GAM-0.5_EB*D
            F2     = GAM+0.5_EB*D
            EE     = Q2*GAM/(A*A*TEMP)
            UNFLO1 = EE*DELTA*(1.0_EB + 0.5_EB*A/GAM)

            IF(UNFLO1<=-78.0_EB) CYCLE L202A
            UNFLO2 = EE*2.0_EB*GAM*F1

            IF(UNFLO2>=78.0_EB) CYCLE L202A

            FF     = EXP(EE*DELTA*(1.0_EB + 0.5_EB*A/GAM))
            SMINUS = CC*QQ/OMVBAR*ABS(F1)*FF*EXP(-EE*2.0_EB*GAM*F1)
            UNFLO3 = EE*2.0_EB*GAM*F2

            IF (UNFLO3 >= 78.0_EB) THEN
               SPLUS=0.0_EB
            ELSE
               SPLUS=CC*QQ/OMVBAR*ABS(F2)*FF*EXP(-EE*2.0_EB*GAM*F2)
            ENDIF

            GG     = SDWEAK
            SDWEAK = (SMINUS+SPLUS)/D+SDWEAK
            TEST   = (SDWEAK-GG)/SDWEAK

            IF(TEST<1.0E-4_EB) CYCLE L202A

            SDSTRG=SQRT(0.5_EB*G)*(SQRT(SMINUS)+SQRT(SPLUS))/D+SDSTRG

         ENDDO L201A
      ENDDO L202A

      IF(SDWEAK<=TINY_EB) THEN
         SDWEAK = 0.0_EB
         GDINV  = 1.0_EB
         GDDINV = 1.0_EB
      ELSE
         DINV   = SDSTRG**2/SDWEAK
         GDINV  = GC1*DINV
         GDDINV = GD*DINV
!***  EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
         SDWEAK = SDWEAK*TEMP/273.0_EB
      ENDIF
   ELSE
!CALCULATE ABSORPTION COEF. AND LINE SPACING PARAMETER FOR 2.7 MICRON BAND
      L=1
!CONTRIBUTION TO 2.7 MICRON BAND FROM (000)-(021) AND (010)-(031) TRANS.
      ALPHA  = 28.5_EB
      OMPRIM = 2.0_EB*OM2+OM3

      L120A: DO
         AA = ALPHA*B*Q2/(A*(1.0_EB-EXP(-OM3*Q2/T0))*(1.0_EB-EXP(-OM12*Q2 /T0))**3* &
                              (1.0_EB+EXP(-OM12*Q2/T0))*(1.0_EB-EXP(-OMPRIM*Q2/T0)))
         BB = (1._EB-EXP(-Q2*OMEGA/TEMP))*(1.0_EB-EXP(-Q2*OM3/TEMP))*    &
               (1._EB-EXP(-OM12*Q2/TEMP))**3*(1.0_EB+EXP(-OM12*Q2/TEMP))* &
               (1._EB-EXP(-Q2*OMPRIM/TEMP))

         CC = AA*BB*OMEGA/TEMP*T0/TEMP

         L102A: DO J = 1, 20
            V=REAL(J-1,EB)

            IF(J/2*2==J) G = 0.25_EB*(V+1.0_EB)*(V+3.0_EB) ! USE MODULO
            IF(J/2*2/=J) G = 0.25_EB*(V+2.0_EB)*(V+2.0_EB) ! USE MODULO

            VBAR1 = -1.0_EB + (V+3.0_EB)*(V+4.0_EB)/(6.0_EB*(V+2.0_EB))
            IF(J/2*2==J) VBAR1 = -1.0_EB+(V+5.0_EB)/6.0_EB ! USE MODULO

            L101A: DO K = 1, 10

               V3  = REAL(K-1,EB)
               QQ  = (V3+1)*G*EXP( -(V3*OM3 + V*OM12)*Q2/TEMP )*(VBAR1+1.0_EB)
               GAM = B-A*(V3+1.0_EB)

               IF(L==2) THEN
                  OMVV3 = 3728.0_EB - 5.0_EB*V - 47.0_EB*V3
                  IF(V<=TWO_EPSILON_EB) OMVV3 = 3715.0_EB - 47.0_EB*V3
               ELSE
                  OMVV3 = 3598.0_EB - 18.0_EB*V - 47.0_EB*V3
                  IF(V<=TWO_EPSILON_EB) OMVV3 = 3613.0_EB - 47.0_EB*V3
               ENDIF

               DELTA = A*(OMEGA-OMVV3)
               IF((GAM**2)<=DELTA) CYCLE L102A

               D      = 2.0_EB*SQRT(GAM**2-DELTA)
               OMVBAR = OMVV3*(1.0_EB-EXP(-OMVV3*Q2/TEMP))
               F1     = GAM-0.5_EB*D
               F2     = GAM+0.5_EB*D

               EE     = Q2*GAM/(A**2*TEMP)
               UNFLO1 = EE*DELTA*(1.0_EB+0.5_EB*A/GAM)

               IF(UNFLO1<= -78.0_EB) CYCLE L102A

               UNFLO2 = EE*2.0_EB*GAM*F1

               IF(UNFLO2>= 78._EB) CYCLE L102A

               FF     = EXP(EE*DELTA*(1.0_EB+0.5_EB*A/GAM))
               SMINUS = CC*QQ/OMVBAR*ABS(F1)*FF*EXP(-EE*2.0_EB*GAM*F1)
               UNFLO3 = EE*2.0_EB*GAM*F2

               IF(UNFLO3>=78._EB) THEN
                  SPLUS = 0._EB
               ELSE
                  SPLUS = CC*QQ/OMVBAR*ABS(F2)*FF*EXP(-EE*2.0_EB*GAM*F2)
               ENDIF

               GG     = SDWEAK
               SDWEAK = (SMINUS+SPLUS)/D+SDWEAK
               TEST   = (SDWEAK-GG)/SDWEAK

               IF(TEST<1.0E-4_EB) CYCLE L102A

               SDSTRG = SQRT(0.5_EB*G)*(SQRT(SMINUS)+SQRT(SPLUS))/D+SDSTRG

            ENDDO L101A
         ENDDO L102A

         IF(L==2) EXIT L120A
!CONTRIBUTION TO 2.7 MICRON BAND FROM (000)-(101) AND (010)-(111) TRANS.
         ALPHA  = 42.3_EB
         OMPRIM = OM1+OM3
         L      = 2
      ENDDO L120A
!CALCULATE ABSORPTION COEF AND LINE SPACING PARAMETER FOR 4.3 MICRON BAND
      IF(SDWEAK<=TINY_EB) THEN
         SDWEAK = 0.0_EB
         GDINV  = 1.0_EB
         GDDINV = 1.0_EB
      ELSE
         DINV   = SDSTRG**2/SDWEAK
         GDINV  = GC1*DINV
         GDDINV = GD*DINV
!***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
         SDWEAK = SDWEAK*TEMP/273.0_EB
      ENDIF
   ENDIF

ELSEIF((OMEGA<=1975.0_EB).AND.(OMEGA>1100.0_EB)) THEN
   SDWEAK = 0.0_EB
   GDINV  = 1.0_EB
   GDDINV = 1.0_EB

! BAND 10 MICRONS_
ELSEIF((OMEGA<=1100.0_EB).AND.(OMEGA>880.0_EB)) THEN
!CONTRIBUTION TO 10.0 MICRON BAND FROM (100)-(001) AND (020)-(001) TRANS.
   BCNT(1) = 960.8_EB
   BCNT(2) = 1063.6_EB

   OMA = OM3
   OMB = (OM1+2.0_EB*OM2)/2.0_EB

   ATOT(1)= 0.0219_EB
   ATOT(2)= 0.0532_EB

   DO K=1,2
      ATOT(K)=T0/TEMP*ATOT(K)*EXP(Q2*OMB*(1.0_EB/T0-1.0_EB/TEMP))*(1.0_EB-EXP(-Q2*(OMA-OMB)/TEMP)) &
                  /( (1.0_EB-EXP(-Q2*OMA/TEMP))*(1.0_EB-EXP(-OMB*Q2/TEMP)) )
   ENDDO
   SDWEAK = 0.0_EB

   DO I = 1, 2
      SDWEAK = SDWEAK + ATOT(I)*Q2/(4.0_EB*BE*TEMP)*ABS(OMEGA-BCNT(I))*EXP(-Q2/(4.0_EB*BE*TEMP)*(OMEGA-BCNT(I))**2)
   ENDDO

   DINV   = 1.0_EB/(4.0_EB*BE)
   GDINV  = GC1*DINV
   GDDINV = GD*DINV
!***EXPRESS S/D AT STP, AS IS IN NASA SP-3080
   SDWEAK = SDWEAK*TEMP/273.0_EB

! TABLULATED BAND: 15 MICRON
   ELSEIF((OMEGA<=880.0_EB).AND.(OMEGA>500.0_EB))  THEN
!CONTRIBUTION TO 15.0 MICRON BAND FROM (000)-(010) TRANS.
      TTEMP = TEMP
      J  = (FLOOR(OMEGA)-495)/5
      W1 = 495.0_EB + 5.0_EB*REAL(J,EB)
      WW = (OMEGA-W1)/5

      IF(TTEMP>=2400._EB) TTEMP = 2399.99_EB
      IF(TTEMP < 300._EB) TTEMP =  300.00_EB

      I = INT(TTEMP/300._EB)

      IF (I<2) THEN
         T1=REAL(I,EB)*300._EB
         TT=(TTEMP-T1)/300._EB
      ELSE
         IF (TTEMP < 1200) THEN
            I=2
            TT=(TTEMP-600._EB)/600._EB
         ELSE
            IF(I > 5) THEN
               I=5
               TT=(TTEMP-1800._EB)/600._EB
            ELSE
               T1=REAL(I,EB)*300._EB
               TT=(TTEMP-T1)/300._EB
               IF(I>=4) I=I-1
            ENDIF
         ENDIF
      ENDIF

      TW=TT*WW
      SDWEAK=SD15(I,J)*(1._EB-TT-WW+TW)+SD15(I+1,J)*(TT-TW) + SD15(I,J+1)*(WW-TW)+SD15(I+1,J+1)*TW

      IF(SDWEAK<=TINY_EB) THEN
         SDWEAK = 0.0_EB
         GDINV  = 1.0_EB
         GDDINV = 1.0_EB
      ELSE
   !CALCULATE LINE SPACING PARAMETER FOR 15.0 MICRON BAND
         DINV1 =  1.2_EB
         DINV2 =  8.0_EB
         DINV3 = 30.0_EB

         TEMP1 = 300.0_EB
         TEMP2 = 550.0_EB
         TEMP3 = 830.0_EB

         DINV   = DINV1*(TTEMP-TEMP2)*(TTEMP-TEMP3)/(TEMP1-TEMP2) / (TEMP1-TEMP3)+DINV2*(TTEMP-TEMP1)*(TTEMP-TEMP3)&
                  /(TEMP2-TEMP1)/(TEMP2-TEMP3)+DINV3*(TTEMP-TEMP1) * (TTEMP-TEMP2)/(TEMP3-TEMP1)/(TEMP3-TEMP2)
         GDINV  = GC1*DINV
         GDDINV = GD*DINV
      ENDIF

   ELSE
      SDWEAK = 0.0_EB
      GDINV  = 1.0_EB
      GDDINV = 1.0_EB
   ENDIF
ENDIF

!------------------------------------------------------------------------------
RETURN
END SUBROUTINE CO2

!==============================================================================
ELEMENTAL SUBROUTINE H2O(OMEGA,TEMP,GC2,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
!! THIS SUBROUTINE RETURNS THE MEAN SPECTRAL ABSORPTION COEFFICIENT (SDWEAK),
!! THE COLLISION BROADENING FINE STRUCTURE PARAMETER (GDINV),
!! THE DOPPLER BROADENING FINE STRUCTURE PARAMETER (GDDINV), AND THE
!! NARROW BAND MODEL TO BE USED (I_MODEL)
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER
!------------------------------------------------------------------------------

IMPLICIT NONE

! VARIABLES PASSED IN

REAL(EB), INTENT(IN) :: OMEGA ! WAVENUMBER,  IN CM-1
REAL(EB), INTENT(IN) :: TEMP  ! TEMPERATURE, IN KELVIN
REAL(EB), INTENT(IN) :: GC2   ! HALF WIDTH AT HALF HEIGHT, IN CM-1

! VARIABLES PASSED OUT

REAL(EB), INTENT(OUT) :: SDWEAK  ! SPECTRAL ABSORPTION COEFFICIENT, IN CM-1
REAL(EB), INTENT(OUT) :: GDDINV  ! LINE WIDTH TO LINE SPACING RATION FOR DOPPLER
REAL(EB), INTENT(OUT) :: GDINV   ! LINE WIDTH TO LINE SPACING RATIO FOR LORENTZ BORADENING
INTEGER,  INTENT(OUT) :: I_MODEL ! MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCAL VARIABLES

INTEGER  :: I,J
REAL(EB) :: W1, WW, T1, TT, TW, D, B, DINV, TTEMP, GD
REAL(EB), PARAMETER :: WM_H2O = 18._EB

I_MODEL = I_MODEL_H2O

SDWEAK = 0._EB
GDINV  = 1._EB
GDDINV = 1._EB

IF (OMEGA>=50._EB.AND.OMEGA<9300) THEN
! COMPUTE DOPLLER HALF-WIDTH. EQ 5-35
GD   = 5.94E-6_EB*OMEGA*SQRT(TEMP/(273._EB*WM_H2O))

J    = INT((OMEGA-25._EB)/25._EB)
TTEMP= TEMP

IF(TEMP>=2500._EB) TTEMP = 2499.99_EB
IF(TEMP<300._EB)   TTEMP = 300._EB

I = FLOOR(TTEMP/500._EB) + 1

IF (I==2.AND.TTEMP<600._EB) I = 1

W1 = 25._EB+25._EB*REAL(J,EB)
WW = (OMEGA-W1)/25._EB

IF(I>2) THEN
   T1=REAL(I-1,EB)*500._EB
   TT=(TTEMP-T1)/500._EB
ELSE
   IF(I==1) TT = (TTEMP-300._EB)/300._EB
   IF(I==2) TT = (TTEMP-600._EB)/400._EB
ENDIF

TW     = TT*WW
! PERFORM INTERPOLATION
SDWEAK = SD(I,J)*(1._EB-TT-WW+TW)+SD(I+1,J)*(TT-TW)+SD(I,J+1) *(WW-TW)+SD(I+1,J+1)*TW

D      = -2.294_EB+.3004E-02_EB*TTEMP-.366E-06_EB*TEMP**2
B      = DSIN(.0036_EB*OMEGA-8.043_EB)
DINV   = EXP(.7941_EB*B+D) ! FROM NASA IR HANDBOOK)

!     DINV=EXP(0.00106*TEMP-1.21)

! UPDATE ARGUMENT OUT VARIABLES
GDINV  = GC2*DINV
GDDINV = GD*DINV

ENDIF

!------------------------------------------------------------------------------
RETURN
END SUBROUTINE H2O


!==============================================================================
SUBROUTINE CO(OMEGA,TEMP,GC4,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER
INTEGER J
REAL(EB) OMEGA,TEMP,GC4,SDWEAK,GDINV,GDDINV,AA,BB,CC,QQ,EE,FF,GG, &
      SMINUS,SPLUS,SDSTRG,B,ALPHA,A,OME,WX,WY,OMPRIM,T0, &
      Q2,WM,GD,V,GAM,OMV,DELTA,D,OMVBAR,F1,F2,TEST,DINV,Q2OT,TOAZ

INTEGER, INTENT(OUT) :: I_MODEL

I_MODEL = I_MODEL_CO

IF(OMEGA<1600._EB .OR. OMEGA>2400._EB) THEN
SDWEAK = 0._EB
GDINV  = 1._EB
GDDINV = 1._EB
ELSE
B     = 1.93139_EB
ALPHA = 260._EB
A     = .017485_EB
OME   = 2170.21_EB
WX    = 13.461_EB
WY    = .0308_EB
OMPRIM= OME-2._EB*WX+3.25_EB*WY

T0 = 300._EB
Q2 = 1.4388_EB

TOAZ = TEMP/273._EB
Q2OT = Q2/TEMP

WM     = 28._EB
! DOPPLER BROADENING HALF-WIDTH
GD     = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM)
SDWEAK = 1.E-99_EB
SDSTRG = 1.E-99_EB

AA = ALPHA*B*Q2/(A*(1._EB-EXP(-OMPRIM*Q2/T0))**2)
BB = (1._EB-EXP(-OMEGA*Q2OT))*(1._EB-EXP(-OMPRIM*Q2OT))**2
CC = AA*BB*OMEGA*T0/TEMP**2

L101: DO J=1,20
   V    = REAL(J-1,EB)
   QQ   = (V+1._EB)*EXP(-V*OME*Q2OT)
   GAM  = B-A*(V+1._EB)
   OMV  = OME-2._EB*(V+1._EB)*WX+(3._EB*(V+1._EB)*(V+1._EB)+.25_EB)*WY
   DELTA= A*(OMEGA-OMV)

   IF(GAM**2<=DELTA) EXIT L101

   D      = 2._EB*SQRT(GAM*GAM-DELTA)
   OMVBAR = OMV*(1._EB-EXP(-OMV*Q2OT))
   F1     = GAM-0.5_EB*D
   F2     = GAM+0.5_EB*D
   EE     = Q2*GAM/(A*A*TEMP)

   FF     = EXP(EE*DELTA*(1._EB+.5_EB*A/GAM))
   SMINUS = CC*QQ/OMVBAR*ABS(F1)*FF*EXP(-EE*2._EB*GAM*F1)
   SPLUS  = CC*QQ/OMVBAR*ABS(F2)*FF*EXP(-EE*2._EB*GAM*F2)
   GG     = SDWEAK
   SDWEAK = (SMINUS+SPLUS)/D+SDWEAK
   TEST   = (SDWEAK-GG)/SDWEAK

   IF(TEST<.0001_EB) EXIT L101

   SDSTRG=(SQRT(SMINUS)+SQRT(SPLUS))/D+SDSTRG
ENDDO L101

DINV   = SDSTRG*SDSTRG/SDWEAK
GDINV  = GC4*DINV
GDDINV = GD*DINV
!***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
SDWEAK = SDWEAK*TOAZ
ENDIF
!------------------------------------------------------------------------------
END SUBROUTINE CO


!==============================================================================
!ELEMENTAL SUBROUTINE POD(OMEGA,SOOT_VOLUME_FRAC,PATH_LENGTH_CM,X_PARTICLE)
SUBROUTINE POD(OMEGA,SOOT_VOLUME_FRAC,PATH_LENGTH_CM,X_PARTICLE,TEMP)
!==============================================================================
!! POD CALCULATES PARTICLE OPTICAL DEPTH, X_PARTICLE, OF THE VOLUME FRACTION OF
!! SOOT PARTICLES IN GAS CLOUD.  RIN AND RIK ARE
!! THE REAL AND IMAGINARY PARTS OF THE INDEX OF REFRACTION. THE PARTICLES ARE
!! ASSUMED TO BE IN THE RAYLEIGH LIMIT.
!
! ARGUMENT IN:
! OMEGA: WAVENUMBER, UNITS: CM-1
! SOOT_VOLUME_FRAC: SOOT VOLUME FRACTION. NO UNITS
! PATH_LENGTH_CM: PHYSICAL PATH LENGTH. UNITS IN CM
! TEMP: TEMPERATURE IN KELVIN
!
! ARGUMENT OUT:
! X_PARTICLE: OPTICLE DEPTH DUE TO THE PRESENCE OF SOOT PARTICLE. NO UNITS.
!*-----------------------------------------------------------------------------

IMPLICIT NONE

! VARIABLES PASSED IN:

REAL(EB), INTENT(IN) :: OMEGA, PATH_LENGTH_CM, SOOT_VOLUME_FRAC, TEMP

! VARIBLES PASSED OUT: X_PARTICLE: PARTICLE OPTICAL DEPTH

REAL(EB), INTENT(OUT) :: X_PARTICLE

! LOCALS

REAL(EB) :: ABCO,FF,LAMBDA,RIN,RIK

LAMBDA = 10000._EB/OMEGA

RIN = 1.6_EB
RIK = 0.5_EB

!FF = 36.0_EB*PI*RIN*RIK/LAMBDA/((RIN**2-RIK**2+2.0_EB)**2+(2.0_EB*RIN*RIK)**2)

! ABSORPTION COEF. IS BASED UPON MEASUREMENTS OF DALZELL AND
! SAROFIM
IF (AEROSOL_AL2O3) THEN
   IF (TEMP > 2570._EB) THEN
      FF = 0.017_EB
   ELSEIF (TEMP < 500._EB) THEN
      FF = 5.E-6_EB
   ELSE
      FF = 0.00073_EB
   ENDIF
ELSE
   FF         = 7.0_EB/LAMBDA
ENDIF
ABCO       = FF*SOOT_VOLUME_FRAC*1.E6_EB
X_PARTICLE = ABCO*PATH_LENGTH_CM*CM_TO_M

!------------------------------------------------------------------------------
RETURN
END SUBROUTINE POD


!==============================================================================
SUBROUTINE CH4_OLD(OMEGA,TEMP,PCH4,PTOT,GC3,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER
INTEGER I,J
REAL(EB) OMEGA,TEMP,PCH4,PTOT,GC3,SDWEAK,GDINV,GDDINV,BE,Q2, &
   WM,GD,OM1,OM2,OM3,OM4,COM1,COM2,COM3,COM4,DINV,PE,W1,SDB, &
   SDA,SDC,Q2OT,AZOT,TOAZ

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB), DIMENSION(4) :: ATOT, BCNT

I_MODEL = I_MODEL_CH4

IF(OMEGA>5000._EB .OR. OMEGA<1125._EB) THEN

SDWEAK=0.0_EB
GDINV=1._EB
GDDINV=1._EB

ELSE

BE=5.2412_EB
Q2=1.4388_EB
WM=16._EB
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

GD=5.94E-6_EB*OMEGA*SQRT(TOAZ)/4._EB

IF(OMEGA>=3400._EB) THEN

! CONTRIBUTION TO 2.4 MICRON BAND FROM (0000)-(0110), (0000)-(0011),
! (0000)-(1001), AND (0000)-(0102) TRANS.  THE INTEGRATED BAND INTENSITIES
! OF VINCENT-GEISSE (ANNALES DE PHYSIQUE SER.12, V. 10, 1955) HAVE
! BEEN MULTIPLIED BY A FACTOR OF 4 AND THE LINE SPACING IS THAT
! OF V4 FROM GRAY AND PENNER (JQSRT V. 5, 1965).

   OM1=2914.2_EB
   OM2=1526.0_EB
   OM3=3020.3_EB
   OM4=1306.2_EB

   BCNT(1) = 4123.0_EB
   BCNT(2) = 4216.3_EB
   BCNT(3) = 4313.2_EB
   BCNT(4) = 4546.0_EB

   COM1=OM2+2._EB*OM4
   COM2=OM1+OM4
   COM3=OM3+OM4
   COM4=OM2+OM3

   ATOT(1)=0.64_EB*AZOT*(1._EB-EXP(Q2OT*COM1))/((1._EB-EXP(Q2OT*OM2))*(1._EB-EXP(Q2OT*OM4))**2)
   ATOT(2)=17.6_EB*AZOT*(1._EB-EXP(Q2OT*COM2))/((1._EB-EXP(Q2OT*OM1))*(1._EB-EXP(Q2OT*OM4)))
   ATOT(3)=14.8_EB*AZOT*(1._EB-EXP(Q2OT*COM3))/((1._EB-EXP(Q2OT*OM3))*(1._EB-EXP(Q2OT*OM4)))
   ATOT(4)=5.04_EB*AZOT*(1._EB-EXP(Q2OT*COM4))/((1._EB-EXP(Q2OT*OM2))*(1._EB-EXP(Q2OT*OM3)))

   DINV=1._EB/5.74_EB
   GDINV=GC3*DINV
   GDDINV=GD*DINV
   SDWEAK=0.0_EB

   DO I=1,4
      SDWEAK=SDWEAK+2._EB*(OMEGA-BCNT(I))**2*(-Q2OT*BE)**1.5_EB*ATOT(I)/SQRTPI*DINV**3*EXP(Q2OT*BE*DINV**2 &
      *(OMEGA-BCNT(I))**2)
   ENDDO
   SDWEAK=SDWEAK*TOAZ
ELSE
   PE=PTOT+.3_EB*PCH4

   IF(OMEGA>=2625._EB) THEN
! CONTRIBUTION TO 3.3 MICRON BAND FROM (0000)-(0010) TRANS.
! REFER TO BROSMER AND TIEN, JQSRT V. 33, P. 521

      GDINV  = .00734_EB*PE*SQRT(AZOT)*EXP(1.02_EB*(TOAZ-1._EB))
      GDDINV = GD/9.4_EB

      J  = INT((OMEGA-2600._EB)/25._EB)
      W1 = 2600._EB+25._EB*REAL(J,EB)
      SDB= SD3(2,J)+(OMEGA-W1)/25._EB*(SD3(2,J+1)-SD3(2,J))

      IF(TEMP>600._EB) THEN
         SDC=SD3(3,J)+(OMEGA-W1)/25._EB*(SD3(3,J+1)-SD3(3,J))
         SDWEAK=SDB+(TEMP-600._EB)/250._EB*(SDC-SDB)
         IF(SDWEAK<0._EB)SDWEAK=0._EB
      ELSE
         SDA=SD3(1,J)+(OMEGA-W1)/25._EB*(SD3(1,J+1)-SD3(1,J))
         SDWEAK=SDA+(TEMP-290._EB)/310._EB*(SDB-SDA)
         IF(SDWEAK<0._EB)SDWEAK=0._EB
      ENDIF

   ELSEIF(OMEGA>1450._EB) THEN

      SDWEAK=0.0_EB
      GDINV=1._EB
      GDDINV=1._EB

   ELSE
! CONTRIBUTION TO 7.7 MICRON BAND FROM (0000)-(0001) TRANS.
! REFER TO BROSMER AND TIEN, JQSRT V. 33, P. 521.
      GDINV  = .0243_EB*PE*(TOAZ)**.8_EB
      GDDINV = GD/5.1_EB

      J   = INT((OMEGA-1100._EB)/25._EB)
      W1  = 1100._EB+25._EB*REAL(J,EB)
      SDB = SD7(2,J)+(OMEGA-W1)/25._EB*(SD7(2,J+1)-SD7(2,J))

      IF(TEMP>600._EB) THEN
         SDC = SD7(3,J)+(OMEGA-W1)/25._EB*(SD7(3,J+1)-SD7(3,J))
         SDWEAK = SDB+(TEMP-600._EB)/250._EB*(SDC-SDB)
         IF(SDWEAK<0._EB)SDWEAK=0._EB
      ELSE
         SDA = SD7(1,J)+(OMEGA-W1)/25._EB*(SD7(1,J+1)-SD7(1,J))
         SDWEAK = SDA+(TEMP-290._EB)/310._EB*(SDB-SDA)
         IF(SDWEAK<0._EB)SDWEAK=0._EB
      ENDIF
   ENDIF
ENDIF
ENDIF
!------------------------------------------------------------------------------
END SUBROUTINE CH4_OLD


!==============================================================================
SUBROUTINE CH4(OMEGA,TEMP,PCH4,PTOT,GC3,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
!! COMPUTE METHANE OPTICAL PROPERTIES USING SINGLE LINE GROUP
!!
!! BAND #1: 1150 CM-1 - 1600 CM-1
!! BAND #2: 2700 CM-1 - 3250 CM-1
!! BAND #3: 3400 CM-1 - 5000 CM-1
!
! VARIABLES PASSED IN
! OMEGA: (REAL) WAVENUMBER IN CM-1
! TEMP : (REAL) TEMPERATURE IN K
! PCH4 : (REAL) PARTIAL PRESSURE OF METHANE IN ATM
! PTOT : (REAL) TOTAL PRESSURE IN ATM
! GC3  : (REAL) COLLISIONAL BROADENING HALF-WIDTH AT HALF HEIGH
!
! VRAIBLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER
! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! BE_CH4             : (REAL) ROTATIONAL CONSTANT FOR CH4 [CM^-1]
! DINV_CH4           : INVERSE LINE SPACING [CM]

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PCH4, PTOT, GC3
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT)  :: I_MODEL

REAL(EB) :: GD, PRESSURE_EFFECTIVE, Q2OT, AZOT, TOAZ, FACT1

REAL(EB), PARAMETER :: Q2     = 1.4388_EB ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_CH4 = 16.0425_EB

!LINE SPACING D2_CH4=5.74 CM^-1 FOR THE V4 FUNDAMENTAL OF METHANE GRAY & PENNER,612
!FOOTNOTE AT BOTTOM
REAL(EB), PARAMETER :: BE_CH4 = 5.248_EB   ! ROTATIONAL CONSTANTS [CM^-1]
REAL(EB), DIMENSION(3), PARAMETER :: DINV_CH4 = (/1._EB/5.1_EB,1._EB/9.4_EB,1._EB/5.74_EB/)

REAL(EB), DIMENSION(4), PARAMETER :: &
               OM_CH4  = (/2914.2_EB, 1526.0_EB, 3020.3_EB, 1306.2_EB/),                                               &
               COM_CH4 = (/1526.0_EB+2._EB*1306.2_EB, 2914.2_EB+1306.2_EB, 3020.3_EB+1306.2_EB, 1526.0_EB+3020.3_EB/), &
               S2_CH4  = (/0.64_EB,17.6_EB,14.8_EB,5.04_EB/)

REAL(EB), DIMENSION(4) :: ATOT

INTEGER I

I_MODEL = I_MODEL_CH4

! INITIALIZE OUTPUT

SDWEAK = 0.0_EB  !SPECTRAL ABSORPTION COEFFICIENT.
GDINV  = 1._EB   !LINE WIDTH TO LINE SPACING RATIO: FINE STRUCTURE PARAMETER
GDDINV = 1._EB   !LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING FINE STRUCTURE PARAMETER

! INITIALIZE SOME COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPLLER HALF-WIDTH. EQ 5-35
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_CH4) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE EFFECTIVE PRESSURE. BROSMER & TIEN,524 EQ. 7 (SPECIFIC TO METHANE)
PRESSURE_EFFECTIVE = PTOT+.3_EB*PCH4

!------------------------------------------------------------------------------
! COMPUTED PROPERTIES 2.4 MICRON BAND, BAND #3: 3400 CM-1 - 5000 CM-1
IF((OM_BND_CH4(N_BAND_CH4,1)<=OMEGA).AND.(OMEGA<=OM_BND_CH4(N_BAND_CH4,2))) THEN

   ! CONTRIBUTION OF THE V1+V4 COMBINAISON BAND (2.4 MICRON)
   ! CONTRIBUTION TO 2.4 MICRON BAND FROM
   ! (0000)-(0102)
   ! (0000)-(1001)
   ! (0000)-(0011)
   ! (0000)-(0110) TRANSITIONS
   ! THE INTEGRATED BAND INTENSITIES (S2_CH4)
   ! OF VINCENT-GEISSE (ANNALES DE PHYSIQUE SER.12, V. 10, 1955) HAVE
   ! BEEN MULTIPLIED BY A FACTOR OF 4
   ! THE LINE SPACING (1/DINV_CH4) IS THAT
   ! OF V4 FROM GRAY AND PENNER (JQSRT V. 5, 1965).

   ATOT(1) = S2_CH4(1)*AZOT*(1._EB-EXP(Q2OT*COM_CH4(1)))/((1._EB-EXP(Q2OT*OM_CH4(2)))*(1._EB-EXP(Q2OT*OM_CH4(4)))**2)
   ATOT(2) = S2_CH4(2)*AZOT*(1._EB-EXP(Q2OT*COM_CH4(2)))/((1._EB-EXP(Q2OT*OM_CH4(1)))*(1._EB-EXP(Q2OT*OM_CH4(4))))
   ATOT(3) = S2_CH4(3)*AZOT*(1._EB-EXP(Q2OT*COM_CH4(3)))/((1._EB-EXP(Q2OT*OM_CH4(3)))*(1._EB-EXP(Q2OT*OM_CH4(4))))
   ATOT(4) = S2_CH4(4)*AZOT*(1._EB-EXP(Q2OT*COM_CH4(4)))/((1._EB-EXP(Q2OT*OM_CH4(2)))*(1._EB-EXP(Q2OT*OM_CH4(3))))


   FACT1  = Q2OT*BE_CH4*DINV_CH4(3)**2

   DO I=1,4
      SDWEAK = SDWEAK+(OMEGA-COM_CH4(I))**2*ATOT(I)*EXP(FACT1*(OMEGA-COM_CH4(I))**2)
   ENDDO

   ! EQ 6 IN GRAY AND PENNER JQSRT, VOL 5, PAGE 611-620, 1965
   SDWEAK=SDWEAK*2._EB*(-Q2OT*BE_CH4)**1.5_EB/SQRTPI*DINV_CH4(3)**3

   !***EXPRESS S/D AT STANDARD TEMPERATURE AND PRESSURE, AS IS IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ

   GDINV  = GC3*DINV_CH4(3)
   GDDINV = GD*DINV_CH4(3)

!------------------------------------------------------------------------------
! TABULATED PROPERTIES
ELSE IF((OM_BND_CH4(2,1)<=OMEGA).AND.(OMEGA<OM_BND_CH4(2,2))) THEN

   !------------------------------------------------------------------------------
   ! CONTRIBUTION TO 3.3 MICRON BAND, BAND #2: 2700 CM-1 - 3250 CM-1
   ! (0000)-(0010) TRANSITION (V3 FUNDAMENTAL)
   ! 9.4 IS THE AVERAGE LINE POSITION FOR THE 3.3 MICRON BAND (CM^-1)
   ! SOURCE: BROSMER & TIEN, JQSRT V. 33, P. 525 (SPECIFIC TO METHANE)

   SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH4_TEMP,OM_BND_CH4(2,:),SD2_CH4)
   ! LINE SHAPE PARAMETER BROSMER & TIEN, JQSRT V. 33, P. 525 EQ. 10 (SPECIFIC TO METHANE)
   GDINV  = .00734_EB*PRESSURE_EFFECTIVE*SQRT(AZOT)*EXP(1.02_EB*(TOAZ-1._EB))
   GDDINV = GD*DINV_CH4(2)
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ

ELSE IF((OM_BND_CH4(1,1)<=OMEGA).AND.(OMEGA<OM_BND_CH4(1,2))) THEN
   !------------------------------------------------------------------------------
   ! CONTRIBUTION TO 7.7 MICRON BAND, BAND #1: 1150 CM-1 - 1600 CM-1
   ! (0000)-(0001) TRANSITION (V4 FUNDAMENTAL)
   ! 5.1 (CM-1) IS THE AVERAGE LINE POSITION FOR THE 7.7 MICRON BAND
   ! SOURCE: BROSMER & TIEN, JQSRT V. 33, P. 525 (SPECIFIC TO METHANE)

   SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH4_TEMP,OM_BND_CH4(1,:),SD1_CH4)
   ! LINE SHAPE PARAMETER BROSMER & TIEN, JQSRT V. 33, 525 EQ. 11 (SPECIFIC TO METHANE)
   GDINV  = .0243_EB*PRESSURE_EFFECTIVE*(TOAZ)**.8_EB
   GDDINV = GD*DINV_CH4(1)
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

!------------------------------------------------------------------------------
END SUBROUTINE CH4

!==============================================================================
SUBROUTINE C3H6(OMEGA,TEMP,PC3H6,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
!! COMPUTE PROPYLENE OPTICAL PROPERTIES USING SINGLE LINE GROUP
!!
!! ABSORPTION BANDS:
!! THERE ARE 3 BANDS FOR PROPYLENE
!!
!! BAND #1: 1250 CM-1 - 1950 CM-1
!! BAND #2: 2700 CM-1 - 3200 CM-1
!! BAND #3: 700 CM-1 - 1150 CM-1
!
! VARIABLES PASSED IN
! OMEGA  : (REAL) WAVENUMBER IN CM-1
! TEMP   : (REAL) TEMPERATURE IN K
! PC3H6 : (REAL) PARTIAL PRESSURE OF PROPYLENE
! PTOT   : (REAL) TOTAL PRESSURE
! GC3    : (REAL) COLLISIONAL BROADENING HALF-WIDTH AT HALF HEIGH
!
! VRAIBLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_C3H6         : INVERSE LINE SPACING [CM] FOR PROPYLENE

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PC3H6, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_C3H6

REAL(EB), PARAMETER :: Q2       = 1.4388_EB  ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_C3H6  = 42.0797_EB ! NIST WEBBOOK DATA

INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_C3H6
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_C3H6) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE AVERAGE LINE SPACING
! MODEL PROPYLENE AS A PROLATE SYMMETRIC TOP. THE LINE SPACING IS APPROXIMATED TO
! 2*ROTATIONAL_CONSTANT B. RECALL THAT ROTATIONAL_CONSTANT B = ROTATIONAL_CONSTANT C
! ROTATIONAL_CONSTANT_B OBTAINED FROM CCCBDV WEBSITE, USING EXPERIMENTAL VALUES
! PROPYLENE BELONGS TO GROUP CS
DINV_C3H6 = 1._EB/(2._EB*0.312_EB)

! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! PROPYLENE HAS THREE BANDS
! BAND #1: 775 CM-1 - 1150 CM-1
! BAND #2: 1225 CM-1 - 1975 CM-1
! BAND #3: 2650 CM-1 - 3275 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_C3H6
   IF ((OM_BND_C3H6(I,1)<=OMEGA).AND.(OMEGA<OM_BND_C3H6(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND

IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_C3H6(I_BAND)-1.0_EB)*PC3H6

   SELECT CASE (I_BAND)
      CASE (1) ! BAND #1: 775 CM-1 - 1150 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H6_TEMP,OM_BND_C3H6(I_BAND,:) ,SD1_C3H6)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H6_TEMP,OM_BND_C3H6(I_BAND,:) ,GAMMAD1_C3H6)
      CASE (2) ! BAND #2: 1225 CM-1 - 1975 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H6_TEMP,OM_BND_C3H6(I_BAND,:) ,SD2_C3H6)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H6_TEMP,OM_BND_C3H6(I_BAND,:) ,GAMMAD2_C3H6)
      CASE (3) ! BAND #3: 2650 CM-1 - 3275 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H6_TEMP,OM_BND_C3H6(I_BAND,:) ,SD3_C3H6)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H6_TEMP,OM_BND_C3H6(I_BAND,:) ,GAMMAD3_C3H6)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_C3H6
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE C3H6

!==============================================================================
SUBROUTINE C3H8(OMEGA,TEMP,PC3H8,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! COMPUTE PROPANE OPTICAL PROPERTIES USING SINGLE LINE GROUP
!
! ABSORPTION BANDS:
! THERE ARE 2 BANDS FOR PROPANE

! BAND #1: 1175 CM-1 - 1675 CM-1
! BAND #2: 2550 CM-1 - 3375 CM-1
!
! VARIABLES PASSED IN
! OMEGA  : (REAL) WAVENUMBER IN CM-1
! TEMP   : (REAL) TEMPERATURE IN K
! PC3H8 : (REAL) PARTIAL PRESSURE OF PROPANE
! PTOT   : (REAL) TOTAL PRESSURE
! GC3    : (REAL) COLLISIONAL BROADENING HALF-WIDTH AT HALF HEIGH
!
! VRAIBLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_C3H8         : INVERSE LINE SPACING [CM] FOR PROPANE

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PC3H8, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_C3H8

REAL(EB), PARAMETER :: Q2       = 1.4388_EB  ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_C3H8  = 44.0956_EB ! NIST WEBBOOK DATA

INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_C3H8
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_C3H8) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE AVERAGE LINE SPACING
! PROPANE IS A PROLATE SYMMETRIC TOP. THE LINE SPACING IS APPROXIMATED TO
! 2*ROTATIONAL_CONSTANT B. RECALL THAT ROTATIONAL_CONSTANT B = ROTATIONAL_CONSTANT C
! ROTATIONAL_CONSTANT_B OBTAINED FROM CCCBDV WEBSITE, USING EXPERIMENTAL VALUES
! PROPANE BELONGS TO GROUP C2V
DINV_C3H8 = 1._EB/(2._EB*0.28173_EB)

! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! PROPANE HAS THREE BANDS
! BAND #1: 1175 CM-1 - 1675 CM-1
! BAND #2: 2550 CM-1 - 3375 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_C3H8
   IF ((OM_BND_C3H8(I,1)<=OMEGA).AND.(OMEGA<OM_BND_C3H8(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND

IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_C3H8(I_BAND)-1.0_EB)*PC3H8

   SELECT CASE (I_BAND)
      CASE (1) ! BAND #1: 1175 CM-1 - 1675 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H8_TEMP,OM_BND_C3H8(I_BAND,:) ,SD1_C3H8)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H8_TEMP,OM_BND_C3H8(I_BAND,:) ,GAMMAD1_C3H8)
      CASE (2) ! BAND #2: 2550 CM-1 - 3375 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H8_TEMP,OM_BND_C3H8(I_BAND,:) ,SD2_C3H8)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C3H8_TEMP,OM_BND_C3H8(I_BAND,:) ,GAMMAD2_C3H8)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_C3H8
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE C3H8

!==============================================================================
SUBROUTINE C7H16(OMEGA,TEMP,PC7H16,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! COMPUTE HEPTANE OPTICAL PROPERTIES USING SINGLE LINE GROUP
!
! ABSORPTION BANDS:
! THERE ARE 2 BANDS FOR HEPTANE

! BAND #1: 1100 CM-1 - 1800 CM-1
! BAND #2: 2550 CM-1 - 3275 CM-1
!
! VARIABLES PASSED IN
! OMEGA  : (REAL) WAVENUMBER IN CM-1
! TEMP   : (REAL) TEMPERATURE IN K
! PC7H16 : (REAL) PARTIAL PRESSURE OF HEPTANE
! PTOT   : (REAL) TOTAL PRESSURE
! GC3    : (REAL) COLLISIONAL BROADENING HALF-WIDTH AT HALF HEIGH
!
! VRAIBLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_C7H16         : INVERSE LINE SPACING [CM] FOR HEPTANE

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PC7H16, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_C7H16

REAL(EB), PARAMETER :: Q2       = 1.4388_EB  ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_C7H16 = 100.2019_EB ! NIST WEBBOOK DATA

INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_C7H16
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_C7H16) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE AVERAGE LINE SPACING
! N_HEPTANE IS A PROLATE SYMMETRIC TOP. THE LINE SPACING IS APPROXIMATED TO
! 2*ROTATIONAL_CONSTANT B. RECALL THAT ROTATIONAL_CONSTANT B = ROTATIONAL_CONSTANT C
! ROTATIONAL_CONSTANT_B OBTAINED FROM CCCBDV WEBSITE, USING EXPERIMENTAL VALUES
! N-HEPTANE BELONGS TO GROUP C2V
DINV_C7H16=1._EB/(2.0_EB*0.024_EB)

! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! HEPTANE HAS THREE BANDS
! BAND #1: 1100 CM-1 - 1800 CM-1
! BAND #2: 2550 CM-1 - 3275 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_C7H16
   IF ((OM_BND_C7H16(I,1)<=OMEGA).AND.(OMEGA<OM_BND_C7H16(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND

IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_C7H16(I_BAND)-1.0_EB)*PC7H16

   SELECT CASE (I_BAND)
      CASE (1) ! BAND #1: 1175 CM-1 - 1675 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H16_TEMP,OM_BND_C7H16(I_BAND,:),SD1_C7H16)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H16_TEMP,OM_BND_C7H16(I_BAND,:),GAMMAD1_C7H16)
      CASE (2) ! BAND #2: 2550 CM-1 - 3375 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H16_TEMP,OM_BND_C7H16(I_BAND,:),SD2_C7H16)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H16_TEMP,OM_BND_C7H16(I_BAND,:),GAMMAD2_C7H16)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_C7H16
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE C7H16

!==============================================================================
SUBROUTINE C7H8(OMEGA,TEMP,PC7H8,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! COMPUTE TOLUENE OPTICAL PROPERTIES USING SINGLE LINE GROUP
!
! ABSORPTION BANDS:
! THERE ARE 5 BANDS FOR TOLUENE

! BAND #1: 700 CM-1 - 825 CM-1
! BAND #2: 975 CM-1 - 1175 CM-1
! BAND #3: 1275 CM-1 - 1675 CM-1
! BAND #4: 1650 CM-1 - 2075 CM-1
! BAND #5: 2675 CM-1 - 3225 CM-1
!
! VARIABLES PASSED IN
! OMEGA  : (REAL) WAVENUMBER IN CM-1
! TEMP   : (REAL) TEMPERATURE IN K
! PC7H8 : (REAL) PARTIAL PRESSURE OF TOLUENE
! PTOT   : (REAL) TOTAL PRESSURE
! GC3    : (REAL) COLLISIONAL BROADENING HALF-WIDTH AT HALF HEIGH
!
! VRAIBLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_C7H8         : INVERSE LINE SPACING [CM] FOR TOLUENE

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PC7H8, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_C7H8

REAL(EB), PARAMETER :: Q2       = 1.4388_EB  ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_C7H8  = 92.1384_EB ! NIST WEBBOOK DATA

INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_C7H8
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_C7H8) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! TOLUENE BELONGS TO GROUP CS
! TOLUENE IS MODEL AS A PROLATE SYMMETRIC TOP. THE LINE SPACING IS APPROXIMATED TO
! 2*ROTATIONAL_CONSTANT B. RECALL THAT ROTATIONAL_CONSTANT B = ROTATIONAL_CONSTANT C
! ROTATIONAL_CONSTANT_B OBTAINED FROM CCCBDV WEBSITE.  USING EXPERIMENTAL VALUES

DINV_C7H8=1._EB/(2.0_EB*0.084_EB)


! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! TOLUENE HAS 5 BANDS
! BAND #1: 700 CM-1 - 825 CM-1
! BAND #2: 975 CM-1 - 1175 CM-1
! BAND #3: 1275 CM-1 - 1675 CM-1
! BAND #4: 1650 CM-1 - 2075 CM-1
! BAND #5: 2675 CM-1 - 3225 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_C7H8
   IF ((OM_BND_C7H8(I,1)<=OMEGA).AND.(OMEGA<OM_BND_C7H8(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND


IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_C7H8(I_BAND)-1.0_EB)*PC7H8

   SELECT CASE (I_BAND)
      CASE (1) ! BAND #1: 700 CM-1 - 825 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),SD1_C7H8)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),GAMMAD1_C7H8)
      CASE (2) ! BAND #2: 975 CM-1 - 1175 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),SD2_C7H8)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),GAMMAD2_C7H8)
      CASE (3) ! BAND #3: 1275 CM-1 - 1675 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),SD3_C7H8)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),GAMMAD3_C7H8)
      CASE (4) ! BAND #4: 1650 CM-1 - 2075 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),SD4_C7H8)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),GAMMAD4_C7H8)
      CASE (5) ! BAND #5: 2675 CM-1 - 3225 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),SD5_C7H8)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C7H8_TEMP,OM_BND_C7H8(I_BAND,:),GAMMAD5_C7H8)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_C7H8
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE C7H8


!==============================================================================
SUBROUTINE CH3OH(OMEGA,TEMP,PCH3OH,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! COMPUTE METHANOL OPTICAL PROPERTIES USING SINGLE LINE GROUP
!
! ABSORPTION BANDS:
! THERE ARE 4 BANDS FOR METHANOL

! BAND #1: 825 CM-1 - 1150 CM-1
! BAND #2: 1125 CM-1 - 1700 CM-1
! BAND #3: 2600 CM-1 - 3225 CM-1
! BAND #4: 3525 CM-1 - 3850 CM-1
!
! VARIABLES PASSED IN
! OMEGA  : (REAL) WAVENUMBER IN CM-1
! TEMP   : (REAL) TEMPERATURE IN K
! PCH3OH : (REAL) PARTIAL PRESSURE OF METHANOL
! PTOT   : (REAL) TOTAL PRESSURE
!
! VARIABLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_CH3OH         : INVERSE LINE SPACING [CM] FOR METHANOL

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PCH3OH, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_CH3OH

REAL(EB), PARAMETER :: Q2       = 1.4388_EB  ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_CH3OH = 32.0419_EB ! NIST WEBBOOK DATA
INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_CH3OH
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_CH3OH) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE AVERAGE LINE SPACING
! METHANOL BELONGS TO POINT GROUP CS. FOR NOW THE LINE SPACING IS APPROXIMATED TO
! 2*ROTATIONAL_CONSTANT B SINCE ROTATIONAL_CONSTANT B ~ ROTATIONAL_CONSTANT C
! ROTATIONAL_CONSTANT_B OBTAINED FROM CCCBDV WEBSITE, USING EXPERIMENTAL VALUES
DINV_CH3OH=1._EB/(2.0_EB*0.82338_EB)

! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! METHANOL HAS 4 BANDS
! BAND #1: 825 CM-1 - 1150 CM-1
! BAND #2: 1125 CM-1 - 1700 CM-1
! BAND #3: 2600 CM-1 - 3225 CM-1
! BAND #4: 3525 CM-1 - 3850 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_CH3OH
   IF ((OM_BND_CH3OH(I,1)<=OMEGA).AND.(OMEGA<OM_BND_CH3OH(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND


IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_CH3OH(I_BAND)-1.0_EB)*PCH3OH

   SELECT CASE (I_BAND)
      CASE (1) ! BAND #1: 825 CM-1 - 1150 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,SD1_CH3OH)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,GAMMAD1_CH3OH)
      CASE (2) ! BAND #2: 1125 CM-1 - 1700 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,SD2_CH3OH)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,GAMMAD2_CH3OH)
      CASE (3) ! BAND #3: 2600 CM-1 - 3225 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,SD3_CH3OH)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,GAMMAD3_CH3OH)
      CASE (4) ! BAND #4: 3525 CM-1 - 3850 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,SD4_CH3OH)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_CH3OH_TEMP,OM_BND_CH3OH(I_BAND,:) ,GAMMAD4_CH3OH)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_CH3OH
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE CH3OH

!==============================================================================
SUBROUTINE C5H8O2(OMEGA,TEMP,PC5H8O2,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! COMPUTE MMA OPTICAL PROPERTIES USING SINGLE LINE GROUP
!
! ABSORPTION BANDS:
! THERE ARE 6 BANDS FOR MMA
! BAND #1: 750 CM-1 - 900 CM-1
! BAND #2: 875 CM-1 - 1075 CM-1
! BAND #3: 1050 CM-1 - 1275 CM-1
! BAND #4: 1250 CM-1 - 1575 CM-1
! BAND #5: 1550 CM-1 - 1975 CM-1
! BAND #6: 2650 CM-1 - 3275 CM-1
!
! VARIABLES PASSED IN
! OMEGA   : (REAL) WAVENUMBER IN CM-1
! TEMP    : (REAL) TEMPERATURE IN K
! PC5H8O2 : (REAL) PARTIAL PRESSURE OF MMA
! PTOT    : (REAL) TOTAL PRESSURE
!
! VARIABLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_C5H8O2         : INVERSE LINE SPACING [CM] FOR MMA

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PC5H8O2, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_C5H8O2

REAL(EB), PARAMETER :: Q2        = 1.4388_EB   ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_C5H8O2 = 100.1158_EB ! NIST WEBBOOK DATA
INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_C5H8O2
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_C5H8O2) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE AVERAGE LINE SPACING
! MMA BELONGS TO POINT GROUP ? FOR NOW USING DATA FROM ACETYLACETONE (C5H8O2)
DINV_C5H8O2=1._EB/(2.0_EB*0.067_EB)

! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! THERE ARE 6 BANDS FOR MMA
! BAND #1: 750 CM-1 - 900 CM-1
! BAND #2: 875 CM-1 - 1075 CM-1
! BAND #3: 1050 CM-1 - 1275 CM-1
! BAND #4: 1250 CM-1 - 1575 CM-1
! BAND #5: 1550 CM-1 - 1975 CM-1
! BAND #6: 2650 CM-1 - 3275 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_C5H8O2
   IF ((OM_BND_C5H8O2(I,1)<=OMEGA).AND.(OMEGA<OM_BND_C5H8O2(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND

IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_C5H8O2(I_BAND)-1.0_EB)*PC5H8O2

   SELECT CASE (I_BAND)
      CASE (1) ! BAND #1: 750 CM-1 - 900 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),SD1_C5H8O2)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),GAMMAD1_C5H8O2)
      CASE (2) ! BAND #2: 875 CM-1 - 1075 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),SD2_C5H8O2)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),GAMMAD2_C5H8O2)
      CASE (3) ! BAND #3: 1050 CM-1 - 1275 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),SD3_C5H8O2)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),GAMMAD3_C5H8O2)
      CASE (4) ! BAND #4: 1250 CM-1 - 1575 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),SD4_C5H8O2)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),GAMMAD4_C5H8O2)
      CASE (5) ! BAND #5: 1550 CM-1 - 1975 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),SD5_C5H8O2)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),GAMMAD5_C5H8O2)
      CASE (6) ! BAND #6: 2650 CM-1 - 3275 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),SD6_C5H8O2)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C5H8O2_TEMP,OM_BND_C5H8O2(I_BAND,:),GAMMAD6_C5H8O2)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_C5H8O2
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE C5H8O2


!==============================================================================
SUBROUTINE C2H6(OMEGA,TEMP,PC2H6,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! COMPUTE ETHANE OPTICAL PROPERTIES USING SINGLE LINE GROUP
!
! ABSORPTION BANDS:
! THERE ARE 3 BANDS FOR ETHANE
! BAND #1: 750 CM-1 - 1100 CM-1
! BAND #2: 1250 CM-1 - 1700 CM-1
! BAND #3: 2550 CM-1 - 3375 CM-1

!
! VARIABLES PASSED IN
! OMEGA   : (REAL) WAVENUMBER IN CM-1
! TEMP    : (REAL) TEMPERATURE IN K
! PC2H6 : (REAL) PARTIAL PRESSURE OF ETHANE
! PTOT    : (REAL) TOTAL PRESSURE
!
! VARIABLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_C2H6         : INVERSE LINE SPACING [CM] FOR ETHANE

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PC2H6, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_C2H6

REAL(EB), PARAMETER :: Q2      = 1.4388_EB  ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_C2H6 = 30.0690_EB ! NIST WEBBOOK DATA
INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_C2H6
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_C2H6) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE AVERAGE LINE SPACING
! ETHANE BELONGS TO POINT GROUP D3D WITH ROTATIONAL CONSTANT A = 2.51967 CM-1
! B = 0.68341 CM-1, AND C = 0.68341 CM-1.
DINV_C2H6=1._EB/(2.0_EB*0.68341_EB)

! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! THERE ARE 3 BANDS FOR ETHANE
! BAND #1: 750 CM-1 - 1100 CM-1
! BAND #2: 1250 CM-1 - 1700 CM-1
! BAND #3: 2550 CM-1 - 3375 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_C2H6
   IF ((OM_BND_C2H6(I,1)<=OMEGA).AND.(OMEGA<OM_BND_C2H6(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND

IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_C2H6(I_BAND)-1.0_EB)*PC2H6

   SELECT CASE (I_BAND)
      CASE (1)  ! BAND #1:  750 CM-1 - 1100 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H6_TEMP,OM_BND_C2H6(I_BAND,:),SD1_C2H6)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H6_TEMP,OM_BND_C2H6(I_BAND,:),GAMMAD1_C2H6)
      CASE (2) ! BAND #2: 1250 CM-1 - 1700 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H6_TEMP,OM_BND_C2H6(I_BAND,:),SD2_C2H6)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H6_TEMP,OM_BND_C2H6(I_BAND,:),GAMMAD2_C2H6)
      CASE (3) ! BAND #3: 2550 CM-1 - 3375 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H6_TEMP,OM_BND_C2H6(I_BAND,:),SD3_C2H6)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H6_TEMP,OM_BND_C2H6(I_BAND,:),GAMMAD3_C2H6)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_C2H6
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE C2H6


!==============================================================================
SUBROUTINE C2H4(OMEGA,TEMP,PC2H4,PTOT,SDWEAK,GDINV,GDDINV,I_MODEL)
!==============================================================================
! COMPUTE ETHYLENE OPTICAL PROPERTIES USING SINGLE LINE GROUP
!
! ABSORPTION BANDS:
! THERE ARE 4 BANDS FOR ETHYLENE
! BAND #1: 800 CM-1  - 1250 CM-1
! BAND #2: 1325 CM-1 - 1575 CM-1
! BAND #3: 1775 CM-1 - 2050 CM-1
! BAND #4: 2825 CM-1 - 3350 CM-1
!
! VARIABLES PASSED IN
! OMEGA   : (REAL) WAVENUMBER IN CM-1
! TEMP    : (REAL) TEMPERATURE IN K
! PC2H4 : (REAL) PARTIAL PRESSURE OF ETHYLENE
! PTOT    : (REAL) TOTAL PRESSURE
!
! VARIABLES PASSED OUT
! SDWEAK : (REAL) SPECTRAL ABSORPTION COEFFICIENT
! GDINV  : (REAL) LINE WIDTH TO LINE SPACING RATIO, FINE STRUCTURE PARAMETER
! GDDINV : (REAL) LINE WIDTH TO LINE SPACING RATIO FOR DOPPLER BORADENING,
!                 FINE STRUCTURE PARAMETER
! I_MODEL: (INTEGER) MODEL USED FOR SNB FITTING: 1: GOODY, 2 MALKMUS, 3 ELSASSER

! LOCALS
! PRESSURE_EFFECTIVE : (REAL) EFFECTIVE PRESSURE, (FORMELY PE)
! DINV_C2H4          : INVERSE LINE SPACING [CM] FOR ETHYLENE

REAL(EB), INTENT(IN)  :: OMEGA, TEMP, PC2H4, PTOT
REAL(EB), INTENT(OUT) :: SDWEAK, GDINV, GDDINV

INTEGER, INTENT(OUT) :: I_MODEL

REAL(EB) :: Q2OT, AZOT, TOAZ
REAL(EB) :: GD, PRESSURE_EFFECTIVE
REAL(EB) :: DINV_C2H4

REAL(EB), PARAMETER :: Q2      = 1.4388_EB  ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL(EB), PARAMETER :: WM_C2H4 = 28.0532_EB ! NIST WEBBOOK DATA
INTEGER :: I_BAND, I

LOGICAL :: IN_BAND ! TRUE IF OMEGA IS WITHIN SOME TABULATED BAND, FALSE OTHERWISE

! GET MODEL FOR SNB FITTING

I_MODEL = I_MODEL_C2H4
I_BAND  = 0
IN_BAND = .FALSE.

! SET INITIAL VALUES TO SDWEAK, GDINV, GDDINV - VALUES ARE RETURNED IF OMEGA IS NOT
! WITHIN ABSORPTION BANDS
SDWEAK = 0.0_EB
GDINV  = 1._EB
GDDINV = 1._EB

! COMPUTE THERMAL COEFFICIENTS
Q2OT = -Q2/TEMP
AZOT = 273._EB/TEMP
TOAZ = TEMP/273._EB

! COMPUTE DOPPLER BROADENING HALF WIDTH GD
GD   = 5.94E-6_EB*OMEGA*SQRT(TOAZ/WM_C2H4) !DOPPLER HALF WIDTH [CM^-1]. NASA,222

! COMPUTE AVERAGE LINE SPACING
! ETHYLENE BELONGS TO POINT GROUP D2H WITH ROTATIONAL CONSTANT A = 4.828 CM-1
! B = 1.0012 CM-1, AND C = 0.8282 CM-1. WE MAKE THE ASSUMPTION B ~ C
DINV_C2H4=1._EB/(2.0_EB*1.0012_EB)

! IF OMEGA IS WITHIN ABSORPTION BAND, THEN COMPUTE SDWEAK FROM TABULATED DATA
! THERE ARE 4 BANDS FOR ETHYLENE
! BAND #1: 800 CM-1  - 1250 CM-1
! BAND #2: 1325 CM-1 - 1575 CM-1
! BAND #3: 1775 CM-1 - 2050 CM-1
! BAND #4: 2825 CM-1 - 3350 CM-1

! GET THE INDEX OF THE BAND
BAND: DO I = 1, N_BAND_C2H4
   IF ((OM_BND_C2H4(I,1)<=OMEGA).AND.(OMEGA<OM_BND_C2H4(I,2))) THEN
      IN_BAND = .TRUE.
      I_BAND  = I
      EXIT BAND
   ENDIF
ENDDO BAND

IF (IN_BAND) THEN

   PRESSURE_EFFECTIVE = PTOT+(BE_C2H4(I_BAND)-1.0_EB)*PC2H4

   SELECT CASE (I_BAND)
      CASE (1)  ! BAND #1: 800 CM-1  - 1250 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),SD1_C2H4)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),GAMMAD1_C2H4)
      CASE (2) ! BAND #2: 1325 CM-1 - 1575 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),SD2_C2H4)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),GAMMAD2_C2H4)
      CASE (3) ! BAND #3: 1775 CM-1 - 2050 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),SD3_C2H4)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),GAMMAD3_C2H4)
      CASE (4) ! BAND #4: 2825 CM-1 - 3350 CM-1
         SDWEAK = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),SD4_C2H4)
         GDINV  = GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_C2H4_TEMP,OM_BND_C2H4(I_BAND,:),GAMMAD4_C2H4)
   END SELECT

   ! COMPUTE PROPERTIES
   GDINV  = GDINV*PRESSURE_EFFECTIVE
   GDDINV = GD*DINV_C2H4
   !***EXPRESS S/D AT STP, AS IS K IN NASA SP-3080
   SDWEAK = SDWEAK*TOAZ
ENDIF

RETURN
!------------------------------------------------------------------------------
END SUBROUTINE C2H4

!==============================================================================
PURE REAL(EB) FUNCTION GET_SPECTRAL_ABSORPTION(OMEGA,TEMP,SD_TEMP,BOUNDS,ABSORPTION_DATA)
!==============================================================================
! RETURNS INTERPOLATED VALUES OF SPECTRAL ABSORPTION
! NOTE ASSUMED SHAPE
!
! INPUT:
! OMEGA  : WAVENUMBER, UNIT = 1/CM
! TEMP   : GAS TEMPERATURE, UNIT = KELVIN
! SD_TEMP: VECTOR OF TEMPERATURE MEASUREMENTS, DIMENSION(*) UNIT = KELVIN
! BOUNDS : VECTOR OF WAVENUMBER BOUNDS AND SPACING FOR A GIVEN BAND, DIMENSIONS(3)
! ABSORPTION_DATA: ARRAY OF SPECIES SPECTRAL MEASUREMENTS, DIMENSIONS(*,*)
!                 NOTE = FIRST DIMENSION EQUALS TO DIMENSION OF SD_TEMP
!
! ACTION PERFORMED: BILINEAR INTERPOLATION OF ABSORPTION_DATA
!------------------------------------------------------------------------------
! VARIABLES PASSED IN

REAL(EB), INTENT(IN) :: OMEGA
REAL(EB), INTENT(IN) :: TEMP
REAL(EB), DIMENSION(:), INTENT(IN)   :: SD_TEMP, BOUNDS
REAL(EB), DIMENSION(:,:), INTENT(IN) :: ABSORPTION_DATA

REAL(EB) :: DELTA_T
REAL(EB) :: DELTA_W
REAL(EB) :: W1
REAL(EB) :: OMEGA_MIN, OMEGA_MAX, DELTA_OMEGA
REAL(EB) :: TTEMP
REAL(EB) :: INTERPOLATED_VALUE

INTEGER :: N_TEMP, N_BOUNDS, N_ABS_DATA1, N_ABS_DATA2
INTEGER :: I_OMEGA, I_TEMP, I_OMEGA1, I_OMEGA2

LOGICAL :: CROSS

CROSS = .FALSE.

! CVL NEED TO LINK MODULE THAT DEFINES ERRMSG
! PART OF THE CODE COMMENTED FOR NOW
!  CHARACTER(255) MESSAG
!  LOGICAL FATAL

N_TEMP      = UBOUND(SD_TEMP,1)  ! GET NUMBER OF TEMPERATURE MEASUREMENTS
N_BOUNDS    = UBOUND(BOUNDS,1)   ! GET NUMBER OF ELEMENTS IN BOUNDS. SHOULD BE 3
N_ABS_DATA1 = UBOUND(ABSORPTION_DATA,1) ! GET NUMBER OF TEMPERATURE DATA
N_ABS_DATA2 = UBOUND(ABSORPTION_DATA,2) ! GET NUMBER OF WAVENUMBER DATA

! ! CVL NEED TO LINK MODULE THAT DEFINES ERRMSG
! ! PART OF THE CODE COMMENTED FOR NOW
! !-----------------------------TEST DIMENSIONS OF BOUNDS AND ABSORPTION_DATA----
! ! N_BOUNDS SHOULD BE 3
!
!  IF (N_BOUNDS/=3) THEN
!     MESSAG='ERROR IN RADCAL. N_BOUNDS SHOULD BE EQUAL TO 3'
!     FATAL = .TRUE.
!     CALL ERRMSG( TRIM(MESSAG), FATAL )
!  ENDIF
!
! ! N_ABS_DATA1 SHOULD BE EQUAL TO N_TEMP
!
!  IF (N_ABS_DATA1/=N_TEMP) THEN
!     MESSAG='ERROR IN RADCAL. FIRST DIMENSION OF ABSORPTION_DATA SHOULD BE EQUAL TO N_TEMP'
!     FATAL = .TRUE.
!     CALL ERRMSG( TRIM(MESSAG), FATAL )
!  ENDIF

!------------------------------------------------------------------------------
OMEGA_MIN   = BOUNDS(1)
OMEGA_MAX   = BOUNDS(2)
DELTA_OMEGA = BOUNDS(3)

! TREAT SPECIAL CASE WHEN BAND CROSSES THE 1100 CM-1 MARK.
! CHANGES OF STRIDE
! BELOW 1100 CM-1, DELTA_OMEGA = 5 CM-1
! ABOVE 1100 CM-1, DELTA_OMEGA = 25 CM-1

IF ((OMEGA_MIN<=1100.0_EB).AND.(OMEGA_MAX>1100.0_EB)) THEN
CROSS = .TRUE.
IF (OMEGA>1100._EB) DELTA_OMEGA = 25.0_EB
ENDIF

! BOUNDS TTEMP SO THAT MIN(SD_TEMP) <= TTEMP <= MAX(SD_TEMP)
! IF TTEMP < MIN(SD_TEMP) THEN TTEMP := MIN(SD_TEMP)
! IF MAX(SD_TEMP) < TTEMP THEN TTEMP := MAX(SD_TEMP)
TTEMP       = MIN(MAXVAL(SD_TEMP),MAX(TEMP,MINVAL(SD_TEMP)))

IF (TTEMP <= MINVAL(SD_TEMP)) THEN
   I_TEMP  = 1
   DELTA_T = 0._EB
ELSE IF (TTEMP >= MAXVAL(SD_TEMP)) THEN
   I_TEMP  = N_TEMP-1
   DELTA_T = 1.0_EB
ELSE
! FIND INDEX I_TEMP SUCH THAT SD_TEMP(I_TEMP)<TTEMP<=SD_TEMP(I_TEMP+1)
   DO I_TEMP = 1, N_TEMP-1
      IF ((SD_TEMP(I_TEMP)<TTEMP).AND.(TTEMP<=SD_TEMP(I_TEMP+1))) EXIT
   ENDDO

   DELTA_T = (TTEMP-SD_TEMP(I_TEMP))/(SD_TEMP(I_TEMP+1)-SD_TEMP(I_TEMP))
ENDIF

IF ((OMEGA_MIN<=OMEGA).AND.(OMEGA<=OMEGA_MAX)) THEN

   IF(.NOT.(CROSS)) THEN ! THE BAND IS ON EITHER SIDE OF THE 1100 CM-1 MARK

      IF(OMEGA<OMEGA_MAX) THEN
! FIND I_OMEGA SUCH THAT OMEGA(I_OMEGA)<=OMEGA<OMEGA(I_OMEGA+1)
         I_OMEGA = FLOOR((OMEGA-OMEGA_MIN)/DELTA_OMEGA)+1
         I_OMEGA = MIN(I_OMEGA,N_ABS_DATA2-1)
         W1      = OMEGA_MIN+DELTA_OMEGA*(REAL(I_OMEGA-1,EB))
         DELTA_W = (OMEGA-W1)/DELTA_OMEGA
      ELSE
! CASE I_OMEGA = OMEGA_MAX => ASSUME THEN THAT I_OMEGA = N_ABS_DATA2 -1
         I_OMEGA = N_ABS_DATA2 - 1
         W1      = OMEGA_MIN+DELTA_OMEGA*(REAL(I_OMEGA-1,EB))
         DELTA_W = (OMEGA-W1)/DELTA_OMEGA
      ENDIF

   ELSE ! THE BAND IS ON BOTH SIDE OF 1100 CM-1
      ! TEST IF OMEGA <= 1100
      IF(OMEGA<=1100) THEN
! FIND I_OMEGA SUCH THAT OMEGA(I_OMEGA)<=OMEGA<OMEGA(I_OMEGA+1)
         I_OMEGA = FLOOR((OMEGA-OMEGA_MIN)/DELTA_OMEGA)+1
         I_OMEGA = MIN(I_OMEGA,N_ABS_DATA2-1)
         W1      = OMEGA_MIN+DELTA_OMEGA*(REAL(I_OMEGA-1,EB))
         DELTA_W = (OMEGA-W1)/DELTA_OMEGA
      ELSE
! CASE OMEGA > 1100 CM-1
         I_OMEGA1 = FLOOR((1100-OMEGA_MIN)/5.0_EB)+1
         I_OMEGA2 = FLOOR((OMEGA-1100)/DELTA_OMEGA)
         I_OMEGA  = MIN(I_OMEGA1+I_OMEGA2-1,N_ABS_DATA2-1)
         W1      = OMEGA_MIN+5.0_EB*(REAL(I_OMEGA1,EB))+DELTA_OMEGA*(REAL(I_OMEGA2-1,EB))
         DELTA_W = (OMEGA-W1)/DELTA_OMEGA
      ENDIF

   ENDIF

! PERFORM BILINEAR INTERPOLATION TO OBTAIN VALUES OF SDWEAK
   INTERPOLATED_VALUE = (1._EB-DELTA_W)*((1._EB-DELTA_T)*ABSORPTION_DATA(I_TEMP,I_OMEGA)+ &
                        ABSORPTION_DATA(I_TEMP+1,I_OMEGA)*DELTA_T) +                      &
                        DELTA_W*(ABSORPTION_DATA(I_TEMP,I_OMEGA+1)*(1._EB-DELTA_T)+       &
                        ABSORPTION_DATA(I_TEMP+1,I_OMEGA+1)*DELTA_T)

ELSE
   INTERPOLATED_VALUE = 0._EB
ENDIF

! ENSURE POSITIVITY
GET_SPECTRAL_ABSORPTION = MAX(0._EB,INTERPOLATED_VALUE)
!------------------------------------------------------------------------------
END FUNCTION GET_SPECTRAL_ABSORPTION


!==============================================================================
ELEMENTAL REAL(EB) FUNCTION PLANCK(TEMP,LAMBDA)
!==============================================================================
! COMPUTES BLACKBODY FUNCTION IN UNITS OF W/M^2/MICRON/SR
! INPUT:
! TEMP: TEMPERATURE, UNIT = KELVIN
! LAMBDA: WAVELENGTH, UNIT = MICROMETERS
!------------------------------------------------------------------------------
! VARIABLES PASSED IN
REAL (EB), INTENT(IN):: TEMP, LAMBDA

! LOCAL VARIABLES
REAL (EB), PARAMETER :: Q1 = 1.19088E8_EB ! Q1 = 2*SPEED_OF_LIGHT^2*PLANCK_CNST
REAL (EB), PARAMETER :: Q2 = 14388._EB    ! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL (EB) :: C   ! C = LAMBDA * TEMP

IF(ABS(TEMP)<TWO_EPSILON_EB) THEN
   PLANCK = 0._EB
ELSE
   C = TEMP * LAMBDA
   IF (Q2/C > 38._EB) THEN
      PLANCK = 0._EB
   ELSE
      PLANCK=Q1*(LAMBDA**(-5))/(EXP(Q2/C)-1._EB)
END IF
ENDIF
!------------------------------------------------------------------------------
END FUNCTION PLANCK

!==============================================================================
ELEMENTAL REAL(EB) FUNCTION PLANCK_WN(TEMP,OMEGA)
!==============================================================================
! COMPUTES BLACKBODY FUNCTION IN UNITS OF W/M^2/CM^-1/SR
! INPUT:
! TEMP : TEMPERATURE, UNIT = KELVIN
! OMEGA: WAVENUMBER,  UNIT = CM-1
! CHECKED BY VLECOUS1 ON APRIL 15TH 2013
!------------------------------------------------------------------------------
! VARIABLES PASSED IN
REAL (EB), INTENT(IN):: TEMP, OMEGA

! LOCAL VARIABLES
REAL (EB), PARAMETER :: Q1 = 1.19088E-8_EB ! Q1 = 2*SPEED_OF_LIGHT^2*PLANCK_CNST
REAL (EB), PARAMETER :: Q2 = 1.4388_EB
! Q2 = SPEED_OF_LIGHT*PLANCK_CNS/BOLTZMANN
REAL (EB) :: C   ! C = TEMP/OMEGA

IF(ABS(TEMP)<TWO_EPSILON_EB) THEN
   PLANCK_WN = 0._EB
ELSE
   C = TEMP/OMEGA
   IF (Q2/C > 38._EB) THEN
      PLANCK_WN = 0._EB
   ELSE
      PLANCK_WN=Q1*(OMEGA**3)/(EXP(Q2/C)-1._EB)
END IF
ENDIF
!------------------------------------------------------------------------------
END FUNCTION PLANCK_WN


!==============================================================================
REAL(EB) FUNCTION INTEGRATION(X,Y)
!==============================================================================
! NUMERICAL INTEGRATION OF THE VECTOR Y OVER THE RANGE GIVEN BY VECTOR X
! DOES NOT ASSUME ANY REGULARITY OF X
! INTEGRATION BASED ON SIMPSON RULE OVER NON REGULAR ABS(CISSA
! BASED ON LAGRANGIAN INTERPOLATION USING 3 POINT (QUADRATIC INTERPOLATION)
! NOTE: IO IS UNIT OF OUTPUT FILE. NEEDED FOR PRINTING ERROR MESSAGE
!------------------------------------------------------------------------------
! VARIABLES PASSED IN
REAL(EB), INTENT(IN), DIMENSION(:) :: X,Y

! LOCAL VARIABLE
REAL(EB) :: ALPHA, BETA, GAMMA, B, A, SEGMENT_INT

INTEGER  :: I_POINT, N_POINTS, I, J ,K

! GET NUMBER OF ELEMENTS CONTAINS IN X AND Y

N_POINTS = SIZE(Y)

! TEST WHETHER SIZE(X) = SIZE(Y). STOP IF CONDITION NOT MET
IF (SIZE(X)/=N_POINTS) THEN
   WRITE(MESSAGE, '(A)') 'ERROR 2: (INTERNAL) VECTOR X AND Y FOR INTEGRATION SHOULD HAVE SAME SIZE(.'
   CALL SHUTDOWN(MESSAGE)
ENDIF

! TEST WHETHER N_POINTS IS GREATER OR EQUAL TO 3. IF NOT, PRINT ERROR MSG AND STOP RADCAL
IF (N_POINTS.LT.2) THEN
   WRITE(MESSAGE, '(A)') 'ERROR 3: (INTERNAL) NOT ENOUGH POINTS FOR INTEGRATION.'
   CALL SHUTDOWN(MESSAGE)
ENDIF
! STARTS INTEGRATION
INTEGRATION = 0.0_EB

! TREAT FIRST AND LAST SEGMENT IN THE LOOP
DO I_POINT = 1,N_POINTS

   IF (I_POINT == 1) THEN
      I = I_POINT
      J = I_POINT+1
      K = I_POINT+2

      A = X(I)
      B = 0.5_EB*(X(I)+X(J))
   ELSEIF(I_POINT == N_POINTS) THEN
      I = I_POINT-2
      J = I_POINT-1
      K = I_POINT

      A = 0.5_EB*(X(J)+X(K))
      B = X(K)
   ELSE
      I = I_POINT-1
      J = I_POINT
      K = I_POINT+1

      A = 0.5_EB*(X(I)+X(J))
      B = 0.5_EB*(X(J)+X(K))
   ENDIF

   ALPHA = Y(I)/((X(I)-X(J))*(X(I)-X(K)))
   BETA  = Y(J)/((X(J)-X(I))*(X(J)-X(K)))
   GAMMA = Y(K)/((X(K)-X(I))*(X(K)-X(J)))

   SEGMENT_INT = 1.0_EB/3.0_EB*(B**3-A**3)*(ALPHA+BETA+GAMMA)
   SEGMENT_INT = SEGMENT_INT - 0.5_EB*(B**2-A**2)*((X(J)+X(K))*ALPHA+BETA*(X(I)+X(K))+GAMMA*(X(I)+X(J)))
   SEGMENT_INT = SEGMENT_INT + (B-A)*(ALPHA*X(J)*X(K)+BETA*X(K)*X(I)+GAMMA*X(J)*X(I))

   INTEGRATION = INTEGRATION + SEGMENT_INT

END DO

! FINALLY INTEGRATION = SIGN(B-A)*INTEGRATION. THIS ACCOUNTS FOR
! DESCENDING ORDER VECTOR X.

INTEGRATION = SIGN(1.0_EB,B-A)*INTEGRATION

RETURN

!------------------------------------------------------------------------------
END FUNCTION INTEGRATION


!==============================================================================
SUBROUTINE RCALLOC
!==============================================================================

! OLD RADCAL VARIABLES
ALLOCATE(GAMMA(4,7))
ALLOCATE(SD15(6,80))
ALLOCATE(SD(6,376))
ALLOCATE(SD7(3,16))
ALLOCATE(SD3(3,32))

!-------------------------METHANE DATA-------------------


! THERE ARE 3 BANDS FOR METHANE

! BAND #1: 1150 CM-1 - 1600 CM-1
! BAND #2: 2700 CM-1 - 3250 CM-1
! BAND #3: 3400 CM-1 - 5000 CM-1

ALLOCATE(SD_CH4_TEMP(N_TEMP_CH4))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR METHANE ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., THEN THE BAND IS CALCULATED AND NOT TABULATED.

ALLOCATE(OM_BND_CH4(N_BAND_CH4,3))

OM_BND_CH4 = RESHAPE((/ &
   1150._EB, 2700._EB, 3400._EB,&
   1600._EB, 3250._EB, 5000._EB,&
   25._EB,   25._EB,    0._EB/),(/N_BAND_CH4,3/))

SD_CH4_TEMP = (/ &
   300._EB, 350._EB, 400._EB, 450._EB, 500._EB, 550._EB,&
   600._EB, 650._EB, 700._EB, 750._EB, 800._EB, 850._EB,&
   900._EB, 950._EB, 1000._EB, 1050._EB, 1100._EB, 1150._EB,&
   1200._EB, 1250._EB, 1300._EB, 1350._EB, 1400._EB/)

ALLOCATE(SD1_CH4(N_TEMP_CH4,19))

! BAND #1: 1150 CM-1 - 1600 CM-1

SD1_CH4(1:23,1:8) = RESHAPE((/ &  ! 1150-1325 CM-1
   2.674955E-03_EB, 4.122616E-03_EB, 6.116504E-03_EB, 8.394889E-03_EB, 1.066588E-02_EB, 1.270929E-02_EB, &
   1.431368E-02_EB, 1.543500E-02_EB, 1.607798E-02_EB, 1.623394E-02_EB, 1.594103E-02_EB, 1.537005E-02_EB, &
   1.456109E-02_EB, 1.356591E-02_EB, 1.253062E-02_EB, 1.145338E-02_EB, 1.037156E-02_EB, 9.319748E-03_EB, &
   8.296535E-03_EB, 7.359273E-03_EB, 6.512254E-03_EB, 5.764075E-03_EB, 5.034367E-03_EB,  &
   1.056047E-02_EB, 1.689522E-02_EB, 2.382293E-02_EB, 3.013931E-02_EB, 3.518013E-02_EB, 3.873081E-02_EB, &
   4.056505E-02_EB, 4.099451E-02_EB, 4.031689E-02_EB, 3.867720E-02_EB, 3.630281E-02_EB, 3.360423E-02_EB, &
   3.072697E-02_EB, 2.769650E-02_EB, 2.480529E-02_EB, 2.203561E-02_EB, 1.945492E-02_EB, 1.710571E-02_EB, &
   1.490918E-02_EB, 1.297063E-02_EB, 1.127359E-02_EB, 9.806206E-03_EB, 8.433200E-03_EB,  &
   6.535368E-02_EB, 8.634735E-02_EB, 1.035285E-01_EB, 1.149338E-01_EB, 1.211259E-01_EB, 1.231160E-01_EB, &
   1.212403E-01_EB, 1.167139E-01_EB, 1.104665E-01_EB, 1.027695E-01_EB, 9.409726E-02_EB, 8.537322E-02_EB, &
   7.679556E-02_EB, 6.836183E-02_EB, 6.060003E-02_EB, 5.339354E-02_EB, 4.679438E-02_EB, 4.089038E-02_EB, &
   3.545789E-02_EB, 3.072468E-02_EB, 2.660506E-02_EB, 2.306236E-02_EB, 1.979016E-02_EB,  &
   2.797310E-01_EB, 2.878774E-01_EB, 2.844637E-01_EB, 2.718418E-01_EB, 2.544158E-01_EB, 2.351542E-01_EB, &
   2.144282E-01_EB, 1.940107E-01_EB, 1.744123E-01_EB, 1.554283E-01_EB, 1.374171E-01_EB, 1.210226E-01_EB, &
   1.060935E-01_EB, 9.238361E-02_EB, 8.033186E-02_EB, 6.960028E-02_EB, 6.015950E-02_EB, 5.186871E-02_EB, &
   4.448682E-02_EB, 3.815557E-02_EB, 3.272683E-02_EB, 2.816055E-02_EB, 2.399102E-02_EB,  &
   6.055294E-01_EB, 5.085203E-01_EB, 4.314546E-01_EB, 3.661128E-01_EB, 3.120083E-01_EB, 2.675996E-01_EB, &
   2.295433E-01_EB, 1.976798E-01_EB, 1.706085E-01_EB, 1.470613E-01_EB, 1.263463E-01_EB, 1.086762E-01_EB, &
   9.339644E-02_EB, 7.991483E-02_EB, 6.850499E-02_EB, 5.859531E-02_EB, 5.007218E-02_EB, 4.275497E-02_EB, &
   3.636644E-02_EB, 3.094626E-02_EB, 2.637845E-02_EB, 2.255529E-02_EB, 1.911152E-02_EB,  &
   7.291248E-01_EB, 6.247580E-01_EB, 5.567927E-01_EB, 5.047634E-01_EB, 4.622435E-01_EB, 4.253591E-01_EB, &
   3.901550E-01_EB, 3.566718E-01_EB, 3.248608E-01_EB, 2.936112E-01_EB, 2.628808E-01_EB, 2.344160E-01_EB, &
   2.078641E-01_EB, 1.829068E-01_EB, 1.606407E-01_EB, 1.403181E-01_EB, 1.221762E-01_EB, 1.060864E-01_EB, &
   9.154534E-02_EB, 7.892764E-02_EB, 6.808643E-02_EB, 5.882322E-02_EB, 5.031768E-02_EB,  &
   1.503577E+00_EB, 1.216438E+00_EB, 1.009787E+00_EB, 8.445781E-01_EB, 7.110515E-01_EB, 6.021463E-01_EB, &
   5.099874E-01_EB, 4.327438E-01_EB, 3.680691E-01_EB, 3.125303E-01_EB, 2.644163E-01_EB, 2.242054E-01_EB, &
   1.897708E-01_EB, 1.602130E-01_EB, 1.354708E-01_EB, 1.143770E-01_EB, 9.654761E-02_EB, 8.152145E-02_EB, &
   6.854917E-02_EB, 5.774320E-02_EB, 4.875464E-02_EB, 4.129146E-02_EB, 3.467888E-02_EB,  &
   1.113787E+00_EB, 8.666106E-01_EB, 6.937600E-01_EB, 5.636603E-01_EB, 4.651973E-01_EB, 3.899245E-01_EB, &
   3.293625E-01_EB, 2.804438E-01_EB, 2.403481E-01_EB, 2.062563E-01_EB, 1.767199E-01_EB, 1.518103E-01_EB, &
   1.304026E-01_EB, 1.116263E-01_EB, 9.571162E-02_EB, 8.192030E-02_EB, 7.003835E-02_EB, 5.986022E-02_EB, &
   5.092725E-02_EB, 4.337782E-02_EB, 3.700459E-02_EB, 3.164349E-02_EB, 2.682395E-02_EB/),(/23,8/))

SD1_CH4(1:23,9:16) = RESHAPE((/ &  ! 1350-1525 CM-1
   5.859637E-01_EB, 5.741640E-01_EB, 5.446284E-01_EB, 5.016955E-01_EB, 4.533548E-01_EB, 4.052375E-01_EB, &
   3.578427E-01_EB, 3.138042E-01_EB, 2.741405E-01_EB, 2.379703E-01_EB, 2.050667E-01_EB, 1.765003E-01_EB, &
   1.515652E-01_EB, 1.294551E-01_EB, 1.106049E-01_EB, 9.430003E-02_EB, 8.028460E-02_EB, 6.831086E-02_EB, &
   5.788843E-02_EB, 4.905327E-02_EB, 4.167658E-02_EB, 3.549325E-02_EB, 2.997796E-02_EB,  &
   5.143640E-02_EB, 7.394302E-02_EB, 9.480931E-02_EB, 1.108135E-01_EB, 1.211319E-01_EB, 1.262273E-01_EB, &
   1.262909E-01_EB, 1.227527E-01_EB, 1.167145E-01_EB, 1.087709E-01_EB, 9.943737E-02_EB, 9.002538E-02_EB, &
   8.060527E-02_EB, 7.137067E-02_EB, 6.299760E-02_EB, 5.512164E-02_EB, 4.801455E-02_EB, 4.169764E-02_EB, &
   3.593587E-02_EB, 3.095498E-02_EB, 2.664278E-02_EB, 2.295572E-02_EB, 1.959682E-02_EB,  &
   7.616134E-03_EB, 9.957002E-03_EB, 1.281279E-02_EB, 1.589717E-02_EB, 1.894916E-02_EB, 2.174268E-02_EB, &
   2.395920E-02_EB, 2.555557E-02_EB, 2.648904E-02_EB, 2.671274E-02_EB, 2.629477E-02_EB, 2.540283E-02_EB, &
   2.418621E-02_EB, 2.265546E-02_EB, 2.101542E-02_EB, 1.928221E-02_EB, 1.752831E-02_EB, 1.582992E-02_EB, &
   1.416926E-02_EB, 1.261250E-02_EB, 1.121041E-02_EB, 9.951538E-03_EB, 8.728757E-03_EB,  &
   1.277402E-02_EB, 1.327177E-02_EB, 1.322108E-02_EB, 1.267471E-02_EB, 1.186190E-02_EB, 1.093562E-02_EB, &
   9.920972E-03_EB, 8.920419E-03_EB, 7.971129E-03_EB, 7.057028E-03_EB, 6.197372E-03_EB, 5.429023E-03_EB, &
   4.732635E-03_EB, 4.092017E-03_EB, 3.548360E-03_EB, 3.062048E-03_EB, 2.633809E-03_EB, 2.263255E-03_EB, &
   1.935254E-03_EB, 1.656795E-03_EB, 1.416023E-03_EB, 1.215289E-03_EB, 1.031774E-03_EB,  &
   8.014464E-03_EB, 7.194752E-03_EB, 6.553713E-03_EB, 6.000398E-03_EB, 5.531684E-03_EB, 5.121393E-03_EB, &
   4.722456E-03_EB, 4.343916E-03_EB, 3.983544E-03_EB, 3.621247E-03_EB, 3.268422E-03_EB, 2.936068E-03_EB, &
   2.617151E-03_EB, 2.322363E-03_EB, 2.051315E-03_EB, 1.804840E-03_EB, 1.578471E-03_EB, 1.378963E-03_EB, &
   1.197596E-03_EB, 1.036958E-03_EB, 9.007021E-04_EB, 7.820103E-04_EB, 6.693842E-04_EB,  &
   2.792665E-03_EB, 2.575454E-03_EB, 2.537125E-03_EB, 2.578233E-03_EB, 2.648059E-03_EB, 2.713453E-03_EB, &
   2.741224E-03_EB, 2.722787E-03_EB, 2.662395E-03_EB, 2.559402E-03_EB, 2.415338E-03_EB, 2.254521E-03_EB, &
   2.077537E-03_EB, 1.891657E-03_EB, 1.708933E-03_EB, 1.533948E-03_EB, 1.367874E-03_EB, 1.212162E-03_EB, &
   1.064882E-03_EB, 9.339250E-04_EB, 8.169687E-04_EB, 7.180516E-04_EB, 6.220503E-04_EB,  &
   8.239408E-04_EB, 8.207289E-04_EB, 8.769232E-04_EB, 9.471812E-04_EB, 1.013068E-03_EB, 1.072041E-03_EB, &
   1.109555E-03_EB, 1.125480E-03_EB, 1.120373E-03_EB, 1.094883E-03_EB, 1.043855E-03_EB, 9.887558E-04_EB, &
   9.221581E-04_EB, 8.517148E-04_EB, 7.786200E-04_EB, 7.086512E-04_EB, 6.363394E-04_EB, 5.709486E-04_EB, &
   5.037963E-04_EB, 4.478632E-04_EB, 3.933607E-04_EB, 3.461225E-04_EB, 3.036065E-04_EB,  &
   1.787674E-02_EB, 1.651301E-02_EB, 1.538618E-02_EB, 1.427021E-02_EB, 1.321135E-02_EB, 1.220556E-02_EB, &
   1.120017E-02_EB, 1.022271E-02_EB, 9.293408E-03_EB, 8.381487E-03_EB, 7.491250E-03_EB, 6.673691E-03_EB, &
   5.914957E-03_EB, 5.201979E-03_EB, 4.566356E-03_EB, 3.988961E-03_EB, 3.472458E-03_EB, 3.017625E-03_EB, &
   2.605493E-03_EB, 2.248445E-03_EB, 1.941774E-03_EB, 1.676885E-03_EB, 1.438512E-03_EB/),(/23,8/))

SD1_CH4(1:23,17:19) = RESHAPE((/ &  ! 1550-1600 CM-1
   3.842192E-03_EB, 4.909077E-03_EB, 5.876585E-03_EB, 6.637230E-03_EB, 7.159498E-03_EB, 7.449619E-03_EB, &
   7.502474E-03_EB, 7.374704E-03_EB, 7.105819E-03_EB, 6.720456E-03_EB, 6.241450E-03_EB, 5.729337E-03_EB, &
   5.213056E-03_EB, 4.685851E-03_EB, 4.183214E-03_EB, 3.714659E-03_EB, 3.276286E-03_EB, 2.879051E-03_EB, &
   2.508044E-03_EB, 2.182116E-03_EB, 1.899894E-03_EB, 1.653926E-03_EB, 1.423668E-03_EB,  &
   1.587677E-03_EB, 1.446538E-03_EB, 1.552901E-03_EB, 1.804439E-03_EB, 2.128742E-03_EB, 2.458565E-03_EB, &
   2.743898E-03_EB, 2.955974E-03_EB, 3.088835E-03_EB, 3.134690E-03_EB, 3.095698E-03_EB, 3.002805E-03_EB, &
   2.862955E-03_EB, 2.683299E-03_EB, 2.486550E-03_EB, 2.282038E-03_EB, 2.073847E-03_EB, 1.873143E-03_EB, &
   1.674361E-03_EB, 1.489429E-03_EB, 1.321804E-03_EB, 1.173554E-03_EB, 1.026702E-03_EB,  &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB/),(/23,3/))

ALLOCATE(SD2_CH4(N_TEMP_CH4,23))

! BAND #2: 2700 CM-1 - 3250 CM-1

SD2_CH4(1:23,1:8) = RESHAPE((/ &  ! 2700-2875 CM-1
   1.217530E-02_EB, 1.117112E-02_EB, 1.018335E-02_EB, 9.113159E-03_EB, 8.069237E-03_EB, 7.089143E-03_EB, &
   6.164984E-03_EB, 5.339743E-03_EB, 4.605161E-03_EB, 3.952813E-03_EB, 3.369126E-03_EB, 2.873535E-03_EB, &
   2.442004E-03_EB, 2.069306E-03_EB, 1.752468E-03_EB, 1.483929E-03_EB, 1.255931E-03_EB, 1.056565E-03_EB, &
   8.920522E-04_EB, 7.508758E-04_EB, 6.317838E-04_EB, 5.352971E-04_EB, 4.530464E-04_EB,  &
   2.513916E-02_EB, 2.114376E-02_EB, 1.780513E-02_EB, 1.491293E-02_EB, 1.247243E-02_EB, 1.043356E-02_EB, &
   8.698933E-03_EB, 7.256269E-03_EB, 6.066735E-03_EB, 5.057876E-03_EB, 4.209670E-03_EB, 3.508303E-03_EB, &
   2.922322E-03_EB, 2.428483E-03_EB, 2.028794E-03_EB, 1.689558E-03_EB, 1.409261E-03_EB, 1.176715E-03_EB, &
   9.788265E-04_EB, 8.167162E-04_EB, 6.823326E-04_EB, 5.730157E-04_EB, 4.786727E-04_EB,  &
   2.557782E-02_EB, 2.133345E-02_EB, 1.832454E-02_EB, 1.586992E-02_EB, 1.378920E-02_EB, 1.202910E-02_EB, &
   1.045539E-02_EB, 9.073223E-03_EB, 7.878108E-03_EB, 6.813401E-03_EB, 5.858711E-03_EB, 5.036798E-03_EB, &
   4.322566E-03_EB, 3.689893E-03_EB, 3.157259E-03_EB, 2.690401E-03_EB, 2.292743E-03_EB, 1.951361E-03_EB, &
   1.658405E-03_EB, 1.405787E-03_EB, 1.195120E-03_EB, 1.018165E-03_EB, 8.624014E-04_EB,  &
   3.267234E-02_EB, 2.794974E-02_EB, 2.414894E-02_EB, 2.075791E-02_EB, 1.780983E-02_EB, 1.531517E-02_EB, &
   1.312860E-02_EB, 1.127375E-02_EB, 9.704346E-03_EB, 8.352011E-03_EB, 7.159191E-03_EB, 6.165400E-03_EB, &
   5.304681E-03_EB, 4.552437E-03_EB, 3.918329E-03_EB, 3.371374E-03_EB, 2.897155E-03_EB, 2.490584E-03_EB, &
   2.139129E-03_EB, 1.834560E-03_EB, 1.579448E-03_EB, 1.362473E-03_EB, 1.164728E-03_EB,  &
   4.729257E-02_EB, 4.020146E-02_EB, 3.655117E-02_EB, 3.486375E-02_EB, 3.451183E-02_EB, 3.486976E-02_EB, &
   3.526911E-02_EB, 3.549943E-02_EB, 3.539419E-02_EB, 3.477334E-02_EB, 3.359574E-02_EB, 3.209631E-02_EB, &
   3.033091E-02_EB, 2.826035E-02_EB, 2.617425E-02_EB, 2.398793E-02_EB, 2.182750E-02_EB, 1.975692E-02_EB, &
   1.770980E-02_EB, 1.581216E-02_EB, 1.409330E-02_EB, 1.256203E-02_EB, 1.106706E-02_EB,  &
   3.967221E-02_EB, 3.764359E-02_EB, 3.949747E-02_EB, 4.277841E-02_EB, 4.623172E-02_EB, 4.903094E-02_EB, &
   5.058120E-02_EB, 5.097742E-02_EB, 5.029388E-02_EB, 4.863988E-02_EB, 4.605751E-02_EB, 4.308689E-02_EB, &
   3.983596E-02_EB, 3.638363E-02_EB, 3.298206E-02_EB, 2.967115E-02_EB, 2.650179E-02_EB, 2.356866E-02_EB, &
   2.077236E-02_EB, 1.827032E-02_EB, 1.605444E-02_EB, 1.411180E-02_EB, 1.226114E-02_EB,  &
   7.223868E-02_EB, 8.544183E-02_EB, 9.962661E-02_EB, 1.104863E-01_EB, 1.171420E-01_EB, 1.196787E-01_EB, &
   1.182118E-01_EB, 1.139658E-01_EB, 1.078438E-01_EB, 1.002471E-01_EB, 9.161471E-02_EB, 8.298437E-02_EB, &
   7.452297E-02_EB, 6.622584E-02_EB, 5.857678E-02_EB, 5.147962E-02_EB, 4.505849E-02_EB, 3.932179E-02_EB, &
   3.408714E-02_EB, 2.948172E-02_EB, 2.551286E-02_EB, 2.211846E-02_EB, 1.897647E-02_EB,  &
   1.759866E-01_EB, 2.015363E-01_EB, 2.180490E-01_EB, 2.232613E-01_EB, 2.203020E-01_EB, 2.118597E-01_EB, &
   1.990075E-01_EB, 1.841205E-01_EB, 1.683887E-01_EB, 1.521963E-01_EB, 1.360646E-01_EB, 1.209307E-01_EB, &
   1.069128E-01_EB, 9.382435E-02_EB, 8.206146E-02_EB, 7.155641E-02_EB, 6.219310E-02_EB, 5.392479E-02_EB, &
   4.648837E-02_EB, 4.008493E-02_EB, 3.458414E-02_EB, 2.988156E-02_EB, 2.558176E-02_EB/),(/23,8/))

SD2_CH4(1:23,9:16) = RESHAPE((/ &  ! 2900-3075 CM-1
   3.311709E-01_EB, 3.344762E-01_EB, 3.267663E-01_EB, 3.099531E-01_EB, 2.886161E-01_EB, 2.659881E-01_EB, &
   2.423456E-01_EB, 2.191913E-01_EB, 1.973904E-01_EB, 1.764196E-01_EB, 1.564157E-01_EB, 1.382986E-01_EB, &
   1.218590E-01_EB, 1.067421E-01_EB, 9.331013E-02_EB, 8.132822E-02_EB, 7.068471E-02_EB, 6.139407E-02_EB, &
   5.297295E-02_EB, 4.571219E-02_EB, 3.948284E-02_EB, 3.416733E-02_EB, 2.931231E-02_EB,  &
   8.235613E-01_EB, 7.238968E-01_EB, 6.324968E-01_EB, 5.466365E-01_EB, 4.712117E-01_EB, 4.058971E-01_EB, &
   3.486088E-01_EB, 2.993153E-01_EB, 2.575278E-01_EB, 2.209771E-01_EB, 1.890168E-01_EB, 1.618217E-01_EB, &
   1.384064E-01_EB, 1.180022E-01_EB, 1.008139E-01_EB, 8.601549E-02_EB, 7.332027E-02_EB, 6.253544E-02_EB, &
   5.308446E-02_EB, 4.515135E-02_EB, 3.847721E-02_EB, 3.288247E-02_EB, 2.786602E-02_EB,  &
   6.821008E-01_EB, 5.401363E-01_EB, 4.353503E-01_EB, 3.533680E-01_EB, 2.892836E-01_EB, 2.389427E-01_EB, &
   1.981961E-01_EB, 1.653239E-01_EB, 1.387030E-01_EB, 1.163771E-01_EB, 9.777404E-02_EB, 8.236369E-02_EB, &
   6.949908E-02_EB, 5.853014E-02_EB, 4.942064E-02_EB, 4.176497E-02_EB, 3.531768E-02_EB, 2.985698E-02_EB, &
   2.516886E-02_EB, 2.128060E-02_EB, 1.802439E-02_EB, 1.532867E-02_EB, 1.292454E-02_EB,  &
   5.823907E-01_EB, 4.331161E-01_EB, 3.353248E-01_EB, 2.655224E-01_EB, 2.144294E-01_EB, 1.761169E-01_EB, &
   1.472744E-01_EB, 1.241615E-01_EB, 1.057407E-01_EB, 9.045740E-02_EB, 7.745806E-02_EB, 6.667193E-02_EB, &
   5.744892E-02_EB, 4.942743E-02_EB, 4.263289E-02_EB, 3.676096E-02_EB, 3.166692E-02_EB, 2.731256E-02_EB, &
   2.343114E-02_EB, 2.014747E-02_EB, 1.734533E-02_EB, 1.497211E-02_EB, 1.281431E-02_EB,  &
   3.210821E+00_EB, 2.793045E+00_EB, 2.475891E+00_EB, 2.198508E+00_EB, 1.954455E+00_EB, 1.740419E+00_EB, &
   1.543274E+00_EB, 1.366589E+00_EB, 1.208668E+00_EB, 1.063917E+00_EB, 9.309610E-01_EB, 8.135837E-01_EB, &
   7.091172E-01_EB, 6.150707E-01_EB, 5.334009E-01_EB, 4.611724E-01_EB, 3.980142E-01_EB, 3.431784E-01_EB, &
   2.944666E-01_EB, 2.526486E-01_EB, 2.171496E-01_EB, 1.870189E-01_EB, 1.596229E-01_EB,  &
   3.517151E-01_EB, 2.541996E-01_EB, 1.936159E-01_EB, 1.517283E-01_EB, 1.221750E-01_EB, 1.004308E-01_EB, &
   8.351777E-02_EB, 7.016371E-02_EB, 5.949379E-02_EB, 5.053966E-02_EB, 4.298934E-02_EB, 3.667062E-02_EB, &
   3.130112E-02_EB, 2.669085E-02_EB, 2.279490E-02_EB, 1.946235E-02_EB, 1.660600E-02_EB, 1.417075E-02_EB, &
   1.204697E-02_EB, 1.025668E-02_EB, 8.756988E-03_EB, 7.490243E-03_EB, 6.351877E-03_EB,  &
   7.921358E-01_EB, 6.055366E-01_EB, 4.798618E-01_EB, 3.867414E-01_EB, 3.174199E-01_EB, 2.642038E-01_EB, &
   2.216893E-01_EB, 1.874171E-01_EB, 1.594862E-01_EB, 1.359448E-01_EB, 1.159382E-01_EB, 9.908354E-02_EB, &
   8.473452E-02_EB, 7.233655E-02_EB, 6.187191E-02_EB, 5.286391E-02_EB, 4.518233E-02_EB, 3.863453E-02_EB, &
   3.288079E-02_EB, 2.803528E-02_EB, 2.394608E-02_EB, 2.051761E-02_EB, 1.743587E-02_EB,  &
   1.334409E+00_EB, 1.095178E+00_EB, 9.120448E-01_EB, 7.631562E-01_EB, 6.425576E-01_EB, 5.444728E-01_EB, &
   4.621443E-01_EB, 3.938260E-01_EB, 3.368579E-01_EB, 2.880892E-01_EB, 2.459378E-01_EB, 2.104005E-01_EB, &
   1.799327E-01_EB, 1.535802E-01_EB, 1.314063E-01_EB, 1.121907E-01_EB, 9.581696E-02_EB, 8.185813E-02_EB, &
   6.964125E-02_EB, 5.934825E-02_EB, 5.067442E-02_EB, 4.339843E-02_EB, 3.686411E-02_EB/),(/23,8/))

SD2_CH4(1:23,17:23) = RESHAPE((/ &  ! 3100-3250 CM-1
   9.674073E-01_EB, 9.013091E-01_EB, 8.256467E-01_EB, 7.423397E-01_EB, 6.619262E-01_EB, 5.879801E-01_EB, &
   5.191373E-01_EB, 4.573467E-01_EB, 4.022801E-01_EB, 3.523929E-01_EB, 3.072101E-01_EB, 2.674592E-01_EB, &
   2.326179E-01_EB, 2.012004E-01_EB, 1.742107E-01_EB, 1.505433E-01_EB, 1.298551E-01_EB, 1.119451E-01_EB, &
   9.604141E-02_EB, 8.242372E-02_EB, 7.089244E-02_EB, 6.110245E-02_EB, 5.221546E-02_EB,  &
   3.658653E-01_EB, 4.184364E-01_EB, 4.441884E-01_EB, 4.457809E-01_EB, 4.319013E-01_EB, 4.088715E-01_EB, &
   3.793819E-01_EB, 3.474065E-01_EB, 3.151904E-01_EB, 2.830904E-01_EB, 2.516607E-01_EB, 2.228145E-01_EB, &
   1.964195E-01_EB, 1.718178E-01_EB, 1.502324E-01_EB, 1.307846E-01_EB, 1.135071E-01_EB, 9.842138E-02_EB, &
   8.481592E-02_EB, 7.309981E-02_EB, 6.307130E-02_EB, 5.451710E-02_EB, 4.669207E-02_EB,  &
   8.373191E-02_EB, 1.229238E-01_EB, 1.573970E-01_EB, 1.825905E-01_EB, 1.975753E-01_EB, 2.040052E-01_EB, &
   2.024706E-01_EB, 1.957505E-01_EB, 1.855755E-01_EB, 1.726368E-01_EB, 1.580202E-01_EB, 1.433211E-01_EB, &
   1.287520E-01_EB, 1.144879E-01_EB, 1.013973E-01_EB, 8.925648E-02_EB, 7.818868E-02_EB, 6.826146E-02_EB, &
   5.920229E-02_EB, 5.129713E-02_EB, 4.441206E-02_EB, 3.851923E-02_EB, 3.306979E-02_EB,  &
   1.457007E-02_EB, 2.736439E-02_EB, 4.340085E-02_EB, 6.002981E-02_EB, 7.506875E-02_EB, 8.727601E-02_EB, &
   9.581284E-02_EB, 1.007875E-01_EB, 1.027285E-01_EB, 1.017913E-01_EB, 9.839048E-02_EB, 9.365514E-02_EB, &
   8.789001E-02_EB, 8.123320E-02_EB, 7.449732E-02_EB, 6.766345E-02_EB, 6.100922E-02_EB, 5.470686E-02_EB, &
   4.865358E-02_EB, 4.309088E-02_EB, 3.810993E-02_EB, 3.369467E-02_EB, 2.945916E-02_EB,  &
   1.191095E-03_EB, 1.552045E-03_EB, 2.059789E-03_EB, 2.649547E-03_EB, 3.236595E-03_EB, 3.762561E-03_EB, &
   4.165025E-03_EB, 4.448429E-03_EB, 4.600555E-03_EB, 4.628224E-03_EB, 4.534948E-03_EB, 4.374710E-03_EB, &
   4.151710E-03_EB, 3.880527E-03_EB, 3.599880E-03_EB, 3.300161E-03_EB, 2.998249E-03_EB, 2.715486E-03_EB, &
   2.425276E-03_EB, 2.164309E-03_EB, 1.929373E-03_EB, 1.713087E-03_EB, 1.506010E-03_EB,  &
   4.538291E-04_EB, 4.599393E-04_EB, 4.587246E-04_EB, 4.526653E-04_EB, 4.375098E-04_EB, 4.167315E-04_EB, &
   3.918634E-04_EB, 3.644121E-04_EB, 3.335314E-04_EB, 3.072189E-04_EB, 2.753771E-04_EB, 2.453914E-04_EB, &
   2.183408E-04_EB, 1.968316E-04_EB, 1.704817E-04_EB, 1.507858E-04_EB, 1.319531E-04_EB, 1.134243E-04_EB, &
   9.801919E-05_EB, 8.900765E-05_EB, 7.481087E-05_EB, 6.511635E-05_EB, 5.390619E-05_EB,  &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, &
   0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB, 0.000000E+00_EB/),(/23,7/))

!-------------------------ETHANE DATA-------------------


! THERE ARE 3 BANDS FOR ETHANE

! BAND #1: 730 CM-1 - 1095 CM-1
! BAND #2: 1250 CM-1 - 1700 CM-1
! BAND #3: 2550 CM-1 - 3375 CM-1

ALLOCATE(SD_C2H6_TEMP(N_TEMP_C2H6))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR ETHANE ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_C2H6(N_BAND_C2H6,3))
ALLOCATE(BE_C2H6(N_BAND_C2H6))

OM_BND_C2H6 = RESHAPE((/ &
   730._EB, 1250._EB, 2550._EB, &
   1095._EB, 1700._EB, 3375._EB, &
   5._EB, 25._EB, 25._EB/),(/N_BAND_C2H6,3/))

SD_C2H6_TEMP = (/ &
   296._EB, 400._EB, 450._EB, 500._EB, 600._EB, 800._EB,&
   1000._EB/)

BE_C2H6 = (/ &
   1.000_EB, 1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_C2H6(N_TEMP_C2H6,74))

! BAND #1: 730 CM-1 - 1095 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.69096 %

SD1_C2H6(1:7,1:8) = RESHAPE((/ &  ! 730-765 CM-1
   2.794554E-04_EB, 3.019036E-04_EB, 3.080672E-04_EB, 2.762484E-04_EB, 2.579677E-04_EB, 3.235741E-04_EB, &
   3.596358E-04_EB, &
   3.869441E-04_EB, 3.054332E-04_EB, 4.065149E-04_EB, 3.649137E-04_EB, 3.706177E-04_EB, 3.794478E-04_EB, &
   3.418980E-04_EB, &
   3.961398E-04_EB, 3.859808E-04_EB, 4.060006E-04_EB, 3.536294E-03_EB, 3.591171E-04_EB, 3.754312E-04_EB, &
   3.669844E-04_EB, &
   7.822607E-04_EB, 3.453771E-04_EB, 2.262803E-02_EB, 1.021724E-02_EB, 1.653952E-02_EB, 6.505243E-03_EB, &
   3.164663E-04_EB, &
   1.410615E-02_EB, 3.210967E-04_EB, 6.360977E-02_EB, 1.166115E-02_EB, 1.932412E+00_EB, 1.575805E-02_EB, &
   3.405423E-04_EB, &
   1.497481E-02_EB, 3.592094E-04_EB, 1.309596E-02_EB, 7.475906E-03_EB, 1.892167E+00_EB, 1.483042E-02_EB, &
   2.983060E-04_EB, &
   4.020853E-02_EB, 4.386772E-04_EB, 7.714318E-01_EB, 2.671185E-02_EB, 1.498279E+00_EB, 5.703491E-03_EB, &
   4.203063E-04_EB, &
   5.124663E-02_EB, 1.551881E-01_EB, 3.195236E-01_EB, 2.525727E-02_EB, 1.027149E+00_EB, 1.160139E-02_EB, &
   3.688548E-04_EB/),(/7,8/))

SD1_C2H6(1:7,9:16) = RESHAPE((/ &  ! 770-805 CM-1
   6.559937E-02_EB, 3.631953E-01_EB, 1.309011E-01_EB, 5.307552E-02_EB, 1.606841E-01_EB, 1.891214E-02_EB, &
   3.202870E-04_EB, &
   1.060830E-01_EB, 2.002917E-01_EB, 3.031810E-01_EB, 6.540746E-02_EB, 2.598325E-01_EB, 3.968316E-01_EB, &
   3.244087E-04_EB, &
   1.068733E-01_EB, 2.931690E-01_EB, 1.405611E-01_EB, 5.788228E-02_EB, 8.277802E-01_EB, 9.351959E-01_EB, &
   3.515174E-04_EB, &
   1.515345E-01_EB, 2.276208E-01_EB, 1.256607E-01_EB, 4.591616E-01_EB, 2.628483E+00_EB, 8.399321E-01_EB, &
   3.078236E-04_EB, &
   1.710208E-01_EB, 7.250193E-01_EB, 9.620347E-02_EB, 1.004229E-01_EB, 1.608952E+00_EB, 9.263801E-01_EB, &
   4.007032E-04_EB, &
   2.129896E-01_EB, 3.313998E-01_EB, 1.381292E-01_EB, 2.290180E-01_EB, 1.075200E+00_EB, 7.619251E-01_EB, &
   4.125753E-04_EB, &
   2.536966E-01_EB, 2.974149E-01_EB, 1.376301E-01_EB, 2.140445E-01_EB, 1.112248E+00_EB, 1.347530E+00_EB, &
   3.626265E-04_EB, &
   2.932385E-01_EB, 2.623213E-01_EB, 1.391961E-01_EB, 1.394845E-01_EB, 4.998654E-01_EB, 1.363158E+00_EB, &
   3.577957E-04_EB/),(/7,8/))

SD1_C2H6(1:7,17:24) = RESHAPE((/ &  ! 810-845 CM-1
   3.183925E-01_EB, 2.940497E-01_EB, 1.942798E-01_EB, 1.955972E-01_EB, 2.139000E-01_EB, 1.687942E+00_EB, &
   4.048250E-04_EB, &
   3.562908E-01_EB, 2.514123E-01_EB, 1.948319E-01_EB, 2.290250E-01_EB, 6.104920E-01_EB, 1.456760E+00_EB, &
   3.298797E-04_EB, &
   3.457713E-01_EB, 2.350119E-01_EB, 2.001751E-01_EB, 3.161435E-01_EB, 5.531987E-01_EB, 1.962497E-01_EB, &
   3.488040E-04_EB, &
   3.720387E-01_EB, 2.431526E-01_EB, 2.192578E-01_EB, 1.640348E-01_EB, 9.663389E-01_EB, 6.727370E-02_EB, &
   3.578073E-04_EB, &
   3.527769E-01_EB, 2.743749E-01_EB, 2.170241E-01_EB, 1.992689E-01_EB, 6.608995E-01_EB, 6.062421E-02_EB, &
   3.421209E-04_EB, &
   3.345994E-01_EB, 2.376781E-01_EB, 2.379620E-01_EB, 2.238005E-01_EB, 1.419365E-01_EB, 6.753641E-02_EB, &
   2.941644E-04_EB, &
   3.110441E-01_EB, 2.342208E-01_EB, 1.884871E-01_EB, 2.033274E-01_EB, 3.626503E-01_EB, 7.846996E-02_EB, &
   3.468572E-03_EB, &
   2.854259E-01_EB, 2.338848E-01_EB, 2.014852E-01_EB, 1.903667E-01_EB, 4.180199E-01_EB, 1.112186E-01_EB, &
   1.298106E-02_EB/),(/7,8/))

SD1_C2H6(1:7,25:32) = RESHAPE((/ &  ! 850-885 CM-1
   2.751807E-01_EB, 2.333134E-01_EB, 1.748430E-01_EB, 1.437778E-01_EB, 2.133839E-01_EB, 1.309889E-01_EB, &
   2.178703E-02_EB, &
   2.593350E-01_EB, 2.501009E-01_EB, 1.770492E-01_EB, 1.695213E-01_EB, 1.244732E-01_EB, 1.113414E-01_EB, &
   4.564412E-02_EB, &
   2.216701E-01_EB, 1.761779E-01_EB, 1.650476E-01_EB, 1.312825E-01_EB, 1.553354E-01_EB, 1.429574E-01_EB, &
   3.560235E-02_EB, &
   2.022044E-01_EB, 1.510443E-01_EB, 2.176619E-01_EB, 1.572246E-01_EB, 1.135620E-01_EB, 7.703252E-02_EB, &
   2.142633E-02_EB, &
   1.812600E-01_EB, 1.353210E-01_EB, 1.613117E-01_EB, 2.734768E-01_EB, 3.022567E-01_EB, 6.876571E-02_EB, &
   2.406288E-02_EB, &
   1.343889E-01_EB, 1.259396E-01_EB, 1.578672E-01_EB, 1.373682E-01_EB, 2.232547E-01_EB, 6.066181E-02_EB, &
   5.509858E-02_EB, &
   1.095603E-01_EB, 8.840111E-02_EB, 1.614927E-01_EB, 1.024326E-01_EB, 4.054561E-01_EB, 5.622874E-02_EB, &
   8.081915E-02_EB, &
   8.280573E-02_EB, 1.072322E-01_EB, 1.083568E-01_EB, 8.976527E-02_EB, 7.891590E-02_EB, 8.667161E-02_EB, &
   1.032621E-01_EB/),(/7,8/))

SD1_C2H6(1:7,33:40) = RESHAPE((/ &  ! 890-925 CM-1
   6.315700E-02_EB, 1.839109E-01_EB, 9.563206E-02_EB, 1.558355E-01_EB, 1.009801E-01_EB, 4.350811E-01_EB, &
   2.056632E-01_EB, &
   4.243703E-02_EB, 1.021804E-01_EB, 5.212402E-02_EB, 1.014279E-01_EB, 1.625632E-01_EB, 6.953883E-01_EB, &
   7.419429E-02_EB, &
   3.637108E-02_EB, 3.530674E-01_EB, 3.834881E-02_EB, 4.227429E-01_EB, 1.077391E-01_EB, 6.239971E-01_EB, &
   2.133274E-01_EB, &
   2.862217E-02_EB, 6.251917E-02_EB, 2.879530E-02_EB, 3.363107E-02_EB, 3.988006E-02_EB, 8.289923E-01_EB, &
   2.326386E-01_EB, &
   5.105750E-02_EB, 2.945599E-03_EB, 2.656328E-02_EB, 2.709003E-01_EB, 7.963168E-01_EB, 9.731125E-01_EB, &
   2.237348E-01_EB, &
   2.094406E-02_EB, 2.532325E-04_EB, 8.603313E-02_EB, 1.258408E-01_EB, 6.987088E-02_EB, 9.704437E-01_EB, &
   7.049559E-01_EB, &
   1.382001E-02_EB, 2.164374E-04_EB, 5.300388E-03_EB, 4.040185E-01_EB, 5.065259E-01_EB, 1.779278E+00_EB, &
   9.229669E-02_EB, &
   1.013101E-02_EB, 3.460394E-04_EB, 2.566832E-03_EB, 6.388567E-02_EB, 2.147372E-01_EB, 1.237580E+00_EB, &
   1.062539E-01_EB/),(/7,8/))

SD1_C2H6(1:7,41:48) = RESHAPE((/ &  ! 930-965 CM-1
   2.805218E-03_EB, 3.631322E-04_EB, 2.361065E-03_EB, 2.853649E-03_EB, 1.391981E-02_EB, 1.135302E+00_EB, &
   1.345327E-01_EB, &
   2.616753E-03_EB, 3.519962E-04_EB, 2.459903E-04_EB, 2.436123E-01_EB, 3.966290E-03_EB, 1.048639E+00_EB, &
   5.178554E-02_EB, &
   5.158410E-04_EB, 3.832391E-04_EB, 2.145139E-04_EB, 4.559609E-02_EB, 7.809468E-04_EB, 1.207831E+00_EB, &
   8.442327E-02_EB, &
   2.986755E-04_EB, 3.753807E-04_EB, 2.668659E-04_EB, 2.070056E-02_EB, 2.165559E-04_EB, 1.341963E+00_EB, &
   1.118404E-01_EB, &
   4.009616E-04_EB, 3.667449E-04_EB, 3.857376E-04_EB, 2.884928E-04_EB, 2.499297E-04_EB, 1.164307E+00_EB, &
   1.179168E-01_EB, &
   3.855251E-04_EB, 4.011809E-04_EB, 4.248678E-04_EB, 4.035946E-04_EB, 9.176744E-04_EB, 1.155933E+00_EB, &
   1.250155E-01_EB, &
   2.767518E-04_EB, 3.807194E-04_EB, 3.312430E-04_EB, 3.468054E-04_EB, 3.815901E-04_EB, 5.697397E-01_EB, &
   1.342966E-01_EB, &
   3.522300E-04_EB, 3.241922E-04_EB, 2.938979E-04_EB, 3.128230E-04_EB, 3.702311E-04_EB, 1.288695E-01_EB, &
   1.274616E-01_EB/),(/7,8/))

SD1_C2H6(1:7,49:56) = RESHAPE((/ &  ! 970-1005 CM-1
   3.734602E-04_EB, 3.927916E-04_EB, 3.867028E-04_EB, 3.488727E-04_EB, 3.129810E-04_EB, 5.653561E-03_EB, &
   1.114115E-01_EB, &
   3.090640E-04_EB, 3.182551E-04_EB, 3.234946E-04_EB, 3.977928E-04_EB, 3.092056E-04_EB, 3.223704E-04_EB, &
   2.110842E-01_EB, &
   3.694516E-04_EB, 3.386979E-04_EB, 3.942745E-04_EB, 4.268219E-04_EB, 4.823033E-04_EB, 3.062293E-04_EB, &
   1.371742E-01_EB, &
   3.449803E-04_EB, 3.878879E-04_EB, 4.122692E-04_EB, 3.550996E-04_EB, 3.586092E-04_EB, 3.305017E-04_EB, &
   4.857086E-01_EB, &
   3.341938E-04_EB, 4.611333E-04_EB, 2.824504E-04_EB, 3.440987E-04_EB, 3.632731E-04_EB, 3.190974E-04_EB, &
   1.036874E-01_EB, &
   3.732784E-04_EB, 3.561815E-04_EB, 3.078580E-04_EB, 4.006334E-04_EB, 3.049231E-04_EB, 3.978448E-04_EB, &
   4.974197E-02_EB, &
   2.979406E-04_EB, 3.116971E-04_EB, 3.889874E-04_EB, 3.786843E-04_EB, 3.072701E-04_EB, 2.867540E-04_EB, &
   4.862621E-02_EB, &
   3.325163E-04_EB, 3.138932E-04_EB, 4.250213E-04_EB, 3.119730E-04_EB, 3.732476E-04_EB, 3.717123E-04_EB, &
   4.454824E-02_EB/),(/7,8/))

SD1_C2H6(1:7,57:64) = RESHAPE((/ &  ! 1010-1045 CM-1
   3.350681E-04_EB, 3.530821E-04_EB, 4.528042E-04_EB, 3.516044E-04_EB, 3.923830E-04_EB, 3.532259E-04_EB, &
   6.542840E-02_EB, &
   3.204957E-04_EB, 3.836292E-04_EB, 3.138970E-04_EB, 3.100875E-04_EB, 3.990551E-04_EB, 3.191793E-04_EB, &
   7.406912E-02_EB, &
   3.472691E-04_EB, 3.920082E-04_EB, 3.234488E-04_EB, 4.026391E-04_EB, 3.384464E-04_EB, 2.968447E-04_EB, &
   2.322588E-01_EB, &
   3.347127E-04_EB, 3.546956E-04_EB, 2.920812E-04_EB, 4.081782E-04_EB, 3.386853E-04_EB, 3.307871E-04_EB, &
   4.463887E-01_EB, &
   3.248741E-04_EB, 4.609033E-04_EB, 3.231879E-04_EB, 3.456361E-04_EB, 3.298612E-04_EB, 3.329381E-04_EB, &
   1.035782E+00_EB, &
   4.690522E-04_EB, 3.215237E-04_EB, 4.686739E-04_EB, 3.787632E-04_EB, 3.650086E-04_EB, 3.038461E-04_EB, &
   6.376563E-01_EB, &
   3.712414E-04_EB, 3.428482E-04_EB, 3.459720E-04_EB, 3.001890E-04_EB, 3.653620E-04_EB, 3.093595E-04_EB, &
   1.783476E-01_EB, &
   3.413817E-04_EB, 3.624958E-04_EB, 3.637267E-04_EB, 3.228750E-04_EB, 2.920456E-04_EB, 3.181545E-04_EB, &
   2.620076E-01_EB/),(/7,8/))

SD1_C2H6(1:7,65:72) = RESHAPE((/ &  ! 1050-1085 CM-1
   3.325905E-04_EB, 3.330314E-04_EB, 3.617122E-04_EB, 3.505622E-04_EB, 3.869601E-04_EB, 3.838024E-04_EB, &
   5.425202E-01_EB, &
   3.314643E-04_EB, 3.409742E-04_EB, 2.922515E-04_EB, 2.908445E-04_EB, 3.423906E-04_EB, 2.970385E-04_EB, &
   6.238502E-01_EB, &
   3.492482E-04_EB, 3.606574E-04_EB, 3.274995E-04_EB, 3.213303E-04_EB, 3.179339E-04_EB, 3.480095E-04_EB, &
   4.163639E-02_EB, &
   3.199232E-04_EB, 3.955618E-04_EB, 2.798966E-04_EB, 3.155772E-04_EB, 2.916205E-04_EB, 3.756559E-04_EB, &
   1.227547E-02_EB, &
   3.705657E-04_EB, 3.404134E-04_EB, 3.590301E-04_EB, 3.440815E-04_EB, 3.347934E-04_EB, 3.248160E-04_EB, &
   1.130962E-02_EB, &
   3.884066E-04_EB, 3.017404E-04_EB, 3.954654E-04_EB, 3.017435E-04_EB, 2.609871E-04_EB, 2.759450E-04_EB, &
   2.756706E-03_EB, &
   3.676700E-04_EB, 3.019733E-04_EB, 3.533386E-04_EB, 3.287691E-04_EB, 4.033519E-04_EB, 3.370772E-04_EB, &
   3.213983E-04_EB, &
   3.262600E-04_EB, 3.615855E-04_EB, 3.142415E-04_EB, 3.309319E-04_EB, 4.228572E-04_EB, 3.478703E-04_EB, &
   3.234945E-04_EB/),(/7,8/))

SD1_C2H6(1:7,73:74) = RESHAPE((/ &  ! 1090-1095 CM-1
   2.711945E-04_EB, 2.496329E-04_EB, 3.150593E-04_EB, 2.603387E-04_EB, 2.992890E-04_EB, 3.589836E-04_EB, &
   2.656451E-04_EB, &
   2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, &
   2.145139E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_C2H6(N_TEMP_C2H6,74))

! BAND #1: 730 CM-1 - 1095 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.69096 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_C2H6(1:7,1:8) = RESHAPE((/ &  ! 730-765 CM-1
   2.006565E-04_EB, 2.337401E-04_EB, 2.177856E-04_EB, 2.048546E-04_EB, 1.869983E-04_EB, 2.713666E-04_EB, &
   2.905961E-04_EB, &
   2.306868E-04_EB, 2.167323E-04_EB, 3.350447E-04_EB, 2.526474E-04_EB, 2.261799E-04_EB, 2.480626E-04_EB, &
   2.521419E-04_EB, &
   2.854020E-04_EB, 3.329316E-04_EB, 1.598485E-04_EB, 3.267121E-02_EB, 2.234431E-04_EB, 2.957501E-04_EB, &
   2.173206E-04_EB, &
   2.137117E-04_EB, 2.266418E-04_EB, 4.728511E-05_EB, 2.245423E-01_EB, 3.064851E-05_EB, 8.222848E-03_EB, &
   1.859738E-04_EB, &
   3.462857E-03_EB, 2.519057E-04_EB, 3.480152E-05_EB, 3.925807E-02_EB, 1.572683E-05_EB, 3.101866E-01_EB, &
   1.995997E-04_EB, &
   9.741100E-03_EB, 2.710087E-04_EB, 2.101240E-03_EB, 7.469797E-02_EB, 2.264929E-05_EB, 1.275440E-01_EB, &
   1.730011E-04_EB, &
   7.831446E-03_EB, 2.356564E-04_EB, 2.537222E-05_EB, 5.290921E-02_EB, 3.544110E-05_EB, 1.081723E-01_EB, &
   2.635023E-04_EB, &
   1.830629E-02_EB, 5.427184E-06_EB, 1.638831E-04_EB, 1.590424E-02_EB, 4.986288E-05_EB, 2.169141E-01_EB, &
   2.505642E-04_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,9:16) = RESHAPE((/ &  ! 770-805 CM-1
   3.746173E-02_EB, 1.736644E-05_EB, 6.085991E-04_EB, 4.704301E-01_EB, 6.606423E-04_EB, 7.502293E-02_EB, &
   2.108317E-04_EB, &
   2.746203E-02_EB, 1.840834E-04_EB, 5.921780E-04_EB, 2.043773E-02_EB, 6.107989E-04_EB, 1.388557E-04_EB, &
   1.893995E-04_EB, &
   2.420680E-01_EB, 4.191667E-04_EB, 3.132099E-03_EB, 1.886145E+00_EB, 3.070676E-04_EB, 1.079068E-04_EB, &
   2.691653E-04_EB, &
   8.925252E-02_EB, 1.124913E-03_EB, 8.738915E-03_EB, 1.559750E-03_EB, 2.395539E-04_EB, 1.072791E-04_EB, &
   1.752395E-04_EB, &
   1.067081E-01_EB, 6.895830E-04_EB, 4.690182E-02_EB, 4.174376E-02_EB, 4.527850E-04_EB, 2.137964E-04_EB, &
   2.588101E-04_EB, &
   8.746003E-02_EB, 2.440063E-03_EB, 2.130236E-02_EB, 6.354666E-03_EB, 6.671299E-04_EB, 3.378605E-04_EB, &
   2.343097E-04_EB, &
   9.852102E-02_EB, 4.561646E-03_EB, 3.949280E-02_EB, 1.123806E-02_EB, 8.233218E-04_EB, 2.522466E-04_EB, &
   2.226954E-04_EB, &
   1.207618E-01_EB, 8.941124E-03_EB, 1.085484E-01_EB, 3.349804E-02_EB, 1.725479E-03_EB, 3.076476E-04_EB, &
   1.938371E-04_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,17:24) = RESHAPE((/ &  ! 810-845 CM-1
   1.241365E-01_EB, 9.920833E-03_EB, 2.540899E-02_EB, 1.887039E-02_EB, 6.826400E-03_EB, 2.126909E-04_EB, &
   2.773335E-04_EB, &
   1.417207E-01_EB, 2.020016E-02_EB, 3.589721E-02_EB, 1.758195E-02_EB, 1.968415E-03_EB, 2.289210E-04_EB, &
   2.228792E-04_EB, &
   1.684374E-01_EB, 2.315876E-02_EB, 3.432510E-02_EB, 1.023915E-02_EB, 2.040480E-03_EB, 2.528208E-03_EB, &
   2.461503E-04_EB, &
   2.993636E-01_EB, 4.679863E-02_EB, 4.993611E-02_EB, 5.799012E-01_EB, 1.375382E-03_EB, 5.540907E+00_EB, &
   2.729510E-04_EB, &
   3.763363E-01_EB, 3.147273E-02_EB, 5.066712E-02_EB, 5.649368E-02_EB, 2.734156E-03_EB, 8.088974E+00_EB, &
   2.352079E-04_EB, &
   3.923696E-01_EB, 4.535395E-02_EB, 3.848111E-02_EB, 2.837120E-02_EB, 3.903926E-02_EB, 4.193583E+00_EB, &
   1.708561E-04_EB, &
   6.116601E-01_EB, 3.911633E-02_EB, 9.404679E-02_EB, 3.598464E-02_EB, 4.985811E-03_EB, 2.536374E+00_EB, &
   1.671554E-02_EB, &
   6.488105E-01_EB, 3.434365E-02_EB, 5.620492E-02_EB, 4.579674E-02_EB, 3.831981E-03_EB, 2.040147E-02_EB, &
   1.535947E-01_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,25:32) = RESHAPE((/ &  ! 850-885 CM-1
   2.185282E-01_EB, 2.624089E-02_EB, 8.281180E-02_EB, 1.337021E-01_EB, 1.164389E-02_EB, 1.441667E-02_EB, &
   1.326192E+00_EB, &
   1.054436E-01_EB, 1.564415E-02_EB, 4.232910E-02_EB, 2.542335E-02_EB, 9.182689E-02_EB, 2.979608E-02_EB, &
   3.412059E+00_EB, &
   1.037085E-01_EB, 2.307616E-02_EB, 2.791608E-02_EB, 7.390912E-02_EB, 1.792121E-02_EB, 8.709794E-03_EB, &
   9.123564E-01_EB, &
   6.557347E-02_EB, 2.400623E-02_EB, 1.184967E-02_EB, 1.759675E-02_EB, 4.025549E-02_EB, 1.650755E-01_EB, &
   4.126380E-01_EB, &
   3.854672E-02_EB, 1.545381E-02_EB, 1.176956E-02_EB, 5.353178E-03_EB, 3.790587E-03_EB, 2.857260E+00_EB, &
   4.896543E-02_EB, &
   3.823510E-02_EB, 1.163165E-02_EB, 7.527145E-03_EB, 1.044536E-02_EB, 4.277846E-03_EB, 2.782661E+00_EB, &
   9.862962E-03_EB, &
   2.698248E-02_EB, 1.682716E-02_EB, 4.532103E-03_EB, 2.286498E-02_EB, 1.772522E-03_EB, 2.075720E+00_EB, &
   2.124687E-02_EB, &
   2.066380E-02_EB, 4.723103E-03_EB, 6.343220E-03_EB, 2.397806E-02_EB, 2.607826E-02_EB, 7.702274E-03_EB, &
   1.637205E-02_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,33:40) = RESHAPE((/ &  ! 890-925 CM-1
   1.658402E-02_EB, 9.717357E-04_EB, 4.999287E-03_EB, 3.360172E-03_EB, 7.640958E-03_EB, 5.866049E-04_EB, &
   4.203039E-03_EB, &
   9.306719E-03_EB, 1.059841E-03_EB, 1.554113E-02_EB, 4.017085E-03_EB, 1.562439E-03_EB, 2.629554E-04_EB, &
   2.723941E-02_EB, &
   6.324845E-03_EB, 1.323745E-04_EB, 2.015294E-02_EB, 4.021824E-04_EB, 2.200845E-03_EB, 1.909660E-04_EB, &
   2.384092E-03_EB, &
   6.442116E-03_EB, 1.976663E-04_EB, 1.301740E-02_EB, 7.108478E-03_EB, 9.667620E-03_EB, 1.093856E-04_EB, &
   1.651473E-03_EB, &
   6.177417E-04_EB, 5.252625E-04_EB, 3.292456E-03_EB, 9.638064E-05_EB, 1.157061E-04_EB, 9.272832E-05_EB, &
   1.772000E-03_EB, &
   3.119150E-03_EB, 1.688032E-04_EB, 1.157769E-04_EB, 4.344200E-05_EB, 6.951343E-04_EB, 9.547409E-05_EB, &
   5.716592E-04_EB, &
   3.033519E-03_EB, 1.700977E-04_EB, 1.923551E-03_EB, 1.441479E-05_EB, 9.010921E-05_EB, 4.649518E-05_EB, &
   3.993180E-03_EB, &
   4.505246E-03_EB, 2.206088E-04_EB, 4.709851E-04_EB, 4.044407E-05_EB, 9.427052E-05_EB, 4.271051E-05_EB, &
   4.728030E-03_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,41:48) = RESHAPE((/ &  ! 930-965 CM-1
   8.515529E-04_EB, 2.523157E-04_EB, 3.215017E-03_EB, 8.896712E-05_EB, 2.653323E-03_EB, 2.395564E-05_EB, &
   5.128282E-03_EB, &
   1.598460E-03_EB, 2.391773E-04_EB, 2.024249E-04_EB, 4.570502E-06_EB, 1.197295E-03_EB, 2.420553E-05_EB, &
   7.176382E-01_EB, &
   5.405162E-04_EB, 2.754095E-04_EB, 1.688326E-04_EB, 1.901197E-05_EB, 1.882451E-04_EB, 2.561929E-05_EB, &
   1.225580E+00_EB, &
   2.092858E-04_EB, 2.662552E-04_EB, 2.026153E-04_EB, 1.070822E-05_EB, 1.684702E-04_EB, 2.442314E-05_EB, &
   2.444828E+01_EB, &
   2.529925E-04_EB, 2.765314E-04_EB, 2.642207E-04_EB, 2.200703E-04_EB, 1.995939E-04_EB, 2.380865E-05_EB, &
   2.685955E+01_EB, &
   2.173973E-04_EB, 2.813395E-04_EB, 2.874743E-04_EB, 3.035869E-04_EB, 6.183918E-04_EB, 2.202463E-05_EB, &
   2.519809E+00_EB, &
   1.980501E-04_EB, 2.678571E-04_EB, 2.137226E-04_EB, 2.556760E-04_EB, 2.222342E-04_EB, 1.634015E-05_EB, &
   3.727180E-02_EB, &
   2.678728E-04_EB, 2.424340E-04_EB, 1.979944E-04_EB, 2.135470E-04_EB, 2.146239E-04_EB, 1.781477E-05_EB, &
   7.049181E-03_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,49:56) = RESHAPE((/ &  ! 970-1005 CM-1
   2.364062E-04_EB, 2.712640E-04_EB, 2.842134E-04_EB, 2.567382E-04_EB, 2.256683E-04_EB, 2.498984E-05_EB, &
   4.004321E-03_EB, &
   2.157659E-04_EB, 2.064345E-04_EB, 2.288569E-04_EB, 2.738793E-04_EB, 2.050141E-04_EB, 2.197032E-04_EB, &
   1.028210E-03_EB, &
   2.197468E-04_EB, 2.679068E-04_EB, 2.304862E-04_EB, 2.905492E-04_EB, 2.725375E-04_EB, 2.159399E-04_EB, &
   1.419923E-03_EB, &
   2.335319E-04_EB, 2.681889E-04_EB, 3.239557E-04_EB, 2.569741E-04_EB, 2.238526E-04_EB, 2.007748E-04_EB, &
   2.537471E-04_EB, &
   2.263569E-04_EB, 3.026600E-04_EB, 2.089141E-04_EB, 2.243273E-04_EB, 2.228419E-04_EB, 2.279698E-04_EB, &
   7.317457E-04_EB, &
   2.753690E-04_EB, 2.292202E-04_EB, 2.318102E-04_EB, 2.597276E-04_EB, 2.082393E-04_EB, 2.415048E-04_EB, &
   2.144539E-03_EB, &
   2.130814E-04_EB, 2.160667E-04_EB, 3.005135E-04_EB, 2.457929E-04_EB, 2.262402E-04_EB, 2.271924E-04_EB, &
   1.982460E-03_EB, &
   2.260545E-04_EB, 2.078552E-04_EB, 2.252198E-04_EB, 2.292910E-04_EB, 2.617146E-04_EB, 2.719938E-04_EB, &
   9.539658E-03_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,57:64) = RESHAPE((/ &  ! 1010-1045 CM-1
   2.185131E-04_EB, 2.784174E-04_EB, 2.724240E-04_EB, 2.166217E-04_EB, 2.454826E-04_EB, 2.343101E-04_EB, &
   4.681303E-03_EB, &
   2.140851E-04_EB, 2.908903E-04_EB, 2.238573E-04_EB, 1.936346E-04_EB, 2.929472E-04_EB, 1.997998E-04_EB, &
   5.199463E-03_EB, &
   2.704701E-04_EB, 2.762782E-04_EB, 2.520038E-04_EB, 2.774566E-04_EB, 2.294524E-04_EB, 1.897369E-04_EB, &
   1.101146E-03_EB, &
   2.261254E-04_EB, 2.439262E-04_EB, 2.109405E-04_EB, 2.752945E-04_EB, 2.380712E-04_EB, 2.219084E-04_EB, &
   4.729845E-04_EB, &
   2.442952E-04_EB, 3.469938E-04_EB, 2.322461E-04_EB, 2.311834E-04_EB, 2.364839E-04_EB, 2.042171E-04_EB, &
   2.219015E-04_EB, &
   2.634592E-04_EB, 2.110525E-04_EB, 3.161689E-04_EB, 2.414846E-04_EB, 2.423626E-04_EB, 2.110280E-04_EB, &
   3.145896E-04_EB, &
   2.659955E-04_EB, 2.470635E-04_EB, 2.700914E-04_EB, 2.258166E-04_EB, 2.553061E-04_EB, 2.172449E-04_EB, &
   8.521657E-04_EB, &
   2.210763E-04_EB, 2.410264E-04_EB, 2.690630E-04_EB, 2.296574E-04_EB, 1.980775E-04_EB, 2.236759E-04_EB, &
   5.582951E-04_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,65:72) = RESHAPE((/ &  ! 1050-1085 CM-1
   2.066800E-04_EB, 2.209692E-04_EB, 2.738395E-04_EB, 2.085763E-04_EB, 2.168987E-04_EB, 2.377505E-04_EB, &
   1.629293E-04_EB, &
   2.355869E-04_EB, 2.212229E-04_EB, 2.006984E-04_EB, 2.157291E-04_EB, 2.203006E-04_EB, 2.060948E-04_EB, &
   9.639746E-05_EB, &
   2.268691E-04_EB, 2.800409E-04_EB, 1.984453E-04_EB, 2.045951E-04_EB, 1.977593E-04_EB, 2.051878E-04_EB, &
   9.621759E-04_EB, &
   2.273679E-04_EB, 2.692054E-04_EB, 2.117256E-04_EB, 2.039894E-04_EB, 1.777981E-04_EB, 2.609597E-04_EB, &
   1.607451E-03_EB, &
   2.293047E-04_EB, 2.421597E-04_EB, 2.833909E-04_EB, 2.164108E-04_EB, 2.199777E-04_EB, 2.398280E-04_EB, &
   1.282936E-03_EB, &
   2.497990E-04_EB, 2.311811E-04_EB, 2.716080E-04_EB, 2.074448E-04_EB, 1.857482E-04_EB, 1.792751E-04_EB, &
   9.092886E-04_EB, &
   2.440601E-04_EB, 2.141159E-04_EB, 2.414717E-04_EB, 2.388990E-04_EB, 2.516826E-04_EB, 2.158533E-04_EB, &
   2.049667E-04_EB, &
   2.220025E-04_EB, 2.592544E-04_EB, 1.988151E-04_EB, 1.978750E-04_EB, 2.337158E-04_EB, 2.300715E-04_EB, &
   1.991380E-04_EB/),(/7,8/))

GAMMAD1_C2H6(1:7,73:74) = RESHAPE((/ &  ! 1090-1095 CM-1
   1.847719E-04_EB, 1.895118E-04_EB, 2.362824E-04_EB, 1.933361E-04_EB, 2.035927E-04_EB, 2.735696E-04_EB, &
   2.160458E-04_EB, &
   1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, &
   1.688326E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_C2H6(N_TEMP_C2H6,19))

! BAND #2: 1250 CM-1 - 1700 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 1.536 %

SD2_C2H6(1:7,1:8) = RESHAPE((/ &  ! 1250-1425 CM-1
   2.890736E-04_EB, 2.831224E-04_EB, 2.500860E-04_EB, 2.757140E-04_EB, 3.408382E-04_EB, 8.798483E-04_EB, &
   9.996940E-04_EB, &
   3.173085E-04_EB, 3.511457E-04_EB, 3.615941E-04_EB, 3.532323E-04_EB, 3.063903E-03_EB, 5.665886E-01_EB, &
   6.009682E-01_EB, &
   7.762194E-03_EB, 3.886970E-04_EB, 3.445713E-04_EB, 3.346007E-04_EB, 9.791971E-03_EB, 3.101724E-01_EB, &
   3.854181E-01_EB, &
   5.755406E-03_EB, 2.965005E-03_EB, 3.003864E-03_EB, 2.921568E-02_EB, 2.167033E-02_EB, 2.481666E-01_EB, &
   4.975186E-01_EB, &
   2.729491E-02_EB, 4.272374E-02_EB, 4.078658E-02_EB, 5.957540E-02_EB, 5.312857E-02_EB, 7.121161E-02_EB, &
   4.478193E-02_EB, &
   1.116554E-01_EB, 1.211306E-01_EB, 1.109669E-01_EB, 1.148103E-01_EB, 9.457225E-02_EB, 9.397124E-02_EB, &
   6.487302E-02_EB, &
   2.433578E-01_EB, 2.251434E-01_EB, 1.980541E-01_EB, 1.663919E-01_EB, 1.429235E-01_EB, 1.679094E-01_EB, &
   1.043870E-01_EB, &
   3.085264E-01_EB, 2.668615E-01_EB, 2.316373E-01_EB, 1.816814E-01_EB, 1.675634E-01_EB, 1.397869E-01_EB, &
   9.888312E-02_EB/),(/7,8/))

SD2_C2H6(1:7,9:16) = RESHAPE((/ &  ! 1450-1625 CM-1
   4.739873E-01_EB, 3.630012E-01_EB, 2.939668E-01_EB, 2.366840E-01_EB, 2.000756E-01_EB, 1.538908E-01_EB, &
   1.749130E-01_EB, &
   5.226241E-01_EB, 3.793992E-01_EB, 3.109342E-01_EB, 2.476162E-01_EB, 1.999044E-01_EB, 1.438092E-01_EB, &
   3.486559E-01_EB, &
   4.230308E-01_EB, 3.178107E-01_EB, 2.777402E-01_EB, 1.894468E-01_EB, 1.805602E-01_EB, 1.554106E-01_EB, &
   1.422521E-01_EB, &
   2.761096E-01_EB, 2.597293E-01_EB, 2.257682E-01_EB, 1.514748E-01_EB, 1.401495E-01_EB, 1.253687E-01_EB, &
   7.720757E-02_EB, &
   1.111228E-01_EB, 1.672047E-01_EB, 1.760444E-01_EB, 8.528705E-02_EB, 1.082694E-01_EB, 9.903350E-02_EB, &
   6.563257E-02_EB, &
   3.538374E-02_EB, 1.038124E-01_EB, 5.979240E-02_EB, 7.230511E-02_EB, 5.579668E-02_EB, 6.549601E-02_EB, &
   1.002546E-01_EB, &
   4.798824E-03_EB, 1.584507E-02_EB, 1.391213E-02_EB, 7.436389E-01_EB, 2.920638E-02_EB, 1.145160E-01_EB, &
   1.008028E-01_EB, &
   3.514365E-04_EB, 3.078667E-04_EB, 6.289715E-04_EB, 7.024049E-03_EB, 1.079463E-02_EB, 2.595491E-01_EB, &
   1.608240E-02_EB/),(/7,8/))

SD2_C2H6(1:7,17:19) = RESHAPE((/ &  ! 1650-1700 CM-1
   3.438068E-04_EB, 3.506371E-04_EB, 3.282171E-04_EB, 3.346269E-04_EB, 1.558845E-03_EB, 6.506371E-02_EB, &
   2.338883E-01_EB, &
   2.659394E-04_EB, 3.048186E-04_EB, 2.844540E-04_EB, 2.662972E-04_EB, 3.138724E-04_EB, 3.077023E-03_EB, &
   7.859120E-03_EB, &
   2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, &
   2.145139E-04_EB/),(/7,3/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_C2H6(N_TEMP_C2H6,19))

! BAND #2: 1250 CM-1 - 1700 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 1.536 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_C2H6(1:7,1:8) = RESHAPE((/ &  ! 1250-1425 CM-1
   2.119975E-04_EB, 2.146327E-04_EB, 1.903252E-04_EB, 2.022541E-04_EB, 2.602782E-04_EB, 4.275791E-04_EB, &
   2.592334E-04_EB, &
   2.197390E-04_EB, 2.683644E-04_EB, 2.472885E-04_EB, 2.304333E-04_EB, 8.267459E-03_EB, 1.628893E-05_EB, &
   3.112078E-05_EB, &
   1.421119E-02_EB, 2.763312E-04_EB, 2.321580E-04_EB, 2.291520E-04_EB, 1.958198E-02_EB, 6.419678E-05_EB, &
   1.177209E-04_EB, &
   5.134518E-02_EB, 1.180015E-03_EB, 5.665049E-03_EB, 9.959465E-05_EB, 7.884425E-02_EB, 2.789112E-04_EB, &
   2.062441E-04_EB, &
   6.493055E-01_EB, 8.533910E-03_EB, 1.523050E-02_EB, 5.127453E-03_EB, 6.635840E-01_EB, 1.697909E-02_EB, &
   1.007998E+00_EB, &
   6.644469E+00_EB, 5.703315E-02_EB, 7.526340E-02_EB, 3.982739E-02_EB, 2.373472E+00_EB, 2.681325E-02_EB, &
   6.481877E-01_EB, &
   4.838580E+01_EB, 1.082873E-01_EB, 1.011916E-01_EB, 2.605553E-01_EB, 2.237840E+00_EB, 1.611963E-02_EB, &
   3.893042E-01_EB, &
   7.997133E+01_EB, 1.230787E-01_EB, 1.231629E-01_EB, 3.381738E+00_EB, 2.828329E-01_EB, 5.186004E-02_EB, &
   4.989001E-02_EB/),(/7,8/))

GAMMAD2_C2H6(1:7,9:16) = RESHAPE((/ &  ! 1450-1625 CM-1
   7.999999E+01_EB, 1.520030E-01_EB, 1.960215E-01_EB, 3.752618E-01_EB, 2.469986E-01_EB, 4.882190E-02_EB, &
   1.110526E-02_EB, &
   7.998354E+01_EB, 1.855013E-01_EB, 1.836005E-01_EB, 2.792553E-01_EB, 2.412743E-01_EB, 4.914911E-02_EB, &
   2.585712E-03_EB, &
   8.000000E+01_EB, 1.653039E-01_EB, 1.276875E-01_EB, 4.450008E+01_EB, 1.282813E-01_EB, 1.923521E-02_EB, &
   9.527892E-03_EB, &
   8.000000E+01_EB, 7.676701E-02_EB, 7.528859E-02_EB, 3.303007E+01_EB, 3.079957E-01_EB, 2.422473E-02_EB, &
   3.925627E-02_EB, &
   7.999946E+01_EB, 2.618451E-02_EB, 2.459894E-02_EB, 1.811547E+01_EB, 7.028942E-02_EB, 2.413638E-02_EB, &
   3.482649E-02_EB, &
   3.034575E+00_EB, 3.670718E-03_EB, 3.712350E-02_EB, 1.721291E-02_EB, 2.149261E+00_EB, 1.626810E-02_EB, &
   2.799184E-03_EB, &
   2.512595E-02_EB, 8.026413E-04_EB, 5.409205E-02_EB, 1.139540E-04_EB, 7.373019E-01_EB, 1.105238E-03_EB, &
   1.244382E-03_EB, &
   2.370622E-04_EB, 2.109883E-04_EB, 4.709529E-04_EB, 1.239090E-03_EB, 2.096932E-01_EB, 1.101878E-04_EB, &
   6.601257E-03_EB/),(/7,8/))

GAMMAD2_C2H6(1:7,17:19) = RESHAPE((/ &  ! 1650-1700 CM-1
   2.490089E-04_EB, 2.329594E-04_EB, 2.331607E-04_EB, 2.360833E-04_EB, 1.239584E-03_EB, 1.730171E-04_EB, &
   6.965657E-05_EB, &
   1.969514E-04_EB, 2.180366E-04_EB, 2.121351E-04_EB, 1.905845E-04_EB, 2.199848E-04_EB, 9.700224E-04_EB, &
   1.623935E-03_EB, &
   1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, &
   1.688326E-04_EB/),(/7,3/))

!---------------------------------------------------------------------------
ALLOCATE(SD3_C2H6(N_TEMP_C2H6,34))

! BAND #3: 2550 CM-1 - 3375 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.3765 %

SD3_C2H6(1:7,1:8) = RESHAPE((/ &  ! 2550-2725 CM-1
   1.273212E-03_EB, 2.818357E-04_EB, 2.919712E-04_EB, 2.725676E-04_EB, 3.224456E-04_EB, 1.576641E-03_EB, &
   7.098264E-02_EB, &
   1.760425E-02_EB, 2.470202E-03_EB, 6.517460E-04_EB, 1.488758E-03_EB, 2.133431E-03_EB, 7.058130E-03_EB, &
   5.189907E-01_EB, &
   1.544358E-02_EB, 7.033132E-03_EB, 6.497614E-03_EB, 7.444432E-03_EB, 7.481133E-03_EB, 9.709862E-03_EB, &
   1.866477E-01_EB, &
   2.776353E-02_EB, 1.788161E-02_EB, 1.440452E-02_EB, 1.497498E-02_EB, 1.098338E-02_EB, 1.398847E-02_EB, &
   1.830505E-01_EB, &
   4.053879E-02_EB, 2.592830E-02_EB, 1.797176E-02_EB, 1.619226E-02_EB, 1.327884E-02_EB, 1.766187E-02_EB, &
   1.018847E-02_EB, &
   4.388545E-02_EB, 3.290936E-02_EB, 2.841386E-02_EB, 1.850778E-02_EB, 1.703843E-02_EB, 2.108276E-02_EB, &
   2.254688E-02_EB, &
   3.693893E-02_EB, 3.278040E-02_EB, 3.020784E-02_EB, 2.524477E-02_EB, 2.353619E-02_EB, 2.930089E-02_EB, &
   3.312656E-02_EB, &
   5.710175E-02_EB, 5.583766E-02_EB, 4.445106E-02_EB, 4.789108E-02_EB, 3.783170E-02_EB, 4.950327E-02_EB, &
   5.137114E-02_EB/),(/7,8/))

SD3_C2H6(1:7,9:16) = RESHAPE((/ &  ! 2750-2925 CM-1
   9.437750E-02_EB, 7.826799E-02_EB, 6.481721E-02_EB, 7.020402E-02_EB, 6.017769E-02_EB, 7.963757E-02_EB, &
   9.068663E-02_EB, &
   1.449953E-01_EB, 1.110552E-01_EB, 9.612305E-02_EB, 1.060636E-01_EB, 8.633211E-02_EB, 1.043293E-01_EB, &
   1.280871E-01_EB, &
   1.001382E-01_EB, 8.904423E-02_EB, 8.899341E-02_EB, 1.080298E-01_EB, 9.748937E-02_EB, 1.507712E-01_EB, &
   2.152193E-01_EB, &
   1.789013E-01_EB, 1.576201E-01_EB, 1.688876E-01_EB, 1.893222E-01_EB, 2.046234E-01_EB, 2.778477E-01_EB, &
   3.326868E-01_EB, &
   4.519875E-01_EB, 4.722980E-01_EB, 5.002552E-01_EB, 5.176398E-01_EB, 5.259894E-01_EB, 5.581873E-01_EB, &
   5.217227E-01_EB, &
   1.942344E+00_EB, 1.479985E+00_EB, 1.319915E+00_EB, 1.189402E+00_EB, 1.020515E+00_EB, 8.684235E-01_EB, &
   7.596269E-01_EB, &
   2.588837E+00_EB, 1.924667E+00_EB, 1.726463E+00_EB, 1.589788E+00_EB, 1.374889E+00_EB, 1.152320E+00_EB, &
   9.331464E-01_EB, &
   4.349487E+00_EB, 3.315472E+00_EB, 2.890903E+00_EB, 2.568487E+00_EB, 2.077789E+00_EB, 1.524372E+00_EB, &
   1.129152E+00_EB/),(/7,8/))

SD3_C2H6(1:7,17:24) = RESHAPE((/ &  ! 2950-3125 CM-1
   4.686779E+00_EB, 3.462217E+00_EB, 2.991646E+00_EB, 2.656886E+00_EB, 2.157561E+00_EB, 1.584726E+00_EB, &
   1.167425E+00_EB, &
   5.661792E+00_EB, 4.018616E+00_EB, 3.411886E+00_EB, 2.975545E+00_EB, 2.380752E+00_EB, 1.690561E+00_EB, &
   1.242062E+00_EB, &
   4.918064E+00_EB, 3.441837E+00_EB, 2.903757E+00_EB, 2.525786E+00_EB, 2.015014E+00_EB, 1.429458E+00_EB, &
   1.053831E+00_EB, &
   3.342233E+00_EB, 2.430853E+00_EB, 2.070409E+00_EB, 1.791861E+00_EB, 1.453571E+00_EB, 1.041375E+00_EB, &
   8.022528E-01_EB, &
   1.298165E+00_EB, 1.255308E+00_EB, 1.160176E+00_EB, 1.048498E+00_EB, 9.329619E-01_EB, 7.444407E-01_EB, &
   5.802852E-01_EB, &
   2.926016E-01_EB, 4.494914E-01_EB, 4.813448E-01_EB, 4.684675E-01_EB, 5.012802E-01_EB, 4.679462E-01_EB, &
   4.003903E-01_EB, &
   6.131343E-02_EB, 1.274991E-01_EB, 1.601441E-01_EB, 1.663262E-01_EB, 2.229073E-01_EB, 2.672821E-01_EB, &
   2.381398E-01_EB, &
   4.322732E-02_EB, 6.015505E-02_EB, 6.743728E-02_EB, 6.674353E-02_EB, 1.110929E-01_EB, 1.464795E-01_EB, &
   1.569146E-01_EB/),(/7,8/))

SD3_C2H6(1:7,25:32) = RESHAPE((/ &  ! 3150-3325 CM-1
   3.604160E-02_EB, 1.194845E-01_EB, 4.058106E-02_EB, 4.315323E-02_EB, 9.519510E-02_EB, 8.917098E-02_EB, &
   1.405300E-01_EB, &
   4.315334E-02_EB, 4.712174E-01_EB, 4.312845E-02_EB, 3.219296E-02_EB, 3.525944E-01_EB, 1.060124E-01_EB, &
   5.046058E-01_EB, &
   1.257731E-01_EB, 8.432473E-01_EB, 9.486353E-02_EB, 2.860309E-01_EB, 4.209545E-01_EB, 1.897258E-01_EB, &
   1.586487E+00_EB, &
   5.478996E-02_EB, 5.751983E-01_EB, 1.238009E-01_EB, 1.250768E-01_EB, 9.468278E-01_EB, 3.490872E-01_EB, &
   1.256817E+00_EB, &
   7.050923E-02_EB, 6.595339E-01_EB, 1.143623E-01_EB, 3.234956E-01_EB, 7.050330E-01_EB, 3.260248E-01_EB, &
   5.549485E-01_EB, &
   4.360067E-02_EB, 5.450163E-01_EB, 1.152844E-01_EB, 2.285287E-01_EB, 6.637506E-01_EB, 2.501009E-01_EB, &
   3.535923E-01_EB, &
   1.872369E-01_EB, 3.564558E-01_EB, 4.090535E-02_EB, 1.306462E-01_EB, 4.246963E-01_EB, 7.148910E-02_EB, &
   1.438647E-01_EB, &
   5.015709E-03_EB, 1.912379E-01_EB, 9.285764E-03_EB, 1.392927E-03_EB, 2.040005E-01_EB, 9.061367E-02_EB, &
   9.395617E-04_EB/),(/7,8/))

SD3_C2H6(1:7,33:34) = RESHAPE((/ &  ! 3350-3375 CM-1
   2.644024E-04_EB, 1.032051E-02_EB, 3.153028E-03_EB, 3.073098E-04_EB, 1.028061E-02_EB, 1.242207E-03_EB, &
   2.965961E-04_EB, &
   2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, 2.145139E-04_EB, &
   2.145139E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD3_C2H6(N_TEMP_C2H6,34))

! BAND #3: 2550 CM-1 - 3375 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.3765 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD3_C2H6(1:7,1:8) = RESHAPE((/ &  ! 2550-2725 CM-1
   5.906363E-04_EB, 2.102426E-04_EB, 2.132346E-04_EB, 2.065179E-04_EB, 2.655299E-04_EB, 8.306361E-04_EB, &
   3.355503E-05_EB, &
   2.516050E-04_EB, 3.355121E-03_EB, 4.172370E-04_EB, 1.011463E-03_EB, 2.596114E-02_EB, 1.182928E-03_EB, &
   2.503340E-05_EB, &
   4.163269E-03_EB, 1.125552E-01_EB, 4.980295E-03_EB, 2.696283E-03_EB, 3.105873E-02_EB, 4.712694E-03_EB, &
   6.033084E-05_EB, &
   2.623691E-02_EB, 1.923864E-02_EB, 3.655389E-02_EB, 4.141023E-03_EB, 3.609314E-02_EB, 3.824795E-03_EB, &
   1.219453E-04_EB, &
   5.265718E-02_EB, 8.436915E-03_EB, 1.960254E-02_EB, 1.875421E-03_EB, 2.111866E-02_EB, 3.776296E-03_EB, &
   3.441962E-02_EB, &
   3.937315E-02_EB, 8.289911E-03_EB, 4.840661E-03_EB, 4.421232E-03_EB, 7.814993E-02_EB, 1.662021E-02_EB, &
   1.455835E-01_EB, &
   3.720731E-02_EB, 8.723819E-03_EB, 7.153107E-03_EB, 7.013544E-03_EB, 7.319968E-02_EB, 1.767570E-01_EB, &
   8.122963E-01_EB, &
   4.868277E-02_EB, 1.269629E-02_EB, 3.095084E-02_EB, 1.285202E-02_EB, 3.311119E-01_EB, 4.745323E-02_EB, &
   5.448600E-01_EB/),(/7,8/))

GAMMAD3_C2H6(1:7,9:16) = RESHAPE((/ &  ! 2750-2925 CM-1
   7.027332E-02_EB, 3.651191E-02_EB, 1.128884E-01_EB, 2.630950E-02_EB, 2.289043E-01_EB, 3.168499E-02_EB, &
   2.263308E-02_EB, &
   1.007340E-01_EB, 9.210111E-02_EB, 1.687337E-01_EB, 2.959829E-02_EB, 1.810977E-01_EB, 9.553982E-02_EB, &
   3.918733E-02_EB, &
   4.509352E-02_EB, 1.086112E-01_EB, 7.647679E-02_EB, 2.814778E-02_EB, 8.981709E-01_EB, 1.273344E-01_EB, &
   4.408137E-02_EB, &
   9.590449E-02_EB, 2.770484E-01_EB, 1.477733E-01_EB, 1.049910E-01_EB, 3.105849E-01_EB, 2.918868E-01_EB, &
   1.291524E-01_EB, &
   3.138946E-01_EB, 5.413508E-01_EB, 3.949461E-01_EB, 3.999787E-01_EB, 6.266594E-01_EB, 4.691785E-01_EB, &
   3.754321E-01_EB, &
   1.106384E+00_EB, 1.162295E+00_EB, 1.094260E+00_EB, 1.051491E+00_EB, 1.215748E+00_EB, 7.154832E-01_EB, &
   4.545049E-01_EB, &
   1.517339E+00_EB, 1.415673E+00_EB, 1.368748E+00_EB, 1.312400E+00_EB, 1.453900E+00_EB, 9.649236E-01_EB, &
   7.807490E-01_EB, &
   4.303118E+00_EB, 2.952375E+00_EB, 2.831306E+00_EB, 2.384206E+00_EB, 2.387748E+00_EB, 1.208513E+00_EB, &
   8.645917E-01_EB/),(/7,8/))

GAMMAD3_C2H6(1:7,17:24) = RESHAPE((/ &  ! 2950-3125 CM-1
   3.315145E+00_EB, 2.599752E+00_EB, 2.743646E+00_EB, 2.629391E+00_EB, 2.506127E+00_EB, 1.333128E+00_EB, &
   9.626906E-01_EB, &
   1.456169E+00_EB, 1.207547E+00_EB, 1.363886E+00_EB, 1.509996E+00_EB, 1.699983E+00_EB, 1.288544E+00_EB, &
   8.734672E-01_EB, &
   1.081840E+00_EB, 1.007445E+00_EB, 1.186416E+00_EB, 1.295045E+00_EB, 1.467006E+00_EB, 1.089437E+00_EB, &
   7.294066E-01_EB, &
   1.769766E+00_EB, 1.424872E+00_EB, 1.588295E+00_EB, 1.745279E+00_EB, 1.453194E+00_EB, 9.836635E-01_EB, &
   4.884795E-01_EB, &
   8.210794E-01_EB, 7.811090E-01_EB, 9.322482E-01_EB, 1.373898E+00_EB, 1.084482E+00_EB, 6.956616E-01_EB, &
   5.118669E-01_EB, &
   1.964022E-01_EB, 3.039042E-01_EB, 3.956607E-01_EB, 1.328063E+00_EB, 5.899273E-01_EB, 6.132437E-01_EB, &
   5.047836E-01_EB, &
   2.041406E-02_EB, 8.312137E-02_EB, 1.318700E-01_EB, 7.573599E+00_EB, 2.512530E-01_EB, 2.496958E-01_EB, &
   1.051791E+00_EB, &
   5.023107E-03_EB, 9.491129E-03_EB, 3.053563E-02_EB, 9.379598E-02_EB, 3.673982E-02_EB, 7.455316E-02_EB, &
   2.454365E-01_EB/),(/7,8/))

GAMMAD3_C2H6(1:7,25:32) = RESHAPE((/ &  ! 3150-3325 CM-1
   6.352743E-03_EB, 7.943354E-04_EB, 1.083748E-02_EB, 9.588981E-03_EB, 4.904702E-03_EB, 1.679226E-02_EB, &
   1.304495E-02_EB, &
   6.766775E-03_EB, 1.596208E-04_EB, 6.130026E-03_EB, 7.152650E-03_EB, 3.279571E-04_EB, 1.794526E-03_EB, &
   4.823723E-04_EB, &
   1.510899E-03_EB, 1.196785E-04_EB, 1.094313E-03_EB, 1.868976E-04_EB, 1.950517E-04_EB, 3.851491E-04_EB, &
   8.135632E-05_EB, &
   1.592258E-02_EB, 2.062254E-04_EB, 8.766044E-04_EB, 4.062596E-04_EB, 1.031744E-04_EB, 1.363841E-04_EB, &
   4.489251E-05_EB, &
   5.336883E-03_EB, 1.621021E-04_EB, 8.122952E-04_EB, 9.938396E-05_EB, 1.057126E-04_EB, 8.283648E-05_EB, &
   2.421683E-05_EB, &
   7.894809E-03_EB, 1.025986E-04_EB, 4.657427E-04_EB, 6.874881E-05_EB, 6.535834E-05_EB, 3.906150E-05_EB, &
   4.299378E-05_EB, &
   1.168996E-04_EB, 6.996908E-05_EB, 5.712580E-04_EB, 3.973347E-05_EB, 4.446145E-05_EB, 8.103887E-05_EB, &
   3.469019E-05_EB, &
   1.835174E-04_EB, 4.182929E-05_EB, 1.449024E-03_EB, 6.981809E-04_EB, 3.965878E-05_EB, 3.732167E-05_EB, &
   1.777260E-02_EB/),(/7,8/))

GAMMAD3_C2H6(1:7,33:34) = RESHAPE((/ &  ! 3350-3375 CM-1
   1.966771E-04_EB, 1.267047E-04_EB, 5.949690E-04_EB, 2.287935E-04_EB, 1.200025E-04_EB, 2.518698E-04_EB, &
   2.116261E-04_EB, &
   1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, 1.688326E-04_EB, &
   1.688326E-04_EB/),(/7,2/))

!-------------------------ETHYLENE DATA-------------------


! THERE ARE 4 BANDS FOR ETHYLENE

! BAND #1: 750 CM-1 - 1250 CM-1
! BAND #2: 1300 CM-1 - 1600 CM-1
! BAND #3: 1750 CM-1 - 2075 CM-1
! BAND #4: 2800 CM-1 - 3400 CM-1

ALLOCATE(SD_C2H4_TEMP(N_TEMP_C2H4))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR ETHYLENE ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_C2H4(N_BAND_C2H4,3))
ALLOCATE(BE_C2H4(N_BAND_C2H4))

OM_BND_C2H4 = RESHAPE((/ &
   750._EB, 1300._EB, 1750._EB, 2800._EB, &
   1250._EB, 1600._EB, 2075._EB, 3400._EB, &
   5._EB, 25._EB, 25._EB, 25._EB/),(/N_BAND_C2H4,3/))

SD_C2H4_TEMP = (/ &
   296._EB, 400._EB, 450._EB, 500._EB, 601._EB, 801._EB,&
   1000._EB/)

BE_C2H4 = (/ &
   1.000_EB, 1.000_EB, 1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_C2H4(N_TEMP_C2H4,77))

! BAND #1: 750 CM-1 - 1250 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.6588 %

SD1_C2H4(1:7,1:8) = RESHAPE((/ &  ! 750-785 CM-1
   3.061616E-04_EB, 2.767365E-04_EB, 2.573770E-04_EB, 3.123088E-04_EB, 2.802708E-04_EB, 3.214089E-04_EB, &
   1.645179E-04_EB, &
   3.863930E-04_EB, 3.214740E-04_EB, 3.599651E-04_EB, 2.875226E-04_EB, 3.220027E-04_EB, 3.442021E-04_EB, &
   3.429771E-04_EB, &
   3.479571E-04_EB, 3.988382E-04_EB, 5.410044E-04_EB, 3.711756E-04_EB, 4.049933E-04_EB, 3.780624E-04_EB, &
   1.581304E-04_EB, &
   3.363011E-04_EB, 4.180798E-04_EB, 3.197069E-04_EB, 3.653986E-04_EB, 4.187833E-04_EB, 7.101448E-03_EB, &
   3.809923E-04_EB, &
   3.533603E-04_EB, 3.413308E-04_EB, 3.254621E-04_EB, 4.091800E-04_EB, 3.812469E-04_EB, 2.897062E-01_EB, &
   2.857766E-04_EB, &
   3.471021E-04_EB, 3.981441E-04_EB, 3.391170E-04_EB, 3.788735E-04_EB, 3.733369E-04_EB, 7.154619E-01_EB, &
   2.973615E-04_EB, &
   3.515529E-04_EB, 3.867219E-04_EB, 3.372267E-04_EB, 4.293329E-04_EB, 3.549482E-03_EB, 2.273679E+00_EB, &
   2.916009E-04_EB, &
   4.086122E-04_EB, 3.034723E-04_EB, 3.481328E-04_EB, 4.240499E-04_EB, 2.168522E-02_EB, 8.658210E-01_EB, &
   2.764510E-04_EB/),(/7,8/))

SD1_C2H4(1:7,9:16) = RESHAPE((/ &  ! 790-825 CM-1
   3.370619E-04_EB, 3.449516E-04_EB, 3.068753E-03_EB, 2.166061E-03_EB, 3.067872E-01_EB, 2.126871E+00_EB, &
   3.160891E-04_EB, &
   2.931244E-04_EB, 2.735279E-03_EB, 6.531861E-01_EB, 6.019424E-03_EB, 5.639027E-01_EB, 1.299202E+00_EB, &
   3.396657E-04_EB, &
   4.877063E-04_EB, 9.280827E-03_EB, 1.579979E-02_EB, 1.496897E-02_EB, 4.992397E-01_EB, 1.406058E+00_EB, &
   3.043444E-04_EB, &
   4.001483E-04_EB, 1.891644E-02_EB, 9.491147E-03_EB, 2.267553E-02_EB, 3.421534E-01_EB, 6.382093E-02_EB, &
   1.852766E-04_EB, &
   1.290823E-03_EB, 2.799153E-02_EB, 1.446431E-02_EB, 2.843022E-02_EB, 8.398144E-02_EB, 4.512817E-02_EB, &
   1.294866E-02_EB, &
   2.070200E-01_EB, 3.632895E-02_EB, 3.296335E-02_EB, 4.820045E-02_EB, 1.668313E-01_EB, 1.323383E-01_EB, &
   3.901722E+00_EB, &
   2.001955E-01_EB, 4.074161E-02_EB, 5.504791E-02_EB, 2.033738E-01_EB, 1.370229E-01_EB, 1.403502E+00_EB, &
   9.703462E-01_EB, &
   2.718536E-02_EB, 7.068372E-02_EB, 8.035231E-02_EB, 1.066961E-01_EB, 1.330522E-01_EB, 2.467964E-01_EB, &
   4.709421E+00_EB/),(/7,8/))

SD1_C2H4(1:7,17:24) = RESHAPE((/ &  ! 830-865 CM-1
   8.016587E-02_EB, 1.232301E-01_EB, 1.080896E-01_EB, 1.876693E-01_EB, 1.436722E-01_EB, 1.806775E-01_EB, &
   1.955761E-01_EB, &
   8.134800E-02_EB, 1.769076E-01_EB, 1.680321E-01_EB, 1.930321E-01_EB, 1.757502E-01_EB, 2.398550E-01_EB, &
   2.039868E-01_EB, &
   1.441756E-01_EB, 2.209548E-01_EB, 2.092624E-01_EB, 2.537575E-01_EB, 2.144884E-01_EB, 2.788351E-01_EB, &
   7.848435E-01_EB, &
   1.963952E-01_EB, 2.703514E-01_EB, 2.622330E-01_EB, 3.903483E-01_EB, 3.265592E-01_EB, 3.009724E-01_EB, &
   6.260775E+00_EB, &
   2.991975E-01_EB, 3.352004E-01_EB, 3.564035E-01_EB, 3.796975E-01_EB, 4.082076E-01_EB, 3.402704E-01_EB, &
   2.111753E+00_EB, &
   3.860882E-01_EB, 4.216740E-01_EB, 4.053700E-01_EB, 4.209808E-01_EB, 5.214527E-01_EB, 5.131750E-01_EB, &
   3.472334E-01_EB, &
   4.648316E-01_EB, 4.632385E-01_EB, 4.893760E-01_EB, 5.410936E-01_EB, 5.227714E-01_EB, 3.928599E-01_EB, &
   3.841886E-01_EB, &
   6.022275E-01_EB, 5.990898E-01_EB, 5.570012E-01_EB, 6.192645E-01_EB, 6.317220E-01_EB, 4.457061E-01_EB, &
   4.504311E-01_EB/),(/7,8/))

SD1_C2H4(1:7,25:32) = RESHAPE((/ &  ! 870-905 CM-1
   7.041966E-01_EB, 6.621798E-01_EB, 6.254956E-01_EB, 6.363171E-01_EB, 6.107430E-01_EB, 5.497340E-01_EB, &
   5.235314E-01_EB, &
   8.729129E-01_EB, 7.949380E-01_EB, 7.594193E-01_EB, 7.573864E-01_EB, 6.279590E-01_EB, 5.238851E-01_EB, &
   6.195839E-01_EB, &
   1.076407E+00_EB, 9.480760E-01_EB, 8.612630E-01_EB, 8.823757E-01_EB, 7.150564E-01_EB, 6.412975E-01_EB, &
   5.713233E-01_EB, &
   1.149326E+00_EB, 9.164652E-01_EB, 8.407015E-01_EB, 7.805755E-01_EB, 7.361358E-01_EB, 4.885371E-01_EB, &
   6.246788E-01_EB, &
   1.285145E+00_EB, 1.001830E+00_EB, 9.098842E-01_EB, 8.665704E-01_EB, 7.069087E-01_EB, 5.784402E-01_EB, &
   5.420318E-01_EB, &
   1.656594E+00_EB, 1.243694E+00_EB, 1.132558E+00_EB, 1.004866E+00_EB, 8.573835E-01_EB, 6.721572E-01_EB, &
   5.651920E-01_EB, &
   1.512695E+00_EB, 1.063450E+00_EB, 9.534704E-01_EB, 8.623387E-01_EB, 7.340704E-01_EB, 5.647638E-01_EB, &
   5.138506E-01_EB, &
   1.503391E+00_EB, 1.046223E+00_EB, 9.264262E-01_EB, 8.216305E-01_EB, 7.214293E-01_EB, 4.921282E-01_EB, &
   9.505467E-01_EB/),(/7,8/))

SD1_C2H4(1:7,33:40) = RESHAPE((/ &  ! 910-945 CM-1
   2.136589E+00_EB, 1.441371E+00_EB, 1.265404E+00_EB, 1.131971E+00_EB, 8.629746E-01_EB, 5.974312E-01_EB, &
   6.931755E-01_EB, &
   1.937708E+00_EB, 1.292014E+00_EB, 1.117250E+00_EB, 9.960421E-01_EB, 7.945183E-01_EB, 6.856903E-01_EB, &
   6.607081E-01_EB, &
   1.854239E+00_EB, 1.153175E+00_EB, 9.838961E-01_EB, 9.179450E-01_EB, 6.997797E-01_EB, 5.900011E-01_EB, &
   1.136208E+00_EB, &
   1.823799E+00_EB, 1.165735E+00_EB, 1.007681E+00_EB, 9.292605E-01_EB, 7.896203E-01_EB, 5.208819E-01_EB, &
   1.763502E+00_EB, &
   1.757932E+00_EB, 1.135853E+00_EB, 9.853861E-01_EB, 9.026759E-01_EB, 7.075480E-01_EB, 6.179452E-01_EB, &
   6.802571E-01_EB, &
   1.474799E+00_EB, 1.023497E+00_EB, 9.336669E-01_EB, 8.778123E-01_EB, 8.131435E-01_EB, 6.616974E-01_EB, &
   7.567051E-01_EB, &
   2.289039E+00_EB, 1.607837E+00_EB, 1.440647E+00_EB, 1.347166E+00_EB, 1.239509E+00_EB, 1.038545E+00_EB, &
   1.100592E+00_EB, &
   3.685332E+00_EB, 2.419819E+00_EB, 2.144192E+00_EB, 1.965314E+00_EB, 1.747995E+00_EB, 1.497634E+00_EB, &
   1.667794E+00_EB/),(/7,8/))

SD1_C2H4(1:7,41:48) = RESHAPE((/ &  ! 950-985 CM-1
   9.810246E+00_EB, 7.439967E+00_EB, 6.716823E+00_EB, 6.081792E+00_EB, 4.841636E+00_EB, 3.239145E+00_EB, &
   2.178622E+00_EB, &
   2.747963E+00_EB, 2.185259E+00_EB, 2.110959E+00_EB, 2.104012E+00_EB, 1.967093E+00_EB, 1.887167E+00_EB, &
   1.592410E+00_EB, &
   1.878302E+00_EB, 1.339111E+00_EB, 1.209909E+00_EB, 1.180489E+00_EB, 1.063613E+00_EB, 1.043418E+00_EB, &
   9.802449E-01_EB, &
   1.846689E+00_EB, 1.233103E+00_EB, 1.078269E+00_EB, 1.010819E+00_EB, 7.843025E-01_EB, 7.220401E-01_EB, &
   6.485561E-01_EB, &
   2.135746E+00_EB, 1.274099E+00_EB, 1.072632E+00_EB, 9.574284E-01_EB, 7.244360E-01_EB, 6.000363E-01_EB, &
   5.922618E-01_EB, &
   1.598509E+00_EB, 1.039908E+00_EB, 9.058024E-01_EB, 8.461204E-01_EB, 6.619076E-01_EB, 5.572688E-01_EB, &
   6.188948E-01_EB, &
   2.103202E+00_EB, 1.300976E+00_EB, 1.111634E+00_EB, 9.847661E-01_EB, 7.744278E-01_EB, 5.718997E-01_EB, &
   6.346134E-01_EB, &
   2.261675E+00_EB, 1.454471E+00_EB, 1.243161E+00_EB, 1.093485E+00_EB, 8.306949E-01_EB, 6.753502E-01_EB, &
   5.723775E-01_EB/),(/7,8/))

SD1_C2H4(1:7,49:56) = RESHAPE((/ &  ! 990-1025 CM-1
   2.079467E+00_EB, 1.263897E+00_EB, 1.076642E+00_EB, 9.600844E-01_EB, 7.479525E-01_EB, 5.904677E-01_EB, &
   5.384319E-01_EB, &
   1.966589E+00_EB, 1.324982E+00_EB, 1.148642E+00_EB, 1.030295E+00_EB, 8.285130E-01_EB, 6.028164E-01_EB, &
   5.615524E-01_EB, &
   1.704129E+00_EB, 1.136312E+00_EB, 1.028789E+00_EB, 9.378779E-01_EB, 7.696240E-01_EB, 6.407307E-01_EB, &
   5.740712E-01_EB, &
   2.099061E+00_EB, 1.480171E+00_EB, 1.279590E+00_EB, 1.150617E+00_EB, 8.965259E-01_EB, 6.108346E-01_EB, &
   5.651648E-01_EB, &
   1.532195E+00_EB, 1.072161E+00_EB, 9.586967E-01_EB, 8.739703E-01_EB, 7.409633E-01_EB, 5.463573E-01_EB, &
   5.543106E-01_EB, &
   1.449718E+00_EB, 1.130569E+00_EB, 1.010466E+00_EB, 9.314315E-01_EB, 8.044840E-01_EB, 5.741548E-01_EB, &
   5.103097E-01_EB, &
   1.361629E+00_EB, 1.048216E+00_EB, 9.484550E-01_EB, 8.875882E-01_EB, 7.583217E-01_EB, 6.224803E-01_EB, &
   5.104751E-01_EB, &
   1.215443E+00_EB, 9.557761E-01_EB, 8.676531E-01_EB, 8.218519E-01_EB, 6.763815E-01_EB, 5.340442E-01_EB, &
   4.917932E-01_EB/),(/7,8/))

SD1_C2H4(1:7,57:64) = RESHAPE((/ &  ! 1030-1065 CM-1
   1.134122E+00_EB, 9.472716E-01_EB, 8.758677E-01_EB, 8.193195E-01_EB, 7.086094E-01_EB, 5.766945E-01_EB, &
   6.808968E-01_EB, &
   8.950378E-01_EB, 7.770052E-01_EB, 7.253665E-01_EB, 7.245858E-01_EB, 6.204068E-01_EB, 5.655605E-01_EB, &
   6.332963E-01_EB, &
   6.844087E-01_EB, 6.493525E-01_EB, 6.230393E-01_EB, 6.224226E-01_EB, 5.363550E-01_EB, 4.974590E-01_EB, &
   7.583943E-01_EB, &
   7.556324E-01_EB, 7.049278E-01_EB, 6.726842E-01_EB, 6.598185E-01_EB, 5.886987E-01_EB, 5.424892E-01_EB, &
   5.990280E-01_EB, &
   6.289721E-01_EB, 6.069297E-01_EB, 6.011876E-01_EB, 5.779180E-01_EB, 5.326335E-01_EB, 4.750991E-01_EB, &
   4.588251E-01_EB, &
   4.745725E-01_EB, 4.627839E-01_EB, 4.803008E-01_EB, 4.911406E-01_EB, 4.593009E-01_EB, 3.894394E-01_EB, &
   4.927539E-01_EB, &
   4.243657E-01_EB, 4.444157E-01_EB, 4.520903E-01_EB, 4.522320E-01_EB, 4.526266E-01_EB, 3.758722E-01_EB, &
   4.237227E-01_EB, &
   2.905835E-01_EB, 3.554843E-01_EB, 3.584158E-01_EB, 3.801254E-01_EB, 4.065901E-01_EB, 4.115057E-01_EB, &
   5.026136E-01_EB/),(/7,8/))

SD1_C2H4(1:7,65:72) = RESHAPE((/ &  ! 1070-1125 CM-1
   3.642080E-01_EB, 4.087106E-01_EB, 4.100121E-01_EB, 4.077213E-01_EB, 4.343523E-01_EB, 3.642286E-01_EB, &
   4.076571E-01_EB, &
   2.285227E-01_EB, 2.715118E-01_EB, 2.835998E-01_EB, 3.085575E-01_EB, 3.428083E-01_EB, 2.858531E-01_EB, &
   3.538602E-01_EB, &
   2.072180E-01_EB, 2.508628E-01_EB, 2.687615E-01_EB, 3.061642E-01_EB, 3.126805E-01_EB, 3.288883E-01_EB, &
   3.309101E-01_EB, &
   1.485077E-01_EB, 1.994016E-01_EB, 2.138919E-01_EB, 2.352105E-01_EB, 2.665822E-01_EB, 2.680282E-01_EB, &
   4.173662E-01_EB, &
   1.233539E-01_EB, 1.739268E-01_EB, 2.005570E-01_EB, 2.264850E-01_EB, 2.387994E-01_EB, 2.855644E-01_EB, &
   3.039316E-01_EB, &
   1.208930E-01_EB, 1.604253E-01_EB, 1.886652E-01_EB, 1.954718E-01_EB, 2.106737E-01_EB, 2.480406E-01_EB, &
   2.674441E-01_EB, &
   9.300476E-02_EB, 1.298575E-01_EB, 1.529873E-01_EB, 1.727849E-01_EB, 2.062593E-01_EB, 2.191541E-01_EB, &
   2.754971E-01_EB, &
   2.843120E-02_EB, 4.750139E-02_EB, 6.454680E-02_EB, 7.833683E-02_EB, 1.007079E-01_EB, 1.334723E-01_EB, &
   1.952035E-01_EB/),(/7,8/))

SD1_C2H4(1:7,73:77) = RESHAPE((/ &  ! 1150-1250 CM-1
   1.746127E-02_EB, 7.841172E-03_EB, 2.585562E-02_EB, 3.153723E-02_EB, 3.990568E-02_EB, 6.617133E-02_EB, &
   2.077741E-01_EB, &
   3.428479E-04_EB, 3.716895E-04_EB, 5.481270E-03_EB, 9.380871E-03_EB, 1.423446E-02_EB, 3.592881E-02_EB, &
   5.844943E-01_EB, &
   3.844693E-04_EB, 4.065786E-04_EB, 5.088321E-04_EB, 3.670823E-04_EB, 1.012681E-03_EB, 1.177143E-02_EB, &
   1.441741E+00_EB, &
   2.863454E-04_EB, 2.697429E-04_EB, 2.663683E-04_EB, 2.847322E-04_EB, 2.814940E-04_EB, 4.490652E-04_EB, &
   2.765929E-03_EB, &
   2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, &
   1.035699E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_C2H4(N_TEMP_C2H4,77))

! BAND #1: 750 CM-1 - 1250 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.6588 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_C2H4(1:7,1:8) = RESHAPE((/ &  ! 750-785 CM-1
   2.416099E-04_EB, 1.972502E-04_EB, 2.014288E-04_EB, 2.567319E-04_EB, 2.171052E-04_EB, 2.409977E-04_EB, &
   1.211331E-04_EB, &
   2.468780E-04_EB, 2.281487E-04_EB, 2.453616E-04_EB, 1.924299E-04_EB, 2.048382E-04_EB, 2.401308E-04_EB, &
   2.496562E-04_EB, &
   2.424362E-04_EB, 2.577800E-04_EB, 3.990244E-04_EB, 2.948000E-04_EB, 2.703709E-04_EB, 2.386217E-04_EB, &
   1.097937E-04_EB, &
   2.310838E-04_EB, 3.105661E-04_EB, 2.545394E-04_EB, 2.425587E-04_EB, 2.350566E-04_EB, 5.850300E-04_EB, &
   2.717545E-04_EB, &
   2.188209E-04_EB, 2.694507E-04_EB, 2.395597E-04_EB, 2.852713E-04_EB, 3.015723E-04_EB, 1.785307E-05_EB, &
   1.970987E-04_EB, &
   2.681455E-04_EB, 3.065002E-04_EB, 2.277458E-04_EB, 2.640085E-04_EB, 2.242100E-04_EB, 5.823579E-05_EB, &
   2.107112E-04_EB, &
   2.864384E-04_EB, 2.901989E-04_EB, 2.177055E-04_EB, 2.318611E-04_EB, 8.020695E-03_EB, 6.053782E-05_EB, &
   1.973837E-04_EB, &
   2.864942E-04_EB, 1.888603E-04_EB, 2.876981E-04_EB, 2.918797E-04_EB, 4.025819E-04_EB, 1.500744E-04_EB, &
   1.850902E-04_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,9:16) = RESHAPE((/ &  ! 790-825 CM-1
   2.028540E-04_EB, 2.332145E-04_EB, 8.835388E-04_EB, 1.557804E-03_EB, 6.921471E-05_EB, 1.608572E-04_EB, &
   2.284767E-04_EB, &
   1.964413E-04_EB, 1.354319E-03_EB, 3.099296E-05_EB, 1.230936E-01_EB, 9.177521E-05_EB, 3.735015E-04_EB, &
   2.567094E-04_EB, &
   3.523993E-04_EB, 4.495628E-03_EB, 1.170244E-01_EB, 1.834956E-01_EB, 2.119215E-04_EB, 4.987881E-04_EB, &
   2.324660E-04_EB, &
   2.481552E-04_EB, 3.297622E-03_EB, 3.408731E-01_EB, 1.941870E-02_EB, 5.655524E-04_EB, 6.632397E+00_EB, &
   1.455497E-04_EB, &
   6.200957E-02_EB, 5.552681E-03_EB, 1.304796E+00_EB, 9.167035E-01_EB, 1.036969E-02_EB, 4.543832E+00_EB, &
   4.663762E-02_EB, &
   4.857487E-05_EB, 7.312172E-03_EB, 6.397405E-01_EB, 1.805633E+00_EB, 7.876697E-03_EB, 1.052298E+01_EB, &
   2.200552E-04_EB, &
   1.266088E-04_EB, 3.671828E-01_EB, 1.052101E+00_EB, 4.874822E-03_EB, 1.186820E-02_EB, 1.436321E-03_EB, &
   2.304645E-03_EB, &
   7.234540E-02_EB, 1.423099E-01_EB, 4.250396E+00_EB, 4.471836E-01_EB, 1.394765E-01_EB, 2.304974E-02_EB, &
   9.320014E-04_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,17:24) = RESHAPE((/ &  ! 830-865 CM-1
   5.320200E-03_EB, 3.524362E-02_EB, 1.338375E+01_EB, 2.690068E-02_EB, 2.237749E+01_EB, 3.550085E+01_EB, &
   4.352470E+01_EB, &
   1.362377E+00_EB, 4.799018E-02_EB, 3.645279E-01_EB, 1.401350E-01_EB, 2.192962E+01_EB, 1.052708E-01_EB, &
   4.352470E+01_EB, &
   4.653039E-02_EB, 8.421067E-02_EB, 1.759985E+00_EB, 1.658517E-01_EB, 5.613535E+01_EB, 4.860610E+01_EB, &
   9.812101E-03_EB, &
   2.321729E-01_EB, 2.414939E-01_EB, 1.567072E+01_EB, 7.788555E-02_EB, 1.585202E-01_EB, 2.057238E+01_EB, &
   2.123765E-03_EB, &
   1.555629E-01_EB, 8.669673E-01_EB, 5.965418E-01_EB, 3.491496E-01_EB, 1.276794E-01_EB, 4.796946E-01_EB, &
   7.827500E-03_EB, &
   2.604872E-01_EB, 5.734264E-01_EB, 3.245414E+01_EB, 2.473726E+00_EB, 1.188363E-01_EB, 8.314367E-02_EB, &
   4.352145E+01_EB, &
   3.170365E-01_EB, 1.356303E+00_EB, 6.470392E-01_EB, 2.532365E-01_EB, 1.744398E-01_EB, 7.021255E-01_EB, &
   4.352322E+01_EB, &
   5.909101E-01_EB, 1.099398E+00_EB, 4.090840E+00_EB, 4.223559E-01_EB, 2.001671E-01_EB, 4.828266E+01_EB, &
   4.352310E+01_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,25:32) = RESHAPE((/ &  ! 870-905 CM-1
   6.696240E-01_EB, 9.061473E-01_EB, 1.096877E+00_EB, 6.942716E-01_EB, 5.833627E-01_EB, 7.942786E-01_EB, &
   3.236155E-01_EB, &
   7.130009E-01_EB, 9.362301E-01_EB, 7.015500E-01_EB, 5.223698E-01_EB, 6.248457E+00_EB, 4.863162E+01_EB, &
   1.544108E-01_EB, &
   1.043070E+00_EB, 7.601537E-01_EB, 8.543522E-01_EB, 5.009053E-01_EB, 8.115324E-01_EB, 2.773377E-01_EB, &
   1.734668E+00_EB, &
   9.830346E-01_EB, 1.107606E+00_EB, 7.646673E-01_EB, 7.719984E-01_EB, 3.261238E-01_EB, 2.851084E+00_EB, &
   2.517361E-01_EB, &
   1.198560E+00_EB, 1.081208E+00_EB, 7.434343E-01_EB, 6.010115E-01_EB, 9.163865E-01_EB, 3.554812E-01_EB, &
   2.317131E-01_EB, &
   9.814423E-01_EB, 1.013604E+00_EB, 6.693517E-01_EB, 9.522875E-01_EB, 8.114383E-01_EB, 3.910559E-01_EB, &
   3.337523E+01_EB, &
   7.620617E-01_EB, 7.803621E-01_EB, 5.747705E-01_EB, 6.345028E-01_EB, 5.888869E-01_EB, 4.300627E-01_EB, &
   4.310793E+01_EB, &
   8.991018E-01_EB, 8.909633E-01_EB, 6.980425E-01_EB, 1.113145E+00_EB, 7.124349E-01_EB, 4.862871E+01_EB, &
   6.096945E-02_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,33:40) = RESHAPE((/ &  ! 910-945 CM-1
   5.715478E-01_EB, 5.814464E-01_EB, 5.080659E-01_EB, 4.983276E-01_EB, 8.896438E-01_EB, 3.707252E+00_EB, &
   2.418005E-01_EB, &
   4.564794E-01_EB, 5.408890E-01_EB, 5.281411E-01_EB, 6.161207E-01_EB, 9.304299E-01_EB, 1.822560E-01_EB, &
   1.433739E-01_EB, &
   4.474194E-01_EB, 5.933999E-01_EB, 5.926916E-01_EB, 4.188992E-01_EB, 8.916726E-01_EB, 2.166682E-01_EB, &
   3.172885E-02_EB, &
   6.679241E-01_EB, 7.624342E-01_EB, 6.139160E-01_EB, 4.989378E-01_EB, 3.780157E-01_EB, 4.858759E+01_EB, &
   2.070389E-02_EB, &
   1.572323E+00_EB, 1.507380E+00_EB, 1.025688E+00_EB, 7.931157E-01_EB, 2.355979E+00_EB, 4.083523E-01_EB, &
   4.042678E-01_EB, &
   1.840928E+00_EB, 1.516332E+00_EB, 9.674401E-01_EB, 7.614660E-01_EB, 5.264076E-01_EB, 2.038597E+00_EB, &
   6.604494E-01_EB, &
   2.359084E+00_EB, 1.464881E+00_EB, 1.174436E+00_EB, 1.055262E+00_EB, 8.221130E-01_EB, 1.578212E+00_EB, &
   5.803317E-01_EB, &
   1.353327E+00_EB, 1.007304E+00_EB, 8.550921E-01_EB, 9.242096E-01_EB, 8.460605E-01_EB, 1.145876E+00_EB, &
   3.951015E-01_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,41:48) = RESHAPE((/ &  ! 950-985 CM-1
   1.026068E+00_EB, 7.509719E-01_EB, 6.828323E-01_EB, 7.181245E-01_EB, 8.830267E-01_EB, 1.535879E+00_EB, &
   3.739161E+00_EB, &
   1.923367E+00_EB, 1.440812E+00_EB, 1.103009E+00_EB, 1.030902E+00_EB, 1.337752E+00_EB, 1.153093E+00_EB, &
   2.040631E+00_EB, &
   1.610198E+00_EB, 1.196060E+00_EB, 1.104793E+00_EB, 8.780307E-01_EB, 1.156102E+00_EB, 6.886495E-01_EB, &
   2.238055E+00_EB, &
   1.383929E+00_EB, 1.280808E+00_EB, 1.097115E+00_EB, 7.326487E-01_EB, 5.498608E+01_EB, 4.452870E-01_EB, &
   4.352469E+01_EB, &
   1.188786E+00_EB, 1.159934E+00_EB, 9.631713E-01_EB, 7.953825E-01_EB, 1.129330E+01_EB, 6.701708E-01_EB, &
   5.733374E-01_EB, &
   1.242802E+00_EB, 1.350551E+00_EB, 1.172729E+00_EB, 7.385986E-01_EB, 3.096362E+01_EB, 8.122837E-01_EB, &
   2.708279E-01_EB, &
   6.907316E-01_EB, 7.891104E-01_EB, 7.404165E-01_EB, 7.153440E-01_EB, 1.513590E+00_EB, 8.412093E-01_EB, &
   2.566696E-01_EB, &
   7.900363E-01_EB, 8.478862E-01_EB, 7.713692E-01_EB, 7.316301E-01_EB, 1.686388E+00_EB, 3.127574E-01_EB, &
   9.749592E-01_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,49:56) = RESHAPE((/ &  ! 990-1025 CM-1
   5.092459E-01_EB, 6.454904E-01_EB, 6.133995E-01_EB, 5.823764E-01_EB, 1.459319E+00_EB, 3.862803E-01_EB, &
   4.351923E+01_EB, &
   7.621436E-01_EB, 8.099637E-01_EB, 7.550770E-01_EB, 6.563547E-01_EB, 8.599084E-01_EB, 1.265807E+00_EB, &
   4.352461E+01_EB, &
   8.135765E-01_EB, 9.399794E-01_EB, 6.891443E-01_EB, 6.724564E-01_EB, 1.443782E+00_EB, 4.675502E-01_EB, &
   6.832951E+00_EB, &
   1.261934E+00_EB, 1.187896E+00_EB, 1.050548E+00_EB, 8.976709E-01_EB, 1.264652E+00_EB, 4.862525E+01_EB, &
   8.286417E+00_EB, &
   8.043161E-01_EB, 8.668835E-01_EB, 7.890285E-01_EB, 7.711386E-01_EB, 1.050733E+00_EB, 4.856478E+01_EB, &
   4.350007E+01_EB, &
   1.357377E+00_EB, 1.080455E+00_EB, 1.033605E+00_EB, 1.033196E+00_EB, 8.256082E-01_EB, 8.647109E+00_EB, &
   4.350847E+01_EB, &
   8.412833E-01_EB, 8.456811E-01_EB, 8.447705E-01_EB, 7.603788E-01_EB, 9.443837E-01_EB, 2.903532E-01_EB, &
   4.352240E+01_EB, &
   9.318939E-01_EB, 8.628581E-01_EB, 7.702967E-01_EB, 5.812089E-01_EB, 1.478510E+00_EB, 7.524972E-01_EB, &
   4.352454E+01_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,57:64) = RESHAPE((/ &  ! 1030-1065 CM-1
   8.629044E-01_EB, 8.975496E-01_EB, 7.798708E-01_EB, 6.608358E-01_EB, 6.759715E-01_EB, 2.711107E-01_EB, &
   1.220647E-01_EB, &
   7.370489E-01_EB, 8.019656E-01_EB, 1.000196E+00_EB, 5.336084E-01_EB, 1.804552E+00_EB, 2.787416E-01_EB, &
   1.281215E-01_EB, &
   7.787285E-01_EB, 7.334174E-01_EB, 8.721264E-01_EB, 5.887939E-01_EB, 3.552536E+01_EB, 8.474746E-01_EB, &
   8.392367E-02_EB, &
   5.450501E-01_EB, 6.247731E-01_EB, 7.400770E-01_EB, 5.561437E-01_EB, 9.526879E-01_EB, 3.143742E-01_EB, &
   1.138369E-01_EB, &
   5.144688E-01_EB, 6.585084E-01_EB, 5.903669E-01_EB, 7.810567E-01_EB, 8.401970E-01_EB, 6.417739E-01_EB, &
   4.420604E-01_EB, &
   3.665531E-01_EB, 8.719562E-01_EB, 5.223315E-01_EB, 4.375375E-01_EB, 5.743448E-01_EB, 2.121049E+00_EB, &
   2.496608E-01_EB, &
   4.667181E-01_EB, 7.081388E-01_EB, 6.911857E-01_EB, 6.081257E-01_EB, 3.737837E-01_EB, 2.130304E+00_EB, &
   6.904458E-01_EB, &
   3.306129E-01_EB, 3.244247E-01_EB, 1.436316E+00_EB, 6.770932E-01_EB, 3.549036E-01_EB, 2.672621E-01_EB, &
   1.358031E-01_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,65:72) = RESHAPE((/ &  ! 1070-1125 CM-1
   2.919422E-01_EB, 3.828297E-01_EB, 5.395788E-01_EB, 5.826860E-01_EB, 2.566111E-01_EB, 2.715471E+01_EB, &
   2.745597E-01_EB, &
   3.398729E-01_EB, 4.374885E-01_EB, 1.141266E+00_EB, 3.939052E-01_EB, 1.637264E-01_EB, 4.863163E+01_EB, &
   4.352422E+01_EB, &
   1.420571E-01_EB, 4.454635E-01_EB, 1.365866E+00_EB, 2.363442E-01_EB, 2.876851E-01_EB, 4.033493E-01_EB, &
   4.352427E+01_EB, &
   2.987867E-01_EB, 2.946281E-01_EB, 3.722689E+00_EB, 5.611341E-01_EB, 3.323989E-01_EB, 4.184566E+01_EB, &
   8.215267E-02_EB, &
   6.464375E-02_EB, 2.415560E-01_EB, 5.270835E-01_EB, 2.591978E-01_EB, 1.623247E+00_EB, 2.083192E-01_EB, &
   2.703531E+00_EB, &
   1.652449E-01_EB, 7.624040E-01_EB, 3.783911E-01_EB, 6.227587E-01_EB, 4.813467E+00_EB, 4.088673E-01_EB, &
   4.352470E+01_EB, &
   4.426509E-02_EB, 2.377002E-01_EB, 5.194564E+00_EB, 4.416648E-01_EB, 1.768354E-01_EB, 6.614723E-01_EB, &
   8.659489E-01_EB, &
   1.448604E-02_EB, 6.642067E-02_EB, 6.463703E-01_EB, 1.052108E+00_EB, 1.675080E-01_EB, 3.912344E-01_EB, &
   3.760628E-01_EB/),(/7,8/))

GAMMAD1_C2H4(1:7,73:77) = RESHAPE((/ &  ! 1150-1250 CM-1
   2.129075E-04_EB, 2.822799E-02_EB, 2.461798E-02_EB, 8.986262E-02_EB, 1.247997E+00_EB, 1.467934E-01_EB, &
   1.669303E-02_EB, &
   2.363189E-04_EB, 2.307838E-04_EB, 6.030981E-03_EB, 4.177046E-03_EB, 8.698422E-02_EB, 1.673678E-02_EB, &
   1.814471E-03_EB, &
   2.536687E-04_EB, 2.591352E-04_EB, 3.402664E-04_EB, 2.425216E-04_EB, 8.814904E-04_EB, 1.309772E-03_EB, &
   1.809337E-04_EB, &
   2.020888E-04_EB, 2.044892E-04_EB, 1.950677E-04_EB, 2.120114E-04_EB, 2.042227E-04_EB, 3.416414E-04_EB, &
   5.953816E-02_EB, &
   1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, &
   8.125358E-05_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_C2H4(N_TEMP_C2H4,13))

! BAND #2: 1300 CM-1 - 1600 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.84783 %

SD2_C2H4(1:7,1:8) = RESHAPE((/ &  ! 1300-1475 CM-1
   2.882468E-04_EB, 2.872390E-04_EB, 2.844740E-04_EB, 2.830853E-04_EB, 2.869806E-04_EB, 2.911547E-04_EB, &
   1.572859E-04_EB, &
   3.652126E-04_EB, 3.386829E-04_EB, 3.628235E-04_EB, 3.693377E-04_EB, 3.974359E-04_EB, 5.740813E-04_EB, &
   1.406431E+00_EB, &
   3.864016E-04_EB, 3.824772E-04_EB, 4.379786E-03_EB, 2.882465E-03_EB, 2.712238E-02_EB, 1.918484E-02_EB, &
   7.860096E-01_EB, &
   1.282101E-02_EB, 2.679804E-02_EB, 4.946063E-02_EB, 5.785843E-02_EB, 7.524531E-02_EB, 1.070631E-01_EB, &
   1.846526E-01_EB, &
   2.572462E-01_EB, 2.610537E-01_EB, 2.762334E-01_EB, 2.568286E-01_EB, 2.524152E-01_EB, 2.204080E-01_EB, &
   1.917598E-01_EB, &
   5.165021E-01_EB, 3.183373E-01_EB, 2.741011E-01_EB, 2.245903E-01_EB, 1.734576E-01_EB, 1.131539E-01_EB, &
   1.085824E-01_EB, &
   4.620901E-01_EB, 3.014428E-01_EB, 2.695816E-01_EB, 2.383870E-01_EB, 1.998293E-01_EB, 1.716093E-01_EB, &
   1.585133E-01_EB, &
   3.857202E-01_EB, 2.649185E-01_EB, 2.402621E-01_EB, 2.174361E-01_EB, 1.792406E-01_EB, 1.252422E-01_EB, &
   1.649973E-01_EB/),(/7,8/))

SD2_C2H4(1:7,9:13) = RESHAPE((/ &  ! 1500-1600 CM-1
   5.926638E-02_EB, 7.277371E-02_EB, 8.111314E-02_EB, 7.867862E-02_EB, 9.699567E-02_EB, 8.865234E-02_EB, &
   7.947370E-02_EB, &
   1.150376E-03_EB, 2.245129E-03_EB, 7.274074E-03_EB, 7.462653E-03_EB, 2.221317E-02_EB, 4.638888E-02_EB, &
   2.712345E-02_EB, &
   3.976051E-04_EB, 3.647587E-04_EB, 3.852124E-04_EB, 3.887963E-04_EB, 3.797778E-04_EB, 1.070583E-03_EB, &
   3.011592E-04_EB, &
   3.339180E-04_EB, 3.021121E-04_EB, 2.677955E-04_EB, 2.991550E-04_EB, 2.786140E-04_EB, 3.017892E-04_EB, &
   1.540746E-04_EB, &
   2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, &
   1.035699E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_C2H4(N_TEMP_C2H4,13))

! BAND #2: 1300 CM-1 - 1600 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.84783 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_C2H4(1:7,1:8) = RESHAPE((/ &  ! 1300-1475 CM-1
   2.158499E-04_EB, 2.198834E-04_EB, 2.077167E-04_EB, 2.022446E-04_EB, 2.019156E-04_EB, 2.108599E-04_EB, &
   1.180395E-04_EB, &
   2.584797E-04_EB, 2.404707E-04_EB, 2.365446E-04_EB, 2.427590E-04_EB, 2.779223E-04_EB, 3.857109E-04_EB, &
   4.413016E-05_EB, &
   2.651063E-04_EB, 2.657195E-04_EB, 1.443449E-03_EB, 2.228190E-03_EB, 2.843764E-04_EB, 1.574315E-02_EB, &
   3.976791E-04_EB, &
   3.442081E-02_EB, 6.159804E-02_EB, 2.276418E-02_EB, 3.598936E-02_EB, 1.165571E-01_EB, 3.704352E-01_EB, &
   3.636410E-02_EB, &
   4.705384E-01_EB, 6.099726E-01_EB, 2.288867E-01_EB, 5.066479E-01_EB, 3.305657E-01_EB, 3.501155E-01_EB, &
   1.491325E-01_EB, &
   9.295562E-01_EB, 6.218514E-01_EB, 2.607721E-01_EB, 5.342728E-01_EB, 6.235127E-01_EB, 1.582834E+00_EB, &
   1.082596E-01_EB, &
   4.753247E-01_EB, 4.602367E-01_EB, 2.660304E-01_EB, 3.252024E-01_EB, 4.963318E-01_EB, 1.168146E-01_EB, &
   1.189208E-01_EB, &
   5.580701E-01_EB, 5.558061E-01_EB, 2.447941E-01_EB, 1.720602E-01_EB, 1.909313E-01_EB, 2.351203E+00_EB, &
   2.093321E-02_EB/),(/7,8/))

GAMMAD2_C2H4(1:7,9:13) = RESHAPE((/ &  ! 1500-1600 CM-1
   4.272966E-01_EB, 1.013776E+00_EB, 9.918181E-02_EB, 1.052649E+00_EB, 4.393797E-02_EB, 3.996882E-02_EB, &
   9.062687E-02_EB, &
   2.701330E-03_EB, 2.355961E-02_EB, 2.364569E-03_EB, 1.244040E-02_EB, 4.075644E-03_EB, 1.508431E-03_EB, &
   3.184016E-01_EB, &
   2.534164E-04_EB, 2.480212E-04_EB, 2.545401E-04_EB, 2.448742E-04_EB, 2.630816E-04_EB, 2.165252E-04_EB, &
   2.247869E-04_EB, &
   2.330947E-04_EB, 2.194445E-04_EB, 2.014090E-04_EB, 2.218115E-04_EB, 2.076602E-04_EB, 2.253786E-04_EB, &
   1.141333E-04_EB, &
   1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, &
   8.125358E-05_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(SD3_C2H4(N_TEMP_C2H4,14))

! BAND #3: 1750 CM-1 - 2075 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.45198 %

SD3_C2H4(1:7,1:8) = RESHAPE((/ &  ! 1750-1925 CM-1
   2.815672E-04_EB, 3.085650E-04_EB, 3.011997E-04_EB, 2.507052E-04_EB, 3.301713E-04_EB, 2.824203E-04_EB, &
   1.401316E-04_EB, &
   3.584064E-04_EB, 3.532685E-04_EB, 3.752903E-04_EB, 3.459819E-04_EB, 6.155777E-04_EB, 3.497840E-03_EB, &
   1.410505E-01_EB, &
   3.713646E-04_EB, 3.521998E-04_EB, 5.743608E-04_EB, 8.300439E-04_EB, 2.211093E-02_EB, 1.662667E-02_EB, &
   7.102650E-01_EB, &
   1.552528E-02_EB, 1.444417E-02_EB, 2.574225E-02_EB, 2.639928E-02_EB, 4.136403E-02_EB, 4.346707E-02_EB, &
   8.056475E-01_EB, &
   1.377934E-01_EB, 1.025802E-01_EB, 1.048520E-01_EB, 9.315740E-02_EB, 9.685325E-02_EB, 7.199117E-02_EB, &
   6.245123E-01_EB, &
   1.901714E-01_EB, 1.029944E-01_EB, 9.811801E-02_EB, 8.417027E-02_EB, 7.035037E-02_EB, 5.536099E-02_EB, &
   4.099564E-01_EB, &
   2.542475E-01_EB, 1.411985E-01_EB, 1.285533E-01_EB, 1.090056E-01_EB, 9.009588E-02_EB, 7.442026E-02_EB, &
   1.611310E-01_EB, &
   1.937759E-01_EB, 1.535567E-01_EB, 1.503998E-01_EB, 1.340973E-01_EB, 1.208789E-01_EB, 9.651057E-02_EB, &
   1.654443E-01_EB/),(/7,8/))

SD3_C2H4(1:7,9:14) = RESHAPE((/ &  ! 1950-2075 CM-1
   1.548981E-02_EB, 2.175487E-02_EB, 3.117281E-02_EB, 3.251084E-02_EB, 4.215815E-02_EB, 5.301204E-02_EB, &
   1.143451E-01_EB, &
   4.016633E-04_EB, 3.491488E-04_EB, 4.867031E-04_EB, 4.500772E-04_EB, 2.320822E-03_EB, 7.486336E-03_EB, &
   2.340736E-01_EB, &
   3.822381E-04_EB, 3.883093E-04_EB, 3.278614E-04_EB, 3.723909E-04_EB, 3.584776E-04_EB, 3.541946E-04_EB, &
   2.753666E-01_EB, &
   3.861390E-04_EB, 3.578653E-04_EB, 3.623971E-04_EB, 3.869958E-04_EB, 3.611574E-04_EB, 3.348041E-04_EB, &
   5.612807E-03_EB, &
   2.666887E-04_EB, 2.879471E-04_EB, 2.979300E-04_EB, 2.891663E-04_EB, 3.019136E-04_EB, 2.928543E-04_EB, &
   1.433099E-04_EB, &
   2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, &
   1.035699E-04_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD3_C2H4(N_TEMP_C2H4,14))

! BAND #3: 1750 CM-1 - 2075 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.45198 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD3_C2H4(1:7,1:8) = RESHAPE((/ &  ! 1750-1925 CM-1
   2.062935E-04_EB, 2.213095E-04_EB, 2.125965E-04_EB, 1.924384E-04_EB, 2.245648E-04_EB, 2.001139E-04_EB, &
   1.052265E-04_EB, &
   2.525656E-04_EB, 2.298492E-04_EB, 2.561819E-04_EB, 2.386624E-04_EB, 3.763751E-04_EB, 6.701805E-04_EB, &
   3.650678E-05_EB, &
   2.518994E-04_EB, 2.486015E-04_EB, 3.999398E-04_EB, 6.636791E-04_EB, 2.149851E-04_EB, 1.018456E-02_EB, &
   1.523459E-04_EB, &
   9.583444E-03_EB, 5.252649E-02_EB, 2.359414E-02_EB, 1.373517E-01_EB, 1.580336E-02_EB, 3.863765E-01_EB, &
   4.400337E-04_EB, &
   1.817495E-01_EB, 1.535393E+00_EB, 1.403613E-01_EB, 5.133053E-01_EB, 5.338970E-02_EB, 9.422243E-01_EB, &
   8.608663E-04_EB, &
   1.604957E-01_EB, 1.152431E+00_EB, 9.900652E-02_EB, 1.034304E-01_EB, 1.658726E-01_EB, 9.190427E-01_EB, &
   1.104201E-03_EB, &
   1.875157E-01_EB, 6.534721E-01_EB, 1.215243E-01_EB, 1.499339E-01_EB, 1.669372E-01_EB, 1.540611E-01_EB, &
   5.452210E-03_EB, &
   2.586428E-01_EB, 3.173253E-01_EB, 1.281338E-01_EB, 2.222569E-01_EB, 1.791652E-01_EB, 9.692458E-01_EB, &
   1.136010E-02_EB/),(/7,8/))

GAMMAD3_C2H4(1:7,9:14) = RESHAPE((/ &  ! 1950-2075 CM-1
   1.624541E-02_EB, 2.116364E-02_EB, 2.358393E-02_EB, 5.479449E-02_EB, 3.978462E-02_EB, 5.533594E-02_EB, &
   7.054981E-03_EB, &
   2.704645E-04_EB, 2.392112E-04_EB, 3.409077E-04_EB, 3.457806E-04_EB, 1.714364E-03_EB, 1.531894E-03_EB, &
   1.624673E-04_EB, &
   2.502376E-04_EB, 2.651107E-04_EB, 2.268500E-04_EB, 2.246834E-04_EB, 2.477431E-04_EB, 2.372743E-04_EB, &
   1.590970E-05_EB, &
   2.596193E-04_EB, 2.512472E-04_EB, 2.549796E-04_EB, 2.535211E-04_EB, 2.549967E-04_EB, 2.385259E-04_EB, &
   1.055504E-04_EB, &
   1.974440E-04_EB, 2.070414E-04_EB, 2.178990E-04_EB, 2.070114E-04_EB, 2.210429E-04_EB, 2.175134E-04_EB, &
   1.061814E-04_EB, &
   1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, &
   8.125358E-05_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(SD4_C2H4(N_TEMP_C2H4,25))

! BAND #4: 2800 CM-1 - 3400 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.39903 %

SD4_C2H4(1:7,1:8) = RESHAPE((/ &  ! 2800-2975 CM-1
   2.787454E-04_EB, 2.798900E-04_EB, 3.048047E-04_EB, 2.953254E-04_EB, 3.008118E-04_EB, 2.854943E-04_EB, &
   1.499675E-04_EB, &
   1.974546E-03_EB, 3.576021E-04_EB, 3.665839E-04_EB, 3.719708E-04_EB, 4.008233E-04_EB, 3.722047E-04_EB, &
   3.066454E-04_EB, &
   4.675051E-03_EB, 3.770033E-04_EB, 1.005255E-03_EB, 5.231568E-04_EB, 3.365147E-03_EB, 4.031952E-03_EB, &
   1.001045E-02_EB, &
   1.579559E-02_EB, 4.036503E-04_EB, 9.793374E-03_EB, 4.364856E-03_EB, 1.125997E-02_EB, 1.687883E-02_EB, &
   4.109982E-02_EB, &
   1.901642E-02_EB, 7.948388E-03_EB, 2.110029E-02_EB, 1.680041E-02_EB, 3.681206E-02_EB, 5.911015E-02_EB, &
   9.045692E-02_EB, &
   7.057398E-02_EB, 8.615787E-02_EB, 1.147527E-01_EB, 1.077993E-01_EB, 1.296121E-01_EB, 1.341954E-01_EB, &
   1.504629E-01_EB, &
   4.303610E-01_EB, 3.433046E-01_EB, 3.232359E-01_EB, 2.803463E-01_EB, 2.416656E-01_EB, 1.843691E-01_EB, &
   1.701492E-01_EB, &
   6.023500E-01_EB, 3.794196E-01_EB, 3.422745E-01_EB, 2.887488E-01_EB, 2.637113E-01_EB, 2.107903E-01_EB, &
   2.126134E-01_EB/),(/7,8/))

SD4_C2H4(1:7,9:16) = RESHAPE((/ &  ! 3000-3175 CM-1
   7.798028E-01_EB, 4.757862E-01_EB, 3.974023E-01_EB, 3.382752E-01_EB, 2.973165E-01_EB, 2.380506E-01_EB, &
   2.451699E-01_EB, &
   7.172871E-01_EB, 5.831037E-01_EB, 5.314972E-01_EB, 4.800679E-01_EB, 4.358886E-01_EB, 3.600701E-01_EB, &
   3.346981E-01_EB, &
   5.311509E-01_EB, 4.669769E-01_EB, 4.432862E-01_EB, 4.085659E-01_EB, 3.930117E-01_EB, 3.358462E-01_EB, &
   2.929137E-01_EB, &
   8.649072E-01_EB, 5.873164E-01_EB, 4.935644E-01_EB, 4.177742E-01_EB, 3.508014E-01_EB, 2.332914E-01_EB, &
   2.128007E-01_EB, &
   7.478903E-01_EB, 4.640090E-01_EB, 3.775139E-01_EB, 3.190243E-01_EB, 2.690863E-01_EB, 1.975084E-01_EB, &
   1.921879E-01_EB, &
   1.077261E+00_EB, 6.909648E-01_EB, 5.870020E-01_EB, 4.958355E-01_EB, 4.101225E-01_EB, 3.093396E-01_EB, &
   2.599685E-01_EB, &
   7.637944E-01_EB, 5.720239E-01_EB, 5.128777E-01_EB, 4.357737E-01_EB, 3.822630E-01_EB, 3.059218E-01_EB, &
   2.471299E-01_EB, &
   3.843373E-01_EB, 3.060968E-01_EB, 2.831408E-01_EB, 2.396863E-01_EB, 2.332271E-01_EB, 1.957615E-01_EB, &
   1.637279E-01_EB/),(/7,8/))

SD4_C2H4(1:7,17:24) = RESHAPE((/ &  ! 3200-3375 CM-1
   2.099610E-01_EB, 1.791031E-01_EB, 1.774091E-01_EB, 1.486224E-01_EB, 1.585195E-01_EB, 1.392503E-01_EB, &
   1.207946E-01_EB, &
   1.199548E-01_EB, 1.020644E-01_EB, 9.502921E-02_EB, 8.145885E-02_EB, 1.125359E-01_EB, 1.424681E-01_EB, &
   8.669709E-02_EB, &
   4.616169E-01_EB, 7.858449E-02_EB, 7.272250E-02_EB, 2.867532E-02_EB, 1.168091E-01_EB, 1.330542E-01_EB, &
   5.570436E-02_EB, &
   8.691472E-02_EB, 9.220433E-03_EB, 1.896226E-02_EB, 3.974691E-03_EB, 3.169841E-01_EB, 2.218115E-02_EB, &
   5.082209E-02_EB, &
   2.867450E-02_EB, 3.903775E-04_EB, 3.229916E-04_EB, 3.302777E-04_EB, 1.149402E-01_EB, 6.959691E-02_EB, &
   1.770145E-02_EB, &
   3.690812E-03_EB, 3.360817E-04_EB, 3.487733E-04_EB, 4.020878E-04_EB, 3.692208E-04_EB, 3.501982E-04_EB, &
   8.401837E-03_EB, &
   3.318610E-04_EB, 3.541529E-04_EB, 3.588894E-04_EB, 3.595732E-04_EB, 3.703241E-04_EB, 3.649087E-04_EB, &
   2.903059E-04_EB, &
   3.061877E-04_EB, 2.641235E-04_EB, 2.931529E-04_EB, 2.985364E-04_EB, 2.827758E-04_EB, 2.741559E-04_EB, &
   1.355576E-04_EB/),(/7,8/))

SD4_C2H4(1:7,25:25) = RESHAPE((/ &  ! 3400-3400 CM-1
   2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, 2.172015E-04_EB, &
   1.035699E-04_EB/),(/7,1/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD4_C2H4(N_TEMP_C2H4,25))

! BAND #4: 2800 CM-1 - 3400 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.39903 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD4_C2H4(1:7,1:8) = RESHAPE((/ &  ! 2800-2975 CM-1
   2.053802E-04_EB, 2.055037E-04_EB, 2.250544E-04_EB, 2.225113E-04_EB, 2.136803E-04_EB, 2.087893E-04_EB, &
   1.105053E-04_EB, &
   4.253229E-04_EB, 2.343264E-04_EB, 2.233710E-04_EB, 2.452992E-04_EB, 2.751557E-04_EB, 2.508501E-04_EB, &
   2.222309E-04_EB, &
   1.276922E-03_EB, 2.646545E-04_EB, 7.364197E-04_EB, 3.348640E-04_EB, 3.561604E-03_EB, 2.031177E-03_EB, &
   2.949697E-01_EB, &
   1.385628E-02_EB, 2.885267E-04_EB, 9.042876E-01_EB, 2.692734E-03_EB, 9.713740E-03_EB, 4.653156E-03_EB, &
   2.653808E+00_EB, &
   2.397896E-02_EB, 8.918801E-03_EB, 4.080122E-01_EB, 2.625635E-01_EB, 3.017722E-02_EB, 6.429571E-02_EB, &
   6.032175E+00_EB, &
   9.755870E-02_EB, 2.422956E+00_EB, 1.631270E+01_EB, 6.661998E+00_EB, 3.034249E-01_EB, 5.074180E+00_EB, &
   6.298484E+00_EB, &
   4.898871E-01_EB, 2.215583E+00_EB, 6.480333E+01_EB, 1.549944E+01_EB, 1.573448E+00_EB, 4.766076E-01_EB, &
   4.349295E+01_EB, &
   4.836954E-01_EB, 1.451915E+00_EB, 6.485194E+01_EB, 3.798161E+01_EB, 6.188424E-01_EB, 1.109668E+00_EB, &
   4.352421E+01_EB/),(/7,8/))

GAMMAD4_C2H4(1:7,9:16) = RESHAPE((/ &  ! 3000-3175 CM-1
   3.345606E-01_EB, 5.278129E-01_EB, 6.469998E+01_EB, 1.288707E+01_EB, 8.242915E-01_EB, 1.035867E+00_EB, &
   4.352463E+01_EB, &
   1.563175E+00_EB, 2.625564E+00_EB, 6.454977E+01_EB, 6.093823E+01_EB, 2.035803E+00_EB, 1.210794E+00_EB, &
   4.352445E+01_EB, &
   8.952736E-01_EB, 1.070318E+00_EB, 1.932614E+00_EB, 1.478721E+01_EB, 1.380362E+00_EB, 4.608167E-01_EB, &
   4.352008E+01_EB, &
   1.218284E+00_EB, 7.511695E-01_EB, 9.571422E-01_EB, 3.171139E+00_EB, 1.952744E+00_EB, 1.419092E+00_EB, &
   4.351690E+01_EB, &
   1.188336E+00_EB, 5.437254E-01_EB, 7.645704E-01_EB, 1.220082E+00_EB, 1.364237E+00_EB, 2.953192E-01_EB, &
   3.944037E+01_EB, &
   1.069749E+00_EB, 7.136400E-01_EB, 5.480781E-01_EB, 7.511426E-01_EB, 8.072618E-01_EB, 2.442109E-01_EB, &
   1.336082E+01_EB, &
   5.123621E-01_EB, 4.456214E-01_EB, 3.621148E-01_EB, 6.473524E-01_EB, 5.364253E-01_EB, 1.595842E-01_EB, &
   2.523820E+01_EB, &
   2.185443E-01_EB, 1.605263E-01_EB, 1.238794E-01_EB, 2.209229E-01_EB, 1.353909E-01_EB, 5.702117E-02_EB, &
   3.092779E+01_EB/),(/7,8/))

GAMMAD4_C2H4(1:7,17:24) = RESHAPE((/ &  ! 3200-3375 CM-1
   5.244654E-02_EB, 7.872155E-02_EB, 5.156657E-02_EB, 8.595139E-02_EB, 5.856846E-02_EB, 2.220866E-02_EB, &
   1.517203E-01_EB, &
   3.762586E-03_EB, 1.216532E-02_EB, 1.917467E-02_EB, 2.846263E-02_EB, 2.089697E-02_EB, 4.969412E-03_EB, &
   1.387417E-01_EB, &
   4.291422E-05_EB, 7.807504E-04_EB, 1.341258E-03_EB, 6.152627E-03_EB, 2.309479E-03_EB, 1.073258E-03_EB, &
   3.311750E-01_EB, &
   3.584912E-05_EB, 2.671120E-04_EB, 1.957471E-04_EB, 1.342098E-03_EB, 1.278685E-04_EB, 3.324302E-03_EB, &
   3.684340E-03_EB, &
   8.216113E-05_EB, 2.682532E-04_EB, 2.247056E-04_EB, 2.195120E-04_EB, 7.774207E-05_EB, 3.815559E-05_EB, &
   2.872761E-03_EB, &
   1.294855E-04_EB, 2.265159E-04_EB, 2.506069E-04_EB, 2.536762E-04_EB, 2.424375E-04_EB, 2.435020E-04_EB, &
   1.033795E-04_EB, &
   2.206754E-04_EB, 2.418204E-04_EB, 2.208317E-04_EB, 2.403401E-04_EB, 2.291933E-04_EB, 2.353414E-04_EB, &
   2.133339E-04_EB, &
   2.205017E-04_EB, 1.924693E-04_EB, 2.146102E-04_EB, 2.118201E-04_EB, 2.108670E-04_EB, 2.100997E-04_EB, &
   9.661135E-05_EB/),(/7,8/))

GAMMAD4_C2H4(1:7,25:25) = RESHAPE((/ &  ! 3400-3400 CM-1
   1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, 1.709157E-04_EB, &
   8.125358E-05_EB/),(/7,1/))

!-------------------------HEPTANE DATA-------------------


! THERE ARE 2 BANDS FOR HEPTANE

! BAND #1: 1100 CM-1 - 1800 CM-1
! BAND #2: 2550 CM-1 - 3275 CM-1

ALLOCATE(SD_C7H16_TEMP(N_TEMP_C7H16))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR HEPTANE ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_C7H16(N_BAND_C7H16,3))
ALLOCATE(BE_C7H16(N_BAND_C7H16))

OM_BND_C7H16 = RESHAPE((/ &
   1100._EB, 2550._EB, &
   1800._EB, 3275._EB, &
   25._EB, 25._EB/),(/N_BAND_C7H16,3/))

SD_C7H16_TEMP = (/ &
   293._EB, 400._EB, 450._EB, 490._EB, 593._EB, 794._EB,&
   1000._EB/)

BE_C7H16 = (/ &
   1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_C7H16(N_TEMP_C7H16,29))

! BAND #1: 1100 CM-1 - 1800 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.72543 %

SD1_C7H16(1:7,1:8) = RESHAPE((/ &  ! 1100-1275 CM-1
   1.426744E-03_EB, 1.201770E-03_EB, 1.274320E-03_EB, 1.347928E-03_EB, 1.656416E-03_EB, 1.325255E-03_EB, &
   2.057861E-03_EB, &
   1.811130E-03_EB, 2.521922E-03_EB, 1.872758E-03_EB, 1.707256E-03_EB, 1.927722E-03_EB, 1.955475E-03_EB, &
   2.516760E-01_EB, &
   1.747928E-03_EB, 2.120166E-03_EB, 2.046218E-03_EB, 1.649154E-03_EB, 1.666396E-03_EB, 1.992630E-03_EB, &
   2.005874E+00_EB, &
   1.624614E-03_EB, 2.021985E-03_EB, 4.956340E-03_EB, 1.877145E-03_EB, 1.905600E-03_EB, 1.746241E-03_EB, &
   3.502611E-01_EB, &
   1.930396E-03_EB, 1.894347E-03_EB, 1.299287E-02_EB, 1.476436E-02_EB, 4.597890E-03_EB, 2.359195E-02_EB, &
   4.013873E+00_EB, &
   6.004547E-02_EB, 4.017701E-02_EB, 3.934126E-02_EB, 3.083190E-02_EB, 2.179721E-01_EB, 4.014522E-02_EB, &
   2.600484E+00_EB, &
   7.613398E-02_EB, 5.549015E-02_EB, 3.003541E-01_EB, 1.344358E-01_EB, 4.411584E-01_EB, 7.993576E-02_EB, &
   4.665685E+00_EB, &
   1.373972E-01_EB, 1.249248E-01_EB, 1.230316E-01_EB, 1.155939E-01_EB, 3.749324E-01_EB, 1.429788E-01_EB, &
   2.876741E+00_EB/),(/7,8/))

SD1_C7H16(1:7,9:16) = RESHAPE((/ &  ! 1300-1475 CM-1
   1.960350E-01_EB, 7.251248E-01_EB, 4.660543E-01_EB, 2.468984E-01_EB, 3.746538E-01_EB, 1.932187E-01_EB, &
   1.906827E+00_EB, &
   2.182462E-01_EB, 1.037076E+00_EB, 4.532828E-01_EB, 3.341447E-01_EB, 4.124866E-01_EB, 3.832806E-01_EB, &
   1.538802E+00_EB, &
   5.511687E-01_EB, 9.486430E-01_EB, 6.841490E-01_EB, 5.497692E-01_EB, 5.893967E-01_EB, 6.049837E-01_EB, &
   6.095137E+00_EB, &
   2.169296E+00_EB, 1.762238E+00_EB, 1.496288E+00_EB, 1.153363E+00_EB, 9.332138E-01_EB, 6.553914E-01_EB, &
   1.559757E+00_EB, &
   1.152586E+00_EB, 1.398012E+00_EB, 1.008290E+00_EB, 8.691367E-01_EB, 7.857883E-01_EB, 5.751618E-01_EB, &
   3.554067E+00_EB, &
   2.809542E-01_EB, 2.417925E+00_EB, 4.381657E-01_EB, 5.312573E-01_EB, 5.715073E-01_EB, 5.154692E-01_EB, &
   1.149283E+00_EB, &
   2.520765E+00_EB, 2.328593E+00_EB, 1.895157E+00_EB, 1.577062E+00_EB, 1.328041E+00_EB, 1.069522E+00_EB, &
   2.814896E+00_EB, &
   4.096478E+00_EB, 3.000085E+00_EB, 2.158821E+00_EB, 1.809690E+00_EB, 1.347468E+00_EB, 8.991423E-01_EB, &
   2.342350E+00_EB/),(/7,8/))

SD1_C7H16(1:7,17:24) = RESHAPE((/ &  ! 1500-1675 CM-1
   4.388670E-01_EB, 4.751103E-01_EB, 5.711364E-01_EB, 4.125599E-01_EB, 4.895056E-01_EB, 3.320026E-01_EB, &
   2.177032E+00_EB, &
   1.319705E-01_EB, 1.769183E-01_EB, 1.924812E-01_EB, 1.563493E-01_EB, 2.730796E-01_EB, 1.797482E-01_EB, &
   1.084963E+00_EB, &
   1.057772E-01_EB, 1.102722E-01_EB, 1.166527E-01_EB, 7.750779E-02_EB, 1.371680E-01_EB, 7.530555E-02_EB, &
   4.673835E+00_EB, &
   1.595659E-03_EB, 8.558479E-01_EB, 3.816360E-02_EB, 1.698589E-01_EB, 3.558832E-02_EB, 1.126854E+00_EB, &
   6.274235E+00_EB, &
   1.754518E-03_EB, 3.490672E-01_EB, 1.460628E-01_EB, 3.092481E-01_EB, 1.777143E-03_EB, 5.457342E+00_EB, &
   4.901413E+00_EB, &
   1.574552E-03_EB, 2.022684E-03_EB, 1.771614E-03_EB, 8.573879E-02_EB, 1.949107E-03_EB, 8.452931E-01_EB, &
   6.243195E+00_EB, &
   1.827339E-03_EB, 2.129922E-03_EB, 2.108249E-03_EB, 4.442893E-03_EB, 1.724490E-03_EB, 3.038490E-03_EB, &
   3.195029E+00_EB, &
   1.956178E-03_EB, 2.075598E-03_EB, 1.738792E-03_EB, 1.976314E-03_EB, 1.668323E-03_EB, 1.791602E-03_EB, &
   4.975893E+00_EB/),(/7,8/))

SD1_C7H16(1:7,25:29) = RESHAPE((/ &  ! 1700-1800 CM-1
   1.635422E-03_EB, 2.360211E-03_EB, 1.966190E-03_EB, 1.838431E-03_EB, 1.806395E-03_EB, 2.025722E-03_EB, &
   3.595214E+00_EB, &
   1.685412E-03_EB, 1.845260E-03_EB, 1.904451E-03_EB, 1.947180E-03_EB, 1.880024E-03_EB, 1.570401E-03_EB, &
   4.352547E+00_EB, &
   1.693693E-03_EB, 2.153411E-03_EB, 2.016532E-03_EB, 1.811016E-03_EB, 1.826653E-03_EB, 1.799395E-03_EB, &
   1.496407E+00_EB, &
   1.300712E-03_EB, 1.598790E-03_EB, 1.518162E-03_EB, 1.529646E-03_EB, 1.291412E-03_EB, 1.245281E-03_EB, &
   1.384262E-01_EB, &
   9.565465E-04_EB, 9.572222E-04_EB, 9.564413E-04_EB, 9.564850E-04_EB, 9.561277E-04_EB, 9.585852E-04_EB, &
   1.922450E-03_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_C7H16(N_TEMP_C7H16,29))

! BAND #1: 1100 CM-1 - 1800 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.72543 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_C7H16(1:7,1:8) = RESHAPE((/ &  ! 1100-1275 CM-1
   1.079377E-03_EB, 8.321757E-04_EB, 1.106526E-03_EB, 8.589890E-04_EB, 1.218045E-03_EB, 9.003631E-04_EB, &
   1.652303E-03_EB, &
   1.219534E-03_EB, 1.678161E-03_EB, 1.027749E-03_EB, 1.087240E-03_EB, 1.300295E-03_EB, 1.381191E-03_EB, &
   2.766162E-01_EB, &
   1.117750E-03_EB, 1.217712E-03_EB, 1.293043E-03_EB, 1.061960E-03_EB, 1.158654E-03_EB, 1.485883E-03_EB, &
   6.838761E-04_EB, &
   1.138939E-03_EB, 1.252798E-03_EB, 2.807650E-03_EB, 1.167864E-03_EB, 1.225067E-03_EB, 1.314096E-03_EB, &
   2.282961E-02_EB, &
   1.420120E-03_EB, 1.020475E-03_EB, 6.385377E-04_EB, 2.240175E-03_EB, 1.390330E-03_EB, 3.503565E-01_EB, &
   2.532075E-04_EB, &
   1.303079E-02_EB, 1.684651E-02_EB, 1.656972E-02_EB, 6.785015E-03_EB, 1.293076E-04_EB, 1.692435E-02_EB, &
   9.530208E-04_EB, &
   1.370385E-02_EB, 1.586469E-02_EB, 1.596226E-04_EB, 6.447976E-04_EB, 2.699714E-04_EB, 4.064881E-01_EB, &
   7.587172E-04_EB, &
   1.038621E-02_EB, 6.544712E-03_EB, 6.378851E-03_EB, 9.284895E-03_EB, 4.414444E-04_EB, 4.010176E-01_EB, &
   1.276912E-03_EB/),(/7,8/))

GAMMAD1_C7H16(1:7,9:16) = RESHAPE((/ &  ! 1300-1475 CM-1
   9.663644E-02_EB, 3.809049E-04_EB, 7.060483E-04_EB, 2.346751E-03_EB, 1.601612E-03_EB, 2.284259E-01_EB, &
   3.174484E-03_EB, &
   2.648317E-02_EB, 4.915970E-04_EB, 1.642486E-03_EB, 3.803913E-03_EB, 3.050190E-03_EB, 1.055777E-02_EB, &
   4.320127E-03_EB, &
   3.165256E-02_EB, 3.335451E-03_EB, 6.233122E-03_EB, 1.162381E-02_EB, 1.333659E-02_EB, 1.163372E-02_EB, &
   8.774580E-04_EB, &
   7.316298E-02_EB, 3.067612E-02_EB, 2.893574E-02_EB, 4.067615E-02_EB, 5.248547E-02_EB, 4.649271E-01_EB, &
   1.044691E-02_EB, &
   4.533824E-02_EB, 1.035218E-02_EB, 1.958798E-02_EB, 2.401292E-02_EB, 2.064596E-02_EB, 3.122001E-02_EB, &
   3.284470E-03_EB, &
   2.887489E-02_EB, 4.618211E-04_EB, 1.291555E-02_EB, 9.570390E-03_EB, 1.139081E-02_EB, 4.131392E-02_EB, &
   2.376013E-02_EB, &
   9.435371E-02_EB, 3.559624E-02_EB, 4.477145E-02_EB, 5.321919E-02_EB, 5.410966E-02_EB, 4.965735E-02_EB, &
   3.060972E-03_EB, &
   1.434765E-01_EB, 5.756528E-02_EB, 7.898010E-02_EB, 6.856662E-02_EB, 6.369824E-02_EB, 6.660056E-02_EB, &
   3.552779E-03_EB/),(/7,8/))

GAMMAD1_C7H16(1:7,17:24) = RESHAPE((/ &  ! 1500-1675 CM-1
   6.164439E-02_EB, 4.097206E-02_EB, 2.017864E-02_EB, 5.078487E-02_EB, 2.150953E-02_EB, 6.143261E-01_EB, &
   2.140967E-03_EB, &
   4.999099E-02_EB, 6.175675E-01_EB, 5.926015E-01_EB, 5.735879E-01_EB, 3.127071E-03_EB, 8.449651E-02_EB, &
   4.359548E-03_EB, &
   2.308160E-01_EB, 2.876922E+00_EB, 2.678573E+00_EB, 3.088436E-01_EB, 1.022662E-03_EB, 4.348389E-01_EB, &
   5.138371E-04_EB, &
   1.136554E-03_EB, 2.312863E-05_EB, 1.191809E+00_EB, 5.400494E-04_EB, 8.452069E-04_EB, 3.414922E-05_EB, &
   2.732241E-04_EB, &
   1.236428E-03_EB, 7.575429E-06_EB, 1.804662E-05_EB, 1.168575E-04_EB, 1.132712E-03_EB, 1.348325E-05_EB, &
   2.605356E-04_EB, &
   1.005482E-03_EB, 1.099106E-03_EB, 1.146793E-03_EB, 1.793408E-04_EB, 1.319446E-03_EB, 2.932722E-05_EB, &
   2.217645E-04_EB, &
   1.351400E-03_EB, 1.302077E-03_EB, 1.237548E-03_EB, 2.270156E-03_EB, 1.159496E-03_EB, 2.121847E-03_EB, &
   2.405564E-04_EB, &
   1.421491E-03_EB, 1.320464E-03_EB, 1.069063E-03_EB, 1.283897E-03_EB, 9.514195E-04_EB, 1.174814E-03_EB, &
   1.274543E-04_EB/),(/7,8/))

GAMMAD1_C7H16(1:7,25:29) = RESHAPE((/ &  ! 1700-1800 CM-1
   1.029018E-03_EB, 1.522013E-03_EB, 1.190609E-03_EB, 1.137355E-03_EB, 1.227873E-03_EB, 1.521396E-03_EB, &
   4.984208E-05_EB, &
   1.062872E-03_EB, 1.240560E-03_EB, 1.115838E-03_EB, 1.071608E-03_EB, 1.073014E-03_EB, 1.091421E-03_EB, &
   1.043998E-04_EB, &
   1.264467E-03_EB, 1.452995E-03_EB, 1.276660E-03_EB, 1.062100E-03_EB, 1.296189E-03_EB, 1.314713E-03_EB, &
   1.161390E-04_EB, &
   9.490158E-04_EB, 1.088570E-03_EB, 1.017566E-03_EB, 1.036075E-03_EB, 9.453163E-04_EB, 9.054199E-04_EB, &
   1.849978E-04_EB, &
   7.632406E-04_EB, 7.638486E-04_EB, 7.631088E-04_EB, 7.630990E-04_EB, 7.628205E-04_EB, 7.652755E-04_EB, &
   1.536272E-03_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_C7H16(N_TEMP_C7H16,30))

! BAND #2: 2550 CM-1 - 3275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 5.5204 %

SD2_C7H16(1:7,1:8) = RESHAPE((/ &  ! 2550-2725 CM-1
   1.230745E-03_EB, 1.059249E-02_EB, 5.258556E-03_EB, 8.983219E-03_EB, 1.071517E-02_EB, 1.270153E-03_EB, &
   2.238178E-03_EB, &
   1.746456E-02_EB, 4.672224E-02_EB, 3.759388E-02_EB, 4.616045E-02_EB, 5.001133E-02_EB, 1.722111E-02_EB, &
   2.491507E-02_EB, &
   1.081581E-01_EB, 1.027192E-01_EB, 8.800407E-02_EB, 8.251442E-02_EB, 6.268006E-02_EB, 4.814096E-02_EB, &
   7.543431E-02_EB, &
   1.324171E-01_EB, 1.186396E-01_EB, 8.522107E-02_EB, 8.615581E-02_EB, 7.161614E-02_EB, 4.907251E-02_EB, &
   4.189169E-02_EB, &
   1.360592E-01_EB, 1.511535E-01_EB, 9.963712E-02_EB, 9.989906E-02_EB, 7.841100E-02_EB, 6.989163E-02_EB, &
   1.135539E-01_EB, &
   1.798230E-01_EB, 2.192535E-01_EB, 1.208341E-01_EB, 1.243637E-01_EB, 2.183718E-01_EB, 7.565057E-02_EB, &
   8.186008E-02_EB, &
   2.869951E-01_EB, 2.102309E-01_EB, 1.285038E-01_EB, 1.353365E-01_EB, 1.198975E-01_EB, 1.097888E-01_EB, &
   8.920775E-02_EB, &
   4.131559E-01_EB, 3.718172E-01_EB, 1.977670E-01_EB, 1.999998E-01_EB, 2.661506E-01_EB, 1.630822E-01_EB, &
   1.430789E-01_EB/),(/7,8/))

SD2_C7H16(1:7,9:16) = RESHAPE((/ &  ! 2750-2925 CM-1
   4.814798E-01_EB, 4.371742E-01_EB, 2.946385E-01_EB, 2.539099E-01_EB, 3.004489E-01_EB, 2.639523E-01_EB, &
   2.106653E-01_EB, &
   4.711680E-01_EB, 4.231342E-01_EB, 2.816821E-01_EB, 2.662044E-01_EB, 2.714629E-01_EB, 4.380265E-01_EB, &
   2.595817E-01_EB, &
   6.539776E-01_EB, 7.016029E-01_EB, 5.178696E-01_EB, 4.535956E-01_EB, 5.002499E-01_EB, 7.197349E-01_EB, &
   3.199467E-01_EB, &
   1.241588E+00_EB, 1.255753E+00_EB, 9.891572E-01_EB, 8.923575E-01_EB, 8.902290E-01_EB, 1.114933E+00_EB, &
   4.631429E-01_EB, &
   5.884384E+00_EB, 4.563318E+00_EB, 3.951471E+00_EB, 3.339596E+00_EB, 2.714647E+00_EB, 2.641774E+00_EB, &
   7.520423E-01_EB, &
   1.531155E+01_EB, 1.096024E+01_EB, 9.162135E+00_EB, 7.803908E+00_EB, 5.562181E+00_EB, 4.589694E+00_EB, &
   1.056916E+00_EB, &
   1.317309E+01_EB, 1.037752E+01_EB, 9.226521E+00_EB, 8.166911E+00_EB, 6.406625E+00_EB, 6.017800E+00_EB, &
   1.580795E+00_EB, &
   2.423199E+01_EB, 1.841550E+01_EB, 1.601599E+01_EB, 1.391139E+01_EB, 1.034058E+01_EB, 9.135221E+00_EB, &
   2.166977E+00_EB/),(/7,8/))

SD2_C7H16(1:7,17:24) = RESHAPE((/ &  ! 2950-3125 CM-1
   2.666381E+01_EB, 2.072769E+01_EB, 1.799865E+01_EB, 1.588750E+01_EB, 1.152743E+01_EB, 9.524580E+00_EB, &
   2.012903E+00_EB, &
   3.139322E+01_EB, 2.055468E+01_EB, 1.643220E+01_EB, 1.409534E+01_EB, 8.965984E+00_EB, 6.302236E+00_EB, &
   1.518462E+00_EB, &
   2.594752E+00_EB, 2.921881E+00_EB, 2.424761E+00_EB, 2.367343E+00_EB, 1.941979E+00_EB, 1.810548E+00_EB, &
   1.163951E+00_EB, &
   9.124058E-01_EB, 1.242167E+00_EB, 8.462169E-01_EB, 7.637253E-01_EB, 7.930042E-01_EB, 7.315673E-01_EB, &
   9.222189E-01_EB, &
   5.694335E-01_EB, 1.040431E+00_EB, 4.731323E-01_EB, 4.570560E-01_EB, 5.207545E-01_EB, 4.518627E-01_EB, &
   7.258027E-01_EB, &
   4.615311E-01_EB, 7.262374E-01_EB, 2.975859E-01_EB, 3.257008E-01_EB, 3.595477E-01_EB, 3.278019E-01_EB, &
   5.488059E-01_EB, &
   4.537172E-01_EB, 5.187525E-01_EB, 2.166416E-01_EB, 2.523836E-01_EB, 3.336731E-01_EB, 2.793265E-01_EB, &
   4.844761E-01_EB, &
   1.173983E-01_EB, 4.043441E-01_EB, 1.764942E-01_EB, 2.034332E-01_EB, 2.260464E-01_EB, 2.126733E-01_EB, &
   5.112998E-01_EB/),(/7,8/))

SD2_C7H16(1:7,25:30) = RESHAPE((/ &  ! 3150-3275 CM-1
   1.745347E-01_EB, 3.869948E-01_EB, 1.606330E-01_EB, 1.691991E-01_EB, 1.773258E-01_EB, 1.545312E-01_EB, &
   3.754934E-01_EB, &
   2.125656E-01_EB, 3.845411E-01_EB, 1.519506E-01_EB, 1.522530E-01_EB, 1.848037E-01_EB, 1.013866E-01_EB, &
   2.632396E-01_EB, &
   2.543277E-01_EB, 2.891570E-01_EB, 9.975616E-02_EB, 1.088793E-01_EB, 1.322502E-01_EB, 5.343256E-02_EB, &
   1.951975E-01_EB, &
   6.412029E-02_EB, 2.180048E-01_EB, 9.631661E-02_EB, 5.927599E-02_EB, 1.626811E-02_EB, 2.207700E-02_EB, &
   1.199629E-01_EB, &
   4.940691E-03_EB, 1.520018E-02_EB, 8.071023E-03_EB, 1.120400E-02_EB, 5.605272E-03_EB, 1.455284E-03_EB, &
   4.956089E-02_EB, &
   9.565465E-04_EB, 9.572222E-04_EB, 9.564413E-04_EB, 9.564850E-04_EB, 9.561277E-04_EB, 9.585852E-04_EB, &
   1.922450E-03_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_C7H16(N_TEMP_C7H16,30))

! BAND #2: 2550 CM-1 - 3275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 5.5204 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_C7H16(1:7,1:8) = RESHAPE((/ &  ! 2550-2725 CM-1
   9.146079E-04_EB, 1.220358E-02_EB, 3.937261E-03_EB, 7.187786E-03_EB, 5.227505E-03_EB, 9.355109E-04_EB, &
   1.819838E-03_EB, &
   9.874420E-03_EB, 2.084439E-01_EB, 2.173266E-01_EB, 8.182000E-03_EB, 2.973184E-02_EB, 9.482167E-03_EB, &
   5.745211E-01_EB, &
   9.786100E-03_EB, 1.858210E-01_EB, 1.533293E-02_EB, 5.349239E-02_EB, 2.414250E-02_EB, 4.205505E-04_EB, &
   3.283538E-02_EB, &
   1.016087E-02_EB, 5.218556E-02_EB, 6.697325E-02_EB, 2.379606E-01_EB, 8.937135E-03_EB, 9.990945E-03_EB, &
   5.966306E-01_EB, &
   1.783704E-02_EB, 2.299629E-02_EB, 4.295102E-02_EB, 6.377060E-02_EB, 3.169911E-02_EB, 5.529323E-03_EB, &
   7.456241E-01_EB, &
   9.757832E-03_EB, 5.468131E-03_EB, 4.867751E-02_EB, 1.172578E-01_EB, 4.899078E-04_EB, 2.599541E-02_EB, &
   1.519840E+00_EB, &
   1.701122E-03_EB, 1.906085E-02_EB, 8.916232E-02_EB, 3.655990E-01_EB, 7.351637E-03_EB, 3.626665E-02_EB, &
   1.919410E+00_EB, &
   3.999808E-03_EB, 7.531740E-03_EB, 1.072796E-01_EB, 1.003155E+00_EB, 1.614637E-03_EB, 7.721701E-02_EB, &
   1.224303E+00_EB/),(/7,8/))

GAMMAD2_C7H16(1:7,9:16) = RESHAPE((/ &  ! 2750-2925 CM-1
   4.880113E-03_EB, 8.565604E-03_EB, 1.367064E-02_EB, 3.997591E-01_EB, 4.442547E-03_EB, 1.915330E-02_EB, &
   5.051128E+00_EB, &
   1.825295E-03_EB, 7.512748E-03_EB, 1.729303E-02_EB, 1.104825E+00_EB, 1.586989E-02_EB, 1.138387E-02_EB, &
   1.724797E+00_EB, &
   8.938939E-03_EB, 1.035703E-02_EB, 1.615494E-02_EB, 9.389559E-02_EB, 1.525842E-02_EB, 1.850825E-02_EB, &
   3.698894E+00_EB, &
   2.478387E-02_EB, 2.301877E-02_EB, 4.297098E-02_EB, 1.503277E-01_EB, 6.014925E-02_EB, 8.742386E-02_EB, &
   2.622943E+00_EB, &
   1.601922E-01_EB, 1.114843E-01_EB, 1.112220E-01_EB, 1.804939E-01_EB, 2.339949E-01_EB, 1.584145E-01_EB, &
   9.606857E+00_EB, &
   5.154427E-01_EB, 3.320188E-01_EB, 3.025580E-01_EB, 3.557579E-01_EB, 6.640807E-01_EB, 3.241466E-01_EB, &
   2.268534E+01_EB, &
   4.187811E-01_EB, 3.414132E-01_EB, 3.225035E-01_EB, 3.833688E-01_EB, 7.058557E-01_EB, 4.035161E-01_EB, &
   4.330011E+01_EB, &
   8.053331E-01_EB, 5.054746E-01_EB, 4.771400E-01_EB, 5.477855E-01_EB, 1.088287E+00_EB, 5.418580E-01_EB, &
   4.326237E+01_EB/),(/7,8/))

GAMMAD2_C7H16(1:7,17:24) = RESHAPE((/ &  ! 2950-3125 CM-1
   9.236854E-01_EB, 6.215496E-01_EB, 5.912471E-01_EB, 6.204509E-01_EB, 1.340069E+00_EB, 6.140174E-01_EB, &
   4.330342E+01_EB, &
   5.662789E-01_EB, 5.083725E-01_EB, 4.767821E-01_EB, 4.500428E-01_EB, 1.099144E+00_EB, 4.936755E-01_EB, &
   2.497692E+01_EB, &
   5.143360E-02_EB, 5.233965E-02_EB, 8.660531E-02_EB, 8.976548E-02_EB, 1.810249E-01_EB, 2.371358E-01_EB, &
   4.306803E+01_EB, &
   1.542965E-02_EB, 1.268218E-02_EB, 2.772033E-02_EB, 9.250133E-02_EB, 2.649493E-02_EB, 3.172831E-01_EB, &
   8.950148E+00_EB, &
   8.562522E-03_EB, 4.977221E-03_EB, 3.610411E-02_EB, 1.439318E+00_EB, 1.049615E-02_EB, 6.108043E-01_EB, &
   4.534910E+00_EB, &
   1.597126E-03_EB, 3.635349E-03_EB, 5.555783E-02_EB, 3.967871E+00_EB, 6.887030E-03_EB, 5.613768E-01_EB, &
   3.059256E+00_EB, &
   6.799236E-04_EB, 3.384843E-03_EB, 1.010555E-01_EB, 2.884777E+00_EB, 2.380478E-03_EB, 1.681744E-01_EB, &
   2.528805E+00_EB, &
   8.299446E-03_EB, 2.940292E-03_EB, 9.141446E-02_EB, 3.099637E+00_EB, 1.826256E-03_EB, 9.931857E-02_EB, &
   3.001536E+00_EB/),(/7,8/))

GAMMAD2_C7H16(1:7,25:30) = RESHAPE((/ &  ! 3150-3275 CM-1
   1.022601E-02_EB, 2.439564E-03_EB, 2.129847E-02_EB, 1.571134E+00_EB, 1.678148E-03_EB, 9.784188E-02_EB, &
   7.913187E+00_EB, &
   2.072404E-02_EB, 2.738722E-03_EB, 2.883252E-02_EB, 5.459085E-01_EB, 6.513406E-04_EB, 1.967974E-01_EB, &
   1.236417E+00_EB, &
   3.185761E-03_EB, 2.226331E-03_EB, 3.340747E-02_EB, 2.954516E+00_EB, 3.127038E-04_EB, 3.897705E-02_EB, &
   5.973433E-01_EB, &
   3.201661E-03_EB, 5.002671E-04_EB, 5.027477E-04_EB, 5.059749E-01_EB, 2.581852E-03_EB, 5.653921E-03_EB, &
   4.902323E-01_EB, &
   3.091588E-03_EB, 1.287212E-03_EB, 4.025758E-03_EB, 1.683050E-02_EB, 2.725481E-03_EB, 9.810405E-04_EB, &
   1.476595E-02_EB, &
   7.632406E-04_EB, 7.638486E-04_EB, 7.631088E-04_EB, 7.630990E-04_EB, 7.628205E-04_EB, 7.652755E-04_EB, &
   1.536272E-03_EB/),(/7,6/))

!-------------------------METHANOL DATA-------------------


! THERE ARE 4 BANDS FOR METHANOL

! BAND #1: 825 CM-1 - 1150 CM-1
! BAND #2: 1125 CM-1 - 1700 CM-1
! BAND #3: 2600 CM-1 - 3225 CM-1
! BAND #4: 3525 CM-1 - 3850 CM-1

ALLOCATE(SD_CH3OH_TEMP(N_TEMP_CH3OH))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR METHANOL ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_CH3OH(N_BAND_CH3OH,3))
ALLOCATE(BE_CH3OH(N_BAND_CH3OH))

OM_BND_CH3OH = RESHAPE((/ &
   825._EB, 1125._EB, 2600._EB, 3525._EB, &
   1150._EB, 1700._EB, 3225._EB, 3850._EB, &
   5._EB, 25._EB, 25._EB, 25._EB/),(/N_BAND_CH3OH,3/))

SD_CH3OH_TEMP = (/ &
   293._EB, 396._EB, 443._EB, 483._EB, 570._EB, 804._EB,&
   1000._EB/)

BE_CH3OH = (/ &
   1.000_EB, 1.000_EB, 1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_CH3OH(N_TEMP_CH3OH,58))

! BAND #1: 825 CM-1 - 1150 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 3.9209 %

SD1_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 825-860 CM-1
   7.075310E-04_EB, 8.159230E-04_EB, 6.305365E-04_EB, 6.771661E-04_EB, 6.382997E-04_EB, 6.733725E-04_EB, &
   5.317197E-04_EB, &
   1.027754E-03_EB, 7.803944E-04_EB, 9.690581E-04_EB, 8.327490E-04_EB, 1.014832E-03_EB, 1.009424E-03_EB, &
   4.548089E-02_EB, &
   9.923660E-04_EB, 7.762778E-04_EB, 8.630137E-04_EB, 1.137131E-03_EB, 8.918616E-04_EB, 1.053434E-03_EB, &
   2.195153E+00_EB, &
   9.759840E-04_EB, 1.083038E-03_EB, 9.878635E-04_EB, 9.020455E-04_EB, 9.799016E-04_EB, 6.911637E-04_EB, &
   6.514857E-01_EB, &
   1.006816E-03_EB, 9.248700E-04_EB, 8.933603E-04_EB, 9.409542E-04_EB, 7.796524E-04_EB, 1.033193E-03_EB, &
   3.104940E-01_EB, &
   1.262552E-03_EB, 9.296738E-04_EB, 1.134121E-03_EB, 1.041003E-03_EB, 8.506538E-04_EB, 1.025871E-03_EB, &
   4.257321E-01_EB, &
   9.314553E-04_EB, 9.719139E-04_EB, 1.050566E-03_EB, 1.107681E-03_EB, 7.432552E-04_EB, 7.341790E-03_EB, &
   2.976279E-01_EB, &
   9.166184E-04_EB, 9.949250E-04_EB, 1.142691E-03_EB, 1.003991E-03_EB, 9.570837E-04_EB, 2.474105E-02_EB, &
   3.966025E-02_EB/),(/7,8/))

SD1_CH3OH(1:7,9:16) = RESHAPE((/ &  ! 865-900 CM-1
   8.873778E-04_EB, 1.014874E-03_EB, 8.954091E-04_EB, 8.658071E-04_EB, 8.266735E-04_EB, 3.960677E-02_EB, &
   4.795769E-01_EB, &
   9.318468E-04_EB, 1.081375E-03_EB, 8.779975E-04_EB, 6.417478E-04_EB, 8.380587E-04_EB, 5.064673E-02_EB, &
   2.069845E+00_EB, &
   9.172948E-04_EB, 6.246642E-04_EB, 8.418469E-04_EB, 1.002610E-03_EB, 8.748320E-04_EB, 6.309985E-02_EB, &
   2.535762E+00_EB, &
   5.635661E-04_EB, 9.022735E-04_EB, 7.536768E-04_EB, 1.384075E-02_EB, 4.169450E-03_EB, 6.860592E-02_EB, &
   4.255594E+00_EB, &
   5.238187E-04_EB, 9.020580E-04_EB, 8.975650E-04_EB, 2.989369E-02_EB, 2.964629E-03_EB, 5.639540E-01_EB, &
   4.081000E+00_EB, &
   4.272194E-03_EB, 1.285085E-03_EB, 1.101707E-03_EB, 3.363383E-02_EB, 2.223112E-03_EB, 7.068202E-01_EB, &
   5.035905E+00_EB, &
   7.391996E-03_EB, 1.149033E-03_EB, 4.043676E-03_EB, 3.024886E-02_EB, 1.237039E-02_EB, 2.917510E-01_EB, &
   5.127708E+00_EB, &
   1.769445E-02_EB, 1.162378E-03_EB, 1.101607E-02_EB, 3.265720E-02_EB, 2.700163E-02_EB, 2.303736E-01_EB, &
   7.211748E+00_EB/),(/7,8/))

SD1_CH3OH(1:7,17:24) = RESHAPE((/ &  ! 905-940 CM-1
   1.041542E-02_EB, 1.114301E-02_EB, 1.311655E-02_EB, 4.064275E-02_EB, 7.266914E-02_EB, 2.423119E-01_EB, &
   4.957147E+00_EB, &
   2.050418E-02_EB, 1.592394E-02_EB, 3.294495E-02_EB, 5.520246E-02_EB, 7.496393E-02_EB, 2.368645E-01_EB, &
   1.130981E+01_EB, &
   2.109122E-02_EB, 1.222826E-02_EB, 4.727193E-02_EB, 8.561265E-02_EB, 1.062168E-01_EB, 1.049801E+00_EB, &
   7.996949E+00_EB, &
   3.772088E-02_EB, 8.854601E-03_EB, 6.435677E-02_EB, 7.648012E-02_EB, 1.545208E-01_EB, 2.887097E-01_EB, &
   3.363412E+00_EB, &
   5.889149E-02_EB, 2.614746E-02_EB, 8.847350E-02_EB, 1.405710E-01_EB, 1.772912E-01_EB, 3.718949E-01_EB, &
   9.837331E+00_EB, &
   6.654570E-02_EB, 5.749401E-02_EB, 9.802181E-02_EB, 1.688979E-01_EB, 2.243950E-01_EB, 1.220047E+00_EB, &
   1.655983E+01_EB, &
   7.945784E-02_EB, 9.272308E-02_EB, 1.367127E-01_EB, 1.949308E-01_EB, 2.725708E-01_EB, 6.381513E-01_EB, &
   1.228651E+01_EB, &
   9.776093E-02_EB, 1.303068E-01_EB, 1.815588E-01_EB, 2.578179E-01_EB, 3.469963E-01_EB, 5.560945E-01_EB, &
   1.183943E+01_EB/),(/7,8/))

SD1_CH3OH(1:7,25:32) = RESHAPE((/ &  ! 945-980 CM-1
   1.006010E-01_EB, 1.730636E-01_EB, 2.270729E-01_EB, 3.315982E-01_EB, 4.573468E-01_EB, 6.708389E-01_EB, &
   1.024291E+01_EB, &
   1.194413E-01_EB, 2.369771E-01_EB, 3.106016E-01_EB, 4.274250E-01_EB, 5.708632E-01_EB, 7.929006E-01_EB, &
   1.330083E+01_EB, &
   1.736727E-01_EB, 3.344439E-01_EB, 4.261052E-01_EB, 5.527598E-01_EB, 7.561373E-01_EB, 9.600217E-01_EB, &
   1.114027E+01_EB, &
   2.600298E-01_EB, 4.700460E-01_EB, 5.820182E-01_EB, 7.302241E-01_EB, 9.234681E-01_EB, 1.070772E+00_EB, &
   1.188692E+01_EB, &
   3.671640E-01_EB, 6.491391E-01_EB, 7.852073E-01_EB, 9.389176E-01_EB, 1.126578E+00_EB, 1.172724E+00_EB, &
   6.310257E+00_EB, &
   5.738182E-01_EB, 9.168206E-01_EB, 1.052392E+00_EB, 1.223815E+00_EB, 1.363750E+00_EB, 1.356380E+00_EB, &
   1.276687E+01_EB, &
   8.847952E-01_EB, 1.262063E+00_EB, 1.370575E+00_EB, 1.545198E+00_EB, 1.612648E+00_EB, 1.449724E+00_EB, &
   1.203918E+01_EB, &
   1.380983E+00_EB, 1.728933E+00_EB, 1.776913E+00_EB, 1.952279E+00_EB, 1.903225E+00_EB, 1.560147E+00_EB, &
   1.409457E+01_EB/),(/7,8/))

SD1_CH3OH(1:7,33:40) = RESHAPE((/ &  ! 985-1020 CM-1
   1.989792E+00_EB, 2.216523E+00_EB, 2.173252E+00_EB, 2.302944E+00_EB, 2.154186E+00_EB, 1.615229E+00_EB, &
   1.614576E+01_EB, &
   2.901685E+00_EB, 2.807597E+00_EB, 2.638654E+00_EB, 2.691349E+00_EB, 2.424882E+00_EB, 1.732381E+00_EB, &
   1.516364E+01_EB, &
   3.935938E+00_EB, 3.363705E+00_EB, 3.049578E+00_EB, 3.000715E+00_EB, 2.568481E+00_EB, 1.745273E+00_EB, &
   2.252866E+01_EB, &
   5.148871E+00_EB, 3.896474E+00_EB, 3.364794E+00_EB, 3.239344E+00_EB, 2.682511E+00_EB, 1.748386E+00_EB, &
   2.390649E+01_EB, &
   6.170452E+00_EB, 4.190528E+00_EB, 3.476275E+00_EB, 3.307799E+00_EB, 2.673466E+00_EB, 1.690532E+00_EB, &
   2.170835E+01_EB, &
   6.905606E+00_EB, 4.214646E+00_EB, 3.441984E+00_EB, 3.180517E+00_EB, 2.517520E+00_EB, 1.596762E+00_EB, &
   1.932637E+01_EB, &
   6.940706E+00_EB, 3.956671E+00_EB, 3.247771E+00_EB, 2.994245E+00_EB, 2.445520E+00_EB, 1.673228E+00_EB, &
   1.306371E+01_EB, &
   5.958695E+00_EB, 3.454990E+00_EB, 2.948441E+00_EB, 2.807961E+00_EB, 2.435395E+00_EB, 1.873623E+00_EB, &
   1.081489E+01_EB/),(/7,8/))

SD1_CH3OH(1:7,41:48) = RESHAPE((/ &  ! 1025-1060 CM-1
   4.201331E+00_EB, 2.701090E+00_EB, 2.459137E+00_EB, 2.429934E+00_EB, 2.270440E+00_EB, 1.776380E+00_EB, &
   1.211452E+01_EB, &
   6.021324E+00_EB, 4.632103E+00_EB, 4.201019E+00_EB, 4.002494E+00_EB, 3.407819E+00_EB, 2.266345E+00_EB, &
   1.495652E+01_EB, &
   9.188156E+00_EB, 5.163112E+00_EB, 4.909468E+00_EB, 4.147644E+00_EB, 3.320954E+00_EB, 1.941572E+00_EB, &
   1.668271E+01_EB, &
   3.550622E+00_EB, 2.256569E+00_EB, 2.077046E+00_EB, 1.997234E+00_EB, 1.831325E+00_EB, 1.549349E+00_EB, &
   1.016637E+01_EB, &
   6.406905E+00_EB, 3.553412E+00_EB, 2.981786E+00_EB, 2.818896E+00_EB, 2.393274E+00_EB, 1.845333E+00_EB, &
   1.267998E+01_EB, &
   9.119940E+00_EB, 5.130999E+00_EB, 4.171838E+00_EB, 3.898993E+00_EB, 3.183156E+00_EB, 2.173550E+00_EB, &
   1.068448E+01_EB, &
   1.001487E+01_EB, 6.103507E+00_EB, 4.932645E+00_EB, 4.654735E+00_EB, 3.723719E+00_EB, 2.430534E+00_EB, &
   1.889930E+01_EB, &
   8.945286E+00_EB, 6.111634E+00_EB, 5.090176E+00_EB, 4.850566E+00_EB, 3.927607E+00_EB, 2.511476E+00_EB, &
   2.721090E+01_EB/),(/7,8/))

SD1_CH3OH(1:7,49:56) = RESHAPE((/ &  ! 1065-1100 CM-1
   6.548008E+00_EB, 5.234485E+00_EB, 4.593759E+00_EB, 4.456866E+00_EB, 3.759323E+00_EB, 2.389545E+00_EB, &
   1.930330E+01_EB, &
   4.016006E+00_EB, 3.880275E+00_EB, 3.590672E+00_EB, 3.603798E+00_EB, 3.180636E+00_EB, 2.124012E+00_EB, &
   7.098842E+00_EB, &
   2.039641E+00_EB, 2.438251E+00_EB, 2.437403E+00_EB, 2.528162E+00_EB, 2.390431E+00_EB, 1.727051E+00_EB, &
   1.564504E+01_EB, &
   8.890902E-01_EB, 1.306925E+00_EB, 1.439104E+00_EB, 1.573364E+00_EB, 1.582749E+00_EB, 1.325349E+00_EB, &
   1.309975E+01_EB, &
   4.289045E-01_EB, 6.181922E-01_EB, 7.402675E-01_EB, 8.358344E-01_EB, 9.479778E-01_EB, 8.718487E-01_EB, &
   8.578315E+00_EB, &
   3.110189E-01_EB, 3.181081E-01_EB, 3.757707E-01_EB, 4.355864E-01_EB, 4.933576E-01_EB, 9.265380E-01_EB, &
   5.124942E+00_EB, &
   3.078577E-01_EB, 2.309721E-01_EB, 2.504640E-01_EB, 2.713709E-01_EB, 2.640710E-01_EB, 6.768124E-01_EB, &
   7.096896E+00_EB, &
   2.629835E-01_EB, 1.842053E-01_EB, 1.854752E-01_EB, 2.041121E-01_EB, 2.026328E-01_EB, 4.463207E-01_EB, &
   1.030077E+01_EB/),(/7,8/))

SD1_CH3OH(1:7,57:58) = RESHAPE((/ &  ! 1125-1150 CM-1
   1.066160E-01_EB, 8.237972E-02_EB, 7.650034E-02_EB, 7.750927E-02_EB, 8.650473E-02_EB, 5.002688E-01_EB, &
   2.961022E+00_EB, &
   4.699306E-04_EB, 4.696984E-04_EB, 4.697130E-04_EB, 4.702606E-04_EB, 4.699096E-04_EB, 4.682572E-04_EB, &
   4.691749E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_CH3OH(N_TEMP_CH3OH,58))

! BAND #1: 825 CM-1 - 1150 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 3.9209 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 825-860 CM-1
   6.320478E-04_EB, 6.861763E-04_EB, 4.976022E-04_EB, 5.212687E-04_EB, 4.951307E-04_EB, 4.722485E-04_EB, &
   4.126358E-04_EB, &
   5.857091E-04_EB, 4.986862E-04_EB, 7.181338E-04_EB, 6.000370E-04_EB, 8.426048E-04_EB, 6.854139E-04_EB, &
   9.797634E-05_EB, &
   6.687840E-04_EB, 5.631883E-04_EB, 6.493401E-04_EB, 8.974743E-04_EB, 8.364450E-04_EB, 9.128100E-04_EB, &
   2.228170E-05_EB, &
   7.015973E-04_EB, 8.509882E-04_EB, 6.749386E-04_EB, 7.104894E-04_EB, 9.244817E-04_EB, 4.993607E-04_EB, &
   3.618598E-05_EB, &
   6.879541E-04_EB, 6.793889E-04_EB, 6.064846E-04_EB, 7.494890E-04_EB, 6.021521E-04_EB, 8.041830E-04_EB, &
   4.227068E-05_EB, &
   1.115437E-03_EB, 6.123108E-04_EB, 9.950372E-04_EB, 7.526989E-04_EB, 5.762193E-04_EB, 6.564205E-04_EB, &
   4.006551E-05_EB, &
   7.510632E-04_EB, 6.557116E-04_EB, 7.646716E-04_EB, 8.360125E-04_EB, 5.785516E-04_EB, 7.166127E-02_EB, &
   3.307520E-05_EB, &
   7.077134E-04_EB, 6.379859E-04_EB, 8.630039E-04_EB, 8.226252E-04_EB, 6.815535E-04_EB, 2.157192E-01_EB, &
   1.186508E-04_EB/),(/7,8/))

GAMMAD1_CH3OH(1:7,9:16) = RESHAPE((/ &  ! 865-900 CM-1
   6.878550E-04_EB, 7.626337E-04_EB, 7.231979E-04_EB, 6.559557E-04_EB, 6.188317E-04_EB, 1.528514E-01_EB, &
   3.379362E-05_EB, &
   8.329430E-04_EB, 7.234964E-04_EB, 7.086063E-04_EB, 4.799785E-04_EB, 5.996684E-04_EB, 1.058608E-01_EB, &
   2.576308E-05_EB, &
   7.201976E-04_EB, 4.879678E-04_EB, 7.514895E-04_EB, 9.296249E-04_EB, 7.627646E-04_EB, 1.357754E-02_EB, &
   4.112898E-05_EB, &
   4.468231E-04_EB, 7.152109E-04_EB, 5.892145E-04_EB, 6.268495E-03_EB, 6.890662E-04_EB, 8.344058E-03_EB, &
   3.782398E-05_EB, &
   4.115617E-04_EB, 6.608084E-04_EB, 6.730864E-04_EB, 1.653705E-02_EB, 7.764525E-04_EB, 2.258557E-04_EB, &
   7.122008E-05_EB, &
   2.243172E-03_EB, 1.292195E-03_EB, 1.270788E-03_EB, 1.990852E-02_EB, 5.353682E-02_EB, 3.379605E-04_EB, &
   1.094851E-04_EB, &
   2.795631E-03_EB, 8.919100E-04_EB, 5.452501E-03_EB, 5.460032E-02_EB, 4.954888E-01_EB, 2.416852E-03_EB, &
   8.087006E-05_EB, &
   3.021916E-03_EB, 8.815026E-04_EB, 2.682624E-03_EB, 1.363428E-01_EB, 7.040860E-01_EB, 9.532477E-03_EB, &
   4.720845E-05_EB/),(/7,8/))

GAMMAD1_CH3OH(1:7,17:24) = RESHAPE((/ &  ! 905-940 CM-1
   1.302330E-02_EB, 5.102234E-03_EB, 2.422490E-01_EB, 1.768788E-01_EB, 2.586447E-01_EB, 3.526812E-01_EB, &
   4.129581E-04_EB, &
   2.544888E-02_EB, 1.060450E-02_EB, 1.381583E-02_EB, 4.518627E-01_EB, 5.206079E-01_EB, 8.391062E-03_EB, &
   2.632629E-04_EB, &
   2.533298E-01_EB, 6.433475E-03_EB, 2.725983E-02_EB, 2.678852E-01_EB, 2.275531E-01_EB, 1.261497E-03_EB, &
   2.704299E-04_EB, &
   3.122114E-02_EB, 8.346136E-03_EB, 2.487987E-01_EB, 9.374202E-02_EB, 6.604262E-02_EB, 2.488402E-02_EB, &
   2.562359E-04_EB, &
   3.870602E-02_EB, 2.158011E-01_EB, 1.910186E-01_EB, 2.065914E-01_EB, 8.461225E-01_EB, 2.529266E-01_EB, &
   3.603466E-04_EB, &
   7.517862E-02_EB, 2.869129E-01_EB, 5.031265E-01_EB, 3.707385E-01_EB, 3.378211E-01_EB, 7.697419E-03_EB, &
   3.347542E-04_EB, &
   1.159008E-01_EB, 5.649415E-01_EB, 2.185566E-01_EB, 2.313598E+00_EB, 1.481866E+00_EB, 3.664479E-02_EB, &
   3.612492E-04_EB, &
   2.101148E-01_EB, 1.362737E+00_EB, 1.681817E-01_EB, 2.403762E+00_EB, 5.052362E+00_EB, 4.829425E+01_EB, &
   7.256697E-04_EB/),(/7,8/))

GAMMAD1_CH3OH(1:7,25:32) = RESHAPE((/ &  ! 945-980 CM-1
   2.154223E-01_EB, 1.063134E+00_EB, 2.745270E+00_EB, 6.048114E+00_EB, 1.191546E+01_EB, 3.613074E+01_EB, &
   9.758013E-04_EB, &
   3.415582E-01_EB, 5.670554E+00_EB, 1.626968E+00_EB, 8.377516E+00_EB, 2.646977E+01_EB, 4.827207E+01_EB, &
   6.374449E-04_EB, &
   7.233287E-01_EB, 1.170538E+01_EB, 2.699023E+00_EB, 1.811784E+01_EB, 4.833090E+01_EB, 4.829379E+01_EB, &
   6.697140E-04_EB, &
   4.410713E+00_EB, 2.163944E+01_EB, 1.204215E+01_EB, 6.230210E+01_EB, 5.735624E+01_EB, 4.829428E+01_EB, &
   6.621854E-04_EB, &
   6.410135E+00_EB, 5.657130E+01_EB, 3.102339E+01_EB, 6.230874E+01_EB, 5.735677E+01_EB, 4.829401E+01_EB, &
   1.509172E-03_EB, &
   3.756246E+01_EB, 6.880041E+01_EB, 4.129381E+01_EB, 6.230370E+01_EB, 5.735689E+01_EB, 4.829405E+01_EB, &
   8.487264E-04_EB, &
   6.111221E+01_EB, 6.881385E+01_EB, 6.471263E+01_EB, 6.230875E+01_EB, 5.735697E+01_EB, 4.829401E+01_EB, &
   1.318189E-03_EB, &
   7.999989E+01_EB, 6.881377E+01_EB, 6.495080E+01_EB, 6.230820E+01_EB, 5.735697E+01_EB, 4.828988E+01_EB, &
   1.354435E-03_EB/),(/7,8/))

GAMMAD1_CH3OH(1:7,33:40) = RESHAPE((/ &  ! 985-1020 CM-1
   7.999998E+01_EB, 6.881384E+01_EB, 6.494110E+01_EB, 6.230890E+01_EB, 5.735615E+01_EB, 4.829425E+01_EB, &
   1.479087E-03_EB, &
   7.999920E+01_EB, 6.881390E+01_EB, 6.506107E+01_EB, 6.230863E+01_EB, 5.735699E+01_EB, 4.829154E+01_EB, &
   1.456376E-03_EB, &
   7.999917E+01_EB, 6.881386E+01_EB, 6.499928E+01_EB, 6.230892E+01_EB, 5.735688E+01_EB, 1.272502E+01_EB, &
   7.723150E-04_EB, &
   7.999980E+01_EB, 6.881301E+01_EB, 6.505184E+01_EB, 6.230892E+01_EB, 5.735698E+01_EB, 4.828738E+01_EB, &
   6.758482E-04_EB, &
   8.000000E+01_EB, 6.881338E+01_EB, 6.504268E+01_EB, 6.230892E+01_EB, 5.735686E+01_EB, 4.829415E+01_EB, &
   7.919076E-04_EB, &
   8.000000E+01_EB, 6.881339E+01_EB, 6.506098E+01_EB, 6.230892E+01_EB, 5.735694E+01_EB, 4.829379E+01_EB, &
   1.044482E-03_EB, &
   8.000000E+01_EB, 6.881323E+01_EB, 6.503974E+01_EB, 6.230732E+01_EB, 5.735696E+01_EB, 4.829424E+01_EB, &
   1.883098E-03_EB, &
   8.000000E+01_EB, 6.881386E+01_EB, 6.498651E+01_EB, 6.230722E+01_EB, 5.735624E+01_EB, 2.615034E+00_EB, &
   1.731620E-03_EB/),(/7,8/))

GAMMAD1_CH3OH(1:7,41:48) = RESHAPE((/ &  ! 1025-1060 CM-1
   7.999918E+01_EB, 6.881390E+01_EB, 6.501143E+01_EB, 6.230770E+01_EB, 5.735573E+01_EB, 4.828762E+01_EB, &
   1.844304E-03_EB, &
   7.999986E+01_EB, 6.881374E+01_EB, 2.509078E+00_EB, 6.230892E+01_EB, 5.735699E+01_EB, 4.829418E+01_EB, &
   2.140274E-03_EB, &
   5.272622E-01_EB, 6.881390E+01_EB, 7.822208E-01_EB, 6.230892E+01_EB, 5.735699E+01_EB, 4.829426E+01_EB, &
   1.608061E-03_EB, &
   7.999988E+01_EB, 6.881380E+01_EB, 6.416330E+01_EB, 6.230736E+01_EB, 5.735695E+01_EB, 4.829395E+01_EB, &
   1.715694E-03_EB, &
   8.000000E+01_EB, 6.881387E+01_EB, 6.501663E+01_EB, 6.230833E+01_EB, 5.735699E+01_EB, 4.829374E+01_EB, &
   1.856662E-03_EB, &
   8.000000E+01_EB, 6.881386E+01_EB, 6.506116E+01_EB, 6.230710E+01_EB, 5.735699E+01_EB, 4.829358E+01_EB, &
   2.738523E-03_EB, &
   8.000000E+01_EB, 6.881391E+01_EB, 6.506116E+01_EB, 6.230892E+01_EB, 5.735699E+01_EB, 4.829418E+01_EB, &
   1.589317E-03_EB, &
   7.999991E+01_EB, 6.881391E+01_EB, 6.506118E+01_EB, 6.230892E+01_EB, 5.735618E+01_EB, 4.829397E+01_EB, &
   1.507561E-03_EB/),(/7,8/))

GAMMAD1_CH3OH(1:7,49:56) = RESHAPE((/ &  ! 1065-1100 CM-1
   8.000000E+01_EB, 6.881387E+01_EB, 6.506116E+01_EB, 6.230827E+01_EB, 5.735699E+01_EB, 4.829427E+01_EB, &
   2.253348E-03_EB, &
   7.999938E+01_EB, 6.881318E+01_EB, 6.505314E+01_EB, 6.230887E+01_EB, 5.735699E+01_EB, 4.829173E+01_EB, &
   3.424844E-03_EB, &
   7.999997E+01_EB, 6.881353E+01_EB, 6.505405E+01_EB, 6.230890E+01_EB, 5.735697E+01_EB, 4.828966E+01_EB, &
   1.125303E-03_EB, &
   6.451623E+01_EB, 6.881383E+01_EB, 6.492757E+01_EB, 6.230886E+01_EB, 5.735691E+01_EB, 4.823782E+01_EB, &
   9.265505E-04_EB, &
   2.835561E+00_EB, 4.814647E+01_EB, 1.930069E+01_EB, 4.670295E+01_EB, 5.733661E+01_EB, 4.981992E-01_EB, &
   1.101043E-03_EB, &
   3.405886E-01_EB, 8.624550E+00_EB, 8.265512E-01_EB, 4.097800E+00_EB, 1.789592E+01_EB, 1.840689E-02_EB, &
   1.204450E-03_EB, &
   2.136135E+00_EB, 6.560155E+00_EB, 2.276106E-01_EB, 2.826746E-01_EB, 4.031936E+00_EB, 9.675596E-03_EB, &
   8.693678E-04_EB, &
   3.865526E-01_EB, 2.663625E+00_EB, 2.801253E-01_EB, 1.479540E-01_EB, 6.549643E-01_EB, 4.007539E-03_EB, &
   2.400175E-04_EB/),(/7,8/))

GAMMAD1_CH3OH(1:7,57:58) = RESHAPE((/ &  ! 1125-1150 CM-1
   1.031691E-01_EB, 1.063157E+00_EB, 8.796742E-02_EB, 3.121383E-01_EB, 3.923037E-01_EB, 4.063955E-04_EB, &
   1.096427E-04_EB, &
   3.748583E-04_EB, 3.747168E-04_EB, 3.744971E-04_EB, 3.751403E-04_EB, 3.748851E-04_EB, 3.734005E-04_EB, &
   3.740585E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_CH3OH(N_TEMP_CH3OH,24))

! BAND #2: 1125 CM-1 - 1700 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 1.5713 %

SD2_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 1125-1300 CM-1
   8.927831E-02_EB, 6.438189E-02_EB, 6.336484E-02_EB, 7.782879E-02_EB, 7.726760E-02_EB, 3.430637E-01_EB, &
   2.111028E+00_EB, &
   1.428237E-01_EB, 1.080083E-01_EB, 1.208600E-01_EB, 1.421780E-01_EB, 1.436580E-01_EB, 1.355584E+00_EB, &
   5.627932E+00_EB, &
   1.431595E-01_EB, 1.414695E-01_EB, 1.558110E-01_EB, 1.760891E-01_EB, 2.027903E-01_EB, 5.295318E-01_EB, &
   8.429621E+00_EB, &
   2.293295E-01_EB, 2.455796E-01_EB, 2.542758E-01_EB, 2.648440E-01_EB, 2.899987E-01_EB, 4.124789E-01_EB, &
   7.705341E+00_EB, &
   2.983198E-01_EB, 2.968924E-01_EB, 2.931528E-01_EB, 3.015322E-01_EB, 3.289150E-01_EB, 4.332046E-01_EB, &
   1.274789E+01_EB, &
   3.760820E-01_EB, 3.648052E-01_EB, 3.591136E-01_EB, 3.639314E-01_EB, 3.850232E-01_EB, 5.091509E-01_EB, &
   9.900445E+00_EB, &
   4.286786E-01_EB, 4.103485E-01_EB, 4.009699E-01_EB, 4.152912E-01_EB, 4.397207E-01_EB, 4.994214E-01_EB, &
   1.080146E+01_EB, &
   6.526914E-01_EB, 4.766714E-01_EB, 4.276416E-01_EB, 4.115988E-01_EB, 4.189427E-01_EB, 4.846849E-01_EB, &
   8.381233E+00_EB/),(/7,8/))

SD2_CH3OH(1:7,9:16) = RESHAPE((/ &  ! 1325-1500 CM-1
   8.926014E-01_EB, 5.725096E-01_EB, 4.936350E-01_EB, 4.461230E-01_EB, 4.220260E-01_EB, 4.240504E-01_EB, &
   6.735177E+00_EB, &
   9.802829E-01_EB, 6.469535E-01_EB, 5.500127E-01_EB, 4.937137E-01_EB, 4.581570E-01_EB, 4.189329E-01_EB, &
   5.750931E+00_EB, &
   7.911468E-01_EB, 5.793679E-01_EB, 5.044808E-01_EB, 4.592149E-01_EB, 4.299574E-01_EB, 3.354935E-01_EB, &
   5.762184E+00_EB, &
   6.069133E-01_EB, 4.730774E-01_EB, 4.244640E-01_EB, 3.791906E-01_EB, 3.723725E-01_EB, 3.634556E-01_EB, &
   5.212994E+00_EB, &
   5.069178E-01_EB, 3.597810E-01_EB, 3.086202E-01_EB, 2.782571E-01_EB, 2.798365E-01_EB, 3.419783E-01_EB, &
   5.269252E+00_EB, &
   4.541349E-01_EB, 3.336580E-01_EB, 2.970106E-01_EB, 2.657370E-01_EB, 2.766112E-01_EB, 3.361569E-01_EB, &
   5.737162E+00_EB, &
   4.615691E-01_EB, 3.258696E-01_EB, 2.757947E-01_EB, 2.454101E-01_EB, 2.523896E-01_EB, 2.077158E-01_EB, &
   3.866394E+00_EB, &
   3.328678E-01_EB, 2.037433E-01_EB, 1.740683E-01_EB, 1.606349E-01_EB, 1.789734E-01_EB, 3.064642E-01_EB, &
   2.819085E+00_EB/),(/7,8/))

SD2_CH3OH(1:7,17:24) = RESHAPE((/ &  ! 1525-1700 CM-1
   2.653588E-01_EB, 1.712954E-01_EB, 1.458324E-01_EB, 1.336402E-01_EB, 1.497627E-01_EB, 6.752522E-01_EB, &
   1.211105E+00_EB, &
   1.719421E-01_EB, 9.991170E-02_EB, 8.707576E-02_EB, 9.851840E-02_EB, 1.045626E-01_EB, 1.257519E-01_EB, &
   1.729420E-01_EB, &
   9.561175E-02_EB, 1.214100E-01_EB, 1.009499E-01_EB, 8.557814E-02_EB, 1.061461E-01_EB, 5.279653E-02_EB, &
   1.371839E+00_EB, &
   5.679817E-02_EB, 1.064996E-01_EB, 8.829344E-02_EB, 5.535303E-02_EB, 8.540729E-02_EB, 6.869857E-02_EB, &
   2.390774E+00_EB, &
   2.299715E-02_EB, 5.668754E-02_EB, 4.611447E-02_EB, 2.942043E-02_EB, 5.283754E-02_EB, 1.368461E-02_EB, &
   8.704598E-02_EB, &
   9.994737E-04_EB, 6.379671E-03_EB, 5.806672E-03_EB, 3.377512E-03_EB, 2.456039E-02_EB, 4.999823E-04_EB, &
   1.518390E-02_EB, &
   6.504118E-04_EB, 6.258215E-04_EB, 6.777009E-04_EB, 5.831397E-04_EB, 1.258943E-02_EB, 6.613745E-04_EB, &
   3.926253E-03_EB, &
   4.699306E-04_EB, 4.696984E-04_EB, 4.697130E-04_EB, 4.702606E-04_EB, 4.699096E-04_EB, 4.682572E-04_EB, &
   4.691749E-04_EB/),(/7,8/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_CH3OH(N_TEMP_CH3OH,24))

! BAND #2: 1125 CM-1 - 1700 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 1.5713 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 1125-1300 CM-1
   2.036126E-02_EB, 2.341570E-01_EB, 2.115157E-01_EB, 3.745991E-02_EB, 2.969993E-01_EB, 6.950924E-04_EB, &
   1.090294E-04_EB, &
   4.318091E-02_EB, 3.213276E+00_EB, 1.818315E-01_EB, 6.478423E-02_EB, 7.314703E-01_EB, 8.584914E-04_EB, &
   1.822575E-04_EB, &
   1.945666E-02_EB, 1.155771E+00_EB, 8.508445E-01_EB, 1.670850E-01_EB, 2.280568E+00_EB, 4.205261E-03_EB, &
   3.102895E-04_EB, &
   7.511583E-01_EB, 6.453461E+00_EB, 3.392847E+00_EB, 1.506522E+00_EB, 5.522265E+00_EB, 1.714255E-02_EB, &
   3.985647E-04_EB, &
   1.205702E+00_EB, 9.217844E+00_EB, 2.021655E+00_EB, 2.387126E+00_EB, 4.807045E+00_EB, 2.526665E-02_EB, &
   3.559097E-04_EB, &
   2.746003E+00_EB, 9.503086E+00_EB, 2.328040E+00_EB, 4.485330E+00_EB, 9.413318E+00_EB, 1.788133E-02_EB, &
   4.729684E-04_EB, &
   7.531540E+00_EB, 1.737151E+01_EB, 3.326148E+00_EB, 8.198222E+00_EB, 1.532668E+01_EB, 4.522148E-02_EB, &
   4.549597E-04_EB, &
   2.576303E+01_EB, 2.203775E+01_EB, 4.437035E+00_EB, 9.898502E+00_EB, 1.872281E+01_EB, 3.906498E-02_EB, &
   4.821874E-04_EB/),(/7,8/))

GAMMAD2_CH3OH(1:7,9:16) = RESHAPE((/ &  ! 1325-1500 CM-1
   6.250399E+01_EB, 3.711813E+01_EB, 4.149099E+00_EB, 1.502754E+01_EB, 1.389076E+01_EB, 4.031216E-02_EB, &
   5.674173E-04_EB, &
   7.999962E+01_EB, 5.252798E+01_EB, 1.230226E+01_EB, 1.351493E+01_EB, 1.576973E+01_EB, 3.751308E-02_EB, &
   4.823587E-04_EB, &
   6.092331E+01_EB, 3.805490E+01_EB, 1.271376E+01_EB, 1.115539E+01_EB, 1.762731E+01_EB, 7.120929E-02_EB, &
   2.295392E-04_EB, &
   3.106485E+01_EB, 2.102259E+01_EB, 4.586676E+00_EB, 6.831440E+00_EB, 9.421920E+00_EB, 1.574921E-02_EB, &
   2.188143E-04_EB, &
   1.678441E+01_EB, 9.704951E+00_EB, 2.972980E+00_EB, 2.004021E+00_EB, 5.086223E+00_EB, 6.963213E-03_EB, &
   1.950359E-04_EB, &
   9.295170E+00_EB, 9.919157E+00_EB, 2.792442E+00_EB, 3.932078E+00_EB, 5.121134E+00_EB, 9.304463E-03_EB, &
   2.472872E-04_EB, &
   8.180857E+00_EB, 9.293708E+00_EB, 2.066943E+00_EB, 4.076114E+00_EB, 4.294751E+00_EB, 2.727838E-02_EB, &
   4.265301E-04_EB, &
   8.289335E-02_EB, 2.291116E+00_EB, 8.466376E-01_EB, 1.801527E-01_EB, 1.505300E+00_EB, 1.548374E-03_EB, &
   2.852002E-04_EB/),(/7,8/))

GAMMAD2_CH3OH(1:7,17:24) = RESHAPE((/ &  ! 1525-1700 CM-1
   7.067783E-02_EB, 1.479749E+00_EB, 5.187742E-01_EB, 1.372559E-01_EB, 7.400371E-01_EB, 2.949015E-04_EB, &
   4.953459E-04_EB, &
   1.628627E-02_EB, 8.476648E-01_EB, 2.184428E-01_EB, 1.854959E-02_EB, 1.335169E-01_EB, 2.574359E-04_EB, &
   6.601786E-04_EB, &
   1.778534E-02_EB, 2.130514E-01_EB, 1.040976E-01_EB, 1.031289E-01_EB, 1.600152E+00_EB, 2.537685E-02_EB, &
   1.933608E-05_EB, &
   6.344549E-03_EB, 1.274538E-01_EB, 5.382005E-02_EB, 6.261896E-01_EB, 4.338573E-01_EB, 3.757499E-01_EB, &
   6.947226E-05_EB, &
   2.030029E-04_EB, 6.186341E-02_EB, 2.709645E-02_EB, 5.640946E-02_EB, 3.663756E-01_EB, 2.286239E-02_EB, &
   1.077139E-04_EB, &
   8.095102E-04_EB, 6.039758E-03_EB, 3.546569E-03_EB, 3.077717E-03_EB, 3.777927E-01_EB, 4.051839E-04_EB, &
   8.507547E-03_EB, &
   4.920711E-04_EB, 4.753092E-04_EB, 5.221505E-04_EB, 4.531404E-04_EB, 1.775124E-01_EB, 5.028014E-04_EB, &
   4.011702E-03_EB, &
   3.748583E-04_EB, 3.747168E-04_EB, 3.744971E-04_EB, 3.751403E-04_EB, 3.748851E-04_EB, 3.734005E-04_EB, &
   3.740585E-04_EB/),(/7,8/))

!---------------------------------------------------------------------------
ALLOCATE(SD3_CH3OH(N_TEMP_CH3OH,26))

! BAND #3: 2600 CM-1 - 3225 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.0044 %

SD3_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 2600-2775 CM-1
   6.834504E-04_EB, 6.520912E-04_EB, 6.293446E-04_EB, 7.602714E-04_EB, 6.782192E-04_EB, 5.546971E-04_EB, &
   4.991362E-04_EB, &
   9.382971E-04_EB, 9.146731E-04_EB, 9.400615E-04_EB, 9.494733E-04_EB, 9.709058E-04_EB, 8.544557E-04_EB, &
   1.012754E-01_EB, &
   8.700561E-04_EB, 8.964927E-04_EB, 9.462079E-04_EB, 7.740041E-04_EB, 9.999124E-04_EB, 3.951283E-03_EB, &
   5.295951E-02_EB, &
   9.087410E-04_EB, 8.586469E-04_EB, 1.061772E-03_EB, 9.199452E-04_EB, 7.793155E-04_EB, 2.101256E-02_EB, &
   1.014346E-01_EB, &
   9.791713E-04_EB, 7.304428E-04_EB, 3.244007E-03_EB, 2.026606E-03_EB, 4.855695E-03_EB, 6.777374E-02_EB, &
   6.074970E-01_EB, &
   9.694186E-03_EB, 8.820685E-03_EB, 2.014998E-02_EB, 1.559350E-02_EB, 2.203169E-02_EB, 1.271102E-01_EB, &
   1.115356E+00_EB, &
   7.029245E-02_EB, 5.091559E-02_EB, 6.271267E-02_EB, 6.250553E-02_EB, 7.985290E-02_EB, 2.008186E-01_EB, &
   7.823174E-01_EB, &
   1.832465E-01_EB, 1.685849E-01_EB, 1.767557E-01_EB, 1.929714E-01_EB, 2.182618E-01_EB, 3.370225E-01_EB, &
   9.170478E-01_EB/),(/7,8/))

SD3_CH3OH(1:7,9:16) = RESHAPE((/ &  ! 2800-2975 CM-1
   6.431025E-01_EB, 5.822856E-01_EB, 5.579032E-01_EB, 5.709313E-01_EB, 5.423547E-01_EB, 5.399764E-01_EB, &
   1.804940E+00_EB, &
   1.616617E+00_EB, 1.106968E+00_EB, 9.574460E-01_EB, 9.001676E-01_EB, 7.635145E-01_EB, 6.273449E-01_EB, &
   2.500203E+00_EB, &
   2.054107E+00_EB, 1.363124E+00_EB, 1.172320E+00_EB, 1.104390E+00_EB, 9.352928E-01_EB, 7.059677E-01_EB, &
   1.705210E+00_EB, &
   2.373043E+00_EB, 1.746858E+00_EB, 1.529178E+00_EB, 1.473170E+00_EB, 1.254399E+00_EB, 9.322847E-01_EB, &
   2.050430E+00_EB, &
   2.371351E+00_EB, 1.769945E+00_EB, 1.583483E+00_EB, 1.552549E+00_EB, 1.370339E+00_EB, 1.037337E+00_EB, &
   2.789724E+00_EB, &
   3.792115E+00_EB, 2.516241E+00_EB, 2.137945E+00_EB, 2.022427E+00_EB, 1.666670E+00_EB, 1.107211E+00_EB, &
   3.100678E+00_EB, &
   4.157174E+00_EB, 2.644308E+00_EB, 2.220834E+00_EB, 2.075201E+00_EB, 1.691865E+00_EB, 1.086754E+00_EB, &
   2.915371E+00_EB, &
   4.342764E+00_EB, 2.822870E+00_EB, 2.375959E+00_EB, 2.221659E+00_EB, 1.802922E+00_EB, 1.138893E+00_EB, &
   2.903207E+00_EB/),(/7,8/))

SD3_CH3OH(1:7,17:24) = RESHAPE((/ &  ! 3000-3175 CM-1
   2.739997E+00_EB, 1.931251E+00_EB, 1.703000E+00_EB, 1.624103E+00_EB, 1.381073E+00_EB, 9.258508E-01_EB, &
   2.377676E+00_EB, &
   1.657562E+00_EB, 1.227931E+00_EB, 1.089977E+00_EB, 1.050288E+00_EB, 9.156831E-01_EB, 6.349324E-01_EB, &
   1.478204E+00_EB, &
   7.792319E-01_EB, 6.776318E-01_EB, 6.362424E-01_EB, 6.325745E-01_EB, 5.728339E-01_EB, 4.149136E-01_EB, &
   8.597464E-01_EB, &
   2.993662E-01_EB, 3.092377E-01_EB, 3.217485E-01_EB, 3.267839E-01_EB, 3.261281E-01_EB, 2.499769E-01_EB, &
   5.435448E-01_EB, &
   7.556116E-02_EB, 1.013165E-01_EB, 1.286472E-01_EB, 1.311502E-01_EB, 1.509531E-01_EB, 1.344009E-01_EB, &
   4.354789E-01_EB, &
   2.499115E-02_EB, 1.505850E-02_EB, 4.570610E-02_EB, 3.921605E-02_EB, 5.480372E-02_EB, 5.712297E-02_EB, &
   9.435909E-01_EB, &
   4.505936E-03_EB, 1.692846E-03_EB, 7.624509E-03_EB, 2.230769E-03_EB, 1.179331E-02_EB, 1.334128E-02_EB, &
   4.731345E-01_EB, &
   8.990606E-04_EB, 7.925739E-04_EB, 1.307211E-03_EB, 8.823890E-04_EB, 3.090415E-03_EB, 1.825819E-03_EB, &
   5.903177E-03_EB/),(/7,8/))

SD3_CH3OH(1:7,25:26) = RESHAPE((/ &  ! 3200-3225 CM-1
   6.930486E-04_EB, 6.167383E-04_EB, 6.581892E-04_EB, 7.470153E-04_EB, 6.327299E-04_EB, 5.451720E-04_EB, &
   6.008264E-04_EB, &
   4.699306E-04_EB, 4.696984E-04_EB, 4.697130E-04_EB, 4.702606E-04_EB, 4.699096E-04_EB, 4.682572E-04_EB, &
   4.691749E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD3_CH3OH(N_TEMP_CH3OH,26))

! BAND #3: 2600 CM-1 - 3225 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.0044 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD3_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 2600-2775 CM-1
   5.383708E-04_EB, 5.150011E-04_EB, 4.711666E-04_EB, 5.877341E-04_EB, 5.325580E-04_EB, 4.526624E-04_EB, &
   3.953361E-04_EB, &
   6.469173E-04_EB, 6.295204E-04_EB, 6.229213E-04_EB, 6.140709E-04_EB, 7.108644E-04_EB, 5.982461E-04_EB, &
   3.920520E-04_EB, &
   6.148513E-04_EB, 6.637986E-04_EB, 6.390155E-04_EB, 5.400884E-04_EB, 6.784448E-04_EB, 3.991532E-03_EB, &
   4.998359E-03_EB, &
   6.548999E-04_EB, 6.227472E-04_EB, 7.730036E-04_EB, 6.279630E-04_EB, 5.750964E-04_EB, 8.889816E-02_EB, &
   4.610782E-03_EB, &
   7.658461E-04_EB, 5.677730E-04_EB, 3.744908E-03_EB, 1.790178E-03_EB, 4.848659E-03_EB, 5.816371E-02_EB, &
   7.773551E-04_EB, &
   9.155583E-03_EB, 4.468866E-02_EB, 7.356051E-02_EB, 9.249021E-03_EB, 4.330986E-02_EB, 7.759327E-01_EB, &
   1.031984E-03_EB, &
   3.240618E-01_EB, 4.813829E-01_EB, 2.376949E-01_EB, 9.693334E-02_EB, 1.522188E-01_EB, 1.138999E+00_EB, &
   2.244386E-03_EB, &
   1.958993E+00_EB, 1.883288E+00_EB, 9.427313E-01_EB, 2.032231E+00_EB, 2.865213E+00_EB, 8.627391E+00_EB, &
   2.968586E-03_EB/),(/7,8/))

GAMMAD3_CH3OH(1:7,9:16) = RESHAPE((/ &  ! 2800-2975 CM-1
   5.408634E+01_EB, 4.260803E+01_EB, 1.570059E+01_EB, 2.520800E+01_EB, 3.812191E+01_EB, 3.643507E+01_EB, &
   1.949224E-03_EB, &
   8.000000E+01_EB, 6.880853E+01_EB, 4.378312E+01_EB, 6.230866E+01_EB, 5.732026E+01_EB, 3.975357E+01_EB, &
   1.722003E-03_EB, &
   7.999962E+01_EB, 6.881387E+01_EB, 6.160988E+01_EB, 6.230878E+01_EB, 5.734000E+01_EB, 4.829424E+01_EB, &
   3.362841E-03_EB, &
   7.999999E+01_EB, 6.881386E+01_EB, 6.505766E+01_EB, 6.230889E+01_EB, 5.735690E+01_EB, 4.829419E+01_EB, &
   3.574523E-03_EB, &
   7.999997E+01_EB, 6.881383E+01_EB, 6.505113E+01_EB, 6.230881E+01_EB, 5.735699E+01_EB, 4.829428E+01_EB, &
   2.799128E-03_EB, &
   7.999848E+01_EB, 6.881323E+01_EB, 6.505412E+01_EB, 6.230885E+01_EB, 5.735695E+01_EB, 4.829421E+01_EB, &
   2.167350E-03_EB, &
   7.999924E+01_EB, 6.881280E+01_EB, 6.505343E+01_EB, 6.230882E+01_EB, 5.735689E+01_EB, 4.829429E+01_EB, &
   2.206560E-03_EB, &
   7.999955E+01_EB, 6.881390E+01_EB, 6.505410E+01_EB, 6.230862E+01_EB, 5.735674E+01_EB, 4.829421E+01_EB, &
   2.226025E-03_EB/),(/7,8/))

GAMMAD3_CH3OH(1:7,17:24) = RESHAPE((/ &  ! 3000-3175 CM-1
   8.000000E+01_EB, 6.881380E+01_EB, 6.505082E+01_EB, 6.230881E+01_EB, 5.735619E+01_EB, 4.829420E+01_EB, &
   2.316390E-03_EB, &
   7.999999E+01_EB, 6.881373E+01_EB, 4.658312E+01_EB, 6.230870E+01_EB, 5.735696E+01_EB, 4.160676E+01_EB, &
   1.752164E-03_EB, &
   7.996301E+01_EB, 6.881042E+01_EB, 7.394547E+00_EB, 3.799419E+01_EB, 2.909775E+01_EB, 1.888125E+01_EB, &
   1.732017E-03_EB, &
   4.261947E+00_EB, 5.499787E+00_EB, 4.853050E-01_EB, 6.858640E+00_EB, 8.915265E+00_EB, 4.488725E+00_EB, &
   1.039419E-03_EB, &
   3.182069E-01_EB, 2.435677E-01_EB, 7.167320E-02_EB, 3.943245E-01_EB, 1.078816E+00_EB, 5.812817E-01_EB, &
   5.867531E-04_EB, &
   4.059798E-01_EB, 7.132746E-03_EB, 2.110640E-02_EB, 1.885784E-02_EB, 1.959525E+00_EB, 1.180753E-01_EB, &
   1.052399E-04_EB, &
   6.304409E-03_EB, 7.479753E-04_EB, 2.811545E-03_EB, 1.771085E-03_EB, 3.653146E-01_EB, 6.263784E-03_EB, &
   5.856904E-05_EB, &
   6.352137E-04_EB, 5.913538E-04_EB, 7.804488E-04_EB, 6.507595E-04_EB, 5.533275E-03_EB, 5.897126E-04_EB, &
   1.359107E-03_EB/),(/7,8/))

GAMMAD3_CH3OH(1:7,25:26) = RESHAPE((/ &  ! 3200-3225 CM-1
   5.327466E-04_EB, 4.880598E-04_EB, 4.928840E-04_EB, 6.016314E-04_EB, 5.375037E-04_EB, 4.225641E-04_EB, &
   4.726310E-04_EB, &
   3.748583E-04_EB, 3.747168E-04_EB, 3.744971E-04_EB, 3.751403E-04_EB, 3.748851E-04_EB, 3.734005E-04_EB, &
   3.740585E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(SD4_CH3OH(N_TEMP_CH3OH,14))

! BAND #4: 3525 CM-1 - 3850 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.52389 %

SD4_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 3525-3700 CM-1
   6.712614E-04_EB, 6.207569E-04_EB, 6.722895E-04_EB, 6.499733E-04_EB, 6.502738E-04_EB, 6.875865E-04_EB, &
   5.963773E-04_EB, &
   8.659465E-04_EB, 7.116916E-03_EB, 2.432786E-03_EB, 5.672617E-03_EB, 1.138304E-03_EB, 8.423409E-04_EB, &
   3.290015E-02_EB, &
   1.352414E-02_EB, 3.203665E-02_EB, 2.755663E-02_EB, 2.769632E-02_EB, 2.312063E-02_EB, 5.525230E-02_EB, &
   2.986065E-02_EB, &
   4.836204E-02_EB, 5.823870E-02_EB, 5.880972E-02_EB, 6.018123E-02_EB, 6.237734E-02_EB, 6.647353E-02_EB, &
   7.540357E-02_EB, &
   2.393220E-01_EB, 2.245383E-01_EB, 2.030768E-01_EB, 2.229738E-01_EB, 2.012059E-01_EB, 1.546924E-01_EB, &
   9.432293E-02_EB, &
   8.391763E-01_EB, 5.928915E-01_EB, 4.976034E-01_EB, 4.745560E-01_EB, 3.784750E-01_EB, 2.223088E-01_EB, &
   1.157200E-01_EB, &
   1.156927E+00_EB, 6.735305E-01_EB, 5.357276E-01_EB, 4.781313E-01_EB, 3.584284E-01_EB, 1.968791E-01_EB, &
   6.758499E-02_EB, &
   1.214378E+00_EB, 7.022996E-01_EB, 5.366277E-01_EB, 4.875065E-01_EB, 3.656436E-01_EB, 2.096662E-01_EB, &
   6.457647E-01_EB/),(/7,8/))

SD4_CH3OH(1:7,9:14) = RESHAPE((/ &  ! 3725-3850 CM-1
   7.394573E-01_EB, 5.605469E-01_EB, 4.645577E-01_EB, 4.379377E-01_EB, 3.556244E-01_EB, 2.353483E-01_EB, &
   1.093933E+00_EB, &
   1.627441E-01_EB, 1.282880E-01_EB, 1.200820E-01_EB, 1.269546E-01_EB, 1.078069E-01_EB, 1.197911E-01_EB, &
   5.594711E-02_EB, &
   5.446422E-02_EB, 6.975799E-02_EB, 5.570485E-02_EB, 2.768076E-02_EB, 2.935100E-02_EB, 4.703585E-02_EB, &
   5.054507E-01_EB, &
   5.503798E-03_EB, 3.001374E-02_EB, 1.548540E-02_EB, 1.507541E-03_EB, 3.052107E-03_EB, 1.675834E-02_EB, &
   3.138134E-02_EB, &
   6.972876E-04_EB, 3.117091E-03_EB, 2.511646E-03_EB, 6.126449E-04_EB, 6.508959E-04_EB, 6.472961E-04_EB, &
   6.127914E-04_EB, &
   4.699306E-04_EB, 4.696984E-04_EB, 4.697130E-04_EB, 4.702606E-04_EB, 4.699096E-04_EB, 4.682572E-04_EB, &
   4.691749E-04_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD4_CH3OH(N_TEMP_CH3OH,14))

! BAND #4: 3525 CM-1 - 3850 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.52389 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD4_CH3OH(1:7,1:8) = RESHAPE((/ &  ! 3525-3700 CM-1
   5.122880E-04_EB, 4.870888E-04_EB, 5.202731E-04_EB, 4.858325E-04_EB, 4.894141E-04_EB, 5.234560E-04_EB, &
   4.739454E-04_EB, &
   6.149338E-04_EB, 1.383613E-01_EB, 1.894640E-03_EB, 1.652906E-01_EB, 1.023597E-03_EB, 6.016220E-04_EB, &
   4.058911E-03_EB, &
   3.510951E-01_EB, 4.231322E-02_EB, 2.094183E-02_EB, 1.756637E-01_EB, 8.897080E-02_EB, 1.613316E-04_EB, &
   2.603264E-03_EB, &
   4.348848E-01_EB, 1.997004E+00_EB, 9.906747E-02_EB, 6.264650E-02_EB, 3.368153E-02_EB, 5.344243E-03_EB, &
   1.024940E-02_EB, &
   6.050885E+00_EB, 5.037136E+00_EB, 1.610741E+00_EB, 9.196821E-01_EB, 3.334769E+00_EB, 1.286001E-01_EB, &
   1.852205E-02_EB, &
   7.999493E+01_EB, 4.970642E+01_EB, 1.283805E+01_EB, 1.192477E+01_EB, 8.469410E+00_EB, 7.793742E-01_EB, &
   2.887610E-03_EB, &
   7.999963E+01_EB, 6.880115E+01_EB, 1.309018E+01_EB, 1.179133E+01_EB, 1.455179E+01_EB, 1.751987E-01_EB, &
   5.542345E-03_EB, &
   7.999935E+01_EB, 6.881382E+01_EB, 1.685368E+01_EB, 1.646915E+01_EB, 1.087273E+01_EB, 1.258692E+00_EB, &
   2.305314E-04_EB/),(/7,8/))

GAMMAD4_CH3OH(1:7,9:14) = RESHAPE((/ &  ! 3725-3850 CM-1
   6.925146E+01_EB, 3.835584E+01_EB, 7.883505E+00_EB, 1.488897E+01_EB, 1.369263E+01_EB, 1.664808E+00_EB, &
   3.295444E-04_EB, &
   5.002891E-01_EB, 7.416100E-01_EB, 2.838529E-01_EB, 3.020566E-01_EB, 1.523043E+00_EB, 1.133794E-03_EB, &
   4.486303E-03_EB, &
   6.823303E-03_EB, 4.678797E-01_EB, 9.198815E-02_EB, 1.770170E-01_EB, 1.483378E-01_EB, 1.869665E-01_EB, &
   2.436489E-05_EB, &
   4.419159E-03_EB, 3.220528E-01_EB, 2.310260E-01_EB, 1.151947E-03_EB, 2.324707E-03_EB, 6.947554E-03_EB, &
   1.285540E-04_EB, &
   5.281954E-04_EB, 2.447127E-03_EB, 2.223998E-03_EB, 4.730215E-04_EB, 5.179966E-04_EB, 4.804723E-04_EB, &
   4.227659E-04_EB, &
   3.748583E-04_EB, 3.747168E-04_EB, 3.744971E-04_EB, 3.751403E-04_EB, 3.748851E-04_EB, 3.734005E-04_EB, &
   3.740585E-04_EB/),(/7,6/))

!-------------------------MMA DATA-------------------


! THERE ARE 6 BANDS FOR MMA

! BAND #1: 750 CM-1 - 880 CM-1
! BAND #2: 875 CM-1 - 1055 CM-1
! BAND #3: 1050 CM-1 - 1275 CM-1
! BAND #4: 1250 CM-1 - 1575 CM-1
! BAND #5: 1550 CM-1 - 1975 CM-1
! BAND #6: 2650 CM-1 - 3275 CM-1

ALLOCATE(SD_C5H8O2_TEMP(N_TEMP_C5H8O2))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR MMA ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_C5H8O2(N_BAND_C5H8O2,3))
ALLOCATE(BE_C5H8O2(N_BAND_C5H8O2))

OM_BND_C5H8O2 = RESHAPE((/ &
   750._EB,  875._EB, 1050._EB, 1250._EB, 1550._EB, 2650._EB, &
   880._EB, 1055._EB, 1275._EB, 1575._EB, 1975._EB, 3275._EB, &
   5._EB, 5._EB, 5._EB, 25._EB, 25._EB, 25._EB/),(/N_BAND_C5H8O2,3/))

SD_C5H8O2_TEMP = (/ &
   297._EB, 396._EB, 441._EB, 483._EB, 597._EB, 803._EB,&
   1014._EB/)

BE_C5H8O2 = (/ &
   1.000_EB, 1.000_EB, 1.000_EB, 1.000_EB, 1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_C5H8O2(N_TEMP_C5H8O2,27))

! BAND #1: 750 CM-1 - 880 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 8.0046 %

SD1_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 750-785 CM-1
   6.713660E-04_EB, 5.825251E-04_EB, 7.916094E-04_EB, 6.543173E-04_EB, 6.421674E-04_EB, 6.087532E-04_EB, &
   7.107910E-04_EB, &
   9.848065E-04_EB, 9.607899E-04_EB, 7.994027E-04_EB, 1.090851E-03_EB, 1.087202E-03_EB, 7.734671E-04_EB, &
   9.284135E-04_EB, &
   9.110967E-04_EB, 7.183894E-04_EB, 8.010414E-04_EB, 1.063396E-03_EB, 1.000702E-03_EB, 6.657414E-04_EB, &
   6.127113E-04_EB, &
   8.267669E-04_EB, 9.315188E-04_EB, 8.290909E-04_EB, 7.969467E-04_EB, 8.913743E-04_EB, 6.855529E-04_EB, &
   7.046623E-04_EB, &
   1.042663E-03_EB, 1.043538E-03_EB, 8.872311E-04_EB, 8.694322E-04_EB, 9.329158E-04_EB, 9.772556E-04_EB, &
   7.458198E-04_EB, &
   1.006578E-03_EB, 7.209128E-04_EB, 1.019682E-03_EB, 7.712991E-04_EB, 3.187441E-02_EB, 5.248309E-02_EB, &
   8.138133E-04_EB, &
   7.585006E-04_EB, 8.285426E-04_EB, 7.850058E-04_EB, 7.514949E-04_EB, 2.435151E+00_EB, 2.382451E-01_EB, &
   7.995357E-04_EB, &
   5.346436E-03_EB, 1.695093E-03_EB, 1.005551E-02_EB, 1.807196E-02_EB, 4.893034E+00_EB, 1.072109E+00_EB, &
   8.263934E-04_EB/),(/7,8/))

SD1_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 790-825 CM-1
   7.889654E-02_EB, 6.932866E-02_EB, 6.943889E-02_EB, 1.657322E-01_EB, 4.820655E+00_EB, 1.595563E+00_EB, &
   6.604864E-04_EB, &
   2.859272E-01_EB, 2.060379E-01_EB, 1.784774E-01_EB, 2.655142E-01_EB, 5.069557E+00_EB, 1.587178E+00_EB, &
   8.491025E-04_EB, &
   6.960576E-01_EB, 3.989489E-01_EB, 3.440851E-01_EB, 4.569709E-01_EB, 4.118131E+00_EB, 7.958553E-01_EB, &
   7.688756E-04_EB, &
   1.250951E+00_EB, 6.520021E-01_EB, 6.089583E-01_EB, 6.456713E-01_EB, 2.261652E+00_EB, 7.908891E-01_EB, &
   8.388555E-04_EB, &
   2.208710E+00_EB, 1.110849E+00_EB, 1.035737E+00_EB, 1.049217E+00_EB, 1.084162E+00_EB, 7.448165E-01_EB, &
   1.121289E-03_EB, &
   3.745705E+00_EB, 1.437198E+00_EB, 1.241512E+00_EB, 1.206161E+00_EB, 8.198992E-01_EB, 7.316076E-01_EB, &
   7.825346E-04_EB, &
   3.469611E+00_EB, 1.274191E+00_EB, 1.123297E+00_EB, 1.063600E+00_EB, 7.020966E-01_EB, 8.989672E-01_EB, &
   7.836122E-04_EB, &
   3.387588E+00_EB, 1.219143E+00_EB, 1.110425E+00_EB, 9.818704E-01_EB, 6.253886E-01_EB, 1.450947E+00_EB, &
   8.923055E-04_EB/),(/7,8/))

SD1_C5H8O2(1:7,17:24) = RESHAPE((/ &  ! 830-865 CM-1
   2.650263E+00_EB, 9.577831E-01_EB, 8.753717E-01_EB, 8.295316E-01_EB, 6.934556E-01_EB, 1.431367E+00_EB, &
   8.664361E-04_EB, &
   1.730390E+00_EB, 6.490013E-01_EB, 6.546973E-01_EB, 5.810689E-01_EB, 2.534337E+00_EB, 7.747536E-01_EB, &
   7.994517E-04_EB, &
   1.002821E+00_EB, 3.608083E-01_EB, 3.892233E-01_EB, 3.747237E-01_EB, 6.259059E+00_EB, 1.624771E+00_EB, &
   7.025073E-04_EB, &
   4.114065E-01_EB, 1.202763E-01_EB, 1.380148E-01_EB, 2.829570E-01_EB, 7.146988E+00_EB, 1.774384E+00_EB, &
   9.127284E-04_EB, &
   1.274057E-01_EB, 9.521312E-04_EB, 7.626901E-04_EB, 5.309748E-02_EB, 6.530803E+00_EB, 3.950028E-01_EB, &
   8.260701E-04_EB, &
   3.101890E-02_EB, 9.960440E-04_EB, 1.450745E-03_EB, 8.889566E-04_EB, 4.823124E+00_EB, 5.607234E-02_EB, &
   1.289121E-03_EB, &
   1.815363E-02_EB, 6.716907E-04_EB, 1.134173E-03_EB, 7.672105E-04_EB, 2.305115E+00_EB, 1.215781E-02_EB, &
   9.963103E-03_EB, &
   1.671588E-02_EB, 9.534992E-04_EB, 7.842672E-04_EB, 1.050185E-03_EB, 8.279107E-04_EB, 9.473562E-04_EB, &
   1.520855E-02_EB/),(/7,8/))

SD1_C5H8O2(1:7,25:27) = RESHAPE((/ &  ! 870-880 CM-1
   3.388305E-02_EB, 8.968940E-04_EB, 9.267088E-04_EB, 1.004537E-03_EB, 9.822602E-04_EB, 8.968929E-04_EB, &
   3.122362E-02_EB, &
   2.663964E-02_EB, 7.688918E-04_EB, 6.407620E-04_EB, 5.722080E-04_EB, 9.499110E-04_EB, 6.115005E-04_EB, &
   3.351275E-02_EB, &
   4.683959E-04_EB, 4.682485E-04_EB, 4.694920E-04_EB, 4.693301E-04_EB, 4.691565E-04_EB, 4.675221E-04_EB, &
   4.659141E-04_EB/),(/7,3/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_C5H8O2(N_TEMP_C5H8O2,27))

! BAND #1: 750 CM-1 - 880 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 8.0046 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 750-785 CM-1
   4.881844E-04_EB, 4.346454E-04_EB, 5.958338E-04_EB, 4.571093E-04_EB, 4.822105E-04_EB, 4.530250E-04_EB, &
   5.750719E-04_EB, &
   6.727325E-04_EB, 6.267647E-04_EB, 5.477130E-04_EB, 7.657452E-04_EB, 1.077214E-03_EB, 5.499367E-04_EB, &
   6.460319E-04_EB, &
   5.864217E-04_EB, 5.127786E-04_EB, 5.824402E-04_EB, 8.990885E-04_EB, 8.469550E-04_EB, 5.075038E-04_EB, &
   5.043115E-04_EB, &
   6.108399E-04_EB, 7.428392E-04_EB, 5.525420E-04_EB, 5.867387E-04_EB, 7.082636E-04_EB, 5.889027E-04_EB, &
   4.893654E-04_EB, &
   7.434780E-04_EB, 6.504896E-04_EB, 6.589451E-04_EB, 6.802512E-04_EB, 7.417022E-04_EB, 7.262805E-04_EB, &
   5.511207E-04_EB, &
   6.356817E-04_EB, 5.339199E-04_EB, 7.935459E-04_EB, 4.886757E-04_EB, 2.583559E-03_EB, 3.792064E-03_EB, &
   6.024047E-04_EB, &
   5.106586E-04_EB, 5.822074E-04_EB, 5.456352E-04_EB, 5.195168E-04_EB, 1.249121E-04_EB, 2.441600E-03_EB, &
   6.419995E-04_EB, &
   8.911013E-04_EB, 1.366487E-03_EB, 7.184717E-03_EB, 1.764827E-01_EB, 2.574464E-04_EB, 1.205059E-03_EB, &
   5.689721E-04_EB/),(/7,8/))

GAMMAD1_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 790-825 CM-1
   4.218735E-01_EB, 2.260530E-02_EB, 2.192833E-01_EB, 8.393035E-01_EB, 6.309253E-04_EB, 1.299978E-03_EB, &
   4.943829E-04_EB, &
   2.011501E+01_EB, 1.452133E-01_EB, 1.379684E+00_EB, 3.290149E+00_EB, 1.006802E-03_EB, 2.196075E-03_EB, &
   6.147171E-04_EB, &
   7.999647E+01_EB, 3.727151E-01_EB, 4.909546E+00_EB, 1.504204E-01_EB, 1.827877E-03_EB, 1.390294E-02_EB, &
   5.960009E-04_EB, &
   8.000000E+01_EB, 1.192733E+01_EB, 7.335305E+00_EB, 2.477027E+01_EB, 6.343012E-03_EB, 3.167933E-02_EB, &
   6.338454E-04_EB, &
   8.000000E+01_EB, 1.337438E+00_EB, 1.200630E+01_EB, 8.358805E-01_EB, 5.050388E-02_EB, 5.528025E-02_EB, &
   6.645315E-04_EB, &
   8.000000E+01_EB, 1.575943E+00_EB, 2.254763E+01_EB, 5.284652E-01_EB, 4.899995E-01_EB, 4.748211E-02_EB, &
   5.682371E-04_EB, &
   8.000000E+01_EB, 1.559461E+00_EB, 2.060048E+01_EB, 9.864427E-01_EB, 2.822989E+01_EB, 2.089522E-02_EB, &
   5.933048E-04_EB, &
   8.000000E+01_EB, 9.294253E-01_EB, 7.518592E-01_EB, 2.225028E+00_EB, 1.882489E+01_EB, 5.697081E-03_EB, &
   6.342437E-04_EB/),(/7,8/))

GAMMAD1_C5H8O2(1:7,17:24) = RESHAPE((/ &  ! 830-865 CM-1
   8.000000E+01_EB, 5.595499E-01_EB, 7.229848E-01_EB, 4.663622E-01_EB, 4.280954E-02_EB, 3.407433E-03_EB, &
   7.293662E-04_EB, &
   7.999678E+01_EB, 5.297318E-01_EB, 1.408882E-01_EB, 5.990214E-01_EB, 2.104101E-03_EB, 5.806669E-03_EB, &
   5.977527E-04_EB, &
   8.000000E+01_EB, 2.382664E+00_EB, 5.857715E-02_EB, 2.596665E-01_EB, 5.172312E-04_EB, 1.035222E-03_EB, &
   5.002373E-04_EB, &
   5.199739E+01_EB, 2.075323E-01_EB, 1.580424E-02_EB, 9.101537E-03_EB, 3.465216E-04_EB, 4.105029E-04_EB, &
   6.238247E-04_EB, &
   2.523145E+00_EB, 7.869446E-04_EB, 6.466080E-04_EB, 5.984897E-03_EB, 2.719400E-04_EB, 9.547889E-04_EB, &
   6.744617E-04_EB, &
   3.144180E-01_EB, 6.967438E-04_EB, 8.338212E-04_EB, 6.247549E-04_EB, 2.369812E-04_EB, 2.160043E-01_EB, &
   1.146982E-03_EB, &
   3.556347E-01_EB, 5.790029E-04_EB, 9.280773E-04_EB, 6.140435E-04_EB, 1.166505E-04_EB, 4.285636E-03_EB, &
   1.647599E-01_EB, &
   9.611149E-01_EB, 6.993787E-04_EB, 6.536778E-04_EB, 6.470283E-04_EB, 6.300754E-04_EB, 6.508677E-04_EB, &
   8.394488E-01_EB/),(/7,8/))

GAMMAD1_C5H8O2(1:7,25:27) = RESHAPE((/ &  ! 870-880 CM-1
   8.980663E-01_EB, 6.254884E-04_EB, 6.536611E-04_EB, 7.628907E-04_EB, 7.906692E-04_EB, 7.774676E-04_EB, &
   3.406991E+00_EB, &
   1.692425E-01_EB, 6.157700E-04_EB, 4.905940E-04_EB, 4.383728E-04_EB, 8.114566E-04_EB, 5.029333E-04_EB, &
   3.441873E-01_EB, &
   3.735791E-04_EB, 3.732776E-04_EB, 3.745026E-04_EB, 3.743404E-04_EB, 3.741743E-04_EB, 3.727339E-04_EB, &
   3.713571E-04_EB/),(/7,3/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_C5H8O2(N_TEMP_C5H8O2,37))

! BAND #2: 875 CM-1 - 1055 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 16.2578 %

SD2_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 875-910 CM-1
   2.556429E-02_EB, 6.579116E-04_EB, 7.354099E-04_EB, 7.413989E-04_EB, 6.278952E-04_EB, 6.133453E-04_EB, &
   4.127227E-02_EB, &
   4.274772E-02_EB, 9.965553E-04_EB, 9.559177E-04_EB, 9.609587E-04_EB, 7.244017E-04_EB, 9.125642E-04_EB, &
   1.253746E-01_EB, &
   4.062606E-02_EB, 8.758771E-04_EB, 7.717414E-04_EB, 1.136983E-03_EB, 8.565555E-02_EB, 8.384111E-01_EB, &
   8.119652E-01_EB, &
   1.969402E-02_EB, 8.690397E-04_EB, 8.197383E-04_EB, 1.021204E-01_EB, 7.942200E-01_EB, 1.336575E+00_EB, &
   1.034711E+00_EB, &
   6.150252E-03_EB, 8.420963E-04_EB, 7.601258E-03_EB, 5.822673E-01_EB, 1.849210E+00_EB, 2.267047E-01_EB, &
   6.054709E-01_EB, &
   5.255431E-03_EB, 3.759946E-03_EB, 6.674130E-02_EB, 1.933685E-01_EB, 4.747476E+00_EB, 3.382479E-01_EB, &
   2.717558E-01_EB, &
   6.508425E-02_EB, 7.385905E-02_EB, 1.516138E-01_EB, 2.231123E-01_EB, 3.585642E+00_EB, 4.915855E-01_EB, &
   2.578738E-01_EB, &
   2.660815E-01_EB, 2.584587E-01_EB, 3.361944E-01_EB, 4.729659E-01_EB, 4.590394E+00_EB, 7.118124E-01_EB, &
   2.680924E-01_EB/),(/7,8/))

SD2_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 915-950 CM-1
   7.131926E-01_EB, 5.670265E-01_EB, 5.731113E-01_EB, 6.713483E-01_EB, 5.036802E+00_EB, 1.022991E+00_EB, &
   2.779647E-01_EB, &
   1.616994E+00_EB, 1.062612E+00_EB, 9.571218E-01_EB, 9.882622E-01_EB, 4.253259E+00_EB, 9.880129E-01_EB, &
   3.040041E-01_EB, &
   3.093637E+00_EB, 1.592723E+00_EB, 1.358023E+00_EB, 1.283095E+00_EB, 2.191786E+00_EB, 8.168738E-01_EB, &
   3.592899E-01_EB, &
   4.490720E+00_EB, 2.020538E+00_EB, 1.689196E+00_EB, 1.592450E+00_EB, 1.670002E+00_EB, 9.479666E-01_EB, &
   4.149322E-01_EB, &
   5.933235E+00_EB, 2.516838E+00_EB, 2.089816E+00_EB, 1.959977E+00_EB, 1.677729E+00_EB, 1.102576E+00_EB, &
   4.329370E-01_EB, &
   7.889794E+00_EB, 2.918784E+00_EB, 2.348617E+00_EB, 2.059360E+00_EB, 1.457291E+00_EB, 9.460046E-01_EB, &
   4.177430E-01_EB, &
   5.687576E+00_EB, 2.199276E+00_EB, 1.750801E+00_EB, 1.589524E+00_EB, 9.841957E-01_EB, 7.638453E-01_EB, &
   4.276039E-01_EB, &
   3.912971E+00_EB, 1.521382E+00_EB, 1.199146E+00_EB, 1.132591E+00_EB, 7.312452E-01_EB, 5.627116E-01_EB, &
   4.340964E-01_EB/),(/7,8/))

SD2_C5H8O2(1:7,17:24) = RESHAPE((/ &  ! 955-990 CM-1
   2.475675E+00_EB, 1.047082E+00_EB, 8.418699E-01_EB, 7.489753E-01_EB, 5.673221E-01_EB, 3.884845E-01_EB, &
   4.323566E-01_EB, &
   1.320393E+00_EB, 6.735843E-01_EB, 5.433291E-01_EB, 5.032234E-01_EB, 1.017197E+00_EB, 2.871344E-01_EB, &
   4.412980E-01_EB, &
   5.592797E-01_EB, 4.019616E-01_EB, 3.047795E-01_EB, 3.003325E-01_EB, 2.137201E+00_EB, 2.538634E-01_EB, &
   4.343141E-01_EB, &
   2.669833E-01_EB, 2.361810E-01_EB, 1.839138E-01_EB, 1.867551E-01_EB, 2.802026E+00_EB, 2.467341E-01_EB, &
   8.431414E-01_EB, &
   3.073555E-01_EB, 1.587138E-01_EB, 1.753897E-01_EB, 1.623549E-01_EB, 2.391695E+00_EB, 2.295551E-01_EB, &
   1.916065E+00_EB, &
   5.710281E-01_EB, 2.449444E-01_EB, 2.284392E-01_EB, 2.247888E-01_EB, 2.482041E+00_EB, 2.586464E-01_EB, &
   2.918096E+00_EB, &
   8.249745E-01_EB, 3.134360E-01_EB, 2.982082E-01_EB, 2.831453E-01_EB, 1.314234E+00_EB, 3.428023E-01_EB, &
   2.320833E+00_EB, &
   9.299117E-01_EB, 3.934637E-01_EB, 3.941485E-01_EB, 4.276205E-01_EB, 6.380472E-01_EB, 3.539805E-01_EB, &
   1.717216E+00_EB/),(/7,8/))

SD2_C5H8O2(1:7,25:32) = RESHAPE((/ &  ! 995-1030 CM-1
   1.174651E+00_EB, 5.076500E-01_EB, 5.123310E-01_EB, 5.072349E-01_EB, 6.017282E-01_EB, 4.040867E-01_EB, &
   9.071058E-01_EB, &
   1.220989E+00_EB, 6.322818E-01_EB, 6.487477E-01_EB, 6.488986E-01_EB, 5.856188E-01_EB, 4.402415E-01_EB, &
   6.313728E-01_EB, &
   1.342002E+00_EB, 7.952593E-01_EB, 7.945733E-01_EB, 8.003597E-01_EB, 7.389281E-01_EB, 4.724643E-01_EB, &
   1.867048E-01_EB, &
   2.059673E+00_EB, 1.016101E+00_EB, 9.573528E-01_EB, 9.061740E-01_EB, 6.905226E-01_EB, 5.789912E-01_EB, &
   1.181913E-01_EB, &
   2.708834E+00_EB, 1.261292E+00_EB, 1.056622E+00_EB, 9.704385E-01_EB, 7.470068E-01_EB, 3.592001E-01_EB, &
   1.004398E-01_EB, &
   3.415071E+00_EB, 1.258792E+00_EB, 1.017232E+00_EB, 8.840290E-01_EB, 7.155771E-01_EB, 2.368274E-01_EB, &
   1.095937E-01_EB, &
   3.371272E+00_EB, 1.241748E+00_EB, 9.412804E-01_EB, 7.974575E-01_EB, 8.067485E-01_EB, 1.865056E-01_EB, &
   1.178328E-01_EB, &
   2.426657E+00_EB, 8.931409E-01_EB, 6.775266E-01_EB, 5.586942E-01_EB, 6.967074E-01_EB, 1.154895E-01_EB, &
   1.127254E-01_EB/),(/7,8/))

SD2_C5H8O2(1:7,33:37) = RESHAPE((/ &  ! 1035-1055 CM-1
   1.800189E+00_EB, 5.884496E-01_EB, 4.478842E-01_EB, 3.611785E-01_EB, 6.015273E-01_EB, 2.659912E-02_EB, &
   1.137885E-01_EB, &
   6.470612E-01_EB, 2.347793E-01_EB, 1.941059E-01_EB, 1.585485E-01_EB, 1.218801E+00_EB, 6.925778E-04_EB, &
   6.825179E-01_EB, &
   1.671523E-01_EB, 4.958688E-02_EB, 7.155642E-02_EB, 4.212993E-02_EB, 4.481255E-02_EB, 8.744158E-04_EB, &
   6.166493E-01_EB, &
   4.210092E-02_EB, 5.872012E-03_EB, 1.202013E-02_EB, 8.673460E-03_EB, 7.247323E-04_EB, 7.563806E-04_EB, &
   1.496397E-01_EB, &
   4.683959E-04_EB, 4.682485E-04_EB, 4.694920E-04_EB, 4.693301E-04_EB, 4.691565E-04_EB, 4.675221E-04_EB, &
   4.659141E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_C5H8O2(N_TEMP_C5H8O2,37))

! BAND #2: 875 CM-1 - 1055 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 16.2578 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 875-910 CM-1
   2.823227E-01_EB, 4.610224E-04_EB, 4.906774E-04_EB, 5.417213E-04_EB, 5.214023E-04_EB, 4.619138E-04_EB, &
   2.996879E-01_EB, &
   1.589262E+01_EB, 7.112738E-04_EB, 6.863480E-04_EB, 6.954244E-04_EB, 5.636836E-04_EB, 6.909238E-04_EB, &
   1.362275E+00_EB, &
   1.255210E+01_EB, 6.773562E-04_EB, 5.735201E-04_EB, 8.272739E-04_EB, 1.724787E-04_EB, 1.661576E-04_EB, &
   1.138271E-03_EB, &
   7.051283E-01_EB, 6.891085E-04_EB, 7.024554E-04_EB, 1.678129E-04_EB, 1.524293E-04_EB, 5.547075E-04_EB, &
   1.418645E-03_EB, &
   7.121508E-02_EB, 6.384517E-04_EB, 1.642701E-03_EB, 1.694248E-04_EB, 1.470911E-04_EB, 6.320090E-02_EB, &
   4.163931E-03_EB, &
   2.508564E-03_EB, 4.326233E-03_EB, 1.052820E-02_EB, 5.592027E-03_EB, 2.571364E-04_EB, 3.075452E-02_EB, &
   7.237307E-02_EB, &
   3.527481E-01_EB, 6.803584E-02_EB, 4.286749E-02_EB, 3.844453E-02_EB, 5.858598E-04_EB, 2.259184E-02_EB, &
   3.898073E+00_EB, &
   1.571067E+01_EB, 6.225180E-02_EB, 6.038369E-02_EB, 2.741791E-02_EB, 1.292359E-03_EB, 2.289985E-02_EB, &
   1.760940E+01_EB/),(/7,8/))

GAMMAD2_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 915-950 CM-1
   8.000000E+01_EB, 1.428430E-01_EB, 2.419731E+00_EB, 2.230224E-01_EB, 2.255022E-03_EB, 2.094570E-02_EB, &
   2.051410E+01_EB, &
   8.000000E+01_EB, 1.709830E-01_EB, 3.380268E+00_EB, 5.240909E-01_EB, 4.853055E-03_EB, 4.030274E-02_EB, &
   2.349415E+01_EB, &
   8.000000E+01_EB, 2.636763E-01_EB, 3.653768E+00_EB, 1.717666E+00_EB, 2.060690E-02_EB, 2.522187E-01_EB, &
   2.863894E+01_EB, &
   8.000000E+01_EB, 3.872301E-01_EB, 3.145524E+01_EB, 1.984056E+00_EB, 7.238493E-02_EB, 1.744457E-01_EB, &
   1.234446E+01_EB, &
   8.000000E+01_EB, 5.199554E-01_EB, 5.201416E+01_EB, 1.219270E+00_EB, 1.366230E-01_EB, 1.024375E-01_EB, &
   6.859851E+00_EB, &
   8.000000E+01_EB, 6.059595E-01_EB, 6.357106E+01_EB, 2.510435E+00_EB, 2.483816E-01_EB, 1.870400E-01_EB, &
   6.030901E+00_EB, &
   8.000000E+01_EB, 4.613948E-01_EB, 6.558012E+01_EB, 9.318435E-01_EB, 8.280329E-01_EB, 1.702565E-01_EB, &
   3.671889E+00_EB, &
   8.000000E+01_EB, 2.805617E-01_EB, 4.753473E+01_EB, 4.243871E-01_EB, 1.998842E-01_EB, 1.897322E-01_EB, &
   3.182135E-01_EB/),(/7,8/))

GAMMAD2_C5H8O2(1:7,17:24) = RESHAPE((/ &  ! 955-990 CM-1
   8.000000E+01_EB, 1.917580E-01_EB, 3.160507E+01_EB, 3.013932E+00_EB, 8.610150E-02_EB, 2.680998E+00_EB, &
   1.225672E-01_EB, &
   8.000000E+01_EB, 8.181705E-02_EB, 1.393264E+01_EB, 4.846622E+00_EB, 5.364990E-03_EB, 7.410589E+00_EB, &
   7.715299E-02_EB, &
   7.999744E+01_EB, 2.071421E-02_EB, 2.646321E+00_EB, 1.308531E+00_EB, 1.066053E-03_EB, 5.993038E+00_EB, &
   7.459897E-02_EB, &
   1.600979E+01_EB, 6.164883E-03_EB, 3.149166E-01_EB, 1.578432E+00_EB, 6.172180E-04_EB, 3.880878E+00_EB, &
   7.684969E-03_EB, &
   2.587709E+01_EB, 3.007808E-02_EB, 1.472547E-01_EB, 1.384504E+00_EB, 6.759922E-04_EB, 4.013279E+00_EB, &
   2.292324E-03_EB, &
   7.999663E+01_EB, 3.784955E-02_EB, 4.354468E-01_EB, 2.323880E+00_EB, 8.249666E-04_EB, 2.529707E-01_EB, &
   1.448969E-03_EB, &
   8.000000E+01_EB, 6.887494E-02_EB, 3.383023E-01_EB, 3.534843E+00_EB, 2.269583E-03_EB, 7.251788E-02_EB, &
   1.620063E-03_EB, &
   8.000000E+01_EB, 1.396664E-01_EB, 2.500157E-01_EB, 1.154998E-01_EB, 1.324480E-02_EB, 1.665776E+00_EB, &
   1.569506E-03_EB/),(/7,8/))

GAMMAD2_C5H8O2(1:7,25:32) = RESHAPE((/ &  ! 995-1030 CM-1
   7.999999E+01_EB, 2.622383E-01_EB, 3.680714E-01_EB, 3.431326E-01_EB, 3.214335E-02_EB, 3.028902E+00_EB, &
   2.505970E-03_EB, &
   8.000000E+01_EB, 2.931223E-01_EB, 4.763129E-01_EB, 5.158658E-01_EB, 1.564583E-01_EB, 4.068234E-01_EB, &
   3.470044E-03_EB, &
   8.000000E+01_EB, 2.591539E-01_EB, 9.760853E-01_EB, 3.725883E-01_EB, 6.472451E-02_EB, 7.215647E-02_EB, &
   3.438220E+00_EB, &
   8.000000E+01_EB, 2.874591E-01_EB, 8.902719E-01_EB, 3.997568E-01_EB, 1.273205E-01_EB, 2.106275E-02_EB, &
   1.404359E+00_EB, &
   8.000000E+01_EB, 2.627488E-01_EB, 2.860341E+01_EB, 4.718599E-01_EB, 6.177797E-02_EB, 7.129412E-02_EB, &
   1.080505E+00_EB, &
   8.000000E+01_EB, 3.044334E-01_EB, 1.893461E+01_EB, 7.512970E-01_EB, 3.525571E-02_EB, 8.428393E+00_EB, &
   1.655302E+00_EB, &
   8.000000E+01_EB, 2.042339E-01_EB, 2.401079E+01_EB, 1.842385E+00_EB, 1.678758E-02_EB, 1.160390E+01_EB, &
   7.709535E-02_EB, &
   8.000000E+01_EB, 1.532358E-01_EB, 2.398691E+01_EB, 7.959409E+00_EB, 8.237636E-03_EB, 1.840439E+00_EB, &
   1.574438E-01_EB/),(/7,8/))

GAMMAD2_C5H8O2(1:7,33:37) = RESHAPE((/ &  ! 1035-1055 CM-1
   8.000000E+01_EB, 1.344383E-01_EB, 2.700694E+00_EB, 2.144500E+00_EB, 2.397408E-03_EB, 2.813670E-01_EB, &
   5.280750E-02_EB, &
   7.999478E+01_EB, 7.115998E-02_EB, 2.822221E-01_EB, 9.009493E-01_EB, 3.364759E-04_EB, 5.682134E-04_EB, &
   7.024865E-04_EB, &
   1.748846E+01_EB, 1.217636E-01_EB, 4.857488E-02_EB, 5.398096E-01_EB, 4.000099E-03_EB, 5.990465E-04_EB, &
   6.551947E-04_EB, &
   1.394167E+01_EB, 2.046399E-01_EB, 8.910555E-03_EB, 6.157030E-03_EB, 5.599763E-04_EB, 5.579403E-04_EB, &
   9.116727E-04_EB, &
   3.735791E-04_EB, 3.732776E-04_EB, 3.745026E-04_EB, 3.743404E-04_EB, 3.741743E-04_EB, 3.727339E-04_EB, &
   3.713571E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(SD3_C5H8O2(N_TEMP_C5H8O2,18))

! BAND #3: 1050 CM-1 - 1275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 38.3736 %

SD3_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 1050-1085 CM-1
   2.283540E-02_EB, 4.922172E-04_EB, 2.560169E-03_EB, 3.717395E-03_EB, 5.679232E-04_EB, 5.948702E-04_EB, &
   9.070861E-02_EB, &
   1.330120E-02_EB, 4.682485E-04_EB, 6.438257E-03_EB, 1.249445E-02_EB, 1.158898E-03_EB, 8.034057E-04_EB, &
   3.528869E-01_EB, &
   5.312158E-04_EB, 4.682485E-04_EB, 3.462862E-03_EB, 4.802455E-03_EB, 1.325886E-03_EB, 9.860835E-04_EB, &
   5.966300E-01_EB, &
   6.730728E-04_EB, 5.383842E-04_EB, 1.541199E-02_EB, 1.216344E-03_EB, 8.711834E-04_EB, 2.095446E-02_EB, &
   2.344768E-01_EB, &
   4.842466E-03_EB, 5.362735E-03_EB, 2.344159E-02_EB, 1.759081E-02_EB, 9.888946E-04_EB, 5.883826E-02_EB, &
   1.524721E+00_EB, &
   1.962391E-02_EB, 1.842385E-02_EB, 4.296535E-02_EB, 3.232193E-02_EB, 8.684561E-04_EB, 1.277349E-01_EB, &
   1.441831E+00_EB, &
   3.957410E-02_EB, 3.609774E-02_EB, 7.027800E-02_EB, 6.897527E-02_EB, 1.805630E-02_EB, 1.769389E-01_EB, &
   2.219797E+00_EB, &
   6.033916E-02_EB, 5.989100E-02_EB, 1.135187E-01_EB, 1.032544E-01_EB, 9.054237E-02_EB, 2.674346E-01_EB, &
   1.717514E+00_EB/),(/7,8/))

SD3_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 1090-1225 CM-1
   1.000494E-01_EB, 9.414817E-02_EB, 1.444499E-01_EB, 1.465825E-01_EB, 1.695405E-01_EB, 3.353830E-01_EB, &
   6.307295E-01_EB, &
   1.689087E-01_EB, 1.393048E-01_EB, 1.815843E-01_EB, 1.798559E-01_EB, 3.961284E-01_EB, 3.870093E-01_EB, &
   5.212840E-01_EB, &
   2.316504E-01_EB, 1.905944E-01_EB, 2.231146E-01_EB, 2.373041E-01_EB, 6.108176E-01_EB, 5.517315E-01_EB, &
   5.514970E-01_EB, &
   7.470458E-01_EB, 7.308402E-01_EB, 9.133596E-01_EB, 1.083670E+00_EB, 1.643403E+00_EB, 2.654844E+00_EB, &
   2.192364E+00_EB, &
   9.479777E+00_EB, 8.554789E+00_EB, 7.931693E+00_EB, 8.343454E+00_EB, 7.636411E+00_EB, 6.546753E+00_EB, &
   3.098983E+00_EB, &
   2.594269E+01_EB, 1.307137E+01_EB, 1.100488E+01_EB, 1.045080E+01_EB, 8.049716E+00_EB, 4.871530E+00_EB, &
   1.930570E+00_EB, &
   1.934931E+01_EB, 7.224029E+00_EB, 5.871145E+00_EB, 5.292871E+00_EB, 3.621708E+00_EB, 1.992252E+00_EB, &
   7.777877E-01_EB, &
   4.140490E+00_EB, 2.139609E+00_EB, 1.821476E+00_EB, 1.658745E+00_EB, 1.493490E+00_EB, 7.032707E-01_EB, &
   2.294717E+00_EB/),(/7,8/))

SD3_C5H8O2(1:7,17:18) = RESHAPE((/ &  ! 1250-1275 CM-1
   5.271264E-01_EB, 2.912938E-01_EB, 2.917520E-01_EB, 2.643227E-01_EB, 5.514219E-01_EB, 1.728159E-01_EB, &
   9.935754E-01_EB, &
   4.683959E-04_EB, 4.682485E-04_EB, 4.694920E-04_EB, 4.693301E-04_EB, 4.691565E-04_EB, 4.675221E-04_EB, &
   4.659141E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD3_C5H8O2(N_TEMP_C5H8O2,18))

! BAND #3: 1050 CM-1 - 1275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 38.3736 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD3_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 1050-1085 CM-1
   6.752973E-01_EB, 3.965369E-04_EB, 1.852648E-03_EB, 6.404227E-03_EB, 4.395271E-04_EB, 4.775384E-04_EB, &
   2.331620E-03_EB, &
   2.308916E-01_EB, 3.732776E-04_EB, 1.932620E-03_EB, 2.619408E-03_EB, 7.975806E-04_EB, 6.318845E-04_EB, &
   1.556634E-03_EB, &
   4.424232E-04_EB, 3.732776E-04_EB, 2.729364E-03_EB, 1.540902E-03_EB, 9.180318E-04_EB, 7.423991E-04_EB, &
   8.482000E-04_EB, &
   5.798906E-04_EB, 4.109335E-04_EB, 4.684962E-03_EB, 1.748840E-03_EB, 5.924212E-04_EB, 8.917151E-02_EB, &
   3.547868E-03_EB, &
   2.445999E-01_EB, 1.709414E-03_EB, 8.476824E-03_EB, 1.070389E-02_EB, 6.275957E-04_EB, 4.118998E-01_EB, &
   5.365592E-04_EB, &
   2.903925E+00_EB, 1.089566E-02_EB, 2.453372E-02_EB, 1.655663E-01_EB, 5.991526E-04_EB, 4.446612E+00_EB, &
   5.688486E-04_EB, &
   1.701839E+01_EB, 1.480657E-01_EB, 7.478547E-02_EB, 3.425426E-01_EB, 6.291144E-01_EB, 9.461117E-01_EB, &
   4.469811E-04_EB, &
   6.073385E-01_EB, 1.511806E-01_EB, 1.786293E-01_EB, 1.777844E+00_EB, 9.776316E-01_EB, 7.402583E+00_EB, &
   1.041732E-03_EB/),(/7,8/))

GAMMAD3_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 1090-1225 CM-1
   1.122342E+00_EB, 2.158159E-01_EB, 6.598921E-01_EB, 8.611218E-01_EB, 1.210188E-01_EB, 1.184061E+01_EB, &
   8.335214E-03_EB, &
   5.162273E+00_EB, 3.637833E-01_EB, 7.101749E-01_EB, 1.752143E+00_EB, 7.256519E-03_EB, 3.453023E+01_EB, &
   6.017312E-02_EB, &
   1.631154E+01_EB, 1.875513E-01_EB, 2.486087E+00_EB, 3.547150E+00_EB, 7.521178E-03_EB, 2.393089E+01_EB, &
   3.039081E+01_EB, &
   8.000000E+01_EB, 3.366139E-01_EB, 1.096627E+01_EB, 2.588432E+01_EB, 2.813109E-01_EB, 1.156638E+00_EB, &
   1.806175E-01_EB, &
   8.000000E+01_EB, 6.558841E-01_EB, 1.962907E+00_EB, 1.479647E+00_EB, 1.072565E+00_EB, 2.709515E+00_EB, &
   3.334761E-01_EB, &
   8.000000E+01_EB, 1.895544E+00_EB, 9.110170E+00_EB, 3.023831E+00_EB, 8.442824E-01_EB, 1.836944E+00_EB, &
   1.622069E-01_EB, &
   8.000000E+01_EB, 1.580597E+00_EB, 6.536920E+01_EB, 2.757392E+00_EB, 6.660127E-01_EB, 4.829505E+01_EB, &
   1.550614E-01_EB, &
   7.999934E+01_EB, 3.683494E-01_EB, 5.153029E+00_EB, 1.387986E+00_EB, 7.620403E-02_EB, 3.958904E+01_EB, &
   1.950315E-03_EB/),(/7,8/))

GAMMAD3_C5H8O2(1:7,17:18) = RESHAPE((/ &  ! 1250-1275 CM-1
   7.999908E+01_EB, 9.980022E-02_EB, 3.152017E-01_EB, 1.039694E+00_EB, 3.550883E-03_EB, 1.085236E+00_EB, &
   8.696632E-04_EB, &
   3.735791E-04_EB, 3.732776E-04_EB, 3.745026E-04_EB, 3.743404E-04_EB, 3.741743E-04_EB, 3.727339E-04_EB, &
   3.713571E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(SD4_C5H8O2(N_TEMP_C5H8O2,14))

! BAND #4: 1250 CM-1 - 1575 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 26.5551 %

SD4_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 1250-1425 CM-1
   2.651267E-01_EB, 1.759638E-01_EB, 1.769296E-01_EB, 1.699091E-01_EB, 2.046845E-01_EB, 2.341468E-01_EB, &
   1.105642E+00_EB, &
   1.022795E+00_EB, 9.266422E-01_EB, 1.017170E+00_EB, 1.118834E+00_EB, 1.488453E+00_EB, 1.472991E+00_EB, &
   1.015529E+00_EB, &
   9.008874E+00_EB, 4.815679E+00_EB, 4.305873E+00_EB, 4.164932E+00_EB, 3.844447E+00_EB, 2.320999E+00_EB, &
   1.064962E+00_EB, &
   1.317933E+01_EB, 5.235972E+00_EB, 4.254435E+00_EB, 3.735700E+00_EB, 2.736514E+00_EB, 1.279456E+00_EB, &
   5.251123E-01_EB, &
   3.335495E+00_EB, 1.313429E+00_EB, 1.062255E+00_EB, 8.454394E-01_EB, 1.978232E+00_EB, 2.277879E-01_EB, &
   3.643094E-02_EB, &
   1.208610E+00_EB, 5.576549E-01_EB, 5.119243E-01_EB, 4.375281E-01_EB, 2.260111E+00_EB, 2.451094E-01_EB, &
   9.103457E-02_EB, &
   1.929443E+00_EB, 8.888330E-01_EB, 8.250508E-01_EB, 7.384043E-01_EB, 2.462957E+00_EB, 6.121275E-01_EB, &
   1.297550E-01_EB, &
   2.558288E+00_EB, 1.354005E+00_EB, 1.257364E+00_EB, 1.190951E+00_EB, 1.431084E+00_EB, 8.234527E-01_EB, &
   1.945175E-01_EB/),(/7,8/))

SD4_C5H8O2(1:7,9:14) = RESHAPE((/ &  ! 1450-1575 CM-1
   7.269681E+00_EB, 2.804495E+00_EB, 2.300589E+00_EB, 2.018439E+00_EB, 1.754481E+00_EB, 8.665175E-01_EB, &
   1.851434E-01_EB, &
   3.164709E+00_EB, 1.268072E+00_EB, 1.073237E+00_EB, 9.062173E-01_EB, 1.779966E+00_EB, 4.279472E-01_EB, &
   6.749668E-02_EB, &
   4.203220E-01_EB, 2.293462E-01_EB, 2.285557E-01_EB, 1.794185E-01_EB, 1.212331E+00_EB, 7.133443E-02_EB, &
   9.727769E-04_EB, &
   3.205661E-01_EB, 1.369195E-01_EB, 1.279659E-01_EB, 9.148681E-02_EB, 2.184744E-01_EB, 3.056242E-03_EB, &
   9.133049E-04_EB, &
   1.379741E-01_EB, 6.041193E-02_EB, 5.056159E-02_EB, 3.540082E-02_EB, 4.510264E-03_EB, 3.305673E-03_EB, &
   6.976907E-04_EB, &
   4.683959E-04_EB, 4.682485E-04_EB, 4.694920E-04_EB, 4.693301E-04_EB, 4.691565E-04_EB, 4.675221E-04_EB, &
   4.659141E-04_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD4_C5H8O2(N_TEMP_C5H8O2,14))

! BAND #4: 1250 CM-1 - 1575 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 26.5551 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD4_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 1250-1425 CM-1
   1.900794E+01_EB, 4.523774E-02_EB, 1.704620E-01_EB, 3.550652E-01_EB, 3.540509E-02_EB, 1.516513E+00_EB, &
   1.492599E-03_EB, &
   8.000000E+01_EB, 2.226090E-01_EB, 4.037264E+01_EB, 2.183364E+01_EB, 1.078975E-01_EB, 5.536881E-01_EB, &
   4.287697E-02_EB, &
   8.000000E+01_EB, 1.040937E+00_EB, 6.528460E+01_EB, 3.062031E+00_EB, 2.657658E-01_EB, 1.018869E+00_EB, &
   4.868239E-02_EB, &
   8.000000E+01_EB, 1.163412E+00_EB, 6.557988E+01_EB, 3.224408E+00_EB, 2.125073E-01_EB, 2.326661E-01_EB, &
   7.389632E-03_EB, &
   8.000000E+01_EB, 2.308729E-01_EB, 6.363109E-01_EB, 5.086314E+00_EB, 4.111352E-03_EB, 1.626691E-01_EB, &
   1.308121E+00_EB, &
   8.000000E+01_EB, 1.969078E-01_EB, 1.609102E+00_EB, 5.383790E+00_EB, 1.498115E-03_EB, 3.386863E-02_EB, &
   5.998714E-02_EB, &
   8.000000E+01_EB, 3.410180E-01_EB, 6.907342E+00_EB, 7.887927E+00_EB, 3.683657E-03_EB, 2.330242E-02_EB, &
   2.134694E+00_EB, &
   8.000000E+01_EB, 3.694747E-01_EB, 4.202343E+00_EB, 1.302414E+00_EB, 3.479140E-02_EB, 7.927698E-02_EB, &
   7.049147E+00_EB/),(/7,8/))

GAMMAD4_C5H8O2(1:7,9:14) = RESHAPE((/ &  ! 1450-1575 CM-1
   8.000000E+01_EB, 7.141208E-01_EB, 6.536640E+01_EB, 2.342027E+00_EB, 7.034590E-02_EB, 1.139549E-01_EB, &
   5.068590E+00_EB, &
   8.000000E+01_EB, 3.439583E-01_EB, 2.158187E+00_EB, 1.101709E+00_EB, 6.060588E-03_EB, 2.185348E-02_EB, &
   3.183783E+00_EB, &
   5.006561E+01_EB, 3.299422E-01_EB, 1.190294E+00_EB, 8.974212E-01_EB, 3.345373E-04_EB, 2.520096E-02_EB, &
   1.139827E-03_EB, &
   2.293095E+01_EB, 1.421608E-01_EB, 7.208656E-01_EB, 9.099123E-01_EB, 6.014407E-05_EB, 2.971282E-03_EB, &
   6.193551E-04_EB, &
   2.442962E+00_EB, 4.484537E-02_EB, 1.690359E-01_EB, 1.259316E-01_EB, 9.393127E-04_EB, 1.423587E-03_EB, &
   5.673518E-04_EB, &
   3.735791E-04_EB, 3.732776E-04_EB, 3.745026E-04_EB, 3.743404E-04_EB, 3.741743E-04_EB, 3.727339E-04_EB, &
   3.713571E-04_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(SD5_C5H8O2(N_TEMP_C5H8O2,18))

! BAND #5: 1550 CM-1 - 1975 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 33.9665 %

SD5_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 1550-1725 CM-1
   1.283781E-01_EB, 5.564827E-02_EB, 4.552881E-02_EB, 3.248586E-02_EB, 2.962603E-03_EB, 5.775666E-03_EB, &
   6.549796E-04_EB, &
   3.228302E-01_EB, 1.330461E-01_EB, 1.436585E-01_EB, 1.010144E-01_EB, 3.539658E+00_EB, 7.633235E-01_EB, &
   3.006478E-03_EB, &
   4.297863E-01_EB, 2.130911E-01_EB, 3.958264E-01_EB, 2.086835E-01_EB, 5.228719E+00_EB, 1.073793E+00_EB, &
   5.259787E-02_EB, &
   1.636030E+00_EB, 7.901740E-01_EB, 7.307309E-01_EB, 6.594717E-01_EB, 1.967717E+00_EB, 3.028161E-01_EB, &
   7.555184E-02_EB, &
   2.786162E+00_EB, 9.811787E-01_EB, 7.947490E-01_EB, 6.679210E-01_EB, 3.937599E-01_EB, 1.926412E-01_EB, &
   4.895888E-02_EB, &
   5.371774E-01_EB, 2.528496E-01_EB, 2.792431E-01_EB, 2.552566E-01_EB, 4.163329E-01_EB, 2.237360E-01_EB, &
   1.885545E-01_EB, &
   1.178231E+00_EB, 6.416768E-01_EB, 6.572166E-01_EB, 7.189543E-01_EB, 7.686419E-01_EB, 9.688893E-01_EB, &
   6.754928E-01_EB, &
   7.583351E+00_EB, 5.772869E+00_EB, 5.148166E+00_EB, 5.223398E+00_EB, 4.376104E+00_EB, 3.382685E+00_EB, &
   1.483712E+00_EB/),(/7,8/))

SD5_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 1750-1925 CM-1
   2.394789E+01_EB, 1.094859E+01_EB, 8.981291E+00_EB, 8.316621E+00_EB, 5.789605E+00_EB, 3.531464E+00_EB, &
   1.436978E+00_EB, &
   1.830194E+00_EB, 1.308594E+00_EB, 1.243723E+00_EB, 1.234798E+00_EB, 1.062593E+00_EB, 8.475205E-01_EB, &
   4.802389E-01_EB, &
   6.055532E-01_EB, 3.245427E-01_EB, 3.283205E-01_EB, 2.996324E-01_EB, 2.752568E-01_EB, 1.605966E-01_EB, &
   1.734273E-01_EB, &
   1.185795E-01_EB, 5.910661E-02_EB, 5.293524E-02_EB, 4.089505E-02_EB, 8.894114E-02_EB, 1.564737E-02_EB, &
   4.355749E-02_EB, &
   4.355125E-02_EB, 1.987367E-02_EB, 8.799089E-02_EB, 2.271459E-02_EB, 2.895065E-02_EB, 3.488458E-03_EB, &
   2.838444E-02_EB, &
   3.376642E-01_EB, 1.368154E-01_EB, 1.357689E-01_EB, 1.064263E-01_EB, 6.258461E-02_EB, 1.334385E-02_EB, &
   8.818944E-03_EB, &
   2.105506E-01_EB, 6.949000E-02_EB, 6.742734E-02_EB, 5.316113E-02_EB, 2.371958E-02_EB, 9.426557E-04_EB, &
   1.609069E-03_EB, &
   1.850030E-02_EB, 4.412077E-03_EB, 7.600489E-04_EB, 2.224295E-03_EB, 7.330402E-04_EB, 7.865066E-04_EB, &
   7.116352E-04_EB/),(/7,8/))

SD5_C5H8O2(1:7,17:18) = RESHAPE((/ &  ! 1950-1975 CM-1
   9.010576E-04_EB, 6.552187E-04_EB, 7.003811E-04_EB, 6.830920E-04_EB, 7.111960E-04_EB, 6.441936E-04_EB, &
   6.424584E-04_EB, &
   4.683959E-04_EB, 4.682485E-04_EB, 4.694920E-04_EB, 4.693301E-04_EB, 4.691565E-04_EB, 4.675221E-04_EB, &
   4.659141E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD5_C5H8O2(N_TEMP_C5H8O2,18))

! BAND #5: 1550 CM-1 - 1975 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 33.9665 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD5_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 1550-1725 CM-1
   2.001538E+00_EB, 1.043359E-01_EB, 1.744313E-01_EB, 5.342194E-02_EB, 1.360786E-03_EB, 2.592801E-03_EB, &
   5.075444E-04_EB, &
   1.725159E+01_EB, 1.172078E-01_EB, 4.594855E-02_EB, 2.031343E-01_EB, 1.111304E-04_EB, 2.683498E-04_EB, &
   1.034216E-01_EB, &
   4.985857E+01_EB, 1.394095E-01_EB, 7.496345E-03_EB, 1.380074E-01_EB, 3.724253E-04_EB, 1.238566E-03_EB, &
   7.511094E-01_EB, &
   7.999340E+01_EB, 3.572097E-01_EB, 1.449507E+00_EB, 3.819534E+00_EB, 3.386923E-03_EB, 2.420339E+00_EB, &
   1.689780E+00_EB, &
   8.000000E+01_EB, 4.078442E-01_EB, 2.746087E+01_EB, 5.496911E+00_EB, 7.681082E-02_EB, 9.142768E+00_EB, &
   9.256224E-01_EB, &
   8.000000E+01_EB, 1.120101E+00_EB, 2.154457E-01_EB, 2.496980E-01_EB, 5.035481E-03_EB, 4.567797E+00_EB, &
   1.613476E+01_EB, &
   8.000000E+01_EB, 2.505163E-01_EB, 7.289726E+00_EB, 3.688168E-01_EB, 1.262311E-01_EB, 4.852373E+01_EB, &
   4.327932E+01_EB, &
   8.000000E+01_EB, 4.925207E-01_EB, 1.508704E+00_EB, 1.040179E+00_EB, 6.046273E-01_EB, 2.007314E+01_EB, &
   3.883409E+01_EB/),(/7,8/))

GAMMAD5_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 1750-1925 CM-1
   8.000000E+01_EB, 1.703895E+00_EB, 9.191951E+00_EB, 2.709061E+00_EB, 9.359203E-01_EB, 3.834274E+01_EB, &
   7.078759E-01_EB, &
   7.999983E+01_EB, 1.728338E-01_EB, 6.428302E-01_EB, 6.480871E-01_EB, 2.403467E-01_EB, 4.865268E+01_EB, &
   1.451594E+01_EB, &
   7.999999E+01_EB, 7.815232E-02_EB, 8.730432E-02_EB, 2.873412E-01_EB, 2.837574E-02_EB, 3.159028E+00_EB, &
   3.214034E+00_EB, &
   6.362515E-02_EB, 6.189950E-03_EB, 5.670266E-03_EB, 1.609798E-02_EB, 4.919947E-04_EB, 1.569334E-01_EB, &
   1.400792E-01_EB, &
   4.873130E-03_EB, 6.913205E-03_EB, 1.797921E-04_EB, 1.336251E-02_EB, 8.086060E-03_EB, 1.891584E-02_EB, &
   2.687108E-02_EB, &
   2.972716E+01_EB, 4.496649E-02_EB, 1.698626E-02_EB, 7.124485E-02_EB, 3.445376E-01_EB, 2.327743E-01_EB, &
   6.610266E-01_EB, &
   7.398272E+00_EB, 4.698677E-02_EB, 1.308793E-02_EB, 3.824701E-02_EB, 6.967306E-02_EB, 1.066862E-03_EB, &
   8.184279E-04_EB, &
   9.082357E-03_EB, 3.297959E-03_EB, 4.284375E-04_EB, 1.604779E-03_EB, 5.294588E-04_EB, 6.058801E-04_EB, &
   5.178529E-04_EB/),(/7,8/))

GAMMAD5_C5H8O2(1:7,17:18) = RESHAPE((/ &  ! 1950-1975 CM-1
   7.219679E-04_EB, 5.013825E-04_EB, 4.812947E-04_EB, 5.404546E-04_EB, 5.475653E-04_EB, 4.887417E-04_EB, &
   4.941073E-04_EB, &
   3.735791E-04_EB, 3.732776E-04_EB, 3.745026E-04_EB, 3.743404E-04_EB, 3.741743E-04_EB, 3.727339E-04_EB, &
   3.713571E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(SD6_C5H8O2(N_TEMP_C5H8O2,26))

! BAND #6: 2650 CM-1 - 3275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 11.3243 %

SD6_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 2650-2825 CM-1
   6.617679E-04_EB, 7.038166E-04_EB, 5.823242E-04_EB, 6.498762E-04_EB, 6.305493E-04_EB, 6.702982E-04_EB, &
   2.492250E-02_EB, &
   8.413829E-04_EB, 9.287068E-04_EB, 9.382552E-04_EB, 8.391021E-04_EB, 8.961598E-04_EB, 8.253163E-04_EB, &
   5.284118E-02_EB, &
   9.964645E-03_EB, 3.825227E-03_EB, 5.258108E-03_EB, 3.430243E-03_EB, 8.600200E-04_EB, 6.425559E-03_EB, &
   7.468932E-02_EB, &
   4.273055E-02_EB, 1.049000E-02_EB, 1.562321E-02_EB, 1.351817E-02_EB, 7.806216E-03_EB, 2.714191E-02_EB, &
   1.221449E-01_EB, &
   8.007875E-02_EB, 2.912543E-02_EB, 3.379663E-02_EB, 3.252406E-02_EB, 1.200427E-02_EB, 2.850065E-02_EB, &
   1.385424E-01_EB, &
   9.973924E-02_EB, 2.113446E-02_EB, 3.076355E-02_EB, 2.639593E-02_EB, 6.644964E-03_EB, 2.242511E-02_EB, &
   1.660444E-01_EB, &
   1.028910E-01_EB, 5.220431E-02_EB, 4.300251E-02_EB, 4.795173E-02_EB, 3.143119E-02_EB, 6.642097E-02_EB, &
   1.979488E-01_EB, &
   2.202342E-01_EB, 1.345745E-01_EB, 1.259318E-01_EB, 1.271074E-01_EB, 1.238737E-01_EB, 1.520918E-01_EB, &
   2.525918E-01_EB/),(/7,8/))

SD6_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 2850-3025 CM-1
   1.136213E+00_EB, 4.781891E-01_EB, 4.249774E-01_EB, 3.828941E-01_EB, 3.367838E-01_EB, 2.329863E-01_EB, &
   2.717789E-01_EB, &
   8.634086E-01_EB, 4.219181E-01_EB, 3.728310E-01_EB, 3.478836E-01_EB, 2.675079E-01_EB, 2.733947E-01_EB, &
   3.309081E-01_EB, &
   1.146984E+00_EB, 5.316939E-01_EB, 5.062337E-01_EB, 4.805370E-01_EB, 3.860685E-01_EB, 3.811075E-01_EB, &
   4.090385E-01_EB, &
   2.056434E+00_EB, 9.985803E-01_EB, 9.030141E-01_EB, 8.563341E-01_EB, 6.732570E-01_EB, 6.443301E-01_EB, &
   5.434919E-01_EB, &
   4.615588E+00_EB, 2.244759E+00_EB, 2.030504E+00_EB, 1.928204E+00_EB, 1.523456E+00_EB, 1.179250E+00_EB, &
   7.021791E-01_EB, &
   5.650492E+00_EB, 2.472471E+00_EB, 2.136749E+00_EB, 1.930516E+00_EB, 1.442591E+00_EB, 9.997534E-01_EB, &
   5.978827E-01_EB, &
   4.609191E+00_EB, 1.782318E+00_EB, 1.525555E+00_EB, 1.347833E+00_EB, 9.229266E-01_EB, 6.392344E-01_EB, &
   4.770802E-01_EB, &
   2.820790E+00_EB, 1.104460E+00_EB, 9.776057E-01_EB, 8.631595E-01_EB, 5.721892E-01_EB, 4.422352E-01_EB, &
   3.522752E-01_EB/),(/7,8/))

SD6_C5H8O2(1:7,17:24) = RESHAPE((/ &  ! 3050-3225 CM-1
   1.187353E+00_EB, 5.460448E-01_EB, 4.928400E-01_EB, 4.472267E-01_EB, 3.395637E-01_EB, 2.809475E-01_EB, &
   2.653517E-01_EB, &
   4.704535E-01_EB, 2.491073E-01_EB, 2.537560E-01_EB, 2.409873E-01_EB, 1.982473E-01_EB, 2.279718E-01_EB, &
   2.611720E-01_EB, &
   6.039772E-01_EB, 3.431934E-01_EB, 3.284617E-01_EB, 3.129996E-01_EB, 2.653921E-01_EB, 2.526170E-01_EB, &
   2.011828E-01_EB, &
   5.920773E-01_EB, 2.451875E-01_EB, 2.368206E-01_EB, 2.072703E-01_EB, 2.138315E-01_EB, 1.682738E-01_EB, &
   1.552495E-01_EB, &
   1.700742E-01_EB, 7.290921E-02_EB, 8.217440E-02_EB, 6.629481E-02_EB, 5.469629E-02_EB, 8.994315E-02_EB, &
   9.732156E-02_EB, &
   1.118992E-01_EB, 4.366084E-02_EB, 4.890966E-02_EB, 2.885185E-02_EB, 1.279711E-02_EB, 6.143421E-02_EB, &
   5.932504E-02_EB, &
   5.320012E-02_EB, 1.864948E-02_EB, 1.833215E-02_EB, 8.372681E-03_EB, 9.556242E-04_EB, 2.724051E-02_EB, &
   4.060600E-02_EB, &
   1.562049E-02_EB, 9.377158E-03_EB, 1.509016E-03_EB, 2.400877E-03_EB, 7.719203E-04_EB, 4.833588E-03_EB, &
   3.145233E-03_EB/),(/7,8/))

SD6_C5H8O2(1:7,25:26) = RESHAPE((/ &  ! 3250-3275 CM-1
   1.234745E-03_EB, 1.894526E-03_EB, 7.790315E-04_EB, 6.688577E-04_EB, 6.556824E-04_EB, 6.274349E-04_EB, &
   6.008281E-04_EB, &
   4.683959E-04_EB, 4.682485E-04_EB, 4.694920E-04_EB, 4.693301E-04_EB, 4.691565E-04_EB, 4.675221E-04_EB, &
   4.659141E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD6_C5H8O2(N_TEMP_C5H8O2,26))

! BAND #6: 2650 CM-1 - 3275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 11.3243 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD6_C5H8O2(1:7,1:8) = RESHAPE((/ &  ! 2650-2825 CM-1
   5.202784E-04_EB, 5.525656E-04_EB, 4.434412E-04_EB, 4.998255E-04_EB, 4.935419E-04_EB, 5.063850E-04_EB, &
   1.092982E-02_EB, &
   5.613238E-04_EB, 6.098207E-04_EB, 7.086604E-04_EB, 6.039142E-04_EB, 6.289973E-04_EB, 6.322323E-04_EB, &
   1.194772E+00_EB, &
   2.420843E-03_EB, 1.761218E-03_EB, 3.068546E-03_EB, 3.918268E-03_EB, 6.421563E-04_EB, 2.061317E-03_EB, &
   4.004519E-01_EB, &
   8.686816E-04_EB, 5.690975E-03_EB, 1.368498E-01_EB, 2.276203E-01_EB, 1.896994E-03_EB, 1.942151E-03_EB, &
   1.781156E+00_EB, &
   8.775885E-02_EB, 2.065817E-02_EB, 1.307324E-01_EB, 4.176628E-02_EB, 1.462109E-02_EB, 2.246227E-03_EB, &
   1.423947E+00_EB, &
   1.400226E-03_EB, 7.669502E-03_EB, 2.007661E-01_EB, 5.908664E-02_EB, 9.431682E-02_EB, 1.143177E-02_EB, &
   1.718195E+00_EB, &
   1.151669E-01_EB, 1.850972E-02_EB, 1.407721E-01_EB, 1.023565E-01_EB, 3.200614E-02_EB, 1.001680E-01_EB, &
   2.390609E+00_EB, &
   6.151837E+00_EB, 6.526566E-02_EB, 8.225336E-01_EB, 7.234620E-01_EB, 1.810297E-02_EB, 6.092961E-01_EB, &
   6.507145E+00_EB/),(/7,8/))

GAMMAD6_C5H8O2(1:7,9:16) = RESHAPE((/ &  ! 2850-3025 CM-1
   7.999998E+01_EB, 4.402301E-01_EB, 8.184655E+00_EB, 7.049003E+00_EB, 2.465981E-02_EB, 4.666996E-01_EB, &
   1.050980E+01_EB, &
   8.000000E+01_EB, 1.344989E-01_EB, 5.238427E+00_EB, 2.779348E+00_EB, 5.844913E-02_EB, 2.117354E-01_EB, &
   4.746041E+00_EB, &
   7.999999E+01_EB, 1.706723E+00_EB, 1.717115E+01_EB, 7.059894E+00_EB, 8.912304E-02_EB, 2.663843E-01_EB, &
   1.132361E+01_EB, &
   7.999995E+01_EB, 4.086512E-01_EB, 4.839686E+01_EB, 2.308238E+01_EB, 2.068307E-01_EB, 8.659772E+00_EB, &
   1.089381E+00_EB, &
   8.000000E+01_EB, 7.844302E-01_EB, 6.546691E+01_EB, 3.816010E+00_EB, 3.501117E-01_EB, 4.860275E+01_EB, &
   2.192674E+01_EB, &
   8.000000E+01_EB, 7.277374E-01_EB, 6.563413E+01_EB, 6.128970E+01_EB, 2.795272E-01_EB, 4.863825E+01_EB, &
   1.715815E+01_EB, &
   8.000000E+01_EB, 7.372699E-01_EB, 6.564590E+01_EB, 5.533248E+01_EB, 2.910770E-01_EB, 9.688164E+00_EB, &
   2.950219E+01_EB, &
   8.000000E+01_EB, 5.982188E-01_EB, 6.554359E+01_EB, 3.587869E+01_EB, 3.437982E-01_EB, 2.068358E+00_EB, &
   2.249578E+01_EB/),(/7,8/))

GAMMAD6_C5H8O2(1:7,17:24) = RESHAPE((/ &  ! 3050-3225 CM-1
   8.000000E+01_EB, 4.235551E-01_EB, 2.497231E+01_EB, 9.302322E+00_EB, 8.931722E-02_EB, 5.150630E+00_EB, &
   1.414719E+01_EB, &
   6.010851E+01_EB, 1.879884E+00_EB, 2.783897E+00_EB, 3.173126E+00_EB, 8.402729E-02_EB, 3.467785E+00_EB, &
   1.176859E+01_EB, &
   7.499988E+01_EB, 1.744287E-01_EB, 2.563845E+00_EB, 4.952657E+00_EB, 6.326135E-02_EB, 3.738165E+00_EB, &
   8.586505E+00_EB, &
   7.999518E+01_EB, 3.410346E-01_EB, 4.666935E-01_EB, 2.152000E+00_EB, 7.943992E-03_EB, 9.823071E-01_EB, &
   3.227657E+00_EB, &
   6.470047E-01_EB, 2.696803E-01_EB, 4.931105E-02_EB, 1.777548E-01_EB, 6.548798E-03_EB, 4.997383E-02_EB, &
   2.368120E+00_EB, &
   1.850193E-02_EB, 2.343943E-01_EB, 2.130779E-02_EB, 1.645282E-01_EB, 5.228779E-03_EB, 1.024445E-01_EB, &
   3.884470E-01_EB, &
   7.638521E-03_EB, 1.078876E-01_EB, 1.679201E-02_EB, 8.467687E-03_EB, 6.767255E-04_EB, 1.893862E-01_EB, &
   1.745921E-01_EB, &
   9.543385E-03_EB, 5.137643E-04_EB, 8.369605E-04_EB, 1.442443E-03_EB, 5.585763E-04_EB, 6.679236E-03_EB, &
   1.904481E-03_EB/),(/7,8/))

GAMMAD6_C5H8O2(1:7,25:26) = RESHAPE((/ &  ! 3250-3275 CM-1
   4.938200E-04_EB, 5.333619E-04_EB, 5.374406E-04_EB, 5.112562E-04_EB, 5.068363E-04_EB, 4.776250E-04_EB, &
   4.658301E-04_EB, &
   3.735791E-04_EB, 3.732776E-04_EB, 3.745026E-04_EB, 3.743404E-04_EB, 3.741743E-04_EB, 3.727339E-04_EB, &
   3.713571E-04_EB/),(/7,2/))

!-------------------------PROPANE DATA-------------------


! THERE ARE 2 BANDS FOR PROPANE

! BAND #1: 1175 CM-1 - 1675 CM-1
! BAND #2: 2550 CM-1 - 3375 CM-1

ALLOCATE(SD_C3H8_TEMP(N_TEMP_C3H8))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR PROPANE ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_C3H8(N_BAND_C3H8,3))
ALLOCATE(BE_C3H8(N_BAND_C3H8))

OM_BND_C3H8 = RESHAPE((/ &
   1175._EB, 2550._EB, &
   1675._EB, 3375._EB, &
   25._EB, 25._EB/),(/N_BAND_C3H8,3/))

SD_C3H8_TEMP = (/ &
   295._EB, 396._EB, 435._EB, 513._EB, 578._EB, 790._EB,&
   1009._EB/)

BE_C3H8 = (/ &
   1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_C3H8(N_TEMP_C3H8,21))

! BAND #1: 1175 CM-1 - 1675 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.76397 %

SD1_C3H8(1:7,1:8) = RESHAPE((/ &  ! 1175-1350 CM-1
   2.856475E-02_EB, 1.979289E-02_EB, 1.707802E-02_EB, 1.759506E-02_EB, 5.992940E-04_EB, 6.107200E-04_EB, &
   6.336991E-04_EB, &
   1.484719E-02_EB, 1.365281E-02_EB, 1.474969E-02_EB, 1.061569E+00_EB, 1.945253E-03_EB, 1.757434E-03_EB, &
   4.662084E-03_EB, &
   6.963764E-04_EB, 8.500329E-04_EB, 1.107148E-02_EB, 4.507160E-01_EB, 1.264992E-03_EB, 2.619320E-03_EB, &
   7.816939E-01_EB, &
   7.226016E-04_EB, 9.075238E-04_EB, 1.125338E-02_EB, 2.194527E+00_EB, 3.638673E-03_EB, 7.038039E-03_EB, &
   8.347445E-03_EB, &
   7.167376E-04_EB, 2.529382E-03_EB, 1.471445E-02_EB, 5.689394E-01_EB, 5.129735E-03_EB, 3.199170E-01_EB, &
   4.714861E-02_EB, &
   1.973741E-02_EB, 3.150872E-02_EB, 4.659984E-02_EB, 9.243516E-01_EB, 5.980704E-02_EB, 5.046688E-01_EB, &
   2.447019E+00_EB, &
   1.692269E-01_EB, 1.375209E-01_EB, 1.214367E-01_EB, 4.508625E-01_EB, 8.448186E-02_EB, 1.539041E-01_EB, &
   2.379877E+00_EB, &
   3.080345E-01_EB, 2.355289E-01_EB, 2.170087E-01_EB, 2.766257E-01_EB, 1.745191E-01_EB, 3.225558E-01_EB, &
   2.527352E-01_EB/),(/7,8/))

SD1_C3H8(1:7,9:16) = RESHAPE((/ &  ! 1375-1550 CM-1
   5.345587E-01_EB, 3.413403E-01_EB, 2.898386E-01_EB, 4.388584E-01_EB, 1.966052E-01_EB, 2.312638E-01_EB, &
   1.022691E+00_EB, &
   4.993827E-01_EB, 3.759253E-01_EB, 3.463795E-01_EB, 4.015651E-01_EB, 2.676596E-01_EB, 2.708667E-01_EB, &
   5.754884E-01_EB, &
   3.709223E-01_EB, 3.014857E-01_EB, 3.021999E-01_EB, 9.358461E-01_EB, 2.624617E-01_EB, 2.519600E-01_EB, &
   3.362995E-01_EB, &
   8.886871E-01_EB, 5.747695E-01_EB, 5.132261E-01_EB, 6.330728E-01_EB, 3.344818E-01_EB, 2.389406E-01_EB, &
   1.497565E+00_EB, &
   1.197095E+00_EB, 6.989701E-01_EB, 6.042183E-01_EB, 8.699088E-01_EB, 3.814121E-01_EB, 2.349714E-01_EB, &
   2.153495E-01_EB, &
   6.712891E-01_EB, 3.823312E-01_EB, 3.697114E-01_EB, 8.758687E-01_EB, 2.396388E-01_EB, 1.894728E-01_EB, &
   1.039864E+00_EB, &
   1.822239E-01_EB, 1.242199E-01_EB, 1.389102E-01_EB, 1.652805E+00_EB, 1.094348E-01_EB, 9.737249E-02_EB, &
   6.800781E-02_EB, &
   1.683063E-02_EB, 9.880390E-03_EB, 3.283471E-02_EB, 4.035590E+00_EB, 3.286196E-02_EB, 7.512521E-02_EB, &
   1.698967E+00_EB/),(/7,8/))

SD1_C3H8(1:7,17:21) = RESHAPE((/ &  ! 1575-1675 CM-1
   4.515698E-03_EB, 1.676975E-02_EB, 9.512287E-03_EB, 1.896304E+00_EB, 1.105889E-02_EB, 3.093088E-01_EB, &
   1.288482E+00_EB, &
   9.765260E-04_EB, 6.840034E-01_EB, 1.034854E-02_EB, 3.287022E-03_EB, 7.493779E-03_EB, 2.811245E-01_EB, &
   4.924827E-02_EB, &
   6.957537E-04_EB, 2.871557E-03_EB, 6.957746E-04_EB, 8.150887E-04_EB, 1.138683E-03_EB, 4.575742E-03_EB, &
   1.445285E-02_EB, &
   5.346547E-04_EB, 6.489883E-04_EB, 5.962380E-04_EB, 5.749846E-04_EB, 5.575030E-04_EB, 6.786480E-04_EB, &
   6.619676E-04_EB, &
   4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, &
   4.634348E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_C3H8(N_TEMP_C3H8,21))

! BAND #1: 1175 CM-1 - 1675 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.76397 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_C3H8(1:7,1:8) = RESHAPE((/ &  ! 1175-1350 CM-1
   1.992031E-02_EB, 1.141762E-02_EB, 1.366155E-01_EB, 1.973038E-02_EB, 4.249794E-04_EB, 4.342393E-04_EB, &
   4.468799E-04_EB, &
   8.577876E-03_EB, 6.212098E-03_EB, 1.781483E-01_EB, 9.603555E-05_EB, 4.057086E-03_EB, 1.500925E-04_EB, &
   6.183171E-04_EB, &
   4.465589E-04_EB, 5.682785E-04_EB, 2.127940E-01_EB, 5.658019E-05_EB, 2.225962E-04_EB, 3.326833E-04_EB, &
   5.709512E-06_EB, &
   5.278412E-04_EB, 5.926586E-04_EB, 2.442104E-01_EB, 3.443200E-05_EB, 3.645364E-04_EB, 1.452036E-03_EB, &
   7.864828E-02_EB, &
   5.090058E-04_EB, 2.365350E-03_EB, 6.569843E-01_EB, 4.564938E-05_EB, 6.851756E-03_EB, 3.833493E-05_EB, &
   6.658254E-01_EB, &
   1.552597E-02_EB, 2.378523E-02_EB, 8.041370E-01_EB, 1.425241E-04_EB, 5.497853E-03_EB, 2.341884E-04_EB, &
   8.855023E-05_EB, &
   1.117621E-01_EB, 2.577861E-02_EB, 1.116130E+00_EB, 1.259624E-03_EB, 1.821509E-01_EB, 3.535192E-03_EB, &
   8.671143E-05_EB, &
   1.808233E-01_EB, 6.199510E-02_EB, 4.856790E+00_EB, 1.205230E-02_EB, 3.669786E-02_EB, 2.932383E-03_EB, &
   4.381239E-03_EB/),(/7,8/))

GAMMAD1_C3H8(1:7,9:16) = RESHAPE((/ &  ! 1375-1550 CM-1
   2.669661E-01_EB, 1.045895E-01_EB, 1.300781E+00_EB, 9.033578E-03_EB, 2.323335E-01_EB, 1.276771E-02_EB, &
   8.417636E-04_EB, &
   3.073214E-01_EB, 1.882989E-01_EB, 1.392323E+01_EB, 2.514566E-02_EB, 1.753049E-01_EB, 1.668277E-02_EB, &
   1.564119E-03_EB, &
   1.834951E-01_EB, 1.904052E+00_EB, 1.048436E+01_EB, 3.971260E-03_EB, 8.256188E-02_EB, 1.231218E-02_EB, &
   4.278796E-03_EB, &
   4.067671E-01_EB, 3.098373E-01_EB, 1.683329E+01_EB, 2.321737E-02_EB, 4.469660E+00_EB, 9.982721E-02_EB, &
   5.564305E-04_EB, &
   4.351927E-01_EB, 3.169913E-01_EB, 3.797676E-01_EB, 1.482328E-02_EB, 1.277859E-01_EB, 9.112234E-02_EB, &
   7.464685E-03_EB, &
   3.429172E-01_EB, 1.903809E+01_EB, 1.993602E+01_EB, 4.962290E-03_EB, 2.386361E-01_EB, 1.458725E-02_EB, &
   3.066693E-04_EB, &
   1.761592E-01_EB, 6.137305E+00_EB, 2.041270E+00_EB, 3.861036E-04_EB, 1.018629E+00_EB, 1.406376E-02_EB, &
   1.289657E-02_EB, &
   2.874081E-02_EB, 4.620945E-01_EB, 1.299484E+00_EB, 3.273246E-05_EB, 8.688233E-02_EB, 4.030533E-04_EB, &
   2.700077E-05_EB/),(/7,8/))

GAMMAD1_C3H8(1:7,17:21) = RESHAPE((/ &  ! 1575-1675 CM-1
   1.208990E-03_EB, 6.799487E-03_EB, 3.701547E-01_EB, 1.237888E-05_EB, 1.464532E-02_EB, 2.595773E-05_EB, &
   4.545433E-05_EB, &
   5.930652E-04_EB, 2.514086E-05_EB, 6.469106E-03_EB, 2.000295E-03_EB, 3.669583E-03_EB, 3.705261E-05_EB, &
   6.244904E-03_EB, &
   4.994954E-04_EB, 9.490064E-04_EB, 4.754384E-04_EB, 5.579883E-04_EB, 9.783514E-04_EB, 1.672579E-03_EB, &
   1.298467E-04_EB, &
   4.041578E-04_EB, 4.765211E-04_EB, 4.390729E-04_EB, 4.199440E-04_EB, 4.075633E-04_EB, 4.679761E-04_EB, &
   4.596989E-04_EB, &
   3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, &
   3.684034E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_C3H8(N_TEMP_C3H8,34))

! BAND #2: 2550 CM-1 - 3375 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.8902 %

SD2_C3H8(1:7,1:8) = RESHAPE((/ &  ! 2550-2725 CM-1
   5.503672E-04_EB, 5.994891E-04_EB, 8.662153E-04_EB, 6.007139E-04_EB, 5.508955E-04_EB, 6.062568E-04_EB, &
   9.609866E-04_EB, &
   9.165832E-03_EB, 6.514927E-03_EB, 5.498817E-03_EB, 2.820723E-03_EB, 4.346969E-03_EB, 3.366148E-03_EB, &
   1.036676E-01_EB, &
   3.816653E-02_EB, 2.423984E-02_EB, 2.297132E-02_EB, 1.131018E-02_EB, 1.290251E-02_EB, 9.585606E-03_EB, &
   7.230381E-01_EB, &
   6.229937E-02_EB, 3.989116E-02_EB, 3.343177E-02_EB, 2.820251E-02_EB, 1.841367E-02_EB, 1.058259E-02_EB, &
   1.643570E-02_EB, &
   8.226034E-02_EB, 4.582033E-02_EB, 3.980542E-02_EB, 2.801891E-02_EB, 1.700839E-02_EB, 1.139598E-02_EB, &
   2.021613E-01_EB, &
   7.640614E-02_EB, 3.373610E-02_EB, 3.023935E-02_EB, 2.980953E-02_EB, 2.095503E-02_EB, 1.799163E-02_EB, &
   3.412770E-01_EB, &
   3.340838E-02_EB, 2.504949E-02_EB, 2.777922E-02_EB, 2.817541E-02_EB, 3.213869E-02_EB, 3.140060E-02_EB, &
   1.470802E+00_EB, &
   8.685682E-02_EB, 6.091073E-02_EB, 7.104290E-02_EB, 7.142576E-02_EB, 6.557492E-02_EB, 6.248472E-02_EB, &
   2.113951E+00_EB/),(/7,8/))

SD2_C3H8(1:7,9:16) = RESHAPE((/ &  ! 2750-2925 CM-1
   1.699556E-01_EB, 1.209358E-01_EB, 1.186218E-01_EB, 1.132869E-01_EB, 1.150490E-01_EB, 1.001152E-01_EB, &
   1.402520E-01_EB, &
   1.666365E-01_EB, 1.410150E-01_EB, 1.367199E-01_EB, 1.403357E-01_EB, 1.627667E-01_EB, 1.378065E-01_EB, &
   3.019538E-01_EB, &
   2.031290E-01_EB, 1.630134E-01_EB, 1.593693E-01_EB, 1.694187E-01_EB, 1.815236E-01_EB, 2.026379E-01_EB, &
   2.932861E-01_EB, &
   3.542177E-01_EB, 3.249511E-01_EB, 3.280019E-01_EB, 3.450738E-01_EB, 3.407162E-01_EB, 3.576253E-01_EB, &
   5.198445E-01_EB, &
   1.000630E+00_EB, 9.246602E-01_EB, 9.305942E-01_EB, 9.189938E-01_EB, 8.909863E-01_EB, 7.810609E-01_EB, &
   9.521582E-01_EB, &
   3.753627E+00_EB, 2.683351E+00_EB, 2.484390E+00_EB, 2.041213E+00_EB, 1.810422E+00_EB, 1.239899E+00_EB, &
   1.183325E+00_EB, &
   5.234499E+00_EB, 3.770775E+00_EB, 3.516223E+00_EB, 2.940932E+00_EB, 2.625379E+00_EB, 1.827482E+00_EB, &
   1.807152E+00_EB, &
   5.572566E+00_EB, 4.685164E+00_EB, 4.488098E+00_EB, 3.945955E+00_EB, 3.562082E+00_EB, 2.450846E+00_EB, &
   2.095864E+00_EB/),(/7,8/))

SD2_C3H8(1:7,17:24) = RESHAPE((/ &  ! 2950-3125 CM-1
   1.055304E+01_EB, 7.617962E+00_EB, 7.020237E+00_EB, 5.655513E+00_EB, 4.924307E+00_EB, 2.987908E+00_EB, &
   2.490259E+00_EB, &
   1.309229E+01_EB, 8.614162E+00_EB, 7.735378E+00_EB, 5.958736E+00_EB, 5.015764E+00_EB, 2.751758E+00_EB, &
   2.176924E+00_EB, &
   5.664769E+00_EB, 4.285205E+00_EB, 3.984661E+00_EB, 3.391957E+00_EB, 2.931808E+00_EB, 1.747475E+00_EB, &
   1.598403E+00_EB, &
   7.428057E-01_EB, 8.846817E-01_EB, 9.107382E-01_EB, 1.009605E+00_EB, 9.708428E-01_EB, 7.720345E-01_EB, &
   9.265547E-01_EB, &
   7.186847E-02_EB, 1.458443E-01_EB, 1.735809E-01_EB, 2.489900E-01_EB, 2.701093E-01_EB, 3.154699E-01_EB, &
   6.322075E-01_EB, &
   4.223055E-02_EB, 7.998122E-02_EB, 1.014124E-01_EB, 1.476562E-01_EB, 1.444812E-01_EB, 1.778681E-01_EB, &
   7.948678E-01_EB, &
   5.301805E-02_EB, 7.789503E-02_EB, 9.507427E-02_EB, 1.304625E-01_EB, 1.288331E-01_EB, 1.398066E-01_EB, &
   9.260605E-01_EB, &
   8.229182E-02_EB, 7.916035E-02_EB, 9.156865E-02_EB, 1.229671E-01_EB, 1.154257E-01_EB, 1.069746E-01_EB, &
   1.474222E+00_EB/),(/7,8/))

SD2_C3H8(1:7,25:32) = RESHAPE((/ &  ! 3150-3325 CM-1
   1.097023E-01_EB, 9.711792E-02_EB, 1.018166E-01_EB, 1.145716E-01_EB, 1.629561E-01_EB, 7.870503E-02_EB, &
   1.054101E+00_EB, &
   1.528366E-01_EB, 1.042414E-01_EB, 1.045964E-01_EB, 9.557668E-02_EB, 9.257045E-02_EB, 5.716098E-02_EB, &
   2.230393E+00_EB, &
   1.448776E-01_EB, 8.693362E-02_EB, 8.143001E-02_EB, 8.822165E-02_EB, 7.154741E-02_EB, 3.844782E-02_EB, &
   5.459571E+00_EB, &
   7.739671E-02_EB, 4.498684E-02_EB, 5.166537E-02_EB, 4.834891E-02_EB, 4.653514E-02_EB, 1.900636E-02_EB, &
   3.668670E+00_EB, &
   2.193118E-02_EB, 1.636816E-02_EB, 1.616765E-02_EB, 2.883862E-02_EB, 6.071064E-02_EB, 3.305292E-03_EB, &
   4.440979E+00_EB, &
   7.607103E-04_EB, 2.864772E-03_EB, 1.569824E-03_EB, 5.600370E-03_EB, 2.231541E-03_EB, 1.431965E-03_EB, &
   4.371189E+00_EB, &
   6.767298E-04_EB, 7.265903E-04_EB, 6.896754E-04_EB, 7.704584E-04_EB, 7.843659E-04_EB, 2.492767E-03_EB, &
   3.017731E+00_EB, &
   6.753685E-04_EB, 8.082467E-04_EB, 8.419705E-04_EB, 8.502049E-04_EB, 7.333137E-04_EB, 2.689703E-03_EB, &
   2.546579E-01_EB/),(/7,8/))

SD2_C3H8(1:7,33:34) = RESHAPE((/ &  ! 3350-3375 CM-1
   5.953124E-04_EB, 6.647555E-04_EB, 6.295086E-04_EB, 6.270639E-04_EB, 5.605951E-04_EB, 1.004627E-03_EB, &
   5.997674E-04_EB, &
   4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, 4.634348E-04_EB, &
   4.634348E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_C3H8(N_TEMP_C3H8,34))

! BAND #2: 2550 CM-1 - 3375 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.8902 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_C3H8(1:7,1:8) = RESHAPE((/ &  ! 2550-2725 CM-1
   4.136492E-04_EB, 4.360494E-04_EB, 7.953702E-04_EB, 4.660382E-04_EB, 4.005291E-04_EB, 4.511091E-04_EB, &
   6.072020E-04_EB, &
   2.759961E-03_EB, 1.062067E-01_EB, 1.215285E-02_EB, 2.640006E-03_EB, 3.656880E-03_EB, 1.938678E-03_EB, &
   5.487963E-05_EB, &
   1.584875E-02_EB, 2.833951E-01_EB, 3.264932E-01_EB, 1.739356E-02_EB, 5.773845E-03_EB, 4.395147E-02_EB, &
   1.322632E-05_EB, &
   1.193671E-01_EB, 2.854658E-01_EB, 4.644661E-02_EB, 2.765612E-02_EB, 1.053011E-02_EB, 2.667963E-01_EB, &
   5.817185E-03_EB, &
   2.623899E-02_EB, 1.875377E-01_EB, 5.240968E-02_EB, 2.355271E-02_EB, 7.559665E-02_EB, 2.293658E-01_EB, &
   7.216244E-05_EB, &
   5.846743E-03_EB, 2.105678E-01_EB, 2.701368E-01_EB, 2.713083E-02_EB, 1.035818E-02_EB, 3.749967E-02_EB, &
   8.197520E-05_EB, &
   1.507777E-02_EB, 2.373195E-01_EB, 2.803885E-01_EB, 2.381271E-02_EB, 2.374969E-03_EB, 3.007283E-01_EB, &
   3.941570E-05_EB, &
   5.460024E-02_EB, 2.148020E-01_EB, 1.102276E-02_EB, 2.048424E-02_EB, 1.381473E-02_EB, 2.265009E-01_EB, &
   1.113763E-04_EB/),(/7,8/))

GAMMAD2_C3H8(1:7,9:16) = RESHAPE((/ &  ! 2750-2925 CM-1
   7.609446E-02_EB, 6.650037E-01_EB, 3.402647E-01_EB, 7.088071E-02_EB, 1.528056E-02_EB, 1.623423E+00_EB, &
   1.555020E-02_EB, &
   1.068152E-01_EB, 3.599354E-01_EB, 9.327433E-01_EB, 8.061623E-02_EB, 1.044707E-02_EB, 1.207247E+00_EB, &
   6.148015E-03_EB, &
   1.347777E-01_EB, 1.366370E-01_EB, 1.464407E+00_EB, 8.162167E-02_EB, 3.748354E-02_EB, 6.697765E+00_EB, &
   5.715348E-02_EB, &
   1.784972E-01_EB, 1.113695E-01_EB, 1.456847E-01_EB, 8.124496E-02_EB, 8.031435E-02_EB, 3.755111E+01_EB, &
   4.466462E-02_EB, &
   3.769734E-01_EB, 1.378926E-01_EB, 2.066963E-01_EB, 1.647811E-01_EB, 1.988552E-01_EB, 4.888631E+01_EB, &
   6.856421E-02_EB, &
   1.209828E+00_EB, 4.483110E-01_EB, 4.819971E-01_EB, 3.919982E-01_EB, 3.657716E-01_EB, 4.888633E+01_EB, &
   1.716140E-01_EB, &
   1.907010E+00_EB, 6.962287E-01_EB, 6.668651E-01_EB, 5.363387E-01_EB, 5.417758E-01_EB, 4.888614E+01_EB, &
   1.732383E-01_EB, &
   2.035309E+00_EB, 8.005331E-01_EB, 8.689003E-01_EB, 6.922331E-01_EB, 7.737400E-01_EB, 4.888562E+01_EB, &
   3.566562E-01_EB/),(/7,8/))

GAMMAD2_C3H8(1:7,17:24) = RESHAPE((/ &  ! 2950-3125 CM-1
   3.521357E+00_EB, 1.169922E+00_EB, 1.245608E+00_EB, 9.879083E-01_EB, 9.805019E-01_EB, 4.888630E+01_EB, &
   2.805866E-01_EB, &
   8.194584E+00_EB, 1.715767E+00_EB, 1.407065E+00_EB, 1.043498E+00_EB, 9.327292E-01_EB, 4.888633E+01_EB, &
   2.729909E-01_EB, &
   9.096142E-01_EB, 9.519580E-01_EB, 6.722470E-01_EB, 5.099619E-01_EB, 4.841140E-01_EB, 4.888628E+01_EB, &
   2.097093E-01_EB, &
   1.758692E-01_EB, 8.005995E-01_EB, 3.028875E-01_EB, 1.771971E-01_EB, 1.723475E-01_EB, 4.888622E+01_EB, &
   9.485115E-02_EB, &
   1.552016E-02_EB, 1.106460E+00_EB, 3.508781E-01_EB, 8.578778E-02_EB, 7.481873E-02_EB, 2.362585E+01_EB, &
   2.190639E-02_EB, &
   8.141419E-03_EB, 2.936549E-01_EB, 1.051647E-01_EB, 3.583341E-02_EB, 8.451454E-02_EB, 4.015433E+00_EB, &
   4.672611E-03_EB, &
   6.825195E-03_EB, 2.860617E-01_EB, 9.805913E-02_EB, 3.356674E-02_EB, 3.562043E-02_EB, 1.354576E+00_EB, &
   2.504862E-03_EB, &
   1.731937E-02_EB, 8.103158E-02_EB, 3.656235E-02_EB, 1.650823E-02_EB, 1.716846E-02_EB, 1.104930E+00_EB, &
   1.025944E-03_EB/),(/7,8/))

GAMMAD2_C3H8(1:7,25:32) = RESHAPE((/ &  ! 3150-3325 CM-1
   2.761892E-02_EB, 5.927558E-02_EB, 6.812671E-02_EB, 1.536547E-02_EB, 2.866389E-03_EB, 2.743296E-01_EB, &
   8.973476E-04_EB, &
   4.898929E-02_EB, 2.217375E-01_EB, 1.147533E-01_EB, 5.759265E-02_EB, 1.511087E-02_EB, 9.820712E-02_EB, &
   2.297806E-04_EB, &
   1.075239E-02_EB, 9.660209E-02_EB, 7.236142E-02_EB, 1.745200E-02_EB, 1.115524E-02_EB, 1.145421E-01_EB, &
   4.285123E-05_EB, &
   1.476159E-02_EB, 2.024331E-01_EB, 7.181537E-03_EB, 4.721618E-02_EB, 3.344976E-03_EB, 1.461756E+00_EB, &
   3.907613E-05_EB, &
   7.912117E-03_EB, 2.082529E-02_EB, 6.432685E-03_EB, 1.913236E-02_EB, 7.216503E-05_EB, 1.281409E-03_EB, &
   1.568408E-05_EB, &
   5.304482E-04_EB, 1.965226E-03_EB, 8.885309E-04_EB, 1.080202E-02_EB, 6.214399E-04_EB, 2.290412E-04_EB, &
   1.111330E-05_EB, &
   4.741388E-04_EB, 5.040738E-04_EB, 4.522965E-04_EB, 5.090295E-04_EB, 4.634938E-04_EB, 3.121426E-03_EB, &
   1.265450E-05_EB, &
   4.771043E-04_EB, 5.489571E-04_EB, 5.406306E-04_EB, 6.198226E-04_EB, 4.746614E-04_EB, 1.079570E-03_EB, &
   2.377068E-05_EB/),(/7,8/))

GAMMAD2_C3H8(1:7,33:34) = RESHAPE((/ &  ! 3350-3375 CM-1
   4.399730E-04_EB, 5.039732E-04_EB, 4.232400E-04_EB, 4.772033E-04_EB, 4.129961E-04_EB, 5.016760E-04_EB, &
   4.444099E-04_EB, &
   3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, 3.684034E-04_EB, &
   3.684034E-04_EB/),(/7,2/))

!-------------------------PROPYLENE DATA-------------------


! THERE ARE 3 BANDS FOR PROPYLENE

! BAND #1: 775 CM-1 - 1150 CM-1
! BAND #2: 1225 CM-1 - 1975 CM-1
! BAND #3: 2650 CM-1 - 3275 CM-1

ALLOCATE(SD_C3H6_TEMP(N_TEMP_C3H6))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR PROPYLENE ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_C3H6(N_BAND_C3H6,3))
ALLOCATE(BE_C3H6(N_BAND_C3H6))

OM_BND_C3H6 = RESHAPE((/ &
   775._EB, 1225._EB, 2650._EB, &
   1150._EB, 1975._EB, 3275._EB, &
   5._EB, 25._EB, 25._EB/),(/N_BAND_C3H6,3/))

SD_C3H6_TEMP = (/ &
   296._EB, 390._EB, 444._EB, 491._EB, 594._EB, 793._EB,&
   1003._EB/)

BE_C3H6 = (/ &
   1.000_EB, 1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_C3H6(N_TEMP_C3H6,68))

! BAND #1: 775 CM-1 - 1150 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.3448 %

SD1_C3H6(1:7,1:8) = RESHAPE((/ &  ! 775-810 CM-1
   2.921820E-04_EB, 1.885331E-04_EB, 1.578479E-04_EB, 1.195361E-04_EB, 1.447279E-04_EB, 1.698242E-04_EB, &
   3.510553E-04_EB, &
   5.786383E-04_EB, 4.698237E-04_EB, 3.471753E-04_EB, 3.368251E-04_EB, 1.561044E-04_EB, 2.151079E-04_EB, &
   4.880960E-04_EB, &
   5.146434E-04_EB, 4.282047E-04_EB, 3.953252E-04_EB, 3.606827E-04_EB, 3.453987E-04_EB, 3.100075E-04_EB, &
   5.373971E-04_EB, &
   1.002033E-03_EB, 3.754947E-04_EB, 3.830171E-04_EB, 3.397481E-04_EB, 4.382293E-03_EB, 1.980601E-04_EB, &
   4.996406E-04_EB, &
   5.974180E-04_EB, 4.192863E-04_EB, 2.153249E-04_EB, 1.486372E-03_EB, 3.085406E-02_EB, 4.786688E-03_EB, &
   4.055803E-04_EB, &
   3.186779E-02_EB, 1.217059E-02_EB, 8.908478E-03_EB, 1.366154E-02_EB, 1.604611E-01_EB, 1.277554E-02_EB, &
   3.868758E-04_EB, &
   2.287860E-02_EB, 1.347562E-02_EB, 1.761538E-02_EB, 1.670141E-02_EB, 3.006989E-01_EB, 1.352943E-02_EB, &
   4.562671E-04_EB, &
   3.847401E-01_EB, 1.311464E-01_EB, 1.806964E-02_EB, 2.138006E-02_EB, 6.075345E-01_EB, 2.375582E-02_EB, &
   4.382283E-04_EB/),(/7,8/))

SD1_C3H6(1:7,9:16) = RESHAPE((/ &  ! 815-850 CM-1
   3.612608E-01_EB, 3.310333E-02_EB, 2.135533E-02_EB, 6.703345E-02_EB, 3.383135E-01_EB, 3.802107E-02_EB, &
   1.684642E-03_EB, &
   4.242283E-01_EB, 3.658435E-02_EB, 3.805267E-02_EB, 5.077876E-02_EB, 3.413346E-01_EB, 6.513409E-02_EB, &
   1.587704E-04_EB, &
   1.644787E-01_EB, 5.568292E-02_EB, 5.486572E-02_EB, 6.891789E-02_EB, 1.332364E-01_EB, 6.265865E-02_EB, &
   4.187924E-03_EB, &
   7.882443E-02_EB, 7.779061E-02_EB, 7.929456E-02_EB, 1.096667E-01_EB, 1.347277E-01_EB, 9.669708E-02_EB, &
   2.546892E-02_EB, &
   9.754850E-02_EB, 1.076154E-01_EB, 1.043330E-01_EB, 1.111524E-01_EB, 1.883759E-01_EB, 1.329948E-01_EB, &
   5.845340E-02_EB, &
   1.538045E-01_EB, 1.691789E-01_EB, 1.519359E-01_EB, 1.496250E-01_EB, 2.160998E-01_EB, 1.631159E-01_EB, &
   9.210403E-02_EB, &
   2.041607E-01_EB, 2.369334E-01_EB, 1.944714E-01_EB, 2.025469E-01_EB, 2.541597E-01_EB, 1.848773E-01_EB, &
   1.157566E-01_EB, &
   2.823440E-01_EB, 3.067145E-01_EB, 2.545703E-01_EB, 3.082214E-01_EB, 3.115871E-01_EB, 2.118905E-01_EB, &
   1.367771E-01_EB/),(/7,8/))

SD1_C3H6(1:7,17:24) = RESHAPE((/ &  ! 855-890 CM-1
   3.934069E-01_EB, 3.902770E-01_EB, 3.324288E-01_EB, 3.655227E-01_EB, 4.852182E-01_EB, 2.278497E-01_EB, &
   1.418258E-01_EB, &
   5.098671E-01_EB, 4.707600E-01_EB, 3.953366E-01_EB, 4.123636E-01_EB, 5.023609E-01_EB, 3.093041E-01_EB, &
   2.172109E-01_EB, &
   6.861408E-01_EB, 5.874598E-01_EB, 4.823945E-01_EB, 4.930154E-01_EB, 5.112207E-01_EB, 3.097694E-01_EB, &
   1.945893E-01_EB, &
   8.621571E-01_EB, 7.069132E-01_EB, 5.626431E-01_EB, 5.778790E-01_EB, 6.253640E-01_EB, 4.501915E-01_EB, &
   1.364634E-01_EB, &
   1.117672E+00_EB, 8.597340E-01_EB, 6.702957E-01_EB, 6.824590E-01_EB, 6.721777E-01_EB, 4.204596E-01_EB, &
   2.336953E-01_EB, &
   1.296785E+00_EB, 9.660400E-01_EB, 7.508644E-01_EB, 7.657156E-01_EB, 6.694906E-01_EB, 4.150445E-01_EB, &
   2.912663E-01_EB, &
   1.612024E+00_EB, 1.132199E+00_EB, 8.825466E-01_EB, 8.439805E-01_EB, 7.306116E-01_EB, 4.492608E-01_EB, &
   2.586394E-01_EB, &
   1.769679E+00_EB, 1.240089E+00_EB, 9.826761E-01_EB, 9.126471E-01_EB, 7.901846E-01_EB, 4.958428E-01_EB, &
   2.975731E-01_EB/),(/7,8/))

SD1_C3H6(1:7,25:32) = RESHAPE((/ &  ! 895-930 CM-1
   2.035152E+00_EB, 1.372164E+00_EB, 1.084943E+00_EB, 9.759957E-01_EB, 8.670922E-01_EB, 5.511628E-01_EB, &
   5.051868E-01_EB, &
   2.110903E+00_EB, 1.411810E+00_EB, 1.107883E+00_EB, 1.032360E+00_EB, 8.891386E-01_EB, 6.607676E-01_EB, &
   4.105057E-01_EB, &
   2.241734E+00_EB, 1.509671E+00_EB, 1.226689E+00_EB, 1.120333E+00_EB, 1.019009E+00_EB, 6.960237E-01_EB, &
   6.300682E-01_EB, &
   4.690927E+00_EB, 3.301112E+00_EB, 2.617543E+00_EB, 2.271022E+00_EB, 1.868783E+00_EB, 1.104123E+00_EB, &
   6.791524E-01_EB, &
   3.538563E+00_EB, 2.596313E+00_EB, 2.151089E+00_EB, 1.982625E+00_EB, 1.674390E+00_EB, 1.121363E+00_EB, &
   6.319589E-01_EB, &
   2.529174E+00_EB, 1.690171E+00_EB, 1.371477E+00_EB, 1.268574E+00_EB, 1.124475E+00_EB, 7.538665E-01_EB, &
   4.997968E-01_EB, &
   2.442382E+00_EB, 1.621890E+00_EB, 1.292151E+00_EB, 1.185851E+00_EB, 1.045526E+00_EB, 6.515910E-01_EB, &
   4.350142E-01_EB, &
   2.359003E+00_EB, 1.574882E+00_EB, 1.255634E+00_EB, 1.148501E+00_EB, 1.005751E+00_EB, 6.742696E-01_EB, &
   4.032512E-01_EB/),(/7,8/))

SD1_C3H6(1:7,33:40) = RESHAPE((/ &  ! 935-970 CM-1
   2.292826E+00_EB, 1.573504E+00_EB, 1.257992E+00_EB, 1.148404E+00_EB, 9.819016E-01_EB, 6.786328E-01_EB, &
   3.987199E-01_EB, &
   2.256277E+00_EB, 1.614009E+00_EB, 1.308043E+00_EB, 1.170636E+00_EB, 1.005170E+00_EB, 6.614059E-01_EB, &
   3.824164E-01_EB, &
   2.128388E+00_EB, 1.579098E+00_EB, 1.312822E+00_EB, 1.165846E+00_EB, 1.012764E+00_EB, 6.750425E-01_EB, &
   4.478263E-01_EB, &
   2.002676E+00_EB, 1.570187E+00_EB, 1.337531E+00_EB, 1.201187E+00_EB, 1.060851E+00_EB, 6.651784E-01_EB, &
   5.518473E-01_EB, &
   1.769536E+00_EB, 1.476601E+00_EB, 1.281675E+00_EB, 1.168120E+00_EB, 1.059359E+00_EB, 6.557750E-01_EB, &
   4.612030E-01_EB, &
   1.513007E+00_EB, 1.314704E+00_EB, 1.154175E+00_EB, 1.078702E+00_EB, 9.708818E-01_EB, 7.985200E-01_EB, &
   4.095132E-01_EB, &
   1.300819E+00_EB, 1.112939E+00_EB, 9.729632E-01_EB, 9.305722E-01_EB, 9.107075E-01_EB, 6.817647E-01_EB, &
   3.869006E-01_EB, &
   1.122524E+00_EB, 9.240820E-01_EB, 7.997622E-01_EB, 8.045075E-01_EB, 8.225817E-01_EB, 5.935068E-01_EB, &
   4.076619E-01_EB/),(/7,8/))

SD1_C3H6(1:7,41:48) = RESHAPE((/ &  ! 975-1010 CM-1
   1.090787E+00_EB, 8.705917E-01_EB, 7.392297E-01_EB, 7.356904E-01_EB, 7.567491E-01_EB, 5.944416E-01_EB, &
   3.882052E-01_EB, &
   1.097680E+00_EB, 8.498178E-01_EB, 7.146480E-01_EB, 7.049228E-01_EB, 7.300468E-01_EB, 5.868185E-01_EB, &
   4.235053E-01_EB, &
   1.175184E+00_EB, 9.551431E-01_EB, 8.117959E-01_EB, 7.958744E-01_EB, 7.668238E-01_EB, 5.942202E-01_EB, &
   4.784101E-01_EB, &
   2.200917E+00_EB, 1.489225E+00_EB, 1.189617E+00_EB, 1.045546E+00_EB, 8.888219E-01_EB, 6.345618E-01_EB, &
   3.753116E-01_EB, &
   1.026273E+00_EB, 8.138120E-01_EB, 7.081956E-01_EB, 6.489008E-01_EB, 5.962778E-01_EB, 4.346981E-01_EB, &
   3.144834E-01_EB, &
   8.545948E-01_EB, 6.501326E-01_EB, 5.807415E-01_EB, 5.228005E-01_EB, 5.069268E-01_EB, 3.587344E-01_EB, &
   2.535124E-01_EB, &
   8.182404E-01_EB, 6.020895E-01_EB, 5.272828E-01_EB, 4.843306E-01_EB, 4.757196E-01_EB, 3.642910E-01_EB, &
   2.463011E-01_EB, &
   7.867336E-01_EB, 5.627888E-01_EB, 4.859257E-01_EB, 4.529062E-01_EB, 4.373186E-01_EB, 2.999539E-01_EB, &
   2.195520E-01_EB/),(/7,8/))

SD1_C3H6(1:7,49:56) = RESHAPE((/ &  ! 1015-1050 CM-1
   7.370735E-01_EB, 5.261814E-01_EB, 4.365308E-01_EB, 4.196529E-01_EB, 3.794978E-01_EB, 3.100276E-01_EB, &
   2.316864E-01_EB, &
   6.545905E-01_EB, 4.754812E-01_EB, 3.879959E-01_EB, 3.680691E-01_EB, 3.486950E-01_EB, 2.537942E-01_EB, &
   2.192666E-01_EB, &
   5.636151E-01_EB, 4.220184E-01_EB, 3.368352E-01_EB, 3.294973E-01_EB, 3.480874E-01_EB, 2.278502E-01_EB, &
   1.435619E-01_EB, &
   4.794628E-01_EB, 3.716607E-01_EB, 2.986875E-01_EB, 2.920226E-01_EB, 2.849022E-01_EB, 2.098157E-01_EB, &
   1.086582E-01_EB, &
   3.987250E-01_EB, 3.273405E-01_EB, 2.706509E-01_EB, 2.488392E-01_EB, 2.796603E-01_EB, 1.937969E-01_EB, &
   1.318512E-01_EB, &
   3.618884E-01_EB, 3.043278E-01_EB, 2.752540E-01_EB, 2.348242E-01_EB, 2.449431E-01_EB, 1.840862E-01_EB, &
   1.365370E-01_EB, &
   4.327196E-01_EB, 3.118666E-01_EB, 2.824791E-01_EB, 2.285493E-01_EB, 2.297279E-01_EB, 1.660089E-01_EB, &
   6.776279E-02_EB, &
   2.158610E-01_EB, 1.711786E-01_EB, 1.890422E-01_EB, 1.508251E-01_EB, 1.710093E-01_EB, 1.189088E-01_EB, &
   3.722276E-01_EB/),(/7,8/))

SD1_C3H6(1:7,57:64) = RESHAPE((/ &  ! 1055-1090 CM-1
   1.808366E-01_EB, 1.285694E-01_EB, 1.347601E-01_EB, 1.159091E-01_EB, 1.328783E-01_EB, 9.806016E-02_EB, &
   4.385238E-02_EB, &
   1.688521E-01_EB, 1.098959E-01_EB, 9.863147E-02_EB, 1.007307E-01_EB, 1.218736E-01_EB, 1.096671E-01_EB, &
   3.048481E-02_EB, &
   1.509152E-01_EB, 9.784628E-02_EB, 8.122169E-02_EB, 7.571922E-02_EB, 8.082920E-02_EB, 6.716860E-02_EB, &
   6.805299E-01_EB, &
   1.236023E-01_EB, 8.420975E-02_EB, 6.449748E-02_EB, 5.721581E-02_EB, 8.699266E-02_EB, 4.794369E-02_EB, &
   5.275249E-02_EB, &
   1.005482E-01_EB, 7.333190E-02_EB, 5.246391E-02_EB, 4.854230E-02_EB, 8.581595E-02_EB, 3.857051E-02_EB, &
   1.047535E+00_EB, &
   8.140574E-02_EB, 6.139425E-02_EB, 4.191643E-02_EB, 3.330029E-02_EB, 6.128860E-02_EB, 3.771955E-02_EB, &
   5.547067E-01_EB, &
   5.876207E-02_EB, 4.748412E-02_EB, 3.403509E-02_EB, 2.491123E-02_EB, 1.741583E-01_EB, 3.685622E-02_EB, &
   1.173988E+00_EB, &
   5.067589E-02_EB, 3.249100E-02_EB, 3.614176E-02_EB, 1.695154E-02_EB, 5.607839E-02_EB, 4.126856E-02_EB, &
   4.414252E-02_EB/),(/7,8/))

SD1_C3H6(1:7,65:68) = RESHAPE((/ &  ! 1095-1150 CM-1
   3.475507E-02_EB, 1.803126E-02_EB, 2.233876E-01_EB, 1.137397E-02_EB, 1.233139E-02_EB, 2.244437E-02_EB, &
   3.055494E-04_EB, &
   8.633613E-02_EB, 9.384641E-03_EB, 3.751848E-01_EB, 5.532895E-03_EB, 6.183470E-03_EB, 7.005080E-03_EB, &
   5.030021E-04_EB, &
   4.004646E-04_EB, 1.996236E-04_EB, 3.332406E-04_EB, 1.767862E-04_EB, 1.467998E-04_EB, 1.391831E-04_EB, &
   3.497050E-04_EB, &
   1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, &
   1.027560E-04_EB/),(/7,4/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_C3H6(N_TEMP_C3H6,68))

! BAND #1: 775 CM-1 - 1150 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.3448 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_C3H6(1:7,1:8) = RESHAPE((/ &  ! 775-810 CM-1
   1.865203E-04_EB, 1.481280E-04_EB, 1.121426E-04_EB, 8.618315E-05_EB, 9.939553E-05_EB, 1.228308E-04_EB, &
   2.571832E-04_EB, &
   2.784672E-04_EB, 2.629293E-04_EB, 2.613280E-04_EB, 2.166322E-04_EB, 1.060275E-04_EB, 1.512432E-04_EB, &
   3.364746E-04_EB, &
   3.584490E-04_EB, 2.571836E-04_EB, 2.937264E-04_EB, 2.336679E-04_EB, 2.672244E-04_EB, 2.128761E-04_EB, &
   2.711168E-04_EB, &
   1.050069E-03_EB, 2.060006E-04_EB, 2.640878E-04_EB, 2.179849E-04_EB, 2.169826E-04_EB, 1.354118E-04_EB, &
   2.786454E-04_EB, &
   5.631752E-04_EB, 2.967683E-04_EB, 1.474684E-04_EB, 6.104105E-04_EB, 4.865310E-04_EB, 1.882053E-01_EB, &
   2.951885E-04_EB, &
   8.382309E-05_EB, 1.999192E-03_EB, 2.110707E-01_EB, 2.657106E-03_EB, 4.279171E-04_EB, 8.580741E-01_EB, &
   2.380492E-04_EB, &
   4.133686E-04_EB, 2.280390E-03_EB, 8.751086E-01_EB, 3.295157E-03_EB, 5.879263E-04_EB, 8.275870E-01_EB, &
   2.164440E-04_EB, &
   6.177217E-05_EB, 2.052275E-04_EB, 1.783767E+00_EB, 1.180004E-02_EB, 4.686960E-04_EB, 2.654668E+00_EB, &
   2.743441E-04_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,9:16) = RESHAPE((/ &  ! 815-850 CM-1
   1.362853E-04_EB, 9.486143E-03_EB, 2.766580E+00_EB, 2.523733E-03_EB, 1.516289E-03_EB, 8.166695E+00_EB, &
   1.693790E-01_EB, &
   2.370331E-04_EB, 1.279314E+00_EB, 1.307893E+01_EB, 1.368493E-01_EB, 2.662950E-03_EB, 9.028076E+00_EB, &
   1.501266E-04_EB, &
   1.605909E-03_EB, 3.888338E+00_EB, 3.291564E+01_EB, 2.430327E+00_EB, 1.840118E-02_EB, 4.830532E+00_EB, &
   1.460557E-01_EB, &
   1.751757E-02_EB, 7.388926E+00_EB, 6.531922E+01_EB, 9.190465E-02_EB, 7.887271E-02_EB, 7.702450E+00_EB, &
   3.065075E+00_EB, &
   4.005618E-02_EB, 5.793342E+00_EB, 6.531359E+01_EB, 3.481985E+01_EB, 7.477765E-02_EB, 1.306801E+01_EB, &
   3.031110E+01_EB, &
   4.670461E-02_EB, 1.511945E-01_EB, 6.531029E+01_EB, 5.058553E+01_EB, 1.162566E-01_EB, 1.593096E+01_EB, &
   4.345904E+01_EB, &
   6.963672E-02_EB, 1.180935E-01_EB, 6.530489E+01_EB, 8.198354E-01_EB, 1.298311E-01_EB, 4.887623E+01_EB, &
   4.345787E+01_EB, &
   1.032951E-01_EB, 1.651780E-01_EB, 6.527527E+01_EB, 1.583085E-01_EB, 1.439537E-01_EB, 4.887639E+01_EB, &
   4.345745E+01_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,17:24) = RESHAPE((/ &  ! 855-890 CM-1
   1.554114E-01_EB, 2.463349E-01_EB, 1.552737E+00_EB, 2.489268E-01_EB, 7.175075E-02_EB, 4.877618E+01_EB, &
   4.344856E+01_EB, &
   2.216480E-01_EB, 3.911085E-01_EB, 6.528132E+01_EB, 4.847826E-01_EB, 1.055938E-01_EB, 7.890568E-01_EB, &
   4.345834E+01_EB, &
   3.344937E-01_EB, 5.937512E-01_EB, 6.440865E+01_EB, 5.685496E-01_EB, 2.154384E-01_EB, 4.887601E+01_EB, &
   4.345865E+01_EB, &
   4.563788E-01_EB, 6.544545E-01_EB, 6.527427E+01_EB, 5.645359E-01_EB, 1.603139E-01_EB, 2.220849E-01_EB, &
   4.345957E+01_EB, &
   6.971662E-01_EB, 8.419195E-01_EB, 6.602209E+00_EB, 6.185587E-01_EB, 2.185881E-01_EB, 1.026546E+00_EB, &
   4.345911E+01_EB, &
   9.177302E-01_EB, 8.698788E-01_EB, 4.219718E+00_EB, 5.938638E-01_EB, 3.550496E-01_EB, 4.887400E+01_EB, &
   4.345951E+01_EB, &
   1.024886E+00_EB, 8.931182E-01_EB, 1.906388E+00_EB, 7.671322E-01_EB, 4.049625E-01_EB, 4.886302E+01_EB, &
   4.345956E+01_EB, &
   1.138590E+00_EB, 8.055441E-01_EB, 1.266173E+00_EB, 7.247405E-01_EB, 4.313090E-01_EB, 2.888925E+00_EB, &
   4.345947E+01_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,25:32) = RESHAPE((/ &  ! 895-930 CM-1
   1.258896E+00_EB, 8.879635E-01_EB, 1.183716E+00_EB, 8.697144E-01_EB, 3.994743E-01_EB, 1.020734E+00_EB, &
   1.469295E-01_EB, &
   1.489911E+00_EB, 9.449031E-01_EB, 1.364975E+00_EB, 7.600728E-01_EB, 4.690280E-01_EB, 5.532825E-01_EB, &
   2.320330E+01_EB, &
   1.791634E+00_EB, 1.068040E+00_EB, 1.344772E+00_EB, 1.025605E+00_EB, 5.570787E-01_EB, 4.887568E+01_EB, &
   1.866191E-01_EB, &
   6.582008E-01_EB, 6.151496E-01_EB, 7.320593E-01_EB, 7.674633E-01_EB, 5.984986E-01_EB, 1.256776E+00_EB, &
   1.505021E+00_EB, &
   8.496429E-01_EB, 6.647729E-01_EB, 8.138054E-01_EB, 6.996761E-01_EB, 6.263920E-01_EB, 7.887495E-01_EB, &
   4.345957E+01_EB, &
   2.585850E+00_EB, 1.312862E+00_EB, 1.669178E+00_EB, 1.005278E+00_EB, 5.484994E-01_EB, 1.645085E+00_EB, &
   4.301288E+01_EB, &
   2.304432E+00_EB, 1.337031E+00_EB, 1.903523E+00_EB, 1.080752E+00_EB, 4.772959E-01_EB, 4.887638E+01_EB, &
   3.954397E+01_EB, &
   2.284405E+00_EB, 1.252628E+00_EB, 1.761317E+00_EB, 1.081162E+00_EB, 5.126407E-01_EB, 9.460859E-01_EB, &
   4.345934E+01_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,33:40) = RESHAPE((/ &  ! 935-970 CM-1
   2.011991E+00_EB, 1.160491E+00_EB, 1.606206E+00_EB, 1.044964E+00_EB, 6.166918E-01_EB, 8.541470E-01_EB, &
   4.345947E+01_EB, &
   1.951894E+00_EB, 1.077050E+00_EB, 1.340221E+00_EB, 1.005762E+00_EB, 5.377001E-01_EB, 1.813695E+00_EB, &
   4.345915E+01_EB, &
   1.862497E+00_EB, 1.110545E+00_EB, 1.226215E+00_EB, 1.004332E+00_EB, 5.640741E-01_EB, 1.260546E+00_EB, &
   2.061631E+01_EB, &
   1.735085E+00_EB, 1.152792E+00_EB, 1.235271E+00_EB, 1.029092E+00_EB, 5.304049E-01_EB, 3.364825E+00_EB, &
   4.109795E-01_EB, &
   1.514333E+00_EB, 1.096526E+00_EB, 1.190673E+00_EB, 9.738878E-01_EB, 4.991815E-01_EB, 2.598609E+01_EB, &
   4.345955E+01_EB, &
   1.323596E+00_EB, 1.020536E+00_EB, 1.276590E+00_EB, 9.362680E-01_EB, 5.648663E-01_EB, 3.325554E-01_EB, &
   4.345900E+01_EB, &
   1.090160E+00_EB, 8.522171E-01_EB, 1.383531E+00_EB, 8.696999E-01_EB, 3.985636E-01_EB, 6.142456E-01_EB, &
   4.345952E+01_EB, &
   9.568718E-01_EB, 7.965634E-01_EB, 2.061656E+00_EB, 6.937459E-01_EB, 3.185588E-01_EB, 1.043529E+00_EB, &
   4.345956E+01_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,41:48) = RESHAPE((/ &  ! 975-1010 CM-1
   8.780028E-01_EB, 7.171719E-01_EB, 1.797413E+00_EB, 7.191605E-01_EB, 3.123087E-01_EB, 5.370041E-01_EB, &
   6.031057E-01_EB, &
   7.683322E-01_EB, 6.697888E-01_EB, 1.564974E+00_EB, 6.922381E-01_EB, 2.866901E-01_EB, 6.100381E-01_EB, &
   5.296325E-01_EB, &
   7.710645E-01_EB, 6.686977E-01_EB, 1.316990E+00_EB, 7.356423E-01_EB, 3.962548E-01_EB, 7.103715E-01_EB, &
   3.121685E-01_EB, &
   7.152139E-01_EB, 7.181107E-01_EB, 9.272079E-01_EB, 7.998523E-01_EB, 4.476511E-01_EB, 4.403214E-01_EB, &
   1.432304E+00_EB, &
   7.215480E-01_EB, 5.222809E-01_EB, 7.034660E-01_EB, 5.939143E-01_EB, 3.557110E-01_EB, 7.369583E+00_EB, &
   4.343647E+01_EB, &
   7.509675E-01_EB, 5.541920E-01_EB, 6.013325E-01_EB, 5.827632E-01_EB, 2.516345E-01_EB, 4.887235E+01_EB, &
   4.345657E+01_EB, &
   6.655689E-01_EB, 5.909646E-01_EB, 7.175681E-01_EB, 5.191324E-01_EB, 2.139999E-01_EB, 4.270810E-01_EB, &
   3.186474E+00_EB, &
   5.936443E-01_EB, 6.111940E-01_EB, 8.318915E-01_EB, 4.526249E-01_EB, 2.171504E-01_EB, 4.874889E+01_EB, &
   4.345936E+01_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,49:56) = RESHAPE((/ &  ! 1015-1050 CM-1
   5.108597E-01_EB, 5.449935E-01_EB, 1.467166E+00_EB, 4.175570E-01_EB, 2.800624E-01_EB, 2.807650E-01_EB, &
   1.226529E-01_EB, &
   4.396741E-01_EB, 4.212443E-01_EB, 1.991238E+00_EB, 4.853656E-01_EB, 1.994488E-01_EB, 3.606763E+00_EB, &
   1.195304E-01_EB, &
   3.521564E-01_EB, 3.419630E-01_EB, 9.696171E+00_EB, 3.944128E-01_EB, 1.086559E-01_EB, 4.887097E+01_EB, &
   4.345940E+01_EB, &
   2.826655E-01_EB, 2.596188E-01_EB, 1.720670E+00_EB, 3.339207E-01_EB, 1.507599E-01_EB, 1.947669E+00_EB, &
   4.345956E+01_EB, &
   2.103926E-01_EB, 1.918390E-01_EB, 5.187904E-01_EB, 5.571757E-01_EB, 8.852491E-02_EB, 1.040530E+00_EB, &
   4.345956E+01_EB, &
   2.190684E-01_EB, 2.166025E-01_EB, 2.770976E-01_EB, 6.790640E-01_EB, 1.398958E-01_EB, 4.887609E+01_EB, &
   1.273985E+01_EB, &
   2.372381E-01_EB, 2.099844E-01_EB, 1.691730E-01_EB, 3.370993E-01_EB, 9.039351E-02_EB, 5.560844E-01_EB, &
   3.729305E+01_EB, &
   1.338894E-01_EB, 2.270514E-01_EB, 8.598819E-02_EB, 2.237195E-01_EB, 4.915478E-02_EB, 3.404705E+01_EB, &
   2.634276E-03_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,57:64) = RESHAPE((/ &  ! 1055-1090 CM-1
   8.590987E-02_EB, 2.939130E-01_EB, 1.322061E-01_EB, 2.325564E-01_EB, 4.121415E-02_EB, 1.890644E+01_EB, &
   6.653137E+00_EB, &
   5.828702E-02_EB, 1.919845E-01_EB, 3.558396E+00_EB, 8.935135E-02_EB, 2.151974E-02_EB, 1.813839E-02_EB, &
   5.291639E+00_EB, &
   4.560095E-02_EB, 8.801874E-02_EB, 4.240838E+00_EB, 1.971226E-01_EB, 2.805033E-02_EB, 3.159379E-01_EB, &
   4.231868E-04_EB, &
   5.236543E-02_EB, 4.750499E-02_EB, 7.719299E+00_EB, 2.230632E+00_EB, 1.029862E-02_EB, 3.970081E+00_EB, &
   1.254647E-02_EB, &
   4.584112E-02_EB, 2.781221E-02_EB, 3.521327E+00_EB, 1.069332E-01_EB, 5.513817E-03_EB, 7.390687E-01_EB, &
   8.785510E-05_EB, &
   2.952585E-02_EB, 1.659069E-02_EB, 1.232312E+00_EB, 1.610279E+00_EB, 3.746132E-03_EB, 1.888989E-02_EB, &
   1.172808E-05_EB, &
   2.783898E-02_EB, 1.191524E-02_EB, 1.737513E-01_EB, 6.337914E-01_EB, 3.927326E-04_EB, 1.256543E-02_EB, &
   9.593758E-05_EB, &
   1.161773E-02_EB, 1.234249E-02_EB, 7.173288E-03_EB, 4.651250E-01_EB, 6.905694E-04_EB, 6.625808E-03_EB, &
   1.875822E-05_EB/),(/7,8/))

GAMMAD1_C3H6(1:7,65:68) = RESHAPE((/ &  ! 1095-1150 CM-1
   7.371375E-03_EB, 3.892307E-02_EB, 2.066948E-04_EB, 2.148055E-01_EB, 5.123043E-03_EB, 8.826795E-03_EB, &
   1.519386E-04_EB, &
   1.685293E-04_EB, 4.052437E-02_EB, 5.766893E-05_EB, 1.099394E-01_EB, 1.056916E-02_EB, 3.703491E-03_EB, &
   4.534644E-04_EB, &
   2.576720E-04_EB, 1.502775E-04_EB, 2.394291E-04_EB, 1.106949E-04_EB, 1.070129E-04_EB, 1.035473E-04_EB, &
   2.424308E-04_EB, &
   8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, &
   8.047518E-05_EB/),(/7,4/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_C3H6(N_TEMP_C3H6,31))

! BAND #2: 1225 CM-1 - 1975 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.89939 %

SD2_C3H6(1:7,1:8) = RESHAPE((/ &  ! 1225-1400 CM-1
   3.558545E-04_EB, 3.234771E-04_EB, 3.304950E-04_EB, 1.651299E-04_EB, 1.336215E-04_EB, 1.510735E-04_EB, &
   1.993574E-04_EB, &
   6.741050E-04_EB, 3.773009E-04_EB, 4.527377E-04_EB, 3.445857E-04_EB, 2.150432E-04_EB, 3.112911E-04_EB, &
   3.405452E-03_EB, &
   8.500802E-04_EB, 5.928536E-03_EB, 5.109505E-02_EB, 5.939970E-03_EB, 5.544562E-03_EB, 4.216340E-02_EB, &
   7.821472E-02_EB, &
   6.900284E-04_EB, 1.754929E-02_EB, 1.351477E-01_EB, 1.875618E-02_EB, 1.689271E-02_EB, 5.437608E-01_EB, &
   8.042160E-02_EB, &
   6.597844E-03_EB, 3.076774E-02_EB, 7.267186E-02_EB, 3.875256E-02_EB, 4.198566E-02_EB, 1.377841E-01_EB, &
   4.353146E-02_EB, &
   1.425388E-01_EB, 7.292994E-02_EB, 1.006956E-01_EB, 8.530846E-02_EB, 8.207286E-02_EB, 1.024217E-01_EB, &
   9.040341E-02_EB, &
   2.757070E-01_EB, 2.011514E-01_EB, 2.183300E-01_EB, 1.901067E-01_EB, 1.708424E-01_EB, 1.802992E-01_EB, &
   1.306124E-01_EB, &
   5.487648E-01_EB, 3.944423E-01_EB, 3.612309E-01_EB, 3.179025E-01_EB, 2.635343E-01_EB, 2.207712E-01_EB, &
   1.537228E-01_EB/),(/7,8/))

SD2_C3H6(1:7,9:16) = RESHAPE((/ &  ! 1425-1600 CM-1
   7.347914E-01_EB, 4.838524E-01_EB, 4.349597E-01_EB, 3.807440E-01_EB, 3.206784E-01_EB, 2.496395E-01_EB, &
   1.756546E-01_EB, &
   1.230079E+00_EB, 7.176913E-01_EB, 5.767768E-01_EB, 4.861231E-01_EB, 3.530715E-01_EB, 2.411707E-01_EB, &
   1.663700E-01_EB, &
   1.076204E+00_EB, 6.242017E-01_EB, 5.169408E-01_EB, 4.238453E-01_EB, 3.142849E-01_EB, 2.371514E-01_EB, &
   1.547102E-01_EB, &
   4.989020E-01_EB, 3.259007E-01_EB, 2.647985E-01_EB, 2.506597E-01_EB, 2.113725E-01_EB, 1.667162E-01_EB, &
   1.268622E-01_EB, &
   3.050060E-01_EB, 1.549283E-01_EB, 1.437705E-01_EB, 1.418842E-01_EB, 1.284408E-01_EB, 1.165257E-01_EB, &
   1.181179E-01_EB, &
   2.901797E-01_EB, 1.013263E-01_EB, 8.839726E-02_EB, 9.945842E-02_EB, 9.936424E-02_EB, 9.382087E-02_EB, &
   1.109026E-01_EB, &
   9.170595E-01_EB, 1.077723E-01_EB, 1.126792E-01_EB, 1.012963E-01_EB, 1.008830E-01_EB, 1.343878E-01_EB, &
   1.433742E-01_EB, &
   1.723540E+00_EB, 1.340775E-01_EB, 1.803447E-01_EB, 1.684967E-01_EB, 1.917971E-01_EB, 2.426751E-01_EB, &
   1.993900E-01_EB/),(/7,8/))

SD2_C3H6(1:7,17:24) = RESHAPE((/ &  ! 1625-1800 CM-1
   7.534624E-01_EB, 5.257745E-01_EB, 4.873627E-01_EB, 4.420277E-01_EB, 3.752945E-01_EB, 2.872981E-01_EB, &
   2.158678E-01_EB, &
   1.173360E+00_EB, 6.714563E-01_EB, 5.383806E-01_EB, 4.764335E-01_EB, 3.802518E-01_EB, 2.635373E-01_EB, &
   1.882440E-01_EB, &
   7.574636E-01_EB, 4.662984E-01_EB, 4.191013E-01_EB, 3.278277E-01_EB, 2.460639E-01_EB, 1.547873E-01_EB, &
   1.018707E-01_EB, &
   3.622467E-02_EB, 3.135298E-02_EB, 1.070575E-02_EB, 1.197498E-02_EB, 2.172120E-02_EB, 1.437764E-02_EB, &
   1.246169E-02_EB, &
   6.790247E-04_EB, 4.477004E-04_EB, 4.003496E-04_EB, 2.372342E-03_EB, 2.112353E-04_EB, 1.899291E-04_EB, &
   4.337606E-04_EB, &
   6.821738E-04_EB, 3.841181E-04_EB, 4.398296E-04_EB, 9.336168E-04_EB, 3.215112E-04_EB, 1.577836E-03_EB, &
   1.387750E-03_EB, &
   5.357513E-03_EB, 9.513971E-03_EB, 9.586573E-03_EB, 1.177708E-02_EB, 1.894358E-02_EB, 1.778693E-02_EB, &
   2.131572E-02_EB, &
   1.558459E-01_EB, 1.086936E-01_EB, 1.035246E-01_EB, 8.484628E-02_EB, 7.171917E-02_EB, 5.505700E-02_EB, &
   3.996848E-02_EB/),(/7,8/))

SD2_C3H6(1:7,25:31) = RESHAPE((/ &  ! 1825-1975 CM-1
   2.622036E-01_EB, 1.516085E-01_EB, 1.303312E-01_EB, 1.109721E-01_EB, 9.231520E-02_EB, 7.419268E-02_EB, &
   4.811877E-02_EB, &
   1.824093E-01_EB, 1.314846E-01_EB, 1.287875E-01_EB, 9.898525E-02_EB, 8.120577E-02_EB, 6.960379E-02_EB, &
   4.479529E-02_EB, &
   4.747023E-02_EB, 3.221598E-02_EB, 3.332202E-02_EB, 2.756103E-02_EB, 2.423984E-02_EB, 2.922304E-02_EB, &
   1.964140E-02_EB, &
   1.531867E-02_EB, 9.915829E-03_EB, 3.275042E-02_EB, 4.962254E-03_EB, 3.664153E-03_EB, 1.736865E-03_EB, &
   2.806612E-03_EB, &
   1.672196E-03_EB, 6.907827E-04_EB, 9.686969E-04_EB, 4.056228E-04_EB, 1.810253E-04_EB, 2.048415E-04_EB, &
   3.860056E-04_EB, &
   4.937814E-04_EB, 1.785822E-04_EB, 3.303675E-04_EB, 1.807431E-04_EB, 1.381312E-04_EB, 1.579684E-04_EB, &
   2.163472E-04_EB, &
   1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, &
   1.027560E-04_EB/),(/7,7/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_C3H6(N_TEMP_C3H6,31))

! BAND #2: 1225 CM-1 - 1975 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.89939 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_C3H6(1:7,1:8) = RESHAPE((/ &  ! 1225-1400 CM-1
   2.622715E-04_EB, 2.415333E-04_EB, 2.285633E-04_EB, 1.138182E-04_EB, 1.023912E-04_EB, 1.165998E-04_EB, &
   1.100001E-04_EB, &
   5.083865E-04_EB, 2.678443E-04_EB, 3.093172E-04_EB, 2.072846E-04_EB, 1.338274E-04_EB, 2.187271E-04_EB, &
   3.932176E-03_EB, &
   5.760161E-04_EB, 3.452147E-02_EB, 1.478839E-04_EB, 3.228596E-03_EB, 3.197521E-03_EB, 2.939038E-05_EB, &
   1.383337E-03_EB, &
   4.544833E-04_EB, 8.199799E-02_EB, 3.788188E-04_EB, 1.070417E-02_EB, 1.812632E-02_EB, 7.216282E-05_EB, &
   3.552942E-03_EB, &
   4.941401E-04_EB, 7.477210E-02_EB, 2.805114E-03_EB, 1.723863E-02_EB, 2.045615E-02_EB, 2.807456E-03_EB, &
   4.838154E+00_EB, &
   3.708619E-03_EB, 1.501171E-01_EB, 2.088233E-02_EB, 5.723047E-02_EB, 2.236513E-01_EB, 5.128140E-02_EB, &
   4.345874E+01_EB, &
   5.695155E-02_EB, 2.842757E-01_EB, 8.565045E-02_EB, 1.664632E-01_EB, 2.596688E-01_EB, 6.733220E-02_EB, &
   4.345955E+01_EB, &
   1.810625E-01_EB, 3.644158E-01_EB, 2.284041E-01_EB, 2.941351E-01_EB, 2.787627E-01_EB, 1.246260E-01_EB, &
   4.345946E+01_EB/),(/7,8/))

GAMMAD2_C3H6(1:7,9:16) = RESHAPE((/ &  ! 1425-1600 CM-1
   2.618227E-01_EB, 4.657918E-01_EB, 2.858630E-01_EB, 3.534168E-01_EB, 2.920606E-01_EB, 2.314596E-01_EB, &
   4.345293E+01_EB, &
   4.735753E-01_EB, 5.151092E-01_EB, 4.261774E-01_EB, 3.947653E-01_EB, 3.773149E-01_EB, 1.853547E-01_EB, &
   4.343606E+01_EB, &
   4.430151E-01_EB, 5.018903E-01_EB, 3.534603E-01_EB, 4.029514E-01_EB, 3.684172E-01_EB, 1.453836E-01_EB, &
   4.345940E+01_EB, &
   1.212938E-01_EB, 3.173309E-01_EB, 1.209401E+00_EB, 3.624133E-01_EB, 2.461181E-01_EB, 2.905692E-01_EB, &
   4.345952E+01_EB, &
   1.084288E-02_EB, 1.935743E-01_EB, 3.826811E-01_EB, 2.209504E-01_EB, 3.384784E-01_EB, 5.030323E+00_EB, &
   4.314517E+00_EB, &
   3.502530E-03_EB, 4.235158E-01_EB, 4.492198E+01_EB, 4.178424E-01_EB, 3.215097E-01_EB, 1.976012E+01_EB, &
   3.306259E+01_EB, &
   9.195743E-04_EB, 5.344729E-02_EB, 3.856228E-02_EB, 8.090926E-02_EB, 1.993074E-01_EB, 8.752817E-02_EB, &
   4.345936E+01_EB, &
   5.991440E-04_EB, 1.174575E-01_EB, 4.116301E-02_EB, 1.163800E-01_EB, 2.048843E-01_EB, 1.170091E-01_EB, &
   4.345936E+01_EB/),(/7,8/))

GAMMAD2_C3H6(1:7,17:24) = RESHAPE((/ &  ! 1625-1800 CM-1
   1.048318E-01_EB, 4.077170E-01_EB, 3.250763E-01_EB, 4.204265E-01_EB, 4.172624E-01_EB, 3.501140E-01_EB, &
   4.345754E+01_EB, &
   3.891245E-01_EB, 5.297782E-01_EB, 5.753653E-01_EB, 4.671660E-01_EB, 3.550334E-01_EB, 5.129497E-01_EB, &
   1.852430E+01_EB, &
   1.315391E-01_EB, 2.233168E-01_EB, 1.643737E-01_EB, 2.846889E-01_EB, 2.529524E-01_EB, 5.109478E-01_EB, &
   5.217994E+00_EB, &
   2.202200E-04_EB, 6.992095E-05_EB, 6.651429E-02_EB, 1.261036E-01_EB, 2.793432E-04_EB, 3.400065E-01_EB, &
   2.976967E-01_EB, &
   4.407006E-04_EB, 3.060631E-04_EB, 3.078173E-04_EB, 1.997468E-02_EB, 1.414429E-04_EB, 1.272939E-04_EB, &
   2.920596E-04_EB, &
   4.075042E-04_EB, 2.572692E-04_EB, 2.809507E-04_EB, 7.570268E-04_EB, 2.248097E-04_EB, 6.557878E-04_EB, &
   1.214879E-02_EB, &
   5.533543E-04_EB, 2.280793E-03_EB, 5.727351E-03_EB, 1.258039E-01_EB, 2.877495E-03_EB, 1.528211E-01_EB, &
   3.641746E-01_EB, &
   2.750324E-02_EB, 5.042429E-02_EB, 3.703191E-02_EB, 1.112856E-01_EB, 1.394191E-01_EB, 2.852003E-01_EB, &
   5.289596E+00_EB/),(/7,8/))

GAMMAD2_C3H6(1:7,25:31) = RESHAPE((/ &  ! 1825-1975 CM-1
   1.249123E-01_EB, 1.287617E-01_EB, 8.856076E-02_EB, 1.063268E-01_EB, 6.381303E-02_EB, 3.721067E-02_EB, &
   1.078497E+01_EB, &
   9.224931E-02_EB, 7.783664E-02_EB, 4.282341E-02_EB, 1.077656E-01_EB, 9.668966E-02_EB, 3.557773E-02_EB, &
   4.459730E+00_EB, &
   1.094300E-02_EB, 1.484660E-02_EB, 1.474169E-02_EB, 2.868794E-02_EB, 2.363462E-02_EB, 7.563694E-03_EB, &
   9.480691E-01_EB, &
   3.143676E-03_EB, 1.423552E-03_EB, 2.803890E-04_EB, 1.197387E-02_EB, 1.810420E-03_EB, 9.517693E-04_EB, &
   1.253779E-03_EB, &
   9.700043E-04_EB, 5.310996E-04_EB, 3.238508E-04_EB, 2.462185E-04_EB, 1.221073E-04_EB, 1.352439E-04_EB, &
   2.806980E-04_EB, &
   3.482794E-04_EB, 1.171460E-04_EB, 2.319252E-04_EB, 1.036437E-04_EB, 1.041061E-04_EB, 1.117176E-04_EB, &
   1.472074E-04_EB, &
   8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, &
   8.047518E-05_EB/),(/7,7/))

!---------------------------------------------------------------------------
ALLOCATE(SD3_C3H6(N_TEMP_C3H6,26))

! BAND #3: 2650 CM-1 - 3275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.84732 %

SD3_C3H6(1:7,1:8) = RESHAPE((/ &  ! 2650-2825 CM-1
   9.842396E-04_EB, 9.773581E-04_EB, 4.737803E-04_EB, 1.731125E-04_EB, 1.636141E-04_EB, 1.470685E-04_EB, &
   1.906099E-04_EB, &
   2.026987E-03_EB, 1.186057E-03_EB, 9.296802E-04_EB, 1.762232E-03_EB, 6.242290E-04_EB, 2.052893E-03_EB, &
   1.454807E-02_EB, &
   1.375163E-02_EB, 8.857038E-03_EB, 1.157316E-02_EB, 1.514073E-02_EB, 7.647375E-03_EB, 2.028955E-02_EB, &
   7.263667E-02_EB, &
   6.091687E-02_EB, 3.794720E-02_EB, 3.692778E-02_EB, 3.669497E-02_EB, 2.246508E-02_EB, 2.837078E-02_EB, &
   4.660372E-02_EB, &
   7.796068E-02_EB, 5.480320E-02_EB, 5.466178E-02_EB, 5.561425E-02_EB, 3.895555E-02_EB, 4.987873E-02_EB, &
   5.885558E-02_EB, &
   5.994425E-02_EB, 4.512610E-02_EB, 5.182436E-02_EB, 5.600091E-02_EB, 4.405629E-02_EB, 5.968023E-02_EB, &
   9.479119E-02_EB, &
   9.789297E-02_EB, 7.603902E-02_EB, 8.316338E-02_EB, 8.497910E-02_EB, 7.201492E-02_EB, 9.033387E-02_EB, &
   1.461816E-01_EB, &
   1.180654E-01_EB, 1.124261E-01_EB, 1.279063E-01_EB, 1.423319E-01_EB, 1.323416E-01_EB, 1.543846E-01_EB, &
   1.754013E-01_EB/),(/7,8/))

SD3_C3H6(1:7,9:16) = RESHAPE((/ &  ! 2850-3025 CM-1
   4.934323E-01_EB, 3.715076E-01_EB, 3.606179E-01_EB, 3.406474E-01_EB, 2.903159E-01_EB, 2.686381E-01_EB, &
   2.632060E-01_EB, &
   9.542159E-01_EB, 7.012988E-01_EB, 6.454962E-01_EB, 5.929394E-01_EB, 5.042463E-01_EB, 4.353109E-01_EB, &
   4.219635E-01_EB, &
   1.390515E+00_EB, 1.083207E+00_EB, 1.007094E+00_EB, 9.304473E-01_EB, 8.055580E-01_EB, 6.623870E-01_EB, &
   5.866959E-01_EB, &
   2.165889E+00_EB, 1.557403E+00_EB, 1.377848E+00_EB, 1.234664E+00_EB, 1.013037E+00_EB, 7.883431E-01_EB, &
   6.628056E-01_EB, &
   2.955798E+00_EB, 2.081509E+00_EB, 1.819270E+00_EB, 1.598535E+00_EB, 1.286353E+00_EB, 9.640864E-01_EB, &
   7.596691E-01_EB, &
   2.749633E+00_EB, 1.810398E+00_EB, 1.545370E+00_EB, 1.334588E+00_EB, 1.045362E+00_EB, 8.023541E-01_EB, &
   6.307507E-01_EB, &
   2.482580E+00_EB, 1.760628E+00_EB, 1.526249E+00_EB, 1.327667E+00_EB, 1.049659E+00_EB, 7.936791E-01_EB, &
   6.269325E-01_EB, &
   1.476540E+00_EB, 1.169856E+00_EB, 1.094611E+00_EB, 9.890179E-01_EB, 8.496338E-01_EB, 7.194488E-01_EB, &
   6.198264E-01_EB/),(/7,8/))

SD3_C3H6(1:7,17:24) = RESHAPE((/ &  ! 3050-3225 CM-1
   1.158768E+00_EB, 9.970946E-01_EB, 9.717818E-01_EB, 9.018919E-01_EB, 8.089673E-01_EB, 6.965923E-01_EB, &
   5.756752E-01_EB, &
   1.617699E+00_EB, 1.161452E+00_EB, 1.056867E+00_EB, 9.368936E-01_EB, 7.790857E-01_EB, 6.211223E-01_EB, &
   4.978033E-01_EB, &
   1.668409E+00_EB, 1.154363E+00_EB, 1.028890E+00_EB, 8.900880E-01_EB, 7.204607E-01_EB, 5.621992E-01_EB, &
   4.193038E-01_EB, &
   6.627374E-01_EB, 5.511257E-01_EB, 5.408762E-01_EB, 4.741726E-01_EB, 4.065990E-01_EB, 3.500514E-01_EB, &
   2.603275E-01_EB, &
   1.943149E-01_EB, 1.812675E-01_EB, 1.997641E-01_EB, 1.725390E-01_EB, 1.584383E-01_EB, 1.668560E-01_EB, &
   1.239099E-01_EB, &
   3.848101E-02_EB, 4.502827E-02_EB, 8.182965E-02_EB, 6.478508E-02_EB, 6.632017E-02_EB, 7.060523E-02_EB, &
   5.939407E-02_EB, &
   1.823763E-02_EB, 6.889882E-03_EB, 9.139773E-02_EB, 8.162446E-02_EB, 2.475465E-02_EB, 3.168406E-02_EB, &
   1.903134E-02_EB, &
   6.725369E-04_EB, 3.888602E-04_EB, 1.980271E-01_EB, 3.039167E-02_EB, 1.104407E-02_EB, 1.220321E-02_EB, &
   4.484167E-03_EB/),(/7,8/))

SD3_C3H6(1:7,25:26) = RESHAPE((/ &  ! 3250-3275 CM-1
   4.840794E-04_EB, 3.312870E-04_EB, 5.789111E-03_EB, 3.214095E-03_EB, 3.311342E-03_EB, 2.947009E-03_EB, &
   1.302455E-04_EB, &
   1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, 1.027560E-04_EB, &
   1.027560E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD3_C3H6(N_TEMP_C3H6,26))

! BAND #3: 2650 CM-1 - 3275 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.84732 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD3_C3H6(1:7,1:8) = RESHAPE((/ &  ! 2650-2825 CM-1
   3.045289E-04_EB, 1.015641E-03_EB, 3.935463E-04_EB, 1.154192E-04_EB, 1.194489E-04_EB, 1.102285E-04_EB, &
   1.137523E-04_EB, &
   5.888900E-04_EB, 1.065118E-03_EB, 5.931355E-04_EB, 8.588542E-04_EB, 5.730959E-04_EB, 5.277001E-04_EB, &
   2.327868E-03_EB, &
   3.553832E-03_EB, 5.225375E-02_EB, 5.508119E-03_EB, 3.764204E-03_EB, 6.360672E-02_EB, 6.490060E-04_EB, &
   1.126737E-03_EB, &
   3.234273E-02_EB, 6.202832E-02_EB, 1.838655E-02_EB, 1.292609E-02_EB, 1.730868E-01_EB, 7.897375E-03_EB, &
   1.175871E-02_EB, &
   5.722213E-02_EB, 8.665120E-02_EB, 3.332681E-02_EB, 1.901355E-02_EB, 3.938769E-02_EB, 7.412539E-03_EB, &
   2.819413E-02_EB, &
   3.022697E-02_EB, 1.129809E-01_EB, 2.531468E-02_EB, 1.970225E-02_EB, 5.471032E-02_EB, 2.097209E-02_EB, &
   2.390018E-02_EB, &
   5.826107E-02_EB, 1.114796E-01_EB, 3.749457E-02_EB, 3.774872E-02_EB, 1.009734E-01_EB, 4.970945E-02_EB, &
   3.868475E-02_EB, &
   7.083218E-02_EB, 1.686291E-01_EB, 7.156561E-02_EB, 6.619537E-02_EB, 2.202441E-01_EB, 1.169745E-01_EB, &
   3.834597E-01_EB/),(/7,8/))

GAMMAD3_C3H6(1:7,9:16) = RESHAPE((/ &  ! 2850-3025 CM-1
   2.387749E-01_EB, 3.465833E-01_EB, 1.768091E-01_EB, 2.165351E-01_EB, 3.664688E-01_EB, 2.483037E-01_EB, &
   8.926665E-01_EB, &
   5.324178E-01_EB, 5.980304E-01_EB, 3.793473E-01_EB, 4.509406E-01_EB, 5.538922E-01_EB, 4.079465E-01_EB, &
   5.945021E-01_EB, &
   8.715520E-01_EB, 9.293503E-01_EB, 6.306967E-01_EB, 7.198743E-01_EB, 7.551117E-01_EB, 6.154785E-01_EB, &
   8.223511E-01_EB, &
   1.499464E+00_EB, 1.399873E+00_EB, 9.866448E-01_EB, 1.025786E+00_EB, 9.581477E-01_EB, 7.125299E-01_EB, &
   1.107012E+00_EB, &
   2.156257E+00_EB, 1.637737E+00_EB, 1.241655E+00_EB, 1.280306E+00_EB, 1.138186E+00_EB, 8.441688E-01_EB, &
   1.384899E+00_EB, &
   3.002594E+00_EB, 1.714061E+00_EB, 1.257507E+00_EB, 1.172490E+00_EB, 1.028409E+00_EB, 6.569541E-01_EB, &
   2.100100E+00_EB, &
   2.734771E+00_EB, 1.525371E+00_EB, 1.227801E+00_EB, 1.124537E+00_EB, 1.012834E+00_EB, 6.561761E-01_EB, &
   1.786964E+00_EB, &
   1.391166E+00_EB, 1.039743E+00_EB, 8.609837E-01_EB, 8.549996E-01_EB, 8.423802E-01_EB, 6.266077E-01_EB, &
   1.102257E+00_EB/),(/7,8/))

GAMMAD3_C3H6(1:7,17:24) = RESHAPE((/ &  ! 3050-3225 CM-1
   8.645946E-01_EB, 9.364314E-01_EB, 7.448637E-01_EB, 7.668057E-01_EB, 7.630071E-01_EB, 5.882839E-01_EB, &
   1.479949E+00_EB, &
   1.230685E+00_EB, 1.046279E+00_EB, 7.526449E-01_EB, 7.500084E-01_EB, 7.007694E-01_EB, 5.329368E-01_EB, &
   1.354177E+00_EB, &
   1.587192E+00_EB, 9.549637E-01_EB, 7.345138E-01_EB, 6.867975E-01_EB, 5.981748E-01_EB, 4.326344E-01_EB, &
   4.822820E+00_EB, &
   6.944559E-01_EB, 3.901526E-01_EB, 3.458891E-01_EB, 3.287273E-01_EB, 2.870524E-01_EB, 2.152839E-01_EB, &
   4.345222E+01_EB, &
   1.550994E-01_EB, 1.096247E-01_EB, 1.314631E-01_EB, 1.210936E-01_EB, 1.139491E-01_EB, 5.956151E-02_EB, &
   3.313368E+01_EB, &
   2.202410E-02_EB, 2.065956E-02_EB, 1.775205E-02_EB, 2.386698E-02_EB, 3.888119E-02_EB, 2.898450E-02_EB, &
   2.058336E+00_EB, &
   1.065388E-04_EB, 1.920534E-03_EB, 6.158394E-04_EB, 4.540651E-04_EB, 1.753716E-02_EB, 7.274816E-03_EB, &
   1.007187E+00_EB, &
   4.230956E-04_EB, 2.525852E-04_EB, 8.709782E-05_EB, 2.954917E-04_EB, 5.762183E-03_EB, 3.267777E-03_EB, &
   1.887507E-02_EB/),(/7,8/))

GAMMAD3_C3H6(1:7,25:26) = RESHAPE((/ &  ! 3250-3275 CM-1
   3.563014E-04_EB, 2.388188E-04_EB, 5.650871E-04_EB, 1.353722E-03_EB, 1.592126E-03_EB, 7.501643E-04_EB, &
   1.070038E-04_EB, &
   8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, 8.047518E-05_EB, &
   8.047518E-05_EB/),(/7,2/))


!-------------------------TOLUENE DATA-------------------


! THERE ARE 5 BANDS FOR TOLUENE

! BAND #1: 700 CM-1 - 805 CM-1
! BAND #2: 975 CM-1 - 1175 CM-1
! BAND #3: 1275 CM-1 - 1675 CM-1
! BAND #4: 1650 CM-1 - 2075 CM-1
! BAND #5: 2675 CM-1 - 3225 CM-1

ALLOCATE(SD_C7H8_TEMP(N_TEMP_C7H8))

! INITIALIZE BANDS WAVENUMBER BOUNDS FOR TOLUENE ABS(OPTION COEFFICIENTS.
! BANDS ARE ORGANIZED BY ROW:
! 1ST COLUMN IS THE LOWER BOUND BAND LIMIT IN CM-1.
! 2ND COLUMN IS THE UPPER BOUND BAND LIMIT IN CM-1.
! 3RD COLUMN IS THE STRIDE BETWEEN WAVENUMBERS IN BAND.
! IF 3RD COLUMN = 0., BAND IS CALCULATED, NOT TABULATED.

ALLOCATE(OM_BND_C7H8(N_BAND_C7H8,3))
ALLOCATE(BE_C7H8(N_BAND_C7H8))

OM_BND_C7H8 = RESHAPE((/ &
   700._EB,  975._EB, 1275._EB, 1650._EB, 2675._EB, &
   805._EB, 1175._EB, 1675._EB, 2075._EB, 3225._EB, &
   5._EB, 5._EB, 25._EB, 25._EB, 25._EB/),(/N_BAND_C7H8,3/))

SD_C7H8_TEMP = (/ &
   300._EB, 396._EB, 440._EB, 477._EB, 587._EB, 795._EB,&
   999._EB/)

BE_C7H8 = (/ &
   1.000_EB, 1.000_EB, 1.000_EB, 1.000_EB, 1.000_EB/)

!---------------------------------------------------------------------------
ALLOCATE(SD1_C7H8(N_TEMP_C7H8,22))

! BAND #1: 700 CM-1 - 805 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 5.5814 %

SD1_C7H8(1:7,1:8) = RESHAPE((/ &  ! 700-735 CM-1
   1.112466E+00_EB, 9.044923E-01_EB, 9.061370E-01_EB, 8.938794E-01_EB, 9.084825E-01_EB, 1.435389E+00_EB, &
   1.297156E+01_EB, &
   2.840815E+00_EB, 2.482407E+00_EB, 2.360535E+00_EB, 2.276458E+00_EB, 2.122272E+00_EB, 1.940697E+00_EB, &
   5.112554E+00_EB, &
   3.455206E+00_EB, 3.166147E+00_EB, 2.863080E+00_EB, 2.620674E+00_EB, 2.467633E+00_EB, 2.005290E+00_EB, &
   7.244894E-01_EB, &
   4.288216E+00_EB, 3.577673E+00_EB, 2.963877E+00_EB, 2.789847E+00_EB, 2.401611E+00_EB, 1.912466E+00_EB, &
   2.024019E+00_EB, &
   4.666515E+00_EB, 3.754488E+00_EB, 3.254065E+00_EB, 3.058494E+00_EB, 2.875391E+00_EB, 2.451298E+00_EB, &
   4.223163E+00_EB, &
   7.363590E+00_EB, 6.122925E+00_EB, 5.403265E+00_EB, 5.328257E+00_EB, 4.641149E+00_EB, 3.466727E+00_EB, &
   1.373953E+01_EB, &
   9.587057E+00_EB, 6.742653E+00_EB, 5.600793E+00_EB, 5.363171E+00_EB, 4.064656E+00_EB, 2.508085E+00_EB, &
   1.195299E+00_EB, &
   4.715829E+00_EB, 3.108857E+00_EB, 2.753587E+00_EB, 2.650163E+00_EB, 2.249724E+00_EB, 1.963447E+00_EB, &
   7.262488E+00_EB/),(/7,8/))

SD1_C7H8(1:7,9:16) = RESHAPE((/ &  ! 740-775 CM-1
   4.190702E+00_EB, 2.766122E+00_EB, 2.537781E+00_EB, 2.386878E+00_EB, 2.084718E+00_EB, 1.344226E+00_EB, &
   3.448390E+00_EB, &
   2.838874E+00_EB, 2.217681E+00_EB, 2.081071E+00_EB, 1.986804E+00_EB, 1.810042E+00_EB, 1.549486E+00_EB, &
   7.126841E-01_EB, &
   1.327883E+00_EB, 1.440906E+00_EB, 1.390571E+00_EB, 1.421333E+00_EB, 1.328347E+00_EB, 1.296316E+00_EB, &
   1.254896E+00_EB, &
   4.533944E-01_EB, 8.351739E-01_EB, 7.591511E-01_EB, 7.962784E-01_EB, 8.186310E-01_EB, 8.370456E-01_EB, &
   2.685634E+00_EB, &
   1.131517E-01_EB, 7.896344E-01_EB, 3.378503E-01_EB, 3.523472E-01_EB, 4.364146E-01_EB, 5.594990E-01_EB, &
   3.259981E+00_EB, &
   3.791615E-02_EB, 8.912653E-01_EB, 4.878342E-01_EB, 1.492598E-01_EB, 2.103530E-01_EB, 3.269232E-01_EB, &
   2.144533E+00_EB, &
   4.206761E-02_EB, 7.530056E-02_EB, 1.233550E-01_EB, 7.429326E-02_EB, 9.115926E-02_EB, 1.525590E-01_EB, &
   1.109247E+00_EB, &
   6.711521E-02_EB, 8.510326E-04_EB, 6.739778E-02_EB, 5.411094E-02_EB, 6.504658E-02_EB, 5.754676E-02_EB, &
   5.875940E-01_EB/),(/7,8/))

SD1_C7H8(1:7,17:22) = RESHAPE((/ &  ! 780-805 CM-1
   6.275862E-02_EB, 6.962504E-04_EB, 4.265822E-02_EB, 2.937821E-02_EB, 3.183374E-02_EB, 1.453078E-02_EB, &
   8.515825E-02_EB, &
   5.193232E-02_EB, 7.365615E-04_EB, 1.939785E-02_EB, 1.719349E-02_EB, 2.278960E-02_EB, 1.331369E-03_EB, &
   5.520836E-02_EB, &
   3.535671E-02_EB, 8.579228E-04_EB, 1.765940E-03_EB, 1.491081E-02_EB, 1.181497E-02_EB, 5.221427E-04_EB, &
   2.334611E-03_EB, &
   6.188348E-03_EB, 7.316561E-04_EB, 8.181678E-04_EB, 6.095900E-03_EB, 3.380075E-03_EB, 5.968569E-04_EB, &
   7.411591E-04_EB, &
   6.328121E-04_EB, 6.599274E-04_EB, 5.510258E-04_EB, 5.104659E-04_EB, 5.023646E-04_EB, 5.757667E-04_EB, &
   6.209242E-04_EB, &
   4.621323E-04_EB, 4.638592E-04_EB, 4.665860E-04_EB, 4.636519E-04_EB, 4.646393E-04_EB, 4.633936E-04_EB, &
   4.615859E-04_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD1_C7H8(N_TEMP_C7H8,22))

! BAND #1: 700 CM-1 - 805 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 5.5814 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD1_C7H8(1:7,1:8) = RESHAPE((/ &  ! 700-735 CM-1
   7.999461E+01_EB, 6.960752E+01_EB, 3.022045E+01_EB, 6.344408E+01_EB, 5.719128E+01_EB, 3.598038E-02_EB, &
   1.819109E-03_EB, &
   7.999993E+01_EB, 6.962945E+01_EB, 6.605713E+01_EB, 6.344357E+01_EB, 5.719074E+01_EB, 3.344331E-01_EB, &
   1.001242E-02_EB, &
   8.000000E+01_EB, 6.922109E+01_EB, 4.454022E+00_EB, 6.344038E+01_EB, 5.719013E+01_EB, 9.751497E-02_EB, &
   4.383971E+01_EB, &
   8.000000E+01_EB, 1.649903E+01_EB, 6.605768E+01_EB, 6.344412E+01_EB, 5.719149E+01_EB, 6.787669E-01_EB, &
   1.717471E-01_EB, &
   8.000000E+01_EB, 5.534136E+00_EB, 6.605782E+01_EB, 6.344412E+01_EB, 5.719151E+01_EB, 4.914284E+01_EB, &
   6.407876E-02_EB, &
   8.000000E+01_EB, 6.962930E+01_EB, 6.605773E+01_EB, 8.838043E+00_EB, 5.719150E+01_EB, 2.254685E+00_EB, &
   1.178609E-02_EB, &
   8.000000E+01_EB, 6.963055E+01_EB, 6.605773E+01_EB, 3.115435E+00_EB, 5.719151E+01_EB, 4.914358E+01_EB, &
   4.383966E+01_EB, &
   8.000000E+01_EB, 6.963104E+01_EB, 6.605612E+01_EB, 6.343942E+01_EB, 5.719109E+01_EB, 1.671127E-01_EB, &
   6.618119E-03_EB/),(/7,8/))

GAMMAD1_C7H8(1:7,9:16) = RESHAPE((/ &  ! 740-775 CM-1
   8.000000E+01_EB, 6.963105E+01_EB, 6.605690E+01_EB, 6.333100E+01_EB, 5.718653E+01_EB, 4.913467E+01_EB, &
   1.209833E-02_EB, &
   8.000000E+01_EB, 6.962982E+01_EB, 6.605767E+01_EB, 6.342460E+01_EB, 5.718540E+01_EB, 2.011652E-01_EB, &
   4.383963E+01_EB, &
   7.999980E+01_EB, 6.960747E+01_EB, 6.602987E+01_EB, 6.326339E+01_EB, 5.714448E+01_EB, 3.303355E-01_EB, &
   2.494003E-02_EB, &
   4.634778E+01_EB, 1.999423E-01_EB, 1.445536E+01_EB, 3.851197E+01_EB, 5.716296E+01_EB, 2.419503E+00_EB, &
   4.403455E-03_EB, &
   1.788269E+00_EB, 8.781616E-03_EB, 1.079139E-01_EB, 1.421403E+01_EB, 9.057111E+00_EB, 1.781522E+01_EB, &
   1.669472E-03_EB, &
   2.864129E-01_EB, 1.515581E-03_EB, 2.119890E-03_EB, 5.013687E+00_EB, 3.747774E+00_EB, 1.844443E+00_EB, &
   1.069786E-03_EB, &
   1.102908E+01_EB, 4.896633E-02_EB, 1.798393E-02_EB, 6.492346E-01_EB, 8.863603E-02_EB, 2.773305E-01_EB, &
   5.841296E-04_EB, &
   1.251724E+00_EB, 6.490441E-04_EB, 2.113061E-02_EB, 1.326897E+00_EB, 8.049318E-03_EB, 8.055391E-03_EB, &
   2.478640E-04_EB/),(/7,8/))

GAMMAD1_C7H8(1:7,17:22) = RESHAPE((/ &  ! 780-805 CM-1
   6.819808E-01_EB, 5.093301E-04_EB, 2.666337E-02_EB, 2.101043E-01_EB, 1.980342E-02_EB, 1.679664E-03_EB, &
   1.906432E-02_EB, &
   4.691940E-01_EB, 5.429738E-04_EB, 1.319467E-02_EB, 4.975402E-03_EB, 2.433969E-02_EB, 1.884068E-03_EB, &
   5.887236E-03_EB, &
   1.875981E-01_EB, 6.598198E-04_EB, 1.786636E-03_EB, 1.408548E-02_EB, 2.119727E-01_EB, 4.335098E-04_EB, &
   1.937100E-03_EB, &
   1.442921E-02_EB, 5.711043E-04_EB, 5.971323E-04_EB, 7.973510E-03_EB, 3.341200E-03_EB, 4.314868E-04_EB, &
   5.200300E-04_EB, &
   5.156874E-04_EB, 5.125082E-04_EB, 4.318513E-04_EB, 3.992184E-04_EB, 3.936730E-04_EB, 4.154189E-04_EB, &
   4.736124E-04_EB, &
   3.682775E-04_EB, 3.700575E-04_EB, 3.724614E-04_EB, 3.699127E-04_EB, 3.709688E-04_EB, 3.696394E-04_EB, &
   3.680801E-04_EB/),(/7,6/))

!---------------------------------------------------------------------------
ALLOCATE(SD2_C7H8(N_TEMP_C7H8,29))

! BAND #2: 975 CM-1 - 1175 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.91355 %

SD2_C7H8(1:7,1:8) = RESHAPE((/ &  ! 975-1010 CM-1
   5.806929E-04_EB, 5.942617E-04_EB, 6.726459E-04_EB, 7.052569E-04_EB, 6.695985E-04_EB, 5.159322E-04_EB, &
   5.551144E-04_EB, &
   6.785508E-04_EB, 8.056701E-04_EB, 8.454611E-04_EB, 7.340346E-04_EB, 7.540252E-04_EB, 6.231487E-04_EB, &
   1.161712E-02_EB, &
   7.608293E-04_EB, 5.613980E-02_EB, 8.330584E-04_EB, 7.050990E-04_EB, 7.281009E-04_EB, 1.902747E-03_EB, &
   2.699830E-02_EB, &
   4.497896E-02_EB, 9.748044E-02_EB, 5.769989E-04_EB, 8.307786E-04_EB, 1.048030E-03_EB, 3.071367E-02_EB, &
   5.746590E-02_EB, &
   4.185662E-02_EB, 5.807183E-02_EB, 5.845956E-03_EB, 4.738381E-03_EB, 1.373288E-02_EB, 1.722897E-02_EB, &
   6.537832E-02_EB, &
   2.338216E-02_EB, 1.675285E-02_EB, 2.368298E-02_EB, 3.338373E-02_EB, 1.128221E-02_EB, 2.346587E-02_EB, &
   3.565197E-02_EB, &
   1.379198E-02_EB, 9.132916E-04_EB, 3.690042E-02_EB, 2.275924E-02_EB, 2.434022E-02_EB, 4.242178E-02_EB, &
   6.940845E-02_EB, &
   2.629167E-02_EB, 2.433785E-03_EB, 7.441229E-02_EB, 7.329327E-02_EB, 8.154865E-02_EB, 8.520170E-02_EB, &
   8.926539E-02_EB/),(/7,8/))

SD2_C7H8(1:7,9:16) = RESHAPE((/ &  ! 1015-1050 CM-1
   8.643728E-02_EB, 8.397464E-02_EB, 1.634758E-01_EB, 1.741891E-01_EB, 1.714205E-01_EB, 1.480038E-01_EB, &
   1.360983E-01_EB, &
   3.004729E-01_EB, 2.793190E-01_EB, 2.877798E-01_EB, 2.821584E-01_EB, 2.456569E-01_EB, 1.463510E-01_EB, &
   1.826942E-01_EB, &
   5.278878E-01_EB, 3.693913E-01_EB, 2.973200E-01_EB, 2.717632E-01_EB, 2.438624E-01_EB, 1.379726E-01_EB, &
   2.050903E-01_EB, &
   5.845808E-01_EB, 4.794762E-01_EB, 3.584954E-01_EB, 3.427664E-01_EB, 2.786454E-01_EB, 1.226379E-01_EB, &
   2.097466E-01_EB, &
   4.729738E-01_EB, 3.788669E-01_EB, 2.664993E-01_EB, 2.564802E-01_EB, 2.264334E-01_EB, 1.535049E-01_EB, &
   1.682791E-01_EB, &
   5.530868E-01_EB, 4.490113E-01_EB, 3.310727E-01_EB, 3.288768E-01_EB, 3.019277E-01_EB, 2.043861E-01_EB, &
   1.497158E-01_EB, &
   4.129235E-01_EB, 3.251444E-01_EB, 2.714955E-01_EB, 2.657956E-01_EB, 2.487687E-01_EB, 1.862679E-01_EB, &
   1.223084E-01_EB, &
   2.614169E-01_EB, 1.802809E-01_EB, 1.975745E-01_EB, 1.888234E-01_EB, 2.189458E-01_EB, 1.348280E-01_EB, &
   9.276862E-02_EB/),(/7,8/))

SD2_C7H8(1:7,17:24) = RESHAPE((/ &  ! 1055-1090 CM-1
   2.730370E-01_EB, 1.767450E-01_EB, 2.218796E-01_EB, 2.155878E-01_EB, 2.223112E-01_EB, 1.469945E-01_EB, &
   1.150109E-01_EB, &
   2.816011E-01_EB, 2.052777E-01_EB, 2.522512E-01_EB, 2.342314E-01_EB, 2.280630E-01_EB, 2.544731E-01_EB, &
   1.975636E-01_EB, &
   3.260058E-01_EB, 2.887673E-01_EB, 2.910159E-01_EB, 2.801621E-01_EB, 2.791987E-01_EB, 2.761989E-01_EB, &
   6.486488E-01_EB, &
   4.240289E-01_EB, 3.848016E-01_EB, 3.433772E-01_EB, 3.147066E-01_EB, 2.873021E-01_EB, 2.027784E-01_EB, &
   6.801384E-01_EB, &
   5.531291E-01_EB, 4.341564E-01_EB, 3.312172E-01_EB, 3.045238E-01_EB, 2.620640E-01_EB, 1.519635E-01_EB, &
   9.321833E-01_EB, &
   4.365089E-01_EB, 3.538039E-01_EB, 2.730701E-01_EB, 2.397118E-01_EB, 2.227185E-01_EB, 2.009481E-01_EB, &
   1.174461E+00_EB, &
   4.683512E-01_EB, 3.712916E-01_EB, 2.862375E-01_EB, 2.540628E-01_EB, 2.485619E-01_EB, 2.060439E-01_EB, &
   9.951951E-01_EB, &
   4.698469E-01_EB, 3.518932E-01_EB, 2.825430E-01_EB, 2.624504E-01_EB, 2.213420E-01_EB, 1.625252E-01_EB, &
   5.760843E-01_EB/),(/7,8/))

SD2_C7H8(1:7,25:29) = RESHAPE((/ &  ! 1095-1175 CM-1
   3.649981E-01_EB, 2.736054E-01_EB, 2.408070E-01_EB, 2.323294E-01_EB, 1.878487E-01_EB, 1.252916E-01_EB, &
   4.029469E-02_EB, &
   2.681413E-01_EB, 1.701855E-01_EB, 1.674993E-01_EB, 1.511744E-01_EB, 1.166087E-01_EB, 7.290456E-02_EB, &
   1.474340E-01_EB, &
   4.990165E-02_EB, 6.292687E-02_EB, 4.665808E-02_EB, 3.377088E-02_EB, 2.693981E-02_EB, 9.990414E-03_EB, &
   9.979516E-03_EB, &
   6.130796E-04_EB, 1.597828E-03_EB, 5.178852E-03_EB, 6.236302E-04_EB, 6.358216E-04_EB, 6.358332E-04_EB, &
   5.729829E-04_EB, &
   4.621323E-04_EB, 4.638592E-04_EB, 4.665860E-04_EB, 4.636519E-04_EB, 4.646393E-04_EB, 4.633936E-04_EB, &
   4.615859E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD2_C7H8(N_TEMP_C7H8,29))

! BAND #2: 975 CM-1 - 1175 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.91355 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD2_C7H8(1:7,1:8) = RESHAPE((/ &  ! 975-1010 CM-1
   4.305850E-04_EB, 4.663364E-04_EB, 5.289580E-04_EB, 5.402269E-04_EB, 5.024478E-04_EB, 4.070836E-04_EB, &
   4.471953E-04_EB, &
   5.202785E-04_EB, 6.465361E-04_EB, 6.452178E-04_EB, 5.771613E-04_EB, 5.746126E-04_EB, 4.735889E-04_EB, &
   1.573544E-04_EB, &
   5.872521E-04_EB, 7.042066E-03_EB, 5.999286E-04_EB, 5.331992E-04_EB, 6.220132E-04_EB, 6.364420E-02_EB, &
   1.528351E-04_EB, &
   5.511126E-01_EB, 1.497331E-02_EB, 4.594486E-04_EB, 5.729761E-04_EB, 7.713361E-04_EB, 7.130935E-03_EB, &
   1.427152E-04_EB, &
   2.212250E+00_EB, 1.945735E-01_EB, 2.212567E-01_EB, 9.922155E-03_EB, 9.234920E-02_EB, 1.013608E-02_EB, &
   4.272371E-04_EB, &
   4.405486E-01_EB, 7.436540E-01_EB, 1.920483E-01_EB, 2.962863E-03_EB, 3.448915E-01_EB, 3.486423E-01_EB, &
   1.313079E-01_EB, &
   2.410517E-01_EB, 1.039535E-03_EB, 1.254824E-01_EB, 1.067258E-01_EB, 2.942350E-01_EB, 5.806212E-01_EB, &
   1.774699E+00_EB, &
   4.514842E-01_EB, 1.759071E-01_EB, 2.902656E-01_EB, 2.267366E-01_EB, 8.401946E-02_EB, 7.681227E+00_EB, &
   1.330635E+00_EB/),(/7,8/))

GAMMAD2_C7H8(1:7,9:16) = RESHAPE((/ &  ! 1015-1050 CM-1
   2.224412E+00_EB, 1.484769E+00_EB, 1.494944E+00_EB, 1.832759E+00_EB, 3.314911E+00_EB, 5.865543E+00_EB, &
   6.106454E-01_EB, &
   2.999242E+01_EB, 7.193297E+00_EB, 3.422464E+00_EB, 4.614503E+00_EB, 2.188441E+00_EB, 7.394814E+00_EB, &
   3.908543E-02_EB, &
   7.999990E+01_EB, 3.581705E+00_EB, 7.707850E+00_EB, 3.420693E+00_EB, 6.146487E+00_EB, 7.025651E+00_EB, &
   3.617788E-02_EB, &
   7.998352E+01_EB, 4.444656E-01_EB, 1.228523E+01_EB, 3.433637E+00_EB, 2.072175E+01_EB, 4.335871E+00_EB, &
   7.248922E-02_EB, &
   7.410363E+01_EB, 5.532136E-01_EB, 5.267901E+00_EB, 5.384235E+00_EB, 5.570719E+00_EB, 7.078253E+00_EB, &
   7.681535E+00_EB, &
   7.999988E+01_EB, 1.364684E+01_EB, 1.337888E+01_EB, 5.824947E+00_EB, 5.724261E+00_EB, 6.510348E+00_EB, &
   2.593089E+00_EB, &
   5.835221E+01_EB, 2.086568E+01_EB, 4.818524E+00_EB, 1.625989E+00_EB, 4.963013E+00_EB, 9.091313E+00_EB, &
   2.862616E+00_EB, &
   1.807887E+01_EB, 1.240441E+01_EB, 1.649639E+00_EB, 3.870645E+00_EB, 4.710413E+00_EB, 5.816990E+00_EB, &
   2.011503E-01_EB/),(/7,8/))

GAMMAD2_C7H8(1:7,17:24) = RESHAPE((/ &  ! 1055-1090 CM-1
   2.242814E+01_EB, 1.324109E+01_EB, 2.772538E+00_EB, 2.132624E+00_EB, 2.967247E+00_EB, 5.544644E+00_EB, &
   1.492977E-02_EB, &
   2.726543E+01_EB, 8.755811E+00_EB, 4.268738E+00_EB, 8.268829E+00_EB, 3.471302E+00_EB, 7.655547E+00_EB, &
   4.678557E-03_EB, &
   3.856386E+01_EB, 1.415885E+01_EB, 4.621129E+00_EB, 7.774568E+00_EB, 5.265771E-01_EB, 7.296652E+00_EB, &
   8.956780E-04_EB, &
   6.779154E+01_EB, 1.689515E+01_EB, 8.227216E+00_EB, 7.685408E+00_EB, 6.275187E+00_EB, 4.111618E+00_EB, &
   1.120381E-03_EB, &
   7.998412E+01_EB, 5.217367E+00_EB, 1.270480E+01_EB, 7.011278E+00_EB, 9.533137E+00_EB, 6.557224E+00_EB, &
   7.187912E-04_EB, &
   6.476932E+01_EB, 4.324490E+00_EB, 7.006941E+00_EB, 3.059077E+00_EB, 2.551623E+00_EB, 6.781168E+00_EB, &
   4.850850E-04_EB, &
   7.999697E+01_EB, 1.683983E+01_EB, 7.523811E+00_EB, 3.934915E+00_EB, 3.190601E+00_EB, 5.987392E+00_EB, &
   4.640575E-04_EB, &
   7.298422E+01_EB, 1.746059E+01_EB, 8.445975E+00_EB, 4.651083E+00_EB, 5.086497E+00_EB, 1.446891E+00_EB, &
   2.261182E-04_EB/),(/7,8/))

GAMMAD2_C7H8(1:7,25:29) = RESHAPE((/ &  ! 1095-1175 CM-1
   5.656308E+01_EB, 1.932458E+01_EB, 5.566997E+00_EB, 2.210155E-01_EB, 4.019081E+00_EB, 7.861722E-02_EB, &
   3.404900E-03_EB, &
   2.190597E+01_EB, 1.019795E+01_EB, 2.244564E+00_EB, 1.625625E+00_EB, 5.393117E-01_EB, 1.054102E+00_EB, &
   7.309670E-05_EB, &
   4.172119E-01_EB, 6.066095E-01_EB, 3.794528E-01_EB, 2.201505E-01_EB, 8.146470E-02_EB, 1.686354E-01_EB, &
   2.521455E-03_EB, &
   4.904118E-04_EB, 8.240680E-03_EB, 2.319123E-01_EB, 4.751613E-04_EB, 5.109468E-04_EB, 5.032463E-04_EB, &
   4.422597E-04_EB, &
   3.682775E-04_EB, 3.700575E-04_EB, 3.724614E-04_EB, 3.699127E-04_EB, 3.709688E-04_EB, 3.696394E-04_EB, &
   3.680801E-04_EB/),(/7,5/))

!---------------------------------------------------------------------------
ALLOCATE(SD3_C7H8(N_TEMP_C7H8,17))

! BAND #3: 1275 CM-1 - 1675 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 1.8457 %

SD3_C7H8(1:7,1:8) = RESHAPE((/ &  ! 1275-1450 CM-1
   6.100102E-04_EB, 7.780830E-03_EB, 1.134938E-03_EB, 8.737574E-04_EB, 6.027841E-03_EB, 4.138198E-03_EB, &
   9.056224E-04_EB, &
   7.403110E-04_EB, 3.544033E-02_EB, 3.023568E-02_EB, 1.508266E-02_EB, 2.813528E-02_EB, 2.043206E-02_EB, &
   4.350956E-01_EB, &
   3.178283E-03_EB, 3.702754E-02_EB, 4.100162E-02_EB, 2.423149E-02_EB, 4.352025E-02_EB, 3.192771E-02_EB, &
   9.676683E-01_EB, &
   5.605844E-02_EB, 6.834484E-02_EB, 8.335619E-02_EB, 6.360984E-02_EB, 6.991888E-02_EB, 5.926399E-02_EB, &
   6.248318E-02_EB, &
   2.335707E-01_EB, 2.104431E-01_EB, 2.014754E-01_EB, 1.664392E-01_EB, 1.529464E-01_EB, 1.145999E-01_EB, &
   4.880863E-02_EB, &
   2.297034E-01_EB, 1.952774E-01_EB, 2.009723E-01_EB, 1.554485E-01_EB, 1.472329E-01_EB, 9.989047E-02_EB, &
   5.114395E-02_EB, &
   2.466205E-01_EB, 2.254143E-01_EB, 2.146406E-01_EB, 1.846656E-01_EB, 1.917496E-01_EB, 1.694292E-01_EB, &
   9.885420E-02_EB, &
   6.338228E-01_EB, 4.862753E-01_EB, 4.470592E-01_EB, 3.954809E-01_EB, 3.379362E-01_EB, 2.524978E-01_EB, &
   2.245216E-01_EB/),(/7,8/))

SD3_C7H8(1:7,9:16) = RESHAPE((/ &  ! 1475-1650 CM-1
   8.091914E-01_EB, 6.844493E-01_EB, 6.417303E-01_EB, 6.044514E-01_EB, 5.830416E-01_EB, 4.886091E-01_EB, &
   3.460225E-01_EB, &
   1.985404E+00_EB, 1.434197E+00_EB, 1.197360E+00_EB, 1.090591E+00_EB, 8.747838E-01_EB, 5.725550E-01_EB, &
   3.573484E-01_EB, &
   5.171385E-01_EB, 4.486683E-01_EB, 4.336080E-01_EB, 3.554049E-01_EB, 3.143224E-01_EB, 2.347278E-01_EB, &
   1.742309E-01_EB, &
   2.372746E-01_EB, 2.197212E-01_EB, 2.373528E-01_EB, 1.855960E-01_EB, 1.760600E-01_EB, 1.693906E-01_EB, &
   1.742322E-01_EB, &
   2.366393E-01_EB, 3.786876E-01_EB, 4.105672E-01_EB, 2.381817E-01_EB, 2.272899E-01_EB, 2.664032E-01_EB, &
   3.863583E-01_EB, &
   8.572620E-01_EB, 1.527668E+00_EB, 1.063149E+00_EB, 7.295884E-01_EB, 4.745380E-01_EB, 3.602757E-01_EB, &
   2.354867E-01_EB, &
   5.567486E-01_EB, 5.472333E-01_EB, 5.016325E-01_EB, 3.512578E-01_EB, 1.717825E-01_EB, 9.711190E-02_EB, &
   2.718347E-02_EB, &
   2.625054E-02_EB, 5.072059E-03_EB, 1.694459E-02_EB, 9.001577E-03_EB, 5.938464E-04_EB, 6.681563E-04_EB, &
   5.196216E-04_EB/),(/7,8/))

SD3_C7H8(1:7,17:17) = RESHAPE((/ &  ! 1675-1675 CM-1
   4.621323E-04_EB, 4.638592E-04_EB, 4.665860E-04_EB, 4.636519E-04_EB, 4.646393E-04_EB, 4.633936E-04_EB, &
   4.615859E-04_EB/),(/7,1/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD3_C7H8(N_TEMP_C7H8,17))

! BAND #3: 1275 CM-1 - 1675 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 1.8457 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD3_C7H8(1:7,1:8) = RESHAPE((/ &  ! 1275-1450 CM-1
   4.795815E-04_EB, 2.124259E-03_EB, 1.024725E-03_EB, 7.095131E-04_EB, 6.311264E-03_EB, 2.305633E-03_EB, &
   5.045893E-04_EB, &
   5.695615E-04_EB, 1.692360E-02_EB, 7.400058E-03_EB, 1.063887E-01_EB, 2.392876E-01_EB, 3.837883E-04_EB, &
   3.113013E-05_EB, &
   2.535772E-03_EB, 3.802603E-02_EB, 2.314277E-02_EB, 4.430523E-01_EB, 1.112554E-01_EB, 1.504970E-01_EB, &
   4.347453E-05_EB, &
   7.174141E-02_EB, 1.160640E+00_EB, 2.191895E-02_EB, 4.952819E-01_EB, 4.273983E-01_EB, 3.451409E-01_EB, &
   2.592647E-01_EB, &
   8.276943E+00_EB, 3.502859E-01_EB, 5.113046E-02_EB, 2.027660E+00_EB, 2.707192E+00_EB, 3.833564E-01_EB, &
   3.464319E-01_EB, &
   1.015475E+01_EB, 4.674500E+00_EB, 4.689087E-02_EB, 3.954451E+00_EB, 3.031215E+00_EB, 1.038644E+00_EB, &
   6.356775E-02_EB, &
   1.948113E+01_EB, 7.058967E+00_EB, 9.857392E-01_EB, 6.992797E+00_EB, 4.529865E+00_EB, 2.876236E+00_EB, &
   1.523489E+00_EB, &
   7.999993E+01_EB, 4.801475E+01_EB, 2.297044E+00_EB, 2.307491E+01_EB, 2.255518E+01_EB, 6.921021E+00_EB, &
   3.793481E+00_EB/),(/7,8/))

GAMMAD3_C7H8(1:7,9:16) = RESHAPE((/ &  ! 1475-1650 CM-1
   7.999994E+01_EB, 3.335220E+01_EB, 5.984703E-01_EB, 2.853393E+01_EB, 3.626523E+01_EB, 2.277723E+01_EB, &
   2.603661E+01_EB, &
   7.999988E+01_EB, 6.963104E+01_EB, 6.603136E+01_EB, 6.344400E+01_EB, 5.716242E+01_EB, 4.913311E+01_EB, &
   1.650821E+01_EB, &
   7.999998E+01_EB, 6.963102E+01_EB, 2.177188E+01_EB, 3.299642E+01_EB, 1.878956E+01_EB, 8.669014E+00_EB, &
   7.333480E+00_EB, &
   5.251899E+01_EB, 2.739552E+01_EB, 6.379395E+00_EB, 1.192897E+01_EB, 1.481504E+01_EB, 6.745394E+00_EB, &
   2.823356E+00_EB, &
   1.643612E-01_EB, 5.735340E-03_EB, 4.062201E-03_EB, 3.334872E-02_EB, 1.282097E+00_EB, 8.493361E-02_EB, &
   6.376486E-03_EB, &
   1.487048E-01_EB, 1.035601E-02_EB, 1.158524E-02_EB, 3.774427E-02_EB, 1.523506E-01_EB, 2.811809E-02_EB, &
   7.547214E-03_EB, &
   7.999361E+01_EB, 1.936478E-02_EB, 9.051617E-03_EB, 1.944160E-02_EB, 2.206877E-01_EB, 2.134674E-02_EB, &
   3.778034E-01_EB, &
   1.287431E-02_EB, 7.240508E-04_EB, 6.507482E-03_EB, 3.950013E-03_EB, 4.716179E-04_EB, 5.110565E-04_EB, &
   4.179497E-04_EB/),(/7,8/))

GAMMAD3_C7H8(1:7,17:17) = RESHAPE((/ &  ! 1675-1675 CM-1
   3.682775E-04_EB, 3.700575E-04_EB, 3.724614E-04_EB, 3.699127E-04_EB, 3.709688E-04_EB, 3.696394E-04_EB, &
   3.680801E-04_EB/),(/7,1/))

!---------------------------------------------------------------------------
ALLOCATE(SD4_C7H8(N_TEMP_C7H8,18))

! BAND #4: 1650 CM-1 - 2075 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.43921 %

SD4_C7H8(1:7,1:8) = RESHAPE((/ &  ! 1650-1825 CM-1
   1.608510E-02_EB, 8.538057E-04_EB, 2.868555E-03_EB, 8.340494E-04_EB, 6.015596E-04_EB, 4.952974E-04_EB, &
   5.998542E-04_EB, &
   4.848768E-02_EB, 9.249328E-03_EB, 2.165734E-02_EB, 1.578699E-02_EB, 7.483581E-04_EB, 5.699315E-04_EB, &
   7.083490E-04_EB, &
   3.748462E-02_EB, 1.617457E-02_EB, 2.578284E-02_EB, 1.643850E-02_EB, 1.017398E-02_EB, 2.017300E-03_EB, &
   7.500522E-03_EB, &
   1.132410E-01_EB, 2.967893E-01_EB, 6.916754E-02_EB, 5.538010E-02_EB, 3.122037E-02_EB, 1.282020E-02_EB, &
   6.409946E-02_EB, &
   4.946037E-01_EB, 4.749997E-02_EB, 4.638123E-02_EB, 4.364247E-02_EB, 3.296949E-02_EB, 3.250858E-02_EB, &
   6.216037E-02_EB, &
   9.754174E-01_EB, 6.827422E-01_EB, 5.097573E-01_EB, 4.113609E-01_EB, 9.436322E-02_EB, 1.578733E-01_EB, &
   1.005549E-01_EB, &
   5.760301E-01_EB, 1.058899E+00_EB, 5.428139E-01_EB, 2.925165E-01_EB, 1.220623E-01_EB, 2.559121E-01_EB, &
   8.881943E-02_EB, &
   1.187295E+00_EB, 1.013370E+00_EB, 7.226648E-01_EB, 5.498312E-01_EB, 8.139112E-02_EB, 4.364089E-01_EB, &
   8.132406E-02_EB/),(/7,8/))

SD4_C7H8(1:7,9:16) = RESHAPE((/ &  ! 1850-2025 CM-1
   8.856836E-01_EB, 9.186222E-01_EB, 4.794847E-01_EB, 4.024356E-01_EB, 1.306238E-01_EB, 3.063448E-01_EB, &
   9.935498E-02_EB, &
   1.034528E+00_EB, 7.215443E-01_EB, 5.929663E-01_EB, 4.319927E-01_EB, 8.863470E-02_EB, 3.544881E-01_EB, &
   5.660195E-02_EB, &
   1.495500E+00_EB, 3.838684E-01_EB, 1.268890E-01_EB, 4.632815E-02_EB, 1.148509E-01_EB, 6.472883E-02_EB, &
   8.665695E-02_EB, &
   5.568983E-01_EB, 5.685082E-01_EB, 5.071619E-01_EB, 3.647283E-01_EB, 1.319697E-01_EB, 1.328525E-01_EB, &
   8.945221E-02_EB, &
   2.939080E-01_EB, 3.996266E-01_EB, 2.679096E-01_EB, 1.696599E-01_EB, 1.249259E-01_EB, 8.650031E-02_EB, &
   7.539025E-02_EB, &
   4.677446E-01_EB, 2.881514E-02_EB, 6.793664E-02_EB, 3.565764E-02_EB, 1.939151E-02_EB, 1.637475E-02_EB, &
   3.310412E-02_EB, &
   3.406776E-02_EB, 1.092503E-02_EB, 1.198880E-02_EB, 1.579425E-02_EB, 1.184388E-02_EB, 1.945188E-03_EB, &
   2.385167E-02_EB, &
   1.429157E-02_EB, 8.354021E-04_EB, 4.409521E-03_EB, 4.730578E-03_EB, 4.878561E-03_EB, 8.332528E-04_EB, &
   2.664558E-02_EB/),(/7,8/))

SD4_C7H8(1:7,17:18) = RESHAPE((/ &  ! 2050-2075 CM-1
   5.862965E-04_EB, 5.938092E-04_EB, 6.747537E-04_EB, 5.948772E-04_EB, 6.040692E-04_EB, 6.075352E-04_EB, &
   3.980874E-03_EB, &
   4.621323E-04_EB, 4.638592E-04_EB, 4.665860E-04_EB, 4.636519E-04_EB, 4.646393E-04_EB, 4.633936E-04_EB, &
   4.615859E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD4_C7H8(N_TEMP_C7H8,18))

! BAND #4: 1650 CM-1 - 2075 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 0.43921 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD4_C7H8(1:7,1:8) = RESHAPE((/ &  ! 1650-1825 CM-1
   9.073750E-01_EB, 7.155535E-04_EB, 1.417709E-01_EB, 6.862278E-04_EB, 4.770609E-04_EB, 3.962002E-04_EB, &
   4.599493E-04_EB, &
   4.897835E-01_EB, 2.228040E-03_EB, 1.889228E-03_EB, 3.037127E-03_EB, 5.957388E-04_EB, 4.412357E-04_EB, &
   5.439496E-04_EB, &
   8.222146E-01_EB, 1.967261E-01_EB, 1.176890E-01_EB, 1.993414E-01_EB, 3.392040E-01_EB, 1.285307E+00_EB, &
   7.330343E-02_EB, &
   1.711715E-02_EB, 2.387489E-04_EB, 4.930892E-03_EB, 6.589700E-03_EB, 1.627007E-01_EB, 8.239901E-03_EB, &
   5.086131E-03_EB, &
   2.105689E-04_EB, 4.813970E-03_EB, 5.785146E-03_EB, 2.367207E-02_EB, 2.418442E-01_EB, 7.693889E-03_EB, &
   3.096368E-02_EB, &
   3.797437E-04_EB, 2.010270E-04_EB, 4.063113E-04_EB, 7.930910E-04_EB, 6.260656E-02_EB, 3.447434E-03_EB, &
   2.052893E+00_EB, &
   5.974715E-03_EB, 9.625448E-04_EB, 1.385440E-03_EB, 4.460370E-03_EB, 7.510642E-02_EB, 1.409188E-03_EB, &
   5.077965E-01_EB, &
   2.598721E-04_EB, 9.700799E-05_EB, 1.875798E-04_EB, 3.395949E-04_EB, 1.967132E-02_EB, 5.386934E-04_EB, &
   7.215293E-01_EB/),(/7,8/))

GAMMAD4_C7H8(1:7,9:16) = RESHAPE((/ &  ! 1850-2025 CM-1
   1.840578E-03_EB, 8.312411E-04_EB, 1.279188E-03_EB, 2.105186E-03_EB, 3.505265E-02_EB, 1.459376E-03_EB, &
   1.277784E+00_EB, &
   1.258299E-03_EB, 4.510848E-04_EB, 4.294071E-04_EB, 8.136257E-04_EB, 9.696057E-03_EB, 2.146063E-04_EB, &
   6.860048E-01_EB, &
   4.671598E-05_EB, 4.670991E-05_EB, 7.143144E-05_EB, 4.655956E-03_EB, 4.081375E-04_EB, 5.438877E-03_EB, &
   8.906804E-02_EB, &
   1.496812E-03_EB, 6.530123E-04_EB, 7.078867E-04_EB, 1.683327E-03_EB, 3.846642E-02_EB, 1.605013E-02_EB, &
   5.460498E-01_EB, &
   6.920106E-02_EB, 4.084630E-03_EB, 4.488361E-03_EB, 4.217654E-02_EB, 7.816064E-02_EB, 2.053699E-02_EB, &
   9.131605E-03_EB, &
   2.787181E-04_EB, 4.069568E-03_EB, 6.638116E-04_EB, 5.502709E-03_EB, 9.104577E-03_EB, 6.898146E-03_EB, &
   1.209388E-01_EB, &
   4.515662E-03_EB, 2.851322E-03_EB, 2.748099E-03_EB, 5.552875E-03_EB, 6.615247E-03_EB, 9.856806E-04_EB, &
   1.115308E-02_EB, &
   3.124492E-03_EB, 5.314268E-04_EB, 2.263215E-03_EB, 1.662162E-03_EB, 2.480485E-03_EB, 6.346151E-04_EB, &
   2.031025E-03_EB/),(/7,8/))

GAMMAD4_C7H8(1:7,17:18) = RESHAPE((/ &  ! 2050-2075 CM-1
   4.587128E-04_EB, 4.524884E-04_EB, 5.170999E-04_EB, 4.681410E-04_EB, 4.819639E-04_EB, 4.673877E-04_EB, &
   8.021840E-04_EB, &
   3.682775E-04_EB, 3.700575E-04_EB, 3.724614E-04_EB, 3.699127E-04_EB, 3.709688E-04_EB, 3.696394E-04_EB, &
   3.680801E-04_EB/),(/7,2/))

!---------------------------------------------------------------------------
ALLOCATE(SD5_C7H8(N_TEMP_C7H8,23))

! BAND #5: 2675 CM-1 - 3225 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.2868 %

SD5_C7H8(1:7,1:8) = RESHAPE((/ &  ! 2675-2850 CM-1
   5.916860E-04_EB, 6.494496E-04_EB, 7.055877E-04_EB, 6.603478E-04_EB, 6.674805E-04_EB, 5.589319E-04_EB, &
   6.137597E-04_EB, &
   7.500569E-04_EB, 7.074450E-04_EB, 9.098169E-04_EB, 7.561953E-04_EB, 1.607193E-03_EB, 1.018018E-02_EB, &
   7.780318E-03_EB, &
   7.029259E-04_EB, 8.479757E-04_EB, 9.039741E-04_EB, 8.123582E-04_EB, 2.347565E-02_EB, 3.011753E-02_EB, &
   2.717306E-02_EB, &
   8.267165E-04_EB, 8.801454E-04_EB, 2.940361E-03_EB, 1.990166E-03_EB, 4.660083E-02_EB, 4.429277E-02_EB, &
   3.863177E-02_EB, &
   8.156025E-04_EB, 1.506400E-03_EB, 6.730104E-03_EB, 3.740574E-03_EB, 2.586356E-02_EB, 3.987065E-02_EB, &
   4.205663E-02_EB, &
   9.809286E-03_EB, 1.335514E-02_EB, 2.488364E-02_EB, 1.572642E-02_EB, 4.731877E-02_EB, 6.718605E-02_EB, &
   6.938131E-02_EB, &
   5.850971E-02_EB, 6.588341E-02_EB, 7.434265E-02_EB, 6.744072E-02_EB, 1.018290E-01_EB, 1.263017E-01_EB, &
   1.302993E-01_EB, &
   2.663902E-01_EB, 2.671000E-01_EB, 2.879771E-01_EB, 2.569076E-01_EB, 2.774278E-01_EB, 2.512113E-01_EB, &
   2.305380E-01_EB/),(/7,8/))

SD5_C7H8(1:7,9:16) = RESHAPE((/ &  ! 2875-3050 CM-1
   1.192316E+00_EB, 9.205767E-01_EB, 8.299791E-01_EB, 7.348094E-01_EB, 6.292028E-01_EB, 4.584459E-01_EB, &
   3.314507E-01_EB, &
   1.111204E+00_EB, 9.394978E-01_EB, 8.641911E-01_EB, 8.038659E-01_EB, 7.457709E-01_EB, 6.165272E-01_EB, &
   4.772364E-01_EB, &
   2.020089E+00_EB, 1.631718E+00_EB, 1.548727E+00_EB, 1.354242E+00_EB, 1.167657E+00_EB, 8.526752E-01_EB, &
   6.021600E-01_EB, &
   1.932741E+00_EB, 1.403246E+00_EB, 1.298233E+00_EB, 1.078566E+00_EB, 8.872613E-01_EB, 5.993978E-01_EB, &
   4.326418E-01_EB, &
   9.805520E-01_EB, 7.107386E-01_EB, 6.778887E-01_EB, 5.479328E-01_EB, 4.804311E-01_EB, 3.935294E-01_EB, &
   3.642029E-01_EB, &
   1.226852E+00_EB, 9.501327E-01_EB, 8.230054E-01_EB, 8.274332E-01_EB, 8.323208E-01_EB, 7.816587E-01_EB, &
   6.650895E-01_EB, &
   3.262768E+00_EB, 2.723650E+00_EB, 2.296188E+00_EB, 2.270140E+00_EB, 1.935929E+00_EB, 1.369446E+00_EB, &
   9.666500E-01_EB, &
   3.556303E+00_EB, 2.816675E+00_EB, 2.390948E+00_EB, 2.282893E+00_EB, 1.948012E+00_EB, 1.461546E+00_EB, &
   1.092304E+00_EB/),(/7,8/))

SD5_C7H8(1:7,17:23) = RESHAPE((/ &  ! 3075-3225 CM-1
   2.694239E+00_EB, 2.217181E+00_EB, 1.960010E+00_EB, 1.867801E+00_EB, 1.644870E+00_EB, 1.225906E+00_EB, &
   8.559281E-01_EB, &
   1.505365E+00_EB, 1.180006E+00_EB, 1.047132E+00_EB, 9.563041E-01_EB, 8.157730E-01_EB, 5.594937E-01_EB, &
   3.588645E-01_EB, &
   3.952204E-01_EB, 3.072414E-01_EB, 2.787226E-01_EB, 2.333774E-01_EB, 1.986791E-01_EB, 1.344281E-01_EB, &
   8.253592E-02_EB, &
   2.640387E-02_EB, 3.373476E-02_EB, 3.927558E-02_EB, 2.200042E-02_EB, 3.743437E-02_EB, 3.283105E-02_EB, &
   1.723376E-02_EB, &
   2.071719E-02_EB, 1.345537E-02_EB, 1.924395E-02_EB, 6.019999E-03_EB, 1.908321E-02_EB, 1.111107E-02_EB, &
   6.304850E-04_EB, &
   1.966078E-03_EB, 1.825102E-03_EB, 3.977009E-03_EB, 6.125614E-04_EB, 1.893366E-03_EB, 2.342482E-03_EB, &
   5.793976E-04_EB, &
   4.621323E-04_EB, 4.638592E-04_EB, 4.665860E-04_EB, 4.636519E-04_EB, 4.646393E-04_EB, 4.633936E-04_EB, &
   4.615859E-04_EB/),(/7,7/))

!---------------------------------------------------------------------------
ALLOCATE(GAMMAD5_C7H8(N_TEMP_C7H8,23))

! BAND #5: 2675 CM-1 - 3225 CM-1

! SNB FIT WITH MALKMUS MODEL

! ERROR ASSOCIATED WITH MALKMUS FIT:
! ON TRANSMISSIVITY: 2.2868 %

! PRINT FINE STRUCTURE ARRAY GAMMA_D

GAMMAD5_C7H8(1:7,1:8) = RESHAPE((/ &  ! 2675-2850 CM-1
   4.498679E-04_EB, 5.206905E-04_EB, 5.314804E-04_EB, 5.225909E-04_EB, 5.382880E-04_EB, 4.354686E-04_EB, &
   4.954396E-04_EB, &
   5.733660E-04_EB, 5.664218E-04_EB, 6.760255E-04_EB, 5.720172E-04_EB, 1.214045E-03_EB, 6.728005E-03_EB, &
   6.901498E-02_EB, &
   5.501549E-04_EB, 6.422611E-04_EB, 6.751496E-04_EB, 6.170879E-04_EB, 2.505333E-02_EB, 2.929665E-02_EB, &
   2.183072E-01_EB, &
   6.226316E-04_EB, 6.634878E-04_EB, 2.204441E-03_EB, 1.315450E-03_EB, 1.427511E-01_EB, 5.684139E-02_EB, &
   1.991618E-01_EB, &
   6.173941E-04_EB, 8.607660E-04_EB, 4.084909E-03_EB, 1.485988E-03_EB, 6.217446E-02_EB, 1.696125E-01_EB, &
   6.362403E-01_EB, &
   2.271432E-01_EB, 2.575315E-02_EB, 1.742774E-02_EB, 4.872962E-03_EB, 1.222993E-01_EB, 2.494679E-01_EB, &
   4.960213E-01_EB, &
   3.499358E-01_EB, 9.426016E-01_EB, 5.241002E-02_EB, 4.920380E-02_EB, 6.407093E-01_EB, 1.208700E+00_EB, &
   3.229169E+00_EB, &
   2.056851E+01_EB, 2.155787E+00_EB, 1.241796E-01_EB, 3.443147E-01_EB, 8.158180E+00_EB, 8.235706E+00_EB, &
   4.544885E+00_EB/),(/7,8/))

GAMMAD5_C7H8(1:7,9:16) = RESHAPE((/ &  ! 2875-3050 CM-1
   7.999651E+01_EB, 4.892463E+01_EB, 2.269762E+00_EB, 1.699238E+01_EB, 3.883167E+01_EB, 2.126359E+01_EB, &
   1.128396E+01_EB, &
   7.999972E+01_EB, 6.309208E+01_EB, 3.234965E+00_EB, 1.718937E+01_EB, 4.942497E+01_EB, 3.563188E+01_EB, &
   2.569318E+01_EB, &
   7.999990E+01_EB, 6.940301E+01_EB, 2.588169E+00_EB, 6.006982E+00_EB, 5.718259E+01_EB, 4.914102E+01_EB, &
   3.825054E+01_EB, &
   7.999998E+01_EB, 6.945636E+01_EB, 5.656023E+01_EB, 4.602265E+01_EB, 5.718979E+01_EB, 4.365360E+01_EB, &
   2.767624E+01_EB, &
   7.999997E+01_EB, 6.310533E+01_EB, 2.112740E+01_EB, 2.715276E+01_EB, 2.826351E+01_EB, 1.692961E+01_EB, &
   1.913465E+01_EB, &
   7.999883E+01_EB, 6.945732E+01_EB, 3.458543E+00_EB, 4.892432E+01_EB, 5.719058E+01_EB, 4.910638E+01_EB, &
   3.932336E+01_EB, &
   7.999636E+01_EB, 6.932335E+01_EB, 2.959489E+01_EB, 2.214939E+01_EB, 5.718273E+01_EB, 4.914071E+01_EB, &
   4.383529E+01_EB, &
   7.999930E+01_EB, 6.963014E+01_EB, 6.599550E+01_EB, 5.266039E+01_EB, 5.718071E+01_EB, 4.910109E+01_EB, &
   4.383604E+01_EB/),(/7,8/))

GAMMAD5_C7H8(1:7,17:23) = RESHAPE((/ &  ! 3075-3225 CM-1
   7.999949E+01_EB, 6.961299E+01_EB, 6.604063E+01_EB, 2.849107E+01_EB, 5.719148E+01_EB, 4.914326E+01_EB, &
   4.383622E+01_EB, &
   7.999993E+01_EB, 6.944137E+01_EB, 6.603883E+01_EB, 3.834261E+01_EB, 5.713287E+01_EB, 3.198295E+01_EB, &
   1.533847E+01_EB, &
   5.103942E+01_EB, 7.534768E+00_EB, 6.064864E+00_EB, 1.229512E+00_EB, 4.034341E+00_EB, 1.526492E+00_EB, &
   9.982477E-01_EB, &
   4.966937E-01_EB, 3.630636E-01_EB, 1.127309E-01_EB, 1.271562E-02_EB, 2.831161E-01_EB, 2.892147E-01_EB, &
   8.723984E-02_EB, &
   2.678214E-01_EB, 1.122623E-01_EB, 2.684864E-02_EB, 1.687467E-03_EB, 8.582904E-03_EB, 1.126356E-01_EB, &
   4.870184E-04_EB, &
   5.319443E-03_EB, 2.133031E-03_EB, 2.657648E-03_EB, 4.684001E-04_EB, 1.112851E-03_EB, 5.919668E-03_EB, &
   4.552670E-04_EB, &
   3.682775E-04_EB, 3.700575E-04_EB, 3.724614E-04_EB, 3.699127E-04_EB, 3.709688E-04_EB, 3.696394E-04_EB, &
   3.680801E-04_EB/),(/7,7/))

!--------------------------------OLD RADCAL DATA----------------------------------------
!  INITIALIZE SD ARRAY

! TEMP,K= 300     600      1000      1500      2000      2500

SD(1:6,1:8) = RESHAPE ((/  & ! 50-225
.950E+00_EB, .103E+00_EB, .420E-01_EB, .114E-01_EB, .450E-02_EB, .300E-02_EB, &
.208E+01_EB, .365E+00_EB, .113E+00_EB, .375E-01_EB, .195E-01_EB, .134E-01_EB, &
.368E+01_EB, .990E+00_EB, .300E+00_EB, .104E+00_EB, .577E-01_EB, .365E-01_EB,  &
.650E+01_EB, .201E+01_EB, .650E+00_EB, .214E+00_EB, .128E+00_EB, .845E-01_EB, &
.825E+01_EB, .325E+01_EB, .121E+01_EB, .415E+00_EB, .260E+00_EB, .168E+00_EB, &
.870E+01_EB, .452E+01_EB, .189E+01_EB, .765E+00_EB, .450E+00_EB, .289E+00_EB, &
.810E+01_EB, .540E+01_EB, .261E+01_EB, .126E+01_EB, .695E+00_EB, .460E+00_EB, &
.682E+01_EB, .600E+01_EB, .337E+01_EB, .179E+01_EB, .101E+01_EB, .679E+00_EB/), &
(/6,8/))
SD(1:6,9:16) = RESHAPE ((/ &  ! 250-425
.493E+01_EB, .622E+01_EB, .407E+01_EB, .230E+01_EB, .135E+01_EB, .935E+00_EB,  &
.316E+01_EB, .592E+01_EB, .456E+01_EB, .281E+01_EB, .172E+01_EB, .122E+01_EB,  &
.199E+01_EB, .528E+01_EB, .479E+01_EB, .328E+01_EB, .213E+01_EB, .149E+01_EB,  &
.113E+01_EB, .450E+01_EB, .484E+01_EB, .361E+01_EB, .249E+01_EB, .179E+01_EB,  &
.585E+00_EB, .370E+01_EB, .471E+01_EB, .383E+01_EB, .284E+01_EB, .208E+01_EB, &
.293E+00_EB, .289E+01_EB, .443E+01_EB, .394E+01_EB, .312E+01_EB, .237E+01_EB, &
.138E+00_EB, .205E+01_EB, .400E+01_EB, .396E+01_EB, .330E+01_EB, .260E+01_EB, &
.620E-01_EB, .143E+01_EB, .347E+01_EB, .388E+01_EB, .341E+01_EB, .280E+01_EB/), &
(/6,8/))
SD(1:6,17:24) = RESHAPE ((/ & ! 450-625
.255E-01_EB, .950E+00_EB, .292E+01_EB, .370E+01_EB, .345E+01_EB, .295E+01_EB,   &
.940E-02_EB, .610E+00_EB, .236E+01_EB, .343E+01_EB, .342E+01_EB, .304E+01_EB,  &
.340E-02_EB, .386E+00_EB, .188E+01_EB, .310E+01_EB, .334E+01_EB, .309E+01_EB,   &
.105E-02_EB, .236E+00_EB, .145E+01_EB, .274E+01_EB, .319E+01_EB, .307E+01_EB,   &
.350E-03_EB, .144E+00_EB, .110E+01_EB, .238E+01_EB, .300E+01_EB, .301E+01_EB,   &
.126E-03_EB, .820E-01_EB, .818E+00_EB, .204E+01_EB, .276E+01_EB, .289E+01_EB,   &
.430E-04_EB, .445E-01_EB, .598E+00_EB, .174E+01_EB, .248E+01_EB, .275E+01_EB,  &
.150E-04_EB, .242E-01_EB, .427E+00_EB, .145E+01_EB, .222E+01_EB, .260E+01_EB/), &
(/6,8/))
SD(1:6,25:32) = RESHAPE ((/  &! 650-825
.510E-05_EB, .127E-01_EB, .294E+00_EB, .118E+01_EB, .195E+01_EB, .241E+01_EB,   &
.170E-05_EB, .630E-02_EB, .200E+00_EB, .950E+00_EB, .169E+01_EB, .221E+01_EB,   &
.570E-06_EB, .300E-02_EB, .134E+00_EB, .748E+00_EB, .146E+01_EB, .200E+01_EB,   &
.195E-06_EB, .140E-02_EB, .902E-01_EB, .580E+00_EB, .124E+01_EB, .178E+01_EB,   &
.680E-07_EB, .620E-03_EB, .590E-01_EB, .443E+00_EB, .103E+01_EB, .156E+01_EB,   &
.385E-07_EB, .275E-03_EB, .450E-01_EB, .330E+00_EB, .845E+00_EB, .136E+01_EB,   &
.670E-07_EB, .113E-03_EB, .355E-01_EB, .242E+00_EB, .695E+00_EB, .117E+01_EB,   &
.113E-06_EB, .500E-04_EB, .289E-01_EB, .174E+00_EB, .560E+00_EB, .100E+01_EB/),  &
(/6,8/))
SD(1:6,33:40) = RESHAPE ((/ & ! 850-1025
.195E-06_EB, .230E-04_EB, .245E-01_EB, .123E+00_EB, .450E+00_EB, .855E+00_EB,   &
.328E-06_EB, .103E-04_EB, .214E-01_EB, .100E+00_EB, .357E+00_EB, .718E+00_EB,  &
.560E-06_EB, .460E-05_EB, .189E-01_EB, .830E-01_EB, .278E+00_EB, .595E+00_EB,  &
.950E-06_EB, .205E-05_EB, .174E-01_EB, .730E-01_EB, .239E+00_EB, .492E+00_EB,  &
.160E-05_EB, .140E-05_EB, .166E-01_EB, .665E-01_EB, .211E+00_EB, .405E+00_EB, &
.275E-05_EB, .350E-05_EB, .165E-01_EB, .630E-01_EB, .195E+00_EB, .352E+00_EB, &
.470E-05_EB, .850E-05_EB, .167E-01_EB, .620E-01_EB, .190E+00_EB, .312E+00_EB, &
.810E-05_EB, .215E-04_EB, .175E-01_EB, .630E-01_EB, .191E+00_EB, .289E+00_EB/), &
(/6,8/))
SD(1:6,41:48) = RESHAPE ((/ & ! 1050-1225
.136E-04_EB, .570E-04_EB, .188E-01_EB, .675E-01_EB, .194E+00_EB, .281E+00_EB,   &
.235E-04_EB, .150E-03_EB, .208E-01_EB, .745E-01_EB, .202E+00_EB, .283E+00_EB,   &
.400E-04_EB, .380E-03_EB, .233E-01_EB, .865E-01_EB, .223E+00_EB, .314E+00_EB,  &
.680E-04_EB, .950E-03_EB, .268E-01_EB, .122E+00_EB, .260E+00_EB, .380E+00_EB,  &
.120E-03_EB, .245E-02_EB, .343E-01_EB, .176E+00_EB, .328E+00_EB, .461E+00_EB,  &
.200E-03_EB, .620E-02_EB, .638E-01_EB, .251E+00_EB, .411E+00_EB, .511E+00_EB,  &
.365E-03_EB, .140E-01_EB, .107E+00_EB, .330E+00_EB, .458E+00_EB, .542E+00_EB,  &
.680E-03_EB, .330E-01_EB, .166E+00_EB, .405E+00_EB, .487E+00_EB, .571E+00_EB/),&
(/6,8/))
SD(1:6,49:56) = RESHAPE ((/ & ! 1250-1425
.130E-02_EB, .635E-01_EB, .244E+00_EB, .459E+00_EB, .535E+00_EB, .557E+00_EB,  &
.250E-02_EB, .123E+00_EB, .341E+00_EB, .477E+00_EB, .502E+00_EB, .562E+00_EB,  &
.500E-02_EB, .212E+00_EB, .407E+00_EB, .547E+00_EB, .531E+00_EB, .514E+00_EB,  &
.103E-01_EB, .285E+00_EB, .489E+00_EB, .592E+00_EB, .497E+00_EB, .486E+00_EB,  &
.219E-01_EB, .328E+00_EB, .491E+00_EB, .558E+00_EB, .489E+00_EB, .485E+00_EB,  &
.485E-01_EB, .345E+00_EB, .505E+00_EB, .521E+00_EB, .477E+00_EB, .484E+00_EB,  &
.114E+00_EB, .361E+00_EB, .538E+00_EB, .563E+00_EB, .503E+00_EB, .502E+00_EB,  &
.249E+00_EB, .460E+00_EB, .621E+00_EB, .624E+00_EB, .538E+00_EB, .538E+00_EB/), &
(/6,8/))
SD(1:6,57:64) = RESHAPE ((/ & ! 1450-1625
.397E+00_EB, .569E+00_EB, .749E+00_EB, .768E+00_EB, .581E+00_EB, .565E+00_EB,  &
.418E+00_EB, .627E+00_EB, .824E+00_EB, .849E+00_EB, .640E+00_EB, .594E+00_EB,  &
.108E+01_EB, .125E+01_EB, .113E+01_EB, .940E+00_EB, .807E+00_EB, .663E+00_EB,  &
.165E+01_EB, .155E+01_EB, .118E+01_EB, .670E+00_EB, .562E+00_EB, .483E+00_EB,  &
.142E+01_EB, .675E+00_EB, .557E+00_EB, .349E+00_EB, .276E+00_EB, .263E+00_EB,  &
.451E+00_EB, .202E+00_EB, .132E+00_EB, .118E+00_EB, .134E+00_EB, .156E+00_EB,  &
.603E-01_EB, .538E-01_EB, .863E-01_EB, .112E+00_EB, .120E+00_EB, .125E+00_EB,  &
.501E+00_EB, .252E+00_EB, .118E+00_EB, .112E+00_EB, .131E+00_EB, .140E+00_EB/), &
(/6,8/))
SD(1:6,65:72) = RESHAPE ((/ & ! 1650-1825
.730E+00_EB, .430E+00_EB, .237E+00_EB, .191E+00_EB, .171E+00_EB, .170E+00_EB,   &
.149E+01_EB, .506E+00_EB, .294E+00_EB, .238E+00_EB, .210E+00_EB, .201E+00_EB,  &
.100E+01_EB, .553E+00_EB, .434E+00_EB, .340E+00_EB, .260E+00_EB, .220E+00_EB,   &
.802E+00_EB, .658E+00_EB, .528E+00_EB, .411E+00_EB, .300E+00_EB, .240E+00_EB,    &
.580E+00_EB, .527E+00_EB, .460E+00_EB, .378E+00_EB, .322E+00_EB, .283E+00_EB,  &
.330E+00_EB, .403E+00_EB, .430E+00_EB, .356E+00_EB, .318E+00_EB, .270E+00_EB,  &
.250E+00_EB, .393E+00_EB, .405E+00_EB, .342E+00_EB, .301E+00_EB, .275E+00_EB,  &
.147E+00_EB, .249E+00_EB, .313E+00_EB, .318E+00_EB, .291E+00_EB, .268E+00_EB/), &
(/6,8/))
SD(1:6,73:80) = RESHAPE ((/ & ! 1850-2025
.910E-01_EB, .252E+00_EB, .298E+00_EB, .295E+00_EB, .269E+00_EB, .253E+00_EB,   &
.580E-01_EB, .158E+00_EB, .214E+00_EB, .244E+00_EB, .244E+00_EB, .245E+00_EB,   &
.370E-01_EB, .113E+00_EB, .184E+00_EB, .218E+00_EB, .214E+00_EB, .218E+00_EB,   &
.244E-01_EB, .118E+00_EB, .156E+00_EB, .188E+00_EB, .195E+00_EB, .200E+00_EB,   &
.162E-01_EB, .606E-01_EB, .976E-01_EB, .141E+00_EB, .166E+00_EB, .179E+00_EB,   &
.112E-01_EB, .425E-01_EB, .903E-01_EB, .133E+00_EB, .148E+00_EB, .156E+00_EB,   &
.780E-02_EB, .400E-01_EB, .765E-01_EB, .112E+00_EB, .129E+00_EB, .137E+00_EB,   &
.540E-02_EB, .352E-01_EB, .647E-01_EB, .876E-01_EB, .110E+00_EB, .118E+00_EB/),  &
(/6,8/))
SD(1:6,81:88) = RESHAPE ((/ & ! 2050-2225
.380E-02_EB, .252E-01_EB, .507E-01_EB, .705E-01_EB, .888E-01_EB, .100E+00_EB,  &
.260E-02_EB, .179E-01_EB, .377E-01_EB, .546E-01_EB, .724E-01_EB, .828E-01_EB,  &
.180E-02_EB, .123E-01_EB, .294E-01_EB, .443E-01_EB, .608E-01_EB, .686E-01_EB,  &
.127E-02_EB, .850E-02_EB, .212E-01_EB, .378E-01_EB, .579E-01_EB, .640E-01_EB,   &
.880E-03_EB, .680E-02_EB, .152E-01_EB, .275E-01_EB, .449E-01_EB, .521E-01_EB,  &
.620E-02_EB, .400E-02_EB, .107E-01_EB, .214E-01_EB, .374E-01_EB, .453E-01_EB,  &
.480E-03_EB, .298E-02_EB, .931E-02_EB, .189E-01_EB, .329E-01_EB, .403E-01_EB,  &
.405E-03_EB, .175E-02_EB, .696E-02_EB, .152E-01_EB, .295E-01_EB, .365E-01_EB/), &
(/6,8/))
SD(1:6,89:96) = RESHAPE ((/ & ! 2250-2425
.321E-03_EB, .120E-02_EB, .452E-02_EB, .101E-01_EB, .252E-01_EB, .331E-01_EB, &
.229E-03_EB, .721E-03_EB, .364E-02_EB, .930E-02_EB, .225E-01_EB, .305E-01_EB,  &
.195E-03_EB, .544E-03_EB, .318E-02_EB, .750E-02_EB, .202E-01_EB, .284E-01_EB, &
.154E-03_EB, .375E-03_EB, .185E-02_EB, .603E-02_EB, .175E-01_EB, .269E-01_EB, &
.101E-03_EB, .263E-03_EB, .119E-02_EB, .480E-02_EB, .156E-01_EB, .253E-01_EB, &
.852E-04_EB, .185E-03_EB, .909E-03_EB, .360E-02_EB, .133E-01_EB, .241E-01_EB,  &
.763E-04_EB, .137E-03_EB, .711E-03_EB, .316E-02_EB, .122E-01_EB, .237E-01_EB,  &
.615E-04_EB, .126E-03_EB, .610E-03_EB, .257E-02_EB, .101E-01_EB, .218E-01_EB/), &
(/6,8/))
SD(1:6,97:104) = RESHAPE ((/ & ! 2450-2625
.480E-04_EB, .113E-03_EB, .518E-03_EB, .201E-02_EB, .920E-02_EB, .200E-01_EB, &
.372E-04_EB, .106E-03_EB, .435E-03_EB, .168E-02_EB, .785E-02_EB, .183E-01_EB,  &
.355E-04_EB, .101E-03_EB, .376E-03_EB, .168E-02_EB, .669E-02_EB, .166E-01_EB, &
.358E-04_EB, .990E-04_EB, .366E-03_EB, .167E-02_EB, .651E-02_EB, .156E-01_EB,  &
.389E-04_EB, .102E-03_EB, .376E-03_EB, .167E-02_EB, .641E-02_EB, .152E-01_EB,  &
.422E-04_EB, .106E-03_EB, .373E-03_EB, .168E-02_EB, .656E-02_EB, .150E-01_EB,  &
.521E-04_EB, .111E-03_EB, .371E-03_EB, .170E-02_EB, .673E-02_EB, .152E-01_EB,  &
.646E-04_EB, .121E-03_EB, .384E-03_EB, .179E-02_EB, .798E-02_EB, .179E-01_EB/), &
(/6,8/))
SD(1:6,105:112) = RESHAPE ((/ & ! 2650-2825
.742E-04_EB, .129E-03_EB, .479E-03_EB, .201E-02_EB, .788E-02_EB, .175E-01_EB,   &
.953E-04_EB, .165E-03_EB, .544E-03_EB, .249E-02_EB, .945E-02_EB, .204E-01_EB,   &
.101E-03_EB, .190E-03_EB, .761E-03_EB, .324E-02_EB, .106E-01_EB, .231E-01_EB,   &
.147E-03_EB, .272E-03_EB, .892E-03_EB, .441E-02_EB, .125E-01_EB, .257E-01_EB,    &
.195E-03_EB, .326E-03_EB, .100E-02_EB, .499E-02_EB, .147E-01_EB, .295E-01_EB,    &
.261E-03_EB, .421E-03_EB, .145E-02_EB, .568E-02_EB, .161E-01_EB, .306E-01_EB,   &
.305E-03_EB, .515E-03_EB, .195E-02_EB, .754E-02_EB, .185E-01_EB, .363E-01_EB,   &
.362E-03_EB, .645E-03_EB, .237E-02_EB, .830E-02_EB, .205E-01_EB, .373E-01_EB/), &
(/6,8/))
SD(1:6,113:120) = RESHAPE ((/ & ! 2850-3025
.507E-03_EB, .850E-03_EB, .274E-02_EB, .888E-02_EB, .234E-01_EB, .431E-01_EB,     &
.799E-03_EB, .118E-02_EB, .322E-02_EB, .110E-01_EB, .262E-01_EB, .451E-01_EB,     &
.935E-03_EB, .160E-02_EB, .386E-02_EB, .126E-01_EB, .292E-01_EB, .530E-01_EB,     &
.108E-02_EB, .231E-02_EB, .451E-02_EB, .140E-01_EB, .306E-01_EB, .536E-01_EB,     &
.192E-02_EB, .271E-02_EB, .563E-02_EB, .159E-01_EB, .357E-01_EB, .629E-01_EB,     &
.263E-02_EB, .300E-02_EB, .625E-02_EB, .179E-01_EB, .385E-01_EB, .666E-01_EB,    &
.295E-02_EB, .330E-02_EB, .701E-02_EB, .203E-01_EB, .460E-01_EB, .782E-01_EB,   &
.310E-02_EB, .370E-02_EB, .846E-02_EB, .220E-01_EB, .519E-01_EB, .889E-01_EB/),  &
(/6,8/))
SD(1:6,121:128) = RESHAPE ((/ & ! 3050-3225
.340E-02_EB, .400E-02_EB, .969E-02_EB, .279E-01_EB, .662E-01_EB, .109E+00_EB,     &
.730E-02_EB, .450E-02_EB, .111E-01_EB, .272E-01_EB, .676E-01_EB, .109E+00_EB,     &
.900E-02_EB, .480E-02_EB, .137E-01_EB, .372E-01_EB, .864E-01_EB, .133E+00_EB,     &
.100E-02_EB, .510E-02_EB, .162E-01_EB, .471E-01_EB, .100E+00_EB, .142E+00_EB,     &
.640E-03_EB, .550E-02_EB, .205E-01_EB, .530E-01_EB, .122E+00_EB, .168E+00_EB,     &
.160E-02_EB, .600E-02_EB, .247E-01_EB, .633E-01_EB, .135E+00_EB, .177E+00_EB,    &
.330E-02_EB, .700E-02_EB, .283E-01_EB, .770E-01_EB, .153E+00_EB, .185E+00_EB,   &
.410E-02_EB, .860E-02_EB, .376E-01_EB, .914E-01_EB, .166E+00_EB, .206E+00_EB/),&
(/6,8/))
SD(1:6,129:136) = RESHAPE ((/ & ! 3250-3425
.410E-02_EB, .103E-01_EB, .514E-01_EB, .117E+00_EB, .194E+00_EB, .228E+00_EB,     &
.290E-02_EB, .129E-01_EB, .664E-01_EB, .147E+00_EB, .220E+00_EB, .254E+00_EB,     &
.220E-02_EB, .161E-01_EB, .834E-01_EB, .171E+00_EB, .237E+00_EB, .263E+00_EB,     &
.220E-02_EB, .212E-01_EB, .103E+00_EB, .201E+00_EB, .268E+00_EB, .283E+00_EB,     &
.250E-02_EB, .285E-01_EB, .135E+00_EB, .240E+00_EB, .295E+00_EB, .295E+00_EB,     &
.310E-02_EB, .385E-01_EB, .169E+00_EB, .272E+00_EB, .312E+00_EB, .301E+00_EB,     &
.420E-02_EB, .540E-01_EB, .214E+00_EB, .309E+00_EB, .329E+00_EB, .307E+00_EB,     &
.600E-02_EB, .770E-01_EB, .267E+00_EB, .343E+00_EB, .332E+00_EB, .314E+00_EB/),  &
(/6,8/))
SD(1:6,137:144) = RESHAPE ((/ & ! 3450-3625
.940E-02_EB, .117E+00_EB, .333E+00_EB, .372E+00_EB, .344E+00_EB, .303E+00_EB,     &
.165E-01_EB, .173E+00_EB, .365E+00_EB, .385E+00_EB, .353E+00_EB, .300E+00_EB,     &
.360E-01_EB, .258E+00_EB, .438E+00_EB, .393E+00_EB, .315E+00_EB, .288E+00_EB,     &
.720E-01_EB, .375E+00_EB, .510E+00_EB, .409E+00_EB, .294E+00_EB, .271E+00_EB,     &
.133E+00_EB, .401E+00_EB, .499E+00_EB, .390E+00_EB, .281E+00_EB, .257E+00_EB,     &
.215E+00_EB, .500E+00_EB, .443E+00_EB, .341E+00_EB, .254E+00_EB, .230E+00_EB,     &
.318E+00_EB, .450E+00_EB, .346E+00_EB, .286E+00_EB, .245E+00_EB, .219E+00_EB,     &
.442E+00_EB, .400E+00_EB, .354E+00_EB, .279E+00_EB, .233E+00_EB, .216E+00_EB/),   &
(/6,8/))
SD(1:6,145:152) = RESHAPE ((/ & ! 3650-3825
.473E+00_EB, .405E+00_EB, .347E+00_EB, .281E+00_EB, .238E+00_EB, .219E+00_EB,     &
.568E+00_EB, .501E+00_EB, .423E+00_EB, .315E+00_EB, .243E+00_EB, .218E+00_EB,     &
.690E+00_EB, .708E+00_EB, .673E+00_EB, .432E+00_EB, .268E+00_EB, .189E+00_EB,     &
.617E+00_EB, .831E+00_EB, .566E+00_EB, .320E+00_EB, .194E+00_EB, .123E+00_EB,     &
.181E+01_EB, .520E+00_EB, .200E+00_EB, .131E+00_EB, .124E+00_EB, .107E+00_EB,     &
.136E+00_EB, .124E+00_EB, .120E+00_EB, .119E+00_EB, .115E+00_EB, .115E+00_EB,     &
.455E+00_EB, .298E+00_EB, .167E+00_EB, .129E+00_EB, .123E+00_EB, .112E+00_EB,     &
.760E+00_EB, .503E+00_EB, .242E+00_EB, .154E+00_EB, .129E+00_EB, .127E+00_EB/),   &
(/6,8/))
SD(1:6,153:160) = RESHAPE ((/ & ! 3850-4025
.836E+00_EB, .584E+00_EB, .277E+00_EB, .184E+00_EB, .161E+00_EB, .145E+00_EB,     &
.840E+00_EB, .728E+00_EB, .422E+00_EB, .236E+00_EB, .197E+00_EB, .167E+00_EB,     &
.505E+00_EB, .500E+00_EB, .379E+00_EB, .276E+00_EB, .227E+00_EB, .192E+00_EB,     &
.117E+00_EB, .400E+00_EB, .423E+00_EB, .315E+00_EB, .243E+00_EB, .202E+00_EB,     &
.460E-01_EB, .300E+00_EB, .358E+00_EB, .290E+00_EB, .230E+00_EB, .202E+00_EB,     &
.183E-01_EB, .205E+00_EB, .269E+00_EB, .235E+00_EB, .195E+00_EB, .192E+00_EB,     &
.730E-02_EB, .135E+00_EB, .186E+00_EB, .179E+00_EB, .159E+00_EB, .168E+00_EB,     &
.557E-02_EB, .790E-01_EB, .113E+00_EB, .124E+00_EB, .124E+00_EB, .134E+00_EB/),  &
(/6,8/))
SD(1:6,161:168) = RESHAPE ((/ & ! 4050-4225
.283E-02_EB, .415E-01_EB, .662E-01_EB, .886E-01_EB, .103E+00_EB, .106E+00_EB,     &
.226E-02_EB, .197E-01_EB, .367E-01_EB, .594E-01_EB, .801E-01_EB, .879E-01_EB,     &
.155E-02_EB, .860E-02_EB, .211E-01_EB, .395E-01_EB, .503E-01_EB, .610E-01_EB,     &
.103E-02_EB, .521E-02_EB, .119E-01_EB, .246E-01_EB, .354E-01_EB, .480E-01_EB,     &
.821E-03_EB, .365E-02_EB, .759E-02_EB, .166E-01_EB, .258E-01_EB, .370E-01_EB,     &
.752E-03_EB, .183E-02_EB, .445E-02_EB, .100E-01_EB, .179E-01_EB, .268E-01_EB,     &
.429E-03_EB, .141E-02_EB, .354E-02_EB, .821E-02_EB, .142E-01_EB, .212E-01_EB,     &
.327E-03_EB, .902E-03_EB, .209E-02_EB, .588E-02_EB, .112E-01_EB, .172E-01_EB/),   &
(/6,8/))
SD(1:6,169:176) = RESHAPE ((/ & ! 4250-4425
.225E-03_EB, .685E-03_EB, .189E-02_EB, .512E-02_EB, .101E-01_EB, .164E-01_EB,     &
.186E-03_EB, .551E-03_EB, .156E-02_EB, .366E-02_EB, .812E-02_EB, .136E-01_EB,     &
.173E-03_EB, .472E-03_EB, .139E-02_EB, .306E-02_EB, .661E-02_EB, .115E-01_EB,     &
.138E-03_EB, .395E-03_EB, .110E-02_EB, .272E-02_EB, .587E-02_EB, .104E-01_EB,     &
.900E-04_EB, .270E-03_EB, .968E-03_EB, .222E-02_EB, .497E-02_EB, .921E-02_EB,     &
.752E-04_EB, .233E-03_EB, .744E-03_EB, .208E-02_EB, .466E-02_EB, .876E-02_EB,     &
.618E-04_EB, .175E-03_EB, .638E-03_EB, .185E-02_EB, .465E-02_EB, .914E-02_EB,     &
.504E-04_EB, .134E-03_EB, .499E-03_EB, .174E-02_EB, .455E-02_EB, .935E-02_EB/),   &
(/6,8/))
SD(1:6,177:184) = RESHAPE ((/ & ! 4450-4625
.375E-04_EB, .123E-03_EB, .485E-03_EB, .182E-02_EB, .456E-02_EB, .971E-02_EB,     &
.305E-04_EB, .892E-04_EB, .338E-03_EB, .134E-02_EB, .460E-02_EB, .104E-01_EB,     &
.257E-04_EB, .790E-04_EB, .329E-03_EB, .154E-02_EB, .477E-02_EB, .112E-01_EB,     &
.242E-04_EB, .740E-04_EB, .308E-03_EB, .135E-02_EB, .497E-02_EB, .122E-01_EB,     &
.215E-04_EB, .653E-04_EB, .282E-03_EB, .131E-02_EB, .521E-02_EB, .133E-01_EB,     &
.218E-04_EB, .660E-04_EB, .272E-03_EB, .152E-02_EB, .573E-02_EB, .148E-01_EB,     &
.215E-04_EB, .671E-04_EB, .268E-03_EB, .134E-02_EB, .607E-02_EB, .159E-01_EB,     &
.217E-04_EB, .695E-04_EB, .285E-03_EB, .161E-02_EB, .677E-02_EB, .173E-01_EB/),  &
(/6,8/))
SD(1:6,185:192) = RESHAPE ((/ & ! 4650-4825
.219E-04_EB, .722E-04_EB, .297E-03_EB, .169E-02_EB, .783E-02_EB, .197E-01_EB,     &
.226E-04_EB, .771E-04_EB, .341E-03_EB, .236E-02_EB, .925E-02_EB, .226E-01_EB,     &
.250E-04_EB, .815E-04_EB, .387E-03_EB, .286E-02_EB, .106E-01_EB, .250E-01_EB,    &
.280E-04_EB, .845E-04_EB, .420E-03_EB, .357E-02_EB, .124E-01_EB, .276E-01_EB,     &
.351E-04_EB, .192E-03_EB, .470E-03_EB, .467E-02_EB, .166E-01_EB, .313E-01_EB,     &
.435E-04_EB, .200E-03_EB, .105E-02_EB, .566E-02_EB, .185E-01_EB, .341E-01_EB,     &
.522E-04_EB, .233E-03_EB, .129E-02_EB, .736E-02_EB, .229E-01_EB, .378E-01_EB,     &
.673E-04_EB, .306E-03_EB, .183E-02_EB, .982E-02_EB, .258E-01_EB, .404E-01_EB/),   &
(/6,8/))
SD(1:6,193:200) = RESHAPE ((/ & ! 4850-5025
.886E-04_EB, .399E-03_EB, .246E-02_EB, .128E-01_EB, .302E-01_EB, .430E-01_EB,     &
.113E-03_EB, .618E-03_EB, .346E-02_EB, .161E-01_EB, .358E-01_EB, .459E-01_EB,     &
.174E-03_EB, .825E-03_EB, .441E-02_EB, .200E-01_EB, .417E-01_EB, .493E-01_EB,     &
.265E-03_EB, .163E-02_EB, .777E-02_EB, .245E-01_EB, .450E-01_EB, .507E-01_EB,     &
.355E-03_EB, .200E-02_EB, .978E-02_EB, .317E-01_EB, .492E-01_EB, .527E-01_EB,     &
.538E-03_EB, .271E-02_EB, .167E-01_EB, .401E-01_EB, .503E-01_EB, .523E-01_EB,     &
.651E-03_EB, .301E-02_EB, .264E-01_EB, .467E-01_EB, .520E-01_EB, .526E-01_EB,     &
.987E-03_EB, .530E-02_EB, .321E-01_EB, .499E-01_EB, .523E-01_EB, .510E-01_EB/),   &
(/6,8/))
SD(1:6,201:208) = RESHAPE ((/ & ! 5050-5225
.135E-02_EB, .860E-02_EB, .389E-01_EB, .528E-01_EB, .513E-01_EB, .492E-01_EB,    &
.226E-02_EB, .130E-01_EB, .472E-01_EB, .559E-01_EB, .500E-01_EB, .469E-01_EB,     &
.431E-02_EB, .198E-01_EB, .526E-01_EB, .557E-01_EB, .480E-01_EB, .452E-01_EB,     &
.628E-02_EB, .282E-01_EB, .488E-01_EB, .495E-01_EB, .451E-01_EB, .430E-01_EB,     &
.900E-02_EB, .390E-01_EB, .471E-01_EB, .449E-01_EB, .430E-01_EB, .423E-01_EB,     &
.180E-01_EB, .462E-01_EB, .412E-01_EB, .391E-01_EB, .403E-01_EB, .415E-01_EB,     &
.348E-01_EB, .710E-01_EB, .402E-01_EB, .360E-01_EB, .384E-01_EB, .414E-01_EB,     &
.718E-01_EB, .590E-01_EB, .399E-01_EB, .360E-01_EB, .376E-01_EB, .420E-01_EB/),   &
(/6,8/))
SD(1:6,209:216) = RESHAPE ((/ & ! 5250-5425
.111E+00_EB, .368E-01_EB, .340E-01_EB, .369E-01_EB, .409E-01_EB, .454E-01_EB,     &
.329E-01_EB, .285E-01_EB, .365E-01_EB, .423E-01_EB, .461E-01_EB, .482E-01_EB,     &
.281E-01_EB, .270E-01_EB, .432E-01_EB, .505E-01_EB, .529E-01_EB, .511E-01_EB,     &
.121E+00_EB, .422E-01_EB, .589E-01_EB, .598E-01_EB, .572E-01_EB, .544E-01_EB,     &
.139E+00_EB, .105E+00_EB, .844E-01_EB, .687E-01_EB, .593E-01_EB, .560E-01_EB,     &
.774E-01_EB, .710E-01_EB, .683E-01_EB, .618E-01_EB, .556E-01_EB, .534E-01_EB,     &
.858E-01_EB, .483E-01_EB, .579E-01_EB, .547E-01_EB, .503E-01_EB, .495E-01_EB,     &
.985E-01_EB, .575E-01_EB, .589E-01_EB, .510E-01_EB, .451E-01_EB, .449E-01_EB/),   &
(/6,8/))
SD(1:6,217:224) = RESHAPE ((/ & ! 5450-5625
.996E-01_EB, .682E-01_EB, .539E-01_EB, .489E-01_EB, .454E-01_EB, .446E-01_EB,     &
.680E-01_EB, .680E-01_EB, .548E-01_EB, .495E-01_EB, .460E-01_EB, .458E-01_EB,     &
.325E-01_EB, .520E-01_EB, .515E-01_EB, .483E-01_EB, .449E-01_EB, .454E-01_EB,     &
.150E-01_EB, .350E-01_EB, .451E-01_EB, .464E-01_EB, .452E-01_EB, .449E-01_EB,     &
.620E-02_EB, .238E-01_EB, .369E-01_EB, .408E-01_EB, .414E-01_EB, .417E-01_EB,     &
.270E-02_EB, .158E-01_EB, .282E-01_EB, .339E-01_EB, .366E-01_EB, .384E-01_EB,     &
.113E-02_EB, .101E-01_EB, .203E-01_EB, .263E-01_EB, .303E-01_EB, .333E-01_EB,     &
.829E-03_EB, .590E-02_EB, .148E-01_EB, .206E-01_EB, .247E-01_EB, .295E-01_EB/),  &
(/6,8/))
SD(1:6,225:232) = RESHAPE ((/ & ! 5650-5825
.365E-03_EB, .310E-02_EB, .969E-02_EB, .154E-01_EB, .203E-01_EB, .258E-01_EB,     &
.240E-03_EB, .130E-02_EB, .589E-02_EB, .112E-01_EB, .164E-01_EB, .222E-01_EB,     &
.158E-03_EB, .400E-03_EB, .417E-02_EB, .850E-02_EB, .134E-01_EB, .190E-01_EB,     &
.103E-03_EB, .262E-03_EB, .208E-02_EB, .594E-02_EB, .109E-01_EB, .162E-01_EB,     &
.741E-04_EB, .181E-03_EB, .142E-02_EB, .455E-02_EB, .907E-02_EB, .141E-01_EB,     &
.625E-04_EB, .135E-03_EB, .816E-03_EB, .316E-02_EB, .698E-02_EB, .121E-01_EB,     &
.499E-04_EB, .111E-03_EB, .624E-03_EB, .230E-02_EB, .551E-02_EB, .102E-01_EB,     &
.325E-04_EB, .677E-04_EB, .425E-03_EB, .124E-02_EB, .385E-02_EB, .818E-02_EB/),   &
(/6,8/))
SD(1:6,233:240) = RESHAPE ((/ & ! 5850-6025
.231E-04_EB, .563E-04_EB, .278E-03_EB, .986E-03_EB, .290E-02_EB, .672E-02_EB,    &
.165E-04_EB, .481E-04_EB, .247E-03_EB, .944E-03_EB, .253E-02_EB, .612E-02_EB,           &
.126E-04_EB, .432E-04_EB, .241E-03_EB, .886E-03_EB, .220E-02_EB, .582E-02_EB,          &
.118E-04_EB, .420E-04_EB, .235E-03_EB, .847E-03_EB, .209E-02_EB, .571E-02_EB,          &
.110E-04_EB, .408E-04_EB, .226E-03_EB, .812E-03_EB, .221E-02_EB, .604E-02_EB,         &
.101E-04_EB, .400E-04_EB, .213E-03_EB, .805E-03_EB, .239E-02_EB, .641E-02_EB,        &
.983E-05_EB, .395E-04_EB, .186E-03_EB, .801E-03_EB, .247E-02_EB, .691E-02_EB,       &
.979E-05_EB, .401E-04_EB, .193E-03_EB, .805E-03_EB, .260E-02_EB, .732E-02_EB/),    &
(/6,8/))
SD(1:6,241:248) = RESHAPE ((/ & ! 6050-6225
.976E-05_EB, .410E-04_EB, .201E-03_EB, .814E-03_EB, .285E-02_EB, .776E-02_EB,     &
.988E-05_EB, .420E-04_EB, .210E-03_EB, .832E-03_EB, .317E-02_EB, .842E-02_EB,    &
.991E-05_EB, .425E-04_EB, .219E-03_EB, .877E-03_EB, .340E-02_EB, .888E-02_EB,        &
.102E-04_EB, .435E-04_EB, .231E-03_EB, .937E-03_EB, .361E-02_EB, .929E-02_EB,       &
.110E-04_EB, .486E-04_EB, .244E-03_EB, .971E-03_EB, .402E-02_EB, .994E-02_EB,      &
.127E-04_EB, .579E-04_EB, .257E-03_EB, .111E-02_EB, .437E-02_EB, .104E-01_EB,     &
.131E-04_EB, .612E-04_EB, .277E-03_EB, .113E-02_EB, .465E-02_EB, .110E-01_EB,        &
.150E-04_EB, .783E-04_EB, .353E-03_EB, .116E-02_EB, .510E-02_EB, .116E-01_EB/),     &
(/6,8/))
SD(1:6,249:256) = RESHAPE ((/ & ! 6250-6425
.178E-04_EB, .922E-04_EB, .394E-03_EB, .157E-02_EB, .555E-02_EB, .123E-01_EB,      &
.203E-04_EB, .115E-03_EB, .481E-03_EB, .188E-02_EB, .601E-02_EB, .131E-01_EB,     &
.230E-04_EB, .145E-03_EB, .617E-03_EB, .183E-02_EB, .644E-02_EB, .139E-01_EB,      &
.280E-04_EB, .187E-03_EB, .723E-03_EB, .202E-02_EB, .686E-02_EB, .146E-01_EB,     &
.305E-04_EB, .209E-03_EB, .811E-03_EB, .243E-02_EB, .779E-02_EB, .157E-01_EB,     &
.455E-04_EB, .244E-03_EB, .935E-03_EB, .243E-02_EB, .844E-02_EB, .166E-01_EB,    &
.661E-04_EB, .320E-03_EB, .989E-03_EB, .288E-02_EB, .902E-02_EB, .173E-01_EB,       &
.723E-04_EB, .397E-03_EB, .122E-02_EB, .359E-02_EB, .100E-01_EB, .184E-01_EB/),    &
(/6,8/))
SD(1:6,257:264) = RESHAPE ((/ & ! 6450-6625
.847E-04_EB, .481E-03_EB, .143E-02_EB, .429E-02_EB, .108E-01_EB, .192E-01_EB,     &
.103E-03_EB, .591E-03_EB, .174E-02_EB, .488E-02_EB, .116E-01_EB, .200E-01_EB,           &
.131E-03_EB, .703E-03_EB, .247E-02_EB, .549E-02_EB, .124E-01_EB, .205E-01_EB,          &
.165E-03_EB, .872E-03_EB, .265E-02_EB, .641E-02_EB, .131E-01_EB, .211E-01_EB,         &
.205E-03_EB, .110E-02_EB, .298E-02_EB, .749E-02_EB, .140E-01_EB, .218E-01_EB,         &
.253E-03_EB, .130E-02_EB, .346E-02_EB, .811E-02_EB, .150E-01_EB, .230E-01_EB,         &
.338E-03_EB, .150E-02_EB, .445E-02_EB, .890E-02_EB, .159E-01_EB, .237E-01_EB,        &
.437E-03_EB, .170E-02_EB, .491E-02_EB, .107E-01_EB, .170E-01_EB, .245E-01_EB/),     &
(/6,8/))
SD(1:6,265:272) = RESHAPE ((/ & ! 6650-6825
.581E-03_EB, .190E-02_EB, .537E-02_EB, .116E-01_EB, .179E-01_EB, .254E-01_EB,         &
.685E-03_EB, .220E-02_EB, .578E-02_EB, .128E-01_EB, .189E-01_EB, .263E-01_EB,        &
.900E-03_EB, .250E-02_EB, .649E-02_EB, .134E-01_EB, .195E-01_EB, .275E-01_EB,       &
.121E-02_EB, .280E-02_EB, .722E-02_EB, .142E-01_EB, .202E-01_EB, .281E-01_EB,      &
.152E-02_EB, .330E-02_EB, .813E-02_EB, .161E-01_EB, .212E-01_EB, .288E-01_EB,     &
.185E-02_EB, .370E-02_EB, .907E-02_EB, .168E-01_EB, .222E-01_EB, .292E-01_EB,    &
.220E-02_EB, .430E-02_EB, .929E-02_EB, .183E-01_EB, .233E-01_EB, .294E-01_EB,       &
.255E-02_EB, .500E-02_EB, .114E-01_EB, .195E-01_EB, .245E-01_EB, .289E-01_EB/),       &
(/6,8/))
SD(1:6,273:280) = RESHAPE ((/ & ! 6850-7025
.290E-02_EB, .580E-02_EB, .167E-01_EB, .215E-01_EB, .260E-01_EB, .291E-01_EB,        &
.320E-02_EB, .670E-02_EB, .208E-01_EB, .237E-01_EB, .274E-01_EB, .293E-01_EB,        &
.360E-02_EB, .880E-02_EB, .220E-01_EB, .253E-01_EB, .282E-01_EB, .300E-01_EB,       &
.400E-02_EB, .920E-02_EB, .238E-01_EB, .273E-01_EB, .290E-01_EB, .304E-01_EB,        &
.460E-02_EB, .108E-01_EB, .272E-01_EB, .279E-01_EB, .298E-01_EB, .310E-01_EB,       &
.530E-02_EB, .128E-01_EB, .304E-01_EB, .292E-01_EB, .297E-01_EB, .312E-01_EB,        &
.620E-02_EB, .152E-01_EB, .344E-01_EB, .303E-01_EB, .293E-01_EB, .310E-01_EB,       &
.760E-02_EB, .182E-01_EB, .341E-01_EB, .297E-01_EB, .290E-01_EB, .300E-01_EB/),    &
(/6,8/))
SD(1:6,281:288) = RESHAPE ((/ & ! 7050-7225
.980E-02_EB, .222E-01_EB, .398E-01_EB, .318E-01_EB, .291E-01_EB, .294E-01_EB,       &
.132E-01_EB, .271E-01_EB, .402E-01_EB, .294E-01_EB, .274E-01_EB, .282E-01_EB,      &
.190E-01_EB, .335E-01_EB, .421E-01_EB, .286E-01_EB, .262E-01_EB, .269E-01_EB,     &
.240E-01_EB, .432E-01_EB, .431E-01_EB, .276E-01_EB, .245E-01_EB, .257E-01_EB,    &
.288E-01_EB, .570E-01_EB, .458E-01_EB, .270E-01_EB, .228E-01_EB, .243E-01_EB,   &
.323E-01_EB, .740E-01_EB, .449E-01_EB, .261E-01_EB, .214E-01_EB, .221E-01_EB,        &
.570E-01_EB, .890E-01_EB, .435E-01_EB, .225E-01_EB, .199E-01_EB, .196E-01_EB,       &
.216E-01_EB, .680E-01_EB, .378E-01_EB, .239E-01_EB, .195E-01_EB, .192E-01_EB/),    &
(/6,8/))
SD(1:6,289:296) = RESHAPE ((/ & ! 7250-7425
.126E-01_EB, .475E-01_EB, .364E-01_EB, .238E-01_EB, .197E-01_EB, .192E-01_EB,      &
.117E-01_EB, .369E-01_EB, .385E-01_EB, .249E-01_EB, .212E-01_EB, .204E-01_EB,     &
.140E-01_EB, .370E-01_EB, .419E-01_EB, .272E-01_EB, .228E-01_EB, .213E-01_EB,    &
.425E-01_EB, .418E-01_EB, .440E-01_EB, .280E-01_EB, .248E-01_EB, .229E-01_EB,   &
.640E-01_EB, .460E-01_EB, .427E-01_EB, .290E-01_EB, .263E-01_EB, .238E-01_EB,   &
.385E-01_EB, .385E-01_EB, .374E-01_EB, .259E-01_EB, .235E-01_EB, .224E-01_EB,        &
.182E-01_EB, .179E-01_EB, .282E-01_EB, .231E-01_EB, .211E-01_EB, .214E-01_EB,       &
.170E-01_EB, .810E-02_EB, .191E-01_EB, .175E-01_EB, .181E-01_EB, .194E-01_EB/),     &
(/6,8/))
SD(1:6,297:304) = RESHAPE ((/ & ! 7450-7625
.161E-01_EB, .370E-02_EB, .105E-01_EB, .127E-01_EB, .152E-01_EB, .171E-01_EB,      &
.145E-01_EB, .170E-02_EB, .554E-02_EB, .855E-02_EB, .113E-01_EB, .131E-01_EB,     &
.175E-02_EB, .140E-02_EB, .385E-02_EB, .595E-02_EB, .803E-02_EB, .945E-02_EB,    &
.772E-03_EB, .751E-03_EB, .384E-02_EB, .575E-02_EB, .537E-02_EB, .594E-02_EB,       &
.491E-03_EB, .600E-03_EB, .301E-02_EB, .453E-02_EB, .380E-02_EB, .434E-02_EB,      &
.275E-03_EB, .410E-03_EB, .193E-02_EB, .366E-02_EB, .319E-02_EB, .332E-02_EB,     &
.185E-01_EB, .280E-03_EB, .131E-02_EB, .232E-02_EB, .247E-02_EB, .256E-02_EB,      &
.101E-03_EB, .160E-03_EB, .915E-03_EB, .150E-02_EB, .186E-02_EB, .197E-02_EB/),   &
(/6,8/))
SD(1:6,305:312) = RESHAPE ((/ & ! 7650-7825
.691E-04_EB, .110E-03_EB, .565E-03_EB, .114E-02_EB, .205E-02_EB, .192E-02_EB,        &
.476E-04_EB, .750E-04_EB, .114E-02_EB, .124E-02_EB, .175E-02_EB, .187E-02_EB,         &
.305E-04_EB, .590E-04_EB, .529E-03_EB, .114E-02_EB, .160E-02_EB, .185E-02_EB,       &
.240E-04_EB, .480E-04_EB, .293E-03_EB, .842E-03_EB, .141E-02_EB, .184E-02_EB,       &
.170E-04_EB, .360E-04_EB, .122E-03_EB, .435E-03_EB, .124E-02_EB, .182E-02_EB,      &
.120E-04_EB, .240E-04_EB, .121E-03_EB, .435E-03_EB, .118E-02_EB, .187E-02_EB,     &
.810E-05_EB, .170E-04_EB, .103E-03_EB, .439E-03_EB, .126E-02_EB, .192E-02_EB,    &
.550E-05_EB, .120E-04_EB, .866E-04_EB, .367E-03_EB, .119E-02_EB, .193E-02_EB/),    &
(/6,8/))
SD(1:6,313:320) = RESHAPE ((/&  ! 7850-8025
.390E-05_EB, .900E-05_EB, .716E-04_EB, .351E-03_EB, .116E-02_EB, .194E-02_EB,        &
.295E-05_EB, .830E-05_EB, .373E-04_EB, .254E-03_EB, .114E-02_EB, .196E-02_EB,       &
.230E-05_EB, .800E-05_EB, .465E-04_EB, .298E-03_EB, .117E-02_EB, .201E-02_EB,      &
.225E-05_EB, .820E-05_EB, .367E-04_EB, .252E-03_EB, .116E-02_EB, .205E-02_EB,     &
.220E-05_EB, .840E-05_EB, .371E-04_EB, .268E-03_EB, .127E-02_EB, .211E-02_EB,    &
.223E-05_EB, .920E-05_EB, .396E-04_EB, .273E-03_EB, .128E-02_EB, .216E-02_EB,   &
.235E-05_EB, .103E-04_EB, .415E-04_EB, .263E-03_EB, .121E-02_EB, .221E-02_EB,      &
.280E-05_EB, .125E-04_EB, .633E-04_EB, .363E-03_EB, .136E-02_EB, .231E-02_EB/),   &
(/6,8/))
SD(1:6,321:328) = RESHAPE ((/ & ! 8050-8225
.310E-05_EB, .150E-04_EB, .979E-04_EB, .492E-03_EB, .150E-02_EB, .241E-02_EB,           &
.370E-05_EB, .180E-04_EB, .120E-03_EB, .580E-03_EB, .167E-02_EB, .251E-02_EB,           &
.420E-05_EB, .200E-04_EB, .987E-04_EB, .509E-03_EB, .171E-02_EB, .257E-02_EB,           &
.510E-05_EB, .240E-04_EB, .134E-03_EB, .547E-03_EB, .173E-02_EB, .267E-02_EB,           &
.600E-05_EB, .270E-04_EB, .121E-03_EB, .534E-03_EB, .172E-02_EB, .274E-02_EB,           &
.720E-05_EB, .300E-04_EB, .204E-03_EB, .684E-03_EB, .184E-02_EB, .285E-02_EB,           &
.820E-05_EB, .330E-04_EB, .276E-03_EB, .819E-03_EB, .199E-02_EB, .297E-02_EB,           &
.100E-04_EB, .380E-04_EB, .317E-03_EB, .859E-03_EB, .214E-02_EB, .308E-02_EB/),         &
(/6,8/))
SD(1:6,329:336) = RESHAPE ((/ & ! 8250-8425
.125E-04_EB, .420E-04_EB, .240E-03_EB, .818E-03_EB, .220E-02_EB, .317E-02_EB,           &
.145E-04_EB, .500E-04_EB, .452E-03_EB, .109E-02_EB, .238E-02_EB, .293E-02_EB,           &
.175E-04_EB, .560E-04_EB, .301E-03_EB, .941E-03_EB, .243E-02_EB, .342E-02_EB,           &
.198E-04_EB, .630E-04_EB, .280E-03_EB, .107E-02_EB, .260E-02_EB, .353E-02_EB,           &
.230E-04_EB, .710E-04_EB, .276E-03_EB, .109E-02_EB, .272E-02_EB, .365E-02_EB,           &
.280E-04_EB, .830E-04_EB, .369E-03_EB, .127E-02_EB, .295E-02_EB, .377E-02_EB,           &
.330E-04_EB, .890E-04_EB, .430E-03_EB, .139E-02_EB, .306E-02_EB, .385E-02_EB,           &
.360E-04_EB, .950E-04_EB, .371E-03_EB, .135E-02_EB, .306E-02_EB, .384E-02_EB/),         &
(/6,8/))
SD(1:6,337:344) = RESHAPE ((/ & ! 8450-8625
.390E-04_EB, .980E-04_EB, .434E-03_EB, .147E-02_EB, .316E-02_EB, .385E-02_EB,           &
.400E-04_EB, .990E-04_EB, .397E-03_EB, .143E-02_EB, .318E-02_EB, .384E-02_EB,           &
.400E-04_EB, .980E-04_EB, .364E-03_EB, .141E-02_EB, .317E-02_EB, .381E-02_EB,           &
.390E-04_EB, .940E-04_EB, .390E-03_EB, .142E-02_EB, .314E-02_EB, .376E-02_EB,          &
.380E-04_EB, .900E-04_EB, .380E-03_EB, .145E-02_EB, .318E-02_EB, .375E-02_EB,         &
.380E-04_EB, .900E-04_EB, .380E-03_EB, .145E-02_EB, .318E-02_EB, .375E-02_EB,        &
.330E-04_EB, .750E-04_EB, .358E-03_EB, .138E-02_EB, .310E-02_EB, .372E-02_EB,       &
.270E-04_EB, .580E-04_EB, .382E-03_EB, .143E-02_EB, .315E-02_EB, .369E-02_EB/),    &
(/6,8/))
SD(1:6,345:352) = RESHAPE ((/ & ! 8650-8825
.240E-04_EB, .500E-04_EB, .343E-03_EB, .136E-02_EB, .306E-02_EB, .363E-02_EB,          &
.200E-04_EB, .450E-04_EB, .309E-03_EB, .134E-02_EB, .306E-02_EB, .359E-02_EB,         &
.180E-04_EB, .400E-04_EB, .281E-03_EB, .127E-02_EB, .294E-02_EB, .341E-02_EB,        &
.170E-04_EB, .360E-04_EB, .276E-03_EB, .124E-02_EB, .290E-02_EB, .336E-02_EB,       &
.160E-04_EB, .310E-04_EB, .272E-03_EB, .122E-02_EB, .283E-02_EB, .323E-02_EB,      &
.140E-04_EB, .280E-04_EB, .241E-03_EB, .117E-02_EB, .273E-02_EB, .309E-02_EB,       &
.120E-04_EB, .250E-04_EB, .237E-03_EB, .115E-02_EB, .269E-02_EB, .297E-02_EB,      &
.100E-04_EB, .220E-04_EB, .218E-03_EB, .111E-02_EB, .259E-02_EB, .284E-02_EB/),   &
(/6,8/))
SD(1:6,353:360) = RESHAPE ((/ & ! 8850-9025
.920E-05_EB, .198E-04_EB, .206E-03_EB, .105E-02_EB, .246E-02_EB, .269E-02_EB,       &
.810E-05_EB, .170E-04_EB, .205E-03_EB, .100E-02_EB, .235E-02_EB, .257E-02_EB,       &
.720E-05_EB, .160E-04_EB, .177E-03_EB, .921E-03_EB, .220E-02_EB, .245E-02_EB,      &
.650E-05_EB, .150E-04_EB, .172E-03_EB, .834E-03_EB, .205E-02_EB, .232E-02_EB,     &
.590E-05_EB, .130E-04_EB, .147E-03_EB, .735E-03_EB, .194E-02_EB, .218E-02_EB,    &
.510E-05_EB, .110E-04_EB, .120E-03_EB, .629E-03_EB, .177E-02_EB, .203E-02_EB,     &
.460E-05_EB, .950E-05_EB, .960E-04_EB, .513E-03_EB, .154E-02_EB, .180E-02_EB,      &
.420E-05_EB, .800E-05_EB, .578E-04_EB, .314E-03_EB, .123E-02_EB, .154E-02_EB/),     &
(/6,8/))
SD(1:6,361:368) = RESHAPE ((/ & ! 9050-9225
.380E-05_EB, .720E-05_EB, .529E-04_EB, .292E-03_EB, .114E-02_EB, .137E-02_EB,      &
.330E-05_EB, .660E-05_EB, .485E-04_EB, .269E-03_EB, .102E-02_EB, .122E-02_EB,      &
.290E-05_EB, .580E-05_EB, .430E-04_EB, .239E-03_EB, .896E-03_EB, .107E-02_EB,      &
.270E-05_EB, .520E-05_EB, .259E-04_EB, .193E-03_EB, .748E-03_EB, .944E-03_EB,     &
.240E-05_EB, .450E-05_EB, .316E-04_EB, .207E-03_EB, .671E-02_EB, .848E-03_EB,      &
.220E-05_EB, .400E-05_EB, .444E-05_EB, .602E-04_EB, .516E-03_EB, .750E-03_EB,     &
.190E-05_EB, .360E-05_EB, .324E-05_EB, .460E-04_EB, .439E-03_EB, .688E-03_EB,     &
.170E-05_EB, .320E-05_EB, .180E-05_EB, .321E-04_EB, .384E-03_EB, .653E-03_EB/),  &
(/6,8/))
SD(1:6,369:376) = RESHAPE ((/ & ! 9250-9300
.140E-05_EB, .280E-05_EB, .171E-05_EB, .344E-04_EB, .340E-03_EB, .616E-03_EB,   &
.130E-05_EB, .250E-05_EB, .299E-05_EB, .600E-04_EB, .343E-03_EB, .619E-03_EB,  &
.120E-05_EB, .220E-05_EB, .299E-05_EB, .600E-04_EB, .343E-03_EB, .619E-03_EB, &
1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB,&
1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB, 1._EB/),&
(/6,8/))

! INTIALIZE GAMMA ARRAY

GAMMA = RESHAPE((/ &
!     LINE BROADENING PARAMETERS,GAMMA(I,J),
!     J=CO2,H2O,CH4,CO,O2,N2,SELF RESONANT.
!   I=     CO2  H2O  CH4   CO
!    J
.09_EB , .12_EB, .0_EB , .07_EB,&
.07_EB , .09_EB, .0_EB , .06_EB,&
!.0_EB  , .0_EB , .16_EB, .0_EB ,&
.0_EB  , .0_EB , .12_EB, .0_EB ,&
.06_EB , .10_EB, .0_EB , .06_EB,&
.055_EB, .04_EB, .0_EB , .05_EB,&
.07_EB , .09_EB, .0_EB , .06_EB, &
.01_EB , .44_EB, .0_EB , .0_EB /),(/4,7/))

! INITIALIZE SD15 ARRAY

!  THE FOLLOWING ARE DATA FOR THE 15.0 MICRON BAND OF CO2
!  TEMP, K=300     600      1200      1500  1800    2400

SD15(1:6,1:8) = RESHAPE ((/  &! 500-535
.000E+00_EB, .000E+00_EB, .000E+00_EB, .105E-01_EB, .300E-01_EB, .880E-01_EB,          &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .180E-01_EB, .490E-01_EB, .880E-01_EB,         &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .300E-01_EB, .540E-01_EB, .740E-01_EB,        &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .300E-01_EB, .560E-01_EB, .890E-01_EB,       &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .330E-01_EB, .690E-01_EB, .990E-01_EB,      &
.000E+00_EB, .000E+00_EB, .880E-02_EB, .380E-01_EB, .720E-01_EB, .970E-01_EB,     &
.000E+00_EB, .000E+00_EB, .110E-01_EB, .530E-01_EB, .950E-01_EB, .124E+00_EB,    &
.000E+00_EB, .000E+00_EB, .285E-01_EB, .630E-01_EB, .990E-01_EB, .140E+00_EB/), &
(/6,8/))
SD15(1:6,9:16) = RESHAPE ((/  &! 540-575
.000E+00_EB, .000E+00_EB, .330E-01_EB, .680E-01_EB, .103E+00_EB, .134E+00_EB,          &
.000E+00_EB, .000E+00_EB, .450E-01_EB, .920E-01_EB, .138E+00_EB, .176E+00_EB,          &
.000E+00_EB, .000E+00_EB, .490E-01_EB, .970E-01_EB, .148E+00_EB, .191E+00_EB,         &
.000E+00_EB, .000E+00_EB, .490E-01_EB, .120E-01_EB, .188E+00_EB, .247E+00_EB,        &
.000E+00_EB, .000E+00_EB, .480E-01_EB, .126E+00_EB, .201E+00_EB, .241E+00_EB,       &
.000E+00_EB, .000E+00_EB, .820E-01_EB, .198E+00_EB, .270E+00_EB, .265E+00_EB,      &
.000E+00_EB, .750E-02_EB, .690E-01_EB, .140E+00_EB, .225E+00_EB, .340E+00_EB,     &
.000E+00_EB, .205E-01_EB, .820E-01_EB, .145E+00_EB, .236E+00_EB, .530E+00_EB/),  &
(/6,8/))
SD15(1:6,17:24) = RESHAPE ((/ & ! 580-615
.000E+00_EB, .355E-01_EB, .117E+00_EB, .193E+00_EB, .295E+00_EB, .550E+00_EB,          &
.157E-01_EB, .520E-01_EB, .170E+00_EB, .235E+00_EB, .305E+00_EB, .410E+00_EB,          &
.150E-01_EB, .880E-01_EB, .270E+00_EB, .330E+00_EB, .440E+00_EB, .520E+00_EB,          &
.510E-01_EB, .130E+00_EB, .400E+00_EB, .530E+00_EB, .560E+00_EB, .540E+00_EB,         &
.120E+00_EB, .165E+00_EB, .275E+00_EB, .320E+00_EB, .420E+00_EB, .560E+00_EB,        &
.880E-01_EB, .190E+00_EB, .430E+00_EB, .540E+00_EB, .620E+00_EB, .680E+00_EB,       &
.110E+00_EB, .350E+00_EB, .710E+00_EB, .760E+00_EB, .760E+00_EB, .690E+00_EB,      &
.180E+00_EB, .470E+00_EB, .920E+00_EB, .970E+00_EB, .910E+00_EB, .670E+00_EB/),   &
(/6,8/))
SD15(1:6,25:32) = RESHAPE ((/ & ! 620-655
.970E-01_EB, .265E+00_EB, .610E+00_EB, .720E+00_EB, .780E+00_EB, .730E+00_EB,          &
.175E+00_EB, .380E+00_EB, .720E+00_EB, .790E+00_EB, .830E+00_EB, .840E+00_EB,          &
.370E+00_EB, .640E+00_EB, .920E+00_EB, .960E+00_EB, .980E+00_EB, .940E+00_EB,          &
.590E+00_EB, .840E+00_EB, .107E+01_EB, .110E+01_EB, .111E+01_EB, .106E+01_EB,         &
.940E+00_EB, .103E+01_EB, .115E+01_EB, .115E+01_EB, .115E+01_EB, .118E+01_EB,        &
.196E+01_EB, .177E+01_EB, .146E+01_EB, .136E+01_EB, .132E+01_EB, .139E+01_EB,       &
.345E+01_EB, .282E+01_EB, .198E+01_EB, .172E+01_EB, .156E+01_EB, .148E+01_EB,      &
.282E+01_EB, .248E+01_EB, .200E+01_EB, .190E+01_EB, .186E+01_EB, .205E+01_EB/),   &
(/6,8/))
SD15(1:6,33:40) = RESHAPE ((/  &! 660-695
.254E+01_EB, .234E+01_EB, .184E+01_EB, .176E+01_EB, .174E+01_EB, .203E+01_EB,          &
.142E+02_EB, .860E+01_EB, .370E+01_EB, .260E+01_EB, .196E+01_EB, .142E+01_EB,          &
.450E+01_EB, .570E+01_EB, .580E+01_EB, .520E+01_EB, .350E+01_EB, .420E+01_EB,         &
.360E+01_EB, .310E+01_EB, .330E+01_EB, .290E+01_EB, .205E+01_EB, .200E+01_EB,        &
.310E+01_EB, .260E+01_EB, .200E+01_EB, .196E+01_EB, .180E+01_EB, .210E+01_EB,       &
.240E+01_EB, .250E+01_EB, .230E+01_EB, .220E+01_EB, .170E+01_EB, .194E+01_EB,      &
.182E+01_EB, .200E+01_EB, .218E+01_EB, .205E+01_EB, .184E+01_EB, .130E+01_EB,     &
.104E+01_EB, .135E+01_EB, .172E+01_EB, .172E+01_EB, .165E+01_EB, .130E+01_EB/),  &
(/6,8/))
SD15(1:6,41:48) = RESHAPE ((/  &! 700-735
.550E+00_EB, .120E+01_EB, .143E+01_EB, .147E+01_EB, .148E+01_EB, .125E+01_EB,          &
.136E+01_EB, .128E+01_EB, .128E+01_EB, .135E+01_EB, .138E+01_EB, .134E+01_EB,          &
.210E+00_EB, .780E+00_EB, .127E+01_EB, .133E+01_EB, .137E+01_EB, .132E+01_EB,          &
.190E+00_EB, .780E+00_EB, .140E+01_EB, .146E+01_EB, .147E+01_EB, .142E+01_EB,         &
.900E+00_EB, .106E+01_EB, .140E+01_EB, .150E+01_EB, .155E+01_EB, .134E+01_EB,        &
.720E-01_EB, .300E+00_EB, .800E+00_EB, .100E+01_EB, .115E+01_EB, .126E+01_EB,       &
.640E-01_EB, .210E+00_EB, .560E+00_EB, .720E+00_EB, .860E+00_EB, .102E+01_EB,      &
.680E-01_EB, .210E+00_EB, .530E+00_EB, .670E+00_EB, .790E+00_EB, .101E+01_EB/),   &
(/6,8/))
SD15(1:6,49:56) = RESHAPE ((/ & ! 740-775
.690E-01_EB, .210E+00_EB, .540E+00_EB, .690E+00_EB, .820E+00_EB, .910E+00_EB,          &
.330E-01_EB, .140E+00_EB, .390E+00_EB, .530E+00_EB, .690E+00_EB, .770E+00_EB,          &
.230E-01_EB, .780E-01_EB, .270E+00_EB, .410E+00_EB, .560E+00_EB, .890E+00_EB,          &
.300E-01_EB, .860E-01_EB, .280E+00_EB, .400E+00_EB, .520E+00_EB, .710E+00_EB,         &
.175E-01_EB, .620E-01_EB, .225E+00_EB, .335E+00_EB, .450E+00_EB, .660E+00_EB,       &
.105E-01_EB, .450E-01_EB, .180E+00_EB, .280E+00_EB, .380E+00_EB, .600E+00_EB,      &
.450E-02_EB, .300E-01_EB, .148E+00_EB, .240E+00_EB, .345E+00_EB, .570E+00_EB,     &
.000E+00_EB, .140E-01_EB, .124E+00_EB, .205E+00_EB, .285E+00_EB, .430E+00_EB/),  &
(/6,8/))
SD15(1:6,57:64) = RESHAPE ((/ & ! 780-815
.000E+00_EB, .115E-01_EB, .110E+00_EB, .185E+00_EB, .260E+00_EB, .375E+00_EB,          &
.000E+00_EB, .135E-01_EB, .840E-01_EB, .140E+00_EB, .205E+00_EB, .335E+00_EB,         &
.000E+00_EB, .430E-02_EB, .650E-01_EB, .120E+00_EB, .185E+00_EB, .325E+00_EB,        &
.000E+00_EB, .000E+00_EB, .540E-01_EB, .115E+00_EB, .180E+00_EB, .315E+00_EB,       &
.000E+00_EB, .000E+00_EB, .440E-01_EB, .950E-01_EB, .150E+00_EB, .270E+00_EB,      &
.000E+00_EB, .000E+00_EB, .360E-01_EB, .790E-01_EB, .125E+00_EB, .205E+00_EB,     &
.000E+00_EB, .000E+00_EB, .250E-01_EB, .650E-01_EB, .110E+00_EB, .178E+00_EB,    &
.000E+00_EB, .000E+00_EB, .180E-01_EB, .620E-01_EB, .103E+00_EB, .153E+00_EB/), &
(/6,8/))
SD15(1:6,65:72) = RESHAPE ((/ & ! 820-855
.000E+00_EB, .000E+00_EB, .320E-01_EB, .580E-01_EB, .860E-01_EB, .147E+00_EB,          &
.000E+00_EB, .000E+00_EB, .800E-02_EB, .510E-01_EB, .870E-01_EB, .134E+00_EB,          &
.000E+00_EB, .000E+00_EB, .600E-02_EB, .480E-01_EB, .830E-01_EB, .133E+00_EB,          &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .430E-01_EB, .780E-01_EB, .118E+00_EB,          &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .420E-01_EB, .700E-01_EB, .108E+00_EB,         &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .360E-01_EB, .640E-01_EB, .980E-01_EB,        &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .350E-01_EB, .610E-01_EB, .870E-01_EB,       &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .320E-01_EB, .580E-01_EB, .860E-01_EB/),    &
(/6,8/))
SD15(1:6,73:80) = RESHAPE ((/ & ! 860-880
.000E+00_EB, .000E+00_EB, .000E+00_EB, .330E-01_EB, .560E-01_EB, .750E-01_EB,     &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .300E-01_EB, .530E-01_EB, .750E-01_EB,    &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .290E-01_EB, .530E-01_EB, .850E-01_EB,   &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .240E-01_EB, .470E-01_EB, .900E-01_EB,  &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .220E-01_EB, .450E-01_EB, .860E-01_EB, &
.000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB,&
.000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB,&
.000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB, .000E+00_EB/), &
(/6,8/))

! INITIALIZE SD7 ARRAY

!   THE FOLLOWING DATA ARE FOR THE 7.7 MICRON BAND OF CH4
! TEMP,K= 290 600   850

SD7(1:3,1:8) = RESHAPE ((/ &
0._EB, 0._EB, 0._EB,&
0._EB, 0._EB, 0.03_EB,&
0._EB, 0._EB, 0.22_EB,&
0.16_EB, 0.20_EB, 0.47_EB,&
0.34_EB, 0.34_EB, 0.62_EB,&
0.69_EB, 0.53_EB, 0.65_EB,&
1.27_EB, 0.88_EB, 1.09_EB,&
1.68_EB, 1.38_EB, 0.87_EB/),(/3,8/))
SD7(1:3,9:16) = RESHAPE ((/&
0.55_EB, 0.28_EB, 0.40_EB,&
1.25_EB, 0.86_EB, 0.93_EB,&
0.34_EB, 0.59_EB, 0.75_EB,&
0._EB, 0.13_EB, 0.25_EB,&
0._EB, 0._EB, 0.06_EB,&
0._EB, 0._EB, 0._EB,&
0._EB, 0._EB, 0._EB,&
0._EB, 0._EB, 0._EB/),(/3,8/))

! INITIALIZE SD3 ARRAY

!  THE FOLLOWING DATA ARE FOR THE 3.3 MICRON BAND OF CH4
! TEMP, K= 290  600   850
SD3(1:3,1:8) = RESHAPE ((/&
0._EB, 0._EB, 0.03_EB,&
0._EB, 0._EB, 0.03_EB,&
0._EB, 0._EB, 0.03_EB,&
0._EB, 0._EB, 0.06_EB,&
0.03_EB, 0.03_EB, 0.09_EB,&
0.07_EB, 0.07_EB, 0.12_EB,&
0.09_EB, 0.09_EB, 0.12_EB,&
0.14_EB, 0.15_EB, 0.22_EB/),(/3,8/))
SD3(1:3,9:16) = RESHAPE ((/&
0.18_EB, 0.22_EB, 0.28_EB,&
0.24_EB, 0.31_EB, 0.37_EB,&
0.33_EB, 0.44_EB, 0.47_EB,&
0.45_EB, 0.50_EB, 0.53_EB,&
0.59_EB, 0.62_EB, 0.62_EB,&
0.74_EB, 0.70_EB, 0.68_EB,&
0.91_EB, 0.77_EB, 0.72_EB,&
1.00_EB, 0.81_EB, 0.75_EB/),(/3,8/))
SD3(1:3,17:24) = RESHAPE ((/&
1.03_EB, 0.84_EB, 0.78_EB,&
1.03_EB, 0.84_EB, 0.78_EB,&
1.00_EB, 0.81_EB, 0.75_EB,&
0.94_EB, 0.77_EB, 0.72_EB,&
0.72_EB, 0.68_EB, 0.68_EB,&
0.52_EB, 0.63_EB, 0.63_EB,&
0.33_EB, 0.50_EB, 0.56_EB,&
0.25_EB, 0.42_EB, 0.50_EB/),(/3,8/))
SD3(1:3,25:32) = RESHAPE ((/&
0.17_EB, 0.26_EB, 0.37_EB,&
0.08_EB, 0.18_EB, 0.31_EB,&
0.04_EB, 0.11_EB, 0.22_EB,&
0._EB, 0.06_EB, 0.16_EB,&
0._EB, 0.02_EB, 0.12_EB,&
0._EB, 0._EB, 0.06_EB,&
0._EB, 0._EB, 0.03_EB,&
0._EB, 0._EB, 0._EB/),(/3,8/))
!------------------------------------------------------------------------------
END SUBROUTINE RCALLOC


!==============================================================================
SUBROUTINE RCDEALLOC2
!==============================================================================
DEALLOCATE(INCIDENT_RADIANCE)
DEALLOCATE(TTAU)
DEALLOCATE(LAMBDA)
DEALLOCATE(AB)
DEALLOCATE(WAVE_NUMBER)

END SUBROUTINE RCDEALLOC2

!==============================================================================
SUBROUTINE RCDEALLOC
!==============================================================================

DEALLOCATE(GAMMA)
DEALLOCATE(SD15)
DEALLOCATE(SD)
DEALLOCATE(SD7)
DEALLOCATE(SD3)

!-------------------------DEALLOCATION METHANE VARIABLES-------------------
DEALLOCATE(SD_CH4_TEMP)
DEALLOCATE(OM_BND_CH4)
DEALLOCATE(SD1_CH4)
DEALLOCATE(SD2_CH4)

!-------------------------DEALLOCATION PROPANE VARIABLES-------------------
DEALLOCATE(SD_C3H8_TEMP)
DEALLOCATE(BE_C3H8)
DEALLOCATE(OM_BND_C3H8)
DEALLOCATE(SD1_C3H8)
DEALLOCATE(SD2_C3H8)
DEALLOCATE(GAMMAD1_C3H8)
DEALLOCATE(GAMMAD2_C3H8)

!-------------------------DEALLOCATION METHANOL VARIABLES-------------------
DEALLOCATE(SD_CH3OH_TEMP)
DEALLOCATE(BE_CH3OH)
DEALLOCATE(OM_BND_CH3OH)
DEALLOCATE(SD1_CH3OH)
DEALLOCATE(SD2_CH3OH)
DEALLOCATE(SD3_CH3OH)
DEALLOCATE(SD4_CH3OH)
DEALLOCATE(GAMMAD1_CH3OH)
DEALLOCATE(GAMMAD2_CH3OH)
DEALLOCATE(GAMMAD3_CH3OH)
DEALLOCATE(GAMMAD4_CH3OH)

!-------------------------DEALLOCATION HEPTANE VARIABLES-------------------
DEALLOCATE(SD_C7H16_TEMP)
DEALLOCATE(BE_C7H16)
DEALLOCATE(OM_BND_C7H16)
DEALLOCATE(SD1_C7H16)
DEALLOCATE(SD2_C7H16)
DEALLOCATE(GAMMAD1_C7H16)
DEALLOCATE(GAMMAD2_C7H16)

!-------------------------DEALLOCATION TOLUENE VARIABLES-------------------
DEALLOCATE(SD_C7H8_TEMP)
DEALLOCATE(BE_C7H8)
DEALLOCATE(OM_BND_C7H8)
DEALLOCATE(SD1_C7H8)
DEALLOCATE(SD2_C7H8)
DEALLOCATE(SD3_C7H8)
DEALLOCATE(SD4_C7H8)
DEALLOCATE(SD5_C7H8)
DEALLOCATE(GAMMAD1_C7H8)
DEALLOCATE(GAMMAD2_C7H8)
DEALLOCATE(GAMMAD3_C7H8)
DEALLOCATE(GAMMAD4_C7H8)
DEALLOCATE(GAMMAD5_C7H8)

!-------------------------DEALLOCATION PROPYLENE VARIABLES-------------------
DEALLOCATE(SD_C3H6_TEMP)
DEALLOCATE(BE_C3H6)
DEALLOCATE(OM_BND_C3H6)
DEALLOCATE(SD1_C3H6)
DEALLOCATE(SD2_C3H6)
DEALLOCATE(SD3_C3H6)
DEALLOCATE(GAMMAD1_C3H6)
DEALLOCATE(GAMMAD2_C3H6)
DEALLOCATE(GAMMAD3_C3H6)

!-------------------------DEALLOCATION MMA VARIABLES-------------------
DEALLOCATE(SD_C5H8O2_TEMP)
DEALLOCATE(BE_C5H8O2)
DEALLOCATE(OM_BND_C5H8O2)
DEALLOCATE(SD1_C5H8O2)
DEALLOCATE(SD2_C5H8O2)
DEALLOCATE(SD3_C5H8O2)
DEALLOCATE(SD4_C5H8O2)
DEALLOCATE(SD5_C5H8O2)
DEALLOCATE(SD6_C5H8O2)
DEALLOCATE(GAMMAD1_C5H8O2)
DEALLOCATE(GAMMAD2_C5H8O2)
DEALLOCATE(GAMMAD3_C5H8O2)
DEALLOCATE(GAMMAD4_C5H8O2)
DEALLOCATE(GAMMAD5_C5H8O2)
DEALLOCATE(GAMMAD6_C5H8O2)

!-------------------------DEALLOCATION ETHANE VARIABLES-------------------
DEALLOCATE(SD_C2H6_TEMP)
DEALLOCATE(BE_C2H6)
DEALLOCATE(OM_BND_C2H6)
DEALLOCATE(SD1_C2H6)
DEALLOCATE(SD2_C2H6)
DEALLOCATE(SD3_C2H6)
DEALLOCATE(GAMMAD1_C2H6)
DEALLOCATE(GAMMAD2_C2H6)
DEALLOCATE(GAMMAD3_C2H6)

!-------------------------DEALLOCATION ETHYLENE VARIABLES-------------------
DEALLOCATE(SD_C2H4_TEMP)
DEALLOCATE(BE_C2H4)
DEALLOCATE(OM_BND_C2H4)
DEALLOCATE(SD1_C2H4)
DEALLOCATE(SD2_C2H4)
DEALLOCATE(SD3_C2H4)
DEALLOCATE(SD4_C2H4)
DEALLOCATE(GAMMAD1_C2H4)
DEALLOCATE(GAMMAD2_C2H4)
DEALLOCATE(GAMMAD3_C2H4)
DEALLOCATE(GAMMAD4_C2H4)

!------------------------------------------------------------------------------
END SUBROUTINE RCDEALLOC

END MODULE RADCAL_CALC


MODULE SPECDATA

USE PRECISION_PARAMETERS
IMPLICIT NONE
INTEGER ::  I,J
INTEGER, PARAMETER :: NWATERK=183
REAL(EB) :: CPLXREF_WATER(NWATERK,2)
INTEGER, PARAMETER :: NFUELK=94
REAL(EB) :: CPLXREF_FUEL(NFUELK,3)

! HALE, G.M. AND QUERRY, M.R., "OPTICAL CONSTANTS OF WATER IN THE 200-NM TO 200-\MU M WAVELENGTH REGION",
! APPLIED OPTICS, 12(3), 555-63 (1973)

DATA ((CPLXREF_WATER(I,J), J = 1,2), I = 1,10)  &
/ 1.0000000E-6_EB,  2.8647890E-6_EB,&
1.0200000E-6_EB,  2.1915636E-6_EB,&
1.0400000E-6_EB,  1.3241691E-6_EB,&
1.0600000E-6_EB,  1.0122254E-6_EB,&
1.0800000E-6_EB,  1.1172677E-6_EB,&
1.1000000E-6_EB,  1.4880987E-6_EB,&
1.1200000E-6_EB,  4.6345919E-6_EB,&
1.1400000E-6_EB,  5.9874090E-6_EB,&
1.1600000E-6_EB,  8.2155782E-6_EB,&
1.1800000E-6_EB,  9.7657473E-6_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =11,20)  &
/ 1.2000000E-6_EB,  9.9312684E-6_EB,&
1.2200000E-6_EB,  9.2230290E-6_EB,&
1.2400000E-6_EB,  8.6834937E-6_EB,&
1.2600000E-6_EB,  8.9238177E-6_EB,&
1.2800000E-6_EB,  9.9821980E-6_EB,&
1.3000000E-6_EB,  1.1483029E-5_EB,&
1.3200000E-6_EB,  1.1953809E-5_EB,&
1.3400000E-6_EB,  1.2614780E-5_EB,&
1.3600000E-6_EB,  2.9978533E-5_EB,&
1.3800000E-6_EB,  6.6988316E-5_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =21,30)  &
/ 1.4000000E-6_EB,  1.3803508E-4_EB,&
1.4200000E-6_EB,  2.4995602E-4_EB,&
1.4400000E-6_EB,  3.3002369E-4_EB,&
1.4600000E-6_EB,  3.2996003E-4_EB,&
1.4800000E-6_EB,  2.5003560E-4_EB,&
1.5000000E-6_EB,  2.0996516E-4_EB,&
1.5200000E-6_EB,  1.6994565E-4_EB,&
1.5400000E-6_EB,  1.4497583E-4_EB,&
1.5600000E-6_EB,  1.2004433E-4_EB,&
1.5800000E-6_EB,  9.9957388E-5_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =31,40)  &
/ 1.6000000E-6_EB,  8.5561825E-5_EB,&
1.6200000E-6_EB,  7.5028952E-5_EB,&
1.6400000E-6_EB,  6.4992643E-5_EB,&
1.6600000E-6_EB,  5.9972898E-5_EB,&
1.6800000E-6_EB,  6.0027012E-5_EB,&
1.7000000E-6_EB,  6.0065211E-5_EB,&
1.7200000E-6_EB,  6.9942231E-5_EB,&
1.7400000E-6_EB,  8.5017388E-5_EB,&
1.7600000E-6_EB,  1.0000023E-4_EB,&
1.7800000E-6_EB,  1.1501809E-4_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =41,50)  &
/ 1.8000000E-6_EB,  1.1502128E-4_EB,&
1.8200000E-6_EB,  1.3005838E-4_EB,&
1.8400000E-6_EB,  1.4993669E-4_EB,&
1.8600000E-6_EB,  2.1003200E-4_EB,&
1.8800000E-6_EB,  4.6497435E-4_EB,&
1.9000000E-6_EB,  1.0000183E-3_EB,&
1.9200000E-6_EB,  1.7500423E-3_EB,&
1.9400000E-6_EB,  1.8499391E-3_EB,&
1.9600000E-6_EB,  1.6500261E-3_EB,&
1.9800000E-6_EB,  1.4500559E-3_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =51,60)  &
/ 2.0000000E-6_EB,  1.1000790E-3_EB,&
2.0200000E-6_EB,  9.0001961E-4_EB,&
2.0400000E-6_EB,  7.3003417E-4_EB,&
2.0600000E-6_EB,  6.3998112E-4_EB,&
2.0800000E-6_EB,  5.2006742E-4_EB,&
2.1000000E-6_EB,  4.5003447E-4_EB,&
2.1200000E-6_EB,  4.0505888E-4_EB,&
2.1400000E-6_EB,  3.4995785E-4_EB,&
2.1600000E-6_EB,  3.2005422E-4_EB,&
2.1800000E-6_EB,  2.9994500E-4_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =61,70)  &
/ 2.2000000E-6_EB,  2.8904129E-4_EB,&
2.2200000E-6_EB,  2.8495578E-4_EB,&
2.2400000E-6_EB,  2.9500960E-4_EB,&
2.2600000E-6_EB,  3.1005293E-4_EB,&
2.2800000E-6_EB,  3.5997028E-4_EB,&
2.3000000E-6_EB,  4.0998313E-4_EB,&
2.3200000E-6_EB,  4.9496551E-4_EB,&
2.3400000E-6_EB,  5.9494505E-4_EB,&
2.3600000E-6_EB,  6.9994116E-4_EB,&
2.3800000E-6_EB,  8.2007768E-4_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =71,80)  &
/ 2.4000000E-6_EB,  9.5607557E-4_EB,&
2.4200000E-6_EB,  1.1500727E-3_EB,&
2.4400000E-6_EB,  1.2999617E-3_EB,&
2.4600000E-6_EB,  1.4999176E-3_EB,&
2.4800000E-6_EB,  1.6999912E-3_EB,&
2.5000000E-6_EB,  1.8000424E-3_EB,&
2.5200000E-6_EB,  2.0500716E-3_EB,&
2.5400000E-6_EB,  2.1999478E-3_EB,&
2.5600000E-6_EB,  2.3500946E-3_EB,&
2.5800000E-6_EB,  2.7000302E-3_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =81,90)  &
/ 2.6000000E-6_EB,  3.1699367E-3_EB,&
2.6500000E-6_EB,  6.7000889E-3_EB,&
2.7000000E-6_EB,  1.8999997E-2_EB,&
2.7500000E-6_EB,  5.9000926E-2_EB,&
2.8000000E-6_EB,  1.1500027E-1_EB,&
2.8500000E-6_EB,  1.8499960E-1_EB,&
2.9000000E-6_EB,  2.6769861E-1_EB,&
2.9500000E-6_EB,  2.9813700E-1_EB,&
3.0000000E-6_EB,  2.7215495E-1_EB,&
3.0500000E-6_EB,  2.4000020E-1_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =91,100)  &
/ 3.1000000E-6_EB,  1.9200142E-1_EB,&
3.1500000E-6_EB,  1.3500032E-1_EB,&
3.2000000E-6_EB,  9.2401540E-2_EB,&
3.2500000E-6_EB,  6.0999713E-2_EB,&
3.3000000E-6_EB,  3.6798931E-2_EB,&
3.3500000E-6_EB,  2.6099958E-2_EB,&
3.4000000E-6_EB,  1.9500046E-2_EB,&
3.4500000E-6_EB,  1.3199993E-2_EB,&
3.5000000E-6_EB,  9.4000888E-3_EB,&
3.6000000E-6_EB,  5.1500311E-3_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =101,110)  &
/ 3.7000000E-6_EB,  3.6000769E-3_EB,&
3.8000000E-6_EB,  3.4001225E-3_EB,&
3.9000000E-6_EB,  3.7999516E-3_EB,&
4.0000000E-6_EB,  4.5998962E-3_EB,&
4.1000000E-6_EB,  5.6118033E-3_EB,&
4.2000000E-6_EB,  6.8850428E-3_EB,&
4.3000000E-6_EB,  8.4519233E-3_EB,&
4.4000000E-6_EB,  1.0294142E-2_EB,&
4.5000000E-6_EB,  1.3392888E-2_EB,&
4.6000000E-6_EB,  1.4715466E-2_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =111,120)  &
/ 4.7000000E-6_EB,  1.5708593E-2_EB,&
4.8000000E-6_EB,  1.5011494E-2_EB,&
4.9000000E-6_EB,  1.3686529E-2_EB,&
5.0000000E-6_EB,  1.2414086E-2_EB,&
5.1000000E-6_EB,  1.1120156E-2_EB,&
5.2000000E-6_EB,  1.0096790E-2_EB,&
5.3000000E-6_EB,  9.7848459E-3_EB,&
5.4000000E-6_EB,  1.0313240E-2_EB,&
5.5000000E-6_EB,  1.1598416E-2_EB,&
5.6000000E-6_EB,  1.4215720E-2_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =121,130)  &
/ 5.7000000E-6_EB,  2.0320903E-2_EB,&
5.8000000E-6_EB,  3.3000777E-2_EB,&
5.9000000E-6_EB,  6.2209688E-2_EB,&
6.0000000E-6_EB,  1.0699987E-1_EB,&
6.1000000E-6_EB,  1.3101555E-1_EB,&
6.2000000E-6_EB,  8.8019050E-2_EB,&
6.3000000E-6_EB,  5.7002139E-2_EB,&
6.4000000E-6_EB,  4.4868962E-2_EB,&
6.5000000E-6_EB,  3.9207820E-2_EB,&
6.6000000E-6_EB,  3.5609327E-2_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =131,140)  &
/ 6.7000000E-6_EB,  3.3696285E-2_EB,&
6.8000000E-6_EB,  3.2684059E-2_EB,&
6.9000000E-6_EB,  3.2176355E-2_EB,&
7.0000000E-6_EB,  3.1974228E-2_EB,&
7.1000000E-6_EB,  3.1979003E-2_EB,&
7.2000000E-6_EB,  3.2085637E-2_EB,&
7.3000000E-6_EB,  3.2182721E-2_EB,&
7.4000000E-6_EB,  3.2388031E-2_EB,&
7.5000000E-6_EB,  3.2586975E-2_EB,&
7.6000000E-6_EB,  3.2779552E-2_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =141,150)  &
/ 7.7000000E-6_EB,  3.3088313E-2_EB,&
7.8000000E-6_EB,  3.3518031E-2_EB,&
7.9000000E-6_EB,  3.3884883E-2_EB,&
8.0000000E-6_EB,  3.4313806E-2_EB,&
8.2000000E-6_EB,  3.5106397E-2_EB,&
8.4000000E-6_EB,  3.6096341E-2_EB,&
8.6000000E-6_EB,  3.1001791E-2_EB,&
8.8000000E-6_EB,  3.8515496E-2_EB,&
9.0000000E-6_EB,  3.9892186E-2_EB,&
9.2000000E-6_EB,  4.1510792E-2_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =151,160)  &
/ 9.4000000E-6_EB,  4.3310835E-2_EB,&
9.6000000E-6_EB,  4.5378257E-2_EB,&
9.8000000E-6_EB,  4.7883356E-2_EB,&
1.0000000E-5_EB,  5.0770427E-2_EB,&
1.0500000E-5_EB,  6.6176625E-2_EB,&
1.1000000E-5_EB,  9.6813952E-2_EB,&
1.1500000E-5_EB,  1.4202987E-1_EB,&
1.2000000E-5_EB,  1.9900734E-1_EB,&
1.2500000E-5_EB,  2.5902467E-1_EB,&
1.3000000E-5_EB,  3.0497270E-1_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =161,170)  &
/ 1.3500000E-5_EB,  3.4302267E-1_EB,&
1.4000000E-5_EB,  3.6998750E-1_EB,&
1.4500000E-5_EB,  3.8804760E-1_EB,&
1.5000000E-5_EB,  4.0202539E-1_EB,&
1.5500000E-5_EB,  4.1394609E-1_EB,&
1.6000000E-5_EB,  4.2195159E-1_EB,&
1.6500000E-5_EB,  4.2804722E-1_EB,&
1.7000000E-5_EB,  4.2897828E-1_EB,&
1.7500000E-5_EB,  4.2906183E-1_EB,&
1.8000000E-5_EB,  4.2599412E-1_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =171,180)  &
/ 1.8500000E-5_EB,  4.2104440E-1_EB,&
1.9000000E-5_EB,  4.1397792E-1_EB,&
1.9500000E-5_EB,  4.0407849E-1_EB,&
2.0000000E-5_EB,  3.9295355E-1_EB,&
3.0000000E-5_EB,  3.2801834E-1_EB,&
3.8000000E-5_EB,  3.6105890E-1_EB,&
5.0000000E-5_EB,  5.1407047E-1_EB,&
6.0000000E-5_EB,  5.8680428E-1_EB,&
7.0000000E-5_EB,  5.7598174E-1_EB,&
8.0000000E-5_EB,  5.4685638E-1_EB /
DATA ((CPLXREF_WATER(I,J), J = 1,2), I =181,NWATERK)  &
/ 9.0000000E-5_EB,  5.3571554E-1_EB,&
1.0000000E-4_EB,  5.3237328E-1_EB,&
2.0000000E-4_EB,  5.0452117E-1_EB /

! HEPTANE PROPERTIES FROM
! L.A. DOMBROVSKY, S.S. SAZHIN, S.V. MIKHALOVSKY, R. WOOD, M.R. HEIKAL
! SPECTRAL PROPERTIES OF DIESEL FUEL PARTICLES
! FUEL, VOL. 82, NO. 1 (2003) PP. 15-22
DATA (CPLXREF_FUEL( 1,J), J=1,3) / 0.7_EB,   1.47_EB, 1.55E-07_EB /
DATA (CPLXREF_FUEL( 2,J), J=1,3) / 0.8_EB,   1.47_EB, 4.33E-07_EB /
DATA (CPLXREF_FUEL( 3,J), J=1,3) / 0.9_EB,   1.47_EB, 9.39E-07_EB /
DATA (CPLXREF_FUEL( 4,J), J=1,3) / 1.0_EB,   1.47_EB, 8.81E-07_EB /
DATA (CPLXREF_FUEL( 5,J), J=1,3) / 1.1_EB,   1.47_EB, 1.19E-06_EB /
DATA (CPLXREF_FUEL( 6,J), J=1,3) / 1.2_EB,   1.47_EB, 2.63E-06_EB /
DATA (CPLXREF_FUEL( 7,J), J=1,3) / 1.3_EB,   1.47_EB, 1.08E-05_EB /
DATA (CPLXREF_FUEL( 8,J), J=1,3) / 1.4_EB,   1.47_EB, 2.24E-05_EB /
DATA (CPLXREF_FUEL( 9,J), J=1,3) / 1.5_EB,   1.46_EB, 3.41E-05_EB /
DATA (CPLXREF_FUEL( 10,J), J=1,3) / 1.6_EB,  1.46_EB, 4.57E-05_EB /
DATA (CPLXREF_FUEL( 11,J), J=1,3) / 1.7_EB,  1.46_EB, 5.73E-05_EB /
DATA (CPLXREF_FUEL( 12,J), J=1,3) / 1.8_EB,  1.46_EB, 6.90E-05_EB /
DATA (CPLXREF_FUEL( 13,J), J=1,3) / 1.9_EB,  1.46_EB, 8.06E-05_EB /
DATA (CPLXREF_FUEL( 14,J), J=1,3) / 2.0_EB,  1.46_EB, 9.22E-05_EB /
DATA (CPLXREF_FUEL( 15,J), J=1,3) / 2.1_EB,  1.46_EB, 1.04E-04_EB /
DATA (CPLXREF_FUEL( 16,J), J=1,3) / 2.2_EB,  1.46_EB, 1.15E-04_EB /
DATA (CPLXREF_FUEL( 17,J), J=1,3) / 2.3_EB,  1.46_EB, 1.27E-04_EB /
DATA (CPLXREF_FUEL( 18,J), J=1,3) / 2.4_EB,  1.45_EB, 1.39E-04_EB /
DATA (CPLXREF_FUEL( 19,J), J=1,3) / 2.5_EB,  1.45_EB, 1.50E-04_EB /
DATA (CPLXREF_FUEL( 20,J), J=1,3) / 2.6_EB,  1.45_EB, 1.62E-04_EB /
DATA (CPLXREF_FUEL( 21,J), J=1,3) / 2.7_EB,  1.45_EB, 1.11E-04_EB /
DATA (CPLXREF_FUEL( 22,J), J=1,3) / 2.8_EB,  1.44_EB, 5.92E-05_EB /
DATA (CPLXREF_FUEL( 23,J), J=1,3) / 2.9_EB,  1.44_EB, 7.45E-05_EB /
DATA (CPLXREF_FUEL( 24,J), J=1,3) / 3.0_EB,  1.43_EB, 9.72E-05_EB /
DATA (CPLXREF_FUEL( 25,J), J=1,3) / 3.1_EB,  1.42_EB, 3.12E-04_EB /
DATA (CPLXREF_FUEL( 26,J), J=1,3) / 3.2_EB,  1.40_EB, 6.09E-04_EB /
DATA (CPLXREF_FUEL( 27,J), J=1,3) / 3.3_EB,  1.17_EB, 5.72E-02_EB /
DATA (CPLXREF_FUEL( 28,J), J=1,3) / 3.4_EB,  1.39_EB, 1.20E-01_EB /
DATA (CPLXREF_FUEL( 29,J), J=1,3) / 3.5_EB,  1.45_EB, 8.24E-02_EB /
DATA (CPLXREF_FUEL( 30,J), J=1,3) / 3.6_EB,  1.51_EB, 1.63E-03_EB /
DATA (CPLXREF_FUEL( 31,J), J=1,3) / 3.7_EB,  1.64_EB, 1.33E-03_EB /
DATA (CPLXREF_FUEL( 32,J), J=1,3) / 3.8_EB,  1.56_EB, 1.02E-03_EB /
DATA (CPLXREF_FUEL( 33,J), J=1,3) / 3.9_EB,  1.52_EB, 6.36E-04_EB /
DATA (CPLXREF_FUEL( 34,J), J=1,3) / 4.0_EB,  1.50_EB, 2.51E-04_EB /
DATA (CPLXREF_FUEL( 35,J), J=1,3) / 4.1_EB,  1.50_EB, 2.59E-04_EB /
DATA (CPLXREF_FUEL( 36,J), J=1,3) / 4.2_EB,  1.49_EB, 3.10E-04_EB /
DATA (CPLXREF_FUEL( 37,J), J=1,3) / 4.3_EB,  1.49_EB, 2.60E-04_EB /
DATA (CPLXREF_FUEL( 38,J), J=1,3) / 4.4_EB,  1.48_EB, 2.11E-04_EB /
DATA (CPLXREF_FUEL( 39,J), J=1,3) / 4.5_EB,  1.48_EB, 1.98E-04_EB /
DATA (CPLXREF_FUEL( 40,J), J=1,3) / 4.6_EB,  1.47_EB, 1.90E-04_EB /
DATA (CPLXREF_FUEL( 41,J), J=1,3) / 4.7_EB,  1.47_EB, 1.58E-04_EB /
DATA (CPLXREF_FUEL( 42,J), J=1,3) / 4.8_EB,  1.47_EB, 1.21E-04_EB /
DATA (CPLXREF_FUEL( 43,J), J=1,3) / 4.9_EB,  1.47_EB, 8.32E-05_EB /
DATA (CPLXREF_FUEL( 44,J), J=1,3) / 5.0_EB,  1.46_EB, 4.58E-05_EB /
DATA (CPLXREF_FUEL( 45,J), J=1,3) / 5.1_EB,  1.46_EB, 6.37E-05_EB /
DATA (CPLXREF_FUEL( 46,J), J=1,3) / 5.2_EB,  1.46_EB, 7.78E-05_EB /
DATA (CPLXREF_FUEL( 47,J), J=1,3) / 5.3_EB,  1.46_EB, 7.66E-05_EB /
DATA (CPLXREF_FUEL( 48,J), J=1,3) / 5.4_EB,  1.45_EB, 7.55E-05_EB /
DATA (CPLXREF_FUEL( 49,J), J=1,3) / 5.5_EB,  1.45_EB, 9.62E-05_EB /
DATA (CPLXREF_FUEL( 50,J), J=1,3) / 5.6_EB,  1.45_EB, 1.17E-04_EB /
DATA (CPLXREF_FUEL( 51,J), J=1,3) / 5.7_EB,  1.45_EB, 1.63E-04_EB /
DATA (CPLXREF_FUEL( 52,J), J=1,3) / 5.8_EB,  1.45_EB, 2.16E-04_EB /
DATA (CPLXREF_FUEL( 53,J), J=1,3) / 5.9_EB,  1.44_EB, 2.68E-04_EB /
DATA (CPLXREF_FUEL( 54,J), J=1,3) / 6.0_EB,  1.44_EB, 3.21E-04_EB /
DATA (CPLXREF_FUEL( 55,J), J=1,3) / 6.1_EB,  1.44_EB, 3.74E-04_EB /
DATA (CPLXREF_FUEL( 56,J), J=1,3) / 6.2_EB,  1.44_EB, 4.36E-04_EB /
DATA (CPLXREF_FUEL( 57,J), J=1,3) / 6.3_EB,  1.43_EB, 5.87E-04_EB /
DATA (CPLXREF_FUEL( 58,J), J=1,3) / 6.4_EB,  1.43_EB, 7.38E-04_EB /
DATA (CPLXREF_FUEL( 59,J), J=1,3) / 6.5_EB,  1.42_EB, 1.35E-03_EB /
DATA (CPLXREF_FUEL( 60,J), J=1,3) / 6.6_EB,  1.41_EB, 6.12E-03_EB /
DATA (CPLXREF_FUEL( 61,J), J=1,3) / 6.7_EB,  1.39_EB, 2.06E-02_EB /
DATA (CPLXREF_FUEL( 62,J), J=1,3) / 6.8_EB,  1.35_EB, 3.51E-02_EB /
DATA (CPLXREF_FUEL( 63,J), J=1,3) / 6.9_EB,  1.37_EB, 2.29E-02_EB /
DATA (CPLXREF_FUEL( 64,J), J=1,3) / 7.0_EB,  1.48_EB, 3.99E-03_EB /
DATA (CPLXREF_FUEL( 65,J), J=1,3) / 7.1_EB,  1.53_EB, 3.24E-03_EB /
DATA (CPLXREF_FUEL( 66,J), J=1,3) / 7.2_EB,  1.48_EB, 2.61E-03_EB /
DATA (CPLXREF_FUEL( 67,J), J=1,3) / 7.3_EB,  1.43_EB, 2.97E-03_EB /
DATA (CPLXREF_FUEL( 68,J), J=1,3) / 7.4_EB,  1.47_EB, 3.33E-03_EB /
DATA (CPLXREF_FUEL( 69,J), J=1,3) / 7.5_EB,  1.50_EB, 2.80E-03_EB /
DATA (CPLXREF_FUEL( 70,J), J=1,3) / 7.6_EB,  1.49_EB, 2.14E-03_EB /
DATA (CPLXREF_FUEL( 71,J), J=1,3) / 7.7_EB,  1.48_EB, 2.28E-03_EB /
DATA (CPLXREF_FUEL( 72,J), J=1,3) / 7.8_EB,  1.47_EB, 2.41E-03_EB /
DATA (CPLXREF_FUEL( 73,J), J=1,3) / 7.9_EB,  1.47_EB, 1.63E-03_EB /
DATA (CPLXREF_FUEL( 74,J), J=1,3) / 8.0_EB,  1.47_EB, 9.84E-04_EB /
DATA (CPLXREF_FUEL( 75,J), J=1,3) / 8.1_EB,  1.46_EB, 9.03E-04_EB /
DATA (CPLXREF_FUEL( 76,J), J=1,3) / 8.2_EB,  1.46_EB, 8.19E-04_EB /
DATA (CPLXREF_FUEL( 77,J), J=1,3) / 8.3_EB,  1.46_EB, 7.13E-04_EB /
DATA (CPLXREF_FUEL( 78,J), J=1,3) / 8.4_EB,  1.46_EB, 6.07E-04_EB /
DATA (CPLXREF_FUEL( 79,J), J=1,3) / 8.5_EB,  1.45_EB, 8.54E-04_EB /
DATA (CPLXREF_FUEL( 80,J), J=1,3) / 8.6_EB,  1.45_EB, 1.10E-03_EB /
DATA (CPLXREF_FUEL( 81,J), J=1,3) / 8.7_EB,  1.45_EB, 1.76E-03_EB /
DATA (CPLXREF_FUEL( 82,J), J=1,3) / 8.8_EB,  1.45_EB, 2.41E-03_EB /
DATA (CPLXREF_FUEL( 83,J), J=1,3) / 8.9_EB,  1.45_EB, 1.58E-03_EB /
DATA (CPLXREF_FUEL( 84,J), J=1,3) / 9.0_EB,  1.45_EB, 5.71E-04_EB /
DATA (CPLXREF_FUEL( 85,J), J=1,3) / 9.1_EB,  1.45_EB, 8.53E-04_EB /
DATA (CPLXREF_FUEL( 86,J), J=1,3) / 9.2_EB,  1.45_EB, 1.28E-03_EB /
DATA (CPLXREF_FUEL( 87,J), J=1,3) / 9.3_EB,  1.45_EB, 1.63E-03_EB /
DATA (CPLXREF_FUEL( 88,J), J=1,3) / 9.4_EB,  1.45_EB, 1.98E-03_EB /
DATA (CPLXREF_FUEL( 89,J), J=1,3) / 9.5_EB,  1.45_EB, 1.69E-03_EB /
DATA (CPLXREF_FUEL( 90,J), J=1,3) / 9.6_EB,  1.45_EB, 1.24E-03_EB /
DATA (CPLXREF_FUEL( 91,J), J=1,3) / 9.7_EB,  1.45_EB, 1.43E-03_EB /
DATA (CPLXREF_FUEL( 92,J), J=1,3) / 9.8_EB,  1.45_EB, 1.69E-03_EB /
DATA (CPLXREF_FUEL( 93,J), J=1,3) / 9.9_EB,  1.45_EB, 1.26E-03_EB /
DATA (CPLXREF_FUEL( 94,J), J=1,3) / 10.0_EB, 1.45_EB, 8.24E-04_EB /

END MODULE SPECDATA



MODULE MIEV

USE PRECISION_PARAMETERS
USE GLOBAL_CONSTANTS, ONLY: NUMBER_SPECTRAL_BANDS,NUMBER_RADIATION_ANGLES,LU_ERR
USE SPECDATA, ONLY: CPLXREF_WATER, NWATERK, CPLXREF_FUEL, NFUELK
USE RADCAL_CALC, ONLY: PLANCK
USE RADCONS
USE MEMORY_FUNCTIONS, ONLY : CHKMEMERR
USE MATH_FUNCTIONS, ONLY : INTERPOLATE1D

IMPLICIT NONE
REAL(EB), ALLOCATABLE :: RDMIE(:),  LMBDMIE(:),LMBDWGHT(:),REAL_REF_INDX(:),CMPLX_REF_INDX(:)
REAL(EB), ALLOCATABLE :: QSCA(:,:), QABS(:,:),  CHI_F(:,:)

PRIVATE
PUBLIC MEAN_CROSS_SECTIONS

CONTAINS


SUBROUTINE MEAN_CROSS_SECTIONS(CLASS_NUMBER)

! THIS SUBROUTINE CALCULATES THE MEAN SCATTERING AND ABSORPTION
! COEFFICIENTS FOR EACH RADIATION BAND AND PARTICLE SIZE GROUP.

USE TYPES, ONLY: LAGRANGIAN_PARTICLE_CLASS_TYPE, LAGRANGIAN_PARTICLE_CLASS, TABLES_TYPE, TABLES, SPECIES_MIXTURE, SPECIES, &
                 SPECIES_MIXTURE_TYPE
USE GLOBAL_CONSTANTS, ONLY: H2O_INDEX
USE MATH_FUNCTIONS, ONLY : INTERPOLATE1D
INTEGER  :: NSB,I,J,IBND,IZERO,NX,NLAMBDALOW(1),NLAMBDAHIGH(1),ND,CLASS_NUMBER,NL_DATA
REAL(EB) :: RMMAX,RMMIN,RDTMP,IB,IBSUM,AVAL,BVAL,ASUM,BSUM,B_WIEN
REAL(EB), ALLOCATABLE, DIMENSION(:,:) :: WQABS,WQSCA
REAL(EB), ALLOCATABLE, DIMENSION(:) :: R50
TYPE (LAGRANGIAN_PARTICLE_CLASS_TYPE), POINTER :: LPC=>NULL()
TYPE (SPECIES_MIXTURE_TYPE), POINTER :: SM2=>NULL()
TYPE (TABLES_TYPE),  POINTER :: TA=>NULL()

IF (CLASS_NUMBER > 0) THEN
   LPC => LAGRANGIAN_PARTICLE_CLASS(CLASS_NUMBER)
ELSE
   SM2 => SPECIES_MIXTURE(-CLASS_NUMBER)
ENDIF
NSB = NUMBER_SPECTRAL_BANDS

! PHYSICAL PARAMETERS
! MINIMUM MEAN RADIUS (M)
RMMIN = 0.5_EB*1.E-6*MIE_MINIMUM_DIAMETER
IF (RMMIN < TWO_EPSILON_EB) RMMIN = 0.5E-6_EB

! MAXIMUM MEAN RADIUS (M)
RMMAX = 0.5_EB*1.E-6*MIE_MAXIMUM_DIAMETER
IF (RMMAX < TWO_EPSILON_EB) THEN
   IF (CLASS_NUMBER > 0) RMMAX = 0.5_EB*LPC%DIAMETER
   IF (CLASS_NUMBER < 0) RMMAX = 0.5_EB*SM2%MEAN_DIAMETER
   ! ALLOW INCREASE OF THE MEAN RADIUS
   RMMAX = 1.5_EB*RMMAX
ENDIF

! OTHER CONSTANTS

B_WIEN = 2.8977685E-3_EB

! CALCULATE PARAMETERS OF THE PARTICLE GROUP LOOKUP TABLE

DGROUP_A = (LOG(RMMAX)-LOG(RMMIN))/(MIE_NDG-1)
DGROUP_B = LOG(RMMAX)-DGROUP_A*MIE_NDG

! GENERATE THE PARTICLE RADII FOR MIE TABLE (MICRONS)

RDTMP = 0.5_EB*RMMIN
NX = 0
DO WHILE (RDTMP < 2._EB*RMMAX)
   NX = NX + 1
   RDTMP = RDTMP + MIN(1.0E6_EB,0.2_EB*RDTMP**(1._EB))
ENDDO
NRDMIE = NX

ALLOCATE(RDMIE(1:NRDMIE),STAT=IZERO)
CALL CHKMEMERR('MIEV','RDMIE',IZERO)

RDTMP = 0.5_EB*RMMIN
RDMIE(1) = RDTMP
DO NX = 2, NRDMIE
   RDTMP = RDTMP + MIN(3.0E6_EB,0.2_EB*RDTMP**(1.0_EB))
   RDMIE(NX) = RDTMP
ENDDO

! RADIATIVE PROPERTIES

NLMBDMIE = NWATERK
IF (CLASS_NUMBER > 0) THEN
   IF (LPC%RADIATIVE_PROPERTY_INDEX>0) THEN
      TA => TABLES(LPC%RADIATIVE_PROPERTY_INDEX)
      NL_DATA = TA%NUMBER_ROWS
   ELSEIF (LPC%Y_INDEX==H2O_INDEX) THEN
      NL_DATA = NLMBDMIE
   ELSEIF (LPC%FUEL) THEN
      NL_DATA = NFUELK
   ELSE
      NL_DATA = 1
   ENDIF
ELSE
   IF (SM2%SINGLE_SPEC_INDEX > 0) THEN
      IF (SM2%SINGLE_SPEC_INDEX==H2O_INDEX) THEN
         NL_DATA = NLMBDMIE
      ELSEIF (SPECIES(SM2%SINGLE_SPEC_INDEX)%ISFUEL) THEN
         NL_DATA = NFUELK
      ELSE
         NL_DATA = 1
      ENDIF
   ELSE
      NL_DATA = 1
   ENDIF
ENDIF

!     ALLOCATE ARRAYS

ALLOCATE(QSCA(1:NRDMIE,1:NLMBDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','QSCA',IZERO)
ALLOCATE(QABS(1:NRDMIE,1:NLMBDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','QABS',IZERO)
ALLOCATE(CHI_F(1:NRDMIE,1:NLMBDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','CHI_F',IZERO)
!
ALLOCATE(LMBDMIE(1:NLMBDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','LMBDMIE',IZERO)
ALLOCATE(LMBDWGHT(1:NLMBDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','LMBDWGHT',IZERO)
ALLOCATE(REAL_REF_INDX(1:NLMBDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','REAL_REF_INDX',IZERO)
ALLOCATE(CMPLX_REF_INDX(1:NLMBDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','CMPLX_REF_INDX',IZERO)

! RADIATIVE PROPERTIES

LMBDMIE(1:NLMBDMIE) = CPLXREF_WATER(1:NLMBDMIE,1)

IF (CLASS_NUMBER > 0) THEN
   IF (LPC%RADIATIVE_PROPERTY_INDEX > 0) THEN
      DO NX = 1,NLMBDMIE
         CALL INTERPOLATE1D(TA%TABLE_DATA(:,1)*1.0E-6_EB,TA%TABLE_DATA(:,2),LMBDMIE(NX),REAL_REF_INDX(NX))
         CALL INTERPOLATE1D(TA%TABLE_DATA(:,1)*1.0E-6_EB,TA%TABLE_DATA(:,3),LMBDMIE(NX),CMPLX_REF_INDX(NX))
      ENDDO
   ELSEIF (LPC%Y_INDEX==H2O_INDEX) THEN
      REAL_REF_INDX              = 1.33_EB
      CMPLX_REF_INDX(1:NLMBDMIE) = CPLXREF_WATER(1:NLMBDMIE,2)
   ELSEIF (LPC%FUEL) THEN
      DO NX = 1,NLMBDMIE
         CALL INTERPOLATE1D(CPLXREF_FUEL(:,1)*1.0E-6_EB,CPLXREF_FUEL(:,2),LMBDMIE(NX),REAL_REF_INDX(NX))
         CALL INTERPOLATE1D(CPLXREF_FUEL(:,1)*1.0E-6_EB,CPLXREF_FUEL(:,3),LMBDMIE(NX),CMPLX_REF_INDX(NX))
      ENDDO
   ELSE
      REAL_REF_INDX              = LPC%REAL_REFRACTIVE_INDEX
      CMPLX_REF_INDX             = LPC%COMPLEX_REFRACTIVE_INDEX
   ENDIF
ELSE
   IF (SM2%SINGLE_SPEC_INDEX > 0) THEN
      IF (SM2%SINGLE_SPEC_INDEX==H2O_INDEX) THEN
         REAL_REF_INDX              = 1.33_EB
         CMPLX_REF_INDX(1:NLMBDMIE) = CPLXREF_WATER(1:NLMBDMIE,2)
      ELSEIF (SPECIES(SM2%SINGLE_SPEC_INDEX)%ISFUEL) THEN
         DO NX = 1,NLMBDMIE
            CALL INTERPOLATE1D(CPLXREF_FUEL(:,1)*1.0E-6_EB,CPLXREF_FUEL(:,2),LMBDMIE(NX),REAL_REF_INDX(NX))
            CALL INTERPOLATE1D(CPLXREF_FUEL(:,1)*1.0E-6_EB,CPLXREF_FUEL(:,3),LMBDMIE(NX),CMPLX_REF_INDX(NX))
         ENDDO
      ELSE
         REAL_REF_INDX              = LPC%REAL_REFRACTIVE_INDEX
         CMPLX_REF_INDX             = LPC%COMPLEX_REFRACTIVE_INDEX
      ENDIF
   ELSE
      REAL_REF_INDX              = LPC%REAL_REFRACTIVE_INDEX
      CMPLX_REF_INDX             = LPC%COMPLEX_REFRACTIVE_INDEX
   ENDIF
ENDIF

CALL MIE_SCATTERING

!     GENERATE INTEGRATION WEIGHTS FOR LAMBDA

IF (NLMBDMIE == 1) THEN
   LMBDWGHT(1) = 1._EB
ELSE
   LMBDWGHT(1) = 0.5_EB*(LMBDMIE(2)-LMBDMIE(1))
   DO I = 2, NLMBDMIE-1
      LMBDWGHT(I) = 0.5_EB*(LMBDMIE(I+1) - LMBDMIE(I-1))
   ENDDO
   LMBDWGHT(NLMBDMIE) = 0.5_EB*(LMBDMIE(NLMBDMIE)-LMBDMIE(NLMBDMIE-1))
ENDIF

ALLOCATE(WQABS(0:MIE_NDG,1:NUMBER_SPECTRAL_BANDS))
CALL ChkMemErr('IRAD','WQABS',IZERO)
WQABS = 0._EB
ALLOCATE(WQSCA(0:MIE_NDG,1:NUMBER_SPECTRAL_BANDS))
CALL ChkMemErr('IRAD','WQSCA',IZERO)
WQSCA = 0._EB
ALLOCATE(R50(0:MIE_NDG))
CALL ChkMemErr('IRAD','R50',IZERO)
R50 = 0._EB

!     LOOP OVER ALL RADIATION BANDS

BANDLOOP: DO IBND = 1,NSB
   !
   IF (NSB == 1) THEN
      NLAMBDALOW = 1
      NLAMBDAHIGH = NLMBDMIE
   ELSE
      NLAMBDALOW  = MINLOC(LMBDMIE,MASK=LMBDMIE>=WL_LOW(IBND)*1.E-6)
      NLAMBDAHIGH = MAXLOC(LMBDMIE,MASK=LMBDMIE<=WL_HIGH(IBND)*1.0E-6)
   ENDIF

   !     LOOP OVER ALL PARTICLE SIZE GROUPS


   DRGROUPLOOP: DO ND = 1, MIE_NDG

      R50(ND) = EXP(DGROUP_A*REAL(ND,EB) + DGROUP_B)

      !     LOOP OVER WAVELENGTHS

      IBSUM = 0._EB

      DO J = NLAMBDALOW(1),NLAMBDAHIGH(1)
         IB = PLANCK(RADTMP, LMBDMIE(J)*1.0E6_EB)
         IBSUM = IBSUM + LMBDWGHT(J)*IB

         ASUM = 0._EB
         BSUM = 0._EB

         ! PROPERTIES AT D32

         CALL INTERPOLATE1D(RDMIE,QSCA(:,J),R50(ND),AVAL)
         CALL INTERPOLATE1D(RDMIE,CHI_F(:,J),R50(ND),BVAL)
         BVAL = (1._EB-BVAL)
         ASUM = AVAL*BVAL
         CALL INTERPOLATE1D(RDMIE,QABS(:,J),R50(ND),BVAL)
         BSUM = BVAL

         WQSCA(ND,IBND) = WQSCA(ND,IBND) + ASUM*LMBDWGHT(J)*IB
         WQABS(ND,IBND) = WQABS(ND,IBND) + BSUM*LMBDWGHT(J)*IB
      ENDDO

      !     NORMALIZE WITH BLACKBODY RADIATION

      WQSCA(ND,IBND)  = WQSCA(ND,IBND)/IBSUM
      WQABS(ND,IBND)  = WQABS(ND,IBND)/IBSUM

ENDDO DRGROUPLOOP
ENDDO BANDLOOP

IF (CLASS_NUMBER > 0) THEN
   LPC%R50 = R50
   LPC%WQSCA = WQSCA
   LPC%WQABS = WQABS
ELSE
   SM2%R50 = R50
   SM2%WQSCA = WQSCA
   SM2%WQABS = WQABS
ENDIF

DEALLOCATE(WQABS)
DEALLOCATE(WQSCA)
DEALLOCATE(R50)

DEALLOCATE(RDMIE)
DEALLOCATE(QSCA)
DEALLOCATE(QABS)
DEALLOCATE(CHI_F)
DEALLOCATE(LMBDMIE)
DEALLOCATE(LMBDWGHT)
DEALLOCATE(REAL_REF_INDX)
DEALLOCATE(CMPLX_REF_INDX)

END SUBROUTINE MEAN_CROSS_SECTIONS



SUBROUTINE MIE_SCATTERING
!
!     CALCULATES THE SCATTERING AND ABSORPTION CROSS SECTIONS
!     AND CALCULATES FORWARD SCATTERING FRACTION BY INTEGRATING
!     THE SCATTERING PHASE FUNCTION.
!
! ----------------------------------------------------------------------
! -----------  SPECIFICATIONS FOR SUBROUTINE  MIEV0  ---------------
! ----------------------------------------------------------------------
INTEGER   MOMDIM
PARAMETER  ( MOMDIM = 200)
LOGICAL   ANYANG, PERFCT, PRNT( 2 )
INTEGER   IPOLZN, NMOM
REAL(EB)  GQSC, MIMCUT, PMOM( 0:MOMDIM, 4 ), SPIKE, QE, QS
REAL(EB), ALLOCATABLE :: XXX(:)
COMPLEX(EB) SFORW, SBACK, TFORW( 2 ), TBACK( 2 ), CREFIN
COMPLEX(EB), ALLOCATABLE :: S1(:), S2(:)

! --------------- LOCAL VARIABLES --------------------------------------

INTEGER   I, J, K, NX, IZERO
INTEGER   NLAMBDA, NRA, NMIEANG2
REAL(EB)  STMP, AIJ, FTMP, XX_MAX
REAL(EB)  MUMIN1, MUMIN2,THETALIM1, THETALIM2, MUDLOC
REAL(EB)  MU1, MU2, NU1, NU2, MUD0LOC, MUDPILOC, MUD1, MUD2, DMUD
REAL(EB),ALLOCATABLE :: XMU1(:), XNU1(:),XMU2(:), &
          ANGLE1(:), ANGLE2(:), MUD(:), MUDX(:),PWGHT(:), &
          PHSFUN(:), PFOR(:,:), MUD0(:,:), MUDPI(:,:)
! ----------------------------------------------------------------------

NRA    = NUMBER_RADIATION_ANGLES

!     MIEV-CODE VARIABLES

MIMCUT = 1.E-6_EB
PERFCT = .FALSE.
ANYANG = .TRUE.
!      IPOLZN = +1234
IPOLZN = 0
NMOM   = 0
PRNT   = .FALSE.

!     LIMIT FOR XX

XX_MAX = 15000.0_EB

!     INTEGRATION LIMITS

THETALIM1 = ACOS(1._EB - 2._EB/REAL(NRA))
MUMIN1 = COS(THETALIM1)
MUMIN2 = MUMIN1**2-(1._EB-MUMIN1**2)
THETALIM2 = ACOS(MUMIN2)
NMIEANG2 = NMIEANG*2

!     ALLOCATE LOCAL ARRAYS

ALLOCATE(XXX(1:NRDMIE),STAT=IZERO)
CALL CHKMEMERR('INIT','XXX',IZERO)
ALLOCATE(S1(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','S1',IZERO)
ALLOCATE(S2(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','S2',IZERO)
ALLOCATE(XMU1(1:NMIEANG),STAT=IZERO)
CALL CHKMEMERR('INIT','XMU1',IZERO)
ALLOCATE(XNU1(1:NMIEANG),STAT=IZERO)
CALL CHKMEMERR('INIT','XNU1',IZERO)
ALLOCATE(XMU2(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','XMU2',IZERO)
ALLOCATE(MUD(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','MUD',IZERO)
ALLOCATE(MUDX(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','MUDX',IZERO)
ALLOCATE(PWGHT(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','PWGHT',IZERO)
ALLOCATE(ANGLE1(1:NMIEANG),STAT=IZERO)
CALL CHKMEMERR('INIT','ANGLE1',IZERO)
ALLOCATE(ANGLE2(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','ANGLE2',IZERO)
ALLOCATE(PHSFUN(1:NMIEANG2),STAT=IZERO)
CALL CHKMEMERR('INIT','PHSFUN',IZERO)
ALLOCATE(PFOR(1:NMIEANG,1:NMIEANG),STAT=IZERO)
CALL CHKMEMERR('INIT','PFOR',IZERO)
ALLOCATE(MUD0(1:NMIEANG,1:NMIEANG),STAT=IZERO)
CALL CHKMEMERR('INIT','MUD0',IZERO)
ALLOCATE(MUDPI(1:NMIEANG,1:NMIEANG),STAT=IZERO)
CALL CHKMEMERR('INIT','MUDPI',IZERO)

!     CREATE SOLID ANGLE INTEGRATION ARRAYS

DO I = 1,NMIEANG
   ANGLE1(I) = THETALIM1*REAL(I-1)/REAL(NMIEANG-1)
   ANGLE1(I) = ANGLE1(I)*(1._EB-0.99*REAL(NMIEANG-I)/REAL(NMIEANG))
ENDDO
DO I = 1,NMIEANG
   XMU1(NMIEANG-I+1) = COS(ANGLE1(I))
   XNU1(NMIEANG-I+1) = SIN(ANGLE1(I))
ENDDO

!    CREATE PHASE FUNCTION INGTEGRATION ARRAYS

DO I = 1,NMIEANG2
   ANGLE2(I) = THETALIM2*REAL(I-1)/REAL(NMIEANG2-1)
   ANGLE2(I) = ANGLE2(I)*(1._EB-0.99*REAL(NMIEANG2-I)/REAL(NMIEANG2))
ENDDO
DO I = 1,NMIEANG2
   XMU2(NMIEANG2-I+1) = COS(ANGLE2(I))
ENDDO

!    CALCULATE PHASE FUNCTION INGETRATION LIMITS

DO J = 1,NMIEANG
   DO I = 1,NMIEANG
      MUD0(I,J)  = XMU1(I)*XMU1(J) + XNU1(I)*XNU1(J)
      MUDPI(I,J) = XMU1(I)*XMU1(J) - XNU1(I)*XNU1(J)
   ENDDO
ENDDO

!     CALCULATE PHASE FUNCTION INTEGRATION WEIGHTS

MU1 = 0.7_EB
MU2 = 0.9_EB
NU1 = SQRT(1-MU1**2)
NU2 = SQRT(1-MU2**2)
MUD0LOC  = MU1*MU2 + NU1*NU2
MUDPILOC = MU1*MU2 - NU1*NU2
MUD1 = MUDPILOC
MUD2 = MUD0LOC
DMUD = (MUD2-MUD1)/(NMIEANG2-1)
DO I = 1,NMIEANG2
   MUD(I) = MUD1+REAL(I-1)*DMUD
ENDDO
MUD(1)       = MUD(1)       + 0.25_EB*DMUD !EMPIRICAL
MUD(NMIEANG2) = MUD(NMIEANG2) - 0.25_EB*DMUD !EMPIRICAL
DO I = 1,NMIEANG2
   MUDX(I) = (MUD(I)-MUD1)/(MUD2-MUD1)
ENDDO
DO I = 1,NMIEANG2
   PWGHT(I) = DMUD/SQRT((NU1*NU2)**2-(MUD(I)-MU1*MU2)**2)
ENDDO
PWGHT(2)         = 0.5_EB*PWGHT(2)
PWGHT(NMIEANG2-1) = 0.5_EB*PWGHT(NMIEANG2-1)

!     LOOP OVER WAVELENGTH

LAMBDALOOP: DO NLAMBDA = 1, NLMBDMIE
!
   CREFIN = CMPLX( REAL_REF_INDX(NLAMBDA), CMPLX_REF_INDX(NLAMBDA), EB )

! CHOOSE PERFECTLY REFLECTING SPHERE, IF LARGE REAL INDEX IS GIVEN.

   IF (REAL_REF_INDX(NLAMBDA) > 10._EB) PERFCT = .TRUE.

!     LOOP OVER PARTICLE RADIUS

   RADIUSLOOP: DO NX = 1, NRDMIE

      XXX(NX) = MIN(XX_MAX,2._EB*PI*RDMIE(NX)/LMBDMIE(NLAMBDA))
      CALL MIEV0( XXX(NX), CREFIN, PERFCT, MIMCUT, ANYANG,   &
                  NMIEANG2, XMU2, NMOM, IPOLZN, MOMDIM, PRNT,  &
                  QE, QS, GQSC, &
                  PMOM, SFORW, SBACK, S1, S2, TFORW, TBACK, SPIKE )

      QSCA(NX,NLAMBDA) = QS
      QABS(NX,NLAMBDA) = QE-QS

!     CALCULATE SINGLE DROP PHASE FUNCTION

      IF (ABS(QS)>TWO_EPSILON_EB) THEN
         DO I = 1,NMIEANG2
            PHSFUN(I) = 2._EB*(ABS(S1(I))**2 + ABS(S2(I))**2 )
         ENDDO
         PHSFUN = PHSFUN/(QS*XXX(NX)**2)
      ELSE
         PHSFUN = 1.0_EB
      ENDIF

!     CALCULATE THE INNERMOST INTEGRAL OF THE FORWARD SCATTERING FRACTION

      PFOR = 0._EB
      DO J = 1,NMIEANG
         DO I = J,NMIEANG
            IF (ABS(MUD0(I,J)-MUDPI(I,J))<=TWO_EPSILON_EB) THEN
               CALL INTERPOLATE1D(XMU2,PHSFUN,MUD0(I,J),FTMP)
               PFOR(I,J) = PI*FTMP
            ELSE
               MUD1 = MUDPI(I,J)
               MUD2 = MUD0(I,J)
               STMP = 0.0_EB
               DO K = 1,NMIEANG2
                  MUDLOC = MUD1+MUDX(K)*(MUD2-MUD1)
                  CALL INTERPOLATE1D(XMU2,PHSFUN,MUDLOC,FTMP)
                  STMP = STMP + PWGHT(K)*FTMP
               ENDDO
               PFOR(I,J) = STMP
            ENDIF
         ENDDO
      ENDDO

!     CALCULATE THE TWO OUTER INTEGRALS OF THE FORWARD FRACTION

      STMP = 0._EB
      DO J = 1,NMIEANG-1
         DO I = J+1,NMIEANG-1
            AIJ = (XMU1(I+1)-XMU1(I))*(XMU1(J+1)-XMU1(J))/2._EB
            STMP = STMP +   (2._EB*PFOR(I,J)+PFOR(I+1,J)+PFOR(I,J+1)+2._EB*PFOR(I+1,J+1))*AIJ/3._EB
         ENDDO
      ENDDO
      DO I = 1,NMIEANG-1
         AIJ = ((XMU1(I+1)-XMU1(I))**2)/2._EB
         STMP = STMP + (PFOR(I,I)+PFOR(I+1,I)+PFOR(I+1,I+1))*AIJ/3._EB
      ENDDO
      CHI_F(NX,NLAMBDA) = 2._EB*STMP/(4._EB*PI/NRA)
   ENDDO RADIUSLOOP
ENDDO LAMBDALOOP

DEALLOCATE(XXX)
DEALLOCATE(S1)
DEALLOCATE(S2)
DEALLOCATE(XMU1)
DEALLOCATE(XNU1)
DEALLOCATE(XMU2)
DEALLOCATE(MUD)
DEALLOCATE(MUDX)
DEALLOCATE(PWGHT)
DEALLOCATE(ANGLE1)
DEALLOCATE(ANGLE2)
DEALLOCATE(PHSFUN)
DEALLOCATE(PFOR)
DEALLOCATE(MUD0)
DEALLOCATE(MUDPI)

END SUBROUTINE MIE_SCATTERING

SUBROUTINE MIEV0( XX, CREFIN, PERFCT, MIMCUT, ANYANG, NUMANG, XMU, &
                  NMOM, IPOLZN, MOMDIM, PRNT, QEXT, QSCA, GQSC, &
                  PMOM, SFORW, SBACK, S1, S2, TFORW, TBACK, &
                  SPIKE )

!     MIE SCATTERING FOR A SINGLE PARTICLE AND WAVELENGTH.
!     AUTHOR:  DR. WARREN J. WISCOMBE (WISCOMBE@CLIMATE.GSFC.NASA.GOV)
!         NASA GODDARD SPACE FLIGHT CENTER
!         CODE 913
!         GREENBELT, MD 20771

!     REFERENCES
!     ----------
!
!     (1) WISCOMBE, W., 1979: MIE SCATTERING CALCULATIONS--ADVANCES
!         IN TECHNIQUE AND FAST, VECTOR-SPEED COMPUTER CODES,
!         NCAR TECH NOTE TN-140+STR, NATIONAL CENTER FOR
!         ATMOSPHERIC RESEARCH, BOULDER, COLORADO (OUT OF PRINT
!         BUT AN UPDATED ELECTRONIC VERSION AVAILABLE)
!
!     (2) WISCOMBE, W., 1980: IMPROVED MIE SCATTERING ALGORITHMS,
!         APPL. OPT. 19, 1505-1509
!

!    COMPUTES MIE SCATTERING AND EXTINCTION EFFICIENCIES; ASYMMETRY
!    FACTOR;  FORWARD- AND BACKSCATTER AMPLITUDE;  SCATTERING
!    AMPLITUDES VS. SCATTERING ANGLE FOR INCIDENT POLARIZATION PARALLEL
!    AND PERPENDICULAR TO THE PLANE OF SCATTERING;
!    COEFFICIENTS IN THE LEGENDRE POLYNOMIAL EXPANSIONS OF EITHER THE
!    UNPOLARIZED PHASE FUNCTION OR THE POLARIZED PHASE MATRIX;
!    SOME QUANTITIES NEEDED IN POLARIZED RADIATIVE TRANSFER;  AND
!    INFORMATION ABOUT WHETHER OR NOT A RESONANCE HAS BEEN HIT.
!
!    INPUT AND OUTPUT VARIABLES ARE DESCRIBED IN FILE MIEV.DOC.
!    MANY STATEMENTS ARE ACCOMPANIED BY COMMENTS REFERRING TO
!    REFERENCES IN MIEV.DOC, NOTABLY THE NCAR MIE REPORT WHICH IS NOW
!    AVAILABLE ELECTRONICALLY AND WHICH IS REFERRED TO USING THE
!    SHORTHAND (RN), MEANING EQ. (N) OF THE REPORT.

!    CALLING TREE:
!
!        MIEV0
!            TESTMI
!                TSTBAD
!                MIPRNT
!                ERRMSG
!            CKINMI
!                WRTBAD
!                WRTDIM
!                ERRMSG
!            SMALL1
!            SMALL2
!            ERRMSG
!            BIGA
!                CONFRA
!                    ERRMSG
!            LPCOEF
!                LPCO1T
!                LPCO2T
!                ERRMSG
!            MIPRNT
!
!   I N P U T   V A R I A B L E S
!   -----------------------------
!
!  ( EVEN IF AN INPUT VARIABLE IS NOT NEEDED FOR A PARTICULAR
!    APPLICATION, MAKE SURE IT HAS A LEGITIMATE VALUE THAT CAN
!    BE WRITTEN OUT AND READ IN -- NO INDEFINITES, ETC. )
!
!  XX        MIE SIZE PARAMETER ( 2 * PI * RADIUS / WAVELENGTH )
!
!  CREFIN    COMPLEX REFRACTIVE INDEX ( IMAG PART CAN BE + OR -,
!            BUT INTERNALLY A NEGATIVE IMAGINARY INDEX IS ASSUMED ).
!            IF IMAG PART IS - ,  SCATTERING AMPLITUDES AS IN VAN
!            DE HULST ARE RETURNED;  IF IMAG PART IS + , COMPLEX
!            CONJUGATES OF THOSE SCATTERING AMPLITUDES ARE RETURNED
!            (THE LATTER IS THE CONVENTION IN PHYSICS).
!            ** NOTE ** IN THE 'PERFECT' CASE, SCATTERING AMPLITUDES
!            IN THE VAN DE HULST (REF. 6 ABOVE) CONVENTION WILL
!            AUTOMATICALLY BE RETURNED UNLESS  IM(CREFIN)  IS
!            POSITIVE;  OTHERWISE, CREFIN PLAYS NO ROLE.
!
!  PERFCT    TRUE, ASSUME REFRACTIVE INDEX IS INFINITE AND USE
!            SPECIAL CASE FORMULAS FOR MIE COEFFICIENTS  'A'
!            AND  'B'  ( SEE KERKER, M., THE SCATTERING OF
!            LIGHT AND OTHER ELECTROMAGNETIC RADIATION, P. 90 ).
!            THIS IS SOMETIMES CALLED THE 'TOTALLY REFLECTING',
!            SOMETIMES THE 'PERFECTLY CONDUCTING' CASE.
!            ( SEE CREFIN FOR ADDITIONAL INFORMATION )
!
!  MIMCUT    (POSITIVE) VALUE BELOW WHICH IMAGINARY REFRACTIVE
!            INDEX IS REGARDED AS ZERO (COMPUTATION PROCEEDS
!            FASTER FOR ZERO IMAGINARY INDEX)
!
!  ANYANG    TRUE, ANY ANGLES WHATSOEVER MAY BE INPUT THROUGH
!            XMU.  FALSE, THE ANGLES ARE MONOTONE INCREASING
!            AND MIRROR SYMMETRIC ABOUT 90 DEGREES (THIS OPTION
!            IS ADVANTAGEOUS BECAUSE THE SCATTERING AMPLITUDES
!            S1,S2 FOR THE ANGLES BETWEEN 90 AND 180 DEGREES
!            ARE EVALUABLE FROM SYMMETRY RELATIONS, AND HENCE
!            ARE OBTAINED WITH LITTLE ADDED COMPUTATIONAL COST.)
!
!  NUMANG    NO. OF ANGLES AT WHICH SCATTERING AMPLITUDES
!            S1,S2 ARE TO BE EVALUATED  ( SET = 0 TO SKIP
!            CALCULATION OF S1,S2 ).  MAKE SURE NUMANG DOES
!            NOT EXCEED THE PARAMETER MAXANG IN THE PROGRAM.
!
!  XMU(N)    COSINES OF ANGLES ( N = 1 TO NUMANG ) AT WHICH S1,S2
!            ARE TO BE EVALUATED.  IF ANYANG = FALSE, THEN
!
!             (A) THE ANGLES MUST BE MONOTONE INCREASING AND
!                 MIRROR SYMMETRIC ABOUT 90 DEGREES (IF 90-A IS
!                 AN ANGLE, THEN 90+A MUST BE ALSO)
!
!             (B) IF NUMANG IS ODD, 90 DEGREES MUST BE AMONG
!                 THE ANGLES
!
!  NMOM       HIGHEST LEGENDRE MOMENT PMOM TO CALCULATE,
!             NUMBERING FROM ZERO ( NMOM = 0 PREVENTS
!             CALCULATION OF PMOM )
!
!  IPOLZN     POSITIVE, COMPUTE LEGENDRE MOMENTS PMOM FOR THE
!                       MUELLER MATRIX ELEMENTS DETERMINED BY THE
!                       DIGITS OF IPOLZN, WITH 1 REFERRING TO M1,
!                       2 TO M2, 3 TO S21, AND 4 TO D21 (REF. 3).
!                       E.G., IF IPOLZN = 14 THEN ONLY MOMENTS FOR
!                       M1 AND D21 WILL BE RETURNED.
!
!             0,        COMPUTE LEGENDRE MOMENTS PMOM FOR THE
!                       UNPOLARIZED UNNORMALIZED PHASE FUNCTION.
!
!             NEGATIVE, COMPUTE LEGENDRE MOMENTS PMOM FOR THE
!                       SEKERA PHASE QUANTITIES DETERMINED BY THE
!                       DIGITS OF ABS(IPOLZN), WITH 1 REFERRING TO
!                       R1, 2 TO R2, 3 TO R3, AND 4 TO R4 (REF. 4).
!                       E.G., IF IPOLZN = -14 THEN ONLY MOMENTS FOR
!                       R1 AND R4 WILL BE RETURNED.
!
!             ( NOT USED IF  NMOM = 0 )
!
!  MOMDIM     DETERMINES FIRST DIMENSION OF PMOM, WHICH IS DIMENSIONED
!             INTERNALLY AS PMOM( 0:MOMDIM, * ) (SECOND DIMENSION MUST
!             BE THE LARGER OF UNITY AND THE HIGHEST DIGIT IN
!             IPOLZN; IF NOT, SERIOUS ERRORS WILL OCCUR).
!             MUST BE GIVEN A VALUE, EVEN IF  NMOM = 0.  MINIMUM: 1.
!
!  PRT(L)     PRINT FLAGS (LOGICAL).  L = 1  PRINTS  S1,S2, THEIR
!             SQUARED ABSOLUTE VALUES, AND DEGREE OF POLARIZATION,
!             PROVIDED NUMANG IS NON-ZERO.   L = 2  PRINTS ALL
!             OUTPUT VARIABLES OTHER THAN  S1,S2.
!
!
! O U T P U T   V A R I A B L E S
! -------------------------------
!
!  QEXT      (REAL) EXTINCTION EFFICIENCY FACTOR  ( REF. 2, EQ. 1A )
!
!  QSCA      (REAL) SCATTERING EFFICIENCY FACTOR  ( REF. 2, EQ. 1B )
!
!  GQSC      (REAL) ASYMMETRY FACTOR TIMES SCATTERING EFFICIENCY
!            ( REF. 2, EQ. 1C )  ( ALLOWS CALCULATION OF RADIATION
!            PRESSURE EFFICIENCY FACTOR  QPR = QEXT - GQSC )
!
!  =====================================================================
!  ==== NOTE --  S1, S2, SFORW, SBACK, TFORW, AND TBACK ARE CALCULATED
!  ====          INTERNALLY FOR NEGATIVE IMAGINARY REFRACTIVE INDEX;
!  ====          FOR POSITIVE IMAGINARY INDEX, THEIR COMPLEX CONJUGATES
!  ====          ARE TAKEN BEFORE THEY ARE RETURNED, TO CORRESPOND TO
!  ====          CUSTOMARY USAGE IN SOME PARTS OF PHYSICS ( IN PARTI-
!  ====          CULAR, IN PAPERS ON CAM APPROXIMATIONS TO MIE THEORY ).
!  =====================================================================
!
!  S1(N),    (COMPLEX) MIE SCATTERING AMPLITUDES AT ANGLES SPECIFIED
!  S2(N)     BY XMU(N) ( N=1 TO NUMANG )  ( REF. 2, EQS. 1D-E ).
!
!  SFORW     (COMPLEX) FORWARD-SCATTERING AMPLITUDE S1 AT
!            0 DEGREES.  ( S2(0 DEG) = S1(0 DEG) )
!
!  SBACK     (COMPLEX) BACKSCATTERING AMPLITUDE S1 AT
!            180 DEGREES.   ( S2(180 DEG) = - S1(180 DEG) )
!
!  TFORW(I)  (COMPLEX) VALUES OF
!
!                I=1:  T1 = ( S2 - (MU)*S1 ) / ( 1 - MU**2 )
!                I=2:  T2 = ( S1 - (MU)*S2 ) / ( 1 - MU**2 )
!
!            AT ANGLE THETA = 0 ( MU = COS(THETA) = 1 ), WHERE THE
!            EXPRESSIONS ON THE RIGHT-HAND SIDE ARE INDETERMINATE.
!            ( THESE QUANTITIES ARE REQUIRED FOR DOING POLARIZED
!            RADIATIVE TRANSFER (REF. 4, APPENDIX). )
!  TBACK(I)  (COMPLEX) VALUES OF  T1 (FOR I=1) OR  T2 (FOR I=2) AT
!            ANGLE  THETA = 180 DEGREES ( MU = COS(THETA) = - 1 ).
!
!  SPIKE     (REAL) MAGNITUDE OF THE SMALLEST DENOMINATOR OF
!            EITHER MIE COEFFICIENT (A-SUB-N OR B-SUB-N),
!            TAKEN OVER ALL TERMS IN THE MIE SERIES PAST
!            N = SIZE PARAMETER XX.  VALUES OF SPIKE BELOW
!            ABOUT 0.3 SIGNIFY A RIPPLE SPIKE, SINCE THESE
!            SPIKES ARE PRODUCED BY ABNORMALLY SMALL DENOMINATORS
!            IN THE MIE COEFFICIENTS (NORMAL DENOMINATORS ARE OF
!            ORDER UNITY OR HIGHER).  DEFAULTS TO 1.0 WHEN NOT
!            ON A SPIKE.  DOES NOT IDENTIFY ALL RESONANCES
!            (WE ARE STILL WORKING ON THAT).
!
! PMOM(M,NP) (REAL) MOMENTS  M = 0 TO NMOM  OF UNNORMALIZED NP-TH
!            PHASE QUANTITY PQ  ( MOMENTS WITH  M > 2*NTRM  ARE
!            ZERO, WHERE  NTRM = NO. TERMS IN MIE SERIES =
!            XX + 4*XX**1/3 + 1 ) :
!
!              PQ( MU, NP ) = SUM( M=0 TO INFINITY ) ( (2M+1)
!                                * PMOM( M,NP ) * P-SUB-M( MU ) )
!
!            WHERE  MU = COS( SCATTERING ANGLE )
!                   P-SUB-M = M-TH LEGENDRE POLYNOMIAL
!
!            AND THE DEFINITION OF 'PQ' IS AS FOLLOWS:
!
!            IPOLZN>0:  PQ(MU,1) = CABS( S1(MU) )**2
!                          PQ(MU,2) = CABS( S2(MU) )**2
!                          PQ(MU,3) = RE( S1(MU)*CONJG( S2(MU) ) )
!                          PQ(MU,4) = - IM( S1(MU)*CONJG( S2(MU) ) )
!                          ( CALLED M1, M2, S21, D21 IN LITERATURE )
!
!            IPOLZN=0:  PQ(MU,1) = ( CABS(S1)**2 + CABS(S2)**2 ) / 2
!                       ( THE UNNORMALIZED PHASE FUNCTION )
!
!            IPOLZN<0:  PQ(MU,1) = CABS( T1(MU) )**2
!                          PQ(MU,2) = CABS( T2(MU) )**2
!                          PQ(MU,3) = RE( T1(MU)*CONJG( T2(MU) ) )
!                          PQ(MU,4) = - IM( T1(MU)*CONJG( T2(MU) ) )
!                          ( CALLED R1, R2, R3, R4 IN LITERATURE )
!
!            THE SIGN OF THE 4TH PHASE QUANTITY IS A SOURCE OF
!            CONFUSION.  IT FLIPS IF THE COMPLEX CONJUGATES OF
!            S1,S2  OR  T1,T2  ARE USED, AS OCCURS WHEN A
!            REFRACTIVE INDEX WITH POSITIVE IMAGINARY PART IS
!            USED (SEE DISCUSSION BELOW).  THE DEFINITION ABOVE
!            IS CONSISTENT WITH A NEGATIVE IMAGINARY PART.
!
!            ** WARNING **  MAKE SURE THE SECOND DIMENSION OF PMOM
!            IN THE CALLING PROGRAM IS AT LEAST AS LARGE AS THE
!            ABSOLUTE VALUE OF IPOLZN.
!
!            FOR SMALL ENOUGH VALUES OF XX, OR LARGE ENOUGH VALUES
!            OF M,  PMOM  WILL TEND TO UNDERFLOW.  THUS, IT IS
!            UNWISE TO ASSUME THE VALUES RETURNED ARE NON-ZERO AND,
!            FOR EXAMPLE, TO DIVIDE SOME QUANTITY BY THEM.
!


!      I N T E R N A L   V A R I A B L E S
!      -----------------------------------

!  AN,BN           MIE COEFFICIENTS A-SUB-N, B-SUB-N ( REF. 1, EQ. 16 )
!  ANM1,BNM1       MIE COEFFICIENTS  A-SUB-(N-1),
!                     B-SUB-(N-1);  USED IN GQSC SUM
!  ANP             COEFFS. IN S+ EXPANSION ( REF. 2, P. 1507 )
!  BNP             COEFFS. IN S- EXPANSION ( REF. 2, P. 1507 )
!  ANPM            COEFFS. IN S+ EXPANSION ( REF. 2, P. 1507 )
!                     WHEN  MU  IS REPLACED BY  - MU
!  BNPM            COEFFS. IN S- EXPANSION ( REF. 2, P. 1507 )
!                     WHEN  MU  IS REPLACED BY  - MU
!  CALCMO(K)       TRUE, CALCULATE MOMENTS FOR K-TH PHASE QUANTITY
!                     (DERIVED FROM IPOLZN)
!  CBIGA(N)        BESSEL FUNCTION RATIO A-SUB-N (REF. 2, EQ. 2)
!                     ( COMPLEX VERSION )
!  CDENAN,         (COMPLEX) DENOMINATORS OF AN,BN
!   CDENBN
!  CIOR            COMPLEX INDEX OF REFRACTION WITH NEGATIVE
!                     IMAGINARY PART (VAN DE HULST CONVENTION)
!  CIORIV          1 / CIOR
!  COEFF           ( 2N + 1 ) / ( N ( N + 1 ) )
!  CSUM1,2         TEMPORARY SUM VARIABLES FOR TFORW, TBACK
!  FN              FLOATING POINT VERSION OF LOOP INDEX FOR
!                     MIE SERIES SUMMATION
!  LITA,LITB(N)    MIE COEFFICIENTS AN, BN, SAVED IN ARRAYS FOR
!                     USE IN CALCULATING LEGENDRE MOMENTS PMOM
!  MAXTRM          MAX. POSSIBLE NO. OF TERMS IN MIE SERIES
!  MM              (-1)^(N+1), WHERE N IS MIE SERIES SUM INDEX
!  MIM             MAGNITUDE OF IMAGINARY REFRACTIVE INDEX
!  MRE             REAL PART OF REFRACTIVE INDEX
!  MAXANG          MAX. POSSIBLE VALUE OF INPUT VARIABLE NUMANG
!  NANGD2          (NUMANG+1)/2 ( NO. OF ANGLES IN 0-90 DEG; ANYANG=F )
!  NOABS           TRUE, SPHERE NON-ABSORBING (DETERMINED BY MIMCUT)
!  NP1DN           ( N + 1 ) / N
!  NPQUAN          HIGHEST-NUMBERED PHASE QUANTITY FOR WHICH MOMENTS ARE
!                     TO BE CALCULATED (THE LARGEST DIGIT IN IPOLZN
!                     IF  IPOLZN /= 0)
!  NTRM            NO. OF TERMS IN MIE SERIES
!  PASS1           TRUE ON FIRST ENTRY, FALSE THEREAFTER; FOR SELF-TEST
!  PIN(J)          ANGULAR FUNCTION PI-SUB-N ( REF. 2, EQ. 3 )
!                     AT J-TH ANGLE
!  PINM1(J)        PI-SUB-(N-1) ( SEE PIN ) AT J-TH ANGLE
!  PSINM1          RICATTI-BESSEL FUNCTION PSI-SUB-(N-1), ARGUMENT XX
!  PSIN            RICATTI-BESSEL FUNCTION PSI-SUB-N OF ARGUMENT XX
!                     ( REF. 1, P. 11 FF. )
!  RBIGA(N)        BESSEL FUNCTION RATIO A-SUB-N (REF. 2, EQ. 2)
!                     ( REAL VERSION, FOR WHEN IMAG REFRAC INDEX = 0 )
!  RIORIV          1 / MRE
!  RN              1 / N
!  RTMP            (REAL) TEMPORARY VARIABLE
!  SP(J)           S+  FOR J-TH ANGLE  ( REF. 2, P. 1507 )
!  SM(J)           S-  FOR J-TH ANGLE  ( REF. 2, P. 1507 )
!  SPS(J)          S+  FOR (NUMANG+1-J)-TH ANGLE ( ANYANG=FALSE )
!  SMS(J)          S-  FOR (NUMANG+1-J)-TH ANGLE ( ANYANG=FALSE )
!  TAUN            ANGULAR FUNCTION TAU-SUB-N ( REF. 2, EQ. 4 )
!                     AT J-TH ANGLE
!  TCOEF           N ( N+1 ) ( 2N+1 ) (FOR SUMMING TFORW,TBACK SERIES)
!  TWONP1          2N + 1
!  YESANG          TRUE IF SCATTERING AMPLITUDES ARE TO BE CALCULATED
!  ZETNM1          RICATTI-BESSEL FUNCTION  ZETA-SUB-(N-1) OF ARGUMENT
!                     XX  ( REF. 2, EQ. 17 )
!  ZETN            RICATTI-BESSEL FUNCTION  ZETA-SUB-N OF ARGUMENT XX
! ----------------------------------------------------------------------
!
IMPLICIT  NONE
!
! ----------------------------------------------------------------------
! --------  I / O SPECIFICATIONS FOR SUBROUTINE MIEV0  -----------------
! ----------------------------------------------------------------------
LOGICAL     ANYANG, PERFCT, PRNT(*)
INTEGER     IPOLZN, MOMDIM, NUMANG, NMOM
REAL(EB)    GQSC, MIMCUT, PMOM( 0:MOMDIM, * ), QEXT, QSCA, SPIKE,XMU(*), XX
COMPLEX(EB) CREFIN, SFORW, SBACK, S1(*), S2(*), TFORW(*), TBACK(*)
! ----------------------------------------------------------------------
!
!                                  ** NOTE --  MAXTRM = 10100  IS NECES-
!                                  ** SARY TO DO SOME OF THE TEST PROBS,
!                                  ** BUT 1100 IS SUFFICIENT FOR MOST
!                                  ** CONCEIVABLE APPLICATIONS
!     .. PARAMETERS ..

INTEGER     MAXANG
PARAMETER   ( MAXANG = 5000 )
INTEGER     MAXTRM
!      PARAMETER ( MAXTRM = 10100 )
PARAMETER   ( MAXTRM = 16000 ) ! WORKS FOR FDS
REAL(EB)    ONETHR
PARAMETER   ( ONETHR = 1._EB / 3._EB )

!     .. LOCAL SCALARS ..

LOGICAL     NOABS, PASS1, YESANG, GT100
INTEGER     I, J, N, NANGD2, NPQUAN, NTRM
REAL(EB)    CHIN, CHINM1, COEFF, DENAN, DENBN, FN, MIM, MM, MRE, &
            NP1DN, PSIN, PSINM1, RATIO, RIORIV, RN, RTMP, TAUN, &
            TCOEF, TWONP1, XINV
COMPLEX(EB) AN, ANM1, ANP, ANPM, BN, BNM1, BNP, BNPM, CDENAN,&
            CDENBN, CIOR, CIORIV, CSUM1, CSUM2, ZET, ZETN, ZETNM1

!     .. LOCAL ARRAYS ..

LOGICAL     CALCMO( 4 )
REAL(EB), ALLOCATABLE :: PIN(:), PINM1(:)
REAL(EB) RBIGA( MAXTRM )
COMPLEX(EB) CBIGA( MAXTRM ), LITA( MAXTRM ), LITB( MAXTRM )
COMPLEX(EB), ALLOCATABLE :: SM(:), SMS(:), SP(:), SPS(:)

SAVE      PASS1
DATA      PASS1 / .TRUE. /
!
ALLOCATE(PIN(1:NUMANG))
ALLOCATE(PINM1(1:NUMANG))
ALLOCATE(SM(1:NUMANG))
ALLOCATE(SMS(1:(NUMANG+1)/2))
ALLOCATE(SP(1:NUMANG))
ALLOCATE(SPS(1:(NUMANG+1)/2))

!                    ** SAVE SOME INPUT VARIABLES AND REPLACE THEM
!                    ** WITH VALUES NEEDED TO DO THE SELF-TEST

IF( PASS1 ) CALL TESTMI( .FALSE., XX, CREFIN, MIMCUT, PERFCT, &
                         ANYANG, NMOM, IPOLZN, NUMANG, XMU, QEXT, &
                         QSCA, GQSC, SFORW, SBACK, S1, S2, TFORW, &
                         TBACK, PMOM, MOMDIM )
MAIN_MIEV: DO
!                                        ** CHECK INPUT AND CALCULATE
!                                        ** CERTAIN VARIABLES FROM INPUT
CALL CKINMI( NUMANG, MAXANG, XX, PERFCT, CREFIN, MOMDIM, NMOM, &
             IPOLZN, ANYANG, XMU, CALCMO, NPQUAN )

IF( PERFCT .AND. XX<=0.1_EB ) THEN
!                                            ** USE TOTALLY-REFLECTING
!                                            ** SMALL-PARTICLE LIMIT
   CALL SMALL1( XX, NUMANG, XMU, QEXT, QSCA, GQSC, SFORW, SBACK, &
                  S1, S2, TFORW, TBACK, LITA, LITB )
   NTRM = 2
ELSE
   NOABS = .TRUE.
   GT100 = .FALSE.

   IF( .NOT.PERFCT ) THEN
      CIOR = CREFIN
      IF( AIMAG(CIOR)>0.0_EB ) CIOR = CONJG( CIOR )
      MRE    = REAL( CIOR )
      MIM    = -AIMAG( CIOR )
      NOABS  = MIM<=MIMCUT
      CIORIV = 1.0_EB / CIOR
      RIORIV = 1.0_EB / MRE

      IF( XX*MAX( 1._EB, ABS(CIOR) )<=0.1_EB ) THEN
!                                    ** USE GENERAL-REFRACTIVE-INDEX
!                                    ** SMALL-PARTICLE LIMIT
         CALL SMALL2( XX, CIOR, MIM>MIMCUT, NUMANG, XMU, QEXT, &
                           QSCA, GQSC, SFORW, SBACK, S1, S2, TFORW, &
                           TBACK, LITA, LITB )
         NTRM = 2
         GT100 = .TRUE.
      END IF
   END IF
   GT100IF: IF (.NOT. GT100) THEN

      NANGD2 = ( NUMANG + 1 ) / 2
      YESANG = NUMANG>0
   !                             ** NUMBER OF TERMS IN MIE SERIES; EQ R50
      IF( XX<=8.0_EB ) THEN
      NTRM = INT(XX + 4._EB*XX**ONETHR + 1._EB)
      ELSE IF( XX<4200._EB ) THEN
      NTRM = INT(XX + 4.05*XX**ONETHR + 2._EB)
      ELSE
      NTRM =INT( XX + 4._EB*XX**ONETHR + 2._EB)
      END IF
      IF( NTRM+1 > MAXTRM )CALL ERRMSG('MIEV0--PARAMETER MAXTRM TOO SMALL',.TRUE.)
   !                            ** CALCULATE LOGARITHMIC DERIVATIVES OF
   !                            ** J-BESSEL-FCN., A-SUB-(1 TO NTRM)
      IF( .NOT.PERFCT ) CALL BIGA( CIOR, XX, NTRM, NOABS, YESANG, RBIGA, CBIGA )
   !                            ** INITIALIZE RICATTI-BESSEL FUNCTIONS
   !                            ** (PSI,CHI,ZETA)-SUB-(0,1) FOR UPWARD
   !                            ** RECURRENCE ( EQ. R19 )
      XINV   = 1.0_EB / XX
      PSINM1 = SIN( XX )
      CHINM1 = COS( XX )
      PSIN   = PSINM1*XINV - CHINM1
      CHIN   = CHINM1*XINV + PSINM1
      ZETNM1 = CMPLX( PSINM1, CHINM1, EB )
      ZETN   = CMPLX( PSIN, CHIN, EB )
   !                                     ** INITIALIZE PREVIOUS COEFFI-
   !                                     ** CIENTS FOR GQSC SERIES
      ANM1 = ( 0.0_EB, 0.0_EB )
      BNM1 = ( 0.0_EB, 0.0_EB )
   !                             ** INITIALIZE ANGULAR FUNCTION  PI
   !                             ** AND SUMS FOR S+, S- ( REF. 2, P. 1507 )
      IF( ANYANG ) THEN
      DO J = 1, NUMANG
   !                             ** EQ. R39
            PINM1( J ) = 0.0_EB
            PIN( J ) = 1.0_EB

               SP( J ) = ( 0.0_EB, 0.0_EB )
               SM( J ) = ( 0.0_EB, 0.0_EB )
         END DO
      ELSE
         DO J = 1, NANGD2
   !                          ** EQ. R39
            PINM1( J ) = 0.0_EB
            PIN( J ) = 1.0_EB
            SP( J ) = ( 0.0_EB, 0.0_EB )
            SM( J ) = ( 0.0_EB, 0.0_EB )
            SPS( J ) = ( 0.0_EB, 0.0_EB )
            SMS( J ) = ( 0.0_EB, 0.0_EB )
         END DO
      END IF
   !                       ** INITIALIZE MIE SUMS FOR EFFICIENCIES, ETC.
      QSCA  = 0.0_EB
      GQSC  = 0.0_EB
      SFORW = ( 0._EB, 0._EB )
      SBACK = ( 0._EB, 0._EB )
      CSUM1 = ( 0._EB, 0._EB )
      CSUM2 = ( 0._EB, 0._EB )
   !
   ! ---------  LOOP TO SUM MIE SERIES  -----------------------------------
      MM     = +1.0_EB
      SPIKE  = 1.0_EB
      DO N = 1, NTRM
   !                           ** COMPUTE VARIOUS NUMERICAL COEFFICIENTS
         FN     = N
         RN     = 1.0_EB / FN
         NP1DN  = 1.0_EB + RN
         TWONP1 = 2*N + 1
         COEFF  = TWONP1 / ( FN * ( N + 1 ) )
         TCOEF  = TWONP1 * ( FN * ( N + 1 ) )
   !                          ** CALCULATE MIE SERIES COEFFICIENTS
         IF( PERFCT ) THEN
   !                                 ** TOTALLY-REFLECTING CASE; EQ R/A.1,2
            AN = ( ( FN*XINV )*PSIN - PSINM1 ) / ( ( FN*XINV )*ZETN - ZETNM1 )
            BN = PSIN / ZETN
         ELSE IF( NOABS ) THEN
   !                                      ** NO-ABSORPTION CASE; EQ (R16)
            CDENAN = ( RIORIV*RBIGA(N) + ( FN*XINV ) ) * ZETN - ZETNM1
            AN   = ( ( RIORIV*RBIGA(N) + ( FN*XINV ) ) * PSIN - PSINM1 ) / CDENAN
            CDENBN = ( MRE*RBIGA(N) + ( FN*XINV ) ) * ZETN - ZETNM1
            BN   = ( ( MRE*RBIGA(N) + ( FN*XINV ) ) * PSIN - PSINM1 ) / CDENBN
         ELSE
   !                                       ** ABSORPTIVE CASE; EQ (R16)
            CDENAN = ( CIORIV*CBIGA( N ) + ( FN*XINV ) )*ZETN - ZETNM1
            CDENBN =   ( CIOR*CBIGA( N ) + ( FN*XINV ) )*ZETN - ZETNM1
            AN   = ( ( CIORIV*CBIGA( N ) + ( FN*XINV ) )*PSIN - PSINM1 ) / CDENAN
            BN     = ( ( CIOR*CBIGA( N ) + ( FN*XINV ) )*PSIN - PSINM1 ) / CDENBN
   !                                         ** EQ (R7)
            QSCA   = QSCA + TWONP1*( SQ( AN ) + SQ( BN ) )
         END IF
   !                       ** SAVE MIE COEFFICIENTS FOR PMOM CALCULATION
         LITA( N ) = AN
         LITB( N ) = BN
   !
         IF( .NOT.PERFCT .AND. N>XX ) THEN
   !                                               ** FLAG RESONANCE SPIKES
            DENAN  = ABS( CDENAN )
            DENBN  = ABS( CDENBN )
   !                                                   ** EQ. R/B.9
            RATIO  = DENAN / DENBN
   !                                                   ** EQ. R/B.10
            IF( RATIO<=0.2_EB .OR. RATIO>=5.0_EB ) SPIKE = MIN( SPIKE, DENAN, DENBN )
         END IF
   !                                  ** INCREMENT MIE SUMS FOR NON-ANGLE-
   !                                  ** DEPENDENT QUANTITIES
   !                                                   ** EQ. R/B.2
         SFORW = SFORW + TWONP1*( AN + BN )
   !                                                   ** EQ. R/B.5,6
         CSUM1 = CSUM1 + TCOEF *( AN - BN )
   !                                                   ** EQ. R/B.1
         SBACK = SBACK + ( MM*TWONP1 )*( AN - BN )
   !                                                   ** EQ. R/B.7,8
         CSUM2 = CSUM2 + ( MM*TCOEF ) *( AN + BN )
   !                                         ** EQ (R8)
         GQSC  = GQSC  + (FN - RN) * REAL(ANM1 * CONJG(AN) + BNM1 * CONJG(BN) ) + COEFF * REAL(AN * CONJG(BN))
         IF( YESANG ) THEN
   !                                      ** PUT MIE COEFFICIENTS IN FORM
   !                                      ** NEEDED FOR COMPUTING S+, S-
   !                                      ** ( EQ R10 )
         ANP = COEFF*( AN + BN )
         BNP = COEFF*( AN - BN )
   !                                      ** INCREMENT MIE SUMS FOR S+, S-
   !                                      ** WHILE UPWARD RECURSING
   !                                      ** ANGULAR FUNCTIONS PI AND TAU
         IF( ANYANG ) THEN
   !                                         ** ARBITRARY ANGLES
   !                                              ** VECTORIZABLE LOOP
            DO J = 1, NUMANG
   !                                                 ** EQ. (R37B)
               RTMP = ( XMU(J) * PIN(J) ) - PINM1( J )
   !                                                 ** EQ. (R38B)
               TAUN   = FN * RTMP - PINM1( J )
   !                                                   ** EQ (R10)
               SP( J ) = SP( J ) + ANP * ( PIN( J ) + TAUN )
               SM( J ) = SM( J ) + BNP * ( PIN( J ) - TAUN )
               PINM1( J ) = PIN( J )
   !                                                 ** EQ. R37C
               PIN( J ) = ( XMU( J ) * PIN( J ) ) + NP1DN * RTMP
            END DO
         ELSE
   !                                  ** ANGLES SYMMETRIC ABOUT 90 DEGREES
            ANPM = MM*ANP
            BNPM = MM*BNP
   !                                          ** VECTORIZABLE LOOP
            DO J = 1, NANGD2
   !                                                 ** EQ. (R37B)
               RTMP = ( XMU(J) * PIN(J) ) - PINM1( J )
   !                                                 ** EQ. (R38B)
               TAUN = FN * RTMP - PINM1( J )
   !                                                 ** EQ (R10,12)
               SP ( J ) = SP ( J ) + ANP * ( PIN( J ) + TAUN )
               SMS( J ) = SMS( J ) + BNPM *( PIN( J ) + TAUN )
               SM ( J ) = SM ( J ) + BNP * ( PIN( J ) - TAUN )
               SPS( J ) = SPS( J ) + ANPM *( PIN( J ) - TAUN )
               PINM1( J ) = PIN( J )
         !                                                 ** EQ. R37C
               PIN( J ) = ( XMU(J) * PIN(J) ) + NP1DN * RTMP
            END DO
         END IF
         END IF
   !                          ** UPDATE RELEVANT QUANTITIES FOR NEXT
   !                          ** PASS THROUGH LOOP
         MM   = - MM
         ANM1 = AN
         BNM1 = BN
   !                           ** UPWARD RECURRENCE FOR RICATTI-BESSEL
   !                           ** FUNCTIONS ( EQ. R17 )
         ZET    = ( TWONP1*XINV ) * ZETN - ZETNM1
         ZETNM1 = ZETN
         ZETN   = ZET
         PSINM1 = PSIN
         PSIN   = REAL( ZETN )
      END DO
   ! ---------- END LOOP TO SUM MIE SERIES --------------------------------
   !
   !                                         ** EQ (R6)
      QEXT = 2._EB / XX**2*REAL( SFORW )
      IF( PERFCT .OR. NOABS ) THEN
         QSCA = QEXT
      ELSE
         QSCA = 2._EB/ XX**2 * QSCA
      END IF
      GQSC   = 4._EB/ XX**2 * GQSC
      SFORW  = 0.5_EB*SFORW
      SBACK  = 0.5_EB*SBACK
      TFORW( 1 ) =  0.5_EB*SFORW - 0.125_EB*CSUM1
      TFORW( 2 ) =  0.5_EB*SFORW + 0.125_EB*CSUM1
      TBACK( 1 ) = -0.5_EB*SBACK + 0.125_EB*CSUM2
      TBACK( 2 ) =  0.5_EB*SBACK + 0.125_EB*CSUM2
      IF( YESANG ) THEN
   !                                ** RECOVER SCATTERING AMPLITUDES
   !                                ** FROM S+, S- ( EQ (R11) )
         IF( ANYANG ) THEN
   !                                         ** VECTORIZABLE LOOP
            DO J = 1, NUMANG
   !                                                  ** EQ (R11)
               S1( J ) = 0.5_EB*( SP( J ) + SM( J ) )
               S2( J ) = 0.5_EB*( SP( J ) - SM( J ) )
            END DO
         ELSE
   !                                        ** VECTORIZABLE LOOP
            DO J = 1, NANGD2
   !                                                   ** EQ (R11)
               S1( J ) = 0.5_EB*( SP( J ) + SM( J ) )
               S2( J ) = 0.5_EB*( SP( J ) - SM( J ) )
            END DO
   !                                        ** VECTORIZABLE LOOP
            DO  J = 1, NANGD2
               S1( NUMANG + 1 - J ) = 0.5_EB*( SPS( J ) + SMS( J ) )
               S2( NUMANG + 1 - J ) = 0.5_EB*( SPS( J ) - SMS( J ) )
            END DO
         END IF
      END IF
   END IF GT100IF
   !                                 ** CALCULATE LEGENDRE MOMENTS
END IF
IF( NMOM>0 ) CALL LPCOEF( NTRM, NMOM, IPOLZN, MOMDIM, CALCMO,NPQUAN, LITA, LITB, PMOM )
IF( AIMAG( CREFIN )>0.0_EB ) THEN
!                                         ** TAKE COMPLEX CONJUGATES
!                                         ** OF SCATTERING AMPLITUDES
   SFORW = CONJG( SFORW )
   SBACK = CONJG( SBACK )
   DO I = 1, 2
      TFORW( I ) = CONJG( TFORW( I ) )
      TBACK( I ) = CONJG( TBACK( I ) )
   END DO
   DO J = 1, NUMANG
      S1( J ) = CONJG( S1( J ) )
      S2( J ) = CONJG( S2( J ) )
   END DO
END IF

IF( PASS1 ) THEN
!                           ** COMPARE TEST CASE RESULTS WITH
!                           ** CORRECT ANSWERS AND ABORT IF BAD;
!                           ** OTHERWISE RESTORE USER INPUT AND PROCEED
   CALL TESTMI( .TRUE., XX, CREFIN, MIMCUT, PERFCT, ANYANG, NMOM, &
                  IPOLZN, NUMANG, XMU, QEXT, QSCA, GQSC, SFORW, &
                  SBACK, S1, S2, TFORW, TBACK, PMOM, MOMDIM )
   PASS1  = .FALSE.
   CYCLE MAIN_MIEV
END IF
EXIT MAIN_MIEV
END DO MAIN_MIEV

IF( PRNT( 1 ) .OR. PRNT( 2 ) )  &
  CALL MIPRNT( PRNT, XX, PERFCT, CREFIN, NUMANG, XMU, QEXT, &
               QSCA, GQSC, NMOM, IPOLZN, MOMDIM, CALCMO, PMOM, &
               SFORW, SBACK, TFORW, TBACK, S1, S2 )
!
DEALLOCATE(PIN)
DEALLOCATE(PINM1)
DEALLOCATE(SM)
DEALLOCATE(SMS)
DEALLOCATE(SP)
DEALLOCATE(SPS)

RETURN

END SUBROUTINE MIEV0

SUBROUTINE CKINMI( NUMANG, MAXANG, XX, PERFCT, CREFIN, MOMDIM, NMOM, IPOLZN, ANYANG, XMU, CALCMO, NPQUAN )

!        CHECK FOR BAD INPUT TO MIEV0 AND CALCULATE CALCMO, NPQUAN

!     ROUTINES CALLED :  ERRMSG, WRTBAD, WRTDIM

!     .. SCALAR ARGUMENTS ..

LOGICAL     ANYANG, PERFCT
INTEGER     IPOLZN, MAXANG, MOMDIM, NMOM, NPQUAN, NUMANG
REAL(EB)    XX
COMPLEX(EB) CREFIN
!     ..
!     .. ARRAY ARGUMENTS ..

LOGICAL     CALCMO( * )
REAL(EB)    XMU( * )
!     ..
!     .. LOCAL SCALARS ..

CHARACTER(4) :: STRING
LOGICAL     INPERR
INTEGER     I, IP, J, L

INPERR = .FALSE.

IF( NUMANG>MAXANG ) INPERR = WRTDIM( 'MAXANG', NUMANG )
IF( NUMANG<0 ) INPERR = WRTBAD( 'NUMANG' )

IF( XX<0._EB ) INPERR = WRTBAD( 'XX' )

IF( .NOT.PERFCT .AND. REAL( CREFIN )<=0._EB )INPERR = WRTBAD( 'CREFIN' )

IF( MOMDIM<0 ) INPERR = WRTBAD( 'MOMDIM' )


IF( NMOM/=0 ) THEN

   IF( NMOM<0 .OR. NMOM>MOMDIM ) INPERR = WRTBAD( 'NMOM' )
   IF( ABS( IPOLZN )>4444 ) INPERR = WRTBAD( 'IPOLZN' )

   NPQUAN = 0

   DO L = 1, 4
      CALCMO( L ) = .FALSE.
   END DO

   IF( IPOLZN/=0 ) THEN
   !                                 ** PARSE OUT IPOLZN INTO ITS DIGITS
   !                                 ** TO FIND WHICH PHASE QUANTITIES ARE
   !                                 ** TO HAVE THEIR MOMENTS CALCULATED
      WRITE( STRING, '(I4)' ) ABS( IPOLZN )
      DO J = 1, 4
         IP = ICHAR( STRING( J:J ) ) - ICHAR( '0' )
         IF( IP>=1 .AND. IP<=4 ) CALCMO( IP ) = .TRUE.
         IF( IP==0 .OR. ( IP>=5 .AND. IP<=9 ) ) INPERR = WRTBAD( 'IPOLZN' )
         NPQUAN = MAX( NPQUAN, IP )
      END DO
   END IF
END IF

IF( ANYANG ) THEN
!                                ** ALLOW FOR SLIGHT IMPERFECTIONS IN
!                                ** COMPUTATION OF COSINE
   DO I = 1, NUMANG
      IF( XMU( I )<-1.00001_EB .OR. XMU( I )>1.00001_EB ) INPERR = WRTBAD( 'XMU' )
   END DO
   ELSE
   DO I = 1, ( NUMANG + 1 ) / 2
      IF( XMU( I )<-0.00001_EB .OR. XMU( I )>1.00001_EB ) INPERR = WRTBAD( 'XMU' )
   END DO
END IF

IF( INPERR ) CALL ERRMSG( 'MIEV0--INPUT ERROR(S).  ABORTING...', .TRUE. )
IF( XX>20000.0_EB .OR. REAL( CREFIN )>10.0_EB .OR. ABS( AIMAG( CREFIN ) )>10.0_EB ) &
    CALL ERRMSG( 'MIEV0--XX OR CREFIN OUTSIDE TESTED RANGE',.FALSE.)
RETURN
END SUBROUTINE CKINMI

SUBROUTINE LPCOEF( NTRM, NMOM, IPOLZN, MOMDIM, CALCMO, NPQUAN, A, B, PMOM )
!         CALCULATE LEGENDRE POLYNOMIAL EXPANSION COEFFICIENTS (ALSO
!         CALLED MOMENTS) FOR PHASE QUANTITIES ( REF. 5 FORMULATION )

!     INPUT:  NTRM                    NUMBER TERMS IN MIE SERIES
!             NMOM, IPOLZN, MOMDIM    MIEV0 ARGUMENTS
!             CALCMO                  FLAGS CALCULATED FROM IPOLZN
!             NPQUAN                  DEFINED IN MIEV0
!             A, B                    MIE SERIES COEFFICIENTS
!
!     OUTPUT: PMOM                   LEGENDRE MOMENTS (MIEV0 ARGUMENT)
!
!     ROUTINES CALLED :  ERRMSG, LPCO1T, LPCO2T
!
!     *** NOTES ***
!
!         (1)  EQS. 2-5 ARE IN ERROR IN DAVE, APPL. OPT. 9,
!         1888 (1970).  EQ. 2 REFERS TO M1, NOT M2;  EQ. 3 REFERS TO
!         M2, NOT M1.  IN EQS. 4 AND 5, THE SUBSCRIPTS ON THE SECOND
!         TERM IN SQUARE BRACKETS SHOULD BE INTERCHANGED.
!
!         (2)  THE GENERAL-CASE LOGIC IN THIS SUBROUTINE WORKS CORRECTLY
!         IN THE TWO-TERM MIE SERIES CASE, BUT SUBROUTINE LPCO2T
!         IS CALLED INSTEAD, FOR SPEED.
!
!         (3)  SUBROUTINE  LPCO1T, TO DO THE ONE-TERM CASE, IS NEVER
!         CALLED WITHIN THE CONTEXT OF MIEV0, BUT IS INCLUDED FOR
!         COMPLETE GENERALITY.
!
!         (4)  SOME IMPROVEMENT IN SPEED IS OBTAINABLE BY COMBINING THE
!         310- AND 410-LOOPS, IF MOMENTS FOR BOTH THE THIRD AND FOURTH
!         PHASE QUANTITIES ARE DESIRED, BECAUSE THE THIRD PHASE QUANTITY
!         IS THE REAL PART OF A COMPLEX SERIES, WHILE THE FOURTH PHASE
!         QUANTITY IS THE IMAGINARY PART OF THAT VERY SAME SERIES.  BUT
!         MOST USERS ARE NOT INTERESTED IN THE FOURTH PHASE QUANTITY,
!         WHICH IS RELATED TO CIRCULAR POLARIZATION, SO THE PRESENT
!         SCHEME IS USUALLY MORE EFFICIENT.
!
!
!           ** DEFINITIONS OF LOCAL VARIABLES ***
!      AM(M)       NUMERICAL COEFFICIENTS  A-SUB-M-SUPER-L
!                     IN DAVE, EQS. 1-15, AS SIMPLIFIED IN REF. 5.
!      BI(I)       NUMERICAL COEFFICIENTS  B-SUB-I-SUPER-L
!                     IN DAVE, EQS. 1-15, AS SIMPLIFIED IN REF. 5.
!      BIDEL(I)    1/2 BI(I) TIMES FACTOR CAPITAL-DEL IN DAVE
!      CM,DM()     ARRAYS C AND D IN DAVE, EQS. 16-17 (MUELLER FORM),
!                     CALCULATED USING RECURRENCE DERIVED IN REF. 5
!      CS,DS()     ARRAYS C AND D IN REF. 4, EQS. A5-A6 (SEKERA FORM),
!                     CALCULATED USING RECURRENCE DERIVED IN REF. 5
!      C,D()       EITHER CM,DM OR CS,DS, DEPENDING ON IPOLZN
!      EVENL       TRUE FOR EVEN-NUMBERED MOMENTS;  FALSE OTHERWISE
!      IDEL        1 + LITTLE-DEL  IN DAVE
!      MAXTRM      MAX. NO. OF TERMS IN MIE SERIES
!      MAXMOM      MAX. NO. OF NON-ZERO MOMENTS
!      NUMMOM      NUMBER OF NON-ZERO MOMENTS
!      RECIP(K)    1 / K

IMPLICIT  NONE

!     .. PARAMETERS ..

INTEGER     MAXTRM, MAXMOM, MXMOM2, MAXRCP
PARAMETER ( MAXTRM=1001, MAXMOM = 2*MAXTRM, MXMOM2 = MAXMOM / 2, MAXRCP = 4*MAXTRM + 2 )
!     ..
!     .. SCALAR ARGUMENTS ..

INTEGER   IPOLZN, MOMDIM, NMOM, NPQUAN, NTRM
!     ..
!     .. ARRAY ARGUMENTS ..

LOGICAL     CALCMO( * )
REAL(EB)    PMOM( 0:MOMDIM, * )
COMPLEX(EB) A( * ), B( * )
!     ..
!     .. LOCAL SCALARS ..

LOGICAL     EVENL, PASS1
INTEGER     I, IDEL, IMAX, J, K, L, LD2, M, MMAX, NUMMOM
REAL(EB)    SUM
!     ..
!     .. LOCAL ARRAYS ..

REAL(EB)    AM( 0:MAXTRM ), BI( 0:MXMOM2 ), BIDEL( 0:MXMOM2 ), RECIP( MAXRCP )
COMPLEX(EB) C( MAXTRM ), CM( MAXTRM ), CS( MAXTRM ), D( MAXTRM ), DM( MAXTRM ), DS( MAXTRM )
!     ..
!     .. EQUIVALENCES ..

EQUIVALENCE ( C, CM ), ( D, DM )
!     ..
SAVE      PASS1, RECIP
DATA      PASS1 / .TRUE. /

IF( PASS1 ) THEN
   DO  K = 1, MAXRCP
      RECIP( K ) = 1.0_EB / K
   END DO
   PASS1  = .FALSE.
END IF

DO J = 1, MAX( 1, NPQUAN )
   DO L = 0, NMOM
      PMOM( L, J ) = 0.0_EB
   END DO
END DO

IF( NTRM==1 ) THEN
   CALL LPCO1T( NMOM, IPOLZN, MOMDIM, CALCMO, A, B, PMOM )
   RETURN
ELSE IF( NTRM==2 ) THEN
   CALL LPCO2T( NMOM, IPOLZN, MOMDIM, CALCMO, A, B, PMOM )
   RETURN
END IF

IF( NTRM + 2>MAXTRM )CALL ERRMSG('LPCOEF--PARAMETER MAXTRM TOO SMALL',.TRUE.)
!                                     ** CALCULATE MUELLER C, D ARRAYS
CM( NTRM + 2 ) = ( 0._EB, 0._EB )
DM( NTRM + 2 ) = ( 0._EB, 0._EB )
CM( NTRM + 1 ) = ( 1._EB - RECIP( NTRM+1 ) ) * B( NTRM )
DM( NTRM + 1 ) = ( 1._EB - RECIP( NTRM+1 ) ) * A( NTRM )
CM( NTRM ) = ( RECIP( NTRM ) + RECIP( NTRM+1 ) ) * A( NTRM ) + ( 1._EB - RECIP( NTRM ) )*B( NTRM-1 )
DM( NTRM ) = ( RECIP( NTRM ) + RECIP( NTRM+1 ) ) * B( NTRM ) + ( 1._EB - RECIP( NTRM ) )*A( NTRM-1 )
DO K = NTRM-1, 2, -1
   CM( K ) = CM( K+2 ) - ( 1._EB + RECIP(K+1) ) * B( K+1 )+ ( RECIP(K) + RECIP(K+1) ) * A( K ) &
                        + ( 1._EB - RECIP(K) ) * B( K-1 )
   DM( K ) = DM( K+2 ) - ( 1._EB + RECIP(K+1) ) * A( K+1 )  + ( RECIP(K) + RECIP(K+1) ) * B( K ) &
                        + ( 1._EB - RECIP(K) ) * A( K-1 )
END DO

CM( 1 ) = CM( 3 ) + 1.5_EB * ( A( 1 ) - B( 2 ) )
DM( 1 ) = DM( 3 ) + 1.5_EB * ( B( 1 ) - A( 2 ) )

IF( IPOLZN>=0 ) THEN
   DO K = 1, NTRM + 2
      C( K ) = ( 2*K - 1 ) * CM( K )
      D( K ) = ( 2*K - 1 ) * DM( K )
   END DO
ELSE
!                                    ** COMPUTE SEKERA C AND D ARRAYS
   CS( NTRM + 2 ) = ( 0._EB, 0._EB )
   DS( NTRM + 2 ) = ( 0._EB, 0._EB )
   CS( NTRM + 1 ) = ( 0._EB, 0._EB )
   DS( NTRM + 1 ) = ( 0._EB, 0._EB )

   DO K = NTRM, 1, -1
      CS( K ) = CS( K+2 ) + ( 2*K + 1 ) * ( CM( K+1 ) - B( K ) )
      DS( K ) = DS( K+2 ) + ( 2*K + 1 ) * ( DM( K+1 ) - A( K ) )
   END DO

   DO K = 1, NTRM + 2
      C( K ) = ( 2*K - 1 ) * CS( K )
      D( K ) = ( 2*K - 1 ) * DS( K )
   END DO

END IF

IF( IPOLZN<0 ) NUMMOM = MIN( NMOM, 2*NTRM - 2 )
IF( IPOLZN>=0 ) NUMMOM = MIN( NMOM, 2*NTRM )

IF( NUMMOM>MAXMOM )  CALL ERRMSG('LPCOEF--PARAMETER MAXTRM TOO SMALL',.TRUE.)
!
!                          ** LOOP OVER MOMENTS
L240: DO L = 0, NUMMOM
   LD2 = L / 2
   EVENL  = MOD( L, 2 )==0
!                                    ** CALCULATE NUMERICAL COEFFICIENTS
!                                    ** A-SUB-M AND B-SUB-I IN DAVE
!                                    ** DOUBLE-SUMS FOR MOMENTS
   IF( L==0 ) THEN
      IDEL = 1
      DO M = 0, NTRM
         AM( M ) = 2.0_EB * RECIP( 2*M + 1 )
      END DO
      BI( 0 ) = 1.0_EB
   ELSE IF( EVENL ) THEN
      IDEL = 1
      DO M = LD2, NTRM
         AM( M ) = ( 1._EB + RECIP( 2*M - L + 1 ) ) * AM( M )
      END DO
      DO I = 0, LD2 - 1
         BI( I ) = ( 1._EB - RECIP( L - 2*I ) ) * BI( I )
      END DO
      BI( LD2 ) = ( 2._EB - RECIP( L ) ) * BI( LD2 - 1 )
   ELSE
   IDEL = 2
      DO M = LD2, NTRM
         AM( M ) = ( 1._EB - RECIP( 2*M + L + 2 ) ) * AM( M )
      END DO

      DO I = 0, LD2
         BI( I ) = ( 1._EB - RECIP( L + 2*I + 1 ) ) * BI( I )
      END DO
   END IF
!                                     ** ESTABLISH UPPER LIMITS FOR SUMS
!                                     ** AND INCORPORATE FACTOR CAPITAL-
!                                     ** DEL INTO B-SUB-I
   MMAX = NTRM - IDEL
   IF( IPOLZN>=0 ) MMAX = MMAX + 1
   IMAX = MIN( LD2, MMAX - LD2 )
   IF( IMAX<0 ) EXIT L240
   DO I = 0, IMAX
      BIDEL( I ) = BI( I )
   END DO
   IF( EVENL ) BIDEL( 0 ) = 0.5_EB*BIDEL( 0 )
!                                    ** PERFORM DOUBLE SUMS JUST FOR
!                                    ** PHASE QUANTITIES DESIRED BY USER
   IF( IPOLZN==0 ) THEN
      DO I = 0, IMAX
!                                           ** VECTORIZABLE LOOP
         SUM = 0.0_EB
         DO M = LD2, MMAX - I
            SUM = SUM + AM(M) * ( REAL(C(M-I+1) * CONJG(C(M+I+IDEL)))+ REAL(D(M-I+1) * CONJG(D(M+I+IDEL))))
         END DO
         PMOM( L, 1 ) = PMOM( L, 1 ) + BIDEL( I ) * SUM
      END DO
      PMOM( L, 1 ) = 0.5_EB*PMOM( L, 1 )
      CYCLE L240
   END IF

   IF( CALCMO( 1 ) ) THEN
      DO I = 0, IMAX
         SUM = 0.0_EB
!                                          ** VECTORIZABLE LOOP
         DO M = LD2, MMAX - I
            SUM = SUM + AM( M ) * REAL( C(M-I+1) * CONJG( C(M+I+IDEL) ) )
         END DO
         PMOM( L, 1 ) = PMOM( L, 1 ) + BIDEL( I ) * SUM
      END DO
   END IF

   IF( CALCMO( 2 ) ) THEN
      DO I = 0, IMAX
         SUM = 0.0_EB
!                                           ** VECTORIZABLE LOOP
         DO M = LD2, MMAX - I
            SUM = SUM + AM( M ) * REAL( D(M-I+1) * CONJG( D(M+I+IDEL) ) )
         END DO
         PMOM( L, 2 ) = PMOM( L, 2 ) + BIDEL( I ) * SUM
      END DO
   END IF

   IF( CALCMO( 3 ) ) THEN
      DO I = 0, IMAX
         SUM = 0.0_EB
!                                           ** VECTORIZABLE LOOP
         DO M = LD2, MMAX - I
            SUM = SUM + AM(M) *( REAL(C(M-I+1) * CONJG(D(M+I+IDEL))) + REAL(C(M+I+IDEL) * CONJG(D(M-I+1))))
         END DO
         PMOM( L, 3 ) = PMOM( L, 3 ) + BIDEL( I ) * SUM
      END DO
      PMOM( L, 3 ) = 0.5_EB*PMOM( L, 3 )
   END IF

   IF( CALCMO( 4 ) ) THEN
      DO I = 0, IMAX
         SUM= 0.0_EB
!                                         ** VECTORIZABLE LOOP
         DO M = LD2, MMAX - I
            SUM = SUM + AM(M) *(AIMAG(C(M-I+1) * CONJG(D(M+I+IDEL)))+ AIMAG(C(M+I+IDEL) * CONJG(D(M-I+1))))
         END DO
         PMOM( L, 4 ) = PMOM( L, 4 ) + BIDEL( I ) * SUM
      END DO

      PMOM( L, 4 ) = - 0.5_EB * PMOM( L, 4 )
   END IF

END DO L240

RETURN
END SUBROUTINE LPCOEF

SUBROUTINE LPCO1T( NMOM, IPOLZN, MOMDIM, CALCMO, A, B, PMOM )
!
!         CALCULATE LEGENDRE POLYNOMIAL EXPANSION COEFFICIENTS (ALSO
!         CALLED MOMENTS) FOR PHASE QUANTITIES IN SPECIAL CASE WHERE
!         NO. TERMS IN MIE SERIES = 1
!
!        INPUT:  NMOM, IPOLZN, MOMDIM     MIEV0 ARGUMENTS
!                CALCMO                   FLAGS CALCULATED FROM IPOLZN
!                A(1), B(1)               MIE SERIES COEFFICIENTS
!
!        OUTPUT: PMOM                     LEGENDRE MOMENTS

!     .. SCALAR ARGUMENTS ..

INTEGER     IPOLZN, MOMDIM, NMOM
!     ..
!     .. ARRAY ARGUMENTS ..

LOGICAL     CALCMO( * )
REAL(EB)    PMOM( 0:MOMDIM, * )
COMPLEX(EB) A( * ), B( * )
!     ..
!     .. LOCAL SCALARS ..

INTEGER     L, NUMMOM
REAL(EB)    A1SQ, B1SQ
COMPLEX(EB) A1B1C

A1SQ   = SQ( A( 1 ) )
B1SQ   = SQ( B( 1 ) )
A1B1C  = A( 1 ) * CONJG( B( 1 ) )


IF( IPOLZN<0 ) THEN
   IF( CALCMO( 1 ) ) PMOM( 0, 1 ) = 2.25_EB*B1SQ
   IF( CALCMO( 2 ) ) PMOM( 0, 2 ) = 2.25_EB*A1SQ
   IF( CALCMO( 3 ) ) PMOM( 0, 3 ) = 2.25_EB*REAL( A1B1C )
   IF( CALCMO( 4 ) ) PMOM( 0, 4 ) = 2.25_EB*AIMAG( A1B1C )
ELSE
   NUMMOM = MIN( NMOM, 2 )
   !                             ** LOOP OVER MOMENTS
   L10: DO L = 0, NUMMOM
      IF( IPOLZN==0 ) THEN
         IF( L==0 ) PMOM( L, 1 ) = 1.5_EB*( A1SQ + B1SQ )
         IF( L==1 ) PMOM( L, 1 ) = 1.5_EB*REAL( A1B1C )
         IF( L==2 ) PMOM( L, 1 ) = 0.15_EB*( A1SQ + B1SQ )
         CYCLE L10
      END IF

      IF( CALCMO( 1 ) ) THEN
         IF( L==0 ) PMOM( L, 1 ) = 2.25_EB*( A1SQ + B1SQ / 3.)
         IF( L==1 ) PMOM( L, 1 ) = 1.5_EB*REAL( A1B1C )
         IF( L==2 ) PMOM( L, 1 ) = 0.3_EB*B1SQ
      END IF

      IF( CALCMO( 2 ) ) THEN
         IF( L==0 ) PMOM( L, 2 ) = 2.25_EB*( B1SQ + A1SQ / 3. )
         IF( L==1 ) PMOM( L, 2 ) = 1.5_EB*REAL( A1B1C )
         IF( L==2 ) PMOM( L, 2 ) = 0.3_EB*A1SQ
      END IF

      IF( CALCMO( 3 ) ) THEN
         IF( L==0 ) PMOM( L, 3 ) = 3.0_EB*REAL( A1B1C )
         IF( L==1 ) PMOM( L, 3 ) = 0.75_EB*( A1SQ + B1SQ )
         IF( L==2 ) PMOM( L, 3 ) = 0.3_EB*REAL( A1B1C )
      END IF

      IF( CALCMO( 4 ) ) THEN
         IF( L==0 ) PMOM( L, 4 ) = -1.5_EB*AIMAG( A1B1C )
         IF( L==1 ) PMOM( L, 4 ) = 0.0_EB
         IF( L==2 ) PMOM( L, 4 ) = 0.3_EB*AIMAG( A1B1C )
      END IF

   END DO L10
END IF

RETURN
END SUBROUTINE LPCO1T

SUBROUTINE LPCO2T( NMOM, IPOLZN, MOMDIM, CALCMO, A, B, PMOM )
!         CALCULATE LEGENDRE POLYNOMIAL EXPANSION COEFFICIENTS (ALSO
!         CALLED MOMENTS) FOR PHASE QUANTITIES IN SPECIAL CASE WHERE
!         NO. TERMS IN MIE SERIES = 2

!        INPUT:  NMOM, IPOLZN, MOMDIM     MIEV0 ARGUMENTS
!                CALCMO                   FLAGS CALCULATED FROM IPOLZN
!                A(1-2), B(1-2)           MIE SERIES COEFFICIENTS
!
!        OUTPUT: PMOM                     LEGENDRE MOMENTS


!     .. SCALAR ARGUMENTS ..

INTEGER     IPOLZN, MOMDIM, NMOM
!     ..
!     .. ARRAY ARGUMENTS ..

LOGICAL     CALCMO( * )
REAL(EB)    PMOM( 0:MOMDIM, * )
COMPLEX(EB) A( * ), B( * )
!     ..
!     .. LOCAL SCALARS ..

INTEGER     L, NUMMOM
REAL(EB)    A2SQ, B2SQ, PM1, PM2
COMPLEX(EB) A2C, B2C, CA, CAC, CAT, CB, CBC, CBT, CG, CH

CA   = 3.*A( 1 ) - 5.*B( 2 )
CAT  = 3.*B( 1 ) - 5.*A( 2 )
CAC  = CONJG( CA )
A2SQ = SQ( A( 2 ) )
B2SQ = SQ( B( 2 ) )
A2C  = CONJG( A( 2 ) )
B2C  = CONJG( B( 2 ) )

IF( IPOLZN<0 ) THEN
!                                   ** LOOP OVER SEKERA MOMENTS
   NUMMOM = MIN( NMOM, 2 )
   DO L = 0, NUMMOM
      IF( CALCMO( 1 ) ) THEN
         IF( L==0 ) PMOM( L, 1 ) = 0.25_EB * ( SQ( CAT ) + (100._EB/3._EB)* B2SQ )
         IF( L==1 ) PMOM( L, 1 ) = (5._EB/3._EB)*REAL( CAT*B2C )
         IF( L==2 ) PMOM( L, 1 ) = (10._EB/3._EB)*B2SQ
      END IF

      IF( CALCMO( 2 ) ) THEN
         IF( L==0 ) PMOM( L, 2 ) = 0.25_EB * ( SQ( CA ) + (100._EB/3._EB) * A2SQ )
         IF( L==1 ) PMOM( L, 2 ) = (5._EB/3._EB)*REAL( CA*A2C )
         IF( L==2 ) PMOM( L, 2 ) = (10._EB/3._EB)*A2SQ
      END IF

      IF( CALCMO( 3 ) ) THEN
         IF( L==0 ) PMOM( L, 3 ) = 0.25_EB * REAL( CAT * CAC  + (100._EB/3._EB) * B(2) * A2C )
         IF( L==1 ) PMOM( L, 3 ) = 5._EB/6._EB* REAL( B(2)*CAC + CAT*A2C )
         IF( L==2 ) PMOM( L, 3 ) = 10._EB/3._EB* REAL( B(2)*A2C )
      END IF

      IF( CALCMO( 4 ) ) THEN
         IF( L==0 ) PMOM( L, 4 ) = -0.25_EB * AIMAG( CAT * CAC + (100._EB/3._EB)* B(2) * A2C )
         IF( L==1 ) PMOM( L, 4 ) = -5._EB/ 6._EB*  AIMAG( B(2)*CAC + CAT*A2C )
         IF( L==2 ) PMOM( L, 4 ) = -10._EB/ 3._EB* AIMAG( B(2)*A2C )
      END IF

   END DO

ELSE

   CB  = 3._EB*B( 1 ) + 5._EB*A( 2 )
   CBT = 3._EB*A( 1 ) + 5._EB*B( 2 )
   CBC = CONJG( CB )
   CG  = ( CBC*CBT + 10._EB*( CAC*A( 2 ) + B2C*CAT ) ) / 3._EB
   CH  = 2._EB*( CBC*A( 2 ) + B2C*CBT )

   !                               ** LOOP OVER MUELLER MOMENTS
   NUMMOM = MIN( NMOM, 4 )

   L20: DO L = 0, NUMMOM

      IF( IPOLZN==0 .OR. CALCMO( 1 ) ) THEN
         IF( L==0 ) PM1 = 0.25_EB*SQ( CA ) + SQ( CB ) / 12._EB + (5._EB/3._EB)*REAL( CA*B2C ) + 5._EB*B2SQ
         IF( L==1 ) PM1 = REAL( CB * ( CAC / 6._EB+ B2C ) )
         IF( L==2 ) PM1 = SQ( CB ) / 30._EB+ (20._EB/7._EB)*B2SQ + (2._EB/3._EB)*REAL( CA*B2C )
         IF( L==3 ) PM1 = (2._EB/7.) * REAL( CB*B2C )
         IF( L==4 ) PM1 = (40._EB/63._EB) * B2SQ
         IF( CALCMO( 1 ) ) PMOM( L, 1 ) = PM1
      END IF

      IF( IPOLZN==0 .OR. CALCMO( 2 ) ) THEN
         IF( L==0 ) PM2 = 0.25_EB*SQ( CAT ) + SQ( CBT ) / 12._EB + ( 5._EB/ 3._EB) * REAL( CAT*A2C ) + 5._EB*A2SQ
         IF( L==1 ) PM2 = REAL( CBT * ( CONJG( CAT ) / 6._EB+ A2C ) )
         IF( L==2 ) PM2 = SQ( CBT ) / 30._EB + ( 20._EB/7._EB) * A2SQ + ( 2._EB/3._EB) * REAL( CAT*A2C )
         IF( L==3 ) PM2 = (2._EB/7._EB) * REAL( CBT*A2C )
         IF( L==4 ) PM2 = (40._EB/63._EB) * A2SQ
         IF( CALCMO( 2 ) ) PMOM( L, 2 ) = PM2
      END IF

      IF( IPOLZN==0 ) THEN
         PMOM( L, 1 ) = 0.5_EB*( PM1 + PM2 )
         CYCLE L20
      END IF

      IF( CALCMO( 3 ) ) THEN
         IF( L==0 ) PMOM( L, 3 ) = 0.25_EB * REAL( CAC*CAT + CG + 20._EB* B2C * A(2) )
         IF( L==1 ) PMOM( L, 3 ) = REAL( CAC*CBT + CBC*CAT + 3._EB*CH ) / 12.
         IF( L==2 ) PMOM( L, 3 ) = 0.1_EB * REAL( CG + (200._EB/7._EB) * B2C * A(2) )
         IF( L==3 ) PMOM( L, 3 ) = REAL( CH ) / 14._EB
         IF( L==4 ) PMOM( L, 3 ) = 40._EB/63._EB* REAL( B2C*A(2) )
      END IF

      IF( CALCMO( 4 ) ) THEN
         IF( L==0 ) PMOM( L, 4 ) = 0.25_EB * AIMAG( CAC*CAT + CG + 20._EB* B2C * A(2) )
         IF( L==1 ) PMOM( L, 4 ) = AIMAG( CAC*CBT + CBC*CAT + 3._EB*CH ) / 12._EB
         IF( L==2 ) PMOM( L, 4 ) = 0.1_EB * AIMAG( CG + (200._EB/7._EB) * B2C * A(2) )
         IF( L==3 ) PMOM( L, 4 ) = AIMAG( CH ) / 14._EB
         IF( L==4 ) PMOM( L, 4 ) = 40._EB/63._EB* AIMAG( B2C*A(2) )
      END IF

   END DO L20

END IF

RETURN
END SUBROUTINE LPCO2T

SUBROUTINE BIGA( CIOR, XX, NTRM, NOABS, YESANG, RBIGA, CBIGA )

!        CALCULATE LOGARITHMIC DERIVATIVES OF J-BESSEL-FUNCTION
!
!     INPUT :  CIOR, XX, NTRM, NOABS, YESANG  (DEFINED IN MIEV0)
!
!    OUTPUT :  RBIGA OR CBIGA  (DEFINED IN MIEV0)
!
!    ROUTINES CALLED :  CONFRA
!
!
!    INTERNAL VARIABLES :
!
!       CONFRA     VALUE OF LENTZ CONTINUED FRACTION FOR CBIGA(NTRM),
!                     USED TO INITIALIZE DOWNWARD RECURRENCE
!
!       DOWN       = TRUE, USE DOWN-RECURRENCE.  FALSE, DO NOT.
!
!       F1,F2,F3   ARITHMETIC STATEMENT FUNCTIONS USED IN DETERMINING
!                     WHETHER TO USE UP-  OR DOWN-RECURRENCE
!                     ( REF. 2, EQS. 6-8 )
!
!       MRE        REAL REFRACTIVE INDEX
!       MIM        IMAGINARY REFRACTIVE INDEX
!
!       REZINV     1 / ( MRE * XX ); TEMPORARY VARIABLE FOR RECURRENCE
!       ZINV       1 / ( CIOR * XX ); TEMPORARY VARIABLE FOR RECURRENCE
!
!     .. SCALAR ARGUMENTS ..

LOGICAL     NOABS, YESANG
INTEGER     NTRM
REAL(EB)    XX
COMPLEX(EB) CIOR

!     .. ARRAY ARGUMENTS ..
REAL(EB)    RBIGA( * )
COMPLEX(EB) CBIGA( * )

!     .. LOCAL SCALARS ..
LOGICAL     DOWN
INTEGER     N
REAL(EB)    MIM, MRE, REZINV, RTMP
COMPLEX(EB) CTMP, ZINV

!                                  ** DECIDE WHETHER BIGA CAN BE
!                                  ** CALCULATED BY UP-RECURRENCE
MRE = REAL( CIOR )
MIM = ABS( AIMAG( CIOR ) )
IF( MRE<1.0_EB .OR. MRE>10.0_EB .OR. MIM>10.0_EB ) THEN
   DOWN = .TRUE.
ELSE IF( YESANG ) THEN
   DOWN = .TRUE.
!                                                    ** EQ. R48
   IF( MIM*XX < F2( MRE ) ) DOWN = .FALSE.
ELSE
   DOWN = .TRUE.
!                                                    ** EQ. R48
   IF( MIM*XX < F1( MRE ) ) DOWN = .FALSE.
END IF

ZINV   = 1.0_EB / ( CIOR*XX )
REZINV = 1.0_EB / ( MRE*XX )


IF( DOWN ) THEN
!                          ** COMPUTE INITIAL HIGH-ORDER BIGA USING
!                          ** LENTZ METHOD ( REF. 1, PP. 17-20 )
   CTMP = CONFRA( NTRM, ZINV )
!                                   *** DOWNWARD RECURRENCE FOR BIGA
   IF( NOABS ) THEN
!                                        ** NO-ABSORPTION CASE; EQ (R23)
      RBIGA( NTRM ) = REAL( CTMP )
      DO N = NTRM, 2, -1
         RBIGA( N - 1 ) = ( N*REZINV ) - 1.0_EB / ( ( N*REZINV ) + RBIGA( N ) )
      END DO
   ELSE
!                                         ** ABSORPTIVE CASE; EQ (R23)
      CBIGA( NTRM ) = CTMP
      DO N = NTRM, 2, -1
         CBIGA( N-1 ) = (N*ZINV) - 1.0_EB / ( (N*ZINV) + CBIGA( N ) )
   END DO
   END IF

ELSE
!                            *** UPWARD RECURRENCE FOR BIGA
   IF( NOABS ) THEN
!                                  ** NO-ABSORPTION CASE; EQ (R20,21)
      RTMP = SIN( MRE*XX )
      RBIGA( 1 ) = - REZINV + RTMP / ( RTMP*REZINV - COS( MRE*XX ) )

      DO N = 2, NTRM
         RBIGA( N ) = -( N*REZINV ) + 1.0_EB / ( ( N*REZINV ) - RBIGA( N - 1 ) )
      END DO

   ELSE
!                                     ** ABSORPTIVE CASE; EQ (R20,22)
      CTMP = EXP( - (0._EB,2._EB)*CIOR*XX )
      CBIGA( 1 ) = - ZINV + (1._EB-CTMP) /  ( ZINV * (1._EB-CTMP) - (0._EB,1._EB)*(1._EB+CTMP) )
      DO N = 2, NTRM
         CBIGA( N ) = - (N*ZINV) + 1.0_EB / ((N*ZINV) - CBIGA( N-1 ))
      END DO

END IF

END IF

CONTAINS

REAL(EB) FUNCTION F1(MRE)
REAL(EB), INTENT(IN) :: MRE
F1 = -8.0_EB + MRE**2*( 26.22_EB + MRE*( -0.4474_EB + MRE**3*( 0.00204_EB - 0.000175_EB*MRE ) ) )
END FUNCTION F1

REAL(EB) FUNCTION F2(MRE)  ! EQ. R47B
REAL(EB), INTENT(IN) :: MRE
F2 = 3.9_EB + MRE*( -10.8_EB + 13.78_EB*MRE )
END FUNCTION F2

END SUBROUTINE BIGA

COMPLEX(EB) FUNCTION CONFRA( N, ZINV )

!         COMPUTE BESSEL FUNCTION RATIO A-SUB-N FROM ITS
!         CONTINUED FRACTION USING LENTZ METHOD
!
!         ZINV = RECIPROCAL OF ARGUMENT OF A
!
!
!    I N T E R N A L    V A R I A B L E S
!    ------------------------------------
!
!    CAK      TERM IN CONTINUED FRACTION EXPANSION OF A (EQ. R25)
!    CAPT     FACTOR USED IN LENTZ ITERATION FOR A (EQ. R27)
!    CNUMER   NUMERATOR   IN CAPT  ( EQ. R28A )
!    CDENOM   DENOMINATOR IN CAPT  ( EQ. R28B )
!    CDTD     PRODUCT OF TWO SUCCESSIVE DENOMINATORS OF CAPT FACTORS
!                 ( EQ. R34C )
!    CNTN     PRODUCT OF TWO SUCCESSIVE NUMERATORS OF CAPT FACTORS
!                 ( EQ. R34B )
!    EPS1     ILL-CONDITIONING CRITERION
!    EPS2     CONVERGENCE CRITERION
!    KK       SUBSCRIPT K OF CAK  ( EQ. R25B )
!    KOUNT    ITERATION COUNTER ( USED TO PREVENT INFINITE LOOPING )
!    MAXIT    MAX. ALLOWED NO. OF ITERATIONS
!    MM       + 1  AND - 1, ALTERNATELY
! --------------------------------------------------------------------

!     .. SCALAR ARGUMENTS ..

INTEGER     N
COMPLEX(EB) ZINV

!     .. LOCAL SCALARS ..

INTEGER     KK, KOUNT, MAXIT, MM
REAL(EB)    EPS1, EPS2
COMPLEX(EB) CAK, CAPT, CDENOM, CDTD, CNTN, CNUMER
!
DATA      EPS1 / 1.E-2_EB / , EPS2 / 1.E-8_EB /
DATA      MAXIT / 10000 /
!                                 ** EQ. R25A
CONFRA = ( N + 1 ) * ZINV
MM     = - 1
KK     = 2*N + 3
!                                 ** EQ. R25B, K=2
CAK    = ( MM*KK ) * ZINV
CDENOM = CAK
CNUMER = CDENOM + 1.0_EB / CONFRA
KOUNT  = 1

L10: DO
   KOUNT = KOUNT + 1

   IF( KOUNT>MAXIT ) CALL ERRMSG('CONFRA--ITERATION FAILED TO CONVERGE',.TRUE.)

   MM  = - MM
   KK  = KK + 2
   !                                 ** EQ. R25B
   CAK = ( MM*KK ) * ZINV
   !                                          ** EQ. R32
   IF( ABS( CNUMER / CAK )<=EPS1 .OR. ABS( CDENOM / CAK )<=EPS1 ) THEN

   !                                  ** ILL-CONDITIONED CASE -- STRIDE
   !                                  ** TWO TERMS INSTEAD OF ONE

   !                                       ** EQ. R34
      CNTN   = CAK * CNUMER + 1.0_EB
      CDTD   = CAK * CDENOM + 1.0_EB
   !                                           ** EQ. R33
      CONFRA = ( CNTN / CDTD ) * CONFRA

      MM  = - MM
      KK  = KK + 2
   !                                 ** EQ. R25B
      CAK = ( MM*KK ) * ZINV
   !                                      ** EQ. R35
      CNUMER = CAK + CNUMER / CNTN
      CDENOM = CAK + CDENOM / CDTD
      KOUNT  = KOUNT + 1
      CYCLE L10

   ELSE
   !                           *** WELL-CONDITIONED CASE

   !                                  ** EQ. R27
      CAPT   = CNUMER / CDENOM
   !                                  ** EQ. R26
      CONFRA = CAPT * CONFRA
   !                                  ** CHECK FOR CONVERGENCE; EQ. R31

      IF (      ABS( REAL (CAPT) - 1.0_EB )>=EPS2 .OR. ABS( AIMAG(CAPT) )      >=EPS2 )  THEN
   !                                        ** EQ. R30
         CNUMER = CAK + 1.0_EB / CNUMER
         CDENOM = CAK + 1.0_EB / CDENOM
         CYCLE L10
      END IF
   END IF
   EXIT L10
END DO L10

RETURN
END FUNCTION CONFRA

SUBROUTINE MIPRNT( PRNT, XX, PERFCT, CREFIN, NUMANG, XMU, QEXT, QSCA, GQSC, NMOM, IPOLZN, MOMDIM, &
                   CALCMO, PMOM, SFORW, SBACK, TFORW, TBACK, S1, S2 )

!         PRINT SCATTERING QUANTITIES OF A SINGLE PARTICLE

!     .. SCALAR ARGUMENTS ..
LOGICAL     PERFCT
INTEGER     IPOLZN, MOMDIM, NMOM, NUMANG
REAL(EB)    GQSC, QEXT, QSCA, XX
COMPLEX(EB) CREFIN, SBACK, SFORW

!     .. ARRAY ARGUMENTS ..
LOGICAL     CALCMO( * ), PRNT( * )
REAL(EB)    PMOM( 0:MOMDIM, * ), XMU( * )
COMPLEX(EB) S1( * ), S2( * ), TBACK( * ), TFORW( * )

!     .. LOCAL SCALARS ..
CHARACTER   FMAT*22
INTEGER     I, J, M
REAL(EB)    FNORM, I1, I2

!     .. INTRINSIC FUNCTIONS ..
!     INTRINSIC AIMAG, CONJG, REAL


IF( PERFCT ) WRITE(LU_ERR, '(''1'',10X,A,1P,E11.4)' )  'PERFECTLY CONDUCTING CASE, SIZE PARAMETER =', XX

IF( .NOT.PERFCT ) WRITE(LU_ERR, '(''1'',10X,3(A,1P,E11.4))' ) &
    'REFRACTIVE INDEX:  REAL ', REAL( CREFIN ), '  IMAG ', &
    AIMAG( CREFIN ), ',   SIZE PARAMETER =', XX


IF( PRNT( 1 ) .AND. NUMANG>0 ) THEN

   WRITE(LU_ERR, '(/,A)' ) &
         '    COS(ANGLE)  ------- S1 ---------  ------- S2 ---------' &
         // '  --- S1*CONJG(S2) ---   I1=S1**2   I2=S2**2  (I1+I2)/2' &
         // '  DEG POLZN'

   DO I = 1, NUMANG
      I1 = REAL( S1( I ) )**2 + AIMAG( S1( I ) )**2
      I2 = REAL( S2( I ) )**2 + AIMAG( S2( I ) )**2
      WRITE(LU_ERR, '( I4, F10.6, 1P,10E11.3 )'   ) I, XMU(I), S1(I), S2(I), &
         S1(I)*CONJG(S2(I)),I1, I2, 0.5_EB*(I1+I2), (I2-I1)/(I2+I1)
   END DO
END IF


IF( PRNT( 2 ) ) THEN

   WRITE (LU_ERR, '(/,A,9X,A,17X,A,17X,A,/,(0P,F7.2, 1P,6E12.3) )' ) &
            '  ANGLE', 'S-SUB-1', 'T-SUB-1', 'T-SUB-2', &
                  0.0_EB,     SFORW,    TFORW(1),  TFORW(2), &
               180._EB,     SBACK,    TBACK(1),  TBACK(2)
   WRITE (LU_ERR, '(/,4(A,1P,E11.4))' ) &
            ' EFFICIENCY FACTORS,  EXTINCTION:', QEXT, &
                                 '   SCATTERING:', QSCA, &
                                 '   ABSORPTION:', QEXT-QSCA, &
                              '   RAD. PRESSURE:', QEXT-GQSC
   IF( NMOM>0 ) THEN
      WRITE(LU_ERR, '(/,A)' ) ' NORMALIZED MOMENTS OF :'
      IF( IPOLZN==0 ) WRITE(LU_ERR, '(''+'',27X,A)' ) 'PHASE FCN'
      IF( IPOLZN>0 ) WRITE(LU_ERR, '(''+'',33X,A)' )  'M1           M2          S21          D21'

      IF( IPOLZN<0 ) WRITE(LU_ERR, '(''+'',33X,A)' )  'R1           R2           R3           R4'
      FNORM = 4._EB/ ( XX**2 * QSCA )
      DO M = 0, NMOM
         WRITE(LU_ERR, '(A,I4)' ) '      MOMENT NO.', M
         DO J = 1, 4
            IF( CALCMO( J ) ) THEN
               WRITE( FMAT, '(A,I2,A)' ) '( ''+'', T', 24+(J-1)*13, ', 1P,E13.4 )'
               WRITE(LU_ERR, FMAT ) FNORM * PMOM( M, J )
            END IF
         END DO
      END DO
   END IF
END IF

RETURN
END SUBROUTINE MIPRNT

SUBROUTINE SMALL1( XX, NUMANG, XMU, QEXT, QSCA, GQSC, SFORW, SBACK, S1, S2, TFORW, TBACK, A, B )

!       SMALL-PARTICLE LIMIT OF MIE QUANTITIES IN TOTALLY REFLECTING
!       LIMIT ( MIE SERIES TRUNCATED AFTER 2 TERMS )
!
!        A,B       FIRST TWO MIE COEFFICIENTS, WITH NUMERATOR AND
!                  DENOMINATOR EXPANDED IN POWERS OF XX ( A FACTOR
!                  OF XX**3 IS MISSING BUT IS RESTORED BEFORE RETURN
!                  TO CALLING PROGRAM )  ( REF. 2, P. 1508 )

!     .. PARAMETERS ..
REAL(EB)  TWOTHR, FIVTHR, FIVNIN
PARAMETER ( TWOTHR = 2._EB/3._EB, FIVTHR = 5._EB/3._EB, FIVNIN = 5._EB/9._EB )

!     .. SCALAR ARGUMENTS ..
INTEGER     NUMANG
REAL(EB)    GQSC, QEXT, QSCA, XX
COMPLEX(EB) SBACK, SFORW

!     .. ARRAY ARGUMENTS ..
REAL(EB)    XMU( * )
COMPLEX(EB) A(*), B( * ), S1( * ), S2( * ), TBACK( * ), TFORW( * )

!     .. LOCAL SCALARS ..
INTEGER     J
REAL(EB)    RTMP

!                                                       ** EQ. R/A.5
A( 1 ) = CMPLX( 0._EB,    TWOTHR*( 1._EB - 0.2_EB*XX**2 ), EB ) / CMPLX( 1._EB - 0.5_EB*XX**2, TWOTHR*XX**3, EB )
!                                                      ** EQ. R/A.6
B( 1 ) = CMPLX( 0._EB, - ( 1._EB - 0.1_EB*XX**2 ) / 3._EB, EB ) / CMPLX( 1._EB + 0.5_EB*XX**2, - XX**3 / 3._EB, EB)
!                                                       ** EQ. R/A.7,8
A( 2 ) = CMPLX( 0._EB,   XX**2 / 30._EB, EB)
B( 2 ) = CMPLX( 0._EB, - XX**2 / 45._EB, EB)
!                                                       ** EQ. R/A.9
QSCA = 6._EB* XX**4 *( SQ( A(1) ) + SQ( B(1) ) + FIVTHR*( SQ( A(2) ) + SQ( B(2) ) ) )
QEXT = QSCA
!                                                       ** EQ. R/A.10
GQSC = 6._EB* XX**4 *REAL( A(1)*CONJG( A(2) + B(1) ) + ( B(1) + FIVNIN*A(2) )*CONJG( B(2) ) )

RTMP   = 1.5_EB * XX**3
SFORW  = RTMP*( A(1) + B(1) + FIVTHR*( A(2) + B(2) ) )
SBACK  = RTMP*( A(1) - B(1) - FIVTHR*( A(2) - B(2) ) )
TFORW( 1 ) = RTMP*( B(1) + FIVTHR*( 2._EB*B(2) - A(2) ) )
TFORW( 2 ) = RTMP*( A(1) + FIVTHR*( 2._EB*A(2) - B(2) ) )
TBACK( 1 ) = RTMP*( B(1) - FIVTHR*( 2._EB*B(2) + A(2) ) )
TBACK( 2 ) = RTMP*( A(1) - FIVTHR*( 2._EB*A(2) + B(2) ) )

DO J = 1, NUMANG
!                                                    ** EQ. R/A.11,12
   S1( J ) = RTMP*( A(1) + B(1)*XMU( J ) + FIVTHR*( A(2)*XMU( J ) +  B(2)*( 2._EB*XMU( J )**2 - 1._EB) ) )
   S2( J ) = RTMP*( B(1) + A(1)*XMU( J ) + FIVTHR*( B(2)*XMU( J ) +  A(2)*( 2._EB*XMU( J )**2 - 1._EB) ) )
END DO
!                                     ** RECOVER ACTUAL MIE COEFFICIENTS
A( 1 ) = XX**3 * A(1)
A( 2 ) = XX**3 * A(2)
B( 1 ) = XX**3 * B(1)
B( 2 ) = XX**3 * B(2)

RETURN
END SUBROUTINE SMALL1

SUBROUTINE SMALL2( XX, CIOR, CALCQE, NUMANG, XMU, QEXT, QSCA,GQSC, SFORW, SBACK, S1, S2, TFORW, TBACK,A, B )

!       SMALL-PARTICLE LIMIT OF MIE QUANTITIES FOR GENERAL REFRACTIVE
!       INDEX ( MIE SERIES TRUNCATED AFTER 2 TERMS )
!
!        A,B       FIRST TWO MIE COEFFICIENTS, WITH NUMERATOR AND
!                  DENOMINATOR EXPANDED IN POWERS OF XX ( A FACTOR
!                  OF XX**3 IS MISSING BUT IS RESTORED BEFORE RETURN
!                  TO CALLING PROGRAM )

!        CIORSQ    SQUARE OF REFRACTIVE INDEX

!     .. PARAMETERS ..
REAL(EB)    TWOTHR, FIVTHR
PARAMETER   ( TWOTHR = 2._EB/3._EB, FIVTHR = 5._EB/3._EB)

!     .. SCALAR ARGUMENTS ..
LOGICAL     CALCQE
INTEGER     NUMANG
REAL(EB)    GQSC, QEXT, QSCA, XX
COMPLEX(EB) CIOR, SBACK, SFORW

!     .. ARRAY ARGUMENTS ..
REAL(EB)    XMU( * )
COMPLEX(EB) A(*), B(*), S1(*), S2(*), TBACK(*), TFORW( * )

!     .. LOCAL SCALARS ..
INTEGER   J
REAL(EB)  RTMP
COMPLEX(EB)   CIORSQ, CTMP


CIORSQ = CIOR**2
CTMP   = CMPLX( 0._EB, TWOTHR, EB )*( CIORSQ - 1.0_EB )

!                                           ** EQ. R42A
A( 1 ) = CTMP*( 1._EB- 0.1_EB*XX**2 +   ( CIORSQ / 350._EB + 1._EB/280._EB)*XX**4 ) / &
          ( CIORSQ + 2._EB+ ( 1._EB- 0.7_EB*CIORSQ )*XX**2 -  &
          ( CIORSQ**2 / 175._EB- 0.275_EB*CIORSQ + 0.25_EB )*XX**4 + &
          XX**3 * CTMP * ( 1._EB- 0.1_EB*XX**2 ) )

!                                           ** EQ. R42B
B( 1 ) = ( XX**2 / 30._EB )*CTMP*( 1._EB+  ( CIORSQ / 35._EB - 1._EB/ 14._EB)*XX**2 ) /  &
         ( 1._EB- ( CIORSQ / 15._EB - 1._EB/6._EB)*XX**2 )

!                                           ** EQ. R42C

A( 2 ) = ( 0.1_EB*XX**2 )*CTMP*( 1._EB- XX**2 / 14._EB ) /  ( 2._EB*CIORSQ + 3._EB- &
         ( CIORSQ / 7._EB- 0.5_EB ) * XX**2 )

!                                           ** EQ. R40A

QSCA = (6._EB*XX**4) * ( SQ( A(1) ) + SQ( B(1) ) + FIVTHR * SQ( A(2) ) )

!                                           ** EQ. R40B
QEXT = QSCA
IF( CALCQE ) QEXT = 6._EB*XX * REAL( A(1) + B(1) + FIVTHR*A(2) )

!                                           ** EQ. R40C

GQSC = (6._EB*XX**4) * REAL( A(1)*CONJG( A(2) + B(1) ) )

RTMP   = 1.5_EB * XX**3
SFORW  = RTMP*( A(1) + B(1) + FIVTHR*A(2) )
SBACK  = RTMP*( A(1) - B(1) - FIVTHR*A(2) )
TFORW( 1 ) = RTMP*( B(1) - FIVTHR*A(2) )
TFORW( 2 ) = RTMP*( A(1) + 2._EB*FIVTHR*A(2) )
TBACK( 1 ) = TFORW(1)
TBACK( 2 ) = RTMP*( A(1) - 2._EB*FIVTHR*A(2) )


DO J = 1, NUMANG
!                                      ** EQ. R40D,E

   S1( J ) = RTMP*( A(1) + ( B(1) + FIVTHR*A(2) )*XMU( J ) )
   S2( J ) = RTMP*( B(1) + A(1)*XMU( J ) + FIVTHR*A(2)*( 2._EB*XMU( J )**2 - 1._EB) )
END DO

!                                     ** RECOVER ACTUAL MIE COEFFICIENTS
A( 1 ) = XX**3 * A(1)
A( 2 ) = XX**3 * A(2)
B( 1 ) = XX**3 * B(1)
B( 2 ) = ( 0._EB, 0._EB)

RETURN
END SUBROUTINE SMALL2

SUBROUTINE TESTMI( COMPAR, XX, CREFIN, MIMCUT, PERFCT, ANYANG, &
                   NMOM, IPOLZN, NUMANG, XMU, QEXT, QSCA, GQSC, &
                   SFORW, SBACK, S1, S2, TFORW, TBACK, PMOM, &
                   MOMDIM )

!         SET UP TO RUN TEST CASE WHEN  COMPAR = FALSE;  WHEN  = TRUE,
!         COMPARE MIE CODE TEST CASE RESULTS WITH CORRECT ANSWERS
!         AND ABORT IF EVEN ONE RESULT IS INACCURATE.

!         THE TEST CASE IS :  MIE SIZE PARAMETER = 10
!                             REFRACTIVE INDEX   = 1.5 - 0.1 I
!                             SCATTERING ANGLE = 140 DEGREES
!                             1 SEKERA MOMENT

!         RESULTS FOR THIS CASE MAY BE FOUND AMONG THE TEST CASES
!         AT THE END OF REFERENCE (1).

!         *** NOTE *** WHEN RUNNING ON SOME COMPUTERS, ESP. IN SINGLE
!         PRECISION, THE ACCUR CRITERION BELOW MAY HAVE TO BE RELAXED.
!         HOWEVER, IF ACCUR MUST BE SET LARGER THAN 10**-3 FOR SOME
!         SIZE PARAMETERS, YOUR COMPUTER IS PROBABLY NOT ACCURATE
!         ENOUGH TO DO MIE COMPUTATIONS FOR THOSE SIZE PARAMETERS.

!     ROUTINES CALLED :  ERRMSG, MIPRNT, TSTBAD

!     .. SCALAR ARGUMENTS ..

LOGICAL     ANYANG, COMPAR, PERFCT
INTEGER     IPOLZN, MOMDIM, NMOM, NUMANG
REAL(EB)    GQSC, MIMCUT, QEXT, QSCA, XX
COMPLEX(EB) CREFIN, SBACK, SFORW

!     .. ARRAY ARGUMENTS ..

REAL(EB)    PMOM( 0:MOMDIM, * ), XMU( * )
COMPLEX(EB) S1( * ), S2( * ), TBACK( * ), TFORW( * )

!     .. LOCAL SCALARS ..

LOGICAL     ANYSAV, OK, PERSAV
INTEGER     IPOSAV, M, N, NMOSAV, NUMSAV
REAL(EB)    MIMSAV, TESTGQ, TESTQE, TESTQS,XMUSAV, XXSAV
COMPLEX(EB) CRESAV, TESTS1, TESTS2, TESTSB, TESTSF

!     .. LOCAL ARRAYS ..

LOGICAL     CALCMO( 4 ), PRNT( 2 )
REAL(EB)    TESTPM( 0:1 )
COMPLEX(EB) TESTTB( 2 ), TESTTF( 2 )


SAVE      XXSAV, CRESAV, MIMSAV, PERSAV, ANYSAV, NMOSAV, IPOSAV,NUMSAV, XMUSAV

DATA      TESTQE / 2.459791_EB /,&
          TESTQS / 1.235144_EB /,&
          TESTGQ / 1.139235_EB /,&
          TESTSF / ( 61.49476_EB, -3.177994_EB ) /,&
          TESTSB / ( 1.493434_EB,  0.2963657_EB ) /,&
          TESTS1 / ( -0.1548380_EB, -1.128972_EB ) /,&
          TESTS2 / ( 0.05669755_EB, 0.5425681_EB ) /,&
          TESTTF / ( 12.95238_EB, -136.6436_EB ),&
                   ( 48.54238_EB, 133.4656_EB ) /,&
          TESTTB / ( 41.88414_EB, -15.57833_EB ),&
                   ( 43.37758_EB, -15.28196_EB ) /,&
          TESTPM / 227.1975_EB, 183.6898_EB /

IF( .NOT.COMPAR ) THEN
!                                   ** SAVE CERTAIN USER INPUT VALUES
   XXSAV  = XX
   CRESAV = CREFIN
   MIMSAV = MIMCUT
   PERSAV = PERFCT
   ANYSAV = ANYANG
   NMOSAV = NMOM
   IPOSAV = IPOLZN
   NUMSAV = NUMANG
   XMUSAV = XMU( 1 )
!                                   ** RESET INPUT VALUES FOR TEST CASE
   XX     = 10.0_EB
   CREFIN = ( 1.5_EB, -0.1_EB )
   MIMCUT = 0.0_EB
   PERFCT = .FALSE.
   ANYANG = .TRUE.
   NMOM   = 1
   IPOLZN = -1
   NUMANG = 1
   XMU( 1 ) = -0.7660444_EB

ELSE
!                                    ** COMPARE TEST CASE RESULTS WITH
!                                    ** CORRECT ANSWERS AND ABORT IF BAD
   OK = .TRUE.

   IF( WRONG( QEXT,TESTQE ) ) OK = TSTBAD( 'QEXT', ABS( ( QEXT - TESTQE ) / TESTQE ) )

   IF( WRONG( QSCA,TESTQS ) )OK = TSTBAD( 'QSCA', ABS( ( QSCA - TESTQS ) / TESTQS ) )

   IF( WRONG( GQSC,TESTGQ ) ) OK = TSTBAD( 'GQSC', ABS( ( GQSC - TESTGQ ) / TESTGQ ) )

   IF( WRONG( REAL( SFORW ),REAL( TESTSF ) ) .OR. WRONG( AIMAG( SFORW ),AIMAG( TESTSF ) ) ) &
         OK = TSTBAD( 'SFORW', ABS( ( SFORW - TESTSF ) / TESTSF ) )

   IF( WRONG( REAL( SBACK ),REAL( TESTSB ) ) .OR. WRONG( AIMAG( SBACK ),AIMAG( TESTSB ) ) ) &
         OK = TSTBAD( 'SBACK', ABS( ( SBACK - TESTSB ) / TESTSB ) )

   IF( WRONG( REAL( S1(1) ),REAL( TESTS1 ) ) .OR. WRONG( AIMAG( S1(1) ),AIMAG( TESTS1 ) ) ) &
         OK = TSTBAD( 'S1', ABS( ( S1(1) - TESTS1 ) / TESTS1 ) )

   IF( WRONG( REAL( S2(1) ),REAL( TESTS2 ) ) .OR. WRONG( AIMAG( S2(1) ),AIMAG( TESTS2 ) ) ) &
         OK = TSTBAD( 'S2', ABS( ( S2(1) - TESTS2 ) / TESTS2 ) )

   DO N = 1, 2
      IF( WRONG( REAL( TFORW(N) ),REAL( TESTTF(N) ) ) .OR. &
               WRONG( AIMAG( TFORW(N) ), &
               AIMAG( TESTTF(N) ) ) ) OK = TSTBAD( 'TFORW', &
               ABS( ( TFORW(N) - TESTTF(N) ) / TESTTF(N) ) )

      IF( WRONG( REAL( TBACK(N) ),REAL( TESTTB(N) ) ) .OR. &
               WRONG( AIMAG( TBACK(N) ), &
               AIMAG( TESTTB(N) ) ) ) OK = TSTBAD( 'TBACK', &
               ABS( ( TBACK(N) - TESTTB(N) ) / TESTTB(N) ) )
   END DO

   DO M = 0, 1
      IF ( WRONG( PMOM(M,1), TESTPM(M) ) ) OK =  TSTBAD( 'PMOM', ABS( (PMOM(M,1)-TESTPM(M)) / TESTPM(M) ) )
   END DO


   IF( .NOT.OK ) THEN
      PRNT( 1 ) = .TRUE.
      PRNT( 2 ) = .TRUE.
      CALCMO( 1 ) = .TRUE.
      CALCMO( 2 ) = .FALSE.
      CALCMO( 3 ) = .FALSE.
      CALCMO( 4 ) = .FALSE.
      CALL MIPRNT( PRNT, XX, PERFCT, CREFIN, NUMANG, XMU, QEXT, &
                        QSCA, GQSC, NMOM, IPOLZN, MOMDIM, CALCMO, PMOM, &
                        SFORW, SBACK, TFORW, TBACK, S1, S2 )
      CALL ERRMSG( 'MIEV0 -- SELF-TEST FAILED',.TRUE.)
   END IF
!                                       ** RESTORE USER INPUT VALUES
   XX     = XXSAV
   CREFIN = CRESAV
   MIMCUT = MIMSAV
   PERFCT = PERSAV
   ANYANG = ANYSAV
   NMOM   = NMOSAV
   IPOLZN = IPOSAV
   NUMANG = NUMSAV
   XMU( 1 ) = XMUSAV

END IF

CONTAINS

LOGICAL FUNCTION WRONG(CALC,EXACT)
REAL(EB) ACCUR
REAL(EB), INTENT(IN) :: CALC,EXACT
DATA  ACCUR / 1.E-4_EB /
WRONG = ABS( ( CALC - EXACT ) / EXACT )>ACCUR
END FUNCTION WRONG

END SUBROUTINE TESTMI


SUBROUTINE ERRMSG( MESSAG, FATAL )
! PRINT OUT A WARNING OR ERROR MESSAGE;  ABORT IF ERROR AFTER MAKING SYMBOLIC DUMP (MACHINE-SPECIFIC)
USE COMP_FUNCTIONS, ONLY: SHUTDOWN
CHARACTER :: MESSAG*(*)
LOGICAL ::  FATAL,MSGLIM
INTEGER ::  MAXMSG, NUMMSG
SAVE      MAXMSG, NUMMSG, MSGLIM
DATA      NUMMSG / 0 /,  MAXMSG / 100 /,  MSGLIM /.FALSE./
IF (FATAL) CALL SHUTDOWN(MESSAG)
NUMMSG = NUMMSG + 1
IF( MSGLIM ) RETURN
IF( NUMMSG<=MAXMSG ) THEN
   WRITE(LU_ERR, '(/,2A,/)' ) ' ****** WARNING *****  ', MESSAG
ELSE
   WRITE(LU_ERR, 9000 )
   MSGLIM = .TRUE.
END IF
RETURN
9000 FORMAT( / , / , ' ****** TOO MANY WARNING MESSAGES --  ','THEY WILL NO LONGER BE PRINTED *******', / , / )
END SUBROUTINE ERRMSG


LOGICAL FUNCTION WRTBAD( VARNAM )
!          WRITE NAMES OF ERRONEOUS VARIABLES AND RETURN 'TRUE'
!      INPUT :   VARNAM = NAME OF ERRONEOUS VARIABLE TO BE WRITTEN
!                         ( CHARACTER, ANY LENGTH )
CHARACTER VARNAM*(*)
INTEGER   MAXMSG, NUMMSG
SAVE      NUMMSG, MAXMSG
DATA      NUMMSG / 0 /, MAXMSG / 50 /
WRTBAD = .TRUE.
NUMMSG = NUMMSG + 1
WRITE(LU_ERR, '(3A)' ) ' ****  INPUT VARIABLE  ', VARNAM,'  IN ERROR  ****'
IF( NUMMSG==MAXMSG ) CALL ERRMSG( 'TOO MANY INPUT ERRORS.  ABORTING...',.TRUE.)
RETURN
END FUNCTION WRTBAD


LOGICAL FUNCTION WRTDIM( DIMNAM, MINVAL )

! WRITE NAME OF TOO-SMALL SYMBOLIC DIMENSION AND THE VALUE IT SHOULD BE INCREASED TO;  RETURN 'TRUE'

!      INPUT :  DIMNAM = NAME OF SYMBOLIC DIMENSION WHICH IS TOO SMALL
!                        ( CHARACTER, ANY LENGTH )
!               MINVAL = VALUE TO WHICH THAT DIMENSION SHOULD BE
!                        INCREASED (AT LEAST)
CHARACTER :: DIMNAM*(*)
INTEGER ::  MINVAL

WRITE(LU_ERR, '(3A,I7)' ) ' ****  SYMBOLIC DIMENSION  ', DIMNAM,'  SHOULD BE INCREASED TO AT LEAST ', MINVAL
WRTDIM = .TRUE.
RETURN
END FUNCTION WRTDIM


LOGICAL FUNCTION TSTBAD( VARNAM, RELERR )
! WRITE NAME (VARNAM) OF VARIABLE FAILING SELF-TEST AND ITS PERCENT ERROR FROM THE CORRECT VALUE; RETURN 'FALSE'
CHARACTER :: VARNAM*(*)
REAL(EB) ::  RELERR
TSTBAD = .FALSE.
WRITE(LU_ERR, '(/,3A,1P,E11.2,A)' ) ' *** OUTPUT VARIABLE ', VARNAM, ' DIFFERED BY ', &
     100.*RELERR,' PER CENT FROM CORRECT VALUE.  SELF-TEST FAILED.'
RETURN
END FUNCTION TSTBAD


REAL(EB) FUNCTION SQ(CTMP)
COMPLEX(EB), INTENT(IN) :: CTMP
SQ = REAL( CTMP )**2 + AIMAG( CTMP )**2
END FUNCTION SQ

END MODULE MIEV
